#include <ctype.h>
#include <assert.h>
#include <string>
#include <map>
#include <stdio.h>

using namespace std;

bool noisy;
bool print_all_types = 0 ;
bool print_user_vars = 0 ;

#define ABS(x)  (((x)<0) ? -(x) : (x))
static char *INULL = "-1" ;
#define PRINT_NOTHING INULL
#define PRINT_DEF_STRING NULL

struct dicom_type_info
{
	int group, elt, gmask, emask;
	string name, type;
	int mult; // mult here means, can there be multiple entries
};

struct dicom_data_info
{
	void *data;
	int len;
	dicom_type_info *type;
	char vr[3];
};

typedef map<pair<int,int>, dicom_type_info*> dicom_type_map;
typedef map<string,dicom_data_info> dicom_data_map;

int strtotype(char *s)
{
	switch(s[0]*256 + s[1])
	{
		case ('S'*256+'Q'):
			//assert(false);
			return 2;
		case ('O'*256+'B'):
		case ('O'*256+'W'):
		case ('O'*256+'F'):
		case ('U'*256+'N'):
		case ('U'*256+'T'):
			return 1;
		default:
			return 0;
	}
}

int tohex(char c)
{
	c = toupper(c);
	if( isdigit(c) ) return c - '0';
	if( c >= 'A' && c <= 'F' ) return 10 + (c-'A');
	assert(false);
}

void get_hex_val(char *s, int &val, int &mask)
{
	val = 0; mask = 0;
	for( int i = 0; i < 4; i++ )
	{
		if( tolower(s[i]) == 'x' ) { val = val << 4; mask = mask << 4; }
		else { val = (val << 4) + tohex(s[i]); mask = (mask << 4) + 15; }
	}
}

bool read_dicom_dict(char *fname, dicom_type_map &type_list)
{
	FILE *f;
	bool v = false;
	char buf[512];
	char sgroup[16], selt[16], stype[16], sname[256], smult[16];
	f = fopen(fname, "r");
	if( f == NULL ) return false;
	while(!feof(f))
	{
		fgets( buf, 511, f );
		int cnt = sscanf(buf,"(%4c,%4c)\t%s\t%s\t%s",
                                 sgroup, selt, stype, sname, smult);
		//printf("%d: buf = %s\n", cnt, buf);
		if( cnt != 5 )
			continue;
		v = true;
		sgroup[4] = selt[4] = 0;
		if( 0 && noisy )
			printf( "grp = %s, elt = %s, type = %s, name = %s, "
				"mult = %s\n", sgroup, selt, stype, sname,
				smult);
		dicom_type_info *ti = new dicom_type_info;
		get_hex_val(sgroup, ti->group, ti->gmask);
		get_hex_val(selt, ti->elt, ti->emask);
		ti->name = sname;
		if( strcmp(smult,"1") ) ti->mult = 1;
		else ti->mult = 0;
		ti->type = stype;
		if( 0 && noisy )
			printf( "grp = %04x, elt = %04x, type = %s, name = %s, "
				"mult = %d\n", ti->group, ti->elt,
				ti->type.c_str(), ti->name.c_str(), ti->mult);
		if( 0 && noisy )
			printf( "gmk = %04x, emk = %04x\n", ti->gmask,
				ti->emask);
		pair<int,int> p;
		p.first = ti->group; p.second = ti->elt;
		type_list[p] = ti;
	}

	fclose(f);
	return v;
}

bool read_record(FILE *f, dicom_type_map &type_list, dicom_data_map &data)
{
	int group, elt;

	unsigned char tag[4];
	char vr[3];

	if( 1 != fread(tag, 4, 1, f) )
	{
		if( noisy ) printf("couldn't read tag\n");
		return false;
	}
	group = (tag[1] << 8) + tag[0];
	elt = (tag[3] << 8) + tag[2];
	
	if( 1 != fread(vr, 2, 1, f) )
	{
		printf("couldn't read VR\n");
		return false;
	}
	vr[2] = 0;

	int len;

	char buf[6];
	switch(strtotype(vr))
	{
		case 1:
		// explicit VR with long length
		
		if( 1 != fread(buf, 6, 1, f) )
		{
			printf("couldn't read long VR len\n");
			return false;
		}
		len = (buf[5] << 24) + (buf[4] << 16) + (buf[3] << 8) + buf[2];
		break;

		case 0:
		// explicit VR with short length

		if( 1 != fread(buf, 2, 1, f) )
		{
			printf("couldn't read short VR len\n");
			return false;
		}
		len = (buf[1] << 8) + buf[0];
		break;

		case 2:
		// sequence with no meaningful length
		if( 1 != fread(buf, 6, 1, f) )
		{
			printf("couldn't read SQ VR len\n");
			return false;
		}
		len = 0;
		break;

	}

	char *dbuf;
	if( len ) dbuf = new char[len];
	else
	{
		if( noisy ) printf("%s has 0 len\n", vr);
		dbuf = NULL;
	}
	if( len && (1 != fread(dbuf, len, 1, f)) )
	{
		printf("couldn't read data packet len = %d\n",
				   len);
		return false;
	}

	pair<int,int> p;
	p.first = group; p.second = elt;
	dicom_type_map::iterator it = type_list.find(p);
	if( it != type_list.end() )
	{
		if( print_all_types )
			printf("found type: %s: \n", it->second->name.c_str());
		if( it->second->name == "TransferSyntaxUID" )
		{
			if( strncmp(dbuf, "1.2.840.10008.1.2.1", len) )
			{
				printf("Unsupported Transfer syntax\n");
				return false;
			} else
			{
				if( noisy )
					printf("Transfer syntax supported\n");
			}
		}
		dicom_data_map::iterator dit =
			data.find(it->second->name.c_str());

		// multiple of the same one
		if( dit != data.end() )
		{
			if( noisy ) printf("doubled up on %s\n",
					   it->second->name.c_str());
			delete [] (char*)dit->second.data;
		}
		data[it->second->name.c_str()].data = dbuf;
		data[it->second->name.c_str()].len = len;
		data[it->second->name.c_str()].type = it->second;
		strcpy(data[it->second->name.c_str()].vr, vr);
		
	} else
	{
			printf("couldn't find type: %04x, %04x\n",
				group, elt);
		delete [] dbuf;
	}
	if( 0 && noisy )
	{
		printf( "grp = %04x, elt = %04x, type = %s, "
			"len = %d data=", group, elt, vr, len);
		if( vr[0]*256 + vr[1] == ('U'*256+'L') )
			printf("%d", *((int*)dbuf));
		else
			for( int i = 0; i < len; i++ )
				if( isprint(dbuf[i]) )
				printf("%c", dbuf[i]);
		printf("\n");
	}

	return true;
}

bool read_dicom_file( char *fname, dicom_type_map &type_list,
			dicom_data_map &data )
{
	FILE *f;
	bool v = false;
	char buf[512];
	char sgroup[16], selt[16], stype[16], sname[256], smult[16];
	f = fopen(fname, "r");
	if( f == NULL ) return false;

	char preamble[129];
	char magic[5];
	if( 1 != fread(preamble, 128, 1, f) )
	{
		if( noisy ) printf("couldn't read preamble\n");
		fclose(f);
		return false;
	}

	if( 1 != fread(magic, 4, 1, f) )
	{
		fclose(f);
		return false;
	}

	magic[4] = 0;
	if( strcmp(magic,"DICM") )
	{
		if( noisy ) printf("bad magic number\n");
		fclose(f);
		return false;
	}

	if( noisy )
	{
		printf("preamble = ");
		for( int i = 0; i < 128; i++ )
			printf("%02x", preamble[i]);
		printf("\n");
	}

	v = true;

	int nr = 0;
	while(v && !feof(f))
	{
		bool r = read_record(f,type_list,data);
		if( feof(f) ) break;
		nr++;
		v = v && r;
	}

	if( noisy ) printf("nr = %d\n", nr);
	fclose(f);

	return v;
}


void instructions(void){
  printf("Usage: dicom dictionary file\n"
	 "Options:\n"
	 "-oname Output file name root\n"
	 "-pall  Print all available dicom variables\n"
	 "-puser Print all GE user variables\n"
	 "-debug Lots of debugging output\n"
	 );
}

void send_help()
{
  printf("No help yet for dicom! .. sorry \n"
	 "Here are the instructions:\n");
  instructions() ;
}

int parse_args(int argc,char *argv[]) 
{
    int   nopt = 0;
    int   need_help = 0;
    int   required_opt = 2;

    for (int i = 1; i < argc; i++) { /* ------- Options ------- */

      if (!strncmp(argv[i], "-debug", 6)) { /* print out arrays */
	if (i >= argc) { instructions(); exit(0); }
	noisy = 1;
	nopt++;
      }
      if (!strncmp(argv[i], "-pall", 5)) { /* print out types */
	if (i >= argc) { instructions(); exit(0); }
	print_all_types = 1;
	nopt++;
      }
      if (!strncmp(argv[i], "-puser", 5)) { /* print out user vars */
	if (i >= argc) { instructions(); exit(0); }
	print_user_vars = 1 ;
	nopt++;
      }
      if (!strncmp(argv[i], "-help", 5)) { /* help */
	if (i >= argc) { instructions(); exit(0); }
	need_help = 1;
	nopt++;
      }
      
    } /* end of options */

    if (need_help){
      send_help();
      exit(0);
    }
	
    if (argc < 1 + required_opt + nopt){
      instructions();
      exit(0);
    }

    return nopt ;
}

float get_dicom_val( dicom_data_map dmap, char *valstr, char *print_str )
{
  int debug = 0;
  char stmp[256] ;

  sprintf(stmp,"%s",dmap[valstr].data);

  float val = atof(stmp);
  if(print_str==PRINT_NOTHING) {

  } else if(print_str==PRINT_DEF_STRING) {
    printf("%s = %f\n", valstr,val);
  } else {
    printf("%s = %f\n", print_str,val);
  }

  if(debug) printf("get_dicom_val: val = %f\n",val);

  return(val) ;
}

char *get_dicom_str( dicom_data_map dmap, char *valstr, char *print_str )
{
  int debug = 0;
  char *strval = (char *)calloc(256,sizeof(char));

  sprintf(strval,"%s",dmap[valstr].data);

  if(print_str==PRINT_NOTHING) {

  } else if(print_str==PRINT_DEF_STRING) {
    printf("%s = %s\n", valstr,strval);
  } else {
    printf("%s = %s\n", print_str,strval);
  }

  if(debug) printf("get_dicom_str: str = %s\n",strval);

  return(strval) ;
}

int main(int argc, char **argv)
{
	char user_var[256];
	int  nopt = 0;
	int nuser = 25 ;	/* why isn't this bigger?  There are more user vars! */
	noisy = false;
	dicom_type_map tl;

	nopt = parse_args(argc,argv) ;    /* Parse arguments */

	char dict_name[256] ;
	strcpy(dict_name,argv[nopt+1]);
	char file_name[256] ;
	strcpy(file_name,argv[nopt+2]);

	/* Read dictionary */ 
	if( read_dicom_dict(dict_name, tl) ) {
	  printf("Successfully read dictionary '%s'\n",dict_name);
	} else {
	  printf("Failed to read dictionary '%s'!\n\a",dict_name);
	  exit(0);
	}

	/* Read file */ 
	dicom_data_map d;
	if( read_dicom_file(file_name, tl, d) ) {
	  printf("Successfully read file '%s'\n",file_name);
	} else {
	  printf("Failed to read file '%s'!\n\a",file_name);
	  exit(0);
	}

	printf("\nInput parameters: \n") ;
	char *psdname      = get_dicom_str(d,"PulseSeqName",PRINT_DEF_STRING);
	float slcpos_first = get_dicom_val(d,"FirstScanLocation",PRINT_NOTHING);
	float slcpos_last  = get_dicom_val(d,"LastScanLocation",PRINT_NOTHING);
	float slcgap       = get_dicom_val(d,"SpacingBetweenSlices",PRINT_NOTHING);
	float slcthk       = get_dicom_val(d,"SliceThickness",PRINT_NOTHING);
	int   numave       = (int) get_dicom_val(d,"NumberOfAverages",PRINT_DEF_STRING);
	int   numbv        = (int) get_dicom_val(d,"UserData8","NumberOfBvals");
	int   numb0        = (int) get_dicom_val(d,"UserData23","NumberOfB0imgs");
	int   nvert        = (int) get_dicom_val(d,"UserData24","NumberOfDirections");

	int nslices ;
	if(slcgap==0.0) {
	  nslices = (int) (ABS(slcpos_last - slcpos_first)/slcthk + 1.0);
	} else {
	  nslices = (int) (ABS(slcpos_last - slcpos_first)/slcgap + 1.0);
	}
	printf("nslices = %i\n",nslices);

	if(print_user_vars) {
	  for(int k=0; k<nuser; k++) {
	    sprintf(user_var,"UserData%i",k) ;
	    printf("%s = %s\n", user_var,d[user_var].data);
	  }
	}

	//((float*)d["PixelData"].data)[i*j+k]

	for( dicom_type_map::iterator it = tl.begin(); it != tl.end(); it++ )
		delete it->second;
	for( dicom_data_map::iterator it2 = d.begin(); it2 != d.end(); it2++ )
		delete [] (char*)it2->second.data;
}
