#include <ctype.h>
#include <assert.h>
#include <string>
#include <map>
#include <stdio.h>
#include "ioutils.h"
#include "afni_note.h"
#include "mrilib.h"
#include "3ddata.h"

using namespace std;

bool noisy;
bool cleanup;
bool fixscale;
bool print_all_types = 0;
bool old_style;
//int nx_last,ny_last ;		// to catch dicom errors in get_dicom_img

#define TMP_EXIT {fprintf(stderr,"Temporary exit from program %s:\n",program_name);exit(0);}
#define TMP_BREAK break

#define ABS(x)  (((x)<0) ? -(x) : (x))
#define MIN(a,b) (((a)>(b)) ? (b) : (a))
#define MAX(a,b) (((a)<(b)) ? (b) : (a))
static char *INULL = "-1" ;
#define PRINT_NOTHING INULL
#define PRINT_DEF_STRING NULL
#define MAXLEN 150

int ALLOW_OVERWRITE = 1;	/* Allow files to be overwritten (hw for now) */
int DO_UNWARPING = 0;	/* Field map correction */

struct dicom_type_info
{
	int group, elt, gmask, emask;
	string name, type;
	int mult; // mult here means, can there be multiple entries
};

struct dicom_data_info
{
	void *data;
	int len;
	dicom_type_info *type;
	char vr[3];
};

typedef map<pair<int,int>, dicom_type_info*> dicom_type_map;
typedef map<string,dicom_data_info> dicom_data_map;

int strtotype(char *s)
{
	switch(s[0]*256 + s[1])
	{
		case ('S'*256+'Q'):
			//assert(false);
			return 2;
		case ('O'*256+'B'):
		case ('O'*256+'W'):
		case ('O'*256+'F'):
		case ('U'*256+'N'):
			return 1;
		default:
			return 0;
	}
}

int tohex(char c)
{
	c = toupper(c);
	if( isdigit(c) ) return c - '0';
	if( c >= 'A' && c <= 'F' ) return 10 + (c-'A');
	assert(false);
}

void get_hex_val(char *s, int &val, int &mask)
{
	val = 0; mask = 0;
	for( int i = 0; i < 4; i++ )
	{
		if( tolower(s[i]) == 'x' ) { val = val << 4; mask = mask << 4; }
		else { val = (val << 4) + tohex(s[i]); mask = (mask << 4) + 15; }
	}
}

bool read_dicom_dict(char *fname, dicom_type_map &type_list)
{
	FILE *f;
	bool v = false;
	char buf[512];
	char sgroup[16], selt[16], stype[16], sname[256], smult[16];
	f = fopen(fname, "r");
	if( f == NULL ) return false;
	while(!feof(f))
	{
		fgets( buf, 511, f );
		int cnt = sscanf(buf,"(%4c,%4c)\t%s\t%s\t%s",
                                 sgroup, selt, stype, sname, smult);
		//printf("%d: buf = %s\n", cnt, buf);
		if( cnt != 5 )
			continue;
		v = true;
		sgroup[4] = selt[4] = 0;
		if( 0 && noisy )
			printf( "grp = %s, elt = %s, type = %s, name = %s, "
				"mult = %s\n", sgroup, selt, stype, sname,
				smult);
		dicom_type_info *ti = new dicom_type_info;
		get_hex_val(sgroup, ti->group, ti->gmask);
		get_hex_val(selt, ti->elt, ti->emask);
		ti->name = sname;
		if( strcmp(smult,"1") ) ti->mult = 1;
		else ti->mult = 0;
		ti->type = stype;
		if( 0 && noisy )
			printf( "grp = %04x, elt = %04x, type = %s, name = %s, "
				"mult = %d\n", ti->group, ti->elt,
				ti->type.c_str(), ti->name.c_str(), ti->mult);
		if( 0 && noisy )
			printf( "gmk = %04x, emk = %04x\n", ti->gmask,
				ti->emask);
		pair<int,int> p;
		p.first = ti->group; p.second = ti->elt;
		type_list[p] = ti;
	}

	fclose(f);
	return v;
}

bool read_record(FILE *f, dicom_type_map &type_list, dicom_data_map &data)
{
	int group, elt;

	unsigned char tag[4];
	char vr[3];

	if( 1 != fread(tag, 4, 1, f) )
	{
		if( noisy ) printf("couldn't read tag\n");
		return false;
	}
	group = (tag[1] << 8) + tag[0];
	elt = (tag[3] << 8) + tag[2];
	
	if( 1 != fread(vr, 2, 1, f) )
	{
		printf("couldn't read VR\n");
		return false;
	}
	vr[2] = 0;

	int len;

	char buf[6];
	switch(strtotype(vr))
	{
		case 1:
		// explicit VR with long length
		
		if( 1 != fread(buf, 6, 1, f) )
		{
			printf("couldn't read long VR len\n");
			return false;
		}
		len = (buf[5] << 24) + (buf[4] << 16) + (buf[3] << 8) + buf[2];
		break;

		case 0:
		// explicit VR with short length

		if( 1 != fread(buf, 2, 1, f) )
		{
			printf("couldn't read short VR len\n");
			return false;
		}
		len = (buf[1] << 8) + buf[0];
		break;

		case 2:
		// sequence with no meaningful length
		if( 1 != fread(buf, 6, 1, f) )
		{
			printf("couldn't read SQ VR len\n");
			return false;
		}
		len = 0;
		break;

	}

	char *dbuf;
	if( len > 0) {		/* lrf - 03.21.05 */ 
	  dbuf = new char[len];
	} else {
	  if( noisy ) fprintf(stderr,"%s has len %i\n", vr,len);
	  len = 0 ;
	  dbuf = NULL;
	}

	if( len && (1 != fread(dbuf, len, 1, f)) )
	{
	  fprintf(stderr,"couldn't read data packet len = %d\n",len);
	  return false;
	}

	pair<int,int> p;
	p.first = group; p.second = elt;
	dicom_type_map::iterator it = type_list.find(p);
	if( it != type_list.end() )
	{
		if( print_all_types )
			printf("found type: %s: \n", it->second->name.c_str());
		if( it->second->name == "TransferSyntaxUID" )
		{
			if( strncmp(dbuf, "1.2.840.10008.1.2.1", len) )
			{
				printf("Unsupported Transfer syntax\n");
				return false;
			} else
			{
				if( noisy )
					printf("Transfer syntax supported\n");
			}
		}
		dicom_data_map::iterator dit =
			data.find(it->second->name.c_str());

		// multiple of the same one
		if( dit != data.end() )
		{
			if( noisy ) printf("doubled up on %s\n",
					   it->second->name.c_str());
			delete [] (char*)dit->second.data;
		}
		data[it->second->name.c_str()].data = dbuf;
		data[it->second->name.c_str()].len  = len;
		data[it->second->name.c_str()].type = it->second;
		strcpy(data[it->second->name.c_str()].vr, vr);
		
	} else
	{
			printf("couldn't find type: %04x, %04x\n",
				group, elt);
		delete [] dbuf;
	}

	if( 0 && noisy )
	{
		printf( "grp = %04x, elt = %04x, type = %s, "
			"len = %d data=", group, elt, vr, len);
		if( vr[0]*256 + vr[1] == ('U'*256+'L') )
			printf("%d", *((int*)dbuf));
		else
			for( int i = 0; i < len; i++ )
				if( isprint(dbuf[i]) )
				printf("%c", dbuf[i]);
		printf("\n");
	}

	return true;
}

bool read_dicom_file( char *fname, dicom_type_map &type_list,
			dicom_data_map &data )
{
	FILE *f;
	bool v = false;
	char buf[512];
	char sgroup[16], selt[16], stype[16], sname[256], smult[16];

	f = fopen(fname, "r");
	if( f == NULL ) return false;

	char preamble[129];
	char magic[5];

	if( 1 != fread(preamble, 128, 1, f) )
	{
		if( noisy ) printf("couldn't read preamble\n");
		fclose(f);
		return false;
	}

	if( 1 != fread(magic, 4, 1, f) )
	{
		fclose(f);
		return false;
	}

	magic[4] = 0;
	if( strcmp(magic,"DICM") )
	{
		if( noisy ) printf("bad magic number\n");
		fclose(f);
		return false;
	}

	if( noisy )
	{
		printf("preamble = ");
		for( int i = 0; i < 128; i++ )
			printf("%02x", preamble[i]);
		printf("\n");
	}

	v = true;

	int nr = 0;
	while(v && !feof(f))
	{
		bool r = read_record(f,type_list,data);
		if( feof(f) ) break;
		nr++;
		v = v && r;
	}

	if( noisy ) printf("nr = %d\n", nr);
	fclose(f);

	return v;
}

float get_dicom_val( dicom_data_map dmap, const char *valstr, char *print_str )
{
  char *program_name = "get_dicom_val";
  char stmp[256] ;
  char val_type[256] ;
  int debug = 0;
  string tmp_str(valstr);

  sprintf(val_type,"%s",dmap[tmp_str].vr);
  sprintf(stmp,"%s",dmap[tmp_str].data);
  if(debug) printf("%s\n",stmp);

  float val = atof(stmp);
  if(print_str==PRINT_NOTHING) {

  } else if(print_str==PRINT_DEF_STRING) {
    printf("%s = %f\n", valstr,val);
  } else {
    printf("%s = %f\n", print_str,val);
  }

  if(debug) printf("%s: type = %s, val = %f\n",program_name,val_type,val);

  return(val) ;
}

int get_dicom_ival( dicom_data_map dmap, char *valstr, char *print_str )
{
  char *program_name = "get_dicom_val";
  char stmp[256] ;
  char val_type[256] ;
  int debug = 0;

  sprintf(val_type,"%s",dmap[valstr].vr);
  sprintf(stmp,"%s",dmap[valstr].data);

  int *ival_ptr = (int *)dmap[valstr].data;
  int  ival = *ival_ptr;

  if(print_str==PRINT_NOTHING) {

  } else if(print_str==PRINT_DEF_STRING) {
    printf("%s = %i\n", valstr,ival);
  } else {
    printf("%s = %i\n", print_str,ival);
  }

  if(debug) printf("%s: type = %s, ival = %i\n",program_name,val_type,ival);

  return(ival) ;
}

char *get_dicom_str( dicom_data_map dmap, char *valstr, char *print_str )
{
  char *strval = (char *)calloc(256,sizeof(char));
  char val_type[256] ;
  int debug = 0;

  sprintf(val_type,"%s",dmap[valstr].vr);

  sprintf(strval,"%s",dmap[valstr].data);

  if(print_str==PRINT_NOTHING) {

  } else if(print_str==PRINT_DEF_STRING) {
    printf("%s = %s\n", valstr,strval);
  } else {
    printf("%s = %s\n", print_str,strval);
  }

  if(debug) printf("get_dicom_str: val_type = %s, str = %s\n",val_type,strval);

  return(strval) ;
}

int get_dicom_img( short int *data, dicom_data_map dmap, char *valstr, char *print_str )
{
  char *program_name = "get_dicom_img";
  char stmp[256] ;
  short int img_min =  32767 ;
  short int img_max = -32768 ;
  int debug = 0;

  int *nx = (int *)dmap["Rows"].data;
  int *ny = (int *)dmap["Columns"].data;
  if(debug) {
    printf("Rows type = %s\n",dmap["Rows"].vr);
    printf("nx = %i\n",*nx);
    printf("ny = %i\n",*ny);
  }

  /* Temporary AHB !!! */
  if(*nx!=256) {
    fprintf(stderr,"\n\n\t*************************************\n");
    fprintf(stderr,"\t%s: nx = %i not equal 256\a\n",program_name,*nx);
    fprintf(stderr,"\t*************************************\n");
    *nx = 256;
  }
  if(*ny!=256) {
    fprintf(stderr,"\t*************************************\n");
    fprintf(stderr,"\t%s: ny = %i not equal 256\a\n",program_name,*ny);
    fprintf(stderr,"\t*************************************\n");
    *ny = 256;
  }
  /* end temporary AHB !!! */

  short int *val ;
  int index ;
  for(int ix=0; ix<*nx; ix++) {
    for(int iy=0; iy<*ny; iy++) {

      index = ix*(*ny)+iy ;

      *val = ((short *)dmap[valstr].data)[index];
      if(*val<img_min) img_min = *val ;
      if(*val>img_max) img_max = *val ;

      if(print_str==PRINT_NOTHING) {

      } else if(print_str==PRINT_DEF_STRING) {
	printf("%s(%i,%i) = %i\n", valstr,ix,iy,*val);
      } else {
	printf("%s(%i,%i) = %i\n", print_str,ix,iy,*val);
      }

      data[index] = *val ;
    }
  }

  printf("\t min = %i, max = %i\n",img_min,img_max);

  return(1);
}

void instructions(void);
void send_help() ;
int parse_args(int argc,char *argv[]) ;
char *out_root;
char *dict_name_input = NULL;
char *fieldmap1 = NULL;
char *fieldmap2 = NULL;
int minarg = 1;			/* Minimum number of input args required */ 

int main(int argc, char **argv)
{
        char *program_name = "main";
	char user_var[256];
	char dict_name[256] ;
	char header_text[256];

	int  nopt = 0;
	int nuser = 25 ; /* why isn't this bigger?  There are more user vars! */
	noisy = false;
	cleanup = true;
	fixscale = false;
	old_style = false;
	dicom_type_map tl;

	int debug = 0;

	char *default_tensor_path     = getenv("DEFAULT_DICT_PATH");
	char *default_dicom_dict_path = getenv("DEFAULT_DICT_PATH");
	char *default_dicom_dict_name = "gems-dicom-dict_lrf.txt";

	out_root = "x";		          /* default outfile root */
	nopt = parse_args(argc,argv) ;    /* Parse arguments */

	char file_name[256] ;
	strcpy(file_name,argv[nopt+1]);

	/* Read dictionary */ 
	
	if(dict_name_input==NULL) {
	  sprintf(dict_name,"%s/%s",default_dicom_dict_path,default_dicom_dict_name);
	  printf("No dicom dictionary specified ...\n"
		 "Reading default file %s\n",dict_name) ; fflush(stdout);
	} else if(!file_exists(dict_name_input)){
	  printf("Dictionary file %s not found!  Exiting ...\n",dict_name_input) ; 
	  fflush(stdout);
	  exit(0);
	} else {
	  strcpy(dict_name,dict_name_input) ;
	}

	if( read_dicom_dict(dict_name, tl) ) {
	  printf("Successfully read dictionary '%s'\n",dict_name);
	} else {
	  exit(0);
	}

	/* Read file */ 

	dicom_data_map d;
	fprintf(stderr,"Reading file '%s' ... ",file_name);
	if( read_dicom_file(file_name, tl, d) ) {
	  fprintf(stderr,"\nSuccessfully read file '%s'\n",file_name);
	} else {
	  fprintf(stderr,"\nFailed to read file '%s'!\n\a",file_name);
	  exit(0);
	}

	printf("***** Input parameters *****\n") ;
	char *psdname      = get_dicom_str(d,"PulseSeqName",PRINT_DEF_STRING);
	float slcpos_first = get_dicom_val(d,"FirstScanLocation",PRINT_DEF_STRING);
	float slcpos_last  = get_dicom_val(d,"LastScanLocation",PRINT_DEF_STRING);
	float slcgap       = get_dicom_val(d,"SpacingBetweenSlices",PRINT_DEF_STRING);
	float slcthk       = get_dicom_val(d,"SliceThickness",PRINT_DEF_STRING);
	/*	float fovorig      = get_dicom_val(d,"FieldofViewOrigin",PRINT_DEF_STRING);
		float fovdisp      = get_dicom_val(d,"DisplayFieldofView",PRINT_DEF_STRING); */
	float optr         = get_dicom_val(d,"RepetitionTime",PRINT_DEF_STRING);

	int   imatrix_X    = get_dicom_ival(d,"Rows",PRINT_DEF_STRING);
	int   imatrix_Y    = get_dicom_ival(d,"Columns",PRINT_DEF_STRING);

	int   numave       = (int) get_dicom_val(d,"NumberOfAverages",PRINT_DEF_STRING);
	int   dsamp        = (int) get_dicom_val(d,"UserData5","DiffSampling");
	int   ntess        = (int) get_dicom_val(d,"UserData7","TesselationLevel");
	int   bvalue       = (int) get_dicom_val(d,"UserData1","BValue");
	int   numbv        = (int) get_dicom_val(d,"UserData8","NumberOfBvals");
	int   bstep        = (int) get_dicom_val(d,"UserData9","BStepMode");
	int   polytope     = (int) get_dicom_val(d,"UserData11","TessPolytope");
	int   numt2        = (int) get_dicom_val(d,"UserData2","NumberOfT2imgs");
	int   numb0        = (int) get_dicom_val(d,"UserData23","NumberOfB0imgs");
	int   nvert        = (int) get_dicom_val(d,"UserData24","NumberOfDirections");

	if(!strcmp(psdname,"epi2")) {
	  numbv = 1; // this parameter doesn't exist in epi2
	  polytope = 2; // this parameter doesn't exist in epi2
	  dsamp = 3; /* this parameter doesn't exist in epi2
			SHOULD REALLY BE ANOTHER DIFF_SAMP_GE or something */
	  //float ge_bval = get_dicom_val(d,"FuncBValue","FuncBValue");
	  bvalue = 3000;
	}

	float bval = (float)bvalue;
	// I don't think bvalue works! We're going to use UserData13 instead.
	int   ndif = nvert + numb0 ;

	/* Determine number of slices */
	float slcsiz ;
	if(slcgap==0.0) slcsiz = slcthk;
	else            slcsiz = slcgap;
	if(debug) printf("slcthk = %f, slcgap = %f, slcsiz = %f\n",slcthk,slcgap,slcsiz);

	/* Image dimensions */ 
	int nx = imatrix_X, ny = imatrix_Y;

	float zz  ;
	{ 
	  float xx,yy;
	  xx = round(100.*slcpos_last)/100.;
	  yy = round(100.*slcpos_first)/100.;
	  zz = ABS((xx-yy)/slcsiz)+1.0;
	}
	int nslices = (int) zz;
	printf("nslices = %i\n",nslices);

	/* Print out user info */

	int print_user_vars = 0 ;
	if(print_user_vars) {
	  for(int k=0; k<nuser; k++) {
	    sprintf(user_var,"UserData%i",k) ;
	    printf("%s = %s\n", user_var,d[user_var].data);
	  }
	}

	/**************** Write out header file using previous input *************/

	if(old_style) {
	  char header_file[80] ;

	  FILE *hfp ;

	  sprintf(header_file,"%s.header",out_root);
	  if ((hfp = fopen(header_file,"w")) == NULL) {
	    printf("Problem opening header file %s\n",header_file);
	    exit(0);
	  } else {
	    printf("Writing header file %s\n",header_file);fflush(stdout);
	  }

	  /* Write some descriptive text */ 

	  sprintf(header_text,"#");
	  fprintf(hfp,"%s\n",header_text); 
	  sprintf(header_text,"# Header file for EXCITE data");
	  fprintf(hfp,"%s\n",header_text); 
	  sprintf(header_text,"#");
	  fprintf(hfp,"%s\n",header_text); 

	  int print_file_params = 1;
	  if(print_file_params) {

	    printf("psdname %s\n",psdname); 
	    printf("nx %i\n",nx);
	    printf("ny %i\n",ny);
	    printf("nsl %i\n",nslices); 
	    printf("bval %f\n",bval);
	    printf("nbvals %i\n",numbv);
	    printf("bstep %i\n",bstep);
	    printf("nb0 %i\n",numb0);
	    printf("dsamp %i\n",dsamp);
	    printf("polytope %i\n",polytope); 
	    printf("ntess %i\n",ntess);
	    printf("nvert %i\n",nvert);

	  }

	  /* Write params to header file */ 

	  fprintf(hfp,"psdname %s\n",psdname); 
	  fprintf(hfp,"nx %i\n",nx);
	  fprintf(hfp,"ny %i\n",ny);
	  fprintf(hfp,"nsl %i\n",nslices); 
	  fprintf(hfp,"bval %f\n",bval);
	  fprintf(hfp,"nbvals %i\n",numbv);
	  fprintf(hfp,"bstep %i\n",bstep);
	  fprintf(hfp,"nb0 %i\n",numb0);
	  fprintf(hfp,"dsamp %i\n",dsamp);
	  fprintf(hfp,"polytope %i\n",polytope); 
	  fprintf(hfp,"ntess %i\n",ntess);
	  fprintf(hfp,"nvert %i\n",nvert);

	  fclose(hfp);
	}
	else {
	  int print_file_params = 1;
	  if(print_file_params) {
	    printf("psdname %s\n",psdname); 
	    printf("nx %i\n",nx);
	    printf("ny %i\n",ny);
	    printf("nsl %i\n",nslices); 
	    printf("bval %f\n",bval);
	    printf("nbvals %i\n",numbv);
	    printf("bstep %i\n",bstep);
	    printf("nb0 %i\n",numb0);
	    printf("dsamp %i\n",dsamp);
	    printf("polytope %i\n",polytope); 
	    printf("ntess %i\n",ntess);
	    printf("nvert %i\n",nvert);
	  }
	  /* prepare to write an AFNI note */
	  sprintf(header_text, "# afni-diff-plugin-extra-data\npsdname %s\nnx %i\nny %i\nnsl %i\nbval %f\nnbvals %i\nbstep %i\nnb0 %i\ndsamp %i\npolytope %i\nntess %i\nnvert %i\n",
		  psdname, nx, ny, nslices, bval, numbv, bstep, numb0,
		  dsamp, polytope, ntess, nvert);
// 	  sprintf(header_text, "nx %i\n",nx);
// 	  sprintf(header_text, "ny %i\n",ny);
// 	  sprintf(header_text, "nsl %i\n",nslices); 
// 	  sprintf(header_text, "bval %f\n",bval);
// 	  sprintf(header_text, "nbvals %i\n",numbv);
// 	  sprintf(header_text, "bstep %i\n",bstep);
// 	  sprintf(header_text, "nb0 %i\n",numb0);
// 	  sprintf(header_text, "dsamp %i\n",dsamp);
// 	  sprintf(header_text, "polytope %i\n",polytope); 
// 	  sprintf(header_text, "ntess %i\n",ntess);
// 	  sprintf(header_text, "nvert %i\n",nvert);
	}

	/**************** Read in tensor_lrf.data and write out gradient files *************/

	float *gdx = (float *)calloc(nvert,sizeof(float));
	float *gdy = (float *)calloc(nvert,sizeof(float));
	float *gdz = (float *)calloc(nvert,sizeof(float));
	char gdxname[80],gdyname[80],gdzname[80];
	char gradient_file[256] ;
	if(!strcmp(psdname,"epi3") || !strcmp(psdname,"hepi")) { 
	  strcpy(gradient_file,"tensor_lrf.dat");
	} else if(!strcmp(psdname,"epi2")) { 
	  sprintf(gradient_file,"%s/tensor.dat",default_tensor_path);
	}
	printf("Reading tensor file %s\n",gradient_file); fflush(stdout);

	FILE *gfp, *fd ;

	if ((gfp = fopen(gradient_file,"r")) == NULL) {
	  printf("Problem opening gradient file %s\n",gradient_file);
	  exit(0);
	}
	
	char gfile_line[MAXLEN];
	int gfile_nvert = -1;
	int start_read = 0;
	float gxval,gyval,gzval;
	int gcount = 0;

	char tempstring[MAXLEN];           /* buffer to access file */
	if(!strcmp(psdname,"epi2")) { 
 
            int read_skip = 1;
	    char compstring[MAXLEN];           /* buffer used to open data file */ 
	    long floc ;
	    sprintf( compstring, "%d", nvert );
	    int num_tensor_len = strlen( compstring );

            while( read_skip ) {
	      floc = ftell(gfp) ;
	      fgets( tempstring, MAXLEN, gfp );
	      read_skip = strncmp( compstring, tempstring, num_tensor_len );
            }
	    fseek(gfp,floc,SEEK_SET);
	}

	while (nvert>gcount && fgets(gfile_line, MAXLEN, gfp) != NULL) {

	  if(debug) printf("gfile_line = %s",gfile_line);
	  if(!strncmp(gfile_line,"#",1)) {
	    if(debug) printf("Comment line\n");
	  } else if (!start_read) {
	    sscanf(gfile_line,"%d",&gfile_nvert);
	    if(debug) printf("gfile_nvert = %i\n",gfile_nvert);
	  }

	  if(start_read) {
	    sscanf(gfile_line,"%f %f %f",&gdx[gcount],&gdy[gcount],&gdz[gcount]);
	    if(debug) printf("gd = (%f,%f,%f)\n",gdx[gcount],gdy[gcount],gdz[gcount]);
	    gcount++;
	  }
	  if(gfile_nvert==nvert) start_read = 1;

	}
	if(gcount!=nvert){
	  printf("Number of gradient directions read (%i) not equal to nvert from header (%i)!!\a\n",
		 gcount,nvert);
	  printf("Exiting ...\n");
	  exit(0);
	}

	if (old_style) {

	  /* X grad */

	  sprintf(gdxname,"%s.GDX",out_root);
	  printf("Writing xdiff gradient file to %s\n",gdxname);
	  if ((fd = fopen(gdxname,"w")) == NULL) {
	    printf("Problem opening %s\n",gdxname);
	    exit(0);
	  }
	  for (int idif=0; idif<nvert; idif++) fprintf(fd,"%g\n",gdx[idif]);

	  fclose(fd);

	  /* Y grad */

	  /* sprintf(gdyname,"%s_gdy.1D",out_root); */
	  sprintf(gdyname,"%s.GDY",out_root);
	  printf("Writing ydiff gradient file to %s\n",gdyname);
	  if ((fd = fopen(gdyname,"w")) == NULL) {
	    printf("Problem opening %s\n",gdyname);
	    exit(0);
	  }
	  for (int idif=0; idif<nvert; idif++) fprintf(fd,"%g\n",gdy[idif]);

	  fclose(fd);

	  /* Z grad */

	  /* sprintf(gdzname,"%s_gdz.1D",out_root); */
	  sprintf(gdzname,"%s.GDZ",out_root);
	  printf("Writing zdiff gradient file to %s\n",gdzname);
	  if ((fd = fopen(gdzname,"w")) == NULL) {
	    printf("Problem opening %s\n",gdzname);
	    exit(0);
	  }
	  for (int idif=0; idif<nvert; idif++) fprintf(fd,"%g\n",gdz[idif]);

	  fclose(fd);
	}
	fclose(gfp);

	free(gdx) ; free(gdy) ; free(gdz) ;

	/********************************* end grad write ***************************/

	/* Determine images size and allocate space for data */

	char cleanup_files[80];
	char basename[80], dname[80], hname[80] ;
	char dirname[80], brikname[80] ;
	char ppge_dir[80] ;
	char fmat[80],fname[80],orient_str[80];
	char cd_str[10000];	       /* command line to change directory */
	char to3d_str[10000];	       /* command line for program to3d */
	char ppge_str[10000];	       /* command line for program ppge */
	char buck3d_str[10000];        /* command line for program 3dbucket */
	char copy_str[10000];        /* command line for program 3dbucket */
	char move_str[10000];        /* command line for program 3dbucket */
	strcpy(brikname,out_root);

	/* Read in data */
	int kk = minarg+nopt ;

	for(int ibvl=0; ibvl<numbv; ibvl++) { /* bvalue loop */ 
	  for(int idif=0; idif<ndif; idif++) { /* diffusion direction loop */ 

	    strcpy(dirname,"tempXXXXXX");
	    mkdtemp(dirname);	     

	    if(DO_UNWARPING) {	    
	      strcpy(ppge_dir,"ppge_tempXXXXXX");
	      mkdtemp(ppge_dir);	     
	    }

	    for(int islc=0; islc<nslices; islc++) {	

	      if(debug) printf("\nargv[%i] = %s, idif=%i, islc=%i, ibvl=%i\n",
				 kk,argv[kk],idif,islc,ibvl);

	      if(!file_exists(argv[kk])) {
		printf("Uh oh!  File doesn't exist!\n\a"
		       "Exiting ...\n");
		exit(0) ;
	      } else {
		strcpy(file_name,argv[kk]);
	      }

	      sprintf(copy_str,"/bin/cp -f %s %s",file_name,dirname);
	      if(debug) printf("++++++++++++++++++++++++ buck3d_str = %s\n",copy_str);
	      system (copy_str); 

	      kk++ ;

	    } /* islc */


	    /* Do unwarping here in directory "dirname" */ 
	    if(DO_UNWARPING) {
	      fprintf(stderr,"\nUnwarping diffusion direction %i of b-value %i ...\n",idif,ibvl);
	      if (cleanup && fixscale) {
		sprintf(ppge_str,"ppgedti -d1 %s -d2 %s -i %s -o %sppge -TR %f -tmpdir %s -nomoco -halfdouble -refnum 0",
			fieldmap1,fieldmap2,dirname,brikname,optr,ppge_dir);
	      }
	      else if (!cleanup && fixscale) {
		sprintf(ppge_str,"ppgedti -d1 %s -d2 %s -i %s -o %sppge -TR %f -tmpdir %s -nomoco -nocleanup -halfdouble -refnum 0",
			fieldmap1,fieldmap2,dirname,brikname,optr,ppge_dir);
	      }
	      else if (cleanup && !fixscale) {
		sprintf(ppge_str,"ppgedti -d1 %s -d2 %s -i %s -o %sppge -TR %f -tmpdir %s -nomoco -refnum 0",
			fieldmap1,fieldmap2,dirname,brikname,optr,ppge_dir);
	      }
	      else {
		sprintf(ppge_str,"ppgedti -d1 %s -d2 %s -i %s -o %sppge -TR %f -tmpdir %s -nomoco -nocleanup -refnum 0",
			fieldmap1,fieldmap2,dirname,brikname,optr,ppge_dir);
	      }
	      if(debug) printf("ppge_str = %s\n",ppge_str);
	      sprintf(move_str,"/bin/mv -f %sppge+orig.* %s",brikname,dirname);
	      if(debug) printf("++++++++++++++++++++++++ move_str = %s\n",move_str);
              system (ppge_str);
	      system (move_str);
	    }
	    else{
	    /*----- Construct command line for program to3d -----*/
  
	    sprintf(to3d_str,"to3d -prefix %s -session %s %s/i*",brikname,dirname,dirname);
	    system (to3d_str);
	    if(debug) {
	      printf("\nto3d_str = %s\n\n",to3d_str); 
	      printf("\nargv[%i] = %s\n",kk,argv[kk]);
	    }
            }
	    if(idif==0 && ibvl==0) {
	      if (DO_UNWARPING)
		sprintf(copy_str,"/bin/cp -f %s/%sppge+orig.BRIK ./%s+orig.BRIK",dirname,brikname,out_root);
	      else
		sprintf(copy_str,"/bin/cp -f %s/%s+orig.BRIK ./%s+orig.BRIK",dirname,brikname,out_root);	      

	      if(debug) printf("++++++++++++++++++++++++ copy_str = %s\n",copy_str);
	      system (copy_str); 
	      if (DO_UNWARPING)
		sprintf(copy_str,"/bin/cp -f %s/%sppge+orig.HEAD ./%s+orig.HEAD",dirname,brikname,out_root);
	      else
		sprintf(copy_str,"/bin/cp -f %s/%s+orig.HEAD ./%s+orig.HEAD",dirname,brikname,out_root);	      
	      if(debug) printf("++++++++++++++++++++++++ copy_str = %s\n",copy_str);
	      system (copy_str); 

	    } else {
	      if (DO_UNWARPING)
		sprintf(buck3d_str,"3dbucket -abuc -glueto %s+orig %s/%sppge+orig",out_root,dirname,brikname);
	      else
		sprintf(buck3d_str,"3dbucket -abuc -glueto %s+orig %s/%s+orig",out_root,dirname,brikname);

	      if(debug) printf("++++++++++++++++++++++++ buck3d_str = %s\n",buck3d_str);
	      system (buck3d_str); 
	    }

	    sprintf(cleanup_files,"/bin/rm -fr %s*",dirname);
	    if(debug) printf("cleanup_files = %s\n",cleanup_files);
	    system(cleanup_files);

	    //TMP_EXIT;

	  } /* idif */
	} /* ibvl */

	/* Put scanner type in AFNI header */ 

	char scanner_str[10000];
	char sname[80];
	char cleanup_note[80];
	NOTE *scanner_note = set_scanner_type("SCANNER_TYPE_GE_EXCITE");
	if(debug) printf("scanner_note = %s\n",scanner_note->afni_note);
	sprintf(fname,"%s+orig.HEAD",out_root);
	if(debug) printf("header file name = %s\n",fname);
	strcpy(sname,"crapXXXXXX"); /* template required for mkstemp */
	mkstemp(sname);		   /* create unique filename because */
	if ((fd = fopen(sname,"w")) == NULL) {
	  printf("Problem opening %s\n",sname);
	  exit(0);
	}
	fprintf(fd,"%s",scanner_note->afni_note);
	fclose(fd);
	sprintf(scanner_str,"cat %s >> %s",sname,fname);
	if(debug) printf("scanner_str = %s\n",scanner_str);
	system(scanner_str);
	
	sprintf(cleanup_note,"/bin/rm -fr %s*",sname);
	if(debug) printf("cleanup_note = %s\n",cleanup_note);
	system (cleanup_note);

	/* Put other header information in AFNI header */
	if (!old_style) {
	  sprintf(fname,"%s+orig",out_root);
	  THD_3dim_dataset *dset = THD_open_dataset(fname);
	  tross_Add_Note(dset, header_text);
	  THD_write_3dim_dataset(NULL, NULL, dset, False);
	  THD_delete_3dim_dataset(dset, False);
	}
}

void instructions(void){
  printf("Usage: edif2afni [options] infile_root \n"
	 "Options:\n"
	 "-fieldmap1 field map 1\n"
	 "-fieldmap2 field map 2\n"
	 "-dict  Dictionary file\n"
	 "-prefix Output file name root\n"
	 "-pall  Print all available dicom variables\n"
	 "-debug Lots of debugging output\n"
	 "-old Produce old-style files (.header, .GD{X|Y|Z})\n"
	 "-help  Help\n"
	 );
}

void send_help()
{
  printf("No help yet for dicom! .. sorry \n"
	 "Here are the instructions:\n");
  instructions() ;
}

int parse_args( int argc, char *argv[] ) 
{
    int   nopt = 0;
    int   need_help = 0;
    int   num_field_maps = 0;

    for (int i = 1; i < argc; i++) { /* ------- Options ------- */

      if (!strncmp(argv[i], "-prefix", 6)) {  /* Outfile root */
	if (++i >= argc) { instructions(); exit(0); }
	out_root = argv[i];
	nopt++; nopt++;
      } else if (!strncmp(argv[i], "-dict", 5)) {  /* Dictionary */
	if (++i >= argc) { instructions(); exit(0); }
	dict_name_input = argv[i];
	nopt++; nopt++;
      } else if (!strncmp(argv[i], "-fieldmap1", 10)) {  /* Field map 1*/
	if (++i >= argc) { instructions(); exit(0); }
	fieldmap1 = argv[i];
	num_field_maps++ ;
	nopt++; nopt++;
      } else if (!strncmp(argv[i], "-fieldmap2", 10)) {  /* Field map 2 */
	if (++i >= argc) { instructions(); exit(0); }
	fieldmap2 = argv[i];
	num_field_maps++ ;
	nopt++; nopt++;
      } else if (!strncmp(argv[i], "-noclean", 8)) { /* don't clean up */
	cleanup = false;
	nopt++;
      } else if (!strncmp(argv[i], "-fixresolution", 8)) { /* fix resolution mismatch*/
	fixscale = true;
	nopt++;
      } else if (!strncmp(argv[i], "-debug", 6)) { /* print out arrays */
	if (i >= argc) { instructions(); exit(0); }
	noisy = 1;
	nopt++;
      } else if (!strncmp(argv[i], "-pall", 5)) { /* print out types */
	if (i >= argc) { instructions(); exit(0); }
	print_all_types = 1;
	nopt++;
      } else if (!strncmp(argv[i], "-old", 4)) { /* produce old-style output */
	if (i >= argc) { instructions(); exit(0); }
	old_style = true;
	nopt++;
      } else if (!strncmp(argv[i], "-help", 5)) { /* help */
	if (i >= argc) { instructions(); exit(0); }
	need_help = 1;
	nopt++;
      }
      
    } /* end of options */

    if(num_field_maps==1) {
      fprintf(stderr,"\n\n\t*************************************\n");
      fprintf(stderr,"\t Need two field maps to undistort!!\a\n");
      fprintf(stderr,"\t Exiting ...\a\n");
      fprintf(stderr,"\t*************************************\n");
      exit(0) ;
    } else if(num_field_maps==2) {
      DO_UNWARPING = 1;
    }

    if (need_help){
      send_help();
      exit(0);
    }
	
    if (argc < 1 + minarg + nopt){
      instructions();
      exit(0);
    }

    return nopt ;
}
