/* header file for diffuse directory */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/fcntl.h> 

#include "afni.h"

#define RECON_FLAG		/* for the next header include */
#include "../d2a/imagedb.h"
#include "../d2a/rdbm.h"

#include "./coords.h"		/* new coordinate definitions */
#include "../diffuse/grads.h"   /* Needs to be before dif_structs.h and alm.h */

#include "./thd_difmaker.h"
#include "./dif_patterns.h"
#include "../fiber/fiber.h"	/* must come before dif_structs for FTM_state defn */
#include "./dif_structs.h" 
#include "../utils/alloc.h"
#include "../utils/io.h"
#include "../vector/vector.h"
#include "../vector/matrix.h"

/* Macros */

#define DEG2RAD (M_PI/180.)
#define RAD2DEG (180./M_PI)

#define SINC(x) ( ((x)==0.) ? (1) : (sin((x))/(x)) )
#define COSC(x) ( ((x)==0.) ? (1) : (cos((x))/(x)) )

#undef  FREEUP
#define FREEUP(x) do{ if((x) != NULL){free((x)); (x)=NULL;} } while(0)

#ifndef FATAL_ERROR
#define FATAL_ERROR(msg) \
        {fprintf(stderr,"%s\a\n in file: %s at line %d\n ... exiting\n", \
		 msg,__FILE__,__LINE__);exit(1);}
#endif

#ifndef NEAR_FATAL_ERROR
#define NEAR_FATAL_ERROR(msg) \
        {fprintf(stderr,"%s\a\n in file: %s at line %d\n ... returning\n", \
		 msg,__FILE__,__LINE__);return(NULLVAL);}
#endif

#ifndef NEAR_FATAL_ZERROR
#define NEAR_FATAL_ZERROR(msg) \
        {fprintf(stderr,"%s\a\n in file: %s at line %d\n ... returning\n", \
		 msg,__FILE__,__LINE__);return(ZNULLVAL);}
#endif

#ifndef WARNING
#define WARNING(msg) \
        {fprintf(stderr,"%s\a\n in file: %s at line %d\n", \
		 msg,__FILE__,__LINE__);}
#endif

#ifndef MATCH_STR
#  define MATCH_STR(s1,s2) ((strcmp(s1,s2))?(0):(1))
#endif

#ifndef EVEN
#define EVEN(x) (((x) % 2) ? (0) : (1))
#endif
#ifndef ODD
#define ODD(x)  (((x) % 2) ? (1) : (0))
#endif

/* Print out dataset_link */

#ifndef PRINT_DSL
#define PRINT_DSL(dslink)                                 \
  do{ if((dslink)==NULL) { \
        printf("\nPRINT_DSL: Dataset Link doesn't exist!\n");  \
      } else { \
        int id ; \
	printf("\n      ndsl = %i\n",(ndsl));  \
        for( id=0 ; id < ndsl ; id++ ){ \
	  printf("  title[%i] = %s\n",id,(dsl)[id].title) ; \
	  printf("   date[%i] = %s\n",id,(dsl)[id].idcode.date) ; \
	  printf("id_code[%i] = %s\n\n",id,(dsl)[id].idcode.str) ; \
        } \
      }; \
  } while(0)
#endif

#define PRINT_ZV(v,n,label) \
do{ int i; \
    for(i=0; i<(n); i++) { \
        printf("%s[%i] = (%13.6g,%13.6g)\n",(label),i,(v)[i].r,(v)[i].i); \
    }\
} while(0) 

#define PRINT_FV(v,n,label) \
do{ int i; \
    for(i=0; i<(n); i++) { \
        printf("%s[%i] = %13.6g\n",(label),i,(v)[i]); \
    }\
} while(0) 

#define PRINT_IV(v,n,label) \
do{ int i; \
    for(i=0; i<(n); i++) { \
        printf("%s[%i] = %d\n",(label),i,(v)[i]); \
    }\
} while(0) 

#ifndef BIGVAL
#define BIGVAL 1.e+10
#endif

#ifndef LITTLE
#define LITTLE 1.e-10
#endif

#ifndef EMPTY_LABEL
#  define EMPTY_LABEL "Empty"
#endif
#ifndef EMPTY
#  define EMPTY EMPTY_LABEL
#endif

#define  NULLVAL -666		   /* NULL when numerical value needed */
#define ZNULLVAL zcplx((double)(NULLVAL),(double)(NULLVAL)) /* NULL when complex numerical value needed */

/* IMAGE STUFF */

#define US2SEC 1.e-6		/* microsec to second conversion */

/* These next two are from ses/diff.h */
#define MAXNBVALS 128		/* Maximum number of b-values */
#define MAXNDIF  3000		/* Maximum number of diffusion directions
				   Currently, the actual max is max(nvert)+1
				   where max(nvert) = max(polygon_nvert),
				   defined in sphses.h */
/* GE/ghg */
#define MXNIM        256	/* max image size          */
#define IMD    (2*MXNIM)	/* reconstruction size     */
#define MXSAMP     16390	/* max acquisition length  */
#define MXNVW        128	/* max num views   	   */
#define MXNBV        128	/* max num b-values - lrf  */
#define MAG            1	/* flags for write_image() */
#define PHASE          2
#define REAL           4
#define IMAG           8
#define NOSUBSC       16
#define	NGRID	     512	/* cal factor used in making traj files */
#define FERMIW   (10.0/128)	/* normalized width of fermi window  */
#define REVGHG   "31: 5/16/99"	/* rev number of ghg source I began with  */
#define REV      "3: 99/10/21 (lrf)" /* lrf - rev number of this source  */
#define MINARGC  3

/* DTI Stuff */

#ifndef GAMMA          
#  define GAMMA  4257.0		/* proton gyromagnetic ratio in Hz/gauss */  
#endif                   
				      
#define Gamma = 2*M_PI*GAMMA

#define MAXDATASIZE 72704    /* In WORDS */
#define SIGNASIZE   72704    /* In WORDS */
#define SIGNAHEAD    7168    /* In WORDS */
#define MAX_SLICES    128

#define NVAR  4
#define NTEN  9
#define NEVAL 3

#define TDIM 3			/* Tensor dimension TDIM x TDIM */

#define MAX_SHORT 32767
#define MAX_BYTE  255

/**************** Ylm routines ****************************/

complex ***ylm_alloc(int Lmax, int npts) ;

#define printYlm(Ylm,Lmax,nn) \
do{ int i,j,k,nl = Lmax+1; \
    for(i=0; i<nl; i++) { \
     printf("\n"); \
     for(j=-i; j<=i; j++) { \
      for(k=0; k<nn; k++) { \
	printf("Ylm[%i][%2i][%i] = (%13.6g,%13.6g)\n",\
                i,j,k,Ylm[i][j][k].r,Ylm[i][j][k].i); \
      } \
     }\
    }\
} while(0) ;

#define fillYlm(Ylm,Lmax,nn) \
do{   int i,j,k,nl = Lmax+1; \
    for(k=0; k<nn; k++) { \
    for(i=0; i<nl; i++) { \
      for(j=-i; j<=i; j++) { \
	Ylm[i][j][k].r = i*k; \
	Ylm[i][j][k].i = j*k; \
      } \
    }\
    }\
} while(0) ;

#include "../diffuse/alm.h"
#include "../diffuse/gsl_utils.h"

/*************** GSL EIG routines (Mar 04, 2004) ***********************/

double *dm_symeig( double **x, double **xevec, double *xeval, int n,
                   GSL_eig *gsle) ;

/**************** SHD routines ******************************/

/* Functions */

int sphtes( float *gx, float *gy, float *gz, int *tris, 
	    int polytope, int ntess ) ;

/**************** DTI routines ******************************/
float **setB( DIF_state *difstr ) ;
void setPattern( IMG_state *img ) ;

/* Diffusion tensor */
THD_mat33 basserD( float *data, float bval ) ;
float *diffusion_tensor( float *D, DIF_state *dif ) ;
float *tensor_eig( float *D, float *evals, DTI_eig_state *eigstr ) ;

/* Spherical Harmonic Decomposition */
int init_shd( DIF_state *dif ) ;
int sphere_parms( DIF_state *dif ) ;
int ylm_scale( DIF_state *dif ); 

/* Spherical Wave Decomposition - 16Jul04 lrf */
int init_swd( DIF_state *dif ) ;

/* SHD (cyrus) */
int init_shd_scalars( int METH_SHD_SCALAR , DIF_state *dif );
int init_dti_scalars( int METH_DTI_SCALAR , DIF_state *dif );

/* Scalar quantities */
int init_dti( DIF_state *dif ) ;
double dat2dap( float *dap , float *dat, int nvert, int nb0 ) ;
double dat2dat( float *dap , float *dat, int nvert, int nb0 ) ;
float aveD( float *ds , DIF_state *dif );
float calc_fai( float *ds, int n);
float dti_fai( float *ds , DIF_state *dif );
float dti_rai( float *ds , DIF_state *dif );
float lat( float *ds , DIF_state *dif );
float sdv( float *ds , DIF_state *dif );

/**************** least-squares (start) ******************************/

DTI_lsq_state *init_lsq( int    nt, 
			 double dt, 
			 int    nrts ) ;

void free_lsqstr( DTI_lsq_state *lsqstr, 
		  int    nt, 
		  double dt, 
		  int    nrts ) ;

//void lsqfit_dif( );
void lsqfit_dif ( DTI_lsq_state *lsqstr, float *data); 

double *startup_lsqfit_dif( int veclen , 
			    float *wt , int nref , float *ref[] ) ;
float *delayed_lsqfit_dif( int veclen ,
			   float * data , int nref , float *ref[] , double * cc ) ;

#define LABEL_LEN  4096           /* used by lsqfit_dif.c */

/**************** least-squares (end) ******************************/

/* END DTI Stuff */

/* SHD Stuff */

/* Defaults for viewing dimensions */

#define CALC_shd_grid_size_DEF 50
#define CALC_full_mcalc_range_DEF 1
#define CALC_Lmax_MAX 50   /* maximum value allowable for Lmax (arbitrary but used) */
#define CALC_Mmax_MAX  CALC_Lmax_MAX
#define CALC_Lmax_DEF 4
#define CALC_Lmin_DEF 0
#define CALC_nl_DEF   (CALC_Lmax_DEF-CALC_Lmin_DEF)+1
#define CALC_Mmax_DEF  CALC_Lmax_DEF
#define CALC_Mmin_DEF -CALC_Lmax_DEF
#define CALC_nm_DEF   (CALC_Mmax_DEF-CALC_Mmin_DEF)+1

/* END SHD Stuff */

extern int anisokill ;		/* now global (yikes!) */

/* Routines */

DIF_state     *init_dif_state( char * , DSET_stats * , char * ) ;
IMG_state     *init_img_state( char *, char * ) ;
IMG_state     *init_img_state_GE_LX( char *) ;
IMG_state     *init_img_state_GE_EXCITE( char *) ;
IMG_state     *init_img_state_VARIAN( char *) ;
DTI_state     *init_dti_state( IMG_state * ,  DSET_stats * ) ;
SHD_state     *init_shd_state( IMG_state * ) ;
DIF_status    *init_dif_status( DIF_state * ) ;
DTI_lsq_state *init_dti_lsq_state( DTI_prc_state * ) ;
DTI_eig_state *init_dti_eig_state( DTI_prc_state * ) ;
DTI_prc_state *init_dti_prc_state( IMG_state * , DSET_stats * ) ;

/* Pulse sequence stuff */

/* B-value sampling mode (cv: bstep) */

#define BSAMP_LINB 0
#define BSAMP_LOGB 1
#define BSAMP_LING 2
#define BSAMP_LOGG 3

/* Pulse sequence type definitions */

#define HYPER_SEQ 0
#define   STE_SEQ 1
#define    SE_SEQ 2
#define    GE_SEQ 3

/* Diffusion encoding strategies. dif->encode */

/*#define NODIFF 0
  #define MANUAL 1
  #define BASSER 2
  #define SPHERE 3*/

/* Pulse sequence diffusion sampling modes 
   These must correspond to equivalent DIFF_SAMP
   variables in epi3.e:sphtes.h */

#define DIFF_SAMP_NONE 0		/* no diffusion */
#define DIFF_SAMP_XYZ0 1		/* xyz diffusion */
//#define DIFF_SAMP_BASS 2		/* basser diffusion */
#define DIFF_SAMP_GE 2		/* GE's electrostatic repulsion */
#define DIFF_SAMP_TESS 3		/* spherical tesselation */


/* Polytope definitions */

#define POLY_OCT 0		/* octahedron */
#define POLY_TET 1		/* tetrahedron */
#define POLY_ICO 2		/* icosahedron */

static char tmpbuf[256];		/* for general use */

#define PRINT_RETURN printf("\n")
#define BLANK_LINE PRINT_RETURN 

               // compute 4th-order DT from SHD coefficients
THD_3dim_dataset *shd2tensor4th( THD_3dim_dataset *shd_dset, char *prefix );

