/** dmsymeig.c *************************************************************

     Title:  Double SYM Matrix EIGensystem

   Purpose:  Compute eigenvalues and eigenvectors of
             a real, symmetric matrix and return in
	     descending order.

      Call: double *dm_symeig( double **x,    
                               double **xevec,
			       double *xeval,
			       int n,        
			       GSL_eig *gsle)

     Input: x[n][n] = input matrix
		  n = dimensions of arrays
               gsle = GSL_eig structure

    Output:  xevec[][k] = normalized eigvenvector corresponding to d[k]
               xeval[n] = eigenvalues

 Algorithm: see GSL manual

   History: 2004-Jan-21   Initial verions
            2004-Feb-28   Add gsle
*/
#include "gsl_utils.h"

double *dm_symeig( double **x, double **xevec, double *xeval, int n, 
		   GSL_eig *gsle)
{
  int gsle_null = 0;

  if(gsle==NULL) {  /* internal allocation */

    gsle_null = 1;
    gsle = init_GSL_eig(n) ;

  }

  /* configure */

  dm2gsl(gsle->gmat,x,n,n) ;

  /* Eigensystem */

  gsl_eigen_symmv(gsle->gmat,gsle->eval,gsle->evec,gsle->work) ;
  gsl_eigen_symmv_sort(gsle->eval,gsle->evec,GSL_EIGEN_SORT_VAL_DESC) ;

  /* un-configure */

  gsl2dm(xevec,gsle->evec,n,n) ;
  gsl2dv(xeval,gsle->eval,n) ;

  /* clean up if internally allocated */

  if(gsle_null) free_GSL_eig( gsle );

  return (xeval) ;
}
      
      
