/**<dti.c>*******************************************************

     Title: Diffusion Tensor Imaging methods

   Purpose: Create diffusion tensor ellipsoids

    Author: Lawrence R. Frank

   History: 08-May-01 cp dell.c -> dti.c

*/
#include "diffuse.h"

/******************************************************************************

     Title: INITialize DTI

      Call: int init_dti( DIF_state *dif )

   Purpose: Allocate space for dti.c

     Input: DIF_state *dif

    Output: prc->Dapp(nxy,ntens) 
            prc->Bval(ndif,lsqdim) 


******************************************************************************/

int init_dti( DIF_state *dif )
{
  DIF_status    *sta = dif->sta ;
  IMG_state     *img = dif->img ;
  DTI_prc_state *prc = dif->dti->prc ;
  DTI_lsq_state *lsq = dif->dti->lsq ;

  prc->Dapp  = fm_alloc( img->nxy  , prc->nten  );   /* Allocate diffusion matrix */
  prc->Bmat = fm_alloc( img->dgrad->nvert*img->nbvals , lsq->lsqdim );  /* Allocate diffusion matrix */
  prc->Evals = fm_alloc( img->nxy ,  prc->edim );    /* Allocate eigenvalue image */
  prc->fv    = fv_alloc( img->nbvals * img->dgrad->nvert ) ; /* work space */
  prc->dv    = dv_alloc( img->nbvals * img->dgrad->nvert ) ; /* work space */

  if (prc->Dapp==NULL || prc->Bmat==NULL) {
    sta->dti_initialized = 0;
    return(0) ;
 } else {
    sta->dti_initialized = 1;
    return(1) ;
  }
}
/******************************************************************************

     Title: BASSER mode diffusion tensor reconstruction

   Purpose: Compute diffusion tensor for a single voxel

      Call: float *basser( float *D , float *data, DIF_state *dif )

     Input: D[3][3] = diffusion tensor 
            data[7] = data[0] (b=0), data[1-6] (6 encoded directions)
    Output: D = diffusion tensor subvector (ie, 9 element vector)

 Algorithm: Basser, JMR SerB 103:247-254 (1994)

******************************************************************************/

THD_mat33 basserD( float *data, float bval )
{
  int   i;
  float S0 = data[0];		/* b=0 assumed to be in first slot */
  float Si ;			/* 6 diff directions (data[1-6]) */

  THD_mat33 sum_mat33,tmp_mat33 ;

  CLEAR_MAT33(sum_mat33) ;

  for (i=0; i<6; i++) {

    Si = (log(S0) - log(data[i+1])) / bval ;

    MAT2THD(tmp_mat33,basser_dual[i]) ;
    SCALE_MAT33(tmp_mat33,Si);
    ACCUM_MAT33(sum_mat33,tmp_mat33) ;

  }

  return (sum_mat33) ;
}

/******************************************************************************

     Title: DIFFUSION TENSOR

      Call: float *diffusion_tensor( float *D , DIF_state *dif )

   Purpose: Compute diffusion tensor for a single voxel

     Input: idata = magnitude image data from voxel with b=bval
	     bval = b-value
	      nbv = number of bvalues
	      ndd = number of diffusion directions

    Output: D = diffusion tensor subvector (ie, 9 element vector)

 Algorithm: Basser, JMR SerB 103:247-254 (1994)

******************************************************************************/

float *diffusion_tensor( float *D , DIF_state *dif )
{
  char *program_name = "diffusion_tensor" ;
  int   i,ir;
  float A0 = D[0];		/* b=0 assumed to be in first slot */
  int   debug = 0;

  IMG_state     *img = dif->img ;
  DTI_lsq_state *lsq = dif->dti->lsq ;
  DTI_prc_state *prc = dif->dti->prc ;

  if(debug) printf("Inside %s: ndif = %i\n",program_name,img->ndif);

  for (i=img->nb0; i<img->ndif; i++) {

    prc->fv[i-img->nb0] = (D[i]==0)? 1: (float) log( (double)D[i]/A0 ) ;
    if(debug) printf("%s: D[%i] = %lf, prc->fv[%i] = %lf\n",
		     program_name,i,D[i],i-img->nb0,prc->fv[i-img->nb0]);

  }

  /* regress! */

  lsqfit_dif( lsq , prc->fv );

  if(debug) for( ir=0 ; ir < lsq->nref ; ir++ ) 
    printf("coeff[%i] = %lf\n",ir,lsq->alpha[ir]);

  /* b=0 */

  //  lsq->alpha[0] = A0;	/* A(0) */

  /* Put back in tensor form */

  D[0] = lsq->alpha[0] ;	/* dxx */
  D[4] = lsq->alpha[1] ;	/* dyy */
  D[8] = lsq->alpha[2] ;	/* dzz */

  D[1] = lsq->alpha[3] ;	/* dxy */
  D[2] = lsq->alpha[4] ;	/* dxz */
  D[5] = lsq->alpha[5] ;	/* dyz */

  D[3] = D[1] ;			/* dyx = dxy */
  D[6] = D[2] ;			/* dzx = dxz */
  D[7] = D[5] ;			/* dzy = dyz */


  return(D);
}

/******************************************************************************

     Title: TENSOR EIGenvalues

      Call: double *tensor_eig(float *D, double *evals, DTI_eig_state *eigstr )

   Purpose: Calculate tensor eigenvalues

     Input: D[lxy][ndd] = unique elements of diffusion tensor at each voxel

    Output: evals

   History: lrf 12/27/98 

******************************************************************************/

float *tensor_eig( float *D, float *evals, DTI_eig_state *eig )
{
  int i;
  int   debug = 0;

  /* fv2dm( eig->dm, D, eig->nr, eig->nc ); */

  for (i=0;i<eig->nten;i++) eig->dv[i] = (double) D[i]; 

  /* Eigenvalues are in ascending order */

  if(debug)
  {
    printf("tensor_eig:\n");
    for (i = 0; i <eig->nten; ++i)
    {
      printf("dv[%d] = %f\n", i, eig->dv[i]);
    }
  }

  symeig_double( eig->nr, eig->dv, eig->evals );

  if(debug)
  {
    for (i = 0; i < eig->nc; ++i)
    {
      printf("evals[%d] = %f\n", i, eig->evals[i]);
    }
    for (i = 0; i < eig->nten; ++i)
    {
      printf("dv[%d] = %f\n", i, eig->dv[i]);
    }
  }

  /* Reorder eigenvalues into descending order */

  for ( i=0; i<eig->nc; i++ ) evals[i] = (float) eig->evals[eig->nc - i - 1]; 

  return(evals);
}

/**<setB.c>************************************************************

     Title: SET up B matrix 

   Purpose: Compute B matrix from diffusion gradient
            amplitude and timings.

      Call: float **setB( float **B, DIF_state *dif)

     Input: B[nbv*ndd][img->lsqdim] = empty input matrix
                   dif = diffusion structure containing all 
		            diff grad amps and timings

    Output: float **B

     Notes: This routine is essentially "init_diff.e" in ses
*/

float **setB( DIF_state *dif )
{
  char *program_name = "setB";
  void  setBscale( float *bsclx, float *bscly, float *bsclz, IMG_state *imgs );
  void  setStep( IMG_state *imgs );
  void  setPattern( IMG_state *imgs ) ;
  float **genB( float **B, float bsclx,float bscly,float bsclz, IMG_state *imgs );
  float bsclx,bscly,bsclz ;
  int debug = 1;

  IMG_state *img = dif->img ;
  Grad *dgrad = img->dgrad ;

  /* B-factor */

  if(img->bval_available) {
    bsclx = -img->bval ;
    if(debug) printf("%s: bsclx = %f\n",program_name,bsclx); fflush(stdout);
  } else {
    printf("\tCalculating gradient timing scale factor...\n");fflush(stdout);
    setBscale(&bsclx,&bscly,&bsclz,img) ; /* timings scale factors */
  }

  /* Gradient pattern */

  if(img->bpat_available) {
    if(debug) printf("%s: Copying input grads ...\n",program_name); fflush(stdout);
    for (int i=0; i<dgrad->nvert; i++) {  
      img->gdxpat[i+img->nb0] = img->GX[i];
      img->gdypat[i+img->nb0] = img->GY[i];
      img->gdzpat[i+img->nb0] = img->GZ[i];
    }
  } else {
    printf("\tGenerating gradient pattern...\n");
    setPattern(img);		             /* setup gradient pattern */
  }

  /* B-value stepping */

  /* CURRENT HACK!!  13-Apr-03, lrf
     It is assumed here that img->bstep = BSAMP_LINB,
     ie, that we are using linear B sampling.
     Since right now there is only that available in epi3.e */

  if(img->bval_available) {

    if(debug) printf("%s: Generating b-value steps ... ",program_name); fflush(stdout);
    for (int ibv=0; ibv < img->nbvals; ibv++){
      img->gdbstep[ibv] = (float) (ibv + 1)/img->nbvals ;
    }
    if(debug) printf("done.\n"); fflush(stdout);

  } else {
    printf("\tGenerating b-value stepping...\n");
    setStep(img);		             /* setup b-value stepping */
  }

  /* Construct b-matrix */
  float **Bmat = dif->dti->prc->Bmat ; 
  printf("\tConstructing b matrix...\n");fflush(stdout);
  genB(Bmat,bsclx,bscly,bsclz,img);    /* generate B matrix */

  return( Bmat );
}

/*<genB>*********************************************************

     Title: GENerate B matrix

   Purpose: 

*/
static int UTRI[] = {0,4,8,1,2,5,3,6,7} ;

float **genB( float **B, 
	      float bsclx, float bscly, float bsclz, 
	      IMG_state *img )
{
  char *program_name = "genB";
  int i,j,indx,jndx;
  /*  float fx,fy,fz ; */
  float fx ; 
  Grad *dgrad = img->dgrad ;

  int debug = 0;

  for (j=0; j<img->nbvals; j++) { /* b-values */

    jndx = j*img->dgrad->nvert ;

    fx = bsclx * img->gdbstep[j] ;

    printf("Constructing B-matrix for bvalue[%i] = %f\n",j,fx);fflush(stdout);
    if(debug) printf("genB: bsclx = %f, img->gdbstep[%i] = %f, fx = %lf\n",
		     bsclx,j,img->gdbstep[j],fx); fflush(stdout);

    for (i=img->nb0; i<img->ndif; i++) {   /* diffusion directions */

      indx = (i-img->nb0) + jndx;

      /* b=0 */
//      B[indx][0] = 1.0;		                          /* A0  */

      /* diagonals */
      B[indx][0] = fx * img->gdxpat[i] * img->gdxpat[i] ; /* Bxx */
      B[indx][1] = fx * img->gdypat[i] * img->gdypat[i] ; /* Byy */
      B[indx][2] = fx * img->gdzpat[i] * img->gdzpat[i] ; /* Bzz */

      /* off-diagonals */
      B[indx][3] = fx * img->gdxpat[i] * img->gdypat[i] ; /* Bxy */
      B[indx][4] = fx * img->gdxpat[i] * img->gdzpat[i] ; /* Bxz */
      B[indx][5] = fx * img->gdypat[i] * img->gdzpat[i] ; /* Byz */

      if(debug)
      {
	      printf("gradient[%d] = (%f %f %f)\n", i, img->gdxpat[i],
		     img->gdypat[i], img->gdzpat[i]);
      }

    }
  }

  if(debug) printf("B matrix:\n");
  if(debug) for (i = 0; i < img->dgrad->nvert * img->nbvals; ++i) 
	  printf("%f %f %f %f %f %f\n",
		 B[i][0], B[i][1], B[i][2], B[i][3], B[i][4], B[i][5]);
  return(B);

}
/*<setPattern>*********************************************************

     Title: SET up gradient Pattern

   Purpose: 

     Notes: This is the same stuff as in init_diff.e in ses.e

   History: 01/12/28 lrf - Changed img->gdipat[] output to reflect changes
                           in sphtes, where (0,0,0) point is no longer included.
			   img->gdipat requires it at first entry, so we just
			   put it in manually here. (see sphtes.c)
*/
void setPattern( IMG_state *img )
{
  char *program_name = "setPattern" ;
  int i;
  int nvert;
  char buf[256];

  Grad *dgrad = img->dgrad ;

  /* Diffusion pattern */

  switch (img->difmode) {

    case DIFF_SAMP_NONE:			/* no diffusion */

      img->gdxpat[0] = 0.;
      img->gdypat[0] = 0.;
      img->gdzpat[0] = 0.;

      break;

    case DIFF_SAMP_XYZ0:			/* manual method */

      for (i=0;i<img->ndif;i++) {
	  img->gdxpat[i] = (float) manpat[i][0];
	  img->gdypat[i] = (float) manpat[i][1];
	  img->gdzpat[i] = (float) manpat[i][2];
      }

      break;

    case DIFF_SAMP_TESS:			/* Tesselated polytope */

      /* Tesselation of a sphere by platonic solid "polytope" at level "ntess".
         Returns diffusion directions as vertices in gd[x,y,z]pat */

      printf("inside %s: dgrad->ntess = %i, dgrad->polytope = %i\n",
	     program_name,dgrad->ntess,dgrad->polytope);

      nvert = sphtes( dgrad->gx,dgrad->gy,dgrad->gz,
		      dgrad->tris, dgrad->polytope, dgrad->ntess) ;
      
      for (i=0;i<img->nb0;i++) {
	  img->gdxpat[i] = 0.0 ;
	  img->gdypat[i] = 0.0 ;
	  img->gdzpat[i] = 0.0 ;
      }
      for (i=0;i<dgrad->nvert;i++) {
	  img->gdxpat[i+img->nb0] = dgrad->gx[i] ;
	  img->gdypat[i+img->nb0] = dgrad->gy[i] ;
	  img->gdzpat[i+img->nb0] = dgrad->gz[i] ;
      }

      if(nvert!=dgrad->nvert) {
	sprintf(buf,"Number of vertices (=%i) calculated by " 
		"sphtes() in setPattern()\n"
		"does not equal that retrieved "
		"from image header file (=%i)!\a\n",
		nvert,dgrad->nvert);
	FATAL_ERROR(buf);
      } else {
	printf("%s: Determined %i vertices.\n",program_name,nvert);
      }

      break;

    default:
      printf("Invalid diffusion mode selected.\n");
      exit(0);
  }
}

/*<setStep>*********************************************************

     Title: SET up gradient STEPping

   Purpose: 

     Notes: This is the same stuff as in init_diff.e in ses.e

*/
void setStep( IMG_state *img )
{
  int   i;
  float ginc, fstep, nbv2 ;

  /* B-value stepping */

  switch (img->bstep) {

    case BSAMP_LINB:			/* linear b stepping */


      nbv2 = sqrt((double)img->nbvals) ;
      ginc = (img->nbvals==0) ? GDMAX : (GDMAX/nbv2) ;
      for (i=1; i<=img->nbvals; i++) {
	  fstep = sqrt((double)i) * ginc;
	  img->gdbstep[i-1] = fstep;
      }
      break;

    case BSAMP_LOGB:			/* linear G stepping */

      ginc = (img->nbvals==0) ? GDMAX : (GDMAX/img->nbvals) ;
      for (i=1; i<=img->nbvals; i++) {
	  fstep = i * ginc;
	  img->gdbstep[i-1] = fstep;
      }
      break;

    case BSAMP_LING:			/* log b stepping */

      nbv2 = img->nbvals * img->nbvals ;
      ginc = (img->nbvals==0) ? (img->bslope*GDMAX) : (img->bslope*GDMAX/nbv2) ;
      for (i=1; i<=img->nbvals; i++) {
	  fstep = i * i * ginc;
	  img->gdbstep[i-1] = pow(10.,fstep)/pow(10.,img->bslope*GDMAX);
      }

      break;

    case BSAMP_LOGG:			/* log G stepping */

      ginc = (img->nbvals==0) ? (img->bslope*GDMAX) : (img->bslope*GDMAX/img->nbvals) ;
      for (i=1; i<=img->nbvals; i++) {
	  fstep = i * ginc;
	  img->gdbstep[i-1] = pow(10.,fstep)/pow(10.,img->bslope*GDMAX);
      }

      break;

    default:

      printf("Invalid b-mode %i selected.\n",img->bstep); 
      exit(0);

  }

}

/*<setBscale>*********************************************************

     Title: B value SCALE factor

   Purpose: Calculate the timing-related scaling factor for b-value diffusion 

     Units: Standard unit for D are mm^2/s
            so we want to calculate b in s/mm^2

	    Pulse sequence timings are in ms (converted from us in init_dif)
	    GAMMA is in Hz/G
	    amps are in G/cm, so we need to convert to G/mm

     Notes: This is the same stuff as in diff_calcs.e in ses.e

            ONLY RETURNS X VALUE NOW!
*/
void setBscale( float *bsclx, float *bscly, float *bsclz, IMG_state *img )
{
  float gam   = TWOPI * GAMMA ; 
  float gam2  = gam * gam ;
  float ampx2 = img->agdx1 * img->agdx1 ;
  float ampy2 = img->agdy1 * img->agdy1 ;
  float ampz2 = img->agdz1 * img->agdz1 ;
  int debug = 1;

  /* width of diff encoding grads */

  float del = img->pwgd - img->pwdramp ;

  /* standard factor */
  float a1 = del * del * ( img->Delta - (del/3.0) ) ;

  /* ramp factor */
  float a2 = pow(img->pwdramp,3.0)/30. - del*pow(img->pwdramp,2.0)/6. ;

  /* Complete b scaling factor 
     This assumes all amps are scaled to same max value! */

  *bsclx = -gam2 * (a1 + a2) * ampx2 ;
  *bscly = -gam2 * (a1 + a2) * ampy2 ;
  *bsclz = -gam2 * (a1 + a2) * ampz2 ;

  img->bsclx = *bsclx ;		/* 21-Oct-03 */
  img->bscly = *bscly ;
  img->bsclz = *bsclz ;

  if(debug) printf("\n   del = %g\n"
		   "    a1 = %g\n"
		   "    a2 = %g\n"
		   "    Delta = %g\n"
		   "    pwdramp = %g\n"
		   "  ampx2 = %g, ampy2 = %g, ampz2 = %g\n"
		   "  bsclx = %g, bscly = %g, bsclz = %g\n\n",
		   del,a1,a2,img->Delta,img->pwdramp,ampx2,ampy2,ampz2,*bsclx,*bscly,*bsclz);

}
