/**<dti_scalars.c>*******************************************************

     Title: DTI SCALARS from tensor data

   Purpose: Compute scalar measures from diffusion tensor

    Author: Lawrence R. Frank

   Methods: average diffusion coefficient
            fractional anisotropy index
            relative anisotropy index
            lattice anisotropy index
            spherical diffusion variance

   History: 02-Oct-01
            27-Oct-02 get rid of all masks (which are now done in Maker)

*/
#include "diffuse.h"

/**<init_dti_scalars.c>******************************************************

     Title: INITialize SCALARS

      Call: int init_dti_scalars( DIF_state *dif )

   Purpose: Allocate space for dti_scalars.c

     Input: DIF_state *dif

    Output: prc->fv(ndif) 

******************************************************************************/

int init_dti_scalars( int METH_DTI_SCALAR , DIF_state *dif )
{
  DIF_status    *sta = dif->sta ;
  IMG_state     *img = dif->img ;
  DTI_prc_state *prc = dif->dti->prc ;
  int ival ;

  switch ( METH_DTI_SCALAR ) {

      default: 
	printf("\adti_scalars: Unrecognized value %i of METH_DTI_SCALAR! \n",
	       METH_DTI_SCALAR);
	printf("...returning...\n");
	return(-1);

      case METH_DTI_AVE:
	
	if (prc->dv==NULL ) {
	  ival = 0 ;
	} else {
	  ival = 1 ;
	}

	sta->dti_ave_initialized = ival;
	break;

      case METH_DTI_FAI:

	if (prc->dv==NULL ) {
	  ival = 0 ;
	} else {
	  ival = 1 ;
	}

	sta->dti_fai_initialized = ival;
	break;

      case METH_DTI_RAI:

	if (prc->dv==NULL ) {
	  ival = 0 ;
	} else {
	  ival = 1 ;
	}

	sta->dti_rai_initialized = ival;
	break;

      case METH_DTI_LAT:

	if (prc->dv==NULL ) {
	  ival = 0 ;
	} else {
	  ival = 1 ;
	}

	sta->dti_lat_initialized = ival;
	break;

      case METH_DTI_SDV:

	prc->fv = fv_alloc( img->nbvals * img->ndif ) ; /* work space */

	if (prc->fv==NULL ) {
	  ival = 0 ;
	} else {
	  ival = 1 ;
	}

	sta->dti_sdv_initialized = ival;
	break;

   }

  return ( ival );
}

/**<aveD.c>*********************************************************************

     Title: AVErage Diffusion coefficient

   Purpose: Compute average diffusion coefficient

      Call: float aveD( float *fs , DIF_state *dif )

     Input: fs[3] = eigvalues 1,2,3

    Output: aveD = sum(eigenvalues)/3

******************************************************************************/

float aveD( float *fs , DIF_state *dif )
{
  DIF_status    *sta = dif->sta ;
  DTI_eig_state *eig = dif->dti->eig ;

  float val = vmean( fs, eig->nevals ) ;

  return(val);

}

/**<dti_fai.c>*****************************************************************

     Title: Fractional Anisotropy Index

   Purpose: Compute fractional anisotropy index

      Call: float dti_fai( float *fs , DIF_state *dif )

     Input: fs[3] = eigvalues 1,2,3

    Output: fai = fractional anisotropy index

    Algorithm: 

******************************************************************************/

float calc_fai( float *evals, int n )
{
#if 1
#define fai_scal sqrt(3./2.) 

  float dvar = vvar( evals, n ) ;
  float dmsq = vmsq( evals, n ) ;
  
  float  rat ;
  if(dmsq>LITTLE) rat = dvar/dmsq ;
  else            rat = 0.0 ;
  float   val = (float)(fai_scal * sqrt(rat)) ;
//  float   val = (float) sqrt(rat) ;

  return(val);
#else
  return sqrt(pow(evals[0] - evals[1], 2) + 
	      pow(evals[1] - evals[2], 2) +
	      pow(evals[0] - evals[2], 2))/
	  (sqrt(pow(evals[0], 2) + pow(evals[1], 2) + pow(evals[2], 2)) *
	   sqrt(2.0));
#endif
}

float dti_fai( float *fs , DIF_state *dif )
{

  float fai = calc_fai(fs,3);

  return(fai);

}

/**<dti_rai.c>*****************************************************************

     Title: Relative Anisotropy Index

   Purpose: Compute Relative Anisotropy Index

      Call:

     Input: fs[3] = eigvalues 1,2,3

    Output: rai = 

    Algorithm: 

******************************************************************************/

float dti_rai( float *fs , DIF_state *dif )
{
  float val ;
  DIF_status    *sta = dif->sta ;
  DTI_eig_state *eig = dif->dti->eig ;

  float fvar = vvar(fs,eig->nevals) ;
  float fave = vmean(fs,eig->nevals) ;

  val = sqrt(fvar)/sqrt(fave*fave) ;

  return(val);

}

/**<lat.c>*********************************************************************

     Title: LATtice anisotropy index

   Purpose: Compute lattice anisotropy index

      Call: float lat( float *fs , DIF_state *dif )

     Input: fs[3] = eigvalues 1,2,3

    Output: lat = 

******************************************************************************/

float lat( float *fs , DIF_state *dif )
{
  DTI_eig_state *eig = dif->dti->eig ;
  DIF_status    *sta = dif->sta ;

  /* float val = vsum( fs, eig->nevals ) / 3.0; */
  float val = 0.;

  printf("dti_scalars: lat not available yet!\n\a");
  return(val);

}
/******************************************************************************

     Title: Spherical Diffusion Variance

   Purpose: Compute spherical diffusion variance

      Call:

     Input: idata = magnitude image data from voxel with b=bval
	     bval = b-value

    Output: sdv = variance of apparent diffusion coefficient

    Algorithm: Frank, MRM (2001)

******************************************************************************/

float sdv( float *fs , DIF_state *dif )
{
  DTI_prc_state *prc = dif->dti->prc ;
  IMG_state     *img = dif->img ;
  Grad        *dgrad = img->dgrad ;
  float  val ;
  int    nvert = dgrad->nvert ;

  int debug = 0;

  val = vvar( prc->fv, dgrad->nvert ) ;

  if(debug) {
    PRINT_FV(prc->fv,dgrad->nvert,"prc->fv");
    printf("------------------------------ sdv: val = %f\n",val);
  }

  return(val);

}

/**<dat2dap>******************************************************************

     Title: DATa TO Dapp

   Purpose: Convert data vector into Apparent Diffusion Coefficient
            Dapp = -log(data(az,po)/data(b=0))

      Call: double dat2dap( double *dap , float *dat, int nvert, int nb0 )

     Input: dat[ndif=nvert+nb0] = measured diffusion values
	    nvert = # vertices measured
	    nb0 = number of b0 images

    Output: dap[nvert] = apparent diffusion coeff

     Notes: 1) This routine assumes that the b0 data is contained in
               the first nb0 slots of dat.

            2) This return the value of <b0 images>, rather than the 
               usual pointed to the resultant vector, since <> is 
	       used for thresholding in most routines that call dat2dap.

******************************************************************************/
double dat2dap( float *dap , float *dat, int nvert, int nb0 )
{
  register int i,ii;
  float    dd,db0 ;
  
  if(nb0>1) {
    db0 = vmean(dat,nb0) ;
  } else {
    db0 = dat[0];
  }

  for(i=0; i<nvert; i++) {
    ii = i + nb0 ;
    dd = (double) (dat[ii]/db0) ;
    dap[i] = -log(dd);
  }

  return db0;
}

/* return dap without b=0, otherwise unchanged */
double dat2dat( float *dap , float *dat, int nvert, int nb0 )
{
  register int i,ii;
  float    dd,db0 ;
  
  for(i=0; i<nvert; i++) {
    ii = i + nb0 ;
    dap[i] = dat[ii];
  }

  return(0.0);
}
