/**<Legendre.c>*******************************************************

     Title: associated LEGENDRE polynomial

   Purpose: Compute the associated Legendre polynomial P_{l}^{m}(x)
            where 0 <= m <= l and -1 <= x <= 1 for the vector x[n]

      Call: float *Legendre(float *y, int l, int m, float *x, int n)

     Input:    l = order
               m = degree
	    x[n] = argument
	       n = length(x)

    Output: y[n] = P_{l}^{m}(x[0],...,x[n-1])

 Algorithm: Numerical Recipes pg 254.

*/
#include "diffuse.h"

float Legendre1(int l, int m, float x);

/* Legendre function at all points x[n] for one value of (L,M) */

float *Legendre(float *y, int l, int m, float *x, int n)
{
  int i;

  if (m < 0 || m > l) {
    printf("Error! Legendre: \a");
    printf("m=%i not in range [0,l] = [%i,%i]\n",m,0,l);
    return NULL;
  }

  for (i=0; i<n; i++) {
    if (fabs(x[i]) > 1.0) {
      printf("Error! Legendre: \a");
      printf("abs(x[%i]) = %lf greater than 1.0\n",i,x[i]);
      return NULL;
    }
  }

  for (i=0; i<n; i++) {
    y[i] = Legendre1(l,m,x[i]);
  }

  return(y);
}

/* Legendre function at one point x */

float Legendre1(int l, int m, float x)
{
  float fact, pll, pmm, pmmp1, somx2;
  int i,ll;

  pmm = 1.0;

  if (m > 0) {
    somx2 = sqrt((1.0-x)*(1.0+x));
    fact = 1.0;
    for (i=1; i<=m; i++) {
      pmm *= -fact*somx2;
      fact += 2.0;
    }
  }

  if (l==m) {

    return pmm;

  } else {

    pmmp1 = x * (2*m +1)*pmm;
    if (l==(m+1)) {
      return pmmp1;
    } else {
      for (ll=m+2; ll<=l; ll++) {
	pll = (x*(2*ll-1)*pmmp1 - (ll+m-1)*pmm)/(ll-m);
	pmm =  pmmp1;
	pmmp1 = pll;
      }
      return pll;
    }

  }

}
