/**<SphericalHarmonicY.c>****************************************************

     Title: SPHERICAL HARMONIC Y_{l}^{m}

   Purpose: Compute the Spherical Harmonic of order l and degree m
            where l = any integer
	         -l <= m <= l 
	          0 <= theta <= pi    (polar angle)
		  0 <=  phi  <= 2*pi  (azimuthal angle)

      Call: complex **SphericalHarmonicY(complex **ylm, int l, int m, int n)

     Input:    l = order
               m = degree
	       n = angular resolution (in both azim and polr angles)

    Output: y[n] = Y_{l}^{m}(theta,phi)

 Algorithm: 

*/
#include "diffuse.h"

complex **SphericalHarmonicY(complex **ylm, int l, int m, int npolr, int nazim)
{
  int     i,j;
  int     am = abs(m);
  float   c1,c2;
  float   *y1, *polr, *azim, *cth ;
  float   min_polr = 0., max_polr = PI ;
  float   min_azim = 0., max_azim = TWOPI ;
  complex *y2;

  /* Check input */

  if (l < 0) {
    printf("Error! SphericalHarmonicY: \a");
    printf("l=%i must be >= 0 \n",l);
    return NULL;
  }
  if (am > l) {
    printf("Error! SphericalHarmonicY: \a");
    printf("m=%i not in range [0,l] = [%i,%i]\n",m,0,l);
    return NULL;
  }

  /* Create angular ranges */

  polr = (float *) calloc(npolr,sizeof(float));
  azim = (float *) calloc(nazim,sizeof(float));

  vlinsp(polr,min_polr,max_polr,npolr);
  vlinsp(azim,min_azim,max_azim,nazim);

  /* Associated Legendre function P_{l}^{m}(cos(polr)) */

  cth = (float *) calloc(npolr,sizeof(float));
  y1  = (float *) calloc(npolr,sizeof(float));

  for(i=0; i<npolr; i++) cth[i] = cos(polr[i]);
  Legendre(y1,l,am,cth,npolr);

  free(polr); 
  free(cth);

  /* Compute exp(i*m*azim) */

  y2 = (complex *) calloc(nazim,sizeof(complex));
  for(j=0; j<nazim; j++) y2[j] = CEXPIT(am*azim[j]);
  free(azim); 

  c1 = sqrt((2*l+1)/(4*PI)) ;
  c2 = sqrt(factorial(l-am)/factorial(l+am)) ;

  if(m<0) {

    float f = pow(-1.0,(double)m);
    complex ztmp;

    for(i=0; i< npolr; i++) {
      for(j=0; j< nazim; j++) {
	ztmp = zscal(c1*c2*y1[i],y2[j]); 
	ylm[i][j] = zscal(f,zcnjg(ztmp)); 
      }
    }

  } else {

    for(i=0; i< npolr; i++) {
      for(j=0; j< nazim; j++) {
	ylm[i][j] = zscal(c1*c2*y1[i],y2[j]); 
      }
    }

  }

  /* Clean up */
  free(y1);
  free(y2);

  return ylm;
}
