/**<dti.c>*******************************************************

     Title: Diffusion Tensor Imaging methods

   Purpose: Create diffusion tensor ellipsoids

    Author: Lawrence R. Frank

   History: 08-May-01 cp dell.c -> dti.c

*/
#include "dti.h"

/*********************************************************************************

     Title: INITialize DTI

      Call: int init_dti( DIF_state *dif )

   Purpose: Allocate space for dti.c

     Input: DIF_state *dif

    Output: prc->Dapp(nxy,ntens) 
            prc->Bval(ndif,lsqdim) 


*********************************************************************************/

int init_dti( DIF_state *dif )
{
  DIF_img_state *img = dif->dis ;
  DIF_prc_state *prc = dif->dps ;
  DIF_lsq_state *lsq = dif->dls ;

  prc->Dapp  = fm_alloc( img->nxy  , prc->nten  );  /* Allocate diffusion matrix */
  prc->Bvals = fm_alloc( img->ndif , lsq->lsqdim ); /* Allocate diffusion matrix */
  prc->Evals = fm_alloc( img->nxy ,  prc->edim );   /* Allocate eigenvalue image */
  prc->fv    = fv_alloc( img->nbvals * img->ndif ) ; /* work space */
  prc->dv    = dv_alloc( img->nbvals * img->ndif ) ; /* work space */

  if (prc->Dapp==NULL || prc->Bvals==NULL) {
    return(0) ;
    prc->dti_initialized = 0;
  } else {
    prc->dti_initialized = 1;
    return(1) ;
  }
}

/*********************************************************************************

       Title: SET THRESHold

     Purpose: Set threshold and return mask.

   Algorithm: primitive.

*/
float set_thresh( short int *mask, short int *x, int n)
{
  int i;
  float thresh = 0.;

  for (i=0; i<n; i++) thresh += x[i];
  thresh /= n;			/* mean */

  for (i=0; i<n; i++) mask[i] = (x[i]>thresh)? 1:0;

  return(thresh);
} 

/*********************************************************************************

     Title: DIFFUSION TENSOR IMAGES

      Call: void diffusion_tensor_images( float     **D,
                                          short int **iout,
			                  double    **evals,
					  int       lxy,
					  DIF_eig_state *eigstr )

   Purpose: Compute various images from eigenvalues of diffusion tensor 
            computed at every voxel in an image in fit_tensors

     Input: iout[lxy] = 
            evals[lxy][nevals] = eigenvalues of diffusion tensor 
	                         (ie, 3 element vector at each voxel)

    Output: iout[i][0] = trace at voxel i
            iout[i][1] = relative anisotropy index at voxel i
            iout[i][1] = eval 1 at voxel i
            iout[i][2] = eval 2 at voxel i
            iout[i][3] = eval 3 at voxel i

*********************************************************************************/

short int **diffusion_tensor_images( float     **D,
				     short int **iout,
				     double    **evals,
				     int       lxy,
				     DIF_eig_state *eigstr )
{
  int i;
  double dscale = 1.e3;

  for (i=0; i<lxy; i++) {
    iout[0][i] = (short int) (dscale * dvsum(evals[i],eigstr->nc)) ;
    iout[1][i] = (short int) 
      (dscale * relative_anisotropy_index(D[i],evals[i],eigstr->dm,eigstr->nc) );
    iout[2][i] = (short int) (dscale * evals[i][0]) ;
    iout[3][i] = (short int) (dscale * evals[i][1]) ;
    iout[4][i] = (short int) (dscale * evals[i][2]) ;
  }

  return(iout);
}

/*********************************************************************************

     Title: High Angular Resolution Diffusion STATisticiS IMAGES

      Call: short int **hard_stats_images( float     **D,
			                   float     **B,
		 	                   short int ***idata,
					   short int **iout,
					   double    **evals,
					   int       lxy,
					   DIF_eig_state *eigstr )

   Purpose: Compute various images from eigenvalues of diffusion tensor 
            computed at every voxel in an image in fit_tensors

     Input: iout[lxy] = 
            evals[lxy][nevals] = eigenvalues of diffusion tensor 
	                         (ie, 3 element vector at each voxel)

    Output: iout[i][0] = trace at voxel i
            iout[i][1] = relative anisotropy index at voxel i
            iout[i][1] = eval 1 at voxel i
            iout[i][2] = eval 2 at voxel i
            iout[i][3] = eval 3 at voxel i

*********************************************************************************/

short int **hard_stats_images( float     **D,
			       float     **B,
			       short int ***idata,
			       short int *imask,
			       short int **iout,
			       float     ***fout,
			       int       lxy,
			       DIF_state *dif )
{
  int i,j,k;
  /* double dscale = 1.e5; */
  double dmin,dmax;
  double d0,dk;
  int    idmin,idmax;
  DIF_img_state *img = dif->dis ;
  DIF_prc_state *prc = dif->dps ;
  int nd = img->ndif - 1;
  double imwidth,imrange;

  double *dvar = (double *)calloc(lxy, sizeof(double)); 

  for (j=0; j<img->nbvals; j++) {

    for (i=0; i<lxy; i++) {

      if(imask[i]) {

	d0 = (double) idata[i][j][0];

	for (k=0; k<nd; k++) {

	  /* img->dv[k] = (double)(idata[i][j][0] - idata[i][j][k+1]); */
	  dk = (double) idata[i][j][k+1];
	  prc->dv[k] = -log((dk / d0)); 
	}

	dvar[i] = dvvar(prc->dv,nd) ;

      } else {

	dvar[i] = 0.;

      }

      dvran(prc->dv,&dmin,&idmin,&dmax,&idmax,nd) ;
      fout[j][i][0] = (float) prc->dv[idmax];     /* var */
      fout[j][i][1] = (float) img->gdxpat[idmax]; /* x coord */
      fout[j][i][2] = (float) img->gdypat[idmax]; /* y coord */
      fout[j][i][3] = (float) img->gdzpat[idmax]; /* z coord */
      /* iout[j][i] = (short int) (dvar[i]*dscale); */
    }

    /* scale output image from 0-255 */

    dvran(dvar,&dmin,&idmin,&dmax,&idmax,lxy) ;
    /* printf("dmin = %lf, dmax = %lf\n",dmin,dmax); */
    imwidth = 255;
    imrange = dmax - dmin;
    for (i=0; i<lxy; i++) 
      if(imask[i]) {
	iout[j][i] = imwidth * (dvar[i]/imrange);
      } else {
	iout[j][i] = 0;
      }
  }

  free(dvar);

  return(iout);
}

/*********************************************************************************

     Title: FIT diffusion TENSORS

      Call: float **fit_tensors( float     **D,
			         float     **B,
				 double    **evals,
				 short int ***idata,
				 int       nbv,
				 int       ndd,
				 int       lxy,
				 DIF_state *dif )

   Purpose: Compute diffusion tensor ellipse at every voxel in an image

     Input: idata[lxy][ndd] = magnitude image data from image with b=bval 
                              and ndd diffusion directions
		       bval = b-value
		        nbv = number of bvalues directions
		        ndd = number of diffusion directions

    Output:     D[lxy][ncc] = diffusion tensor subvector (ie, 6 element vector) 
                              at each voxel
         evals[lxy][nevals] = eigenvalues of diffusion tensor (ie, 3 element vector)

*********************************************************************************/

void fit_tensors( DIF_state *dif )
{
  int i,j;

  DIF_img_state *img = dif->dis ;
  DIF_prc_state *prc = dif->dps ;
  DIF_eig_state *eig = dif->des ;
  DIF_lsq_state *lsq = dif->dls ;

  /* Put B into lsq */

  for (j=0; j<lsq->lsqdim; j++) { /* size of alpha */
    for (i=0; i<prc->nimsl; i++) {     /* ndif*nbvals */  
      lsq->tsim[j][i] = prc->Bvals[i][j] ;
    }
  }

  for (i=0; i<img->nxy; i++) {
    /* diffusion_tensor( lsq, prc->Dapp[i], idata[i], img->nbvals, img->ndif ); */
    diffusion_tensor( prc->Dapp[i] , dif );
    tensor_eig( prc->Dapp[i], prc->Evals[i], eig ); 
  }

  free_lsqstr( lsq, lsq->nt, lsq->dt, lsq->lsqdim );

  return ;
}

/*********************************************************************************

     Title: DIFFUSION TENSOR

      Call: float *diffusion_tensor( DIF_lsq_state *lsqstr,
                                     float     *D, 
			             short int **idata, 
				     int       nbv,
				     int       ndd )

   Purpose: Compute diffusion tensor for a single voxel

     Input: idata = magnitude image data from voxel with b=bval
	     bval = b-value
	      nbv = number of bvalues
	      ndd = number of diffusion directions

    Output: D = diffusion tensor subvector (ie, 9 element vector)

    Algorithm: Basser, JMR SerB 103:247-254 (1994)

*********************************************************************************/

/* float *diffusion_tensor( DIF_lsq_state *lsq,
   float     *D, 
   short int **idata, 
   int       nbv,
   int       ndd) */

float *diffusion_tensor( float *D , DIF_state *dif )
{
  int i,ir;
  float A0 = D[0];		/* b=0 assumed to be in first slot */
  int debug = 0;
  DIF_img_state *img = dif->dis ;
  DIF_lsq_state *lsq = dif->dls ;
  DIF_prc_state *prc = dif->dps ;

  for (i=0; i<img->ndif; i++) {	/* check why i=0 is start and not i=1 */

    prc->fv[i] = (D[i]==0)? 1: (float) log( (double)D[i]) ;
    if(debug) printf("D[%i] = %lf, prc->fv[%i] = %lf\n",i,D[i],i,prc->fv[i]);

  }

  /* regress! */

  lsqfit_dif( lsq , prc->fv );

  if(debug) for( ir=0 ; ir < lsq->nref ; ir++ ) 
    printf("coeff[%i] = %lf\n",ir,lsq->alpha[ir]);

  /* b=0 */

  lsq->alpha[0] = A0;	/* A(0) */

  /* Put back in tensor form */

  D[0] = lsq->alpha[1] ;	/* dxx */
  D[4] = lsq->alpha[2] ;	/* dyy */
  D[8] = lsq->alpha[3] ;	/* dzz */

  D[1] = lsq->alpha[4] ;	/* dxy */
  D[2] = lsq->alpha[5] ;	/* dxz */
  D[5] = lsq->alpha[6] ;	/* dyz */

  D[3] = D[1] ;			/* dyx = dxy */
  D[6] = D[2] ;			/* dzx = dxz */
  D[7] = D[5] ;			/* dzy = dyz */

  return(D);
}

/*********************************************************************************

     Title: TENSOR EIGenvalues

      Call: double *tensor_eig(float *D, double *evals, DIF_eig_state *eigstr )

   Purpose: Calculate tensor eigenvalues

     Input: D[lxy][ndd] = unique elements of diffusion tensor at each voxel

    Output: evals

   History: lrf 12/27/98 

*********************************************************************************/

float *tensor_eig( float *D, float *evals, DIF_eig_state *eig )
{
  int i;

  /* fv2dm( eig->dm, D, eig->nr, eig->nc ); */

  for (i=0;i<eig->nten;i++) eig->dv[i] = (double) D[i]; 

  /* Eigenvalues are in ascending order */
  symeig_double( eig->nr, eig->dv, eig->evals );

  /* Reorder eigenvalues into descending order */
  for ( i=0; i<eig->nc; i++ ) evals[i] = (float) eig->evals[eig->nc - i - 1]; 

  return(evals);
}

/*********************************************************************************

     Title: RELATIVE ANISOTROPY INDEX

      Call: double relative_anisotropy_index(float *D, double *evals, 
                                             double **dm, int n )

   Purpose: calculate relative anisotropy index from diffusion tensor

     Input: D[lxy][ndd] = unique elements of diffusion tensor at each voxel
            evals[lxy][ndd] = eigenvalues of D
	    n = diffusion tensor matrix dimensions (n x n)

    Output: rai = relative anisotropy index

   History: lrf 06/08/99

*********************************************************************************/

double relative_anisotropy_index( float *D, double *evals , double **dm, int n )
{
  register int i,j;
  double  rai;

  double trace = dvsum(evals,n) ;

  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) {
      dm[i][j] = D[i+j*n] - (i==j)*trace/3.0;
      dm[i][j] *= dm[i][j] ;
    }
  }

  rai = sqrt(3.0 * dmtrace(dm,n,n)) / trace ;
  if(rai<0.) rai = 0.0;

  return(rai);
}

/**<setB.c>************************************************************

     Title: SET up B matrix 

   Purpose: Compute B matrix from diffusion gradient
            amplitude and timings.

      Call: float **setB( float **B, DIF_state *dif)

     Input: B[nbv*ndd][img->lsqdim] = empty input matrix
                   dif = diffusion structure containing all 
		            diff grad amps and timings

    Output: float **B

     Notes: This routine is essentially "init_diff.e" in ses
*/

float **setB( DIF_state *dif )
{
  void  setBscale( float *bsclx, float *bscly, float *bsclz, DIF_img_state *img );
  void  setStep( DIF_img_state *img );
  void  setPattern( DIF_img_state *img ) ;
  float **genB( float **B, float bsclx,float bscly,float bsclz, DIF_img_state *img );
  float bsclx,bscly,bsclz ;

  float **Bvals = dif->dps->Bvals ; 

  printf("\tCalculating gradient timing scale factor...\n");
  setBscale(&bsclx,&bscly,&bsclz,dif->dis) ; /* timings scale factors */
  printf("\tGenerating gradient pattern...\n");
  setPattern(dif->dis);		           /* setup gradient pattern */
  printf("\tGenerating b-value stepping...\n");
  setStep(dif->dis);		           /* setup b-value stepping */
  printf("\tConstructing b matrix...\n");
  genB(Bvals,bsclx,bscly,bsclz,dif->dis); /* generate B matrix */

  return( Bvals );
}

/*<genB>*********************************************************

     Title: GENerate B matrix

   Purpose: 

*/
#define UTRI[] = {0,4,8,1,2,5,3,6,7};

float **genB( float **B, 
	      float bsclx, float bscly, float bsclz, 
	      DIF_img_state *img )
{
  int debug = 0;
  int i,j,indx,jndx;
  /*  float fx,fy,fz ; */
  float fx ; 

  for (j=0; j<img->nbvals; j++) { /* b-values */

    jndx = j*img->ndif ;
    fx = bsclx * img->gdbstep[j] ;
    if(debug) printf("\ngenB: fx = %lf\n",fx);

    for (i=0; i<img->ndif; i++) {   /* diffusion directions */

      indx = i + jndx;

      /* b=0 */
      B[indx][0] = 1.0;		                                /* A0  */

      /* diagonals */
      B[indx][1] = fx * img->gdxpat[i] * img->gdxpat[i] ; /* Bxx */
      B[indx][2] = fx * img->gdypat[i] * img->gdypat[i] ; /* Byy */
      B[indx][3] = fx * img->gdzpat[i] * img->gdzpat[i] ; /* Bzz */

      /* off-diagonals */
      B[indx][4] = fx * img->gdxpat[i] * img->gdypat[i] ; /* Bxy */
      B[indx][5] = fx * img->gdxpat[i] * img->gdzpat[i] ; /* Bxz */
      B[indx][6] = fx * img->gdypat[i] * img->gdzpat[i] ; /* Byz */

    }
  }

  return(B);

}
/*<setPattern>*********************************************************

     Title: SET up gradient Pattern

   Purpose: 

     Notes: This is the same stuff as in init_diff.e in ses.e

*/
void setPattern( DIF_img_state *img )
{
  int   i;

  /* Diffusion pattern */

  switch (img->difmode) {

    case 0:			/* no diffusion */

      img->gdxpat[0] = 0.;
      img->gdypat[0] = 0.;
      img->gdzpat[0] = 0.;

      break;

    case 1:			/* manual method */

      for (i=0;i<img->ndif;i++) {
	  img->gdxpat[i] = (float) manpat[i][0];
	  img->gdypat[i] = (float) manpat[i][1];
	  img->gdzpat[i] = (float) manpat[i][2];
      }

      break;

    case 2:			/* Basser method */

      for (i=0;i<img->ndif;i++) {
	  img->gdxpat[i] = (float) bgpat[i][0];
	  img->gdypat[i] = (float) bgpat[i][1];
	  img->gdzpat[i] = (float) bgpat[i][2];
      }

      break;

    case 3:			/* Sphere method */

      /* Tesselation of a sphere by platonic solid "polytope" at level "ntess".
         Returns diffusion directions as vertices in gd[x,y,z]pat */

      sphses( img ); 

      break;

    default:
      printf("Invalid diffusion mode selected.\n");
      exit(0);
  }
}

/*<setStep>*********************************************************

     Title: SET up gradient STEPping

   Purpose: 

     Notes: This is the same stuff as in init_diff.e in ses.e

*/
void setStep( DIF_img_state *img )
{
  int   i;
  float ginc, fstep, nbv2 ;

  /* B-value stepping */

  switch (img->bmode) {

    case 1:			/* linear b stepping */

      nbv2 = img->nbvals * img->nbvals ;
      ginc = (img->nbvals==0) ? GDMAX : (GDMAX/nbv2) ;
      for (i=1; i<=img->nbvals; i++) {
	  fstep = i * i * ginc;
	  img->gdbstep[i-1] = fstep;
      }
      break;

    case 2:			/* linear G stepping */

      ginc = (img->nbvals==0) ? GDMAX : (GDMAX/img->nbvals) ;
      for (i=1; i<=img->nbvals; i++) {
	  fstep = i * ginc;
	  img->gdbstep[i-1] = fstep;
      }
      break;

    case 3:			/* log b stepping */

      nbv2 = img->nbvals * img->nbvals ;
      ginc = (img->nbvals==0) ? (img->bslope*GDMAX) : (img->bslope*GDMAX/nbv2) ;
      for (i=1; i<=img->nbvals; i++) {
	  fstep = i * i * ginc;
	  img->gdbstep[i-1] = powf(10.,fstep)/powf(10.,img->bslope*GDMAX);
      }

      break;

    case 4:			/* log G stepping */

      ginc = (img->nbvals==0) ? (img->bslope*GDMAX) : (img->bslope*GDMAX/img->nbvals) ;
      for (i=1; i<=img->nbvals; i++) {
	  fstep = i * ginc;
	  img->gdbstep[i-1] = powf(10.,fstep)/powf(10.,img->bslope*GDMAX);
      }

      break;

    default:
      printf("Invalid b-value selected.\n"); 
      exit(0);

  }

}

/*<setBscale>*********************************************************

     Title: B value SCALE factor

   Purpose: Calculate the timing-related scaling factor for b-value diffusion 

     Units: Standard unit for D are mm^2/s
            so we want to calculate b in s/mm^2

	    Pulse sequence timings are in ms (converted from us in init_dif)
	    GAMMA is in Hz/G
	    amps are in G/cm, so we need to convert to G/mm

     Notes: This is the same stuff as in diff_calcs.e in ses.e

            ONLY RETURNS X VALUE NOW!
*/
void setBscale( float *bsclx, float *bscly, float *bsclz, DIF_img_state *img )
{
  int debug = 0;
  float gam   = TWOPI * GAMMA ; 
  float gam2  = gam * gam ;
  float ampx2 = img->agdx1 * img->agdx1 ;
  float ampy2 = img->agdy1 * img->agdy1 ;
  float ampz2 = img->agdz1 * img->agdz1 ;

  /* width of diff encoding grads */
  float del = img->pwgd - img->pwdramp ;

  /* standard factor */
  float a1 = del * del * ( img->delta - (del/3.0) ) ;

  /* ramp factor */
  float a2 = powf(img->pwdramp,3.0)/30. - del*powf(img->pwdramp,2.0)/6. ;

  /* Complete b scaling factor 
     This assumes all amps are scaled to same max value! */

  *bsclx = -gam2 * (a1 + a2) * ampx2 ;
  *bscly = -gam2 * (a1 + a2) * ampy2 ;
  *bsclz = -gam2 * (a1 + a2) * ampz2 ;

  if(debug) printf("\n   del = %g\n"
		   "    a1 = %g\n"
		   "    a2 = %g\n"
		   "  ampx2 = %g, ampy2 = %g, ampz2 = %g\n"
		   "  bsclx = %g, bscly = %g, bsclz = %g\n\n",
		   del,a1,a2,ampx2,ampy2,ampz2,*bsclx,*bscly,*bsclz);

}
