/*
 * sphses - SPHere for SES
 *  Generate a triangle mesh approximating a sphere by
 *  recursive subdivision. First approximation is a platonic
 *  solid; each level of refinement increases the number of
 *  triangles by a factor of 4.
 *
 *  Input: sphses [level] [-options]
 *	 level is an integer >= 1 setting the tessellation level (default 1).
 *	 Options:
 *	 -o starts with an octahedron [def] 
 *	 -t starts with a tetrahedron 
 *	 -i starts with a icosahedron 
 *	 -v vertex output only
 *	 -f face output only
 *	 -m mathematica output
 *	 -p prints output
 *	 -c  generate triangles with vertices in counterclockwise
 *	     order as viewed from the outside in a RHS coordinate system.
 *	     The default is clockwise order.
 *
 * 3/24/89       Jon Leech (leech@cs.unc.edu) 
 *    5/93       icosahedral code added by Jim Buddenhagen 
 *               (jb1556@daditz.sbc.com) 
 *
 * 99/9/15  lrf - add vertex structure
 *                add mathematica output
 *                vertex determination (naive brute force)
 * 99/9/16  lrf - faster determination of vertices
 *                using a hashtable (!)
 * 99/9/21  lrf - cp sphere.c sphses.c
 *                ses optimized version of sphere.c
 *                switches from vertex_union to vertex_unhash
 *                at HASHTHRESH < difstr->ntess
 * 99/9/22  lrf - dome mode
 * 99/11/02 lrf - difstr sole input, update difstr->ndif,difstr->nvert
 * 01/12/28 lrf - more normal input for more general use; returns nvert
 *                Gradients no longer include superfluous (0,0,0) at
 *		  first points - only actual vertices in g[] arrays!
 *                Also, sphses no longer requires dif_structs.h (good!)
 *                polytope now has allowable values (0,1,2) instead of (1,2,3)
 */
#include "sphses.h"
#include <stdlib.h>
/*--------------------------------------------------------------------------*/

int sdebug   = 0;   /* Don't debug */
int doprint  = 0;   /* Don't print out */
int FACEflag = 0;   /* Faces only */
int VERTflag = 1;   /* Vertices only */
int MATHflag = 1;   /* Don't generate Mathematica format output */
int Flatflag = 0;   /* Don't generate per-vertex normals */

/* Forward declarations */
int same_pt(cpoint *pt1, cpoint *pt2) ;
void flip_object( object *obj );
void print_object( object *obj, int level );
void print_triangle(object *obj,int indx) ;
void print_vertex(object *obj,int indx) ;
void print_pt(cpoint *a, char *name) ;
cpoint *normalize( cpoint *p );
cpoint *midpoint( cpoint *a, cpoint *b );
int vertex_union(object *obj) ;
int vertex_unhash( object *old, int **hashtable, 
		   int polr_size, int azim_size ) ;

/* comment out for sphses.e */
/* extern char *malloc( unsigned ); */

int sphses( float *gx, float *gy, float *gz, 
	    int polytope, int ntess, int DOME )
{
    int    nvert, union_count,unhash_count;
    indx2  ahash,bhash,chash;
    int    **hashtable ;

    object *old,		/* Default is octahedron */
	   *new_var;
    int     ccwflag = 0,	/* Reverse vertex order if true */
	    level,		/* Current subdivision level */
	    i;

    int polr_size = HASH_POLR ;
    int azim_size = HASH_AZIM ;

    if(sdebug) printf("polr_size = %i, azim_size = %i\n",polr_size,azim_size);

    switch (polytope) {
      case POLY_OCT:
	if(DOME)
	  old = &octd;
	else
	  old = &oct;
	break;
      case POLY_TET:
	if(DOME)
	  old = &tetd;
	else
	  old = &tet;
	break;
      case POLY_ICO:
	if(DOME)
	  old = &icod;
	else
	  old = &ico;
	break;
      default:
	printf("\nsphses: Invalid polytope type (= %i)!\a ... exiting\n\n",
	       polytope);
	exit(0);
    }

    /* only need 0 - Pi */
    /* if(DOME) polr_size /= 2; */

    if(doprint) 
      printf("polyvert = %d\n",polygon_nvert[polytope][ntess-1]);

    if (ccwflag) flip_object(old);

    hashtable = (int **) calloc(polr_size , sizeof(int *));
    for(i=0; i<polr_size; i++)
      hashtable[i] = (int *) calloc(azim_size , sizeof(int));

    /* Load original vertices into hashtable */
    for(i=0; i<old->nvert; i++) {
      ahash = hashfunc2(old->vert[i]) ;
      hashtable[ahash.ix][ahash.iy] = 1;
    }

    if(doprint) {
      printf("\nPolygon = %s: #faces = %d, #vertices = %d\n",
	     old->type,old->nface,old->nvert);
      printf("\nAllocating hashing table of size = [%i][%i]:\n\n",
	     polr_size,azim_size);

    }

    /* Subdivide each starting triangle (ntess - 1) times */

    for (level = 1; level < ntess; level++) {
	new_var = (object *)malloc(sizeof(object));	/* Allocate a new_var object */
	if (new_var == NULL) {
	  printf("sphere: Out of memory on subdivision level %d\n",
		  level);
	  exit(-1);
	}
	new_var->nface = old->nface * 4;

	/* Allocate 4* number of face points in the current approximation */
	new_var->face  = (triangle *)calloc(new_var->nface , sizeof(triangle));
	if (new_var->face == NULL) {
	  printf("sphere: Out of memory on subdivision level %d\n",
		  level);
	  exit(-1);
	}

	/* Subdivide each triangle in the old approximation and normalize
	 *  the new_var points thus generated to lie on the surface of the unit
	 *  sphere.
	 * Each input triangle with vertices labelled [0,1,2] as shown
	 *  below will be turned into four new_var triangles:
	 *
	 *			Make new_var points
	 *			    a = (0+2)/2
	 *			    b = (0+1)/2
	 *			    c = (1+2)/2
	 *	  1
	 *	 /\		Normalize a, b, c
	 *	/  \
	 *    b/____\ c		Construct new_var triangles
	 *    /\    /\		    [0,b,a]
	 *   /	\  /  \		    [b,1,c]
	 *  /____\/____\	    [a,b,c]
	 * 0	  a	2	    [a,c,2]
	 *
	 */

	for (i = 0; i < old->nface; i++) {
	    triangle
		 *oldt = &old->face[i],
		 *new_vart = &new_var->face[i*4];
	    cpoint a, b, c;

	    /* New vertices */
	    a = *normalize(midpoint(&oldt->pt[0], &oldt->pt[2]));
	    b = *normalize(midpoint(&oldt->pt[0], &oldt->pt[1]));
	    c = *normalize(midpoint(&oldt->pt[1], &oldt->pt[2]));
	      
	    ahash = hashfunc2(a) ;
	    bhash = hashfunc2(b) ;
	    chash = hashfunc2(c) ;

	    if(doprint && sdebug) {
	      printcpt(a,"a");
	      printcpt(b,"b");
	      printcpt(c,"c");

	      printindx2(ahash,"a-hash");
	      printindx2(bhash,"b-hash");
	      printindx2(chash,"c-hash");

	      printcpt(spt2cpt(unhashfunc2(ahash)),"a-unhash");
	      printcpt(spt2cpt(unhashfunc2(bhash)),"b-unhash");
	      printcpt(spt2cpt(unhashfunc2(chash)),"c-unhash");

	      printf("\n");
	    }

	    /* Store in hash table */
	    hashtable[ahash.ix][ahash.iy] = 1;
	    hashtable[bhash.ix][bhash.iy] = 1;
	    hashtable[chash.ix][chash.iy] = 1;

	    new_vart->pt[0] = oldt->pt[0];
	    new_vart->pt[1] = b;
	    new_vart->pt[2] = a;
	    new_vart++;

	    new_vart->pt[0] = b;
	    new_vart->pt[1] = oldt->pt[1];
	    new_vart->pt[2] = c;
	    new_vart++;

	    new_vart->pt[0] = a;
	    new_vart->pt[1] = b;
	    new_vart->pt[2] = c;
	    new_vart++;

	    new_vart->pt[0] = a;
	    new_vart->pt[1] = c;
	    new_vart->pt[2] = oldt->pt[2];
	}

	if (level > 1) {
	    free(old->face);
	    free(old);
	}

	/* Continue subdividing new_var triangles */
	old = new_var;
    }

    if(ntess<HASHTHRESH) {
      /* Find unique vertices by union (ie bubble) method */
      if(doprint) printf(" \nFinding unique vertices...\n");
      union_count = vertex_union(old) ; 
      nvert = union_count;
    } else {
      /* Find unique vertices by unhashing method */
      if(doprint) printf(" \nUnhashing vertices...\n");
      unhash_count = vertex_unhash(old,hashtable,polr_size,azim_size) ;
      nvert = unhash_count;
    }
    /* ndif = nvert + 1; */

    /* Transfer vertices to img */

    /* sphere components */
    for (i=0;i<nvert;i++) { /* nvert = ndif-1 */
      gx[i] = old->vert[i].x ;
      gy[i] = old->vert[i].y ;
      gz[i] = old->vert[i].z ;
    }

    /* Print out sorted object */
    if(doprint) print_object(old, ntess);

    return(nvert);
}

int vertex_unhash( object *old, int **hashtable, 
		   int polr_size, int azim_size )
{
  int    i,j ;
  indx2  hashpt;
  spoint spt ;
  cpoint cpt ;
  static int vcount = 0;
  object *new_var;

  int    tmp_nvert;

  /* Total number of vertices */

  tmp_nvert = 3000;		/* HACK!!! nvert for level 5 = 2562 */

  /* Allocate new_var object */

  new_var = (object *)malloc(sizeof(object));
  if (new_var == NULL) {
    printf("vertex_unhash: Out of memory on subdivision level \n");
    exit(-1);
  }

  new_var->vert = (cpoint *)calloc(tmp_nvert,sizeof(cpoint));
  if (new_var->vert == NULL) {
    printf("vertex_unhash: Out of memory on subdivision level \n");
    exit(-1);
  }

  for (i = 0; i < polr_size; i++) {
    for (j = 0; j < azim_size; j++) {
      if(hashtable[i][j]) {
	hashpt = newhashpt(i,j);
	spt = unhashfunc2(hashpt);
	cpt = spt2cpt(spt);
	new_var->vert[vcount] = cpt ;
	vcount++ ;
      }
    }
  }

  if(doprint) 
    printf(" old vertex count = %d, new_var vertex count = %d\n",
	   old->nvert,vcount);

  /* Allocate final vert sub-struct */
  old->nvert = vcount;
  old->vert = (cpoint *)calloc(old->nvert,sizeof(cpoint));
  if (old->vert == NULL) {
    printf("vertex_unhash: Out of memory!\n");
    exit(-1);
  }

  /* Load vertices into old object */
  for (i = 0; i < old->nvert; i++) old->vert[i] = new_var->vert[i];

  /* clean up */
  free(new_var->vert);
  free(new_var);

  return(vcount) ;
}

int vertex_union(object *old) 
{
  int    i,j,k,l;
  int    vcount,keept;
  object *new_var;
  cpoint tstpt ;

  /* Total number of vertices */

  old->nvert = old->nface*3;

  /* Allocate new_var object */

  new_var = (object *)malloc(sizeof(object));
  if (new_var == NULL) {
    printf("vertex_union: Out of memory on subdivision level \n");
    exit(-1);
  }

  new_var->vert = (cpoint *)calloc(old->nvert,sizeof(cpoint));
  if (new_var->vert == NULL) {
    printf("vertex_union: Out of memory on subdivision level \n");
    exit(-1);
  }

  vcount = 0;

  for (i = 0; i < old->nface; i++) { /* cycle through all faces */
    for (k = 0; k < 3; k++) {	     /* and each face vertex */

      keept = 1;
      tstpt = old->face[i].pt[k] ;
      for (j = i+1; j < old->nface; j++) { /* for each face ... */
	for (l = 0; l < 3; l++) { /* check vertices */

	  if(same_pt(&tstpt,&old->face[j].pt[l])) {
	    keept = 0;
	    break ;
	  }

	} /* test face vertex loop */
      }	/* test face loop */

      if(keept) {
	new_var->vert[vcount] = tstpt ;
	vcount++;
      }

    } /* face vertex loop */
  } /* face loop */

  if(doprint) 
    printf(" old vertex count = %d, new_var vertex count = %d\n",
	   old->nvert,vcount);

  /* Allocate final vert sub-struct */
  old->nvert = vcount;
  old->vert = (cpoint *)calloc(old->nvert,sizeof(cpoint));
  if (old->vert == NULL) {
    printf("vertex_union: Out of memory!\n");
    exit(-1);
  }

  /* Load vertices into old object */
  for (i = 0; i < old->nvert; i++) old->vert[i] = new_var->vert[i];

  /* clean up */
  free(new_var->vert);
  free(new_var);

  return(vcount) ;
}

int same_pt(cpoint *pt1, cpoint *pt2) 
{

  if( (pt1->x==pt2->x) && (pt1->y==pt2->y) && (pt1->z==pt2->z) )
    return(1) ;
  else
    return(0) ;

}

/* 
   Normalize a point p 
*/
cpoint *normalize(cpoint *p)
{
    static cpoint r;
    double mag;

    r = *p;
    mag = r.x * r.x + r.y * r.y + r.z * r.z;
    if (mag != 0.0) {
	mag = 1.0 / sqrt(mag);
	r.x *= mag;
	r.y *= mag;
	r.z *= mag;
    }

    return &r;
}

/*
  Return the midpoint on the line between two points 
*/
cpoint *midpoint(cpoint *a, cpoint *b)
{
    static cpoint r;

    r.x = (a->x + b->x) * 0.5;
    r.y = (a->y + b->y) * 0.5;
    r.z = (a->z + b->z) * 0.5;

    return &r;
}

/* 
   Reverse order of points in each triangle 
*/
void flip_object(object *obj)
{
    int i;
    for (i = 0; i < obj->nface; i++) {
	cpoint tmp;
		       tmp = obj->face[i].pt[0];
	obj->face[i].pt[0] = obj->face[i].pt[2];
	obj->face[i].pt[2] = tmp;
    }
}

void print_pt(cpoint *a,char *name)
{
  printf("%s = (%g %g %g)\n", name,a->x, a->y, a->z);
}

/* 
   Write out all triangles in an object 
*/
void print_object(object *obj,int level)
{
    int i;

    /* Spit out coordinates for each face (triangle) */
    if(FACEflag) {
      if(doprint) printf("\n Faces:\n\n");
      for (i = 0; i < obj->nface; i++)
	print_triangle(obj,i);
    }

    /* Spit out coordinates for each vertex */
    if(VERTflag) {
      if(doprint) printf("\n Vertices:\n\n");
      for (i = 0; i < obj->nvert; i++)
	print_vertex(obj,i);
    }

}

/* 
   Output vertex 
*/
void print_vertex(object *obj,int indx)
{
    cpoint *v = &obj->vert[indx];

    if (MATHflag) {
      if(indx==0) printf("{");
      printf("{%g,%g,%g}", v->x, v->y, v->z);
      if(indx==obj->nvert-1)
	printf("}\n");
      else
	printf(",\n");
    } else {
      printf("cpoint\n");
      printf("\t%g %g %g\n", v->x, v->y, v->z);
    }
}

/* 
   Output a triangle 
*/
void print_triangle(object *obj,int indx)
{
    triangle *t = &obj->face[indx];
    int i;

    if (MATHflag) {
      /* printf("MATH triangle\n"); */
      if(indx==0) printf("{");
	printf("{");
	for (i = 0; i < 3; i++) {
	    printf("{%g,%g,%g}", t->pt[i].x, t->pt[i].y, t->pt[i].z);
	    if(i<2) printf(",");
	}
	if(indx==obj->nface-1)
	  printf("}}\n");
	else
	  printf("},\n");
    } else {
	/* Modify this to generate your favorite output format
	 * Triangle vertices are in t->pt[0..2].{x,y,z}
	 * A generic format is provided.
	 */
	printf("triangle\n");
	for (i = 0; i < 3; i++)
	    printf("\t%g %g %g\n", t->pt[i].x, t->pt[i].y, t->pt[i].z);
    }
}
