#include "diffuse.h"

#define DT2ND   // 2nd-order DT

// calculate 4th-order diffusion tensor components from the
// spherical harmonical decomposition (SHD) coefficients
THD_3dim_dataset *shd2tensor4th( THD_3dim_dataset *shd_dset, char *prefix )
{
#ifndef DT2ND
  int numSubBricks = 16;  // 15 distinct tensor elements + GFA
#else
  int numSubBricks = 22;  // 2nd-order DT + 4th-order DT + GFA
#endif
  int nvox, shd_nbricks;  // for SHD dataset
  complex *gfa;   // generalized FA
  THD_3dim_dataset *dt4_dset; // dataset to hold 4th-order DT 
               // SHD coefficients
  complex *a00, *a20, *a21, *a22, *a40, *a41, *a42, *a43, *a44;
  int a00_offset, a20_offset, a40_offset;
  int i; 

  a00_offset = 0; 
  a20_offset = 6; 
  a40_offset = 20;
               // Load SHD dataset into memory in case it is not.
  DSET_load( shd_dset );
  a00 = (complex *)DSET_ARRAY( shd_dset, a00_offset );
  a20 = (complex *)DSET_ARRAY( shd_dset, a20_offset );
  a21 = (complex *)DSET_ARRAY( shd_dset, a20_offset+1 );
  a22 = (complex *)DSET_ARRAY( shd_dset, a20_offset+2 );
  a40 = (complex *)DSET_ARRAY( shd_dset, a40_offset );
  a41 = (complex *)DSET_ARRAY( shd_dset, a40_offset+1 );
  a42 = (complex *)DSET_ARRAY( shd_dset, a40_offset+2 );
  a43 = (complex *)DSET_ARRAY( shd_dset, a40_offset+3 );
  a44 = (complex *)DSET_ARRAY( shd_dset, a40_offset+4 );
	       // GFA is the last sub-brick in SHD.
  shd_nbricks = DSET_NVALS( shd_dset );
  gfa = (complex *)DSET_ARRAY( shd_dset, shd_nbricks-1 );
               // Start making a new dataset with copy of old one.
  dt4_dset = EDIT_empty_copy( shd_dset );
               // edit some of its internal parameters
  i = EDIT_dset_items( dt4_dset,
    ADN_prefix, prefix, // filename prefix
    ADN_malloc_type, DATABLOCK_MEM_MALLOC, // store in memory
    ADN_datum_all, MRI_float, // atomic datum
    ADN_nvals, numSubBricks, // number of sub-bricks
    ADN_ntt, 0,
    ADN_type, ISHEAD( shd_dset ) ? // dataset type
              HEAD_FUNC_TYPE : GEN_FUNC_TYPE, 
    ADN_func_type, FUNC_BUCK_TYPE, // function type       
    ADN_none );
  if( i != 0 ) return NULL;

  THD_init_datablock_labels( dt4_dset->dblk );  
  THD_init_datablock_stataux( dt4_dset->dblk );  
  THD_init_datablock_keywords( dt4_dset->dblk );  

               // arrays of diffusion tensor elements
  nvox = DSET_NVOX( shd_dset );
  float **tensor4th = NULL;
  tensor4th = (float **)malloc( numSubBricks*sizeof( float* ) );
  if( tensor4th == NULL )
  {
    THD_delete_3dim_dataset( dt4_dset, False );
    return NULL;
  }  
  for( i=0; i<numSubBricks; ++i )
  {
    tensor4th[i] = NULL;
    tensor4th[i] = (float *)malloc( nvox * sizeof( float ) );
    if( tensor4th[i] == NULL )
    {
      THD_delete_3dim_dataset( dt4_dset, False );
      return NULL;
    }  
  }
               // calculate tensor elements from SHD coeffiecients
	       // See ref. Ozarslan 2003
#define sqrtpi sqrt(M_PI)
  for( i=0; i<nvox; ++i )
  {
               // Dxxxx
    tensor4th[0][i] = 1/(16*sqrtpi)*(4*sqrt(30)*a22[i].r - 6*sqrt(10)*a42[i].r
      + 3*sqrt(70)*a44[i].r + 8*a00[i].r - 4*sqrt(5)*a20[i].r + 9*a40[i].r);
               // Dxxxy
    tensor4th[1][i] = -sqrt(2.5)/(8*sqrtpi)*(2*sqrt(3)*a22[i].i - 3*a42[i].i
      + 3*sqrt(7)*a44[i].i);
               // Dxxxz
    tensor4th[2][i] = -sqrt(5)/(16*sqrtpi)*(2*sqrt(6)*a21[i].r - 9*a41[i].r
      + 3*sqrt(7)*a43[i].r);
               // Dxxyy
    tensor4th[3][i] = 1/(48*sqrtpi)*(-9*sqrt(70)*a44[i].r + 8*a00[i].r
      -4*sqrt(5)*a20[i].r + 9*a40[i].r);
               // Dxxzz
    tensor4th[4][i] = 1/(24*sqrtpi)*(sqrt(30)*a22[i].r + 9*sqrt(10)*a42[i].r
      + 4*a00[i].r + sqrt(5)*a20[i].r - 18*a40[i].r);
               // Dxxyz
    tensor4th[5][i] = sqrt(5)/(48*sqrtpi)*(2*sqrt(6)*a21[i].i - 9*a41[i].i
      + 9*sqrt(7)*a43[i].i);
               // Dyyyx
    tensor4th[6][i] = sqrt(2.5)/(8*sqrtpi)*(-2*sqrt(3)*a22[i].i + 3*a42[i].i
      + 3*sqrt(7)*a44[i].i);
               // Dzzzx
    tensor4th[7][i] = -sqrt(5)/(8*sqrtpi)*(sqrt(6)*a21[i].r + 6*a41[i].r);
               // Dyyxz
    tensor4th[8][i] = sqrt(5)/(48*sqrtpi)*(-2*sqrt(6)*a21[i].r + 9*a41[i].r
      + 9*sqrt(7)*a43[i].r);
               // Dzzxy
    tensor4th[9][i] = -sqrt(2.5)/(12*sqrtpi)*(sqrt(3)*a22[i].i + 9*a42[i].i);
               // Dyyyy
    tensor4th[10][i] = 1/(16*sqrtpi)*(-4*sqrt(30)*a22[i].r 
      + 6*sqrt(10)*a42[i].r + 3*sqrt(70)*a44[i].r + 8*a00[i].r 
      - 4*sqrt(5)*a20[i].r + 9*a40[i].r);
               // Dyyyz
    tensor4th[11][i] = sqrt(5)/(16*sqrtpi)*(2*sqrt(6)*a21[i].i 
      - 9*a41[i].i - 3*sqrt(7)*a43[i].i);
               // Dyyzz
    tensor4th[12][i] = 1/(24*sqrtpi)*(-sqrt(30)*a22[i].r - 9*sqrt(10)*a42[i].r
      + 4*a00[i].r + sqrt(5)*a20[i].r - 18*a40[i].r);
               // Dzzzy
    tensor4th[13][i] = sqrt(5)/(8*sqrtpi)*(sqrt(6)*a21[i].i + 6*a41[i].i);
               // Dzzzz
    tensor4th[14][i] = 1/(2*sqrtpi)*(a00[i].r + sqrt(5)*a20[i].r + 3*a40[i].r);
               // gfa has only real part
    tensor4th[15][i] = gfa[i].r;
#ifdef DT2ND
               // Dxx
    tensor4th[16][i] = 0.5/sqrtpi*a00[i].r - sqrt(5)/(4*sqrtpi)*a20[i].r
      + sqrt(15.0/8)/sqrtpi*a22[i].r;
               // Dxy
    tensor4th[17][i] = -sqrt(15.0/8)/sqrtpi*a22[i].i;
               // Dyy
    tensor4th[18][i] = 0.5/sqrtpi*a00[i].r - sqrt(5)/(4*sqrtpi)*a20[i].r
      - sqrt(15.0/8)/sqrtpi*a22[i].r;
               // Dxz
    tensor4th[19][i] = -sqrt(15.0/8)/sqrtpi*a21[i].r;
               // Dyz
    tensor4th[20][i] = sqrt(15.0/8)/sqrtpi*a21[i].i;	       
               // Dzz
    tensor4th[21][i] = 0.5/sqrtpi*a00[i].r + 0.5*sqrt(5)/sqrtpi*a20[i].r;
#endif
  }
               // put tensor elements into the new tensor dataset
	       // and label the sub-bricks
#ifndef DT2ND
  char *labels[] = { "Dxxxx", "Dxxxy", "Dxxxz", "Dxxyy", "Dxxzz", 
    "Dxxyz", "Dyyyx", "Dzzzx", "Dyyxz", "Dzzxy", "Dyyyy", "Dyyyz", 
    "Dyyzz", "Dzzzy", "Dzzzz", "GFA" };
#else
  char *labels[] = { "Dxxxx", "Dxxxy", "Dxxxz", "Dxxyy", "Dxxzz", 
    "Dxxyz", "Dyyyx", "Dzzzx", "Dyyxz", "Dzzxy", "Dyyyy", "Dyyyz", 
    "Dyyzz", "Dzzzy", "Dzzzz", "GFA", "Dxx", "Dxy", "Dyy", "Dxz",
    "Dyz", "Dzz" };
#endif
  for( i=0; i<numSubBricks; ++i )
  {
    EDIT_substitute_brick( dt4_dset, i, MRI_float, tensor4th[i] );
    EDIT_BRICK_LABEL( dt4_dset, i, labels[i] );
  }

  return dt4_dset;
}

