/**<thd_makeswdbuc.c>****************************************************

     Title: THree-D MAKE DIFfusion BUCket

   Purpose: Routine to create a 3D 'difbuc' dataset from a 3D+diffusion 
            dataset, using a user supplied function to process each 
	    time series.

      Call: THD_3dim_dataset *
                     MAKER_4D_to_typed_difbuc(THD_3dim_dataset *old_dset ,
		                              char *new_prefix , 
					      int new_datum ,
					      int nbrik , 
					      generic_func *user_func ,
					      void *user_data )

   History: I retooled RWC's thd_makefbuc.c
*/
/*-------------------------------------------------------------------------*/

#include "mrilib.h"
#include "Amalloc.h"
#include "diffuse.h"

/*------------------ macros to return workspace at exit -------------------*/

#undef  FREE_FOUT
#define FREE_FOUT                                            \
  do{ int jv ;                                               \
      if( fout != NULL ){                                    \
         for( jv=0 ; jv < nbrik*img->nbvals ; jv++ ) FREEUP(fout[jv]) ;  \
         free(fout) ;                                        \
      } } while(0)

#undef  FREE_WORKSPACE
#define FREE_WORKSPACE                              \
  do{ FREEUP(bptr) ; FREEUP(sptr) ; FREEUP(fptr) ;  \
      FREEUP(cptr) ; FREEUP(fxar) ; FREEUP(fac)  ;  \
      FREEUP(val)  ; FREE_FOUT    ;  \
    } while(0)

/*-------------------------------------------------------------------------*/

THD_3dim_dataset *MAKER_4D_to_typed_swdbuc( Widget wcancel ,
					    Widget shell ,
					    THD_3dim_dataset *old_dset ,
					    byte *mask ,
					    char *new_prefix , 
					    int new_datum ,
					    generic_func *user_func ,
					    int method , 
					    int *cancelcode ,
					    void *user_data )
{
   char *program_name = "MAKER_4D_to_typed_swdbuc";
   THD_3dim_dataset * qset ;  /* output dataset */
   Widget difmeter ;

   DIF_state *dif = (DIF_state*)user_data ;
   IMG_state *img = dif->img ;

   byte    ** bptr = NULL ;  /* one of these will be the array of */
   short   ** sptr = NULL ;  /* pointers to input dataset sub-bricks */
   float   ** fptr = NULL ;  /* (depending on input datum type) */
   complex ** cptr = NULL ;

   float   *  fxar = NULL ;  /* array loaded from input dataset */
   float   *  fac  = NULL ;  /* array of input brick scaling factors */
   float   ** fout = NULL ;  /* will be arrays of output floats */
   float   *  val  = NULL ;  /* will be array of output values */

   complex *  cxar = NULL ;  /* array loaded from input dataset */
   complex *  cfac = NULL ;  /* array of input brick scaling factors */
   complex ** cout = NULL ;  /* will be arrays of output complexs 01-Sep-02 */
   complex *  cval = NULL ;  /* will be array of output values 01-Sep-02 */

   int ii, old_datum, use_fac, nvox, iv, perc, i, j, jj;
   int nbrik, nvals, ndat ;
   register int kk ;
   int nbad = 0 ;
   int badfinish = 0 ;

   int debug  = 1;
   int debug2 = 1;
   int debug3 = 0;

   /*----------------------------------------------------------*/
   /*----- Check inputs to see if they are reasonable-ish -----*/

   if( ! ISVALID_3DIM_DATASET(old_dset) ) {
     printf("%s: Bad input dataset.  Returning ...\n",program_name);
     return NULL ;
   }

   if( new_datum >= 0         &&
       new_datum != MRI_byte  &&
       new_datum != MRI_short &&
       new_datum != MRI_float &&
       new_datum != MRI_complex   ) {
     printf("%s: Invalid datatype.  Returning ...\n",program_name);
     return NULL ;
   }

   if( user_func == NULL ) return NULL ;

   /*--------- set up pointers to each sub-brick in the input dataset ---------*/

   old_datum = DSET_BRICK_TYPE( old_dset , 0 ) ; /* get old dataset datum */
   nvals     = DSET_NVALS(old_dset);             
   if( nvals < 2 ) {
     printf("%s: NVALS=%i too small.  Returning ...\n",program_name,nvals);
     return NULL ;
   }

   if( new_datum < 0 ) new_datum = old_datum ;   /* output datum = input */

   DSET_load( old_dset ) ;  /* must be in memory before we get pointers to it */

   kk = THD_count_databricks( old_dset->dblk ) ;  /* check if it was */
   if( kk < DSET_NVALS(old_dset) ){               /* loaded correctly */
      DSET_unload( old_dset ) ;
      return NULL ;
   }

   /* Processing method */

   switch(method) {

     /* SWD */

     case METH_SWD_SWD:
       nbrik = dif->shd->calc->nlm * dif->shd->calc->nrad ;
       ndat  = img->ndif * img->nbvals ;
       break;

   default:
     printf("%s: Unknown processing method %i!\n\a",program_name,method);
     printf("Returning...\n");
     return NULL ;
     break;
   }

   if( nbrik <= 0 ) {
     printf("%s: nbrik (=%i) <=0.  Returning ...\n",program_name,nbrik);
     return NULL ;
   }

   /* lrf check section */
   if(debug2) {
     printf("%s:  Number of  input data bricks (nvals) = %i\n",program_name,nvals);
     printf("%s:  Number of output data bricks (nbrik) = %i\n",program_name,nbrik);
     printf("%s:  ndat = %i\n",program_name,ndat);
     printf("%s:  nb0 = %i\n",program_name,img->nb0);
     printf("%s:  nvert = %i\n",program_name,img->dgrad->nvert);
     printf("%s:  nrad = %i\n",program_name,dif->shd->calc->nrad);
     printf("%s:  ndif = %i\n",program_name,img->ndif);
     printf("%s:  nbvals = %i\n",program_name,img->nbvals);
     fflush(stdout);
   }
   /* end lrf check section */

   switch( old_datum ){  /* pointer type depends on input datum type */

      default:                      /** don't know what to do **/
         DSET_unload( old_dset ) ;
         return NULL ;

      /** create array of pointers into old dataset sub-bricks **/

      /*--------- input is bytes ----------*/
      /* voxel #i at sub-brick #k is bptr[k][i] */
      /* for i=0..nvox-1 and k=0..nvals-1.  */

      case MRI_byte:
         bptr = (byte **) malloc( sizeof(byte *) * nvals ) ;
         if( bptr == NULL ) return NULL ;
         for( kk=0 ; kk < nvals ; kk++ )
            bptr[kk] = (byte *) DSET_ARRAY(old_dset,kk) ;
      break ;

      /*--------- input is shorts ---------*/
      /* voxel #i at time #k is sptr[k][i] */
      /* for i=0..nvox-1 and k=0..nvals-1.  */

      case MRI_short:
         sptr = (short **) malloc( sizeof(short *) * nvals ) ;
         if( sptr == NULL ) return NULL ;
         for( kk=0 ; kk < nvals ; kk++ )
            sptr[kk] = (short *) DSET_ARRAY(old_dset,kk) ;
      break ;

      /*--------- input is floats ---------*/
      /* voxel #i at time #k is fptr[k][i] */
      /* for i=0..nvox-1 and k=0..nvals-1.  */

      case MRI_float:
         fptr = (float **) malloc( sizeof(float *) * nvals ) ;
         if( fptr == NULL ) return NULL ;
         for( kk=0 ; kk < nvals ; kk++ )
            fptr[kk] = (float *) DSET_ARRAY(old_dset,kk) ;
      break ;

      /*--------- input is complex ---------*/
      /* voxel #i at time #k is cptr[k][i]  */
      /* for i=0..nvox-1 and k=0..nvals-1.   */

      case MRI_complex:
         cptr = (complex **) malloc( sizeof(complex *) * nvals ) ;
         if( cptr == NULL ) return NULL ;
         for( kk=0 ; kk < nvals ; kk++ )
            cptr[kk] = (complex *) DSET_ARRAY(old_dset,kk) ;
      break ;

   } /* end of switch on input type */

   if(debug) printf("%s: allocating\n",program_name);

   /*---- allocate space for 1 voxel timeseries ----*/

   if(old_datum==MRI_complex) { 
     cxar = (complex *) malloc( sizeof(complex) * nvals ) ;
     if( cxar == NULL ){ FREE_WORKSPACE ; return NULL ; }
   } else {
     fxar = (float *) malloc( sizeof(float) * nvals ) ;
     if( fxar == NULL ){ FREE_WORKSPACE ; return NULL ; }
   }

   /*--- get scaling factors for input sub-bricks ---*/

   if(debug) printf("%s: scaling\n",program_name);

   fac = (float *) malloc( sizeof(float) * nvals ) ;   /* factors */
   if( fac == NULL ){ FREE_WORKSPACE ; return NULL ; }

   use_fac = 0 ;
   for( kk=0 ; kk < nvals ; kk++ ){
     fac[kk] = DSET_BRICK_FACTOR(old_dset,kk) ;
     if( fac[kk] != 0.0 ) use_fac++ ;
     else                 fac[kk] = 1.0 ;
   }
   if( !use_fac ) FREEUP(fac) ;

   /*---------------------- make a new dataset ----------------------*/

   if(debug) printf("%s: making new dataset: new_prefix = %s\n",
		    program_name,new_prefix);

   qset = EDIT_empty_copy( old_dset ) ; /* start with copy of old one */
   
   /*-- edit some of its internal parameters --*/

   ii = EDIT_dset_items(qset ,
			ADN_prefix      , new_prefix ,           /* filename prefix */
			ADN_malloc_type , DATABLOCK_MEM_MALLOC , /* store in memory */
			ADN_datum_all   , new_datum ,            /* atomic datum */
			ADN_nvals       , nbrik,                 /* # sub-bricks */
			ADN_ntt         , 0 ,                    /* # time points */
			ADN_type        , ISHEAD(old_dset)       /* dataset type */
			                  ? HEAD_FUNC_TYPE
			                  : GEN_FUNC_TYPE ,
			ADN_func_type   , FUNC_BUCK_TYPE ,        /* function type */
			ADN_none ) ;
   
   if(debug3) { MCW_idcode *idcold, *idcnew;
           idcnew=DSET_IDCODE(qset);
	   idcold=DSET_IDCODE(old_dset);

	   printf("%s: making new dataset: \n"
		  "new_prefix = %s\n"
		  "DSET_FILECODE(qset) = %s\n"
		  "DSET_FILECODE(old_dset) = %s\n"
		  "DSET_BRIKNAME(qset) = %s\n"
		  "DSET_BRIKNAME(old_dset) = %s\n"
		  "DSET_IDCODE(qset) = %s\n"
		  "DSET_IDCODE(old_dset) = %s\n",
		  program_name,
		  new_prefix,
		  DSET_FILECODE(qset),DSET_FILECODE(old_dset),
		  DSET_BRIKNAME(qset),DSET_BRIKNAME(old_dset),
		  idcnew->str,  idcold->str
		  );
   }

   if( ii != 0 ){
      THD_delete_3dim_dataset( qset , False ) ;  /* some error above */
      FREE_WORKSPACE ; return NULL ;
   }

   THD_init_datablock_labels(   qset->dblk ) ;
   THD_init_datablock_keywords( qset->dblk ) ;
   THD_init_datablock_stataux(  qset->dblk ) ;

   /*------ make floating point output bricks
            (only at the end will scale to byte or shorts) ------*/

   if(debug) printf("%s: making output bricks \n",program_name);

   nvox = DSET_NVOX(old_dset);

   if(new_datum==MRI_complex) {

     cout = (complex **) malloc( sizeof(complex *) * nbrik ) ;

     if( cout == NULL ){
       THD_delete_3dim_dataset( qset , False ) ;
       FREE_WORKSPACE ; return NULL ;
     }

     for( iv=0 ; iv < nbrik ; iv++ ) cout[iv] = NULL ;
     for( iv=0 ; iv < nbrik ; iv++ ){
       cout[iv] = (complex *) malloc( sizeof(complex) * nvox ) ;
       if( cout[iv] == NULL ){
         THD_delete_3dim_dataset( qset , False ) ;
         FREE_WORKSPACE ; return NULL ;
       }
     }

     /*-- complexing point storage for output from 1 voxel --*/

     cval = (complex *) malloc( sizeof(complex) * nbrik ) ;
     if( cval == NULL ){
       THD_delete_3dim_dataset( qset , False ) ;
       FREE_WORKSPACE ; return NULL ;
     }

   } else {

     fout = (float **) malloc( sizeof(float *) * nbrik) ;

     if( fout == NULL ){
       THD_delete_3dim_dataset( qset , False ) ;
       FREE_WORKSPACE ; return NULL ;
     }

     for( iv=0 ; iv < nbrik ; iv++ ) fout[iv] = NULL ;
     for( iv=0 ; iv < nbrik ; iv++ ){
       fout[iv] = (float *) malloc( sizeof(float) * nvox ) ;
       if( fout[iv] == NULL ){
         THD_delete_3dim_dataset( qset , False ) ;
         FREE_WORKSPACE ; return NULL ;
       }
     }

     /*-- floating point storage for output from 1 voxel --*/

     val = (float *) malloc( sizeof(float) * nbrik ) ;
     if( val == NULL ){
       THD_delete_3dim_dataset( qset , False ) ;
       FREE_WORKSPACE ; return NULL ;
     }

   }

   /*----------------------------------------------------*/
   /*----- Setup has ended.  Now do some real work. -----*/

   if(debug) printf("%s: setup ended - doing work... \n",program_name);

   XtManageChild( wcancel ) ; AFNI_add_interruptable( wcancel ) ;
   anisokill = 0 ;

   difmeter = MCW_popup_meter( shell , METER_TOP_WIDE ) ;

   /* start notification */

   if(debug) printf("%s: starting notification ... ",program_name);

#if 0
   user_func( method, NULL , NULL, user_data , nbrik , NULL , NULL ) ;
#else
   AFNI_CALL_VOID_7ARG( user_func, int, method, float*, NULL, complex*, NULL,  
			void *,user_data, int, nbrik, float *, NULL, complex *, NULL );
#endif

   if(debug) printf("done\n");

   /***** loop over voxels *****/

   for( ii=0 ; ii < nvox ; ii++  ){  /* 1 time series at a time */

     if(mask[ii]) {

      /*** load data from input dataset, depending on type ***/

      switch( old_datum ){

         /*** input = bytes ***/

         case MRI_byte:
            for( kk=0 ; kk < ndat ; kk++ ) fxar[kk] = bptr[kk][ii] ;
         break ;

         /*** input = shorts ***/

         case MRI_short:
            for( kk=0 ; kk < ndat ; kk++ ) fxar[kk] = sptr[kk][ii] ;
         break ;

         /*** input = floats ***/

         case MRI_float:
            for( kk=0 ; kk < ndat ; kk++ ) fxar[kk] = fptr[kk][ii] ;
         break ;

         /*** input = complex ***/

         case MRI_complex:
	   for( kk=0 ; kk < ndat ; kk++ )  cxar[kk] = cptr[kk][ii] ;

         break ;

      } /* end of switch over input type */

      /*** scale? ***/

      if(debug3) printf("%s: MRI_complex: scaling....\n",program_name);

      if( use_fac ) {
	if(old_datum==MRI_complex) { 
	  for( kk=0 ; kk < ndat ; kk++ ) {
	    cxar[kk].r *= fac[kk] ;
	    cxar[kk].i *= fac[kk] ;
	  }
	} else {
	  for( kk=0 ; kk < ndat ; kk++ ) fxar[kk] *= fac[kk] ;
	}
      }

      /*** compute output ***/

      if(debug3) printf("%s: computing output \n",program_name);

      if(new_datum==MRI_complex) { /* eg, dap in, shd out */

	for( iv=0 ; iv < nbrik ; iv++ ) {
	  cval[iv].r = 0.0 ;
	  cval[iv].i = 0.0 ;
	}

#if 0
	user_func( method, fxar, NULL, user_data, nbrik, NULL, cval );
#else
        AFNI_CALL_VOID_7ARG( user_func, int, method, float*, fxar, complex*, NULL,  
			     void *,user_data, int, nbrik, float *, NULL, complex *, cval );
#endif

	for( iv=0 ; iv < nbrik ; iv++ ) cout[iv][ii] = cval[iv] ;

	if(0) PRINT_ZV(cval,nbrik,"cval");

      } else if(old_datum==MRI_complex) { /* eg, shd in, fib out */

	for( iv=0 ; iv < nbrik ; iv++ ) val[iv] = 0.0 ;

#if 0
	user_func( method, NULL, cxar, user_data, nbrik, val, NULL );
#else
        AFNI_CALL_VOID_7ARG( user_func, int, method, float*, NULL, complex*, cxar,  
			     void *,user_data, int, nbrik, float *, val, complex *, NULL );
#endif

	for( iv=0 ; iv < nbrik ; iv++ ) fout[iv][ii] = val[iv] ;
	if(0) PRINT_FV(val,nbrik,"val");

      } else {			/* eg, dti in, fai out */

	for( iv=0 ; iv < nbrik ; iv++ ) val[iv] = 0.0 ;

#if 0 
	user_func( method, fxar, NULL, user_data, nbrik, val, NULL );
#else
        AFNI_CALL_VOID_7ARG( user_func, int, method, float*, fxar, complex*,NULL ,  
			     void *,user_data, int, nbrik, float *, val, complex *, NULL );
#endif

        if(method == METH_SWD_SWD) {
	
	  for( iv=0 ; iv < nbrik ; iv++ ) fout[iv][ii] = val[iv] ; 

	} else {
      

	}
      }

     } else {

      if(new_datum==MRI_complex) {
	for( iv=0 ; iv < nbrik ; iv++ ) {
	  cout[iv][ii].r = 0. ;
	  cout[iv][ii].i = 0. ;
	}
      } else {
	for( iv=0 ; iv < nbrik ; iv++ ) fout[iv][ii] = 0. ;
      }

     }

     AFNI_process_interrupts(wcancel) ;
     if( anisokill ) {
       badfinish = 1;
       break ;
     }

     perc = (100 * ii) / nvox ;
     MCW_set_meter( difmeter , perc ) ;

   } /* end of outer loop over 1 voxels at a time */


   DSET_unload( old_dset ) ;  /* don't need this no more */

   /* end notification */

#if 0
   user_func( method, NULL , NULL, user_data , nbrik , NULL, NULL ) ;
#else
   AFNI_CALL_VOID_7ARG( user_func, int, method, float*, NULL, complex*,NULL,  
			void *,user_data, int, nbrik, float *, NULL, complex *, NULL );
#endif

   printf("%s: +++++++++++++++ meter... ",program_name);fflush(stdout);
   MCW_popdown_meter( difmeter ) ;
   XtUnmanageChild( wcancel ) ; 
   AFNI_add_interruptable(NULL) ;
   printf("done\n");fflush(stdout);

   /* Handle cancel operation (prevent writing out bad dset) */

   if (badfinish) {
     *cancelcode = 1;
     FREE_WORKSPACE ;
     return NULL ;
   } else {
     *cancelcode = 0;
   }

   /*---- Count and correct float errors ----*/

   if(new_datum==MRI_complex) {
     for( iv=0 ; iv < nbrik ; iv++ )
       nbad += thd_complexscan( nvox, cout[iv] ) ;
   } else {
     for( iv=0 ; iv < nbrik ; iv++ )
       nbad += thd_floatscan( nvox, fout[iv] ) ;
   }

   if( nbad > 0 )
      fprintf(stderr,
              "++ Warning: %d bad floats computed in %s\n\a",
              nbad,program_name ) ;

   /*------------------------------------------------------------------------*/
   /*------- The output is now in fout[iv][ii], iv=0..nbrik-1, ii=0..nvox-1.
             We must now put this into the output dataset. ------------------*/

   switch( new_datum ){

      case MRI_complex:
	 if(debug) printf("%s: substituting %i complex briks ... \n",program_name,nbrik);
         for( iv=0 ; iv < nbrik ; iv++ ){
            EDIT_substitute_brick( qset , iv , MRI_complex , cout[iv] ) ;
            cout[iv] = NULL ;  /* so it won't be freed later */
         }
      break ;

      /*** output is floats is the simplest:
           we just have to attach the fout brick to the dataset ***/

      case MRI_float:
	 if(debug) printf("%s: substituting %i float briks ... \n",program_name,nbrik);
         for( iv=0 ; iv < nbrik ; iv++ ){
            EDIT_substitute_brick( qset , iv , MRI_float , fout[iv] ) ;
            fout[iv] = NULL ;  /* so it won't be freed later */
         }
      break ;

      /*** output is shorts:
           we have to create scaled sub-bricks from fout ***/

      case MRI_short:{
         short * bout ;
         float sfac ;
	 if(debug) printf("%s: substituting %i short briks ... \n",program_name,nbrik);
         for( iv=0 ; iv < nbrik  ; iv++ ){
            bout = (short *) malloc( sizeof(short) * nvox ) ;
            if( bout == NULL ){
               fprintf(stderr,
                "\nFinal malloc error in MAKER_4D_to_difbuc - is memory exhausted?\n\a");
               exit(1) ;
            }
            sfac = MCW_vol_amax( nvox,1,1 , MRI_float , fout[iv] ) ;
            if( sfac > 0.0 ){
               sfac = 32767.0 / sfac ;
               EDIT_coerce_scale_type( nvox,sfac ,
                                       MRI_float,fout[iv] , MRI_short,bout ) ;
               sfac = 1.0 / sfac ;
            }
            val[iv] = sfac ;
            EDIT_substitute_brick( qset , iv , MRI_short , bout ) ;
         }
         EDIT_dset_items( qset , ADN_brick_fac , val , ADN_none ) ;
      }
      break ;

      /*** output is bytes (byte = unsigned char)
           we have to create a scaled sub-brick from fout ***/

      case MRI_byte:{
         byte * bout ;
         float sfac ;

         for( iv=0 ; iv < nbrik ; iv++ ){
            bout = (byte *) malloc( sizeof(byte) * nvox ) ;
            if( bout == NULL ){
               fprintf(stderr,
                "\nFinal malloc error in %s - is memory exhausted?\n\a",program_name);
               exit(1) ;
            }
            sfac = MCW_vol_amax( nvox,1,1 , MRI_float , fout[iv] ) ;
            if( sfac > 0.0 ){
               sfac = 255.0 / sfac ;
               EDIT_coerce_scale_type( nvox,sfac ,
                                       MRI_float,fout[iv] , MRI_byte,bout ) ;
               sfac = 1.0 / sfac ;
            }
            val[iv] = sfac ;
            EDIT_substitute_brick( qset , iv , MRI_byte , bout ) ;
         }
         EDIT_dset_items( qset , ADN_brick_fac , val , ADN_none ) ;
      }
      break ;

   } /* end of switch on output data type */

   /*  Labeling of sub-bricks */

   switch( method ){

      default:
      break ;

      case METH_SWD_SWD:{
	if(debug) printf("\athd_makedifbuc: No labeling for SWD yet!\n");
      }
      break ;

      case METH_SWD_GFA:{

	char **label_list = cm_alloc(img->nbvals,128) ;
	int nlabel = img->nbvals ;

	for( int ib=0 ; ib < img->nbvals ; ib++ ) {
	  sprintf(label_list[ib],"GFA: b = %f",img->bvals[ib]);
	  if(debug2) printf("label_list[%i] = %s\n",ib,label_list[ib]);
	}

	for( int ib=0 ; ib < nbrik ; ib++ ) {
	  if(debug2) printf("Labeling SWD fbuc %i: %s\n",ib,label_list[ib%nlabel]);
	  EDIT_BRICK_LABEL(qset,ib,label_list[ib%nlabel]);
	}
      }
      break ;

      case METH_SWD_GRA:{

	char **label_list = cm_alloc(img->nbvals,128) ;
	int nlabel = img->nbvals ;

	for( int ib=0 ; ib < img->nbvals ; ib++ ) {
	  sprintf(label_list[ib],"GRA: b = %f",img->bvals[ib]);
	  if(debug2) printf("label_list[%i] = %s\n",ib,label_list[ib]);
	}

	for( int ib=0 ; ib < nbrik ; ib++ ) {
	  if(debug2) printf("Labeling SWD fbuc %i: %s\n",ib,label_list[ib%nlabel]);
	  EDIT_BRICK_LABEL(qset,ib,label_list[ib%nlabel]);
	}
      }
      break ;

   }

   /*-------------- Cleanup and go home ----------------*/

   FREE_WORKSPACE ;
   return qset ;
}

