/***********************************************************************
 * File:        datainit.c
 * Programmer:  Ning Kang @CS.UKY
 * Version of:  11/24/2003
 * ---------------------------------------------------------------------
 * Description:
 * - Serial version.
 * - This file contains the following functions:
 *   Diff_Kernel_ kernelData_new()
 *   void kernelData_free()
 *   ILU_ iluData_new()
 *   void iluData_free()
 *   DAE_ daeData_new()
 *   void daeData_free()
 *   DTI_ dtiData_new()
 *   void dtiData_free()
 ***********************************************************************/
 

#include "dst.h"

 
/***********************************************************************
 * Allocate memory for diffusion kernel data structure 
 * and do initialization.
 ***********************************************************************/
Diff_Kernel_ kernelData_new( void *argsTable, int neq, int nnz, int nx, 
    int ny, int nz, int diffSize )
{
    Diff_Kernel_ kernel;
    Args_Table_ args;
    int i;
    
    args = (Args_Table_)argsTable;
                                // allocate memory for the user data structure
    kernel = (Diff_Kernel_ )malloc( sizeof ( *kernel ) );
                                // assign parameters for grid geometry
    kernel->nx = nx;
    kernel->ny = ny;
    kernel->nz = nz;
    kernel->neq = neq;
    kernel->nnz = nnz;
    kernel->dx = args->dx;
    kernel->dy = args->dy;
    kernel->dz = args->dz;
    kernel->cThreshold = 0;
                                // allocate memory for coefficient matrix
				// stored in CSR format.
    kernel->aa = (realtype *)malloc( nnz * sizeof( realtype ) );
    kernel->ja = (int *)malloc( nnz * sizeof( int ) );
    kernel->ia = (int *)malloc( (neq+1) * sizeof( int ) );
	                        // allocate memory for scale factor
    kernel->sf = (realtype *)malloc( neq * sizeof( realtype ) );
	                        // allocate memory for work array
    kernel->diffval = (realtype *)malloc( neq * sizeof( realtype ) );

    kernel->voxel = (Kernel_Voxel_ *)malloc( neq * sizeof( Kernel_Voxel_ ) );
    for( i = 0; i < neq; ++ i )
        kernel->voxel[i] = (Kernel_Voxel_)malloc( 
	    sizeof( *(kernel->voxel[i]) ) );
    
    memset( kernel->aa, 0, nnz * sizeof( realtype ) );
    memset( kernel->ja, 0, nnz * sizeof( int ) ); 
    memset( kernel->ia, 0, (neq+1) * sizeof( int ) ); 
    memset( kernel->sf, 0, neq * sizeof( realtype ) ); 
    memset( kernel->diffval, 0, neq * sizeof( realtype ) ); 

//    kernel->median = (int)( neq *(1-DIFFVOL) );
    kernel->median = neq - diffSize;
    if( kernel->median < 0 || kernel->median >= neq )
    {
        fprintf( stderr, "DIFFVOL should be in (0, 1]\n" );
	return NULL;
    }
    
    kernel->seedI = nx/2;
    kernel->seedJ = ny/2;
    kernel->seedK = nz/2;
    kernel->seedID = nodeIndex( nx/2+1, ny/2+1, nz/2+1, nx, nx*ny ) - 1; 
    return kernel;
}

/***********************************************************************
 * Release the memory allocated for coefficient matrix data structures.
 ***********************************************************************/
void kernelData_free( void *kernelData )
{
    Diff_Kernel_ kernel;
    int i;

    kernel = (Diff_Kernel_)kernelData;
    for( i = 0; i < kernel->neq; ++ i )
        free( kernel->voxel[i] );
    free( kernel->voxel );
    free( kernel->aa );
    free( kernel->ja );
    free( kernel->ia );
    free( kernel->sf );
    free( kernel->diffval );
    free( kernel );
}


/***********************************************************************
 * Allocate memory for ILU data structure 
 * and do initialization.
 ***********************************************************************/
ILU_ iluData_new( int neq, int nnz, int iwk )
{
    ILU_ ilu;
    
    ilu = (ILU_)malloc( sizeof( *ilu ) );
    ilu->alu = (realtype *)malloc( nnz * sizeof( realtype ) );
    ilu->jlu = (int *)malloc( nnz * sizeof( int ) );
    ilu->ju = (int *)malloc( neq * sizeof( int ) );
/*    
    memset( ilu->alu, 0, nnz * sizeof( realtype ) );
    memset( ilu->jlu, 0, nnz * sizeof( int ) );
    memset( ilu->ju, 0, neq * sizeof( int ) );
*/
  
#ifdef ILU0
    ilu->ierr = 0;
    ilu->jr = (int *)malloc( neq * sizeof( int ) );
    memset( ilu->jr, 0, neq * sizeof( int ) ); 
#endif

#ifdef ILUT
    ilu->lfil = LFIL;
    ilu->tol = ILUT_TOL;
    ilu->iwk = iwk;
    ilu->ierr = 0;
    
    ilu->jr = (int *)malloc( neq * sizeof( int ) );
    ilu->jwu = (int *)malloc( neq * sizeof( int ) );
    ilu->jwl = (int *)malloc( neq * sizeof( int ) ); 
    ilu->wu = (realtype *)malloc( ( neq+1 ) * sizeof(realtype) );
    ilu->wl = (realtype *)malloc( neq * sizeof(realtype) );
	
    memset( ilu->jr, 0, neq * sizeof(int) );
    memset( ilu->jwu, 0, neq * sizeof(int) );
    memset( ilu->jwl, 0, neq * sizeof(int) );
    memset( ilu->wu, 0, ( neq+1 ) * sizeof(realtype) );
    memset( ilu->wl, 0, neq * sizeof(realtype) );
#endif  

    return ilu; 
}


/***********************************************************************
 * Release the memory allocated for ILU data structures.
 ***********************************************************************/
void iluData_free( void *iluData )
{
    ILU_ ilu;
    
    ilu = (ILU_)iluData;
    free( ilu->alu );
    free( ilu->jlu );
    free( ilu->ju );
    
#ifdef ILU0
    free( ilu->jr );
#endif

#ifdef ILUT
    free( ilu->jr );
    free( ilu->jwu );
    free( ilu->jwl );
    free( ilu->wu );
    free( ilu->wl );
#endif

    free( ilu );    
}


/***********************************************************************
 * Allocate memory for DAE data structure 
 * and do initialization.
 ***********************************************************************/
DAE_ daeData_new( void *argsTable )
{
    DAE_ dae;
    Args_Table_ args = (Args_Table_)argsTable;
    int neq = NEQ;
    float tout = args->diffIntegTime;
    float hinit = tout;
    
    dae = (DAE_)malloc( sizeof( *dae ) );
                                // Initialize serial machine environment
    dae->machEnv = M_EnvInit_Serial( neq );  
                                // Allocate N-vectors and do initialization 
    dae->yVector = N_VNew_Serial( neq, dae->machEnv ); 
    dae->ypVector = N_VNew_Serial( neq, dae->machEnv );
    dae->rVector = N_VNew_Serial( neq, dae->machEnv );
    dae->constraintsVector = N_VNew_Serial( neq, dae->machEnv );
    dae->idVector = N_VNew_Serial( neq, dae->machEnv );
                                // Set values for various parameters
    dae->t0 = 0.0;
    dae->t = 0.0;
    dae->tout = tout;
    dae->itol = I_TOL;
    dae->rtol = R_TOL;
    dae->atol = A_TOL;
    dae->itask = I_TASK;
    dae->epliFactor = EPLI_FAC;
    dae->maxKDim = MAX_L;
    dae->maxRestart = MAX_RS;
    dae->optIn = TRUE;
    dae->initPeak = SEED_PEAK;
    dae->loop = 0;

    dae->iopt = (long int *)malloc( OPT_SIZE * sizeof(long int) );
    dae->ropt = (realtype *)malloc( OPT_SIZE * sizeof(realtype) );
    memset( dae->iopt, 0, OPT_SIZE * sizeof( long int ) );
    memset( dae->ropt, 0, OPT_SIZE * sizeof( realtype ) );
    dae->iopt[MAXORD] = MAX_ORD;
    dae->iopt[MXSTEP] = MAX_STEP;
    dae->ropt[HMAX] = H_MAX;
    dae->ropt[HINIT] = hinit;
    dae->ropt[NCONFAC] = N_CONFAC;

    dae->mem = NULL;
    
    return dae;
}
               // for 4th-order DT
DAE_ daeData_new_DT4( void *argsTable )
{
    DAE_ dae;
    Args_Table_ args = (Args_Table_)argsTable;
    int neq = NEQ_DT4;
    float tout = args->dt4_integTime;
    float hinit = tout;
    float hmax = tout;
    float nconfac = 0.2,//0.5,
          eplifac = 0.01,//0.1,
	     rtol = 0.001,//0.005,
	     atol = 0.001;//0.005,
    int      maxl = 10,//20,
            maxrs = 10;//20;
    
    dae = (DAE_)malloc( sizeof( *dae ) );
                                // Initialize serial machine environment
    dae->machEnv = M_EnvInit_Serial( neq );  
                                // Allocate N-vectors and do initialization 
    dae->yVector = N_VNew_Serial( neq, dae->machEnv ); 
    dae->ypVector = N_VNew_Serial( neq, dae->machEnv );
    dae->rVector = N_VNew_Serial( neq, dae->machEnv );
    dae->constraintsVector = N_VNew_Serial( neq, dae->machEnv );
    dae->idVector = N_VNew_Serial( neq, dae->machEnv );
                                // Set values for various parameters
    dae->t0 = 0.0;
    dae->t = 0.0;
    dae->tout = tout;
    dae->itol = I_TOL;
    dae->rtol = rtol;//R_TOL;
    dae->atol = atol;//A_TOL;
    dae->itask = I_TASK;
    dae->epliFactor = eplifac;//EPLI_FAC;
    dae->maxKDim = maxl;//MAX_L;
    dae->maxRestart = maxrs;//MAX_RS;
    dae->optIn = TRUE;
    dae->initPeak = SEED_PEAK;
    dae->loop = 0;

    dae->iopt = (long int *)malloc( OPT_SIZE * sizeof(long int) );
    dae->ropt = (realtype *)malloc( OPT_SIZE * sizeof(realtype) );
    memset( dae->iopt, 0, OPT_SIZE * sizeof( long int ) );
    memset( dae->ropt, 0, OPT_SIZE * sizeof( realtype ) );
    dae->iopt[MAXORD] = MAX_ORD;
    dae->iopt[MXSTEP] = MAX_STEP;
    dae->ropt[HMAX] = hmax;//H_MAX;
    dae->ropt[HINIT] = hinit;//H_INIT;
    dae->ropt[NCONFAC] = nconfac;//N_CONFAC;

    dae->mem = NULL;
    
    return dae;
}


/***********************************************************************
 * Release the memory allocated for DAE data structures.
 ***********************************************************************/
void daeData_free( void *daeData )
{
    DAE_ dae;
    
    dae = (DAE_)daeData;
    
    IDAFree( dae->mem );
    N_VFree_Serial( dae->yVector );
    N_VFree_Serial( dae->ypVector );
    N_VFree_Serial( dae->constraintsVector );
    N_VFree_Serial( dae->idVector );
    N_VFree_Serial( dae->rVector );
    free( dae->iopt );
    free( dae->ropt );

    free( dae );
}


/***********************************************************************
 * Allocate memory for dti data structure 
 * and do initialization.
 ***********************************************************************/
DTI_ dtiData_new( void *argsTable )
{
    Args_Table_ args;
    DTI_ dti;
    int i, nvox;
    
    args = (Args_Table_)argsTable;
    dti = (DTI_)malloc( sizeof( *dti ) );
    nvox = args->nvox; // number of total voxels
    
    dti->num = nvox;
    dti->nx = args->nx;
    dti->ny = args->ny;
    dti->nz = args->nz;
    dti->dx = args->dx;
    dti->dy = args->dy;
    dti->dz = args->dz;
    dti->dtxyzNum = DTXYZ_NUM;
                   // tracking thresholds
    dti->tensorFA = args->tensorFA;
    dti->distRatio = args->distRatio;
    dti->ipVectors = (float)cos( M_PI/180*args->ipVectors );
    dti->ipSeedVecNode = (float)cos( M_PI/180*args->ipSeedVecNode );

    dti->voxel = (Voxel_ *)malloc( nvox * sizeof( Voxel_ ) );
    for( i = 0; i < nvox; ++ i )
        dti->voxel[i] = (Voxel_)malloc( sizeof( *(dti->voxel[i]) ) );
    
    for( i = 0; i < DTXYZ_NUM; ++ i )
        dti->dtxyz[i] = args->dt[i];
 
    dti->fa = args->fa;
    dti->md = args->md;
    dti->b0 = args->dwi_b0;
    dti->b0_lo = args->b0_lo;
    dti->b0_hi = args->b0_hi;
    for( i = 0; i < 3; ++ i ) dti->eigvec[i] = args->eigvec[i];

    return dti;
}

               // for 4th-order DT
DTI_ dtiData_new_DT4( void *argsTable )
{
    Args_Table_ args;
    DTI_ dti;
    int i, nvox;
    
    args = (Args_Table_)argsTable;
    dti = (DTI_)malloc( sizeof( *dti ) );
    nvox = args->nvox; // number of total voxels
    
    dti->num = nvox;
    dti->nx = args->nx;
    dti->ny = args->ny;
    dti->nz = args->nz;
    dti->dx = args->dx;
    dti->dy = args->dy;
    dti->dz = args->dz;
    dti->dtxyzNum = DTXYZ_NUM;
                   // tracking thresholds
    dti->dt4_GFA = args->dt4_GFA;
    dti->dt4_distRatio = args->dt4_distRatio;
    dti->dt4_ipVectors = (float)cos( M_PI/180*args->dt4_transAngle );

    dti->voxel = (Voxel_ *)malloc( nvox * sizeof( Voxel_ ) );
    for( i = 0; i < nvox; ++ i )
        dti->voxel[i] = (Voxel_)malloc( sizeof( *(dti->voxel[i]) ) );
    
    for( i = 0; i < DTXYZ_NUM; ++ i )
        dti->dtxyz[i] = args->dt[i];
    for( i = 0; i < DT4_ELEM; ++ i )
        dti->dt4Elem[i] = args->dt4[i];
 
    dti->dt4_gfa = args->fa;

    return dti;
}


/***********************************************************************
 * Release the memory allocated for DTI data structures.
 ***********************************************************************/
void dtiData_free( void *dtiData )
{
    DTI_ dti;
    Voxel_ *voxel;
    int id;
    
    dti = (DTI_)dtiData;
    voxel = dti->voxel;
        
    for( id = 0; id < dti->num; ++ id )
        free( voxel[id] );
    free( voxel );
    free( dti );
}

//***********************************************************************
// Do initialization for each voxel 
//***********************************************************************
void initDTIvoxel( void *dtiData )
{
    DTI_ dti = (DTI_)dtiData;
    Voxel_ *voxel;
    int nx, ny, nz, nxy;
    int id, i, j, k;
    
    nx = dti->nx;
    ny = dti->ny;
    nz = dti->nz;
    nxy = nx*ny;
    voxel = dti->voxel;
				// initialize the voxel structure
    for( i = 0; i < nx; ++ i )
    for( j = 0; j < ny; ++ j )
    for( k = 0; k < nz; ++ k )
    {
        id = nodeIndex( i+1, j+1, k+1, nx, nxy ) - 1;
	voxel[id]->id = id;
	voxel[id]->i = i;
	voxel[id]->j = j;
	voxel[id]->k = k;
	voxel[id]->color = WHITE;
	voxel[id]->next = NULL;
	voxel[id]->tail = NULL;
	voxel[id]->parent = NULL;
	voxel[id]->child = NULL;
                   // extract the diffusion tensor data
                   // which is stored in column-wise.
//        for( component=0; component<6; ++component )
//            DTXYZ( id, component ) = args->dt[component][id];
//            DTXYZ( id, component ) = args->dt[component][i + j*nx + k*nxy];
    }
}

/***********************************************************************
 * Reinitialize each voxel in the grid.
 ***********************************************************************/
void resetDTIvoxel( void *dtiData )
{
    DTI_ dti;
    Voxel_ *voxel;
    int id;
    
    dti = (DTI_)dtiData;
    voxel = dti->voxel;
			/* reinitialize the voxel data structure */
    for( id = 0; id < dti->num; ++ id )
    {
	voxel[id]->color = WHITE;
	voxel[id]->next = NULL;
	voxel[id]->tail = NULL;
	voxel[id]->parent = NULL;
	voxel[id]->child = NULL;
    }
}


















