/**********************************************************************
 * This is the command-line verison for DST fiber tracking in the 
 * diffusion plugin.
 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "afni.h"
#include "dst_interface.h"

static void help( char *command );
static int setOptions( int argc, char *argv[] );
static int **DST_maskXYZ( int *nnzero, THD_3dim_dataset *dset, int nvox );
static float *DST_b0_image( int b0_type, THD_3dim_dataset *dset, int nvox);

int main( int argc, char *argv[] )
{
  THD_3dim_dataset *dset[3], *target_dset = NULL;
  float threshold[7];
  int i, option = 0, 
    flag = 0;  // check if command-line arguments are legal

  option = setOptions( argc, argv );
                   // 0:dwi, 1:dti, 2:seeds dataset
  for( i = 0; i < 3; ++ i )
  {
    if( (dset[i] = THD_open_dataset( argv[i+1] )) == NULL )
    {
      printf( "%s: argument %d: cannot open dataset %s\n", argv[0], i, 
        argv[i+1] );
      flag = 1;
    }
    if( dset[i] != NULL ) DSET_load( dset[i] );
  }  
                   // fa, distRatio
  for( i = 0; i < 2; ++ i )
  {
    threshold[i] = atof( argv[i+4] );
    if( threshold[i] < 0 || threshold[i] > 1 )
    {
      printf( "%s: argument %d: has to be in range [0,1]\n", argv[0], i+4 );
      flag = 1;
    }
  }  
                   // curv, coher
  for( i = 2; i < 4; ++ i )
  {
    threshold[i] = atof( argv[i+4] );
    if( threshold[i] < 0 || threshold[i] > 90.0001 )
    {
      printf( "%s: argument %d: has to be in range [0,90]\n", argv[0], i+4 );
      flag = 1;
    }
  }  
                   // integTime, scaled to lie in [1,10]
  threshold[4] = atof( argv[8] );
  if( threshold[4] < 0.9999 || threshold[4] > 10.0001 ) 
  {
    printf( "%s: argument %d: has to be in range [1,10]\n",argv[0],8);
    flag = 1;
  }
  threshold[4] *= 1.0e-4;
                   // b0 lower and upper bound
  for( i = 5; i < 7; ++ i )
  {
    threshold[i] = atof( argv[i+4] );
    if( threshold[i] < 0 )
    {
      printf( "%s: argument %d: cannot be negative\n", argv[0],i+4 );
      flag = 1;
    }
  }
                   // "-t" is used to include target dataset
  if( option == 1 )
  if( (target_dset = THD_open_dataset( argv[12] )) == NULL )
  {
    printf( "%s: argument %d: cannot open dataset %s\n", argv[0],12, argv[12]);
    flag = 1;
  }
  if( target_dset != NULL ) DSET_load( target_dset );
                   // some arguments are illegal
  if( flag == 1 ) return 1;
                   //     
                   // retrieve necessary info for the DST fiber tracking
		   // interface function
  int nvox = DSET_NVOX( dset[1] );
  int dims[3];
  float voxSize[3];

  dims[0] = DSET_NX( dset[1] );
  dims[1] = DSET_NY( dset[1] );
  dims[2] = DSET_NZ( dset[1] );
  voxSize[0] = DSET_DX( dset[1] );
  voxSize[1] = DSET_DY( dset[1] );
  voxSize[2] = DSET_DZ( dset[1] );

  fprintf( stderr,"%s: Loading dti dataset '%s': \n"
	      "\t orientation =  (%s,%s,%s)\n"
              "\t origin =  (%f,%f,%f)\n"
	      "\t dimensions (nx,ny,nz) =  (%i,%i,%i)\n"
	      "\t increments (dx,dy,dz) =  (%f,%f,%f)\n",
	      argv[0],DSET_PREFIX(dset[1]),
	      ORIENT_typestr[dset[1]->daxes->xxorient],
	      ORIENT_typestr[dset[1]->daxes->yyorient],
	      ORIENT_typestr[dset[1]->daxes->zzorient],
	      DSET_XORG(dset[1]),DSET_YORG(dset[1]),DSET_ZORG(dset[1]),
	      dims[0],dims[1],dims[2],voxSize[0],voxSize[1],voxSize[2] );
                   // retrieve tensor components
  int tensorOffset = 0;   
                   // pointer to the initial address for each of the six 
		   // tensor components
  float *tensor[6]; 
  for( i = 0; i < 6; ++ i )
    tensor[i] = (float *)DSET_ARRAY( dset[1], tensorOffset+i );
                   // retrieve FA, MD, principal eigenvector
  float *fa, *md, *eigvec[3];	
  int eigvecOffset = 9, faOffset = 18, mdOffset = 19;
  fa = (float *)DSET_ARRAY( dset[1], faOffset );
  md = (float *)DSET_ARRAY( dset[1], mdOffset );
  for( i = 0; i < 3; ++ i )
    eigvec[i] = (float *)DSET_ARRAY( dset[1], eigvecOffset + i );
                   // retrieve the file name for seeds dataset
  char seeds_dset_prefix[THD_MAX_NAME];
  strcpy( seeds_dset_prefix, DSET_PREFIX( dset[2] ) );
                   // retrieve seeds points
  int **seedsXYZ, nnzero;
  seedsXYZ = DST_maskXYZ( &nnzero, dset[2], nvox );
  if( nnzero == 0 )
  {
    fprintf( stderr, "%s: there are no seeds from %s!\n", argv[0],
      DSET_FILECODE( dset[2] ) );
    return 1;
  }  
                   // retrieve target zone points if specified 
  int **targetXYZ, nnzero_target = 0;
  if( target_dset != NULL )
    targetXYZ = DST_maskXYZ( &nnzero_target, target_dset, nvox );
                   // retrieve b0 image in dwi dataset
  float *b0_float = NULL;
  int b0_type = DSET_BRICK_TYPE( dset[0], 0 );
  b0_float = DST_b0_image( b0_type, dset[0], nvox ); 
                   // do NOT need dwi dataset anymore
  DSET_unload( dset[0] ); 

  printf(">> start DST compute with %d seeds\n", nnzero );
  printf(">> DST thresholds: FA=%.2f dist-ratio=%.2f curv=%.2f coher=%.2f\n"
    "                   integ-time=%.4f b0_lo=%g b0_hi=%g\n",
    threshold[0],threshold[1],threshold[2],threshold[3],threshold[4], 
    threshold[5],threshold[6]);
                    // fiber tracking starts here 
  Args_Table_ dstArgs = diffsimFT_interface( nvox, dims, voxSize, 
    tensor, fa, md, eigvec, threshold, b0_float,
    nnzero, seedsXYZ, seeds_dset_prefix, nnzero_target, targetXYZ );
  if( NULL == dstArgs ) return 1;

                   // done with seeds
  for( i = 0; i < 3; ++ i )
    free( seedsXYZ[i] );
  free( seedsXYZ );  
                   // done with target zone points
  if( nnzero_target != 0 )
  {
    for( i = 0; i < 3; ++ i )
      free( targetXYZ[i] );
    free( targetXYZ ); 
  }
                   // done with the b0 image
  free( b0_float );

  printf(">> DST compute: done --> '%s' generated\n", 
    dstArgs->tractFileName );
  free( dstArgs );

  return 0;
}

static int setOptions( int argc, char *argv[] )
{
  if( argc < 11 ) help( argv[0] );
  if( argc == 11 ) return 0; 
  if( argc > 11 && strncmp( argv[11], "-t", 2 ) != 0 )
  {
    fprintf( stderr, "%s: option %s not supported\n", argv[0], argv[11] );
    help( argv[0] );
  }
  if( argc == 12 && strncmp( argv[11], "-t", 2 ) == 0 )
  {
    fprintf( stderr, "%s: argument missing after %s\n", argv[0], argv[11] );
    help( argv[0] );
  }
  return 1;
}

static void help( char *command )
{
  printf( "NAME:\n"
          "  diffdst - generate fiber tract data file in VTK format and its\n"
	  "            statistics, using DST fiber tracking algorithm\n"
          "\nSYNOPSIS:\n" 
	  "  diffdst dwi dti seeds fa dist_ratio curv coher integ_time\n"
          "          b0_lo b0_hi [-t target]\n" 
	  "\nARGUMENTS:\n"
	  "  1. dwi: dwi dataset, in AFNI format\n"
	  "  2. dti: dti dataset, in AFNI format\n"
	  "  3. seeds: seeds dataset, in AFNI format\n"
	  "  4. fa: threshold for FA values, in range [0,1]\n"
	  "  5. dist_ratio: threshold for distance ratio, in range [0,1]\n"
	  "  6. curv: threshold for curvature angle, in range [0,90]\n"
	  "  7. coher: threshold for local coherence angle, in range [0,90]\n"
	  "  8. integ_time: integraton time for diffusion simulations,\n"
	  "                 in range [1,10]\n"
	  "  9. b0_lo: lower bound threshold for b0 image, non-negative\n"
	  " 10. b0_hi: upper bound threshold for b0 image, non-negative\n"
	  "\nOPTIONS:\n"
	  " 11. -t: flag to use target dataset\n"
	  " 12. target: target zone dataset, in AFNI format\n" );
  exit( 0 );
}
                   // retrieve b0 image from dwi dataset
static float *DST_b0_image( int b0_type, THD_3dim_dataset *dwi_dset, int nvox )
{
  float *b0_float = NULL;
  short *b0_short = NULL;
  int *b0_int = NULL;
  float *b0_f = NULL;
  int i;

  b0_float = (float *)malloc( nvox*sizeof( float ) );
  switch( b0_type )
  {
    case MRI_short:
    {
      b0_short = (short *)DSET_ARRAY( dwi_dset, 0 );
      for( i = 0; i < nvox; ++ i ) b0_float[i] = b0_short[i];
    }
      break;
    case MRI_int:
    {
      b0_int = (int *)DSET_ARRAY( dwi_dset, 0 );
      for( i = 0; i < nvox; ++ i ) b0_float[i] = b0_int[i];
    }
      break;
    case MRI_float:
    {
      b0_f = (float *)DSET_ARRAY( dwi_dset, 0 );
      for( i = 0; i < nvox; ++ i ) b0_float[i] = b0_f[i];
    }  
      break;
    default:
    {
      for( i = 0; i < nvox; ++ i ) b0_float[i] = 0;
      fprintf( stderr, "** b0 image: dataset type not supported yet." 
               "Instead, a b0 image with all zero values will be used.\n" );
    }	       
      break;
  }
  return b0_float;
}

                   // create a byte mask to retrieve coordinates of seeds
static int **DST_maskXYZ( int *nnzero, THD_3dim_dataset *seeds_dset, int nvox )
{
  int type, volumeOffset = 0;
  float thresh;
  byte *mask = NULL;
  
  type = DSET_BRICK_TYPE( seeds_dset, volumeOffset );
  switch( type )
  {
    case MRI_short:
      thresh = 0;
      break;
    case MRI_float:
      thresh = 1;
      break;
    default:
    {
      fprintf( stderr, "** %s: dataset type not supported yet.\n",
        DSET_PREFIX( seeds_dset ) ); 
      *nnzero = 0;
      return NULL;
    }
  }

  *nnzero = thd_mask_from_brick( seeds_dset, volumeOffset, thresh, &mask, 1 );
                   // retrieve coordinates of the seeds
/*
      float *seeds = (float *)DSET_ARRAY( seeds_dset, 0 );
      int nnzero = 0;
      for( int i = 0; i < nvox; ++ i )
      if( 0 != seeds[i] ) nnzero ++;
*/
  if( *nnzero == 0 )
    return NULL;

  int **seedsXYZ, cnt;
  seedsXYZ = (int **)malloc( 3 * sizeof(int*) );
  for( int i = 0; i < 3; ++ i )
    seedsXYZ[i] = (int *)malloc( *nnzero * sizeof(int) );

  cnt = 0;	
  for( int i = 0; i < nvox; ++ i )
  if( mask[i] ) 
  {/*
//        seedsXYZ[0][cnt] = DSET_index_to_ix( seeds_dset, i );
    seedsXYZ[0][cnt] = dims[0]-1 - DSET_index_to_ix( seeds_dset, i );
    seedsXYZ[1][cnt] = DSET_index_to_jy( seeds_dset, i );
    seedsXYZ[2][cnt] = dims[2]-1 - DSET_index_to_kz( seeds_dset, i );
//        seedsXYZ[2][cnt] = DSET_index_to_kz( seeds_dset, i );
*/
    // shark
    seedsXYZ[0][cnt] = DSET_index_to_ix( seeds_dset, i );
    seedsXYZ[1][cnt] = DSET_index_to_jy( seeds_dset, i );
    seedsXYZ[2][cnt] = DSET_index_to_kz( seeds_dset, i );
    cnt ++;
  }
                   // done with the mask
  if( mask) free( mask );
  return seedsXYZ;
}
