/**********************************************************************
 * This is the command-line verison for DST fiber tracking in the 
 * diffusion plugin.
 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "afni.h"
#include "dst_interface.h"

static void help( char *command );
static int setOptions( int argc, char *argv[] );
static int **DST_maskXYZ( int *nnzero, THD_3dim_dataset *dset, int nvox );
static int tensor2eig( THD_3dim_dataset *dset, float *tensor[], 
  char *eig_prefix );

int main( int argc, char *argv[] )
{
  THD_3dim_dataset *dset[7], *target_dset = NULL;
  float threshold[7];
  int i, option = 0, 
    flag = 0;  // check if command-line arguments are legal

  option = setOptions( argc, argv );
                   // 0-5: six tensor components, 6:seeds dataset
  for( i = 0; i < 7; ++ i )
  {
    if( (dset[i] = THD_open_dataset( argv[i+1] )) == NULL )
    {
      printf( "%s: argument %d: cannot open dataset %s\n", argv[0], i, 
        argv[i+1] );
      flag = 1;
    }
    if( dset[i] != NULL ) DSET_load( dset[i] );
  }  
                   // fa, distRatio
  for( i = 0; i < 2; ++ i )
  {
    threshold[i] = atof( argv[i+8] );
    if( threshold[i] < 0 || threshold[i] > 1 )
    {
      printf( "%s: argument %d: has to be in range [0,1]\n", argv[0], i+8 );
      flag = 1;
    }
  }  
                   // curv, coher
  for( i = 2; i < 4; ++ i )
  {
    threshold[i] = atof( argv[i+8] );
    if( threshold[i] < 0 || threshold[i] > 90.0001 )
    {
      printf( "%s: argument %d: has to be in range [0,90]\n", argv[0], i+8 );
      flag = 1;
    }
  }  
                   // integTime, scaled to lie in [1,10]
  threshold[4] = atof( argv[12] );
  if( threshold[4] < 0.9999 || threshold[4] > 10.0001 ) 
  {
    printf( "%s: argument %d: has to be in range [1,10]\n",argv[0],12);
    flag = 1;
  }
  threshold[4] *= 1.0e-4;
                   // "-t" is used to include target dataset
  if( option == 1 )
  if( (target_dset = THD_open_dataset( argv[13] )) == NULL )
  {
    printf( "%s: argument %d: cannot open dataset %s\n", argv[0],13, argv[13]);
    flag = 1;
  }
  if( target_dset != NULL ) DSET_load( target_dset );
                   // some arguments are illegal
  if( flag == 1 ) return 1;
                   //     
                   // retrieve necessary info for the DST fiber tracking
		   // interface function
  int nvox = DSET_NVOX( dset[0] );
  int dims[3];
  float voxSize[3];

  dims[0] = DSET_NX( dset[0] );
  dims[1] = DSET_NY( dset[0] );
  dims[2] = DSET_NZ( dset[0] );
  voxSize[0] = DSET_DX( dset[0] );
  voxSize[1] = DSET_DY( dset[0] );
  voxSize[2] = DSET_DZ( dset[0] );

  fprintf( stderr,"%s: Loading dt component dataset '%s': \n"
	      "\t orientation =  (%s,%s,%s)\n"
              "\t origin =  (%f,%f,%f)\n"
	      "\t dimensions (nx,ny,nz) =  (%i,%i,%i)\n"
	      "\t increments (dx,dy,dz) =  (%f,%f,%f)\n",
	      argv[0],DSET_PREFIX(dset[0]),
	      ORIENT_typestr[dset[0]->daxes->xxorient],
	      ORIENT_typestr[dset[0]->daxes->yyorient],
	      ORIENT_typestr[dset[0]->daxes->zzorient],
	      DSET_XORG(dset[0]),DSET_YORG(dset[0]),DSET_ZORG(dset[0]),
	      dims[0],dims[1],dims[2],voxSize[0],voxSize[1],voxSize[2] );

                   // retrieve tensor components
                   // pointer to the initial address for each of the six 
		   // tensor components
  float *tensor[6]; 
  for( i = 0; i < 6; ++ i )
    tensor[i] = (float *)DSET_ARRAY( dset[i], 0 );
               // create dataset with eigen stuff and FA,MD
  char *eig_dset_prefix = "dtEig+orig";	       
  if( tensor2eig( dset[0], tensor, eig_dset_prefix ) != 0 ) return 1;
  THD_3dim_dataset *eig_dset = THD_open_dataset( eig_dset_prefix );
  DSET_load( eig_dset );

                   // retrieve FA, MD, principal eigenvector
  float *fa, *md, *eigvec[3];	
  int eigvecOffset = 3, faOffset = 12, mdOffset = 13;
  fa = (float *)DSET_ARRAY( eig_dset, faOffset );
  md = (float *)DSET_ARRAY( eig_dset, mdOffset );
  for( i = 0; i < 3; ++ i )
    eigvec[i] = (float *)DSET_ARRAY( eig_dset, eigvecOffset + i );
                   // retrieve the file name for seeds dataset
  char seeds_dset_prefix[THD_MAX_NAME];
  strcpy( seeds_dset_prefix, DSET_PREFIX( dset[6] ) );
                   // retrieve seeds points
  int **seedsXYZ, nnzero;
  seedsXYZ = DST_maskXYZ( &nnzero, dset[6], nvox );
  if( nnzero == 0 )
  {
    fprintf( stderr, "%s: there are no seeds from %s!\n", argv[0],
      DSET_FILECODE( dset[6] ) );
    return 1;
  }  
                   // retrieve target zone points if specified 
  int **targetXYZ, nnzero_target = 0;
  if( target_dset != NULL )
    targetXYZ = DST_maskXYZ( &nnzero_target, target_dset, nvox );
                   // the b0 image is not actually used
  float *b0_float = fa;
  threshold[5]=0; threshold[6]=0; // for b0

  printf(">> start DST compute with %d seeds\n", nnzero );
  printf(">> DST thresholds: FA=%.2f dist-ratio=%.2f curv=%.2f coher=%.2f\n"
    "                   integ-time=%.4f\n",
    threshold[0],threshold[1],threshold[2],threshold[3],threshold[4] );
                    // fiber tracking starts here 
  Args_Table_ dstArgs = diffsimFT_interface( nvox, dims, voxSize, 
    tensor, fa, md, eigvec, threshold, b0_float,
    nnzero, seedsXYZ, seeds_dset_prefix, nnzero_target, targetXYZ );
  if( NULL == dstArgs ) return 1;

                   // done with seeds
  for( i = 0; i < 3; ++ i )
    free( seedsXYZ[i] );
  free( seedsXYZ );  
                   // done with target zone points
  if( nnzero_target != 0 )
  {
    for( i = 0; i < 3; ++ i )
      free( targetXYZ[i] );
    free( targetXYZ ); 
  }

  printf(">> DST compute: done --> '%s' generated\n", 
    dstArgs->tractFileName );
  free( dstArgs );

  return 0;
}

static int setOptions( int argc, char *argv[] )
{
  int num = 13;
  if( argc < num ) help( argv[0] );
  if( argc == num ) return 0; 
  if( argc > num && strncmp( argv[num], "-t", 2 ) != 0 )
  {
    fprintf( stderr, "%s: option %s not supported\n", argv[0], argv[num] );
    help( argv[0] );
  }
  if( argc == num+1 && strncmp( argv[num], "-t", 2 ) == 0 )
  {
    fprintf( stderr, "%s: argument missing after %s\n", argv[0], argv[num] );
    help( argv[0] );
  }
  return 1;
}

static void help( char *command )
{
  printf( "NAME:\n"
          "  diffdst6 - generate fiber tract data file in VTK format and its\n"
	  "            statistics, using DST fiber tracking algorithm\n"
          "\nSYNOPSIS:\n" 
	  "  diffdst6 Dxx Dxy Dyy Dxz Dyz Dzz seeds \n"
	  "           fa dist_ratio curv coher integ_time\n"
          "           [-t target]\n" 
	  "\nARGUMENTS:\n"
	  "  1. Dxx, in AFNI/NII format\n"
	  "  2. Dxy, in AFNI/NII format\n"
	  "  3. Dyy, in AFNI/NII format\n"
	  "  4. Dxz, in AFNI/NII format\n"
	  "  5. Dyz, in AFNI/NII format\n"
	  "  6. Dzz, in AFNI/NII format\n"
	  "  7. seeds: seeds dataset, in AFNI format\n"
	  "  8. fa: threshold for FA values, in range [0,1]\n"
	  "  9. dist_ratio: threshold for distance ratio, in range [0,1]\n"
	  " 10. curv: threshold for curvature angle, in range [0,90]\n"
	  " 11. coher: threshold for local coherence angle, in range [0,90]\n"
	  " 12. integ_time: integraton time for diffusion simulations,\n"
	  "                 in range [1,10]\n"
	  "\nOPTIONS:\n"
	  " 13. -t: flag to use target dataset\n"
	  " 14. target: target zone dataset, in AFNI format\n" );
  exit( 0 );
}

                   // create a byte mask to retrieve coordinates of seeds
static int **DST_maskXYZ( int *nnzero, THD_3dim_dataset *seeds_dset, int nvox )
{
  int type, volumeOffset = 0;
  float thresh;
  byte *mask = NULL;
  
  type = DSET_BRICK_TYPE( seeds_dset, volumeOffset );
  switch( type )
  {
    case MRI_short:
      thresh = 0;
      break;
    case MRI_float:
      thresh = 1;
      break;
    default:
    {
      fprintf( stderr, "** %s: dataset type not supported yet.\n",
        DSET_PREFIX( seeds_dset ) ); 
      *nnzero = 0;
      return NULL;
    }
  }

  *nnzero = thd_mask_from_brick( seeds_dset, volumeOffset, thresh, &mask, 1 );
                   // retrieve coordinates of the seeds
/*
      float *seeds = (float *)DSET_ARRAY( seeds_dset, 0 );
      int nnzero = 0;
      for( int i = 0; i < nvox; ++ i )
      if( 0 != seeds[i] ) nnzero ++;
*/
  if( *nnzero == 0 )
    return NULL;

  int **seedsXYZ, cnt;
  seedsXYZ = (int **)malloc( 3 * sizeof(int*) );
  for( int i = 0; i < 3; ++ i )
    seedsXYZ[i] = (int *)malloc( *nnzero * sizeof(int) );

  cnt = 0;	
  for( int i = 0; i < nvox; ++ i )
  if( mask[i] ) 
  {/*
//        seedsXYZ[0][cnt] = DSET_index_to_ix( seeds_dset, i );
    seedsXYZ[0][cnt] = dims[0]-1 - DSET_index_to_ix( seeds_dset, i );
    seedsXYZ[1][cnt] = DSET_index_to_jy( seeds_dset, i );
    seedsXYZ[2][cnt] = dims[2]-1 - DSET_index_to_kz( seeds_dset, i );
//        seedsXYZ[2][cnt] = DSET_index_to_kz( seeds_dset, i );
*/
    // shark
    seedsXYZ[0][cnt] = DSET_index_to_ix( seeds_dset, i );
    seedsXYZ[1][cnt] = DSET_index_to_jy( seeds_dset, i );
    seedsXYZ[2][cnt] = DSET_index_to_kz( seeds_dset, i );
    cnt ++;
  }
                   // done with the mask
  if( mask) free( mask );
  return seedsXYZ;
}

int tensor2eig( THD_3dim_dataset *dset, float *tensor[], char *eig_prefix )
{
  int i, numSubBricks = 6;
  char *dt_prefix = "dtElem+orig";

               // datasets already generated
  if( THD_open_dataset( dt_prefix ) != NULL &&
    THD_open_dataset( eig_prefix ) != NULL )
  {  
printf(">> tensor elements and eigen datasets exist\n");
    return 0;
  }  
               // create DT dataset with 6 sub-bricks 
	       // Dxx,Dxy,Dyy,Dxz,Dyz,Dzz
  THD_3dim_dataset *tensor_dset = EDIT_empty_copy( dset );
  i = EDIT_dset_items( tensor_dset,
    ADN_prefix, dt_prefix, // filename prefix
    ADN_malloc_type, DATABLOCK_MEM_MALLOC, // store in memory
    ADN_datum_all, MRI_float, // atomic datum
    ADN_nvals, numSubBricks, // number of sub-bricks
    ADN_ntt, 0,
    ADN_type, ISHEAD( dset ) ? // dataset type
              HEAD_FUNC_TYPE : GEN_FUNC_TYPE, 
    ADN_func_type, FUNC_BUCK_TYPE, // function type       
    ADN_none );
  if( i != 0 ) return -1;

  THD_init_datablock_labels( tensor_dset->dblk );  
  THD_init_datablock_stataux( tensor_dset->dblk );  
  THD_init_datablock_keywords( tensor_dset->dblk );  

  char *labels[] = {"Dxx","Dxy","Dyy","Dxz","Dyz","Dzz"};
  for( i=0; i<numSubBricks; ++i )
  {
    EDIT_substitute_brick( tensor_dset, i, MRI_float, tensor[i] );
    EDIT_BRICK_LABEL( tensor_dset, i, labels[i] );
  }
  DSET_write( tensor_dset );
  printf(">> save tensor elements dataset %s\n", DSET_BRIKNAME(tensor_dset));

               // compute eigen stuff from DT components
  char sysCommand[100];
  sprintf( sysCommand, "3dDTeig -prefix %s %s", eig_prefix, dt_prefix );
printf(">> compute eigen: %s\n", sysCommand );  
  if( system( sysCommand ) )
    printf("<< **Something went wrong with 3dDTeig\n" );

  return 0;
}
