/***********************************************************************
 * File:        diffkernel.c
 * Programmer:  Ning Kang @CS.UKY.EDU
 * Version of:  6/6/2004
 * ---------------------------------------------------------------------
 * Description:
 * - Serial version.
 * - This file contains the following functions:
 *   int kernelToGridMapping ()
 *   void setupKernelVoxel ()
 *   void diffThreshold ()
 *   realtype quickSelect ()
 ***********************************************************************/
 

#include "dst.h"

static void setupKernelVoxel( void *kernelData, void *dtiData );
static void setupKernelVoxel_DT4( void *kernelData, void *dtiData );
static float quickSelect( float *arr, int n, int median ); 

int kernelToGridMapping( void *kernelData, void *dtiData )
{
    DTI_ dti = (DTI_)dtiData;
    Diff_Kernel_ kernel = (Diff_Kernel_)kernelData;
    Kernel_Voxel_ *kernelVoxel = kernel->voxel;
    int id, gid;

    setupKernelVoxel( kernelData, dtiData );

    for( id = 0; id < kernel->neq; ++ id )
    if( kernelVoxel[id]->zeroTensor == FALSE )
    {
        gid = kernelVoxel[id]->gridID;
	kernelVoxel[id]->fa = dti->fa[gid];
	kernelVoxel[id]->b0 = dti->b0[gid];
//	kernelVoxel[id]->md = dti->md[gid]; 
	                  // principal eigenvector
	kernelVoxel[id]->eigvec[0] = dti->eigvec[0][gid];
	kernelVoxel[id]->eigvec[1] = dti->eigvec[1][gid];
	kernelVoxel[id]->eigvec[2] = dti->eigvec[2][gid];
    }	

    return 0;
}
               // for 4th-order DT
int kernelToGridMapping_DT4( void *kernelData, void *dtiData )
{
    DTI_ dti = (DTI_)dtiData;
    Diff_Kernel_ kernel = (Diff_Kernel_)kernelData;
    Kernel_Voxel_ *kernelVoxel = kernel->voxel;
    int id, gid;

    setupKernelVoxel_DT4( kernelData, dtiData );

    for( id = 0; id < kernel->neq; ++ id )
    if( kernelVoxel[id]->zeroTensor == FALSE )
    {
        gid = kernelVoxel[id]->gridID;
	kernelVoxel[id]->fa = dti->dt4_gfa[gid];
    }	
    return 0;
}

void setupKernelVoxel( void *kernelData, void *dtiData )
{
    Diff_Kernel_ kernel;
    Kernel_Voxel_ *voxel;
    DTI_ dti;
    int i, j, k, m, id, gi, gj, gk;
    int nx, ny, nz, nxy, gnx, gny, gnz, gnxy;
    int seedI, seedJ, seedK;
    realtype **dtxyz;
    
    kernel = (Diff_Kernel_)kernelData;
    dti = (DTI_)dtiData;
    voxel = kernel->voxel;
    dtxyz = dti->dtxyz;
    nx = kernel->nx;
    ny = kernel->ny;
    nz = kernel->nz;
    nxy = nx*ny;
    gnx = dti->nx;
    gny = dti->ny;
    gnz = dti->nz;
    gnxy = gnx*gny;
    seedI = dti->seedI;
    seedJ = dti->seedJ;
    seedK = dti->seedK;
    
    for( i = 0; i < nx; ++ i )
    for( j = 0; j < ny; ++ j )
    for( k = 0; k < nz; ++ k )
    {
        id = nodeIndex( i+1, j+1, k+1, nx, nxy ) - 1;
        voxel[id]->id = id;
	voxel[id]->i = i;
	voxel[id]->j = j;
	voxel[id]->k = k;
	voxel[id]->dist = 0;
	voxel[id]->fa = 0;
	voxel[id]->b0 = 0;
//	voxel[id]->md = 0;
	voxel[id]->zeroTensor = TRUE;
	voxel[id]->next = NULL;
	voxel[id]->tail = NULL;
	gi = seedI - nx/2 + i;
	gj = seedJ - ny/2 + j;
	gk = seedK - nz/2 + k;
	if( (gi >= 0 && gi < gnx) && 
	    (gj >= 0 && gj < gny) &&
	    (gk >= 0 && gk < gnz) )
	{
	    voxel[id]->gridID = nodeIndex( gi+1, gj+1, gk+1, gnx, gnxy ) - 1;
	    for( m = 0; m < dti->dtxyzNum; ++ m )
	        if( DTXYZ(voxel[id]->gridID, m) != 0 && 
		    voxel[id]->zeroTensor == TRUE )
		{
	            voxel[id]->zeroTensor = FALSE;
		    break;
		}	
	}   
	else
	    voxel[id]->gridID = OUT_OF_GRID;
    }
}
               // for 4th-order DT 
void setupKernelVoxel_DT4( void *kernelData, void *dtiData )
{
    Diff_Kernel_ kernel;
    Kernel_Voxel_ *voxel;
    DTI_ dti;
    int i, j, k, m, id, gi, gj, gk;
    int nx, ny, nz, nxy, gnx, gny, gnz, gnxy;
    int seedI, seedJ, seedK;
    realtype **dt4Elem;
    
    kernel = (Diff_Kernel_)kernelData;
    dti = (DTI_)dtiData;
    voxel = kernel->voxel;
    dt4Elem = dti->dt4Elem;
    nx = kernel->nx;
    ny = kernel->ny;
    nz = kernel->nz;
    nxy = nx*ny;
    gnx = dti->nx;
    gny = dti->ny;
    gnz = dti->nz;
    gnxy = gnx*gny;
    seedI = dti->seedI;
    seedJ = dti->seedJ;
    seedK = dti->seedK;
    
    for( i = 0; i < nx; ++ i )
    for( j = 0; j < ny; ++ j )
    for( k = 0; k < nz; ++ k )
    {
        id = nodeIndex( i+1, j+1, k+1, nx, nxy ) - 1;
        voxel[id]->id = id;
	voxel[id]->i = i;
	voxel[id]->j = j;
	voxel[id]->k = k;
	voxel[id]->dist = 0;
	voxel[id]->fa = 0;
	voxel[id]->zeroTensor = TRUE;
	voxel[id]->next = NULL;
	voxel[id]->tail = NULL;
	gi = seedI - nx/2 + i;
	gj = seedJ - ny/2 + j;
	gk = seedK - nz/2 + k;
	if( (gi >= 0 && gi < gnx) && 
	    (gj >= 0 && gj < gny) &&
	    (gk >= 0 && gk < gnz) )
	{
	    voxel[id]->gridID = nodeIndex( gi+1, gj+1, gk+1, gnx, gnxy ) - 1;
	    for( m = 0; m < DT4_ELEM; ++ m )
	        if( DTXYZ_4TH(voxel[id]->gridID, m) != 0 && 
		    voxel[id]->zeroTensor == TRUE )
		{
	            voxel[id]->zeroTensor = FALSE;
		    break;
		}	
	}   
	else
	    voxel[id]->gridID = OUT_OF_GRID;
    }
}

/***********************************************************************
 * Set threshold for determining zero concentration.
 ***********************************************************************/
void diffThreshold( void *kernelData, void *daeData )
{
    Diff_Kernel_ kernel;
    DAE_ dae;
    realtype *yData, *diffval;
    int neq;
    
    kernel = (Diff_Kernel_)kernelData;
    dae = (DAE_)daeData;
    neq = kernel->neq;
    diffval = kernel->diffval;
    
                        /* copy the concentration values to work array */
    yData = NV_DATA_S( dae->yVector );
    memcpy( diffval, yData, neq * sizeof( realtype ) );

    kernel->cThreshold = quickSelect( diffval, neq, kernel->median );
//printf("----> cThreshold = %.4e\n", kernel->cThreshold);
}


/***********************************************************************
 * Returns the k-th ("median") smallest value in the array arr[1..n]. 
 * The input array will be rearranged to have this value in location 
 * arr[k], with all smaller elements moved to arr[1..k-1] (in arbitrary 
 * order) and all larger elements in arr[k+1..n] (also in arbitrary order).
 ***********************************************************************/
float quickSelect( float *arr, int n, int median ) 
{
    int low, high ;
    int middle, ll, hh;

    low = 0 ; 
    high = n-1 ; 
    for( ; ; ) 
    {
                        /* One element only */
        if (high <= low) return arr[median];
                        
			/* Two elements only */
        if (high == low + 1) 
	{  
            if (arr[low] > arr[high])
                ELEM_SWAP( arr[low], arr[high] );
            return arr[median];
        }
                        /* Find median of low, middle and high items; 
			   swap into position low */
        middle = (low + high) / 2;
        if( arr[middle] > arr[high] )    ELEM_SWAP(arr[middle], arr[high]);
        if( arr[low] > arr[high] )       ELEM_SWAP(arr[low], arr[high]);
        if( arr[middle] > arr[low] )     ELEM_SWAP(arr[middle], arr[low]);

                        /* Swap low item (now in position middle) into 
			   position (low+1) */
        ELEM_SWAP( arr[middle], arr[low+1] );

                        /* Nibble from each end towards middle, 
			   swapping items when stuck */
        ll = low + 1;
        hh = high;
        for( ; ; ) 
	{
            do ll++; while( arr[low] > arr[ll] );
            do hh--; while( arr[hh]  > arr[low] );
            if( hh < ll ) break;
            ELEM_SWAP( arr[ll], arr[hh] );
        }

                        /* Swap middle item (in position low) back into 
			   correct position */
        ELEM_SWAP(arr[low], arr[hh]);

                        /* Re-set active partition */
        if( hh <= median ) low = ll;
        if( hh >= median ) high = hh - 1;
    }
} 

