/***********************************************************************
 * File:        diffsim.c
 * Programmer:  Ning Kang @CS.UKY
 * Version of:  5/27/2003
 * ---------------------------------------------------------------------
 * Description:
 * - Serial version.
 * - The functions in this file are used to do anisotropic diffusion
 * simulaiton.
 * - This file contains the following functions:
 *   int rootNodeDiffSim ()
 *   int leafNodeDiffSim ()
 *   void setInitialProfiles ()
 *   int res ()
 *   int precond ()
 *   int pSolve ()
 *   void saveDiffProfileVTK ()
 ***********************************************************************/
 

#include "dst.h"

static void setInitialProfiles( void *kernelData, void *daeData ); 
static int res( int Neq, realtype t, N_Vector yVector, N_Vector ypVector, 
  N_Vector rVector, void *rdata );
static void setInitialProfiles_DT4( void *kernelData, void *daeData );
static int res_DT4( int Neq, realtype t, N_Vector yVector, N_Vector 
  ypVector, N_Vector rVector, void *rdata );
static int precond( int Neq, realtype t, N_Vector yVector, N_Vector ypVector, 
  N_Vector rVector, realtype cj, ResFn res, void *rdata, void *pdata, 
  N_Vector ewtVector, N_Vector constraintsVector, realtype h, realtype uround,
  long int *nrePtr, N_Vector tempv1, N_Vector tempv2, N_Vector tempv3 );
static int pSolve( int Neq, realtype t, N_Vector yVector, N_Vector ypVector, 
  N_Vector rVector, realtype cj, ResFn res, void *rdata, void *pdata, 
  N_Vector ewtVector, realtype delta, N_Vector rhsVector, N_Vector zVector, 
  long int *nrePtr, N_Vector tempv );
static void initialGaussianPeak( N_Vector yVector, void *kernelData );
static void initialGaussianPeak_inner( N_Vector yVector, void *kernelData );
static void saveDiffProfileVTK( void *kernelData, void *daeData );

/***********************************************************************
 * Make the constants more readable in this file.
 ***********************************************************************/
#define itol (dae->itol)
#define rtol (dae->rtol)
#define atol (dae->atol)
#define itask (dae->itask)
#define epliFactor (dae->epliFactor)
#define maxKDim (dae->maxKDim)
#define maxRestart (dae->maxRestart)
#define initPeak (dae->initPeak)
#define optIn (dae->optIn)
#define iopt (dae->iopt)
#define ropt (dae->ropt)


/***********************************************************************
 * Perform diffusion simulation over the root node (start node).
 ***********************************************************************/
int rootNodeDiffSim( void *kernelData, void *daeData, void *iluData )
{
    Diff_Kernel_ kernel;
    DAE_ dae;
    ILU_ ilu;
    N_Vector yVector, ypVector, constraintsVector;
    realtype t, t0, tout;
    int flag, counter;
        
    kernel = (Diff_Kernel_)kernelData;
    ilu = (ILU_)iluData;
    dae = (DAE_)daeData;
    yVector = dae->yVector;
    ypVector = dae->ypVector;
    constraintsVector = dae->constraintsVector;
    t0 = dae->t0;
    t = dae->t;
    tout = dae->tout;
                                // Initializes the DAE system to a consistent 
                                // condition
    setInitialProfiles( kernel, dae );
                                // Call 'IDAMalloc' to initialize IDA.
				// NULL arguments are 'id and 'errfp'.
    dae->mem = IDAMalloc( kernel->neq, res, kernel, t0, yVector, ypVector, 
        itol, &rtol, &atol, NULL, constraintsVector, NULL, optIn, iopt, 
	ropt, dae->machEnv );
    if( dae->mem == NULL )
    {
        printf( "IDAMalloc failed\n" );
	return 1;
    }
                                // Call 'IDASpgmr' to specify the linear solver 
    flag = IDASpgmr( dae->mem, precond, pSolve, MODIFIED_GS, maxKDim, 
        maxRestart, epliFactor, 0.0, ilu ); 
    if( flag != DST_SUCCESS )
    {
        printf( "IDASpgmr failed: %d\n", flag );
	return 1;
    }
                                // Loop over output times, call 'IDASolve'
/*    flag = IDASolve( dae->mem, tout, tout, &t, yVector, ypVector, itask );
    if( flag < 0 )
    {
	printf( "IDASolve failed: %d\n", flag );
	return 1;
    }
*/
    counter = 0;
    while( t < tout )
    {
        flag = IDASolve( dae->mem, tout, tout, &t, yVector, ypVector, itask );	
#ifdef DEBUG_PRINT
//printf( "root diff: t=%12.4e------- counter=%d\n", t,counter );
#endif
	if( flag < 0 )
	{
	    printf( "IDASolve failed: %d\n", flag );
	    return 1;
	}
	counter ++;
	if( counter >= MAX_STEP )
	    break;
    }
    dae->loop = counter; 
//*/    
    
//    saveDiffProfileVTK( kernelData, dae );
    
    return DST_SUCCESS;
}

               // for 4th-order DT
int rootNodeDiffSim_DT4( void *kernelData, void *daeData, void *iluData )
{
    Diff_Kernel_ kernel;
    DAE_ dae;
    ILU_ ilu;
    N_Vector yVector, ypVector, constraintsVector;
    realtype t, t0, tout;
    int flag, counter;
        
    kernel = (Diff_Kernel_)kernelData;
    ilu = (ILU_)iluData;
    dae = (DAE_)daeData;
    yVector = dae->yVector;
    ypVector = dae->ypVector;
    constraintsVector = dae->constraintsVector;
    t0 = dae->t0;
    t = dae->t;
    tout = dae->tout;
                                // Initializes the DAE system to a consistent 
                                // condition
    setInitialProfiles_DT4( kernel, dae );
                                // Call 'IDAMalloc' to initialize IDA.
				// NULL arguments are 'id and 'errfp'.
    dae->mem = IDAMalloc( kernel->neq, res_DT4, kernel, t0, yVector, 
        ypVector, itol, &rtol, &atol, NULL, constraintsVector, NULL, optIn, 
	iopt, ropt, dae->machEnv );
    if( dae->mem == NULL )
    {
        printf( "IDAMalloc failed\n" );
	return 1;
    }
                                // Call 'IDASpgmr' to specify the linear solver 
    flag = IDASpgmr( dae->mem, precond, pSolve, MODIFIED_GS, maxKDim, 
        maxRestart, epliFactor, 0.0, ilu ); 
    if( flag != DST_SUCCESS )
    {
        printf( "IDASpgmr failed: %d\n", flag );
	return 1;
    }
                                // Loop over output times, call 'IDASolve'
/*    flag = IDASolve( dae->mem, tout, tout, &t, yVector, ypVector, itask );
    if( flag < 0 )
    {
	printf( "IDASolve failed: %d\n", flag );
	return 1;
    }
*/
    counter = 0;
    while( t < tout )
    {
        flag = IDASolve( dae->mem, tout, tout, &t, yVector, ypVector, itask );	
#ifdef DEBUG_PRINT
//printf( "root diff: t=%12.4e------- counter=%d\n", t,counter );
#endif
	if( flag < 0 )
	{
	    printf( "IDASolve failed: %d\n", flag );
	    return 1;
	}
	counter ++;
	if( counter >= MAX_STEP )
	    break;
    }
    dae->loop = counter; 
//*/    
//printf("@@@@@@@@@@@@@@@loop counter = %d t=%.4e tout=%.4e\n",counter,t,tout);    
//    saveDiffProfileVTK( kernelData, dae );
    
    return DST_SUCCESS;
}


/***********************************************************************
 * Perform diffusion simulation over the leaf node (internal node).
 ***********************************************************************/
int leafNodeDiffSim( void *kernelData, void *daeData, void *iluData )
{
    Diff_Kernel_ kernel;
    DAE_ dae;
    ILU_ ilu;
    N_Vector yVector, ypVector, constraintsVector;
    realtype t, t0, tout, loop;
    int flag, counter = 0;
        
    kernel = (Diff_Kernel_)kernelData;
    ilu = (ILU_)iluData;
    dae = (DAE_)daeData;
    yVector = dae->yVector;
    ypVector = dae->ypVector;
    constraintsVector = dae->constraintsVector;
    t0 = dae->t0;
    t = dae->t;
    tout = dae->tout;
    loop = dae->loop;
                                // Re-initializes the DAE system 
    setInitialProfiles( kernel, dae );
                                // Re-initialize IDA and IDASPGMR 
				// NULL arguments are 'id', 'errfp', 'machEnv'.
    flag = IDAReInit( dae->mem, res, kernel, t0, yVector, ypVector, 
        itol, &rtol, &atol, NULL, constraintsVector, NULL, optIn, 
	iopt, ropt, NULL );
    if( flag != DST_SUCCESS )
    {
        printf( "IDAReInit failed\n" );
	return 1; 
    }
     
    flag = IDAReInitSpgmr( dae->mem, precond, pSolve, MODIFIED_GS, 
        maxKDim, maxRestart, epliFactor, 0.0, ilu ); 
    if( flag != DST_SUCCESS )
    {
        printf( "IDAReInitSpgmr failed\n" );
	return 1; 
    }
                                // Loop over output times, call 'IDASolve'
/*    flag = IDASolve( dae->mem, tout, tout, &t, yVector, ypVector, itask );
    if( flag < 0 )
    {
	printf( "IDASolve failed: %d\n", flag );
	return 1;
    }
*/
    while( t < tout )
    {
        flag = IDASolve( dae->mem, tout, tout, &t, yVector, ypVector, itask );	
#ifdef DEBUG_PRINT
//printf( "leaf diff: t=%12.4e\n", t );
#endif
	if( flag < 0 )
	{
	    printf( "IDASolve failed: %d\n", flag );
	    return 1;
	}
	counter ++;
	if( counter > loop )
	{
            N_VConst_Serial( 0, yVector );
	    break;
	}
    }
//*/    
    return DST_SUCCESS;
}
               // for 4th-order DT
int leafNodeDiffSim_DT4( void *kernelData, void *daeData, void *iluData )
{
    Diff_Kernel_ kernel;
    DAE_ dae;
    ILU_ ilu;
    N_Vector yVector, ypVector, constraintsVector;
    realtype t, t0, tout, loop;
    int flag, counter = 0;
        
    kernel = (Diff_Kernel_)kernelData;
    ilu = (ILU_)iluData;
    dae = (DAE_)daeData;
    yVector = dae->yVector;
    ypVector = dae->ypVector;
    constraintsVector = dae->constraintsVector;
    t0 = dae->t0;
    t = dae->t;
    tout = dae->tout;
    loop = dae->loop;
                                // Re-initializes the DAE system 
    setInitialProfiles_DT4( kernel, dae );
                                // Re-initialize IDA and IDASPGMR 
				// NULL arguments are 'id', 'errfp', 'machEnv'.
    flag = IDAReInit( dae->mem, res_DT4, kernel, t0, yVector, 
        ypVector, itol, &rtol, &atol, NULL, constraintsVector, NULL, 
	optIn, iopt, ropt, NULL );
    if( flag != DST_SUCCESS )
    {
        printf( "IDAReInit failed\n" );
	return 1; 
    }
     
    flag = IDAReInitSpgmr( dae->mem, precond, pSolve, MODIFIED_GS, 
        maxKDim, maxRestart, epliFactor, 0.0, ilu ); 
    if( flag != DST_SUCCESS )
    {
        printf( "IDAReInitSpgmr failed\n" );
	return 1; 
    }
                                // Loop over output times, call 'IDASolve'
/*    flag = IDASolve( dae->mem, tout, tout, &t, yVector, ypVector, itask );
    if( flag < 0 )
    {
	printf( "IDASolve failed: %d\n", flag );
	return 1;
    }
*/
    while( t < tout )
    {
        flag = IDASolve( dae->mem, tout, tout, &t, yVector, ypVector, itask );	
#ifdef DEBUG_PRINT
//printf( "leaf diff: t=%12.4e\n", t );
#endif
	if( flag < 0 )
	{
	    printf( "IDASolve failed: %d\n", flag );
	    return 1;
	}
	counter ++;
	if( counter > loop )
	{
            N_VConst_Serial( 0, yVector );
	    break;
	}
    }
//*/    
    return DST_SUCCESS;
}


/***********************************************************************
 * Initializes the DAE system to a consistent condition
 ***********************************************************************/
void setInitialProfiles( void *kernelData, void *daeData ) 
{
    int neq, i;
    realtype *sf, *ypData, *rData, *yData, t;
    N_Vector yVector, ypVector, rVector, constraintsVector, idVector;
    Diff_Kernel_ kernel;
    DAE_ dae;
        
    kernel = (Diff_Kernel_)kernelData;
    sf = kernel->sf;
    neq = kernel->neq;
	
    dae = (DAE_)daeData;
    yVector = dae->yVector;
    ypVector = dae->ypVector;
    rVector = dae->rVector;
    constraintsVector = dae->constraintsVector;
    idVector = dae->idVector;
	
    t = 0;
    yData = NV_DATA_S( yVector );
    ypData = NV_DATA_S( ypVector );
    rData = NV_DATA_S( rVector );
                                // Set constraints to all 1's, i.e., 
                                // solution values are constrained to be >= 0. 
                                // since concentration can't be negative values
    N_VConst_Serial( 1, constraintsVector );
    N_VConst_Serial( 1, idVector );
    N_VConst_Serial( 0, yVector );
    N_VConst_Serial( 0, ypVector );
    N_VConst_Serial( 0, rVector );
                                // initial distribution of y: only the seed node
				// has concentration value, other nodes are zero
    yData[kernel->seedID] = initPeak;
//    initialGaussianPeak( yVector, kernel );	       
                                    // initial distribution of y'
    res( neq, t, yVector, ypVector, rVector, kernel );

    for( i = 0; i < neq; ++ i )
        ypData[i] = -rData[i] / sf[i];
}	
               // for 4th-order DT
void setInitialProfiles_DT4( void *kernelData, void *daeData )
{
    int neq, i;
    realtype *sf, *ypData, *rData, *yData, t;
    N_Vector yVector, ypVector, rVector, constraintsVector, idVector;
    Diff_Kernel_ kernel;
    DAE_ dae;
        
    kernel = (Diff_Kernel_)kernelData;
    sf = kernel->sf;
    neq = kernel->neq;
	
    dae = (DAE_)daeData;
    yVector = dae->yVector;
    ypVector = dae->ypVector;
    rVector = dae->rVector;
    constraintsVector = dae->constraintsVector;
    idVector = dae->idVector;
	
    t = 0;
    yData = NV_DATA_S( yVector );
    ypData = NV_DATA_S( ypVector );
    rData = NV_DATA_S( rVector );
                                // Set constraints to all 1's, i.e., 
                                // solution values are constrained to be >= 0. 
                                // since concentration can't be negative values
    N_VConst_Serial( 1, constraintsVector );
    N_VConst_Serial( 1, idVector );
    N_VConst_Serial( 0, yVector );
    N_VConst_Serial( 0, ypVector );
    N_VConst_Serial( 0, rVector );
                                // initial distribution of y: only the seed node
				// has concentration value, other nodes are zero
    yData[kernel->seedID] = initPeak;
               // initial distribution of y: Gaussian
//    initialGaussianPeak_inner( yVector, kernel );	       
                                    // initial distribution of y'
    res_DT4( neq, t, yVector, ypVector, rVector, kernel );

    for( i = 0; i < neq; ++ i )
        ypData[i] = -rData[i] / sf[i];
}	

/***********************************************************************
 * Residual function	
 ***********************************************************************/
int res( int Neq, realtype t, N_Vector yVector, N_Vector ypVector, 
    N_Vector rVector, void *rdata )
{
    int i, k, neq;
    realtype *yData, *ypData, *rData, *aa, *sf;
    int *ia, *ja;
    Diff_Kernel_ kernel;
	
    kernel = (Diff_Kernel_)rdata;
    aa = kernel->aa;
    ia = kernel->ia;
    ja = kernel->ja;
    sf = kernel->sf;
    neq = kernel->neq;
		
    yData = NV_DATA_S( yVector );
    ypData = NV_DATA_S( ypVector );
    rData = NV_DATA_S( rVector );

    for( i = 0; i < neq; ++ i )
    {
	rData[i] = -ypData[i] * sf[i];
	for( k = ia[i]-1; k < ia[i+1]-1; ++ k )
	    rData[i] += aa[k] * yData[ja[k]-1];
    }
    return DST_SUCCESS;
}

/***********************************************************************
 * Residual function for 4th-order tensor case.	(Dirichlet boundary)
 ***********************************************************************/
int res_DT4( int Neq, realtype t, N_Vector yVector, N_Vector 
    ypVector, N_Vector rVector, void *rdata )
{
    int i, k, neq, ii, jj, kk, nx, ny, nz;
    realtype *yData, *ypData, *rData, *aa, *sf;
    int *ia, *ja;
    Diff_Kernel_ kernel;
	
    kernel = (Diff_Kernel_)rdata;
    aa = kernel->aa;
    ia = kernel->ia;
    ja = kernel->ja;
    sf = kernel->sf;
    neq = kernel->neq;
    nx = kernel->nx;
    ny = kernel->ny;
    nz = kernel->nz;
		
    yData = NV_DATA_S( yVector );
    ypData = NV_DATA_S( ypVector );
    rData = NV_DATA_S( rVector );

    for( i = 0; i < neq; ++ i )
    {
               // For each boundary point: res_i = y_i.
        nodeIndex_decode( i, &ii, &jj, &kk, nx, ny );
	if( ii==0 || jj==0 || kk==0 || ii==nx-1 || jj==ny-1 || kk==nz-1 )
	{
	    rData[i] = yData[i];
	    continue;
	}
               // interior points
	rData[i] = -ypData[i] * sf[i];
	for( k = ia[i]-1; k < ia[i+1]-1; ++ k )
	    rData[i] += aa[k] * yData[ja[k]-1];
    }

    return DST_SUCCESS;
}


/***********************************************************************
 * preconditioning function
 ***********************************************************************/
int precond( int Neq, realtype t, N_Vector yVector, N_Vector ypVector, 
    N_Vector rVector, realtype cj, ResFn res, void *rdata, void *pdata, 
    N_Vector ewtVector, N_Vector constraintsVector, realtype h, realtype uround,
    long int *nrePtr, N_Vector tempv1, N_Vector tempv2, N_Vector tempv3 )
{
    int i, j;
    realtype *a;                // Newton iteration matrix
    realtype *aa, *sf;          // variables for matrix A
    int neq, nnz, *ia, *ja;
                                // variables for ILU data
    realtype *alu;
    int *jlu, *ju;
#ifdef ILU0
    int *jr, ierr;
#endif
#ifdef ILUT
    int lfil, iwk, ierr, *jr, *jwu, *jwl;
    realtype tol, *wu, *wl;
#endif
		
    Diff_Kernel_ kernel;
    ILU_ ilu;
	
    kernel = (Diff_Kernel_)rdata;
    neq = kernel->neq;
    nnz = kernel->nnz;
    aa = kernel->aa;
    ia = kernel->ia;
    ja = kernel->ja;
    sf = kernel->sf;
	
    ilu = (ILU_)pdata;
    alu = ilu->alu;
    jlu =ilu->jlu;
    ju = ilu->ju;

#ifdef ILU0
    jr = ilu->jr;
    ierr = ilu->ierr;
#endif

#ifdef ILUT
    lfil = ilu->lfil;
    tol = ilu->tol;
    iwk = ilu->iwk;
    ierr = ilu->ierr;
    jr = ilu->jr;
    jwu = ilu->jwu;
    jwl = ilu->jwl;
    wu = ilu->wu;
    wl = ilu->wl;
#endif  
                                // Newton iteration matrix 
    a = (realtype *)malloc( nnz * sizeof(realtype) );
    memcpy( a, aa, nnz * sizeof(realtype) );
                                // Approximate the system Jacobian matrix:
				// J = dF/dy + cj * dF/dy', where F(t,y,y')=0.
				// In this problem: J = A - cj*I*scalefactor.
    for( i = 0; i < neq; ++ i )
        for( j = ia[i]-1; j < ia[i+1]-1; ++ j )
            if( ja[j] == i + 1 )
	    {
                a[j] -= cj * sf[i];
		break;
            }
                                // use ILU(0) as preconditoner
#ifdef ILU0
    #ifdef FORTRAN_UNDERSCORE
        ilu0_( &neq, a, ja, ia, alu, jlu, ju, jr, &ierr );
    #else
        ilu0( &neq, a, ja, ia, alu, jlu, ju, jr, &ierr );
    #endif
#endif
                                // use ILUT as preconditoner
#ifdef ILUT
    #ifdef FORTRAN_UNDERSCORE
        ilut_( &neq, a, ja, ia, &lfil, &tol, alu, jlu, ju, 
            &iwk, wu, wl, jr, jwl, jwu, &ierr );
    #else
        ilut( &neq, a, ja, ia, &lfil, &tol, alu, jlu, ju, 
            &iwk, wu, wl, jr, jwl, jwu, &ierr );
    #endif
#endif

    free( a );
    return DST_SUCCESS;
}


/***********************************************************************
 * function for solving the preconditioning system
 ***********************************************************************/
int pSolve( int Neq, realtype t, N_Vector yVector, N_Vector ypVector, 
    N_Vector rVector, realtype cj, ResFn res, void *rdata, void *pdata, 
    N_Vector ewtVector, realtype delta, N_Vector rhsVector, N_Vector zVector, 
    long int *nrePtr, N_Vector tempv )
{
    realtype *alu, *rhsData, *zData;
    int *jlu, *ju, neq;
    Diff_Kernel_ kernel;
    ILU_ ilu;
	
    kernel = (Diff_Kernel_)rdata;
    neq = kernel->neq;
    ilu = (ILU_)pdata;
    alu = ilu->alu;
    jlu = ilu->jlu;
    ju = ilu->ju;
	
    zData = NV_DATA_S( zVector );
    rhsData = NV_DATA_S( rhsVector );
                                // call FORTRAN subroutine lusol
                                // subroutine lusol (n, y, x, alu, jlu, ju)
#ifdef FORTRAN_UNDERSCORE
    lusol_( &neq, rhsData, zData, alu, jlu, ju );
#else
    lusol( &neq, rhsData, zData, alu, jlu, ju );
#endif

    return DST_SUCCESS;
}

void initialGaussianPeak( N_Vector yVector, void *kernelData )
{
    Diff_Kernel_ kernel = (Diff_Kernel_)kernelData;
    float *yData = NV_DATA_S( yVector );
    int i, j, k, nx, ny, nz, nxy;
    float dx, dy, dz;
    float sigmaX, sigmaY, sigmaZ, muX, muY, muZ, gX, gY, gZ;
    double aX, aY, aZ;
    float sigma = 0.05;

    nx = kernel->nx; ny = kernel->ny; nz = kernel->nz;
    nxy = nx*ny;
    dx = kernel->dx; dy = kernel->dy; dz = kernel->dz;
    sigmaX = sigma;
    sigmaY = sigma;
    sigmaZ = sigma;
    muX = dx * kernel->seedI;
    muY = dy * kernel->seedJ;
    muZ = dz * kernel->seedK;

    aX = 1.0/(sigmaX*sqrt(2*M_PI));
    aY = 1.0/(sigmaY*sqrt(2*M_PI));
    aZ = 1.0/(sigmaZ*sqrt(2*M_PI));
               // only for inner nodes, 
	       // since value on boundary nodes is zero.
    for( i = 0; i < nx; ++ i )
    for( j = 0; j < ny; ++ j )
    for( k = 0; k < nz; ++ k )
    {
        gX = aX * exp( -(i*dx-muX)*(i*dx-muX) / (2*sigmaX*sigmaX) );
        gY = aY * exp( -(j*dy-muY)*(j*dy-muY) / (2*sigmaY*sigmaY) );
        gZ = aZ * exp( -(k*dz-muZ)*(k*dz-muZ) / (2*sigmaZ*sigmaZ) );
        yData[ nodeIndex_zero(i,j,k,nx,nxy) ] = gX*gY*gZ;	
    }
}
               // only for inner nodes
void initialGaussianPeak_inner( N_Vector yVector, void *kernelData )
{
    Diff_Kernel_ kernel = (Diff_Kernel_)kernelData;
    float *yData = NV_DATA_S( yVector );
    int i, j, k, nx, ny, nz, nxy;
    float dx, dy, dz;
    float sigmaX, sigmaY, sigmaZ, muX, muY, muZ, gX, gY, gZ;
    double aX, aY, aZ;
    float sigma = 0.05;

    nx = kernel->nx; ny = kernel->ny; nz = kernel->nz;
    nxy = nx*ny;
    dx = kernel->dx; dy = kernel->dy; dz = kernel->dz;
    sigmaX = sigma;
    sigmaY = sigma;
    sigmaZ = sigma;
    muX = dx * kernel->seedI;
    muY = dy * kernel->seedJ;
    muZ = dz * kernel->seedK;

    aX = 1.0/(sigmaX*sqrt(2*M_PI));
    aY = 1.0/(sigmaY*sqrt(2*M_PI));
    aZ = 1.0/(sigmaZ*sqrt(2*M_PI));
               // only for inner nodes, 
	       // since value on boundary nodes is zero.
    for( i = 1; i < nx-1; ++ i )
    for( j = 1; j < ny-1; ++ j )
    for( k = 1; k < nz-1; ++ k )
    {
        gX = aX * exp( -(i*dx-muX)*(i*dx-muX) / (2*sigmaX*sigmaX) );
        gY = aY * exp( -(j*dy-muY)*(j*dy-muY) / (2*sigmaY*sigmaY) );
        gZ = aZ * exp( -(k*dz-muZ)*(k*dz-muZ) / (2*sigmaZ*sigmaZ) );
        yData[ nodeIndex_zero(i,j,k,nx,nxy) ] = gX*gY*gZ;	
    }
}


/***********************************************************************
 * save the diffusion profile data to file.
 ***********************************************************************/
void saveDiffProfileVTK( void *kernelData, void *daeData )
{
    FILE *fp;
	
    Diff_Kernel_ kernel;
    DAE_ dae;
    int id, num, nx, ny, nz;
    realtype *yData, dx, dy, dz;

    kernel = (Diff_Kernel_)kernelData;
    dae = (DAE_)daeData;
    num = kernel->neq;
    nx = kernel->nx;
    ny = kernel->ny;
    nz = kernel->nz;
    dx = kernel->dx;
    dy = kernel->dy;
    dz = kernel->dz;
    
    yData = NV_DATA_S( dae->yVector );
                                // create a file for writing data
    if( (fp = fopen( FILE_DIFF_PROFILE, "w" )) == NULL )
    {
        fprintf( stderr, "cannot create file: %s\n", FILE_DIFF_PROFILE );
        exit( 1 );
    }
			        // write header info to data file
    fprintf( fp, "# vtk DataFile Version 2.0\n" );
printf( ">> save diffusion profile (concentration distribution)...\n" );
printf( "# vtk DataFile Version 2.0\n" );
    fprintf( fp, "Diffusion Profile\n" );
    fprintf( fp, "ASCII\n" );
    fprintf( fp, "DATASET STRUCTURED_POINTS\n" );
    fprintf( fp, "DIMENSIONS %d %d %d\n", nx, ny, nz );
    fprintf( fp, "ORIGIN 0 0 0\n" );
    fprintf( fp, "SPACING %f %f %f\n", dx, dy, dz );
    fprintf( fp, "POINT_DATA %d\n", num );
    fprintf( fp, "SCALARS scalars float 1\n" );
    fprintf( fp, "LOOKUP_TABLE default\n" );
    for( id = 0; id < num; id ++ )
        fprintf( fp, "%12.4e\n", (float)(yData[id]) );

    fclose( fp );
    exit( 0 );
}
