/***********************************************************************
 * File: 	dst.c	                                    
 * Programmer:	Ning Kang		
 * Date:	08/15/2005			
 * --------------------------------------------------------------------
 * Description: 							
 * - This file is the driver for the program.
 * - written by using IDA in the SUNDIALS package, one of ACTS 
 * toolkits.						
 * - serial version.		
 * - int and float version  
 ***********************************************************************/
 

#include "dst.h"

static int diffsimFT( void *argsTable );
static int diffsimFT_DT4( void *argsTable );

/**********************************************************************
 * the interface function to AFNI plugin
 **********************************************************************/ 
Args_Table_ diffsimFT_interface( int nvox, int *dims, float *voxSize, 
    float *tensor[], float *fa, float *md, float *eigvec[], float *threshold,
    float *dwi_b0, int numSeeds, int *seedsXYZ[], 
    char *seedsFileName, int targetSize, int *targetXYZ[] )
{
    Args_Table_ args;
    int i;
                   // dataset stuff
    args = (Args_Table_)malloc( sizeof( *args ) );
    args->nvox = nvox;
    args->nx = dims[0];
    args->ny = dims[1];
    args->nz = dims[2];
    args->dx = (float)fabs(voxSize[0]);
    args->dy = (float)fabs(voxSize[1]);
    args->dz = (float)fabs(voxSize[2]);
    args->dxOrig = voxSize[0];
    args->dyOrig = voxSize[1];
    args->dzOrig = voxSize[2];
    args->dt[0] = tensor[0];
    args->dt[1] = tensor[2];
    args->dt[2] = tensor[5];
    args->dt[3] = tensor[1];
    args->dt[4] = tensor[3];
    args->dt[5] = tensor[4];
    args->fa = fa;
    args->md = md;
    args->dwi_b0 = dwi_b0;
    for( i = 0; i < 3; ++ i ) args->eigvec[i] = eigvec[i];

    strcpy( args->tractbase, FT_PREFIX );
                   // specify tract file name
    sprintf( args->tractFileName, "%s%s", seedsFileName, FT_VTK_SUFFIX );
                   // specify the file name for fiber stats
    sprintf(args->fiberStatsFileName, "%s%s", seedsFileName, FT_STATS_SUFFIX); 
                   // pick up tracking thresholds
    args->tensorFA = threshold[0];
    args->distRatio = threshold[1];
    args->ipVectors = threshold[2];
    args->ipSeedVecNode = threshold[3];
    args->diffIntegTime = threshold[4];
                   // b0 thresholds   
    args->b0_lo = threshold[5];
    args->b0_hi = threshold[6];
    if( args->b0_lo > args->b0_hi ) ELEM_SWAP( args->b0_lo, args->b0_hi );
    if( args->b0_hi == 0 ) // b0 thresholds are not specified
	args->b0_hi = DWI_B0_MAX;
                   // starting points in drawing ROI and target points
    args->roiSize = numSeeds;
    args->roiSizeTarget = targetSize;  
    for( i = 0; i < 3; ++ i )
    {
        args->roiPoints[i] = seedsXYZ[i];
	if( targetSize > 0 ) args->roiPtsTarget[i] = targetXYZ[i];
    }	

/*
   // save tenosr components to files
   FILE *fp; int i;
   char *fname[6]={"dxx","dyy","dzz","dxy","dxz","dyz"};
    for( i =0;i<6;++i)
    { 
      fp=fopen(fname[i],"wb");
      fwrite( args->dt[i],sizeof(float),nvox,fp);
      fclose(fp);
    }
    exit(0);  
 
    // save b0 to vtk file
    FILE *fp = fopen( "b0_shark.vtk","w");
    fprintf( fp, "# vtk DataFile Version 2.0\n" );
    fprintf( fp, "b0\n" );
    fprintf( fp, "ASCII\n" );
    fprintf( fp, "DATASET STRUCTURED_POINTS\n" );
    fprintf( fp, "DIMENSIONS %d %d %d\n", 256, 128, 128 );
    fprintf( fp, "ORIGIN 0 0 0\n" );
    fprintf( fp, "SPACING %f %f %f\n", 0.156, 0.156, 0.156 );
    fprintf( fp, "POINT_DATA %d\n", 256*128*128 );
    fprintf( fp, "SCALARS scalars float 1\n" );
    fprintf( fp, "LOOKUP_TABLE default\n" );
    for( i = 0; i < 4194304; i ++ )
        fprintf( fp, "%12.4e\n", dwi_b0[i] );
    fclose( fp );
    exit( 0 );
*/


    if( diffsimFT( args ) != 0 )
    {
        fprintf( stderr, "Something went wrong in DST fiber tracking\n" );
	free( args );
	return NULL;
    }
    return args;
}


/**********************************************************************
 * do the actual fiber tracking stuff              
 **********************************************************************/ 
int diffsimFT( void *argsTable )
{
    Args_Table_ args = (Args_Table_)argsTable;
    Diff_Kernel_ kernel;        // The user data structures.
    ILU_ ilu;
    DAE_ dae;
    DTI_ dti;

#ifdef TIMER
    time_t t1, t2;    
    (void)time( &t1 );
#endif
    
                                // Allocate memory for data structures.
    if( (kernel = kernelData_new( args, NEQ, NNZ_MAX, NX_KERNEL, NY_KERNEL,
        NZ_KERNEL, DIFFSIZ )) == NULL ) return 1;
    if( (dti = dtiData_new( args )) == NULL ) return 1;
    ilu = iluData_new( NEQ, NNZ_MAX, IWK );
    dae = daeData_new( args );
#ifdef DEBUG_PRINT
printf(">> memory allocation: done\n");
#endif
                                // initialize grid voxels 
    initDTIvoxel( dti );
                                // Get coefficient matrix structure for 
				// solving the diffusion equation.
    getCoeffMatrixStructure( kernel ); 
#ifdef DEBUG_PRINT
printf( ">> get coeff matrix structure: done\n" );
#endif

                       // create directory to store files of computed tracts
//    system( "mkdir DST_DATA_DIR" );
    if( mkdir( DST_DATA_DIR, 0755 ) != 0 && errno != EEXIST )
    {
        fprintf( stderr, "Fail to create directory: %s\n", DST_DATA_DIR );
	return 1; 
    }

printf( ">> start fiber tracking....\n");

    if( fiberTracking( kernel, ilu, dae, dti, args ) != 0 ) return 1;
#ifdef DEBUG_PRINT
printf( ">> fiber tracking: done\n" );
#endif
    
#ifdef TIMER
    (void)time( &t2 ); 
    printf( "----------> fiber tracking time: %5d seconds\n", (int)(t2-t1) );
#endif
                               // free memory
    kernelData_free( kernel );
    iluData_free( ilu );
    daeData_free( dae );
    dtiData_free( dti );

printf(">> start converting fibers to vtk format\n");
                       // convert fibers to vtk format and save fiber stats
    if( fiberVTK( args, DST_DATA_DIR, SPLINE_ORDER, NUM_PP ) != 0 ) return -1;
printf(">> convert fibers to vtk format and save fiber stats: done\n");

    return 0;
}

/////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////
               // interface function for rank-4 DT
Args_Table_ diffsimFT_DT4_interface( int nvox, int *dims, 
    float *voxSize, float *tensor[], float *tensor4th[], float *gfa, 
    float *threshold, int numSeeds, int *seedsXYZ[], char *seedsFileName )
{
    Args_Table_ args;
    int i;
                   // dataset stuff
    args = (Args_Table_)malloc( sizeof( *args ) );
    args->nvox = nvox;
    args->nx = dims[0];
    args->ny = dims[1];
    args->nz = dims[2];
    args->dx = (float)fabs(voxSize[0]);
    args->dy = (float)fabs(voxSize[1]);
    args->dz = (float)fabs(voxSize[2]);
    args->dxOrig = voxSize[0];
    args->dyOrig = voxSize[1];
    args->dzOrig = voxSize[2];
    args->dt[0] = tensor[0];
    args->dt[1] = tensor[2];
    args->dt[2] = tensor[5];
    args->dt[3] = tensor[1];
    args->dt[4] = tensor[3];
    args->dt[5] = tensor[4];
    args->fa = gfa;
    args->md = gfa;
    args->dwi_b0 = NULL;
    for( i = 0; i < DT4_ELEM; ++ i )
      args->dt4[i] = tensor4th[i];    

    strcpy( args->tractbase, FT_PREFIX );
                   // specify tract file name
    sprintf( args->tractFileName, "%s%s", seedsFileName, 
      FT_DT4_VTK_SUFFIX );
                   // specify the file name for fiber stats
    sprintf(args->fiberStatsFileName, "%s%s", seedsFileName, 
      FT_DT4_STATS_SUFFIX); 
                   // pick up tracking thresholds
    args->dt4_GFA = threshold[0];
    args->dt4_distRatio = threshold[1];
    args->dt4_transAngle = threshold[2];
    args->dt4_integTime = threshold[3];
                   // starting points in drawing ROI and target points
    args->roiSize = numSeeds;
    args->roiSizeTarget = 0;  
    for( i = 0; i < 3; ++ i )
        args->roiPoints[i] = seedsXYZ[i];

/*
   // save rank-2 tenosr components to files
   FILE *fp; 
   char *fname[6]={"dxx","dyy","dzz","dxy","dxz","dyz"};
    for( i =0;i<6;++i)
    { 
      fp=fopen(fname[i],"wb");
      fwrite( args->dt[i],sizeof(float),nvox,fp);
      fclose(fp);
    }
    exit(0);  
*/
    if( diffsimFT_DT4( args ) != 0 )
    {
        fprintf( stderr, "Something went wrong in DST fiber tracking\n" );
	free( args );
	return NULL;
    }
    return args;
}
               // for 4th-order DT
int diffsimFT_DT4( void *argsTable )
{
    Args_Table_ args = (Args_Table_)argsTable;
    Diff_Kernel_ kernel;        // The user data structures.
    ILU_ ilu;
    DAE_ dae;
    DTI_ dti;

#ifdef TIMER
    time_t t1, t2;    
    (void)time( &t1 );
#endif
    
                                // Allocate memory for data structures.
    if( (kernel = kernelData_new( args, NEQ_DT4, NNZ_MAX_DT4, NX_KERNEL_DT4,
        NY_KERNEL_DT4, NZ_KERNEL_DT4, DIFFSIZ_DT4 )) == NULL ) return 1;
    if( (dti = dtiData_new_DT4( args )) == NULL ) return 1;
    ilu = iluData_new( NEQ_DT4, NNZ_MAX_DT4, IWK_DT4 );
    dae = daeData_new_DT4( args );
#ifdef DEBUG_PRINT
printf(">> memory allocation: done\n");
#endif
                                // initialize grid voxels 
    initDTIvoxel( dti );
#ifdef DEBUG_PRINT
printf(">> voxel initialization: done\n");
#endif
                                // Get coefficient matrix structure for 
				// solving the diffusion equation.
    getCoeffMatrixStructure_DT4( kernel ); 
#ifdef DEBUG_PRINT
printf( ">> get coeff matrix structure: done\n" );
#endif

                       // create directory to store files of computed tracts
//    system( "mkdir DST_DATA_DIR" );
    if( mkdir( DST_DATA_DIR, 0755 ) != 0 && errno != EEXIST )
    {
        fprintf( stderr, "Fail to create directory: %s\n", DST_DATA_DIR );
	return 1; 
    }

printf( ">> start fiber tracking....\n");

    if( fiberTracking_DT4( kernel, ilu, dae, dti, args ) != 0 ) 
        return 1;
#ifdef DEBUG_PRINT
printf( ">> fiber tracking: done\n" );
#endif
    
#ifdef TIMER
    (void)time( &t2 ); 
    printf( "----------> fiber tracking time: %5d seconds\n", (int)(t2-t1) );
#endif
                               // free memory
    kernelData_free( kernel );
    iluData_free( ilu );
    daeData_free( dae );
    dtiData_free( dti );

printf(">> start converting fibers to vtk format\n");
                       // convert fibers to vtk format and save fiber stats
    if( fiberVTK( args, DST_DATA_DIR, SPLINE_ORDER, NUM_PP ) != 0 ) return -1;
printf(">> convert fibers to vtk format and save fiber stats: done\n");

    return 0;
}
//////////////////end of main()




