/***********************************************************************
 * File:        fibertract.c
 * Programmer:  Ning Kang @CS.UKY
 * Version of:  7/29/2004
 * ---------------------------------------------------------------------
 * Description:
 * - Serial version.
 * - The functions in this file are used to do fiber tracing.
 * - This file contains the following functions:
 *   int fiberTracking ()
 *   int setTractFile ()
 *   int buildGraph ()
 *   int traverseGraph ()
 *   Voxel_ getDiffFront ()
 *   int isCollinear ()
 *   realtype distance ()
 *   Kernel_Voxel_ sortDistance ()
 *   void insertionSort ()
 *   void insert ()
 *   void insertSeed ()
 *   void setRootNode ()
 *   void setSeedNode ()
 *   void cuttingPlaneCoeff ()
 *   int inForwardDiffVol () 
 ***********************************************************************/
 

#include "dst.h"

static int buildGraph( void *kernelData, void *iluData, void *daeData, 
  void *dtiData );
static int traverseGraph( void *dtiData, char *filename, int *numNodes, 
  int *numPaths );
static Voxel_ getDiffFront( void *kernelData, void *dtiData, void *daeData, 
  int frontMaxNodes );
static int isCollinear( void *voxelSeed, void *voxelNode, void *dtiData, 
  void *kernelData, realtype *ipSeedVecNode );
static int setRootNode( void *dtiData, int *roiPoints[], int index );
    
/***********************************************************************
 * Do fiber tracking.
 ***********************************************************************/
int fiberTracking( void *kernelData, void *iluData, void *daeData, 
    void *dtiData, void *argsTable )
{
    Args_Table_ args;
    Diff_Kernel_ kernel;           
    ILU_ ilu;
    DAE_ dae;
    DTI_ dti;
    int i, seqno = 0, buildRet, numNodes, numPaths;
    char tractfile[STR_LEN], *tractbase;
//    char indexfile[STR_LEN];

    kernel = (Diff_Kernel_)kernelData;
    ilu = (ILU_)iluData;
    dae = (DAE_)daeData;
    dti = (DTI_)dtiData;
    args = (Args_Table_)argsTable;
    
    tractbase = args->tractbase;
    args->totalNumNodes = 0;
    args->totalNumPaths = 0;

    for( i = 0; i < args->roiSize; ++ i )
    {
        if( setRootNode( dti, args->roiPoints, i ) != 0 ) continue;
#ifdef DEBUG_PRINT
printf( "--->root seed %4d: %3d %3d %3d\n", i,dti->seedI, dti->seedJ, 
 dti->seedK );
#endif
                        // build the graph structure by diffusion simulation 
	buildRet = buildGraph( kernel, ilu, dae, dti );
        if( buildRet == NO_FRONT_AT_ROOT ) 
	    continue;
	else if( buildRet != 0 )
	    return 1;
			// traverse the graph and output fiber paths to file 
        if( setTractFile( tractfile, STR_LEN, DST_DATA_DIR, tractbase, seqno )
	    != 0 )
	    return 1;
	if( traverseGraph( dti, tractfile, &numNodes, &numPaths ) != 0 ) 
	    return 1;
                   // make sure the size of the output file is not zero
        if( numNodes > 0 )
	{
            args->totalNumNodes += numNodes;	    
	    args->totalNumPaths += numPaths;
	    seqno ++;
	}    
                        /* reset the voxel data structure */
        resetDTIvoxel( dti );
    }
                       // number of generated files for computed tracks
    args->numTractFiles = seqno;

    return 0;
}    

/***********************************************************************
 * Set the file name and path for saving fiber tract data.
 *********************************************************************/
int setTractFile( char *file, unsigned int len, char *dir, char *prefix, 
    int seqno )
{
    int quot, digits, ten = 10;
                        /* get the number of digits in seqno */
    if( seqno == 0 ) 
        digits = 1;
    else
    {
        digits = 0;
        quot = seqno;
        while(  quot != 0 )
        {
            quot = (int)(quot/ten);
	    digits ++;
        }
    }
    
    if( ( strlen(dir) + strlen(prefix) + digits ) >= len )
    {
        fprintf( stderr, "The maximum string length of %u exceeded\n", len );
        return 1;
    }
     
    memset( file, '\0', len * sizeof(char) );
    sprintf( file, "%s%s%d", dir, prefix, seqno );
    return 0;
}


/***********************************************************************
 * Build the graph structure for fiber tracing.
 ***********************************************************************/
int buildGraph( void *kernelData, void *iluData, void *daeData, void *dtiData )
{
    Diff_Kernel_ kernel;           
    ILU_ ilu;
    DAE_ dae;
    DTI_ dti;
    Voxel_ node, front = NULL, list = NULL;
    
    kernel = (Diff_Kernel_)kernelData;
    ilu = (ILU_)iluData;
    dae = (DAE_)daeData;
    dti = (DTI_)dtiData;
               // root seed
    if( getCoeffMatrixEntry( kernel, dti ) != 0 ) return 1;
    if( rootNodeDiffSim( kernel, dae, ilu ) != 0 ) return 1;
    diffThreshold( kernel, dae );
    cuttingPlaneCoeff( dti, kernel );    
    front = getDiffFront( kernel, dti, dae, FRONT_MAX_NODES_ROOT );
    
    if( front != NULL ) 
        node = list = front;
    else 
    {
        fprintf( stderr, "No diffusion front at root\n" );
	return NO_FRONT_AT_ROOT;
    }
               // the generated seeds
    while( node != NULL )
    {
        if( node->color == BLACK )
        {
#ifdef DEBUG_PRINT_SEED_NODE
printf( "seed node: %3d %3d %3d\n", node->i, node->j, node->k );
#endif
            setSeedNode( dti, node );
            if( getCoeffMatrixEntry( kernel, dti ) != 0 ) return 1;
	    if( leafNodeDiffSim( kernel, dae, ilu ) != 0 ) return 1;
            diffThreshold( kernel, dae );
            cuttingPlaneCoeff( dti, kernel );    
            front = getDiffFront( kernel, dti, dae, FRONT_MAX_NODES );
	    if( front != NULL ) insertSeed( list, front );
	}
        node = node->next;
    }
    
    return 0; 
}


/***********************************************************************
 * Traverse the graph to print out fiber pathways to file.
 ***********************************************************************/
int traverseGraph( void *dtiData, char *filename, int *numNodes, int *numPaths )
{
    FILE *fp;		        // File pointer for fiber trace output.
    DTI_ dti;
    Voxel_ *voxel, node; 
    int id, num, counter;
    
    dti = (DTI_)dtiData;
    num = dti->num;
    voxel = dti->voxel;

    if( ( fp = fopen( filename, "w" ) ) == NULL )
    {
        fprintf( stderr, "Cannot create file %s\n", filename );
	return 1;
    }
    
    *numNodes = 0;
    counter = 0;
#ifdef SET_END_ROI
    for( id = 0; id < num; ++ id )
    if( voxel[id]->child == NULL && voxel[id]->parent != NULL )
    if( voxel[id]->level >= MIN_LEVEL )
    {
	node = voxel[id]; 
	                // if node is not in END-ROI, skip to the next one
	if( node->i < ROI_X_LO || node->i > ROI_X_HI ||
	    node->j < ROI_Y_LO || node->j > ROI_Y_HI ||
	    node->k < ROI_Z_LO || node->k > ROI_Z_HI )
	    continue;
	
	counter ++;    
        while( node != NULL ) 
	{
	    fprintf( fp, "%6d%4d%4d%4d%10.4f%10.4f\n", counter, node->i, 
	        node->j, node->k, dti->fa[node->id], dti->md[node->id] ); 
            (*numNodes) ++;
	    node = node->parent;
	}
    }
#else
    for( id = 0; id < num; ++ id )
    if( voxel[id]->child == NULL && voxel[id]->parent != NULL )
    if( voxel[id]->level >= MIN_LEVEL )
    {
	node = voxel[id]; 
	counter ++;    
        while( node != NULL ) 
	{
	    fprintf( fp, "%6d%4d%4d%4d%10.4f%10.4f\n", counter, node->i, 
	        node->j, node->k, dti->fa[node->id], dti->md[node->id] ); 
            (*numNodes) ++;
	    node = node->parent;
	}
    }
#endif    
    *numPaths = counter;
    fclose( fp );
    return 0;
}


/***********************************************************************
 * Get the diffusion front.
 ***********************************************************************/
Voxel_ getDiffFront( void *kernelData, void *dtiData, void *daeData, 
    int frontMaxNodes )
{
    Diff_Kernel_ kernel;
    Kernel_Voxel_ *kernelVoxel;
    DAE_ dae;
    DTI_ dti;
    Voxel_ *voxel;
    realtype *yData;
    int m, id, gid, i, j, k, 
        *neighbors, neiID, neiGID;      // indeces of neighbor nodes 
    int seedID, neq, nx, ny, nz, nxy, dtxyzNum;
    int iNext, iPrev, jNext, jPrev, kNext, kPrev;
    realtype distAve = 0;       // average distance of nodes on front to root
    realtype distMax, distRatio;
    realtype ipSeedVecNode;     /* average of inner products among seed,
                                   connection vector, and front node */
    int counter = 0;            // count # of nodes on the front
    Kernel_Voxel_ node, 
        front = NULL;           // linked list of nodes on front
    Voxel_ seedfront = NULL;    // linked list of seed nodes on front
	
    kernel = (Diff_Kernel_)kernelData;
    dae = (DAE_)daeData;
    dti = (DTI_)dtiData;
    kernelVoxel = kernel->voxel;
    neq = kernel->neq;
    nx = kernel->nx;
    ny = kernel->ny;
    nz = kernel->nz;
    nxy = nx*ny;
    voxel = dti->voxel;
    seedID = dti->seedID;
    dtxyzNum = dti->dtxyzNum;

    yData = NV_DATA_S( dae->yVector );
                                // Allocate memory for neighbor nodes indeces.
    neighbors = (int *)malloc( NEI_NUM * sizeof( int ) );
                                // Find nodes with nonzero concentration.
    for( id = 0; id < neq; ++ id )
    {
        gid = kernelVoxel[id]->gridID;
	if( gid == OUT_OF_GRID ) continue;

	if( yData[id] >= kernel->cThreshold
            && kernelVoxel[id]->zeroTensor == FALSE
            && voxel[gid]->color == WHITE 
            && kernelVoxel[id]->fa >= TENSOR_FA
	    && kernelVoxel[id]->b0 >= DWI_B0_LO
	    && kernelVoxel[id]->b0 <= DWI_B0_HI
	    && inForwardDiffVol( kernel, kernelVoxel[id] ) == TRUE ) 
        {
//printf("*******************b0=%g\n",kernelVoxel[id]->b0);	    
//	    voxel[gid]->color = GREY;
	    i = kernelVoxel[id]->i;
	    j = kernelVoxel[id]->j;
	    k = kernelVoxel[id]->k;
	    iNext = i + 1;
	    jNext = j + 1;
	    kNext = k + 1;
	    iPrev = i - 1;
	    jPrev = j - 1;
	    kPrev = k - 1;
	                        // if nodes fall out of boundary
	    if( iNext == nx ) iNext = i;
    	    if( jNext == ny ) jNext = j;
	    if( kNext == nz ) kNext = k;
	    if( iPrev == -1 ) iPrev = i;
	    if( jPrev == -1 ) jPrev = j;
	    if( kPrev == -1 ) kPrev = k;
	                        // Get the indeces of neighbor nodes
	    neighbors[0] = nodeIndex( iNext+1, j+1, k+1, nx, nxy ) - 1; 
	    neighbors[1] = nodeIndex( iPrev+1, j+1, k+1, nx, nxy ) - 1;
	    neighbors[2] = nodeIndex( i+1, jNext+1, k+1, nx, nxy ) - 1;
	    neighbors[3] = nodeIndex( i+1, jPrev+1, k+1, nx, nxy ) - 1;
	    neighbors[4] = nodeIndex( i+1, j+1, kNext+1, nx, nxy ) - 1;
	    neighbors[5] = nodeIndex( i+1, j+1, kPrev+1, nx, nxy ) - 1;
	
	    neighbors[6] = nodeIndex( iPrev+1, jNext+1, kNext+1, nx, nxy ) - 1; 
	    neighbors[7] = nodeIndex( i+1, jNext+1, kNext+1, nx, nxy ) - 1;
	    neighbors[8] = nodeIndex( iNext+1, jNext+1, kNext+1, nx, nxy ) - 1;
	    neighbors[9] = nodeIndex( iPrev+1, jNext+1, k+1, nx, nxy ) - 1;
	    neighbors[10] = nodeIndex( iNext+1, jNext+1, k+1, nx, nxy ) - 1;
	    neighbors[11] = nodeIndex( iPrev+1, jNext+1, kPrev+1, nx, nxy ) - 1; 
	    neighbors[12] = nodeIndex( i+1, jNext+1, kPrev+1, nx, nxy ) - 1;
	    neighbors[13] = nodeIndex( iNext+1, jNext+1, kPrev+1, nx, nxy ) - 1;
	
	    neighbors[14] = nodeIndex( iPrev+1, j+1, kNext+1, nx, nxy ) - 1;
	    neighbors[15] = nodeIndex( iNext+1, j+1, kNext+1, nx, nxy ) - 1; 
	    neighbors[16] = nodeIndex( iPrev+1, j+1, kPrev+1, nx, nxy ) - 1;
	    neighbors[17] = nodeIndex( iNext+1, j+1, kPrev+1, nx, nxy ) - 1;
	
	    neighbors[18] = nodeIndex( iPrev+1, jPrev+1, kNext+1, nx, nxy ) - 1; 
	    neighbors[19] = nodeIndex( i+1, jPrev+1, kNext+1, nx, nxy ) - 1;
	    neighbors[20] = nodeIndex( iNext+1, jPrev+1, kNext+1, nx, nxy ) - 1;
	    neighbors[21] = nodeIndex( iPrev+1, jPrev+1, k+1, nx, nxy ) - 1;
	    neighbors[22] = nodeIndex( iNext+1, jPrev+1, k+1, nx, nxy ) - 1;
	    neighbors[23] = nodeIndex( iPrev+1, jPrev+1, kPrev+1, nx, nxy ) - 1; 
	    neighbors[24] = nodeIndex( i+1, jPrev+1, kPrev+1, nx, nxy ) - 1;
	    neighbors[25] = nodeIndex( iNext+1, jPrev+1, kPrev+1, nx, nxy ) - 1;
	
	                        // Find the nodes on diffusion front, i.e.,
				// at least one of its 26 neighbors has zero
				// concentration value.
	    for( m = 0; m < NEI_NUM; ++ m )
	    { 
	        neiID = neighbors[m];
	        neiGID = kernelVoxel[neiID]->gridID;
		if( neiGID == OUT_OF_GRID ) continue;
	        if( yData[neiID] < kernel->cThreshold  
                    && kernelVoxel[neiID]->zeroTensor == FALSE )
//	            && voxel[neiGID]->color == WHITE ) 
//	            && voxel[neiGID]->color != BLACK ) 
	        {
//	            voxel[gid]->color = DARK_GREY;
                    kernelVoxel[id]->dist = distance( kernel, i, j, k );
	            distAve += kernelVoxel[id]->dist;
	            counter ++;
                    if( front == NULL ) 
	            {
	                front = kernelVoxel[id];
		        front->tail = front;
	            }
                    else insert( front, kernelVoxel[id] );
	            break;
	        }
	    }
        }
    }
                                // get the average distance   
    if( counter != 0 ) distAve /= counter;
    else return seedfront;
                                // Sort the nodes on front in descending order
				// in terms of distance.
    front = sortDistance( front, counter, kernelVoxel );
        
    node = front;
    voxel[seedID]->child = voxel[node->gridID];
    counter = 0;                // reset the counter to zero
    distMax = node->dist;
    while( node != NULL 
//        && counter < FRONT_MAX_NODES
        && counter < frontMaxNodes
        && ( distRatio = node->dist/distMax ) >= DIST_RATIO )
    {
        gid = node->gridID;

	                // if node falls in ERR-ZONE, skip to the next one
#ifdef SET_ERR_ZONE
        if( voxel[gid]->i > ZONE_X_LO && voxel[gid]->i < ZONE_X_HI && 
	    voxel[gid]->j > ZONE_Y_LO && voxel[gid]->j < ZONE_Y_HI &&
	    voxel[gid]->k > ZONE_Z_LO && voxel[gid]->k < ZONE_Z_HI )
	{
	    node = node->next;
	    continue;
	}    
#endif

#ifdef SET_ERR_ZONE_1
        if( voxel[gid]->i > ZONE_1_X_LO && voxel[gid]->i < ZONE_1_X_HI && 
	    voxel[gid]->j > ZONE_1_Y_LO && voxel[gid]->j < ZONE_1_Y_HI &&
	    voxel[gid]->k > ZONE_1_Z_LO && voxel[gid]->k < ZONE_1_Z_HI )
	{
	    node = node->next;
	    continue;
	}    
#endif

        if( isCollinear( kernelVoxel[kernel->seedID], node, dti, kernel,
	    &ipSeedVecNode ) == FALSE )
	{
	    node = node->next;
	    continue;
	}

//        gid = node->gridID;
	voxel[gid]->color = BLACK;
	voxel[gid]->level = voxel[seedID]->level + 1;
	voxel[gid]->parent = voxel[seedID];
        voxel[gid]->tail = NULL;
	                        // construct the seed front list
	if( seedfront == NULL )
	{
	    seedfront = voxel[gid];
	    seedfront->tail = voxel[gid];
	}
	else insertSeed( seedfront, voxel[gid] );
	
	node = node->next;
	counter ++;
    }
    if( counter == 0 ) voxel[seedID]->child = NULL;
   
    free( neighbors );
    return seedfront;
}


/***********************************************************************
 * The general equation of a plane with nonzero unit normal vector 
 * (a, b, c) through the seed point (x0, y0, z0) is: ax+by+cz+d=0,
 * where d=-ax0-by0-cz0.
 ***********************************************************************/
void cuttingPlaneCoeff( void *dtiData, void *kernelData )
{
    Diff_Kernel_ kernel;
    DTI_ dti;
    Voxel_ *voxel, parent, seed;
    realtype a, b, c, d;
    realtype dx, dy, dz, dist;
    
    kernel = (Diff_Kernel_)kernelData;
    dti = (DTI_)dtiData;
    voxel = dti->voxel;
    dx = dti->dx;
    dy = dti->dy;
    dz = dti->dz;
    
    seed = voxel[dti->seedID];
    parent = seed->parent;
                                // normalize the vector (a,b,c) which connects
				// the current seed and its parent node.
    if( parent != NULL )
    {
        dist = (seed->i - parent->i)*(seed->i - parent->i)*dx*dx +
            (seed->j - parent->j)*(seed->j - parent->j)*dy*dy + 
            (seed->k - parent->k)*(seed->k - parent->k)*dz*dz;
        dist = (realtype)sqrt( dist );	 
        a = (seed->i - parent->i) * dx / dist;
        b = (seed->j - parent->j) * dy / dist;
        c = (seed->k - parent->k) * dz / dist;
        d = -a*dx*kernel->seedI - b*dy*kernel->seedJ - c*dz*kernel->seedK;
    }	
    else
        a = b = c = d = 0;

    kernel->pa = a;
    kernel->pb = b;
    kernel->pc = c;
    kernel->pd = d;
}


/***********************************************************************
 * Determine if the voxel is on the correct side of the cutting plane.
 ***********************************************************************/
int inForwardDiffVol( void *kernelData, void *kernelVoxel )
{
    Diff_Kernel_ kernel;
    Kernel_Voxel_ voxel;
    realtype a, b, c, d, dx, dy, dz;

    voxel = (Kernel_Voxel_)kernelVoxel;
    kernel = (Diff_Kernel_)kernelData;
    a = kernel->pa;
    b = kernel->pb;
    c = kernel->pc;
    d = kernel->pd;
    dx = kernel->dx;
    dy = kernel->dy;
    dz = kernel->dz;
                        /* the seed is diffusion root */
    if( a == 0 && b == 0 && c == 0 && d == 0 ) return TRUE;
    
    if( a*dx*voxel->i + b*dy*voxel->j + c*dz*voxel->k + d >= 0 )
        return TRUE;
    else
        return FALSE;
}

/***********************************************************************
 * Check the inner products of several pairs of directions 
 * to prevent sharp transition in the fiber path.
 ***********************************************************************/
int isCollinear( void *voxelSeed, void *voxelNode, void *dtiData, 
    void *kernelData, realtype *ipSeedVecNode )
{
    Kernel_Voxel_ seed, node;
    Diff_Kernel_ kernel;
    DTI_ dti;
    realtype x, y, z, dx, dy, dz;
    realtype ip_vectors = 1, 
        ip_seed_vec, ip_seed_node, ip_node_vec;
    
    seed = (Kernel_Voxel_)voxelSeed;
    node = (Kernel_Voxel_)voxelNode;
    kernel = (Diff_Kernel_)kernelData;
    dti = (DTI_)dtiData;
    dx = kernel->dx;
    dy = kernel->dy;
    dz = kernel->dz;
                                // normalize the vector which connects
				// the seed and the front node.
    x = (node->i - seed->i) * dx / node->dist;
    y = (node->j - seed->j) * dy / node->dist;
    z = (node->k - seed->k) * dz / node->dist;
                                // inner product between two connection vectors
    if( dti->voxel[seed->gridID]->parent != NULL )
        ip_vectors = x*kernel->pa + y*kernel->pb + z*kernel->pc;
                                // inner product between seed eigvec and 
				// connection vector
    ip_seed_vec = seed->eigvec[0]*x + seed->eigvec[1]*y + seed->eigvec[2]*z;
                                // inner product between seed eigvec and 
				// front node eigvec
    ip_seed_node = seed->eigvec[0]*(node->eigvec[0]) 
        + seed->eigvec[1]*(node->eigvec[1]) + seed->eigvec[2]*(node->eigvec[2]);
                                // inner product between node eigvec and 
				// connection vector
    ip_node_vec = node->eigvec[0]*x + node->eigvec[1]*y + node->eigvec[2]*z;
/*    
printf("ip: vectors=%9.4f seed_vec=%9.4f seed_node=%9.4f node_vec=%9.4f\n",
ip_vectors, ip_seed_vec, ip_seed_node, ip_node_vec );
*/

    *ipSeedVecNode = ( ABS(ip_seed_vec) + ABS(ip_seed_node) 
        + ABS(ip_node_vec) ) / 3;

    if( ip_vectors >= IP_VECTORS &&
        ABS(ip_seed_vec) >= IP_SEED_VEC &&
        ABS(ip_node_vec) >= IP_NODE_VEC &&
	ABS(ip_seed_node) >= IP_SEED_NODE )
	return TRUE;
    else
        return FALSE;
}


/***********************************************************************
 * Get the distance from the current node to the seed node.
 ***********************************************************************/
realtype distance( void *kernelData, int i, int j, int k )
{
    Diff_Kernel_ kernel;
    int seedI, seedJ, seedK;
    realtype dx, dy, dz, dist;
    
    kernel = (Diff_Kernel_)kernelData;
    seedI = kernel->seedI;
    seedJ = kernel->seedJ;
    seedK = kernel->seedK;
    dx = kernel->dx;
    dy = kernel->dy;
    dz = kernel->dz;

    dist = (seedI-i)*(seedI-i)*dx*dx
        + (seedJ-j)*(seedJ-j)*dy*dy
        + (seedK-k)*(seedK-k)*dz*dz;
    dist = (realtype)sqrt( dist ); 
    return dist;
}

/***********************************************************************
 * Sort the distance in descending order.
 ***********************************************************************/
Kernel_Voxel_ sortDistance( void *voxelList, int listSize, void *voxelData )
{
    Kernel_Voxel_ list, *voxel, node;
    int nodeID[listSize], counter = 0;
    realtype dist[listSize];
    
    list = (Kernel_Voxel_)voxelList;
    voxel = (Kernel_Voxel_ *)voxelData;
    
    node = list;
    while( node != NULL && counter < listSize )
    {
        nodeID[counter] = node->id;
	dist[counter] = node->dist;
	counter ++;
        node = node->next;
    }
                                // sort the distance in ascending order
    insertionSort( dist, listSize, nodeID );
                                // reconstruct the list with sorted distance
				// in descending order
    list = voxel[nodeID[listSize-1]];
    list->tail = list;
    list->next = NULL;
    for( counter = listSize-2; counter >= 0; --counter )
    {
        node = voxel[nodeID[counter]];
	node->next = NULL;
	node->tail = NULL;
        insert( list, node );
    }
    return list;
}



/***********************************************************************
 * Insertion sort in ascending order.
 ***********************************************************************/
void insertionSort( realtype *numbers, int array_size, int *numIDs )
{
    int i, j, tmpID; 
    realtype index;

    for( i = 1; i < array_size; i++ )
    {
        index = numbers[i];
	tmpID = numIDs[i];
        j = i;
        while( j > 0 && numbers[j-1] > index )
        {
            numbers[j] = numbers[j-1];
	    numIDs[j] = numIDs[j-1];
            j --;
        }
        numbers[j] = index;
	numIDs[j] = tmpID;
    }
}
      
      
/***********************************************************************
 * Insert a node or a list at the end of a list.
 ***********************************************************************/
void insert( void *voxelList, void *voxelNode )
{
    Kernel_Voxel_ list, node;
    
    list = (Kernel_Voxel_)voxelList;
    node = (Kernel_Voxel_)voxelNode;
    
    list->tail->next = node;
                                // insert a node
    if( node->next == NULL )
	list->tail = node;
	                        // insert a list
    else
    {
        list->tail = node->tail;        
	node->tail = NULL;
    }
}
      
/***********************************************************************
 * Insert a seed node or list at the end of a seed list.
 ***********************************************************************/
void insertSeed( void *voxelList, void *voxelNode )
{
    Voxel_ list, node;
    
    list = (Voxel_)voxelList;
    node = (Voxel_)voxelNode;
    
    list->tail->next = node;
                                // insert a node
    if( node->next == NULL )
	list->tail = node;
	                        // insert a list
    else
    {
        list->tail = node->tail;        
	node->tail = NULL;
    }
}
      
/***********************************************************************
 * Set the root node.
 * Check if the root node is in white matter.
 ***********************************************************************/
int setRootNode( void *dtiData, int *roiPoints[], int index )
{
    DTI_ dti = (DTI_)dtiData;
    float fa, b0;
    
    dti->seedI = roiPoints[0][index];
    dti->seedJ = roiPoints[1][index];
    dti->seedK = roiPoints[2][index];

    dti->seedID = nodeIndex( dti->seedI+1, dti->seedJ+1, dti->seedK+1,
        dti->nx, dti->nx*(dti->ny) ) - 1;
    fa = dti->fa[dti->seedID];
    b0 = dti->b0[dti->seedID];
    if( fa < TENSOR_FA || b0 < DWI_B0_LO || b0 > DWI_B0_HI ) return -1;

    dti->voxel[dti->seedID]->color = BLACK;
    dti->voxel[dti->seedID]->level = LEVEL_ZERO;
    return 0;
}


/***********************************************************************
 * Set the seed node which will be diffused.
 ***********************************************************************/
void setSeedNode( void *dtiData, void *voxelNode )
{
    DTI_ dti;
    Voxel_ node;
    int i, j, k;
    
    dti = (DTI_)dtiData;
    node = (Voxel_)voxelNode;
    i = node->i;
    j = node->j;
    k = node->k;
    
    dti->seedID = nodeIndex( i+1, j+1, k+1, dti->nx, dti->nx*(dti->ny) ) - 1;
    dti->seedI = i;
    dti->seedJ = j;
    dti->seedK = k;
}

