/***********************************************************************
 * File:        fibertract.c
 * Programmer:  Ning Kang @CS.UKY
 * Version of:  7/29/2004
 * ---------------------------------------------------------------------
 * Description:
 * - Serial version.
 * - The functions in this file are used to do fiber tracing.
 * - This file contains the following functions:
 *   int fiberTracking ()
 *   int setTractFile ()
 *   int buildGraph ()
 *   int traverseGraph ()
 *   Voxel_ getDiffFront ()
 *   realtype distance ()
 *   Kernel_Voxel_ sortDistance ()
 *   void insertionSort ()
 *   void insert ()
 *   void insertSeed ()
 *   void setRootNode ()
 *   void setSeedNode ()
 *   void cuttingPlaneCoeff ()
 *   int inForwardDiffVol () 
 ***********************************************************************/
 

#include "dst.h"

static int buildGraph_DT4( void *kernelData, void *iluData, 
  void *daeData, void *dtiData );
static int traverseGraph_DT4( void *dtiData, char *filename, 
  int *numNodes, int *numPaths );
static Voxel_ getDiffFront_DT4( void *kernelData, void *dtiData, 
  void *daeData, int frontMaxNodes );
static int isCollinear_DT4( void *voxelSeed, void *voxelNode, 
  void *dtiData, void *kernelData );
static int setRootNode_DT4( void *dtiData, int *roiPoints[], 
  int index );


/***********************************************************************
 * Do fiber tracking.
 ***********************************************************************/
int fiberTracking_DT4( void *kernelData, void *iluData, 
    void *daeData, void *dtiData, void *argsTable )
{
    Args_Table_ args;
    Diff_Kernel_ kernel;           
    ILU_ ilu;
    DAE_ dae;
    DTI_ dti;
    int i, seqno = 0, buildRet, numNodes, numPaths;
    char tractfile[STR_LEN], *tractbase;

    kernel = (Diff_Kernel_)kernelData;
    ilu = (ILU_)iluData;
    dae = (DAE_)daeData;
    dti = (DTI_)dtiData;
    args = (Args_Table_)argsTable;
    
    tractbase = args->tractbase;
    args->totalNumNodes = 0;
    args->totalNumPaths = 0;

    for( i = 0; i < args->roiSize; ++ i )
    {
//if(i==1)break;    
        if( setRootNode_DT4( dti, args->roiPoints, i ) != 0 ) 
	    continue;
#ifdef DEBUG_PRINT
printf( "--->root seed %4d: %3d %3d %3d\n", i,dti->seedI, dti->seedJ, 
 dti->seedK );
#endif
                        // build the graph structure by diffusion simulation 
	buildRet = buildGraph_DT4( kernel, ilu, dae, dti );
        if( buildRet == NO_FRONT_AT_ROOT ) 
	    continue;
	else if( buildRet != 0 )
	    return 1;
			// traverse the graph and output fiber paths to file 
        if( setTractFile( tractfile, STR_LEN, DST_DATA_DIR, tractbase, 
	    seqno ) != 0 ) return 1;
	if( traverseGraph_DT4( dti, tractfile, &numNodes, 
	    &numPaths ) != 0 ) return 1;
                   // make sure the size of the output file is not zero
        if( numNodes > 0 )
	{
            args->totalNumNodes += numNodes;	    
	    args->totalNumPaths += numPaths;
	    seqno ++;
	}    
                        /* reset the voxel data structure */
        resetDTIvoxel( dti );
    }
                       // number of generated files for computed tracks
    args->numTractFiles = seqno;

    return 0;
}    

/***********************************************************************
 * Build the graph structure for fiber tracing.
 ***********************************************************************/
int buildGraph_DT4( void *kernelData, void *iluData, void *daeData, 
  void *dtiData )
{
    Diff_Kernel_ kernel;           
    ILU_ ilu;
    DAE_ dae;
    DTI_ dti;
    Voxel_ node, front = NULL, list = NULL;
    
    kernel = (Diff_Kernel_)kernelData;
    ilu = (ILU_)iluData;
    dae = (DAE_)daeData;
    dti = (DTI_)dtiData;
               // root seed
    if( getCoeffMatrixEntry_DT4( kernel, dti ) != 0 ) return 1;
    if( rootNodeDiffSim_DT4( kernel, dae, ilu ) != 0 ) return 1;
    diffThreshold( kernel, dae );
    cuttingPlaneCoeff( dti, kernel );    
    front = getDiffFront_DT4( kernel, dti, dae, FRONT_MAX_NODES_ROOT );
    
    if( front != NULL ) 
        node = list = front;
    else 
    {
        fprintf( stderr, "No diffusion front at root\n" );
	return NO_FRONT_AT_ROOT;
    }
               // the generated seeds
    while( node != NULL )
    {
        if( node->color == BLACK )
        {
#ifdef DEBUG_PRINT_SEED_NODE
printf( "seed node: %3d %3d %3d\n", node->i, node->j, node->k );
#endif
          setSeedNode( dti, node );
          if( getCoeffMatrixEntry_DT4( kernel, dti ) != 0 ) return 1;
	  if( leafNodeDiffSim_DT4( kernel, dae, ilu ) != 0 ) 
	      return 1;
          diffThreshold( kernel, dae );
          cuttingPlaneCoeff( dti, kernel );    
          front = getDiffFront_DT4( kernel, dti, dae, FRONT_MAX_NODES );
	  if( front != NULL ) insertSeed( list, front );
	}
        node = node->next;
    }
    
    return 0; 
}

/***********************************************************************
 * Traverse the graph to print out fiber pathways to file.
 ***********************************************************************/
int traverseGraph_DT4( void *dtiData, char *filename, int *numNodes,
    int *numPaths )
{
    FILE *fp;		        // File pointer for fiber trace output.
    DTI_ dti;
    Voxel_ *voxel, node; 
    int id, num, counter;
    
    dti = (DTI_)dtiData;
    num = dti->num;
    voxel = dti->voxel;

    if( ( fp = fopen( filename, "w" ) ) == NULL )
    {
        fprintf( stderr, "Cannot create file %s\n", filename );
	return 1;
    }
    
    *numNodes = 0;
    counter = 0;
#ifdef SET_END_ROI
    for( id = 0; id < num; ++ id )
    if( voxel[id]->child == NULL && voxel[id]->parent != NULL )
    if( voxel[id]->level >= MIN_LEVEL )
    {
	node = voxel[id]; 
	                // if node is not in END-ROI, skip to the next one
	if( node->i < ROI_X_LO || node->i > ROI_X_HI ||
	    node->j < ROI_Y_LO || node->j > ROI_Y_HI ||
	    node->k < ROI_Z_LO || node->k > ROI_Z_HI )
	    continue;
	
	counter ++;    
        while( node != NULL ) 
	{
	    fprintf( fp, "%6d%4d%4d%4d%10.4f%10.4f\n", counter, node->i, 
	        node->j, node->k, dti->dt4_gfa[node->id], 
		dti->dt4_gfa[node->id] ); 
            (*numNodes) ++;
	    node = node->parent;
	}
    }
#else
    for( id = 0; id < num; ++ id )
    if( voxel[id]->child == NULL && voxel[id]->parent != NULL )
    if( voxel[id]->level >= MIN_LEVEL )
    {
	node = voxel[id]; 
	counter ++;    
        while( node != NULL ) 
	{
	    fprintf( fp, "%6d%4d%4d%4d%10.4f%10.4f\n", counter, node->i, 
	        node->j, node->k, dti->dt4_gfa[node->id], 
		dti->dt4_gfa[node->id] ); 
            (*numNodes) ++;
	    node = node->parent;
	}
    }
#endif    
    *numPaths = counter;
    fclose( fp );
    return 0;
}


/***********************************************************************
 * Get the diffusion front.
 ***********************************************************************/
Voxel_ getDiffFront_DT4( void *kernelData, void *dtiData, 
  void *daeData, int frontMaxNodes )
{
    Diff_Kernel_ kernel;
    Kernel_Voxel_ *kernelVoxel;
    DAE_ dae;
    DTI_ dti;
    Voxel_ *voxel;
    realtype *yData;
    int m, id, gid, i, j, k, 
        *neighbors, neiID, neiGID;      // indeces of neighbor nodes 
    int seedID, neq, nx, ny, nz, nxy;
    int iNext, iPrev, jNext, jPrev, kNext, kPrev;
    realtype distAve = 0;       // average distance of nodes on front to root
    realtype distMax, distRatio;
    int counter = 0;            // count # of nodes on the front
    Kernel_Voxel_ node, 
        front = NULL;           // linked list of nodes on front
    Voxel_ seedfront = NULL;    // linked list of seed nodes on front
	
    kernel = (Diff_Kernel_)kernelData;
    dae = (DAE_)daeData;
    dti = (DTI_)dtiData;
    kernelVoxel = kernel->voxel;
    neq = kernel->neq;
    nx = kernel->nx;
    ny = kernel->ny;
    nz = kernel->nz;
    nxy = nx*ny;
    voxel = dti->voxel;
    seedID = dti->seedID;

    yData = NV_DATA_S( dae->yVector );
                                // Allocate memory for neighbor nodes indeces.
    neighbors = (int *)malloc( NEI_NUM * sizeof( int ) );
                                // Find nodes with nonzero concentration.
    for( id = 0; id < neq; ++ id )
    {
        gid = kernelVoxel[id]->gridID;
	if( gid == OUT_OF_GRID ) continue;

	if( yData[id] >= kernel->cThreshold
            && kernelVoxel[id]->zeroTensor == FALSE
            && voxel[gid]->color == WHITE 
            && kernelVoxel[id]->fa >= dti->dt4_GFA
	    && inForwardDiffVol( kernel, kernelVoxel[id] ) == TRUE ) 
        {
//	    voxel[gid]->color = GREY;
	    i = kernelVoxel[id]->i;
	    j = kernelVoxel[id]->j;
	    k = kernelVoxel[id]->k;
	    iNext = i + 1;
	    jNext = j + 1;
	    kNext = k + 1;
	    iPrev = i - 1;
	    jPrev = j - 1;
	    kPrev = k - 1;
	                        // if nodes fall out of boundary
	    if( iNext == nx ) iNext = i;
    	    if( jNext == ny ) jNext = j;
	    if( kNext == nz ) kNext = k;
	    if( iPrev == -1 ) iPrev = i;
	    if( jPrev == -1 ) jPrev = j;
	    if( kPrev == -1 ) kPrev = k;
	                        // Get the indeces of neighbor nodes
	    neighbors[0] = nodeIndex( iNext+1, j+1, k+1, nx, nxy ) - 1; 
	    neighbors[1] = nodeIndex( iPrev+1, j+1, k+1, nx, nxy ) - 1;
	    neighbors[2] = nodeIndex( i+1, jNext+1, k+1, nx, nxy ) - 1;
	    neighbors[3] = nodeIndex( i+1, jPrev+1, k+1, nx, nxy ) - 1;
	    neighbors[4] = nodeIndex( i+1, j+1, kNext+1, nx, nxy ) - 1;
	    neighbors[5] = nodeIndex( i+1, j+1, kPrev+1, nx, nxy ) - 1;
	
	    neighbors[6] = nodeIndex( iPrev+1, jNext+1, kNext+1, nx, nxy ) - 1; 
	    neighbors[7] = nodeIndex( i+1, jNext+1, kNext+1, nx, nxy ) - 1;
	    neighbors[8] = nodeIndex( iNext+1, jNext+1, kNext+1, nx, nxy ) - 1;
	    neighbors[9] = nodeIndex( iPrev+1, jNext+1, k+1, nx, nxy ) - 1;
	    neighbors[10] = nodeIndex( iNext+1, jNext+1, k+1, nx, nxy ) - 1;
	    neighbors[11] = nodeIndex( iPrev+1, jNext+1, kPrev+1, nx, nxy ) - 1; 
	    neighbors[12] = nodeIndex( i+1, jNext+1, kPrev+1, nx, nxy ) - 1;
	    neighbors[13] = nodeIndex( iNext+1, jNext+1, kPrev+1, nx, nxy ) - 1;
	
	    neighbors[14] = nodeIndex( iPrev+1, j+1, kNext+1, nx, nxy ) - 1;
	    neighbors[15] = nodeIndex( iNext+1, j+1, kNext+1, nx, nxy ) - 1; 
	    neighbors[16] = nodeIndex( iPrev+1, j+1, kPrev+1, nx, nxy ) - 1;
	    neighbors[17] = nodeIndex( iNext+1, j+1, kPrev+1, nx, nxy ) - 1;
	
	    neighbors[18] = nodeIndex( iPrev+1, jPrev+1, kNext+1, nx, nxy ) - 1; 
	    neighbors[19] = nodeIndex( i+1, jPrev+1, kNext+1, nx, nxy ) - 1;
	    neighbors[20] = nodeIndex( iNext+1, jPrev+1, kNext+1, nx, nxy ) - 1;
	    neighbors[21] = nodeIndex( iPrev+1, jPrev+1, k+1, nx, nxy ) - 1;
	    neighbors[22] = nodeIndex( iNext+1, jPrev+1, k+1, nx, nxy ) - 1;
	    neighbors[23] = nodeIndex( iPrev+1, jPrev+1, kPrev+1, nx, nxy ) - 1; 
	    neighbors[24] = nodeIndex( i+1, jPrev+1, kPrev+1, nx, nxy ) - 1;
	    neighbors[25] = nodeIndex( iNext+1, jPrev+1, kPrev+1, nx, nxy ) - 1;
	
	                        // Find the nodes on diffusion front, i.e.,
				// at least one of its 26 neighbors has zero
				// concentration value.
	    for( m = 0; m < NEI_NUM; ++ m )
	    { 
	        neiID = neighbors[m];
	        neiGID = kernelVoxel[neiID]->gridID;
		if( neiGID == OUT_OF_GRID ) continue;
	        if( yData[neiID] < kernel->cThreshold  
                    && kernelVoxel[neiID]->zeroTensor == FALSE )
//	            && voxel[neiGID]->color == WHITE ) 
//	            && voxel[neiGID]->color != BLACK ) 
	        {
//	            voxel[gid]->color = DARK_GREY;
                    kernelVoxel[id]->dist = distance( kernel, i, j, k );
	            distAve += kernelVoxel[id]->dist;
	            counter ++;
                    if( front == NULL ) 
	            {
	                front = kernelVoxel[id];
		        front->tail = front;
	            }
                    else insert( front, kernelVoxel[id] );
	            break;
	        }
	    }
        }
    }
//printf("****cThreshold=%.4e counter=%d\n",kernel->cThreshold,counter);    
                                // get the average distance   
    if( counter != 0 ) distAve /= counter;
    else return seedfront;
                                // Sort the nodes on front in descending order
				// in terms of distance.
    front = sortDistance( front, counter, kernelVoxel );
        
    node = front;
    voxel[seedID]->child = voxel[node->gridID];
    counter = 0;                // reset the counter to zero
    distMax = node->dist;
    while( node != NULL 
        && counter < frontMaxNodes
        && ( distRatio = node->dist/distMax ) >= dti->dt4_distRatio )
    {
        gid = node->gridID;
	                // if node falls in ERR-ZONE, skip to the next one
#ifdef SET_ERR_ZONE
        if( voxel[gid]->i > ZONE_X_LO && voxel[gid]->i < ZONE_X_HI && 
	    voxel[gid]->j > ZONE_Y_LO && voxel[gid]->j < ZONE_Y_HI &&
	    voxel[gid]->k > ZONE_Z_LO && voxel[gid]->k < ZONE_Z_HI )
	{
	    node = node->next;
	    continue;
	}    
#endif

#ifdef SET_ERR_ZONE_1
        if( voxel[gid]->i > ZONE_1_X_LO && voxel[gid]->i < ZONE_1_X_HI && 
	    voxel[gid]->j > ZONE_1_Y_LO && voxel[gid]->j < ZONE_1_Y_HI &&
	    voxel[gid]->k > ZONE_1_Z_LO && voxel[gid]->k < ZONE_1_Z_HI )
	{
	    node = node->next;
	    continue;
#endif

        if( isCollinear_DT4( kernelVoxel[kernel->seedID], node, 
	    dti, kernel) == FALSE )
	{
	    node = node->next;
	    continue;
	}

//        gid = node->gridID;
	voxel[gid]->color = BLACK;
	voxel[gid]->level = voxel[seedID]->level + 1;
	voxel[gid]->parent = voxel[seedID];
        voxel[gid]->tail = NULL;
	                        // construct the seed front list
	if( seedfront == NULL )
	{
	    seedfront = voxel[gid];
	    seedfront->tail = voxel[gid];
	}
	else insertSeed( seedfront, voxel[gid] );
	
	node = node->next;
	counter ++;
    }
    if( counter == 0 ) voxel[seedID]->child = NULL;
   
    free( neighbors );
    return seedfront;
}

/***********************************************************************
 * Check the inner products of several pairs of directions 
 * to prevent sharp transition in the fiber path.
 ***********************************************************************/
int isCollinear_DT4( void *voxelSeed, void *voxelNode, 
    void *dtiData, void *kernelData )
{
    Kernel_Voxel_ seed, node;
    Diff_Kernel_ kernel;
    DTI_ dti;
    realtype x, y, z, dx, dy, dz;
    realtype ip_vectors = 1; 
    
    seed = (Kernel_Voxel_)voxelSeed;
    node = (Kernel_Voxel_)voxelNode;
    kernel = (Diff_Kernel_)kernelData;
    dti = (DTI_)dtiData;
    dx = kernel->dx;
    dy = kernel->dy;
    dz = kernel->dz;
                                // normalize the vector which connects
				// the seed and the front node.
    x = (node->i - seed->i) * dx / node->dist;
    y = (node->j - seed->j) * dy / node->dist;
    z = (node->k - seed->k) * dz / node->dist;
                                // inner product between two connection vectors
    if( dti->voxel[seed->gridID]->parent != NULL )
        ip_vectors = x*kernel->pa + y*kernel->pb + z*kernel->pc;
    if( ip_vectors >= dti->dt4_ipVectors )
	return TRUE;
    else
        return FALSE;
}



/***********************************************************************
 * Set the root node.
 * Check if the root node is in white matter.
 ***********************************************************************/
int setRootNode_DT4( void *dtiData, int *roiPoints[], int index )
{
    DTI_ dti = (DTI_)dtiData;
    float gfa;
    
    dti->seedI = roiPoints[0][index];
    dti->seedJ = roiPoints[1][index];
    dti->seedK = roiPoints[2][index];

    dti->seedID = nodeIndex( dti->seedI+1, dti->seedJ+1, dti->seedK+1,
        dti->nx, dti->nx*(dti->ny) ) - 1;
    gfa = dti->dt4_gfa[dti->seedID];
    if( gfa < dti->dt4_GFA ) return -1;

    dti->voxel[dti->seedID]->color = BLACK;
    dti->voxel[dti->seedID]->level = LEVEL_ZERO;
    return 0;
}




