/***********************************************************************
 * File: 	ftvtk.c	                                    
 * Programmer:	Ning Kang		
 * Date:	11/7/2004			
 * --------------------------------------------------------------------
 * Description: 							
 * - The functions in this file are used to convert fiber tract data 
 * into the VTK data format (polydata) and write them to a file.
 * - There are five columns in the input data:
 *   col-1: fiber path no.
 *   col-2: x-coordinate of points -> int
 *   col-3: y-coordinate of points -> int
 *   col-4: z-coordinate of points -> int
 *   col-5: attribute value
 * - The output data are approximated by least square with B-splines.
 ***********************************************************************/

#include "dst.h" 
#include "dst_qsort.h"
               // used in QSORT
#define int_lt(a,b) ((*a)<(*b))

static int fiberStats( void *ftData, void *argsTable );
static double calcMeanVariance( double *variance, float *array, int size );
static int saveFTDataVTK( void *ftData, char *filename, float dx, float dy, 
    float dz );
static void splitFTData( void *ftData );
static FT_ ftData_new( void *argsTable, char *dir );
static void ftData_free( void *ftData );
static void approximation( void *ftData, int spline_order, int num_pp );
static void l2apprSpline( float *outPt, int outSize, int *inPt, int inSize, 
    int spline_order, int num_pp );
static void l2apprSpline_attr( float *outPt, int outSize, float *inPt, 
    int inSize, int spline_order, int num_pp );
static int isPrime( int n );
static int hashFunc( int hashTableSize, int x, int y, int z );
static TargetPt_ *hashTable_new( int hashTableSize, int ptArraySize, 
    int *pt[3] );
static void hashInsert( void *hashTable, int hashTableSize, int x, int y, 
    int z );
static int hashLookup( void *hashTable, int hashTableSize, int x, int y, 
    int z );
static void trimFTData( void *ftData, void *argsTable );
static int saveTrimmedFTDataVTK( void *ftData, char *filename, float dx, 
    float dy, float dz );
static int trimmedFiberStats( void *ftData, void *argsTable );

/***********************************************************************
 * Convert all files of computed tracts to a single VTK data file.
 ***********************************************************************/
int fiberVTK( void *argsTable, char *dir, int spline_order, int num_pp )
{
    Args_Table_ args = (Args_Table_)argsTable;
    FT_ ft;
   
printf("total nodes = %d  total paths = %d total tract files = %d\n",
    args->totalNumNodes,args->totalNumPaths,args->numTractFiles);    

    if( args->totalNumNodes == 0 || args->totalNumPaths == 0 )
    { 
        fprintf( stderr, "Oops, no fiber tracts generated\n" );
	return -1;
    }
               // allocate memory and read input data
    if( (ft = ftData_new( args, dir ))  == NULL ) return -1;
printf("ftData_new: done\n");				
               // approximate the data points
    approximation( ft, spline_order, num_pp );    
printf("approximation: done\n");
               // group points of paths in VTK format
    splitFTData( ft );
printf("splitFTdata: done\n");
               // save original data to a file in VTK format 
    if( args->roiSizeTarget == 0 )	
    {
        if( saveFTDataVTK( ft, args->tractFileName, 
	    args->dxOrig, args->dyOrig, args->dzOrig ) != 0 ) return -1;
        if( fiberStats( ft, args ) != 0 ) return -1;
    }	    
	        // save trimmed data to a file in VTK format
    if( args->roiSizeTarget > 0 )
    { 
        trimFTData( ft, args );
        if( saveTrimmedFTDataVTK( ft, args->tractFileName, 
	    args->dxOrig, args->dyOrig, args->dzOrig ) != 0 ) return -1;
        if( trimmedFiberStats( ft, args ) != 0 ) return -1;
printf("saveTrimmedFTDataVTK and trimmedFiberStats: done\n");    
    }
               // free memory
    ftData_free( ft );
printf("ftData_free: done\n");

    return 0;
}

/***********************************************************************
 * Calculate fiber tract statistics.
 ***********************************************************************/
int fiberStats( void *ftData, void *argsTable )
{
    Args_Table_ args = (Args_Table_)argsTable;
    FT_ ft = (FT_)ftData;
    FILE *fp;
    int i, j, x, y, z, ptIdx;
    double fiberFAMean, fiberFAVariance, volumeFAMean, volumeFAVariance,
        fiberMDMean, fiberMDVariance, volumeMDMean, volumeMDVariance,
        scatterXMean, scatterYMean, scatterZMean,
	scatterXVariance, scatterYVariance, scatterZVariance,
        fiberLengthMean, fiberLengthVariance, fiberLengthSum;
    int fiberVolume, brainVolume = 0;
    int *voxelIdx; // array for holding voxel indices
    float *voxelFA, *voxelMD; // array for holding FA, MD values in voxels
    float *endPtX, *endPtY, *endPtZ; // coordinates of fiber ending points
    float *pathLength; // the length of each fiber tract
    float x1, y1, z1, x0, y0, z0;
    float dx, dy, dz;
    int nxy = args->nx*(args->ny);

    dx = args->dx; dy = args->dy; dz = args->dz;

               // calculate the brain volume
//    brainVolume = computeBrainVolume( args->dwi_b0, args->nx, args->ny,
//        args->nz );
    if( args->dwi_b0 != NULL )
    for( i = 0; i < args->nvox; ++ i )
    if( args->dwi_b0[i] >= B0_BRAIN_VOLUME_THRESHOLD ) brainVolume ++;
               // Stats-1: 
	       // mean and variance for FA on individual fibers
    fiberFAMean = calcMeanVariance( &fiberFAVariance, ft->ptFA, ft->size );
    fiberMDMean = calcMeanVariance( &fiberMDVariance, ft->ptMD, ft->size );

               // Stats-2:
               // fiber volume: number of unique voxels fibers go through
    voxelIdx = (int *)malloc( ft->size * sizeof( int ) );
    voxelFA = (float *)malloc( ft->size * sizeof( float ) );
    voxelMD = (float *)malloc( ft->size * sizeof( float ) );
    for( i = 0; i < ft->size; ++ i )
    {          // coordinate
        x = (int)( ft->ptx[i] );
	y = (int)( ft->pty[i] );
	z = (int)( ft->ptz[i] );
        voxelIdx[i] = nodeIndex( x+1, y+1, z+1, args->nx, nxy ) - 1; 
    }
               // sort the array of voxel indices
    QSORT( int, voxelIdx, ft->size, int_lt ); 
               // count the number of unique elements in voxelArray
    fiberVolume = 1;
    voxelFA[0] = args->fa[voxelIdx[0]];
    voxelMD[0] = args->md[voxelIdx[0]];
    for( i = 0; i < ft->size-1; ++ i )
    if( voxelIdx[i] != voxelIdx[i+1] )
    {
        voxelFA[fiberVolume] = args->fa[voxelIdx[i+1]];
        voxelMD[fiberVolume] = args->md[voxelIdx[i+1]];
        fiberVolume ++;
    }    
    volumeFAMean = calcMeanVariance( &volumeFAVariance, voxelFA, 
        fiberVolume );
    volumeMDMean = calcMeanVariance( &volumeMDVariance, voxelMD, 
        fiberVolume );
    free( voxelIdx );
    free( voxelFA );
    free( voxelMD );
               
	       // Stats-3:
	       // calculate the diversification of fiber scattering, i.e.,
	       // calculate the mean and variance of fiber ending points
    endPtX = (float *)malloc( ft->numPaths * sizeof( float ) );
    endPtY = (float *)malloc( ft->numPaths * sizeof( float ) );
    endPtZ = (float *)malloc( ft->numPaths * sizeof( float ) );
    for( i = 0; i < ft->numPaths; ++ i )
    {
        ptIdx = ft->path[i].ptIndex[0];
        endPtX[i] = ft->ptx[ptIdx];
        endPtY[i] = ft->pty[ptIdx];
        endPtZ[i] = ft->ptz[ptIdx];
    }
    scatterXMean = calcMeanVariance( &scatterXVariance, endPtX, ft->numPaths ) 
        * dx;
    scatterXVariance = scatterXVariance * dx*dx;
    scatterYMean = calcMeanVariance( &scatterYVariance, endPtY, ft->numPaths ) 
        * dy;
    scatterYVariance = scatterYVariance * dy*dy;
    scatterZMean = calcMeanVariance( &scatterZVariance, endPtZ, ft->numPaths ) 
        * dz;
    scatterZVariance = scatterZVariance * dz*dz;
    free( endPtX );
    free( endPtY );
    free( endPtZ );

	       // Stats-4:
               // calculate the fiber length 
    pathLength = (float *)malloc( ft->numPaths * sizeof( float ) );
    for( i = 0; i < ft->numPaths; ++ i )
    {
        pathLength[i] = 0;
	for( j = 0; j < ft->path[i].numPts-1; ++ j )
	{
            ptIdx = ft->path[i].ptIndex[j]; 
	    x0 = ft->ptx[ptIdx];
	    y0 = ft->pty[ptIdx];
	    z0 = ft->ptz[ptIdx];
	    ptIdx = ft->path[i].ptIndex[j+1];
	    x1 = ft->ptx[ptIdx];
	    y1 = ft->pty[ptIdx];
	    z1 = ft->ptz[ptIdx];
	    pathLength[i] += sqrt( (x0-x1)*(x0-x1)*dx*dx + 
	        (y0-y1)*(y0-y1)*dy*dy + (z0-z1)*(z0-z1)*dz*dz );
	}
    }
    fiberLengthMean = calcMeanVariance( &fiberLengthVariance, pathLength, 
        ft->numPaths );
    fiberLengthSum = fiberLengthMean * (ft->numPaths);
    free( pathLength );
               // create a file to write fiber statistics 
    if( (fp = fopen( args->fiberStatsFileName, "w" )) == NULL )
    {
        fprintf( stderr, "cannot create file: %s\n", args->fiberStatsFileName );
	return -1;
    }
    fprintf( fp, "# %s\n", args->fiberStatsFileName );
    fprintf( fp, "# Fiber Tract Statistics\n" );    
    fprintf( fp, "fractional anistropy:        %.2f\n", args->tensorFA );
    fprintf( fp, "distance ratio:              %.2f\n", args->distRatio );
    fprintf( fp, "curvature angle:             %.0f\n", args->ipVectors );
    fprintf( fp, "coherence angle:             %.0f\n", args->ipSeedVecNode );
    fprintf( fp, "integration time out:        %.4f\n", args->diffIntegTime );
    fprintf( fp, "b0 lower bound:              %g\n", args->b0_lo );
    fprintf( fp, "b0 upper bound:              %g\n", args->b0_hi );
    fprintf( fp, "-----------------------------------------\n" );
    fprintf( fp, "number of seeds:             %d\n", args->roiSize );
    fprintf( fp, "number of effective seeds:   %d\n", args->numTractFiles );
    fprintf( fp, "number of fiber paths:       %d\n", args->totalNumPaths );
    fprintf( fp, "fiber tract FA mean:         %.4f\n", fiberFAMean );
    fprintf( fp, "fiber tract FA variance:     %f\n", fiberFAVariance );
    fprintf( fp, "fiber tract MD mean:         %.4f\n", fiberMDMean );
    fprintf( fp, "fiber tract MD variance:     %f\n", fiberMDVariance );
    fprintf( fp, "brain volume (voxels):       %d\n", brainVolume );
    fprintf( fp, "brain volume (mm^3):         %.2f\n", brainVolume*dx*dy*dz );
    fprintf( fp, "fiber tract volume (voxels): %d\n", fiberVolume );
    fprintf( fp, "fiber tract volume (mm^3):   %.2f\n", fiberVolume*dx*dy*dz );
    fprintf( fp, "normalized fiber volume:     %.2e\n", 
        fiberVolume*1.0/brainVolume );
    fprintf( fp, "fiber volume FA mean:        %.4f\n", volumeFAMean );
    fprintf( fp, "fiber volume FA variance:    %f\n", volumeFAVariance );
    fprintf( fp, "fiber volume MD mean:        %.4f\n", volumeMDMean );
    fprintf( fp, "fiber volume MD variance:    %f\n", volumeMDVariance );
    fprintf( fp, "scatter X mean:              %.2f\n", scatterXMean );
    fprintf( fp, "scatter X variance:          %.4f\n", scatterXVariance );
    fprintf( fp, "scatter Y mean:              %.2f\n", scatterYMean );
    fprintf( fp, "scatter Y variance:          %.4f\n", scatterYVariance );
    fprintf( fp, "scatter Z mean:              %.2f\n", scatterZMean );
    fprintf( fp, "scatter Z variance:          %.4f\n", scatterZVariance );
    fprintf( fp, "fiber length mean (mm):      %.2f\n", fiberLengthMean );
    fprintf( fp, "fiber length variance:       %.4f\n", fiberLengthVariance );
    fprintf( fp, "fiber length sum (mm):       %.2f", fiberLengthSum );

    fclose( fp );
    return 0;
}

/***********************************************************************
 * Calculate mean and variance.
 * Algorithm source:
 * http://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
 ***********************************************************************/
double calcMeanVariance( double *variance, float *array, int size )
{
    int i;
    double mean, delta, residual;
    float value;

    mean = 0;
    residual = 0;
    for( i = 0; i < size; ++ i )
    {
        value = array[i];
        delta = value - mean;
	mean += delta/(i+1);
	residual += delta * ( value - mean );
    }
    
    *variance = residual/(size-1);
    return mean;
}



/***********************************************************************
 * Save data to file in VTK data format.
 ***********************************************************************/
int saveFTDataVTK( void *ftData, char *filename, float dx, float dy, float dz )
{
    FILE *fp;
    FT_ ft;
    int i, j;

    ft = (FT_)ftData;
                                // create a file for writing data
    if( (fp = fopen( filename, "w" )) == NULL )
    {
        fprintf( stderr, "cannot create file: %s\n", filename );
	return -1;
    }  
			        // write header info to data file
    fprintf( fp, "# vtk DataFile Version 2.0\n" );
printf( "# vtk DataFile Version 2.0\n" );
    fprintf( fp, "Fiber Tract\n" );
    fprintf( fp, "ASCII\n" );
    fprintf( fp, "DATASET POLYDATA\n" );
    fprintf( fp, "POINTS %d float\n", ft->size);
    for( i = 0; i < ft->size; ++ i )
        fprintf( fp, "%.2f %.2f %.2f\n", (ft->ptx[i])*dx,
	    (ft->pty[i])*dy, (ft->ptz[i])*dz );
    fprintf( fp, "LINES %d %d\n", ft->numPaths, ft->size+ft->numPaths );
    for( i = 0; i < ft->numPaths; ++ i )
    {
        fprintf( fp, "%d ", ft->path[i].numPts );
	       // points lined up from root to end
	for( j = ft->path[i].numPts-1; j >= 0; j -- )
	    fprintf( fp, "%d ", ft->path[i].ptIndex[j] );
	fprintf( fp, "\n" );    
    }
    fprintf( fp, "POINT_DATA %d\n", ft->size);
    fprintf( fp, "SCALARS scalars float\n" );
    fprintf( fp, "LOOKUP_TABLE default\n" );
    for( i = 0; i < ft->size; ++ i )
        fprintf( fp, "%.4f\n", ft->ptFA[i] );
    fclose( fp );	
    return 0;	
}

/***********************************************************************
 * Group the points for the corresponding paths according to
 * the VTK polydata format. 
 ***********************************************************************/
void splitFTData( void *ftData )
{
    FT_ ft;
    int i, counter, pathInd;
        
    ft = (FT_)ftData;

    counter = 0;
    pathInd = 1;
    for( i = 0; i < ft->size; ++ i )
    {
        if( ft->ptn[i] == pathInd )
	{
	    ft->path[pathInd-1].ptIndex[counter] = i;
	    counter ++;
	}
	else
	{
	    pathInd ++;    
	    counter = 0;
	    ft->path[pathInd-1].ptIndex[counter] = i;
	    counter ++;
        }
    }
    return;
}

/***********************************************************************
 * Allocate memory for data structures and 
 * read data points of fiber paths.
 ***********************************************************************/
FT_ ftData_new( void *argsTable, char *dir )
{
    Args_Table_ args = (Args_Table_)argsTable;
    FILE *fp;
    FT_ ft;
    int i, n, numPaths, seq, x, y, z, nodeCounter, pathCounter;
    char file[STR_LEN];
    float fa, md;

    n = args->totalNumNodes;
    numPaths = args->totalNumPaths;
                   // allocate memory and initialize
    ft = (FT_)malloc( sizeof( *ft ) );
//    ft = (FT_)malloc( sizeof( __FT_ ) );
    ft->origSize = n;
    ft->numPaths = numPaths;
    
//    ft->origPtn = (int *)malloc( n*sizeof( int ) );
    ft->origX = (int *)malloc( n*sizeof( int ) );
    ft->origY = (int *)malloc( n*sizeof( int ) );
    ft->origZ = (int *)malloc( n*sizeof( int ) );
    ft->origFA = (float *)malloc( n*sizeof( float ) );
    ft->origMD = (float *)malloc( n*sizeof( float ) );
    ft->origPts = (int *)malloc( numPaths * sizeof( int ) );
    memset( ft->origPts, 0, numPaths * sizeof( int ) );
                   // read data points
    nodeCounter = 0; pathCounter = 0;
    for( i = 0; i < args->numTractFiles; ++ i )
    {
        sprintf( file, "%s%s%d", dir, args->tractbase, i );
        if( (fp = fopen( file, "r" )) == NULL )
        {
            fprintf( stderr, "cannot open file: %s\n", file ); return NULL;
        }
        while( !feof( fp ) )
	{
	    fscanf( fp, "%d%d%d%d%f%f", &seq, &x, &y, &z, &fa, &md ); 
//	    ft->origPtn[nodeCounter] = pathCounter + seq;
	    ft->origX[nodeCounter] = x;
	    ft->origY[nodeCounter] = y;
	    ft->origZ[nodeCounter] = z;
	    ft->origFA[nodeCounter] = fa;
	    ft->origMD[nodeCounter] = md;
	    ft->origPts[pathCounter+seq-1] ++;
	    nodeCounter ++;
	           // The last line in the last file should NOT be read in,
		   // since it is empty and will cause accessing memory
		   // out of bounds!
            if( nodeCounter == n ) break;
        } 
	           // if it is not the last file
	if( nodeCounter < n )  
	{
            ft->origPts[pathCounter+seq-1] --; // since the last line is emtpy
	    nodeCounter --; // since the last line is empty
        }
	pathCounter = pathCounter + seq;
	fclose( fp );
    }
/*
fp=fopen( "zzzz","w");
for(i=0;i<n;++i)
fprintf(fp,"%6d%4d%4d%4d%10.4f\n",ft->origPtn[i],ft->origX[i],ft->origY[i],
ft->origZ[i],ft->origFA[i]);
*/
    ft->path = (PATH_)malloc( numPaths*sizeof( *(ft->path) ) );
//    ft->path = (PATH_)malloc( numPaths*sizeof( __PATH_ ) );
	
                        /* now n becomes the number of approximated points */
    n = (n - 1) * APPR_PTS_NUM + n - (numPaths - 1) * APPR_PTS_NUM;
    ft->size = n;
    ft->ptn = (int *)malloc( n*sizeof( int ) );
    ft->ptx = (float *)malloc( n*sizeof( float ) );
    ft->pty = (float *)malloc( n*sizeof( float ) );
    ft->ptz = (float *)malloc( n*sizeof( float ) );
    ft->ptFA = (float *)malloc( n*sizeof( float ) );
    ft->ptMD = (float *)malloc( n*sizeof( float ) );

    ft->tft_head = NULL;
    return ft;
}

/***********************************************************************
 * Free memory.
 ***********************************************************************/
void ftData_free( void *ftData )
{
    FT_ ft = (FT_)ftData;
    TrimmedFT_ tft, tft_cur;
    int i;
    
    for( i = 0; i < ft->numPaths; i ++ )
        free( ft->path[i].ptIndex );
    free( ft->path );
    free( ft->ptn );
    free( ft->ptx );
    free( ft->pty );
    free( ft->ptz );
    free( ft->ptFA ); 
    free( ft->ptMD ); 
//    free( ft->origPtn );
    free( ft->origPts );
    free( ft->origX );
    free( ft->origY );
    free( ft->origZ );
    free( ft->origFA );
    free( ft->origMD );

    tft = ft->tft_head;
    while( tft != NULL )
    {
        tft_cur = tft;
	tft = tft->next;
	free( tft_cur );
    }
  
    free( ft );
    return;
}

/***********************************************************************
 * do approximation
 ***********************************************************************/
void approximation( void *ftData, int spline_order, int num_pp )
{
    FT_ ft;
    int i, j, origDisp, origPts, disp, pts;

    ft = (FT_)ftData;
                        /* do approximation for each path */
    origDisp = 0;
    disp = 0;
//printf("number of paths: %d\n",ft->numPaths);    
    for( i = 0; i < ft->numPaths; ++ i )
    {
        origPts = ft->origPts[i];
	
	pts = origPts + APPR_PTS_NUM * ( origPts - 1 );
//printf("***********i=%d pts=%d\n",i,pts);			
                        /* approximate values in each coordinate */
	l2apprSpline( ft->ptx+disp, pts, ft->origX+origDisp, origPts, 
	    spline_order, num_pp );
	l2apprSpline( ft->pty+disp, pts, ft->origY+origDisp, origPts,
	    spline_order, num_pp );
	l2apprSpline( ft->ptz+disp, pts, ft->origZ+origDisp, origPts,
	    spline_order, num_pp );
	l2apprSpline_attr( ft->ptFA+disp, pts, ft->origFA+origDisp, origPts,
	    spline_order, num_pp );
	l2apprSpline_attr( ft->ptMD+disp, pts, ft->origMD+origDisp, origPts,
	    spline_order, num_pp );
//printf("l2apprSpline: done\n");	
                        /* update the sequence number and the number of points
			   on each path */
        for( j = 0; j < pts; ++ j )
	    ft->ptn[disp+j] = i+1;

	ft->path[i].numPts = pts;
	ft->path[i].ptIndex = (int *)malloc( pts*sizeof( int ) );
	
	origDisp += origPts;
	disp += pts;
    }    
    return;
}

/***********************************************************************
 * least square approximation with b-splines
 ***********************************************************************/
void l2apprSpline( float *outPt, int outSize, int *inPt, int inSize,
    int spline_order, int num_pp )
{
    int l,              /* number of polynomial pieces */ 
        k,              /* order of the spline */
	ntau,           /* number of data points to be fitted */
	n,              /* length of bcoef and dimension of spline space */
	prfun,          /* control switch for printing out error info */ 
	jderiv;         /* the order of the derivative to be evaluated */
    float *tau,         /* (tau[i], gtau[i]), i=1...ntau, are the ntau */
        *gtau,          /* data points to be fitted; tau ranges from 0 to 1. */ 
	*brkpt,         /* breakpoint sequence */
	*weight,        /* the corresponding weights */
        *t,             /* the knot sequence */
        *q,             /* work array of size at least k*n */
        *scrtch,        /* work array of size at least n */ 
        *bcoef,         /* b-spline coefficients */ 
        *coef;          /* coefficients of pp-representation for approx */
    float totwt = 0;
    int i;
    float step, disp, val;
    int lpkmax= 10*inSize, ntmax = 20*inSize, ltkmax = 200*inSize;
    
                        /* information about the function to be approximated 
			   and order and breakpoint sequence of the 
			   approximating pp functions */
    k = spline_order;
    l = num_pp;
    ntau = inSize;
    tau = (float *)malloc( /*ntau*/ntmax * sizeof( float ) );
    gtau = (float *)malloc( /*ntau*/ntmax * sizeof( float ) );
    weight = (float *)malloc( /*ntau*/ntmax * sizeof( float ) );
    brkpt = (float *)malloc( /*(l+1)*/lpkmax * sizeof( float ) );
    step = 1.0/(inSize-1);
    for( i = 0; i < inSize; ++ i )
    {                   /* tau ranges from 0 to 1 */
        tau[i] = i * step;
	gtau[i] = (float)inPt[i];
	weight[i] = 1;
        totwt += weight[i];
    }
    
    step = 1.0/l;
    for( i = 0; i < l+1; ++ i )
        brkpt[i] = i * step;

    n = k - 1 + l;
    t = (float *)malloc( /*(n+k)*/ntmax * sizeof( float ) );
    q = (float *)malloc( /*k*n*/ltkmax * sizeof( float ) );
    scrtch = (float *)malloc( /*n*/ntmax * sizeof( float ) );
    bcoef = (float *)malloc( /*n*/lpkmax * sizeof( float ) );
    coef = (float *)malloc( /*k*l*/ltkmax * sizeof( float ) );
/*
printf("knots:\n");
for(i=0;i<n+k;++i)
printf("t[%d] = %.4f\n",i,t[i]);
*/
                        /* breakpoints are translated into knots, and the 
			   number of b-splines to be used is obtained by */
    l2knts_( brkpt, &l, &k, t, &n ); 

                        /* the b-spline coeffs,'bcoef' of the l2-approx. */
    l2appr_( t, &n, &k, q, scrtch, bcoef, &ntau, tau, gtau, weight );
/* 
printf("b-spline coeffs:\n");
for(i=0;i<n;++i)
printf("%.4f ",bcoef[i]); 
*/    
                        /* convert the b-repr. of the approx to pp repr. */
    bsplpp_( t, bcoef, &n, &k, q, brkpt, coef, &l );

/*
printf("\npp-repr:\n");
int j,ii;
for(i=0;i<l;++i)
{printf("%.3f   ",brkpt[i] );
ii=i*k;
for(j=0;j<k;++j)
printf( "%.3f ",coef[ii+j]);
printf("\n");
}
*/
                        /* compute and print out various error norms */
    prfun = 0;
    l2err_( &prfun, scrtch, q, &ntau, tau, gtau, weight, &totwt, 
       brkpt, coef, &l, &k );
//printf("error printing done\n");
                        /* get the approximated data points */
    jderiv = 0;
    step = 1.0 / (outSize-1);	
    for( i = 0; i < outSize; ++ i )
    {
        disp = i * step;
        ppval_( &val, brkpt, coef, &l, &k, &disp, &jderiv );
//	outPt[i] = (int)(FACTOR * val);
	outPt[i] = val;
    }

    free( tau );
    free( gtau );
    free( brkpt );
    free( weight );
    free( t );
    free( q );
    free( scrtch );
    free( bcoef );
    free( coef );
}

/***********************************************************************
 * least square approximation with b-splines: for attribute
 ***********************************************************************/
void l2apprSpline_attr( float *outPt, int outSize, float *inPt, int inSize,
    int spline_order, int num_pp )
{
    int l,              /* number of polynomial pieces */ 
        k,              /* order of the spline */
	ntau,           /* number of data points to be fitted */
	n,              /* length of bcoef and dimension of spline space */
	prfun,          /* control switch for printing out error info */ 
	jderiv;         /* the order of the derivative to be evaluated */
    float *tau,         /* (tau[i], gtau[i]), i=1...ntau, are the ntau */
        *gtau,          /* data points to be fitted; tau ranges from 0 to 1. */ 
	*brkpt,         /* breakpoint sequence */
	*weight,        /* the corresponding weights */
        *t,             /* the knot sequence */
        *q,             /* work array of size at least k*n */
        *scrtch,        /* work array of size at least n */ 
        *bcoef,         /* b-spline coefficients */ 
        *coef;          /* coefficients of pp-representation for approx */
    float totwt = 0;
    int i;
    float step, disp, val;
    int lpkmax= 10*inSize, ntmax = 20*inSize, ltkmax = 200*inSize;
    
                        /* information about the function to be approximated 
			   and order and breakpoint sequence of the 
			   approximating pp functions */
    k = spline_order;
    l = num_pp;
    ntau = inSize;
    tau = (float *)malloc( /*ntau*/ntmax * sizeof( float ) );
    gtau = (float *)malloc( /*ntau*/ntmax * sizeof( float ) );
    weight = (float *)malloc( /*ntau*/ntmax * sizeof( float ) );
    brkpt = (float *)malloc( /*(l+1)*/lpkmax * sizeof( float ) );
    step = 1.0/(inSize-1);
    for( i = 0; i < inSize; ++ i )
    {                   /* tau ranges from 0 to 1 */
        tau[i] = i * step;
	gtau[i] = inPt[i];
	weight[i] = 1;
        totwt += weight[i];
    }
    
    step = 1.0/l;
    for( i = 0; i < l+1; ++ i )
        brkpt[i] = i * step;

    n = k - 1 + l;
    t = (float *)malloc( /*(n+k)*/ntmax * sizeof( float ) );
    q = (float *)malloc( /*k*n*/ltkmax * sizeof( float ) );
    scrtch = (float *)malloc( /*n*/ntmax * sizeof( float ) );
    bcoef = (float *)malloc( /*n*/lpkmax * sizeof( float ) );
    coef = (float *)malloc( /*k*l*/ltkmax * sizeof( float ) );
/*
printf("knots:\n");
for(i=0;i<n+k;++i)
printf("t[%d] = %.4f\n",i,t[i]);
*/
                        /* breakpoints are translated into knots, and the 
			   number of b-splines to be used is obtained by */
    l2knts_( brkpt, &l, &k, t, &n ); 

                        /* the b-spline coeffs,'bcoef' of the l2-approx. */
    l2appr_( t, &n, &k, q, scrtch, bcoef, &ntau, tau, gtau, weight );
/* 
printf("b-spline coeffs:\n");
for(i=0;i<n;++i)
printf("%.4f ",bcoef[i]); 
*/    
                        /* convert the b-repr. of the approx to pp repr. */
    bsplpp_( t, bcoef, &n, &k, q, brkpt, coef, &l );

/*
printf("\npp-repr:\n");
int j,ii;
for(i=0;i<l;++i)
{printf("%.3f   ",brkpt[i] );
ii=i*k;
for(j=0;j<k;++j)
printf( "%.3f ",coef[ii+j]);
printf("\n");
}
*/                        /* compute and print out various error norms */
    prfun = 0;
    l2err_( &prfun, scrtch, q, &ntau, tau, gtau, weight, &totwt, 
       brkpt, coef, &l, &k );

                        /* get the approximated data points */
    jderiv = 0;
    step = 1.0 / (outSize-1);	
    for( i = 0; i < outSize; ++ i )
    {
        disp = i * step;
        ppval_( &val, brkpt, coef, &l, &k, &disp, &jderiv );
	outPt[i] = val;
    }

    free( tau );
    free( gtau );
    free( brkpt );
    free( weight );
    free( t );
    free( q );
    free( scrtch );
    free( bcoef );
    free( coef );
}


/***********************************************************************
 * test if a given number is a prime number
 ***********************************************************************/
int isPrime( int n )
{
    int test;
    for( test = 2; test <= (int)sqrt(n); ++ test )
    if( n%test != 0 )
        continue;
    else
        return 0;
    return 1;
}

/***********************************************************************
 * the hash function 
 ***********************************************************************/
int hashFunc( int hashTableSize, int x, int y, int z )
{
    int rem;   // remainder
    rem = 3*x + 17*y + 41*z;
    rem %= hashTableSize;
    return rem;
}

/***********************************************************************
 * allocate memory for hash table and do initialization
 ***********************************************************************/
TargetPt_ *hashTable_new( int hashTableSize, int ptArraySize, int *pt[3] )
{
    TargetPt_ *hashTable;
    int i;

    hashTable = (TargetPt_ *)malloc( hashTableSize * sizeof( TargetPt_ ) );
    for( i = 0; i < hashTableSize; ++ i )
        hashTable[i] = NULL;
    for( i = 0; i < ptArraySize; ++ i )
        hashInsert( hashTable, hashTableSize, pt[0][i], pt[1][i], pt[2][i] );
    return hashTable;	
}

/***********************************************************************
 * insert an entry into hash table
 ***********************************************************************/
void hashInsert( void *hashTable, int hashTableSize, int x, int y, int z )
{
    TargetPt_ *table = (TargetPt_ *)hashTable;
    TargetPt_ node, newNode;
    int hashNum;
    
    newNode = (TargetPt_)malloc( sizeof( *newNode ) );
    newNode->x = x;
    newNode->y = y;
    newNode->z = z;
    newNode->next = NULL;
                   // get the hash number
    hashNum = hashFunc( hashTableSize, x, y, z );
                   // no linked list
    if( table[hashNum] == NULL )
        table[hashNum] = newNode;
    else           // linked list exists and insert at the end
    {
        node = table[hashNum];
	while( node->next != NULL )
	    node = node->next;
	node->next = newNode;
    }
}

/***********************************************************************
 * lookup an entry in hash table 
 ***********************************************************************/
int hashLookup( void *hashTable, int hashTableSize, int x, int y, int z )
{
    TargetPt_ *table = (TargetPt_ *)hashTable;
    TargetPt_ node;
    int hashNum;
                   // get the hash number
    hashNum = hashFunc( hashTableSize, x, y, z );
                   // head of the linked list
    node = table[hashNum];
                   // search the linked list
    while( node != NULL )
    {
        if( node->x == x && node->y == y && node->z == z ) return 1;
	node = node->next;
    }
    return 0;
}

/***********************************************************************
 * retrieve fiber tracts that run hrough the target region
 ***********************************************************************/
void trimFTData( void *ftData, void *argsTable )
{
    FT_ ft = (FT_)ftData;
    Args_Table_ args = (Args_Table_)argsTable;
    TargetPt_ *hashTable, node, node_cur;
    int hashTableSize;
    TrimmedFT_ tft, tft_prev;
    int i, j, offset, disp, x, y, z, x0, y0, z0;
                   // find a prime number to be the hash table size
    hashTableSize = args->roiSizeTarget;
    while( !isPrime( hashTableSize ) ) { hashTableSize ++; }
                   // get the hash table storing points in target
    hashTable = hashTable_new( hashTableSize, args->roiSizeTarget, 
        args->roiPtsTarget );
                   // create a linked list to store retrieved paths
    ft->tft_head = (TrimmedFT_)malloc( sizeof( *(ft->tft_head) ) );
    ft->tft_head->next = NULL;
    tft_prev = ft->tft_head;
    disp = 0; 
    ft->trimmedNumPaths = 0; 
    ft->trimmedSize = 0;
    for( i = 0; i < ft->numPaths; ++ i )
    {
        x0 = -1; y0 = -1; z0 = -1;
        for( j = ft->path[i].numPts-1; j >= 0; j -- )
        {
            offset = ft->path[i].ptIndex[j];
            x = (int)ft->ptx[offset];
	    y = (int)ft->pty[offset];
	    z = (int)ft->ptz[offset];
	    if( x == x0 && y == y0 && z == z0 ) continue;
	           // the point got a hit in hash table
	    if( hashLookup( hashTable, hashTableSize, x, y, z ) )   
            {
	        ft->trimmedNumPaths ++;
		ft->trimmedSize += ft->path[i].numPts;
                tft = (TrimmedFT_)malloc( sizeof( *tft ) );
//	        tft->ptOffset = ft->path[i].ptIndex[0];
	        tft->pathOffset = i;
	        tft->ptIdxDisp = disp;
		tft->next = NULL;
		tft_prev->next = tft;
		tft_prev = tft;
                break;
	    }
	    x0 = x; y0 = y; z0 = z;
        }
	           // record total number of points on paths 
		   // which are not running through target
        if( j < 0 ) disp += ft->path[i].numPts;
    }
                   // free the memory allocated to hash table
    for( i = 0; i < hashTableSize; ++ i )
    {
        node = hashTable[i]; 
	while( node != NULL )
	{
	    node_cur = node;
	    node = node->next;
	    free( node_cur );
	}    
    }
    free( hashTable );
    return;
}

/***********************************************************************
 * Save trimmed data to file in VTK data format.
 ***********************************************************************/
int saveTrimmedFTDataVTK( void *ftData, char *filename, float dx, float dy, 
    float dz )
{
    FILE *fp;
    FT_ ft = (FT_)ftData;
    TrimmedFT_ tft;
    int i, idx, offset;

    if( ft->trimmedNumPaths == 0 )
    {
        fprintf( stderr, 
	    ">> NO fiber tract running through the target zone!\n" );
	return -1;
    }

    if( (fp = fopen( filename, "w" )) == NULL )
    {
        fprintf( stderr, "cannot create file: %s\n", filename );
	return -1;
    }        
		   // write header info to data file
    fprintf( fp, "# vtk DataFile Version 2.0\n" );
printf( "# vtk DataFile Version 2.0: trimmed\n" );
    fprintf( fp, "Fiber Tract\n" );
    fprintf( fp, "ASCII\n" );
    fprintf( fp, "DATASET POLYDATA\n" );
    fprintf( fp, "POINTS %d float\n", ft->trimmedSize);
                   // save coordinates of the points 
    tft = ft->tft_head->next;
    while( tft != NULL )
    {
        offset = tft->pathOffset;
        for( i = 0; i < ft->path[offset].numPts; ++ i )
	{
	    idx = ft->path[offset].ptIndex[i];
            fprintf( fp, "%.2f %.2f %.2f\n", (ft->ptx[idx])*dx,
	        (ft->pty[idx])*dy, (ft->ptz[idx])*dz );
	}    
        tft = tft->next;
    }
                   // save paths
    fprintf( fp, "LINES %d %d\n", ft->trimmedNumPaths, 
        ft->trimmedSize + ft->trimmedNumPaths );

    tft = ft->tft_head->next;
    while( tft != NULL )
    {
        offset = tft->pathOffset;
        fprintf( fp, "%d ", ft->path[offset].numPts );
	           // points lined up from root to end
        for( i = ft->path[offset].numPts-1; i >= 0; i -- )
	    fprintf( fp, "%d ", ft->path[offset].ptIndex[i] - tft->ptIdxDisp );
	fprintf( fp, "\n" );
	tft = tft->next;
    }
                   // save attributes
    fprintf( fp, "POINT_DATA %d\n", ft->trimmedSize);
    fprintf( fp, "SCALARS scalars float\n" );
    fprintf( fp, "LOOKUP_TABLE default\n" );
    tft = ft->tft_head->next;
    while( tft != NULL )
    {
        offset = tft->pathOffset;
	for( i = 0; i < ft->path[offset].numPts; ++ i )
	{
	    idx = ft->path[offset].ptIndex[i];
	    fprintf( fp, "%.4f\n", ft->ptFA[idx] );
	}    
	tft = tft->next;    
    }

    fclose( fp );	
    return 0;	
}

/***********************************************************************
 * Calculate statistics for trimmed fiber tracts. 
 ***********************************************************************/
int trimmedFiberStats( void *ftData, void *argsTable )
{
    Args_Table_ args = (Args_Table_)argsTable;
    FT_ ft = (FT_)ftData;
    TrimmedFT_ tft;
    FILE *fp;
    int i, cntpt, cntpath, x, y, z, idx, offset;
    double fiberFAMean, fiberFAVariance, volumeFAMean, volumeFAVariance,
	fiberMDMean, fiberMDVariance, volumeMDMean, volumeMDVariance,
        scatterXMean, scatterYMean, scatterZMean,
	scatterXVariance, scatterYVariance, scatterZVariance,
        fiberLengthMean, fiberLengthVariance, fiberLengthSum;
    int fiberVolume, brainVolume = 0;
    int *voxelIdx; // array for holding voxel indices
    float *fiberFA, *voxelFA; // array for holding FA values 
    float *fiberMD, *voxelMD; // array for holding MD values 
    float *endPtX, *endPtY, *endPtZ; // coordinates of fiber ending points
    float *pathLength; // the length of each fiber tract
    float x1=0, y1=0, z1=0, x0, y0, z0;
    float dx, dy, dz;
    int nxy = args->nx*(args->ny);

    dx = args->dx; dy = args->dy; dz = args->dz;
               // calculate the brain volume using b0 image
//    brainVolume = computeBrainVolume( args->dwi_b0, args->nx, args->ny,
//        args->nz );
    if( args->dwi_b0 != NULL )
    for( i = 0; i < args->nvox; ++ i )
    if( args->dwi_b0[i] >= B0_BRAIN_VOLUME_THRESHOLD ) brainVolume ++;
	       // allocate memory 
    fiberFA = (float *)malloc( ft->trimmedSize * sizeof( float ) );
    fiberMD = (float *)malloc( ft->trimmedSize * sizeof( float ) );
    voxelIdx = (int *)malloc( ft->trimmedSize * sizeof( int ) );
    voxelFA = (float *)malloc( ft->trimmedSize * sizeof( float ) );
    voxelMD = (float *)malloc( ft->trimmedSize * sizeof( float ) );
    endPtX = (float *)malloc( ft->trimmedNumPaths * sizeof( float ) );
    endPtY = (float *)malloc( ft->trimmedNumPaths * sizeof( float ) );
    endPtZ = (float *)malloc( ft->trimmedNumPaths * sizeof( float ) );
    pathLength = (float *)malloc( ft->trimmedNumPaths * sizeof( float ) );
	       // go through the linked list of trimmed fiber paths 
    tft = ft->tft_head->next;
    cntpt = 0; cntpath = 0;
    while( tft != NULL )
    {
        offset = tft->pathOffset;
	           // end point index for a path
        idx = ft->path[offset].ptIndex[0];
	endPtX[cntpath] = ft->ptx[idx];
	endPtY[cntpath] = ft->pty[idx];
	endPtZ[cntpath] = ft->ptz[idx];
	pathLength[cntpath] = 0;
	for( i = 0; i < ft->path[offset].numPts; ++ i )
	{
	    idx = ft->path[offset].ptIndex[i];
	    fiberFA[cntpt] = ft->ptFA[idx];
	    fiberMD[cntpt] = ft->ptMD[idx];
	    x = (int)( x0 = ft->ptx[idx] );
	    y = (int)( y0 = ft->pty[idx] );
	    z = (int)( z0 = ft->ptz[idx] );
	    voxelIdx[cntpt] = nodeIndex( x+1, y+1, z+1, args->nx, nxy ) - 1;
	    cntpt ++;
	           // calculate path length
	    if( i > 0 )
	        pathLength[cntpath] += sqrt( (x0-x1)*(x0-x1)*dx*dx + 
	            (y0-y1)*(y0-y1)*dy*dy + (z0-z1)*(z0-z1)*dz*dz );
	    x1 = x0;
	    y1 = y0;
	    z1 = z0;
	}   
	cntpath ++;
	tft = tft->next;    
    }
               // Stats-1: 
	       // mean and variance for FA on individual fibers
    fiberFAMean = calcMeanVariance( &fiberFAVariance, fiberFA, 
        ft->trimmedSize );
    fiberMDMean = calcMeanVariance( &fiberMDVariance, fiberMD, 
        ft->trimmedSize );
               // Stats-2:
               // fiber volume: number of unique voxels fibers go through
               // sort the array of voxel indices
    QSORT( int, voxelIdx, ft->trimmedSize, int_lt ); 
               // count the number of unique elements in voxelIdx
    fiberVolume = 1;
    voxelFA[0] = args->fa[voxelIdx[0]];
    voxelMD[0] = args->md[voxelIdx[0]];
    for( i = 0; i < ft->trimmedSize-1; ++ i )
    if( voxelIdx[i] != voxelIdx[i+1] )
    {
	voxelFA[fiberVolume] = args->fa[voxelIdx[i+1]];
	voxelMD[fiberVolume] = args->md[voxelIdx[i+1]];
	fiberVolume ++;
    }    
    volumeFAMean = calcMeanVariance( &volumeFAVariance, voxelFA, 
        fiberVolume );
    volumeMDMean = calcMeanVariance( &volumeMDVariance, voxelMD, 
        fiberVolume );
	       // Stats-3:
	       // calculate the diversification of fiber scattering, i.e.,
	       // calculate the mean and variance of fiber ending points
    scatterXMean = calcMeanVariance( &scatterXVariance, endPtX, 
        ft->trimmedNumPaths ) * dx;
    scatterXVariance = scatterXVariance * dx*dx;
    scatterYMean = calcMeanVariance( &scatterYVariance, endPtY, 
        ft->trimmedNumPaths ) * dy;
    scatterYVariance = scatterYVariance * dy*dy;
    scatterZMean = calcMeanVariance( &scatterZVariance, endPtZ, 
        ft->trimmedNumPaths ) * dz;
    scatterZVariance = scatterZVariance * dz*dz;
	       // Stats-4:
               // calculate the fiber length 
    fiberLengthMean = calcMeanVariance( &fiberLengthVariance, pathLength, 
        ft->trimmedNumPaths );
    fiberLengthSum = fiberLengthMean * (ft->trimmedNumPaths);
               // free memory
    free( fiberFA );
    free( fiberMD );
    free( voxelIdx );
    free( voxelFA );
    free( voxelMD );
    free( endPtX );
    free( endPtY );
    free( endPtZ );
    free( pathLength );
               // create a file to write fiber statistics 
    if( (fp = fopen( args->fiberStatsFileName, "w" )) == NULL )
    {
        fprintf(stderr,"cannot create file: %s\n",args->fiberStatsFileName);
	return -1;
    }        
    fprintf( fp, "# %s\n", args->fiberStatsFileName );
    fprintf( fp, "# Fiber Tract Statistics\n" );    
    fprintf( fp, "fractional anistropy:        %.2f\n", args->tensorFA );
    fprintf( fp, "distance ratio:              %.2f\n", args->distRatio );
    fprintf( fp, "curvature angle:             %.0f\n", args->ipVectors );
    fprintf( fp, "coherence angle:             %.0f\n", args->ipSeedVecNode );
    fprintf( fp, "integration time out:        %.4f\n", args->diffIntegTime );
    fprintf( fp, "b0 lower bound:              %g\n", args->b0_lo );
    fprintf( fp, "b0 upper bound:              %g\n", args->b0_hi );
    fprintf( fp, "-----------------------------------------\n" );
    fprintf( fp, "number of seeds:             %d\n", args->roiSize );
    fprintf( fp, "number of effective seeds:   %d\n", args->numTractFiles );
    fprintf( fp, "number of fiber paths:       %d\n", ft->trimmedNumPaths );
    fprintf( fp, "fiber tract FA mean:         %.4f\n", fiberFAMean );
    fprintf( fp, "fiber tract FA variance:     %f\n", fiberFAVariance );
    fprintf( fp, "fiber tract MD mean:         %.4f\n", fiberMDMean );
    fprintf( fp, "fiber tract MD variance:     %f\n", fiberMDVariance );
    fprintf( fp, "brain volume (voxels):       %d\n", brainVolume );
    fprintf( fp, "brain volume (mm^3):         %.2f\n", brainVolume*dx*dy*dz );
    fprintf( fp, "fiber tract volume (voxels): %d\n", fiberVolume );
    fprintf( fp, "fiber tract volume (mm^3):   %.2f\n", fiberVolume*dx*dy*dz );
    fprintf( fp, "normalized fiber volume:     %.2e\n", 
        fiberVolume*1.0/brainVolume );
    fprintf( fp, "fiber volume FA mean:        %.4f\n", volumeFAMean );
    fprintf( fp, "fiber volume FA variance:    %f\n", volumeFAVariance );
    fprintf( fp, "fiber volume MD mean:        %.4f\n", volumeMDMean );
    fprintf( fp, "fiber volume MD variance:    %f\n", volumeMDVariance );
    fprintf( fp, "scatter X mean:              %.2f\n", scatterXMean );
    fprintf( fp, "scatter X variance:          %.4f\n", scatterXVariance );
    fprintf( fp, "scatter Y mean:              %.2f\n", scatterYMean );
    fprintf( fp, "scatter Y variance:          %.4f\n", scatterYVariance );
    fprintf( fp, "scatter Z mean:              %.2f\n", scatterZMean );
    fprintf( fp, "scatter Z variance:          %.4f\n", scatterZVariance );
    fprintf( fp, "fiber length mean (mm):      %.2f\n", fiberLengthMean );
    fprintf( fp, "fiber length variance:       %.4f\n", fiberLengthVariance );
    fprintf( fp, "fiber length sum (mm):       %.2f", fiberLengthSum );

    fclose( fp );
    return 0;
}

