/***********************************************************************
 * File:        matrix.c
 * Programmer:  Ning Kang @CS.UKY
 * Version of:  10/14/2003
 * ---------------------------------------------------------------------
 * Description:
 * - Serial version.
 * - The functions in this file are used to generate the coefficient
 * matrix in CSR format from the 3D Cartesian grid. The nodes in grid
 * are counted in natural order. The entries in the coefficient matrix
 * are specified according to the discretized anisotropic diffusion
 * equation.
 * - This file contains the following functions:
 *   void getCoeffMatrixStructure() 
 *   int getCoeffMatrixEntry()
 *   int nodeIndex ()
 *   void csrMatrix ()
 *   void csrEntry ()
 *   int csrIndex ()
 ***********************************************************************/
 

#include "dst.h"

static void csrMatrix( void *kernelData );
static void csrEntry( void *kernelData, void *dtiData, int nx, int ny, int nz);
static int csrIndex( int i, int j, int k, int node, void *kernelData );

/***********************************************************************
 * Interface for getting the coefficient matrix.
 ***********************************************************************/
void getCoeffMatrixStructure( void *kernelData )
{
    Diff_Kernel_ kernel;
    int neq;

    kernel = (Diff_Kernel_)kernelData;
    neq = kernel->neq;
                                // Initialize ia and ja arrays in CSR format
    csrMatrix( kernel );
                                // Re-allocate memory
    kernel->nnz = *(kernel->ia + neq) - 1;
    kernel->aa = (realtype *)realloc( kernel->aa, 
        kernel->nnz * sizeof( realtype ) );
    kernel->ja = (int *)realloc( kernel->ja, kernel->nnz * sizeof( int ) );
                                // Sort the matrix structure to be the column-
                                // increasing CSR format.

//printf("before matsort is called\nja0=%d ja1=%d ja2=%d\n", 
//        kernel->ja,kernel->ja+1,kernel->ja+2);

#ifdef ILU0
    #ifdef FORTRAN_UNDERSCORE
        matsort_( &neq, kernel->aa, kernel->ja, kernel->ia );
    #else
        matsort( &neq, kernel->aa, kernel->ja, kernel->ia );
    #endif
#endif

}

int getCoeffMatrixEntry( void *kernelData, void *dtiData )
{
    Diff_Kernel_ kernel;
    DTI_ dti;

    kernel = (Diff_Kernel_)kernelData;
    dti = (DTI_)dtiData;

    if( kernelToGridMapping( kernel, dti ) != 0 ) return 1;
                                // Fill in entries in the aa array
    csrEntry( kernel, dti, kernel->nx, kernel->ny, kernel->nz );
    return 0;
}

 
/***********************************************************************
 * This function returns the index of the node
 * in Cartesian grid. The nodes in the grid are
 * counted in natural order, starting from 1,
 * index changing fastest by x first, then by y, then by z.
 ***********************************************************************/
int nodeIndex( int i, int j, int k, int nx, int nxy )
{
    return i + (j-1)*nx + (k-1)*nxy;
}
               // node index, starting from 0
int nodeIndex_zero( int i, int j, int k, int nx, int nxy )
{
    return i + j*nx + k*nxy;
} 


/***********************************************************************
 * This function sets up ja[ia[k-1]-1]=k (puts the
 * diagonal entry as the first entry of each row in
 * ja. The remaining values in a given row of ja
 * is placed without being sorted.
 ***********************************************************************/
void csrMatrix( void *kernelData )
{
    Diff_Kernel_ kernel;
    int nx, ny, nz, nxy;
    int *ia, *ja;
    int i, j, k, counter, id;
    
    kernel = (Diff_Kernel_)kernelData;
    ia = kernel->ia;
    ja = kernel->ja;
    nx = kernel->nx;
    ny = kernel->ny;
    nz = kernel->nz;
    nxy = nx*ny;

    counter = 1;
    for( k = 1; k <= nz; ++ k )
        for( j = 1; j <= ny; ++ j )
            for( i = 1; i <= nx; ++ i )
	    {
                id = nodeIndex( i,j,k,nx,nxy );
		ia[id-1] = counter;
		ja[counter-1] = id;
		counter ++;
		
		if( i != 1 )
		{
		    id = nodeIndex( i-1,j,k,nx,nxy ); 
		    ja[counter-1] = id;
		    counter ++;
		}
		
		if( i != nx )
		{
		    id = nodeIndex( i+1,j,k,nx,nxy );
		    ja[counter-1] = id;
		    counter ++;
		}
		
		if( j != 1 )
		{
		    id = nodeIndex( i,j-1,k,nx,nxy );
		    ja[counter-1] = id;
		    counter ++;
		}
		
		if( j != ny )
		{
		    id = nodeIndex( i,j+1,k,nx,nxy );
		    ja[counter-1] = id;
		    counter ++;
		}
		
		if( k != 1 )
		{
		    id = nodeIndex( i,j,k-1,nx,nxy );
		    ja[counter-1] = id;
		    counter ++;
		}
		
		if( k != nz )
		{
		    id = nodeIndex( i,j,k+1,nx,nxy );
		    ja[counter-1] = id;
		    counter ++;
		}
		
		if( i != nx && j != ny )
		{
		    id = nodeIndex( i+1,j+1,k,nx,nxy );
		    ja[counter-1] = id;
		    counter ++;
		}
		
		if( i != nx && j != 1 )
		{
		    id = nodeIndex( i+1,j-1,k,nx,nxy );
		    ja[counter-1] = id;
		    counter ++;
		}
		
		if( i != 1 && j != ny )
		{
		    id = nodeIndex( i-1,j+1,k,nx,nxy );
		    ja[counter-1] = id;
		    counter ++;
		}
		
		if( i != 1 && j != 1 )
		{
		    id = nodeIndex( i-1,j-1,k,nx,nxy );
		    ja[counter-1] = id;
		    counter ++;
		}
		
		if( i != nx && k != nz )
		{
		    id = nodeIndex( i+1,j,k+1,nx,nxy );
		    ja[counter-1] = id;
		    counter ++;
		}
		
		if( i != nx && k != 1 )
		{ 
		    id = nodeIndex( i+1,j,k-1,nx,nxy );
		    ja[counter-1] = id;
		    counter ++;
		}
		
		if( i != 1 && k != nz )
		{
		    id = nodeIndex( i-1,j,k+1,nx,nxy );
		    ja[counter-1] = id;
		    counter ++;
		}
		
		if( i != 1 && k != 1 )
		{
		    id = nodeIndex( i-1,j,k-1,nx,nxy );
		    ja[counter-1] = id;
		    counter ++;
		}
		
		if( j != ny && k != nz )
		{
		    id = nodeIndex( i,j+1,k+1,nx,nxy );
		    ja[counter-1] = id;
		    counter ++;
		}
		
		if( j != ny && k != 1 )
		{
		    id = nodeIndex( i,j+1,k-1,nx,nxy );
		    ja[counter-1] = id;
		    counter ++;
		}
		
		if( j != 1 && k != nz )
		{
		    id = nodeIndex( i,j-1,k+1,nx,nxy );
		    ja[counter-1] = id;
		    counter ++;
		}
		
		if( j != 1 && k != 1 )
		{
		    id = nodeIndex( i,j-1,k-1,nx,nxy );
		    ja[counter-1] = id;
		    counter ++;
		}
            }
		
    ia[nx*ny*nz] = counter;
}


/***********************************************************************
 * Fill in the entries in the coefficient matrix.
 ***********************************************************************/
void csrEntry( void *kernelData, void *dtiData, int nx, int ny, int nz )
{
    Diff_Kernel_ kernel;
    realtype *aa, *sf, dx, dy, dz;
    DTI_ dti;
    realtype **dtxyz, dtmax;
    int i, j, k, index, id, nxy;
    int neq, *ia, *ja;
    realtype kxx[nx][ny][nz], kyy[nx][ny][nz], kzz[nx][ny][nz], 
             kxy[nx][ny][nz], kxz[nx][ny][nz], kyz[nx][ny][nz];
    double dxx, dxy, dxz, dyy, dyz, dzz;	     
    
    kernel = (Diff_Kernel_)kernelData;
    neq = kernel->neq;
    aa = kernel->aa;
    ia = kernel->ia;
    ja = kernel->ja;
    sf = kernel->sf;
    dx = kernel->dx;
    dy = kernel->dy;
    dz = kernel->dz;
    dti = (DTI_)dtiData;
    dtxyz = dti->dtxyz;
    nxy = nx*ny;
    dxx = dx*dx;
    dxy = dx*dy;
    dxz = dx*dz;
    dyy = dy*dy;
    dyz = dy*dz;
    dzz = dz*dz;
                                // Get the diffusion tensor components
    for( i = 0; i < nx; ++ i )
    for( j = 0; j < ny; ++ j )
    for( k = 0; k < nz; ++ k )
    {
                                // node index in diffusion kernel  
        id = nodeIndex( i+1,j+1,k+1,nx,nxy ) - 1;
	                        // node index in the grid
        id = kernel->voxel[id]->gridID;				
        if( id == OUT_OF_GRID )
	{
	    kxx[i][j][k] = 0;    
            kyy[i][j][k] = 0;
            kzz[i][j][k] = 0;
            kxy[i][j][k] = 0;
            kxz[i][j][k] = 0;
            kyz[i][j][k] = 0;
	}    
	else
	{
	    kxx[i][j][k] = DTXYZ(id, 0);    
            kyy[i][j][k] = DTXYZ(id, 1);
            kzz[i][j][k] = DTXYZ(id, 2);
            kxy[i][j][k] = DTXYZ(id, 3);
            kxz[i][j][k] = DTXYZ(id, 4);
            kyz[i][j][k] = DTXYZ(id, 5);
	}    
    }
                                // Release memory for tensor components 
				// since it is not used any more
//    free( dtxyz );
                                // Fill out the array for coefficient matrix
    memset( aa, 0, kernel->nnz * sizeof( realtype ) );
    memset( sf, 0, neq * sizeof( realtype ) ); 
    for( k = 1; k <= nz; ++ k )
    for( j = 1; j <= ny; ++ j )
    for( i = 1; i <= nx; ++ i )
    {
        id = nodeIndex( i,j,k,nx,nxy );

                                // find the scale factor for the row
/*        if( kxx[i-1][j-1][k-1] == 0.0 && kyy[i-1][j-1][k-1] == 0.0
            && kzz[i-1][j-1][k-1] == 0.0 )
*/
        dtmax = MAX( MAX( ABS(kxx[i-1][j-1][k-1]), 
            ABS(kyy[i-1][j-1][k-1]) ), ABS(kzz[i-1][j-1][k-1]) );
        if( dtmax <= DT_THRESHOLD )
            sf[id-1] = SF_CONST;
        else
            sf[id-1] = 1.0 / dtmax;		

                                // diagonal entry
        index = csrIndex( i,j,k,id,kernel );		
        aa[index] = -2 * ( kxx[i-1][j-1][k-1]/dxx 
	    + kyy[i-1][j-1][k-1]/dyy + kzz[i-1][j-1][k-1]/dzz );

                                // internal nodes
	if( i != 1 && i != nx && j != 1 && j != ny && k != 1 && k != nz )
	{
	    index = csrIndex( i-1,j,k,id,kernel );
	    aa[index] = kxx[i-1][j-1][k-1]/dxx 
	        - (kxx[i][j-1][k-1] - kxx[i-2][j-1][k-1])/(4*dxx)
		- (kxy[i-1][j][k-1] - kxy[i-1][j-2][k-1])/(4*dxy)
		- (kxz[i-1][j-1][k] - kxz[i-1][j-1][k-2])/(4*dxz);

	    index = csrIndex( i+1,j,k,id,kernel );
	    aa[index] = kxx[i-1][j-1][k-1]/dxx
	        + (kxx[i][j-1][k-1] - kxx[i-2][j-1][k-1])/(4*dxx)
		+ (kxy[i-1][j][k-1] - kxy[i-1][j-2][k-1])/(4*dxy)
		+ (kxz[i-1][j-1][k] - kxz[i-1][j-1][k-2])/(4*dxz);

	    index = csrIndex( i,j-1,k,id,kernel );
	    aa[index] = kyy[i-1][j-1][k-1]/dyy 
	        - (kxy[i][j-1][k-1] - kxy[i-2][j-1][k-1])/(4*dxy)
		- (kyy[i-1][j][k-1] - kyy[i-1][j-2][k-1])/(4*dyy)
		- (kyz[i-1][j-1][k] - kyz[i-1][j-1][k-2])/(4*dyz);

	    index = csrIndex( i,j+1,k,id,kernel );
	    aa[index] = kyy[i-1][j-1][k-1]/dyy 
	        + (kxy[i][j-1][k-1] - kxy[i-2][j-1][k-1])/(4*dxy)
		+ (kyy[i-1][j][k-1] - kyy[i-1][j-2][k-1])/(4*dyy)
		+ (kyz[i-1][j-1][k] - kyz[i-1][j-1][k-2])/(4*dyz);

	    index = csrIndex( i,j,k-1,id, kernel );
	    aa[index] = kzz[i-1][j-1][k-1]/dzz 
	        - (kxz[i][j-1][k-1] - kxz[i-2][j-1][k-1])/(4*dxz)
		- (kyz[i-1][j][k-1] - kyz[i-1][j-2][k-1])/(4*dyz)
		- (kzz[i-1][j-1][k] - kzz[i-1][j-1][k-2])/(4*dzz);

	    index = csrIndex( i,j,k+1,id,kernel );
	    aa[index] = kzz[i-1][j-1][k-1]/dzz 
	        + (kxz[i][j-1][k-1] - kxz[i-2][j-1][k-1])/(4*dxz)
		+ (kyz[i-1][j][k-1] - kyz[i-1][j-2][k-1])/(4*dyz)
		+ (kzz[i-1][j-1][k] - kzz[i-1][j-1][k-2])/(4*dzz);

	    index = csrIndex( i+1,j+1,k,id,kernel );
	    aa[index] = kxy[i-1][j-1][k-1]/(2*dxy);

	    index = csrIndex( i+1,j-1,k,id,kernel );
	    aa[index] = -kxy[i-1][j-1][k-1]/(2*dxy);
    
	    index = csrIndex( i-1,j+1,k,id,kernel );
	    aa[index] = -kxy[i-1][j-1][k-1]/(2*dxy);
	
	    index = csrIndex( i-1,j-1,k,id,kernel );
	    aa[index] = kxy[i-1][j-1][k-1]/(2*dxy);
   
	    index = csrIndex( i+1,j,k+1,id,kernel );
	    aa[index] = kxz[i-1][j-1][k-1]/(2*dxz);
    
	    index = csrIndex( i+1,j,k-1,id,kernel );
	    aa[index] = -kxz[i-1][j-1][k-1]/(2*dxz);

	    index = csrIndex( i-1,j,k+1,id,kernel );
	    aa[index] = -kxz[i-1][j-1][k-1]/(2*dxz);

	    index = csrIndex( i-1,j,k-1,id,kernel );
	    aa[index] = kxz[i-1][j-1][k-1]/(2*dxz);
 
	    index = csrIndex( i,j+1,k+1,id,kernel );
	    aa[index] = kyz[i-1][j-1][k-1]/(2*dyz);
  
	    index = csrIndex( i,j+1,k-1,id,kernel );
	    aa[index] = -kyz[i-1][j-1][k-1]/(2*dyz);
   
	    index = csrIndex( i,j-1,k+1,id,kernel );
	    aa[index] = -kyz[i-1][j-1][k-1]/(2*dyz);
    
	    index = csrIndex( i,j-1,k-1,id,kernel );
	    aa[index] = kyz[i-1][j-1][k-1]/(2*dyz);

        }
	
	if( i == 1 && j != 1 && j != ny && k != 1 && k != nz )
	{
	    index = csrIndex( i+1,j,k,id,kernel );
	    aa[index] = 2*kxx[i-1][j-1][k-1]/dxx;
	    
	    index = csrIndex( i,j-1,k,id,kernel );
	    aa[index] = kyy[i-1][j-1][k-1]/dyy + kxy[i-1][j-1][k-1]/dxy
	        - (4*kxy[i][j-1][k-1] - 3*kxy[i-1][j-1][k-1] -
		kxy[i+1][j-1][k-1])/(4*dxy)
		- (kyy[i-1][j][k-1] - kyy[i-1][j-2][k-1])/(4*dyy)
		- (kyz[i-1][j-1][k] - kyz[i-1][j-1][k-2])/(4*dyz);
		
	    index = csrIndex( i,j+1,k,id,kernel );
	    aa[index] = kyy[i-1][j-1][k-1]/dyy - kxy[i-1][j-1][k-1]/dxy
	        + (4*kxy[i][j-1][k-1] - 3*kxy[i-1][j-1][k-1] -
		kxy[i+1][j-1][k-1])/(4*dxy)
		+ (kyy[i-1][j][k-1] - kyy[i-1][j-2][k-1])/(4*dyy)
		+ (kyz[i-1][j-1][k] - kyz[i-1][j-1][k-2])/(4*dyz);
		
	    index = csrIndex( i,j,k-1,id,kernel );
	    aa[index] = kzz[i-1][j-1][k-1]/dzz + kxz[i-1][j-1][k-1]/dxz
	        - (4*kxz[i][j-1][k-1] - 3*kxz[i-1][j-1][k-1] -
		kxz[i+1][j-1][k-1])/(4*dxz)
		- (kyz[i-1][j][k-1] - kyz[i-1][j-2][k-1])/(4*dyz)
		- (kzz[i-1][j-1][k] - kzz[i-1][j-1][k-2])/(4*dzz);
		 	
	    index = csrIndex( i,j,k+1,id,kernel );
	    aa[index] = kzz[i-1][j-1][k-1]/dzz - kxz[i-1][j-1][k-1]/dxz
	        +(4*kxz[i][j-1][k-1] - 3*kxz[i-1][j-1][k-1] -
		kxz[i+1][j-1][k-1])/(4*dxz)
		+ (kyz[i-1][j][k-1] - kyz[i-1][j-2][k-1])/(4*dyz)
		+ (kzz[i-1][j-1][k] - kzz[i-1][j-1][k-2])/(4*dzz);
	
	    index = csrIndex( i+1,j+1,k,id,kernel );
	    aa[index] = kxy[i-1][j-1][k-1]/dxy;
	    
	    index = csrIndex( i+1,j-1,k,id,kernel );
	    aa[index] = -kxy[i-1][j-1][k-1]/dxy;
	    
	    index = csrIndex( i+1,j,k+1,id,kernel );
	    aa[index] = kxz[i-1][j-1][k-1]/dxz;
	
	    index = csrIndex( i+1,j,k-1,id,kernel );
	    aa[index] = -kxz[i-1][j-1][k-1]/dxz;
	    
	    index = csrIndex( i,j+1,k+1,id,kernel );
	    aa[index] = kyz[i-1][j-1][k-1]/(2*dyz);
	    
	    index = csrIndex( i,j+1,k-1,id,kernel );
	    aa[index] = -kyz[i-1][j-1][k-1]/(2*dyz);
	    
	    index = csrIndex( i,j-1,k+1,id,kernel );
	    aa[index] = -kyz[i-1][j-1][k-1]/(2*dyz);
	    
	    index = csrIndex( i,j-1,k-1,id,kernel );
	    aa[index] = kyz[i-1][j-1][k-1]/(2*dyz);
	    
	}
	
	if( i == nx && j != 1 && j != ny && k != 1 && k != nz )
	{
	    index = csrIndex( i-1,j,k,id,kernel );
	    aa[index] = 2*kxx[i-1][j-1][k-1]/dxx;
	    
	    index = csrIndex( i,j-1,k,id,kernel );
	    aa[index] = kyy[i-1][j-1][k-1]/dyy - kxy[i-1][j-1][k-1]/dxy
	        - (3*kxy[i-1][j-1][k-1] - 4*kxy[i-2][j-1][k-1] +
		kxy[i-3][j-1][k-1])/(4*dxy)
		- (kyy[i-1][j][k-1] - kyy[i-1][j-2][k-1])/(4*dyy)
		- (kyz[i-1][j-1][k] - kyz[i-1][j-1][k-2])/(4*dyz);
		
	    index = csrIndex( i,j+1,k,id,kernel );
	    aa[index] = kyy[i-1][j-1][k-1]/dyy + kxy[i-1][j-1][k-1]/dxy
	        + (3*kxy[i-1][j-1][k-1] - 4*kxy[i-2][j-1][k-1] +
		kxy[i-3][j-1][k-1])/(4*dxy)
		+ (kyy[i-1][j][k-1] - kyy[i-1][j-2][k-1])/(4*dyy)
		+ (kyz[i-1][j-1][k] - kyz[i-1][j-1][k-2])/(4*dyz);
		
	    index = csrIndex( i,j,k-1,id,kernel );
	    aa[index] = kzz[i-1][j-1][k-1]/dzz - kxz[i-1][j-1][k-1]/dxz
	        - (3*kxz[i-1][j-1][k-1] - 4*kxz[i-2][j-1][k-1] +
		kxz[i-3][j-1][k-1])/(4*dxz)
		- (kyz[i-1][j][k-1] - kyz[i-1][j-2][k-1])/(4*dyz)
		- (kzz[i-1][j-1][k] - kzz[i-1][j-1][k-2])/(4*dzz);
		 	
	    index = csrIndex( i,j,k+1,id,kernel );
	    aa[index] = kzz[i-1][j-1][k-1]/dzz + kxz[i-1][j-1][k-1]/dxz
	        + (3*kxz[i-1][j-1][k-1] - 4*kxz[i-2][j-1][k-1] +
		kxz[i-3][j-1][k-1])/(4*dxz)
		+ (kyz[i-1][j][k-1] - kyz[i-1][j-2][k-1])/(4*dyz)
		+ (kzz[i-1][j-1][k] - kzz[i-1][j-1][k-2])/(4*dzz);
	
	    index = csrIndex( i-1,j+1,k,id,kernel );
	    aa[index] = -kxy[i-1][j-1][k-1]/dxy;
	    
	    index = csrIndex( i-1,j-1,k,id,kernel );
	    aa[index] = kxy[i-1][j-1][k-1]/dxy;
	    
	    index = csrIndex( i-1,j,k+1,id,kernel );
	    aa[index] = -kxz[i-1][j-1][k-1]/dxz;
	
	    index = csrIndex( i-1,j,k-1,id,kernel );
	    aa[index] = kxz[i-1][j-1][k-1]/dxz;
	    
	    index = csrIndex( i,j+1,k+1,id,kernel );
	    aa[index] = kyz[i-1][j-1][k-1]/(2*dyz);
	    
	    index = csrIndex( i,j+1,k-1,id,kernel );
	    aa[index] = -kyz[i-1][j-1][k-1]/(2*dyz);
	    
	    index = csrIndex( i,j-1,k+1,id,kernel );
	    aa[index] = -kyz[i-1][j-1][k-1]/(2*dyz);
	    
	    index = csrIndex( i,j-1,k-1,id,kernel );
	    aa[index] = kyz[i-1][j-1][k-1]/(2*dyz);
	}
	
	if( j == 1 && i != 1 && i != nx && k != 1 && k != nz )
	{
	    index = csrIndex( i,j+1,k,id,kernel );
	    aa[index] = 2*kyy[i-1][j-1][k-1]/dyy;
	    
	    index = csrIndex( i-1,j,k,id,kernel );
	    aa[index] = kxx[i-1][j-1][k-1]/dxx + kxy[i-1][j-1][k-1]/dxy
		- (kxx[i][j-1][k-1] - kxx[i-2][j-1][k-1])/(4*dxx)
	        - (4*kxy[i-1][j][k-1] - kxy[i-1][j+1][k-1] -
		3*kxy[i-1][j-1][k-1])/(4*dxy)
		- (kxz[i-1][j-1][k] - kxz[i-1][j-1][k-2])/(4*dxz);
		
	    index = csrIndex( i+1,j,k,id,kernel );
	    aa[index] = kxx[i-1][j-1][k-1]/dxx - kxy[i-1][j-1][k-1]/dxy
		+ (kxx[i][j-1][k-1] - kxx[i-2][j-1][k-1])/(4*dxx)
	        + (4*kxy[i-1][j][k-1] - kxy[i-1][j+1][k-1] -
		3*kxy[i-1][j-1][k-1])/(4*dxy)
		+ (kxz[i-1][j-1][k] - kxz[i-1][j-1][k-2])/(4*dxz);
		
	    index = csrIndex( i,j,k-1,id,kernel );
	    aa[index] = kzz[i-1][j-1][k-1]/dzz + kyz[i-1][j-1][k-1]/dyz
		- (kxz[i][j-1][k-1] - kxz[i-2][j-1][k-1])/(4*dxz)
	        - (4*kyz[i-1][j][k-1] - kyz[i-1][j+1][k-1] -
		3*kyz[i-1][j-1][k-1])/(4*dyz)
		- (kzz[i-1][j-1][k] - kzz[i-1][j-1][k-2])/(4*dzz);
		 	
	    index = csrIndex( i,j,k+1,id,kernel );
	    aa[index] = kzz[i-1][j-1][k-1]/dzz - kyz[i-1][j-1][k-1]/dyz
		+ (kxz[i][j-1][k-1] - kxz[i-2][j-1][k-1])/(4*dxz)
	        + (4*kyz[i-1][j][k-1] - kyz[i-1][j+1][k-1] -
		3*kyz[i-1][j-1][k-1])/(4*dyz)
		+ (kzz[i-1][j-1][k] - kzz[i-1][j-1][k-2])/(4*dzz);
	
	    index = csrIndex( i+1,j+1,k,id,kernel );
	    aa[index] = kxy[i-1][j-1][k-1]/dxy;
	    
	    index = csrIndex( i-1,j+1,k,id,kernel );
	    aa[index] = -kxy[i-1][j-1][k-1]/dxy;
	    
	    index = csrIndex( i+1,j,k+1,id,kernel );
	    aa[index] = kxz[i-1][j-1][k-1]/(2*dxz);
	
	    index = csrIndex( i+1,j,k-1,id,kernel );
	    aa[index] = -kxz[i-1][j-1][k-1]/(2*dxz);
	    
	    index = csrIndex( i-1,j,k+1,id,kernel );
	    aa[index] = -kxz[i-1][j-1][k-1]/(2*dxz);
	    
	    index = csrIndex( i-1,j,k-1,id,kernel );
	    aa[index] = kxz[i-1][j-1][k-1]/(2*dxz);
	    
	    index = csrIndex( i,j+1,k+1,id,kernel );
	    aa[index] = kyz[i-1][j-1][k-1]/dyz;
	    
	    index = csrIndex( i,j+1,k-1,id,kernel );
	    aa[index] = -kyz[i-1][j-1][k-1]/dyz;
	}

	if( j == ny && i != 1 && i != nx && k != 1 && k != nz )
	{
	    index = csrIndex( i,j-1,k,id,kernel );
	    aa[index] = 2*kyy[i-1][j-1][k-1]/dyy;
	    
	    index = csrIndex( i-1,j,k,id,kernel );
	    aa[index] = kxx[i-1][j-1][k-1]/dxx - kxy[i-1][j-1][k-1]/dxy
		- (kxx[i][j-1][k-1] - kxx[i-2][j-1][k-1])/(4*dxx)
	        - (3*kxy[i-1][j-1][k-1] - 4*kxy[i-1][j-2][k-1] +
		kxy[i-1][j-3][k-1])/(4*dxy)
		- (kxz[i-1][j-1][k] - kxz[i-1][j-1][k-2])/(4*dxz);
		
	    index = csrIndex( i+1,j,k,id,kernel );
	    aa[index] = kxx[i-1][j-1][k-1]/dxx + kxy[i-1][j-1][k-1]/dxy
		+ (kxx[i][j-1][k-1] - kxx[i-2][j-1][k-1])/(4*dxx)
	        + (3*kxy[i-1][j-1][k-1] - 4*kxy[i-1][j-2][k-1] +
		kxy[i-1][j-3][k-1])/(4*dxy)
		+ (kxz[i-1][j-1][k] - kxz[i-1][j-1][k-2])/(4*dxz);
		
	    index = csrIndex( i,j,k-1,id,kernel );
	    aa[index] = kzz[i-1][j-1][k-1]/dzz - kyz[i-1][j-1][k-1]/dyz
		- (kxz[i][j-1][k-1] - kxz[i-2][j-1][k-1])/(4*dxz)
	        - (3*kyz[i-1][j-1][k-1] - 4*kyz[i-1][j-2][k-1] +
		kyz[i-1][j-3][k-1])/(4*dyz)
		- (kzz[i-1][j-1][k] - kzz[i-1][j-1][k-2])/(4*dzz);
		 	
	    index = csrIndex( i,j,k+1,id,kernel );
	    aa[index] = kzz[i-1][j-1][k-1]/dzz + kyz[i-1][j-1][k-1]/dyz
		+ (kxz[i][j-1][k-1] - kxz[i-2][j-1][k-1])/(4*dxz)
	        + (3*kyz[i-1][j-1][k-1] - 4*kyz[i-1][j-2][k-1] +
		kyz[i-1][j-3][k-1])/(4*dyz)
		+ (kzz[i-1][j-1][k] - kzz[i-1][j-1][k-2])/(4*dzz);
	
	    index = csrIndex( i+1,j-1,k,id,kernel );
	    aa[index] = -kxy[i-1][j-1][k-1]/dxy;
	    
	    index = csrIndex( i-1,j-1,k,id,kernel );
	    aa[index] = kxy[i-1][j-1][k-1]/dxy;
	    
	    index = csrIndex( i+1,j,k+1,id,kernel );
	    aa[index] = kxz[i-1][j-1][k-1]/(2*dxz);
	
	    index = csrIndex( i+1,j,k-1,id,kernel );
	    aa[index] = -kxz[i-1][j-1][k-1]/(2*dxz);
	    
	    index = csrIndex( i-1,j,k+1,id,kernel );
	    aa[index] = -kxz[i-1][j-1][k-1]/(2*dxz);
	    
	    index = csrIndex( i-1,j,k-1,id,kernel );
	    aa[index] = kxz[i-1][j-1][k-1]/(2*dxz);
	    
	    index = csrIndex( i,j-1,k+1,id,kernel );
	    aa[index] = -kyz[i-1][j-1][k-1]/dyz;
	    
	    index = csrIndex( i,j-1,k-1,id,kernel );
	    aa[index] = kyz[i-1][j-1][k-1]/dyz;
	}

	if( k == 1 && i != 1 && i != nx && j != 1 && j != ny )
	{
	    index = csrIndex( i,j,k+1,id,kernel );
	    aa[index] = 2*kzz[i-1][j-1][k-1]/dzz;
	    
	    index = csrIndex( i-1,j,k,id,kernel );
	    aa[index] = kxx[i-1][j-1][k-1]/dxx + kxz[i-1][j-1][k-1]/dxz
		- (kxx[i][j-1][k-1] - kxx[i-2][j-1][k-1])/(4*dxx)
		- (kxy[i-1][j][k-1] - kxy[i-1][j-2][k-1])/(4*dxy)
	        - (4*kxz[i-1][j-1][k] - kxz[i-1][j-1][k+1] -
		3*kxz[i-1][j-1][k-1])/(4*dxz);
		
	    index = csrIndex( i+1,j,k,id,kernel );
	    aa[index] = kxx[i-1][j-1][k-1]/dxx - kxz[i-1][j-1][k-1]/dxz
		+ (kxx[i][j-1][k-1] - kxx[i-2][j-1][k-1])/(4*dxx)
		+ (kxy[i-1][j][k-1] - kxy[i-1][j-2][k-1])/(4*dxy)
	        + (4*kxz[i-1][j-1][k] - kxz[i-1][j-1][k+1] -
		3*kxz[i-1][j-1][k-1])/(4*dxz);
		
	    index = csrIndex( i,j-1,k,id,kernel );
	    aa[index] = kyy[i-1][j-1][k-1]/dyy + kyz[i-1][j-1][k-1]/dyz
		- (kxy[i][j-1][k-1] - kxy[i-2][j-1][k-1])/(4*dxy)
		- (kyy[i-1][j][k-1] - kyy[i-1][j-2][k-1])/(4*dyy)
	        - (4*kyz[i-1][j-1][k] - kyz[i-1][j-1][k+1] -
		3*kyz[i-1][j-1][k-1])/(4*dyz);
		 	
	    index = csrIndex( i,j,k+1,id,kernel );
	    aa[index] = kyy[i-1][j-1][k-1]/dyy - kyz[i-1][j-1][k-1]/dyz
		+ (kxy[i][j-1][k-1] - kxy[i-2][j-1][k-1])/(4*dxy)
		+ (kyy[i-1][j][k-1] - kyy[i-1][j-2][k-1])/(4*dyy)
	        + (4*kyz[i-1][j-1][k] - kyz[i-1][j-1][k+1] -
		3*kyz[i-1][j-1][k-1])/(4*dyz);
	
	    index = csrIndex( i+1,j+1,k,id,kernel );
	    aa[index] = kxy[i-1][j-1][k-1]/(2*dxy);
	    
	    index = csrIndex( i+1,j-1,k,id,kernel );
	    aa[index] = -kxy[i-1][j-1][k-1]/(2*dxy);
	    
	    index = csrIndex( i-1,j+1,k,id,kernel );
	    aa[index] = -kxy[i-1][j-1][k-1]/(2*dxy);
	
	    index = csrIndex( i-1,j-1,k,id,kernel );
	    aa[index] = kxy[i-1][j-1][k-1]/(2*dxy);
	    
	    index = csrIndex( i+1,j,k+1,id,kernel );
	    aa[index] = kxz[i-1][j-1][k-1]/dxz;
	    
	    index = csrIndex( i-1,j,k+1,id,kernel );
	    aa[index] = -kxz[i-1][j-1][k-1]/dxz;
	    
	    index = csrIndex( i,j+1,k+1,id,kernel );
	    aa[index] = kyz[i-1][j-1][k-1]/dyz;
	    
	    index = csrIndex( i,j-1,k+1,id,kernel );
	    aa[index] = -kyz[i-1][j-1][k-1]/dyz;
	}

	if( k == nz && i != 1 && i != nx && j != 1 && j != ny )
	{
	    index = csrIndex( i,j,k-1,id,kernel );
	    aa[index] = 2*kzz[i-1][j-1][k-1]/dzz;
	    
	    index = csrIndex( i-1,j,k,id,kernel );
	    aa[index] = kxx[i-1][j-1][k-1]/dxx - kxz[i-1][j-1][k-1]/dxz
		- (kxx[i][j-1][k-1] - kxx[i-2][j-1][k-1])/(4*dxx)
		- (kxy[i-1][j][k-1] - kxy[i-1][j-2][k-1])/(4*dxy)
	        - (3*kxz[i-1][j-1][k-1] - 4*kxz[i-1][j-1][k-2] +
		kxz[i-1][j-1][k-3])/(4*dxz);
		
	    index = csrIndex( i+1,j,k,id,kernel );
	    aa[index] = kxx[i-1][j-1][k-1]/dxx + kxz[i-1][j-1][k-1]/dxz
		+ (kxx[i][j-1][k-1] - kxx[i-2][j-1][k-1])/(4*dxx)
		+ (kxy[i-1][j][k-1] - kxy[i-1][j-2][k-1])/(4*dxy)
	        + (3*kxz[i-1][j-1][k-1] - 4*kxz[i-1][j-1][k-2] +
		kxz[i-1][j-1][k-3])/(4*dxz);
		
	    index = csrIndex( i,j-1,k,id,kernel );
	    aa[index] = kyy[i-1][j-1][k-1]/dyy - kyz[i-1][j-1][k-1]/dyz
		- (kxy[i][j-1][k-1] - kxy[i-2][j-1][k-1])/(4*dxy)
		- (kyy[i-1][j][k-1] - kyy[i-1][j-2][k-1])/(4*dyy)
	        - (3*kyz[i-1][j-1][k-1] - 4*kyz[i-1][j-1][k-2] +
		kyz[i-1][j-1][k-3])/(4*dyz);
		 	
	    index = csrIndex( i,j,k+1,id,kernel );
	    aa[index] = kyy[i-1][j-1][k-1]/dyy + kyz[i-1][j-1][k-1]/dyz
		+ (kxy[i][j-1][k-1] - kxy[i-2][j-1][k-1])/(4*dxy)
		+ (kyy[i-1][j][k-1] - kyy[i-1][j-2][k-1])/(4*dyy)
	        + (3*kyz[i-1][j-1][k-1] - 4*kyz[i-1][j-1][k-2] +
		kyz[i-1][j-1][k-3])/(4*dyz);
	
	    index = csrIndex( i+1,j+1,k,id,kernel );
	    aa[index] = kxy[i-1][j-1][k-1]/(2*dxy);
	    
	    index = csrIndex( i+1,j-1,k,id,kernel );
	    aa[index] = -kxy[i-1][j-1][k-1]/(2*dxy);
	    
	    index = csrIndex( i-1,j+1,k,id,kernel );
	    aa[index] = -kxy[i-1][j-1][k-1]/(2*dxy);
	
	    index = csrIndex( i-1,j-1,k,id,kernel );
	    aa[index] = kxy[i-1][j-1][k-1]/(2*dxy);
	    
	    index = csrIndex( i+1,j,k-1,id,kernel );
	    aa[index] = -kxz[i-1][j-1][k-1]/dxz;
	    
	    index = csrIndex( i-1,j,k-1,id,kernel );
	    aa[index] = kxz[i-1][j-1][k-1]/dxz;
	    
	    index = csrIndex( i,j+1,k-1,id,kernel );
	    aa[index] = -kyz[i-1][j-1][k-1]/dyz;
	    
	    index = csrIndex( i,j-1,k-1,id,kernel );
	    aa[index] = kyz[i-1][j-1][k-1]/dyz;
	}

        if( i == 1 && j == 1 && k == 1 )
	{
	    index = csrIndex( i,j,k,id,kernel );
	    aa[index] += 2*( kxy[i-1][j-1][k-1]/dxy
	        + kxz[i-1][j-1][k-1]/dxz + kyz[i-1][j-1][k-1]/dyz ); 
	    
	    index = csrIndex( i+1,j,k,id,kernel );
	    aa[index] = 2*( kxx[i-1][j-1][k-1]/dxx
	        - kxy[i-1][j-1][k-1]/dxy - kxz[i-1][j-1][k-1]/dxz );
		
	    index = csrIndex( i,j+1,k,id,kernel );
	    aa[index] = 2*( kyy[i-1][j-1][k-1]/dyy
	        - kxy[i-1][j-1][k-1]/dxy - kyz[i-1][j-1][k-1]/dyz );
		
	    index = csrIndex( i,j,k+1,id,kernel );
	    aa[index] = 2*( kzz[i-1][j-1][k-1]/dzz
	        - kxz[i-1][j-1][k-1]/dxz - kyz[i-1][j-1][k-1]/dyz );
		
	    index = csrIndex( i+1,j+1,k,id,kernel );
	    aa[index] = 2*kxy[i-1][j-1][k-1]/dxy;
	
	    index = csrIndex( i+1,j,k+1,id,kernel );
	    aa[index] = 2*kxz[i-1][j-1][k-1]/dxz;
	    
	    index = csrIndex( i,j+1,k+1,id,kernel );
	    aa[index] = 2*kyz[i-1][j-1][k-1]/dyz;
	}

        if( i == 1 && j == ny && k == 1 )
	{
	    index = csrIndex( i,j,k,id,kernel );
	    aa[index] += 2*( -kxy[i-1][j-1][k-1]/dxy
	        + kxz[i-1][j-1][k-1]/dxz - kyz[i-1][j-1][k-1]/dyz ); 
	    
	    index = csrIndex( i+1,j,k,id,kernel );
	    aa[index] = 2*( kxx[i-1][j-1][k-1]/dxx
	        + kxy[i-1][j-1][k-1]/dxy - kxz[i-1][j-1][k-1]/dxz );
		
	    index = csrIndex( i,j-1,k,id,kernel );
	    aa[index] = 2*( kyy[i-1][j-1][k-1]/dyy
	        + kxy[i-1][j-1][k-1]/dxy + kyz[i-1][j-1][k-1]/dyz );
		
	    index = csrIndex( i,j,k+1,id,kernel );
	    aa[index] = 2*( kzz[i-1][j-1][k-1]/dzz
	        - kxz[i-1][j-1][k-1]/dxz + kyz[i-1][j-1][k-1]/dyz );
		
	    index = csrIndex( i+1,j-1,k,id,kernel );
	    aa[index] = -2*kxy[i-1][j-1][k-1]/dxy;
	
	    index = csrIndex( i+1,j,k+1,id,kernel );
	    aa[index] = 2*kxz[i-1][j-1][k-1]/dxz;
	    
	    index = csrIndex( i,j-1,k+1,id,kernel );
	    aa[index] = -2*kyz[i-1][j-1][k-1]/dyz;
	}

        if( i == nx && j == 1 && k == 1 )
	{
	    index = csrIndex( i,j,k,id,kernel );
	    aa[index] += 2*( -kxy[i-1][j-1][k-1]/dxy
	        - kxz[i-1][j-1][k-1]/dxz + kyz[i-1][j-1][k-1]/dyz ); 
	    
	    index = csrIndex( i-1,j,k,id,kernel );
	    aa[index] = 2*( kxx[i-1][j-1][k-1]/dxx
	        + kxy[i-1][j-1][k-1]/dxy + kxz[i-1][j-1][k-1]/dxz );
		
	    index = csrIndex( i,j+1,k,id,kernel );
	    aa[index] = 2*( kyy[i-1][j-1][k-1]/dyy
	        + kxy[i-1][j-1][k-1]/dxy - kyz[i-1][j-1][k-1]/dyz );
		
	    index = csrIndex( i,j,k+1,id,kernel );
	    aa[index] = 2*( kzz[i-1][j-1][k-1]/dzz
	        + kxz[i-1][j-1][k-1]/dxz - kyz[i-1][j-1][k-1]/dyz );
		
	    index = csrIndex( i-1,j+1,k,id,kernel );
	    aa[index] = -2*kxy[i-1][j-1][k-1]/dxy;
	
	    index = csrIndex( i-1,j,k+1,id,kernel );
	    aa[index] = -2*kxz[i-1][j-1][k-1]/dxz;
	    
	    index = csrIndex( i,j+1,k+1,id,kernel );
	    aa[index] = 2*kyz[i-1][j-1][k-1]/dyz;
	}


        if( i == nx && j == ny && k == 1 )
	{
	    index = csrIndex( i,j,k,id,kernel );
	    aa[index] += 2*( kxy[i-1][j-1][k-1]/dxy
	        - kxz[i-1][j-1][k-1]/dxz - kyz[i-1][j-1][k-1]/dyz ); 
	    
	    index = csrIndex( i-1,j,k,id,kernel );
	    aa[index] = 2*( kxx[i-1][j-1][k-1]/dxx
	        - kxy[i-1][j-1][k-1]/dxy + kxz[i-1][j-1][k-1]/dxz );
		
	    index = csrIndex( i,j-1,k,id,kernel );
	    aa[index] = 2*( kyy[i-1][j-1][k-1]/dyy
	        - kxy[i-1][j-1][k-1]/dxy + kyz[i-1][j-1][k-1]/dyz );
		
	    index = csrIndex( i,j,k+1,id,kernel );
	    aa[index] = 2*( kzz[i-1][j-1][k-1]/dzz
	        + kxz[i-1][j-1][k-1]/dxz + kyz[i-1][j-1][k-1]/dyz );
		
	    index = csrIndex( i-1,j-1,k,id,kernel );
	    aa[index] = 2*kxy[i-1][j-1][k-1]/dxy;
	
	    index = csrIndex( i-1,j,k+1,id,kernel );
	    aa[index] = -2*kxz[i-1][j-1][k-1]/dxz;
	    
	    index = csrIndex( i,j-1,k+1,id,kernel );
	    aa[index] = -2*kyz[i-1][j-1][k-1]/dyz;
	}

        if( i == 1 && j == 1 && k == nz )
	{
	    index = csrIndex( i,j,k,id,kernel );
	    aa[index] += 2*( kxy[i-1][j-1][k-1]/dxy
	        - kxz[i-1][j-1][k-1]/dxz - kyz[i-1][j-1][k-1]/dyz ); 
	    
	    index = csrIndex( i+1,j,k,id,kernel );
	    aa[index] = 2*( kxx[i-1][j-1][k-1]/dxx
	        - kxy[i-1][j-1][k-1]/dxy + kxz[i-1][j-1][k-1]/dxz );
		
	    index = csrIndex( i,j+1,k,id,kernel );
	    aa[index] = 2*( kyy[i-1][j-1][k-1]/dyy
	        - kxy[i-1][j-1][k-1]/dxy + kyz[i-1][j-1][k-1]/dyz );
		
	    index = csrIndex( i,j,k-1,id,kernel );
	    aa[index] = 2*( kzz[i-1][j-1][k-1]/dzz
	        + kxz[i-1][j-1][k-1]/dxz + kyz[i-1][j-1][k-1]/dyz );
		
	    index = csrIndex( i+1,j+1,k,id,kernel );
	    aa[index] = 2*kxy[i-1][j-1][k-1]/dxy;
	
	    index = csrIndex( i+1,j,k-1,id,kernel );
	    aa[index] = -2*kxz[i-1][j-1][k-1]/dxz;
	    
	    index = csrIndex( i,j+1,k-1,id,kernel );
	    aa[index] = -2*kyz[i-1][j-1][k-1]/dyz;
	}

        if( i == 1 && j == ny && k == nz )
	{
	    index = csrIndex( i,j,k,id,kernel );
	    aa[index] += 2*( -kxy[i-1][j-1][k-1]/dxy
	        - kxz[i-1][j-1][k-1]/dxz + kyz[i-1][j-1][k-1]/dyz ); 
	    
	    index = csrIndex( i+1,j,k,id,kernel );
	    aa[index] = 2*( kxx[i-1][j-1][k-1]/dxx
	        + kxy[i-1][j-1][k-1]/dxy + kxz[i-1][j-1][k-1]/dxz );
		
	    index = csrIndex( i,j-1,k,id,kernel );
	    aa[index] = 2*( kyy[i-1][j-1][k-1]/dyy
	        + kxy[i-1][j-1][k-1]/dxy - kyz[i-1][j-1][k-1]/dyz );
		
	    index = csrIndex( i,j,k-1,id,kernel );
	    aa[index] = 2*( kzz[i-1][j-1][k-1]/dzz
	        + kxz[i-1][j-1][k-1]/dxz - kyz[i-1][j-1][k-1]/dyz );
		
	    index = csrIndex( i+1,j-1,k,id,kernel );
	    aa[index] = -2*kxy[i-1][j-1][k-1]/dxy;
	
	    index = csrIndex( i+1,j,k-1,id,kernel );
	    aa[index] = -2*kxz[i-1][j-1][k-1]/dxz;
	    
	    index = csrIndex( i,j-1,k-1,id,kernel );
	    aa[index] = 2*kyz[i-1][j-1][k-1]/dyz;
	}

        if( i == nx && j == 1 && k == nz )
	{
	    index = csrIndex( i,j,k,id,kernel );
	    aa[index] += 2*( -kxy[i-1][j-1][k-1]/dxy
	        + kxz[i-1][j-1][k-1]/dxz - kyz[i-1][j-1][k-1]/dyz ); 
	    
	    index = csrIndex( i-1,j,k,id,kernel );
	    aa[index] = 2*( kxx[i-1][j-1][k-1]/dxx
	        + kxy[i-1][j-1][k-1]/dxy - kxz[i-1][j-1][k-1]/dxz );
		
	    index = csrIndex( i,j+1,k,id,kernel );
	    aa[index] = 2*( kyy[i-1][j-1][k-1]/dyy
	        + kxy[i-1][j-1][k-1]/dxy + kyz[i-1][j-1][k-1]/dyz );
		
	    index = csrIndex( i,j,k-1,id,kernel );
	    aa[index] = 2*( kzz[i-1][j-1][k-1]/dzz
	        - kxz[i-1][j-1][k-1]/dxz + kyz[i-1][j-1][k-1]/dyz );
		
	    index = csrIndex( i-1,j+1,k,id,kernel );
	    aa[index] = -2*kxy[i-1][j-1][k-1]/dxy;
	
	    index = csrIndex( i-1,j,k-1,id,kernel );
	    aa[index] = 2*kxz[i-1][j-1][k-1]/dxz;
	    
	    index = csrIndex( i,j+1,k-1,id,kernel );
	    aa[index] = -2*kyz[i-1][j-1][k-1]/dyz;
	}


        if( i == nx && j == ny && k == nz )
	{
	    index = csrIndex( i,j,k,id,kernel );
	    aa[index] += 2*( kxy[i-1][j-1][k-1]/dxy
	        + kxz[i-1][j-1][k-1]/dxz + kyz[i-1][j-1][k-1]/dyz ); 
	    
	    index = csrIndex( i-1,j,k,id,kernel );
	    aa[index] = 2*( kxx[i-1][j-1][k-1]/dxx
	        - kxy[i-1][j-1][k-1]/dxy - kxz[i-1][j-1][k-1]/dxz );
		
	    index = csrIndex( i,j-1,k,id,kernel );
	    aa[index] = 2*( kyy[i-1][j-1][k-1]/dyy
	        - kxy[i-1][j-1][k-1]/dxy - kyz[i-1][j-1][k-1]/dyz );
		
	    index = csrIndex( i,j,k-1,id,kernel );
	    aa[index] = 2*( kzz[i-1][j-1][k-1]/dzz
	        - kxz[i-1][j-1][k-1]/dxz - kyz[i-1][j-1][k-1]/dyz );
		
	    index = csrIndex( i-1,j-1,k,id,kernel );
	    aa[index] = 2*kxy[i-1][j-1][k-1]/dxy;
	
	    index = csrIndex( i-1,j,k-1,id,kernel );
	    aa[index] = 2*kxz[i-1][j-1][k-1]/dxz;
	    
	    index = csrIndex( i,j-1,k-1,id,kernel );
	    aa[index] = 2*kyz[i-1][j-1][k-1]/dyz;
	}
    }
                                // Apply the scale factor to each row.
    for( i = 0; i < neq; ++ i )
        for( k = ia[i]-1; k < ia[i+1]-1; ++ k )
	    aa[k] *= sf[i]; 
                                // Sort the matrix structure to be the column-
                                // increasing CSR format.
/*
#ifdef ILU0
    #ifdef FORTRAN_UNDERSCORE
        matsort_( &neq, aa, ja, ia );
    #else
        matsort( &neq, aa, ja, ia );
    #endif
#endif
*/
} 


/***********************************************************************
 * Find the index of the entry in the aa array
 * which corresponds to that in the ja array.
 ***********************************************************************/
int csrIndex( int i, int j, int k, int id, void *kernelData )
{
    int index, nx, ny, nxy;
    Diff_Kernel_ kernel;
    int *ia, *ja;        
    
    kernel = (Diff_Kernel_)kernelData;
    ia = kernel->ia;
    ja = kernel->ja;
    nx = kernel->nx;
    ny = kernel->ny;
    nxy = nx*ny;
    
    for( index = ia[id-1]-1; index < ia[id]-1; ++ index )
        if( ja[index] == nodeIndex( i,j,k,nx,nxy ) )
	    break;

    return index;
}



















