/* alloc.c */

#include <stdlib.h>
#include <stdio.h>
#include "alloc.h" 

#define NR_END 1
#define FREE_ARG char*

/**<zv_alloc>****************************************************

     Title:  Double Vector ALLOCation

   Purpose:  Allocates a double vector v of length n

      Call:  double *zv_alloc(n)
                  int n;

    Return: *v[n]

*/
complex *zv_alloc( int n )
{
  complex *v;

  v = (complex *) calloc(n,sizeof(complex));
  if (!v) perror("Allocation failure in zv_alloc()");

  return(v);
}

/**<fv_alloc>****************************************************

     Title:  Float Vector ALLOCation

   Purpose:  Allocates a float vector v of length n

      Call:  float *fv_alloc(int n)

    Return: *v[n]

*/
float *fv_alloc( int n )
{
  float *v;

  v = (float *) calloc(n,sizeof(float));
  if (!v) perror("Allocation failure in fv_alloc()");

  return(v);
}

/**<dv_alloc>****************************************************

     Title:  Double Vector ALLOCation

   Purpose:  Allocates a double vector v of length n

      Call:  double *dv_alloc(n)
                  int n;

    Return: *v[n]

*/
double *dv_alloc( int n )
{
  double *v;

  v = (double *) calloc(n,sizeof(double));
  if (!v) perror("Allocation failure in dv_alloc()");

  return(v);
}

/**<fm_alloc.c>************************************************************

     Title: Float Matrix ALLOCation

   Purpose: Allocate nrows x ncolumns float matrix
            in a contiguous block of memory.

      Call: float **fm_alloc(int nrows, int ncolumns) 

*/
float **fm_alloc(int nr, int nc) 
{
  int i; 
  float **z ;

  z = (float **) calloc(nr,sizeof(float *));
  for(i=0; i<nr; i++) z[i] = (float *) calloc(nc,sizeof(float)); 

  return z;
}

/**<dm_alloc.c>************************************************************

     Title: Double Matrix ALLOCation

   Purpose: Allocate nrows x ncolumns double matrix
            in a contiguous block of memory.

      Call: double **dm_alloc(int nrows, int ncolumns) 

*/
double **dm_alloc(int nr, int nc) 
{
  int i; 
  double **z ;

  z = (double **) calloc(nr,sizeof(double *));
  for(i=0; i<nr; i++) z[i] = (double *) calloc(nc,sizeof(double)); 

  return z;
}

/**<im_alloc.c>************************************************************

     Title: Int Matrix ALLOCation

   Purpose: Allocate nrows x ncolumns int matrix
            in a contiguous block of memory.

      Call: int **im_alloc(int nrows, int ncolumns) 

*/
int **im_alloc(int nr, int nc) 
{
  int i; 
  int **z ;

  z = (int **) calloc(nr,sizeof(int *));
  for(i=0; i<nr; i++) z[i] = (int *) calloc(nc,sizeof(int)); 

  return z;
}

/**<sm_alloc.c>************************************************************

     Title: Short Matrix ALLOCation

   Purpose: Allocate nrows x ncolumns short matrix
            in a contiguous block of memory.

      Call: short **sm_alloc(int nrows, int ncolumns) 

*/
short **sm_alloc(int nr, int nc) 
{
  int i; 
  short **z ;

  z = (short **) calloc(nr,sizeof(short *));
  for(i=0; i<nr; i++) z[i] = (short *) calloc(nc,sizeof(short)); 

  return z;
}

/**<fm_free.c>********************************************************

       Title:  Float Matrix FREE

     Purpose:  Deallocate float matrix d[m][n]

*/
void fm_free( float **fm, int nrow, int ncol )
{
  static int ir ;

  if( fm != NULL ){
    for( ir=0 ; ir < nrow ; ir++ ) if( fm[ir] != NULL ) free(fm[ir]) ;
    free(fm) ;
  }

}

/**<dm_free.c>********************************************************

       Title:  Double Matrix FREE

     Purpose:  Deallocate double matrix d[m][n]

*/
void dm_free( double **dm, int nrow, int ncol )
{
  static int ir ;

  if( dm != NULL ){
    for( ir=0 ; ir < nrow ; ir++ ) if( dm[ir] != NULL ) free(dm[ir]) ;
    free(dm) ;
  }

}

/**<im_free.c>********************************************************

       Title:  Int Matrix FREE

     Purpose:  Deallocate int matrix d[m][n]

*/
void im_free( int **im, int nrow, int ncol )
{
  static int ir ;

  if( im != NULL ){
    for( ir=0 ; ir < nrow ; ir++ ) if( im[ir] != NULL ) free(im[ir]) ;
    free(im) ;
  }

}

/**<sm_free.c>********************************************************

       Title:  Short Matrix FREE

     Purpose:  Deallocate short matrix d[m][n]

*/
void sm_free( short **sm, int nrow, int ncol )
{
  static int ir ;

  if( sm != NULL ){
    for( ir=0 ; ir < nrow ; ir++ ) if( sm[ir] != NULL ) free(sm[ir]) ;
    free(sm) ;
  }

}

/**<zm_alloc.c>*************************************************************

       Title:  float Complex Matrix space ALLOCation

     Purpose:  Allocates a float complex matrix z[nr][nc]
*/
complex **zm_alloc( int nr, int nc )
{
  register int i;
  complex **z ;

  z = (complex **) calloc(nr,sizeof(complex *));
  for(i=0; i<nr; i++) z[i] = (complex *) calloc(nc,sizeof(complex)); 

  return z ;
}

/**<zt3_alloc.c>*************************************************************

       Title:  float Complex 3D Tensor space ALLOCation

     Purpose:  Allocates a float complex matrix z[n1][n2][n3]
*/
complex ***zt3_alloc(int n1, int n2, int n3)
{
  register int i,j;
  complex ***z ;

  z = (complex ***) calloc(n1,sizeof(complex **));
  for(i=0; i<n1; i++) {
    z[i] = (complex **) calloc(n2,sizeof(complex *)); 
    for(j=0; j<n2; j++) {
      z[i][j] = (complex *) calloc(n3,sizeof(complex)); 
    }
  }

  return z ;
}

/**<zt4_alloc.c>*************************************************************

       Title:  float Complex 4D Tensor space ALLOCation

     Purpose:  Allocates a float complex matrix z[n1][n2][n3][n4]
*/
complex ****zt4_alloc(int n1, int n2, int n3, int n4)
{
  register int i,j,k;
  complex ****z ;

  z = (complex ****) calloc(n1,sizeof(complex ***));
  for(i=0; i<n1; i++) {
    z[i] = (complex ***) calloc(n2,sizeof(complex **)); 
    for(j=0; j<n2; j++) {
      z[i][j] = (complex **) calloc(n3,sizeof(complex *)); 
      for(k=0; k<n3; k++) {
	z[i][j][k] = (complex *) calloc(n4,sizeof(complex)); 
      }
    }
  }

  return z ;
}

/** d3tensor.c *************************************************************

       Title:  Double TENSOR space allocation

     Purpose:  Allocates a double tensor with 
               range [nrl..nrh][ncl..nch][ndl..ndh]

   Reference:  Numerical Recipes, Appendix D

*/
double ***d3tensor(int nrl,int nrh,int ncl,int nch,int ndl,int ndh)
{
  int i,j;
  int nrow = nrh-nrl+1;
  int ncol = nch-ncl+1;
  int ndep = ndh-ndl+1;
  double ***t;

  /* Allocate pointers to pointers to rows */

  t = (double ***) calloc((nrow+NR_END),sizeof(double**));
  if (!t) perror("Allocation failure 1 in d3tensor()");
  t += NR_END;
  t -= nrl;

  /* Allocate pointers to rows and set pointers to them */

  t[nrl] = (double **) calloc( (nrow*ncol+NR_END),sizeof(double*));
  if (!t[nrl]) perror("Allocation failure 2 in d3tensor()");
  t[nrl] += NR_END;
  t[nrl] -= ncl;

  /* Allocate rows and set pointers to them */

  t[nrl][ncl] = (double *) calloc( (nrow*ncol*ndep+NR_END),sizeof(double));
  if (!t[nrl][ncl]) perror("Allocation failure 3 in d3tensor()");
  t[nrl][ncl] += NR_END;
  t[nrl][ncl] -= ndl;

  for(j=ncl+1; j<=nch; j++) t[nrl][j] = t[nrl][j-1]+ndep;
  for(i=nrl+1; i<=nrh; i++) {
    t[i] = t[i-1] + ncol;
    t[i][ncl] = t[i-1][ncl] + ncol*ndep;
    for(j=ncl+1; j<=nch; j++) t[i][j] = t[i][j-1] + ndep;
  }

  /* Return pointer to array of pointers to rows. */
  return t;
}

/** f3tensor.c *************************************************************

       Title:  Float TENSOR space allocation

     Purpose:  Allocates a float tensor with 
               range [nrl..nrh][ncl..nch][ndl..ndh]

   Reference:  Numerical Recipes, Appendix D

*/
float ***f3tensor(int nrl, int nrh, int ncl, int nch, int ndl, int ndh)
{
  int i,j;
  int nrow = nrh-nrl+1;
  int ncol = nch-ncl+1;
  int ndep = ndh-ndl+1;
  float ***t;

  /* Allocate pointers to pointers to rows */

  t = (float ***) calloc((nrow+NR_END),sizeof(float**) );
  if (!t) perror("Allocation failure 1 in f3tensor()");
  t += NR_END;
  t -= nrl;

  /* Allocate pointers to rows and set pointers to them */

  t[nrl] = (float **) calloc((nrow*ncol+NR_END),sizeof(float*));
  if (!t[nrl]) perror("Allocation failure 2 in f3tensor()");
  t[nrl] += NR_END;
  t[nrl] -= ncl;

  /* Allocate rows and set pointers to them */

  t[nrl][ncl] = (float *) calloc((nrow*ncol*ndep+NR_END),sizeof(float));
  if (!t[nrl][ncl]) perror("Allocation failure 3 in f3tensor()");
  t[nrl][ncl] += NR_END;
  t[nrl][ncl] -= ndl;

  for(j=ncl+1; j<=nch; j++) t[nrl][j] = t[nrl][j-1]+ndep;
  for(i=nrl+1; i<=nrh; i++) {
    t[i] = t[i-1] + ncol;
    t[i][ncl] = t[i-1][ncl] + ncol*ndep;
    for(j=ncl+1; j<=nch; j++) t[i][j] = t[i][j-1] + ndep;
  }

  /* Return pointer to array of pointers to rows. */
  return t;
}

/** s3tensor.c *************************************************************

       Title:  Short Int TENSOR space allocation

     Purpose:  Allocates a short int tensor with 
               range [nrl..nrh][ncl..nch][ndl..ndh]

   Reference:  Numerical Recipes, Appendix D

*/
short int ***s3tensor(int nrl, int nrh, int ncl, int nch, int ndl, int ndh)
{
  int i,j;
  int nrow = nrh-nrl+1;
  int ncol = nch-ncl+1;
  int ndep = ndh-ndl+1;
  short int ***t;

  /* Allocate pointers to pointers to rows */

  t = (short int ***) calloc((nrow+NR_END),sizeof(short int**));
  if (!t) perror("Allocation failure 1 in s3tensor()");
  t += NR_END;
  t -= nrl;

  /* Allocate pointers to rows and set pointers to them */

  t[nrl] = (short int **) calloc((nrow*ncol+NR_END),sizeof(short int*));
  if (!t[nrl]) perror("Allocation failure 2 in s3tensor()");
  t[nrl] += NR_END;
  t[nrl] -= ncl;

  /* Allocate rows and set pointers to them */

  t[nrl][ncl] = (short int *) calloc((nrow*ncol*ndep+NR_END),sizeof(short int));
  if (!t[nrl][ncl]) perror("Allocation failure 3 in s3tensor()");
  t[nrl][ncl] += NR_END;
  t[nrl][ncl] -= ndl;

  for(j=ncl+1; j<=nch; j++) t[nrl][j] = t[nrl][j-1]+ndep;
  for(i=nrl+1; i<=nrh; i++) {
    t[i] = t[i-1] + ncol;
    t[i][ncl] = t[i-1][ncl] + ncol*ndep;
    for(j=ncl+1; j<=nch; j++) t[i][j] = t[i][j-1] + ndep;
  }

  /* Return pointer to array of pointers to rows. */
  return t;
}

/**<st_alloc.c>*************************************************************

       Title:  Short Int 3-Tensor space ALLOCation

     Purpose:  Allocates a three-dimension matrix d[l][m][n]
*/
short int ***s3t_alloc( int l, int m, int n )
{
  return s3tensor(0,l-1,0,m-1,0,n-1);
}

/**<ft_alloc.c>*************************************************************

       Title:  Float 3-Tensor space ALLOCation

     Purpose:  Allocates a three-dimension matrix d[l][m][n]
*/
float ***f3t_alloc( int l, int m, int n)
{
  return f3tensor(0,l-1,0,m-1,0,n-1);
}

/**<dt_alloc.c>*************************************************************

       Title:  Double 3-Tensor space ALLOCation

     Purpose:  Allocates a three-dimension matrix d[l][m][n]
*/
double ***d3t_alloc( int l, int m, int n)
{
  return d3tensor(0,l-1,0,m-1,0,n-1);
}

