/**<write_data_to_fim.c>*******************************************************

     Title: write data to fim

   Purpose: 

     Input:  PLUGIN_interface  *plint = plugin interface
                     char *new_prefix = output file name
		Boolean OverwriteMode = 0 : don't allow overwrite
		                      = 1 : allow overwrite

*/
#include "utils.h"

/*------------------ macros to return workspace at exit -------------------*/

#undef  FREEUP
#define FREEUP(x) do{ if((x) != NULL){free((x)); (x)=NULL;} } while(0)

#define FREE_WORKSPACE                \
  do{ FREEUP(fout) ; FREEUP(facs)  ;  \
    } while(0)

/*-------------------------------------------------------------------------*/

THD_3dim_dataset *write_data_to_fim( PLUGIN_interface  *plint , 
				     int               nsbricks, /* # sub-bricks */
				     int               nsbsize,	/* sub-brick size */
				     THD_3dim_dataset  *old_dset ,
				     float             **fout ,  /* array of input floats */
				     char              *new_prefix , 
				     Boolean           OverWriteMode )
{
  THD_3dim_dataset *new_dset;
  int new_datum;
  int ii,kk,nvox;
  float  *facs = NULL ;  /* will be array of scale factors */

   if( ! PLUTO_prefix_ok(new_prefix) ) {
     printf("**************************\n"
	    "write_data_to_fim:  bad prefix\n"
	    "**************************\n")  ;
      return NULL ;
   }

   /*-- make a new dataset --*/

   new_dset  = EDIT_empty_copy( old_dset ) ; /* start with copy of old one */
   new_datum = MRI_short;	             /* new type */

   /*-- edit some of its internal parameters --*/

   ii = EDIT_dset_items(
           new_dset ,
              ADN_prefix      , new_prefix ,           /* filename prefix */
              ADN_malloc_type , DATABLOCK_MEM_MALLOC , /* store in memory */
              ADN_datum_all   , new_datum ,            /* atomic datum */
              ADN_nvals       , 1 ,                    /* # sub-bricks */
              ADN_ntt         , 0 ,                    /* # time points */
              ADN_type        , ISHEAD(old_dset)       /* dataset type */
                                 ? HEAD_FUNC_TYPE
                                 : GEN_FUNC_TYPE ,
              ADN_func_type   , FUNC_FIM_TYPE ,        /* function type */
           ADN_none ) ;

   if( ii != 0 ){
      THD_delete_3dim_dataset( new_dset , False ) ;  /* some error above */
      return NULL ;
   }

   if (!OverWriteMode && THD_is_file(DSET_HEADNAME( new_dset )) ) {
     printf("WriteZeroDset: File %s exists! ", DSET_HEADNAME(new_dset));
     printf("No new file created!\n");
     return NULL ;
   }

   /*------ make floating point output brick
            (only at the end will scale to byte or shorts) ------*/

   if( fout == NULL ){
      THD_delete_3dim_dataset( new_dset , False ) ;
      return NULL ;
   }

   nvox = DSET_NVOX(old_dset);

   facs = (float *) malloc( sizeof(float) * nsbricks ) ;

   if( facs == NULL ){
      FREE_WORKSPACE ;
      return "Malloc\nFailure!\n [write_data_to_fim:facs]" ;
   }

   switch( new_datum ){

      /*** output is floats is the simplest:
           we just have to attach the fout brick to the dataset ***/

      case MRI_float:
         for( kk=0 ; kk < nsbricks ; kk++ )
            EDIT_substitute_brick( new_dset , kk , MRI_float , fout[kk] ) ;
      break ;

      /*** output is shorts:
           we have to create a scaled sub-brick from fout ***/

      case MRI_short:{
         short *bout ;
         float sfac ;

         for( kk=0 ; kk < nsbricks ; kk++ ){  /* loop over sub-bricks */

	   /*-- get output sub-brick --*/

	   bout = (short *) calloc(sizeof(short),nvox); 
	   if( bout == NULL ){
	     fprintf(stderr,"\nFinal calloc error in write_data_to_fim!\n\a") ;
	     exit(1) ;
	   }

	   /*-- find scaling and then scale --*/

	   sfac = MCW_vol_amax( nvox,1,1 , MRI_float , fout[kk] ) ;
	   if( sfac > 0.0 ){
	     sfac = 32767.0 / sfac ;
	     EDIT_coerce_scale_type( nvox, sfac ,
				     MRI_float,fout[kk] , MRI_short, bout ) ;
	     sfac = 1.0 / sfac ;
	   }
	   
	   free( fout[kk] ) ;  /* don't need this anymore */

         /*-- put output brick into dataset, and store scale factor --*/

         EDIT_substitute_brick( new_dset , kk , MRI_short , bout ) ;
	 facs[kk] = sfac ;
	 }
      }

      EDIT_dset_items( new_dset , ADN_brick_fac , facs , ADN_none ) ;

      break ;

      /*** output is bytes (byte = unsigned char)
           we have to create a scaled sub-brick from fout ***/

      case MRI_byte:{
         byte * bout ;
         float sfac ;

         for( kk=0 ; kk < nsbricks ; kk++ ){  /* loop over sub-bricks */

	   /*-- get output sub-brick --*/

	   bout = (byte *) calloc(sizeof(byte),nvox); 
	   if( bout == NULL ){
	     fprintf(stderr,"\nFinal calloc error in write_data_to_fim!\n\a") ;
	     exit(1) ;
	   }

	   /*-- find scaling and then scale --*/

	   sfac = MCW_vol_amax( nvox,1,1 , MRI_float , fout[kk] ) ;
	   if( sfac > 0.0 ){
	     sfac = 255.0 / sfac ;
	     EDIT_coerce_scale_type( nvox,sfac ,
				     MRI_float,fout[kk] , MRI_byte, bout ) ;
	     sfac = 1.0 / sfac ;
	   }

	   free( fout[kk] ) ;  /* don't need this anymore */

	   /*-- put output brick into dataset, and store scale factor --*/

	   EDIT_substitute_brick( new_dset , kk , MRI_byte , bout ) ;
	   facs[kk] = sfac ;

	 }
	 EDIT_dset_items( new_dset , ADN_brick_fac , facs , ADN_none ) ;
      }
      break ;

   } /* end of switch on output data type */


   /*----- write new data to disk and new data set visible to afni -----*/

   /* Note: PLUTO_add_dset does both of these things!
            You need to reload to make it visible!  */

   printf("Writing `data' dataset ");
   printf("into %s\n", DSET_HEADNAME(new_dset));
   DSET_FIX_NAMES( new_dset );	/* Need this to create correct labels */
   PLUTO_add_dset( plint , new_dset , DSET_ACTION_MAKE_CURRENT ) ;
   DSET_load( new_dset );

   /*----- write data set -----*/
   
   /*   printf("Writing `zero' dataset ");
	printf("into %s\n", DSET_HEADNAME(new_dset));
	DSET_write(new_dset); */

   return ( new_dset );

}
