#include "io.h"

/* Parse an integer inum into n separate numbers
   eg, if inum = 43512, then n = 5, and
   iv[0-4]={4,3,5,1,2} */

int *iparse(int inum,int n)
{
  int i ;

  int *iv = (int *) calloc(n,sizeof(int));

  for(i=0; i<n; i++) {
    iv[i] = (int) floor( (inum % (int) powf(10,n-i))/powf(10,n-1-i) );
  }

  return(iv) ;
}

/* Concatenate integer from n separate numbers
   eg, if iv={4,3,5,1,2}, then n=5 and
   inum = 43512 */

int icat(int *iv,int n)
{
  int i ;
  int inum = 0;

  for(i=0; i<n; i++) {
    inum += (int) (iv[i] * powf(10,n-i-1)) ;
  }

  return(inum) ;
}

/*===<pinput.c>========================================================

     Title: Parameter INPUT

   Purpose: Prompt user for numerical input

      Call: float pinput(char *pname, float pdef)
      
     Input: char *pname = parameter name for prompt
            float pdef  = parameter default value

    Return: float value of input

   History: 02_11_15 - lrf

*/
float pinput(char *pname, float pdef, float *plims)
{
  char  line[100];
  int   nn;
  float fval;

  int mistakes = 0;
  int too_many_mistakes = 10;

  do {

    printf(" Enter %s [%g] {%g,%g}: ",pname,pdef,plims[0],plims[1]);
    fgets(line,sizeof(line),stdin);
    nn = sscanf(line,"%f",&fval);

    if(nn<0) fval = pdef;

    if(!BETWEEN(fval,plims[0],plims[1])) {
      printf("\n ***** Parameter '%s' must be between %f and %f\a\n",
	     pname,plims[0],plims[1]) ;
      mistakes++ ;
      if(mistakes>too_many_mistakes) {
	printf("You've had your chance!  Later ...\a\n");
	exit(0);
      }
    }

  }  while(!BETWEEN(fval,plims[0],plims[1])) ;

  return(fval) ;
}

/*===<logqyn.c>========================================================

     Title: LOGical Question: Yes or No

   Purpose: Prompt user for yes/no answer

      Call: Boolean logqyn(char *question, char *qdef)
      
     Input: char *question = query to respond to

    Return: YES or NO

   History: 02_12_30 - lrf

*/
Boolean logqyn(char *question, char *qdef)
{
  char line[100], sval[100] ;
  int  nn, answer ;

  do {

    printf(" %s? [%s]: ",question,qdef);
    fgets(line,sizeof(line),stdin);
    nn = sscanf(line,"%s",sval);

    if(nn<0) strcpy(sval,qdef);

         if( !strncmp(sval,"y",1) || !strncmp(sval,"Y",1) ) answer = YES   ;
    else if( !strncmp(sval,"n",1) || !strncmp(sval,"N",1) ) answer = NO    ;
    else                                                    answer = MAYBE ;

  } while (answer!=YES && answer!=NO) ;

  return(answer) ;
}

/*===<sinput.c>========================================================

     Title: String INPUT

   Purpose: Prompt user for string input

      Call: char sinput(char *sval, char *sname, char *sdef)
      
     Input: char *sname = string name for prompt
            char sdef   = string default value

    Return: char value of input

   History: 02_11_15 - lrf

     Notes: Need to put in allowable values
*/
char *sinput(char *sval, char *sname, char *sdef)
{
  char line[100];
  int  nn;

  printf(" Enter %s [%s]: ",sname,sdef);
  fgets(line,sizeof(line),stdin);
  nn = sscanf(line,"%s",sval);

  if(nn<0) strcpy(sval,sdef);

  return(sval) ;
}

/* ----------------------------------------------------------------
   Write file subroutine. A.Jesmanowicz, MCW 1989
	ier   :	return error :	1 - opening problem,
				2 - negative length.
	fname :	file name.
	size  :	size of the file.
	arr   :	file array.
   --------------------------------------------------------------*/
void write_iqm(long int *ier,char fname[],long int *size,char arr[])
{
	int	isize = *size;
	int	fp;				/* file descriptor  */

	{	int i=0;			/* cut junk from the name */
		while(fname[i] != 32 && fname[i]) i++;
		fname[i] = 0;
	}

	if(isize < 0)				/* size has to be real */
	{	*ier=2;				/* or ier = 2.	       */
		return;
	}

	if ((fp = open(fname, O_WRONLY|O_CREAT|O_EXCL,0644)) <= 0)
	{	*ier=1;				/* file must not exist */
		return;				/* or ier = 1.	   */
	}

	*ier=0;					/* no error : ier=0 */
	write(fp, arr, isize);			/* write whole file */
	close(fp);
	return;  
}

/* Report for write_iqm  */

int write_iqm_error( long int *ier , char *fname )
{
  char *errstr;

  if(*ier==1) errstr = "FILE EXISTS";
  if(*ier==2) errstr = "FILE SIZE < 0";

  if (*ier!=0) {
    printf("ERROR WRITING FILE %s! %s\n",fname,errstr);
  }

  return(*ier);

}

int write_file( long int *ier,
		char     fname[],
		long int *size,
		char     arr[] )
{
  write_iqm(ier,fname,size,arr);
  return(write_iqm_error(ier,fname));
}

void read_iqm(long int *ier,char fname[],long int *size,char arr[])
{
	int	isize = *size;
	int	fp;				/* file descriptor  */
	struct stat file_stat;			/* status structure */

	{	int i=0;			/* cut junk from the name */
		while(fname[i] != 32 && fname[i]) i++;
		fname[i] = 0;
	}

	if ((fp = open(fname, O_RDONLY)) <= 0)	/* file must exist */
	{	*ier=1;				/* or ier = 1.	   */
		return;
	}

	fstat(fp, &file_stat);			/* get file size in bytes   */

	if(file_stat.st_size > isize && isize)	/* file can not be too long */
	{	*ier=2;				/* or ier = 2.		    */
		return;
	}
	*size =  file_stat.st_size;		/* return file size */

	read(fp, arr, file_stat.st_size);	/* read whole file  */
	close(fp);
	*ier=0;					/* no error : ier=0 */
	return;
}

/* Report for read_iqm  */

int read_iqm_error( long int *ier , char *fname )
{
  char *errstr;

  if(*ier==1) errstr = "FILE DOES NOT EXIST";
  if(*ier==2) errstr = "FILE SIZE TOO LONG";

  if (*ier!=0) {
    printf("ERROR READING FILE %s! %s\n",fname,errstr);
  }

  return(*ier);

}

int read_file( long int *ier,
	       char     fname[],
	       long int *size,
	       char     arr[] )
{
  read_iqm(ier,fname,size,arr);
  return(read_iqm_error(ier,fname));
}
