/**<mask.c>****************************************************

     Title: MASK

   Purpose: Create byte mask and determine mean and std for
            voxels in briks ivbot->ivtop that are within
	    range mask_bot->mask_top.

      Call: void *mask( THD_3dim_dataset *input_dset, 
                              Mask_state *ms, 
			      int report )

     Input: input_dset = input THD_3dim_dataset
                    ms = Mask_state struct (see init_mask_state.c)
	        report = print out (1) or don't (0) results

   History: Nov 1999 RWC's plug_maskave.c -> mask.c

*/
#include "utils.h"

void *mask( THD_3dim_dataset *input_dset, Mask_state *ms, int report )
{
   MCW_idcode *idc ;
   THD_3dim_dataset *mask_dset ;
   int iv , mcount , nvox , ii , sigmait , nvals , doall ;
   double sum , sigma ;
   float *sumar=NULL , *sigmar=NULL ;
   char *tag , *str , buf[64] , abuf[32],sbuf[32] ;
   int  miv;

   /* Retrieve items from Mask_state */

   byte      *mmm = ms->mask ;
   int      ivbot = ms->ivbot ;
   int      ivtop = ms->ivtop ;
   float mask_bot = ms->mask_bot;
   float mask_top = ms->mask_top;

   /*--------------------------------------------------------------------*/

   if( input_dset == NULL ) {
     printf("mask:  bad input dataset\n");
     return NULL;
   }
   if ( ivbot < 0 ) {
     printf("mask:  min brick = 0\n");
     return NULL;
   }

   miv = ivbot ;
   nvals  = DSET_NVALS(input_dset) ;

   if ( ivtop > nvals - 1 ) {
     printf("mask:  max brick exceded\n");
     return NULL;
   }

   DSET_load(input_dset) ;
   if( DSET_ARRAY(input_dset,ivbot) == NULL ) {
     printf("mask:  can't load dataset\n");
     return NULL;
   }
   nvox = DSET_NVOX(input_dset) ;

   doall = (ivbot = 0 && ivtop == nvals - 1);

   /*---------- At this point, the inputs are OK ----------*/

   /* separate code for each input data type */

   switch( DSET_BRICK_TYPE(input_dset,miv) ){
      default:
         free(mmm) ;
	 printf("mask:  *** Can't use mask dataset -- illegal data type! ***\n");
	 return NULL;

      case MRI_short:{
         short mbot , mtop ;
         short * mar = (short *) DSET_ARRAY(input_dset,miv) ;
         float mfac = DSET_BRICK_FACTOR(input_dset,miv) ;
         if( mfac == 0.0 ) mfac = 1.0 ;
         if( mask_bot <= mask_top ){
            mbot = SHORTIZE(mask_bot/mfac) ;
            mtop = SHORTIZE(mask_top/mfac) ;
         } else {
            mbot = (short) -MRI_TYPE_maxval[MRI_short] ;
            mtop = (short)  MRI_TYPE_maxval[MRI_short] ;
         }
         for( mcount=0,ii=0 ; ii < nvox ; ii++ )
            if( mar[ii] >= mbot && mar[ii] <= mtop && mar[ii] != 0 ){ mmm[ii] = 1 ; mcount++ ; }
            else                                                    { mmm[ii] = 0 ; }
      }
      break ;

      case MRI_byte:{
         byte mbot , mtop ;
         byte * mar = (byte *) DSET_ARRAY(input_dset,miv) ;
         float mfac = DSET_BRICK_FACTOR(input_dset,miv) ;
         if( mfac == 0.0 ) mfac = 1.0 ;
         if( mask_bot <= mask_top ){
            mbot = BYTEIZE(mask_bot/mfac) ;
            mtop = BYTEIZE(mask_top/mfac) ;
            if( mtop == 0 ){
               free(mmm) ;
	       printf("mask:  *** Illegal mask range for mask dataset of bytes. ***\n");
	       return NULL;
            }
         } else {
            mbot = 0 ;
            mtop = (byte) MRI_TYPE_maxval[MRI_short] ;
         }
         for( mcount=0,ii=0 ; ii < nvox ; ii++ )
            if( mar[ii] >= mbot && mar[ii] <= mtop && mar[ii] != 0 ){ mmm[ii] = 1 ; mcount++ ; }
            else                                                    { mmm[ii] = 0 ; }
      }
      break ;

      case MRI_float:{
         float mbot , mtop ;
         float * mar = (float *) DSET_ARRAY(input_dset,miv) ;
         float mfac = DSET_BRICK_FACTOR(input_dset,miv) ;
         if( mfac == 0.0 ) mfac = 1.0 ;
         if( mask_bot <= mask_top ){
            mbot = (float) (mask_bot/mfac) ;
            mtop = (float) (mask_top/mfac) ;
         } else {
            mbot = -WAY_BIG ;
            mtop =  WAY_BIG ;
         }
         for( mcount=0,ii=0 ; ii < nvox ; ii++ )
            if( mar[ii] >= mbot && mar[ii] <= mtop && mar[ii] != 0 ){ mmm[ii] = 1 ; mcount++ ; }
            else                                                    { mmm[ii] = 0 ; }
      }
      break ;
   }

   if( mcount == 0 ){
      mmm = NULL ;
      printf("mask:  *** No voxels survive the masking operations! ***\n");
      return NULL ;
   }
   sigmait = (mcount > 1) ;

   /*-- compute statistics --*/

   if( doall ){
      sumar  = (float *) malloc( sizeof(float) * nvals ) ;
      sigmar = (float *) malloc( sizeof(float) * nvals ) ;
   }

   for( iv=ivbot ; iv <= ivtop ; iv++ ){
      sum = sigma = 0.0 ;                         /* 13 Dec 1999 */
      switch( DSET_BRICK_TYPE(input_dset,iv) ){

         default:
            mmm = NULL ; if( doall ){ free(sumar) ; free(sigmar) ; }
	    printf("mask:  *** Can't use source dataset -- illegal data type! ***\n");
	    return NULL;

         case MRI_short:{
            short * bar = (short *) DSET_ARRAY(input_dset,iv) ;
            float mfac = DSET_BRICK_FACTOR(input_dset,iv) ;
            if( mfac == 0.0 ) mfac = 1.0 ;

            for( ii=0 ; ii < nvox ; ii++ ) if( mmm[ii] ) sum += bar[ii] ;
            sum = sum / mcount ;

            if( sigmait ){
               for( ii=0 ; ii < nvox ; ii++ )
                  if( mmm[ii] ) sigma += SQR(bar[ii]-sum) ;
               sigma = mfac * sqrt( sigma/(mcount-1) ) ;
            }
            sum = mfac * sum ;
         }
         break ;

         case MRI_byte:{
            byte * bar = (byte *) DSET_ARRAY(input_dset,iv) ;
            float mfac = DSET_BRICK_FACTOR(input_dset,iv) ;
            if( mfac == 0.0 ) mfac = 1.0 ;

            for( ii=0 ; ii < nvox ; ii++ ) if( mmm[ii] ) sum += bar[ii] ;
            sum = sum / mcount ;

            if( sigmait ){
               for( ii=0 ; ii < nvox ; ii++ )
                  if( mmm[ii] ) sigma += SQR(bar[ii]-sum) ;
               sigma = mfac * sqrt( sigma/(mcount-1) ) ;
            }
            sum = mfac * sum ;
         }
         break ;

         case MRI_float:{
            float * bar = (float *) DSET_ARRAY(input_dset,iv) ;
            float mfac = DSET_BRICK_FACTOR(input_dset,iv) ;
            if( mfac == 0.0 ) mfac = 1.0 ;

            for( ii=0 ; ii < nvox ; ii++ ) if( mmm[ii] ) sum += bar[ii] ;
            sum = sum / mcount ;

            if( sigmait ){
               for( ii=0 ; ii < nvox ; ii++ )
                  if( mmm[ii] ) sigma += SQR(bar[ii]-sum) ;
               sigma = mfac * sqrt( sigma/(mcount-1) ) ;
            }
            sum = mfac * sum ;
         }
         break ;
      }

      if( doall ){ sumar[iv] = sum ; sigmar[iv] = sigma ; }
   }

   /*-- send report --*/

   if( doall ){
      str = (char *) malloc( 1024 + 64*nvals ) ;
      sprintf(str," ****** ROI statistics ****** \n"
                  " Source  = %s [all sub-bricks] \n"
                  " Mask    = %s [%s]" ,
              DSET_FILECODE(input_dset) ,
              DSET_FILECODE(input_dset)  , DSET_BRICK_LABEL(input_dset,miv) ) ;
      if( mask_bot <= mask_top ){
         sprintf(buf," [range %g .. %g]" , mask_bot , mask_top ) ;
         strcat(str,buf) ;
      }
      strcat(str," \n") ;
      sprintf(buf," Count   = %d voxels\n",mcount) ; strcat(str,buf) ;
      for( iv=0 ; iv < nvals ; iv++ ){
         AV_fval_to_char( sumar[iv]  , abuf ) ;
         AV_fval_to_char( sigmar[iv] , sbuf ) ;
         sprintf(buf," Average = %9.9s  Sigma = %9.9s [%s]  \n",
                 abuf,sbuf , DSET_BRICK_LABEL(input_dset,iv) ) ;
         strcat(str,buf) ;
      }

      if (report)
	printf("mask results: %s\n" ,str ) ;

      free(str) ; free(sumar) ; free(sigmar) ;

   } else if( mask_bot <= mask_top ){
      str = (char *) malloc( 1024 ) ;
      sprintf( str , " *** ROI Statistics *** \n"
                     " Source  = %s [%s] \n"
                     " Mask    = %s [%s] [range %g .. %g] \n"
                     " Count   = %d voxels \n"
                     " Average = %g \n"
                     " Sigma   = %g " ,
               DSET_FILECODE(input_dset) , DSET_BRICK_LABEL(input_dset,ivbot) ,
               DSET_FILECODE(input_dset)  , DSET_BRICK_LABEL(input_dset,miv)    ,
               mask_bot , mask_top , mcount , sum , sigma ) ;

      ms->mcount = mcount ;
      ms->mean   = sum ;
      ms->std    = sigma ;

      if (report)
	printf("mask results: %s\n" ,str ) ;

      free(str) ;

   } else {
      str = (char *) malloc( 1024 ) ;
      sprintf( str , " *** ROI Statistics *** \n"
                     " Source  = %s [%s] \n"
                     " Mask    = %s [%s] \n"
                     " Count   = %d voxels \n"
                     " Average = %g \n"
                     " Sigma   = %g " ,
               DSET_FILECODE(input_dset) , DSET_BRICK_LABEL(input_dset,ivbot) ,
               DSET_FILECODE(input_dset)  , DSET_BRICK_LABEL(input_dset,miv)    ,
               mcount , sum , sigma ) ;

      if (report)
	printf("mask results: %s\n" ,str ) ;

      ms->mcount = mcount ;
      ms->mean   = sum ;
      ms->std    = sigma ;

      free(str) ;
   }

   return NULL ;
}
