/* vmutils.c */

#include "dell.h"

/*	double dvmnmx(x,n,nmin,nmax)
 *	double *x ;
 *	int n,*nmin,*nmax ;
 *
 * Returns the index of the minimum (*nmin) and maximum (*nmax) values in
 *   the double vector x[].  Also value of function is the maximum of the
 *   absolute values of the vector points.
 * If nmin=0 or nmax=0, then does not return these pointer values
 * If n<=0, returns immediately with value 0 and *nmin = *nmax = 0.
*/

double dvmnmx( register double *x, 
	       register int n, 
	       register int *nmin, 
	       register int *nmax)
{
	double *xold = x ;
	register double xmin,xmax ;
	register int nnmin,nnmax ;
	register int count ;

	if(n<=0)
	{
		if(nmin) *nmin = 0 ;
		if(nmax) *nmax = 0 ;
		return(0.) ;
	}
	nnmin = nnmax = 0 ;
	xmin = xmax = *x++ ;
	n-- ;
	count = 1 ;
	if(n>0)
	{
		do {
			if(*x < xmin) {
				nnmin = count++ ;
				xmin = *x++ ;
			} else if(*x > xmax) {
				nnmax = count++ ;
				xmax = *x++ ;
			} else {
				count++ ;
				x++ ;
			}
		} while( --n > 0 ) ;
	}
	if(nmin) *nmin = nnmin ;
	if(nmax) *nmax = nnmax ;
	xmax = mabs(xmax) ;
	xmin = mabs(xmin) ;
	return(max(xmax,xmin)) ;
}


/**<dvran>**************************************************

    Title: Double Vector RANge

  Purpose: Find the minimum and maximum values,
           and their respective indices, of a double vector.
 
    Input: a = double vector
           n = length(x)

   Output:  amin (double) = min(a)
           lamin    (int) = location of amin in a[] 
            amax (double) = max(a)
           lamax    (int) = location of amax in a[] 

    Notes: call must use addresses!
           dvran(a,&amin,&lamin,&amax,&lamax,n);
*/
void dvran(double *a, 
	   double *amin, int *lamin, 
	   double *amax, int *lamax, 
	   int n)
{
  dvmnmx(a,n,lamin,lamax);

  *amin = *(a + *lamin);
  *amax = *(a + *lamax); 
}

/**<dmtrace.c>************************************************************

     Title: Double Matrix TRACE

   Purpose: Compute trace of double matrix

      Call: double dmtrace(x,nr,nc)
                 double **x;
                 int    nr,nc;

     Input: x[nr][nc] = input matrix
                   nr = number of rows in x
                   nc = number of cols in x

    Output: double trace(x)
*/

double dmtrace( double **x, int nr, int nc )
{
  register double sum = 0.0;
  int      i,n = min(nr,nc);

  for (i=0; i<n; i ++) {
      sum += x[i][i];
  }
  return(sum);
}

/**<dvmean>**************************************************

    Title: Double Vector MEAN

  Purpose: Compute mean of double vector.
 
     Call: double dvmean( double *x, int n )

*/

double dvmean( double *x, int n )
{
  double sum, mean;
  double dn = n;

  sum  = dvsum(x,n);
  mean = ((double) sum)/dn;

  return(mean);
}

/**<svmean>**************************************************

    Title: Short integer Vector MEAN

  Purpose: Compute mean of short vector.
 
     Call: double svmean( short int *x, int n )

     Note: Returns double (_not_ short)!
*/

double svmean( short *x, int n )
{
  short sum, mean;
  double dn = n;

  sum  = svsum(x,n);
  mean = ((double) sum)/dn;

  return(mean);
}

/**<dvvar>**************************************************

    Title: Double Vector VARiance

  Purpose: Compute sample variance of double vector.
 
     Call: double dvvar( register double *x, register int n)

*/

double dvvar( double *x, int n )
{
       double dn = n-1 ;
       double mean,var,temp  ;
       double sum=0.;

       mean = dvmean(x,n);

       if(n>0)
	 {
	   do {
	     temp = *x++ - mean ;
	     sum += temp*temp ;
	   } while(--n > 0) ;
	 }
  
       var = sum / dn;
       return(var);
}

/**<dvstd>**************************************************

    Title: Double integer Vector STD

  Purpose: Compute std of double vector.
 
     Call: double dvstd( double int *x, int n )

*/

double dvstd( double *x, int n )
{
  register double sum=0.;
  register double dn = n - 1;
  register double mean,std,tmp;

  mean = dvmean(x,n);

  if(n>0)
    {
      do {
	tmp  = *x++ - mean ;
	sum += tmp*tmp;
      } while(--n > 0) ;
    }
  sum = sqrt(sum/dn);
  
  return(sum);
}

/**<svstd>**************************************************

    Title: Short integer Vector STD

  Purpose: Compute std of short vector.
 
     Call: double svstd( short int *x, int n )

     Note: Returns double (_not_ short)!
*/

double svstd( register short *x, register int n )
{
  register double sum=0.;
  register double dn = n - 1;
  register double mean,std,tmp;

  mean = svmean(x,n);

  if(n>0)
    {
      do {
	tmp  = *x++ - mean ;
	sum += tmp*tmp;
      } while(--n > 0) ;
    }
  sum = sqrt(sum/dn);
  
  return(sum);
}

/**<svsum>*************************************************************
  short svsum(x,n)       Sums elements of vector x[0],...,x[n-1]
  short *x ;             
 int n ;                 sum is returned in short precision.
*/

short svsum( short *x, int n )
{
        register short sum=0;
        if(n>0)
        {
                do {
                        sum += *x++ ;
                } while(--n > 0) ;
        }
        return(sum);
}

/**<dvsum>*************************************************************
  double dvsum(x,n)       Sums elements of vector x[0],...,x[n-1]
  double *x ;             Internal calculation is double precision,
  int n ;                 and the sum is returned in double precision.
*/

double dvsum( double *x, int n )
{
        register double sum=0;
        if(n>0)
        {
                do {
                        sum += *x++ ;
                } while(--n > 0) ;
        }
        return(sum);
}

/**<fv2dm>****************************************************************

    Title: Float Vector TO Double Matrix

  Purpose: Convert float vector to double matrix

     Call: double **fv2dm( double **dmat, float *vec, int nr, int nc)
*/
double **fv2dm( double **dmat, float *fvec, int nr, int nc)
{
  register int i,j;

  for(i=0;i<nr;++i) {
    for(j=0;j<nc;++j) {
      dmat[i][j] = (double) fvec[j + i*nc];
    }
  }

  return(dmat);
}

/**<fm2dv>****************************************************************

    Title: Float Matrix TO Double Vector

  Purpose: Convert float matrix to double vector

     Call: double **fm2dv( double **dmat, float *vec, int nr, int nc)
*/
double *fm2dv( double *dvec, float **fmat, int nr, int nc)
{
  register int i,j;

  for(i=0;i<nr;++i) {
    for(j=0;j<nc;++j) {
      dvec[j + i*nc] = (double) fmat[i][j] ;
    }
  }

  return(dvec);
}
