#include "ParametricSHD.h"
#include "vtkObjectFactory.h"
#include "vtkMath.h"
#include <math.h>
#include "../vector/vector.h"
#include "../diffuse/coords.h"
#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_sf.h>

vtkCxxRevisionMacro(ParametricSHD, "$Revision: 1.2 $");
vtkStandardNewMacro(ParametricSHD);

ParametricSHD::ParametricSHD() :
  XRadius(1),
  YRadius(1),
  ZRadius(1),
  lmax(0),
  Alm(0)
{
  // Preset triangulation parameters
  this->MinimumU = 0;
  this->MinimumV = 0;
  this->MaximumU = 2.0 * vtkMath::Pi();
  this->MaximumV = vtkMath::Pi();

  this->JoinU = 1;
  this->JoinV = 0;
  this->TwistU = 0;
  this->TwistV = 0;
  this->ClockwiseOrdering = 1;
  //  this->DerivativesAvailable = 1;
  this->DerivativesAvailable = 0;
}

ParametricSHD::~ParametricSHD() {
}

void ParametricSHD::SetAlm(complex **alm, int lmax) {
  this->Alm = alm;
  this->Modified();
}

void ParametricSHD::Evaluate(double uvw[3], double Pt[3], double Duvw[9])
{
  double u = uvw[0];
  double v = uvw[1];
  double *Du = Duvw;
  double *Dv = Duvw + 3;

  for ( int i = 0; i < 3; ++i)
    {
    Pt[i] = Du[i] = Dv[i] = 0;
    }

  double cu = cos(u);
  double su = sin(u);
  double cv = cos(v);
  double sv = sin(v);

  double az = u;
  double po = v;

  double cp = cos(po);
  double sp = sin(po);
  double ca, sa;
  int am;
  complex Ylm, zsum, ztmp;
  double Plm;

  double norm = Alm[0][0].r;

  ZCLR(zsum);
  for (int l = 0; l < lmax; l++) {
    for (int m = -l; m <=l; m++) {
      ca = cos(m*az);
      sa = sin(m*az);
      am = ABS(m);

      Plm = gsl_sf_legendre_sphPlm(l, am, cp);
      Ylm.r = Plm*ca/norm;
      Ylm.i = Plm*sa/norm;
      ztmp = zmul(Alm[l][m],Ylm);
      ZADDTO(zsum, ztmp);
    }
  }

  spoint psph = newspt(ZABS(zsum), po, az);
  cpoint pcrt = spt2cpt(psph);

  Pt[0] = pcrt.x;
  Pt[1] = pcrt.y;
  Pt[2] = pcrt.z;

  // The point
//   Pt[0] = this->XRadius*sv*cu;
//   Pt[1] = this->YRadius*sv*su;
//   Pt[2] = this->ZRadius*cv;

  //The derivatives are:
//   Du[0] = -this->XRadius*sv*su;
//   Du[1] = this->YRadius*sv*cu;
//   Du[2] = 0;
//   Dv[0] = this->XRadius*cv*cu;
//   Dv[1] = this->YRadius*cv*su;
//   Dv[2] = -this->ZRadius*sv;

  Du[0] = 0;
  Du[1] = 0;
  Du[2] = 0;
  Dv[0] = 0;
  Dv[1] = 0;
  Dv[2] = 0;

}

double ParametricSHD::EvaluateScalar(double*, double*, double*)
{
  return 0;
}

void ParametricSHD::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os,indent);
  
  os << indent << "X scale factor: " << this->XRadius << "\n";
  os << indent << "Y scale factor: " << this->YRadius << "\n";
  os << indent << "Z scale factor: " << this->ZRadius << "\n";
}

