/*=========================================================================

  Program:   Visualization Toolkit
  Module:    $RCSfile: vtkPolyLineCurvature.cxx,v $
  Language:  C++
  Date:      $Date: 2003/05/13 00:00:00 $
  Version:   $Revision: 1.0 $

  Copyright (c) 1993-2002 Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
#include "vtkPolyLineCurvature.h"

#include "vtkCellArray.h"
#include "vtkDoubleArray.h"
#include "vtkMath.h"
#include "vtkObjectFactory.h"
#include "vtkPointData.h"
#include "vtkPolyData.h"


vtkCxxRevisionMacro(vtkPolyLineCurvature, "$Revision: 1.0 $");
vtkStandardNewMacro(vtkPolyLineCurvature);

vtkPolyLineCurvature::vtkPolyLineCurvature()
{
  this->ScalarOutput = VTK_CURVATURE;
}



void vtkPolyLineCurvature::Execute()
{
  struct frenet
  {
    double T[3];
    double B[3];
    double N[3];
    double ds;
  };

  vtkPolyData *input = this->GetInput();  
  vtkPolyData *output= this->GetOutput();
  output->CopyStructure(input);
  vtkCellArray *lines = input->GetLines();
  vtkPoints *points = input->GetPoints();
  vtkIdType npts=0, *linePts=0;
  vtkIdType i;
  vtkIdType numPts = input->GetNumberOfPoints();
  double p1[3], p2[3];
  double BB[3], kappa, tau;
  vtkDoubleArray *curvature = vtkDoubleArray::New();
    curvature->SetNumberOfComponents(1);
    curvature->SetNumberOfValues(numPts);
    curvature->SetName("Curvature");
  vtkDoubleArray *torsion = vtkDoubleArray::New();
    torsion->SetNumberOfComponents(1);
    torsion->SetNumberOfValues(numPts);
    torsion->SetName("Torsion");
  frenet *f;
  
  //  Loop over all lines
  for (lines->InitTraversal(); lines->GetNextCell(npts,linePts); )
    {
    
    // Need at least 3 points
    if ( npts <= 2 )
      {
      continue;
      }
      
    f = new frenet[npts];
      
    // Loop over points in line
    for (i=0; i<npts-1; i++)
      {
      points->GetPoint(linePts[i],p1);
      points->GetPoint(linePts[(i+1)],p2);
      f[i].ds = sqrt(vtkMath::Distance2BetweenPoints(p1,p2));
      f[i].T[0] = (p2[0]-p1[0])/f[i].ds;
      f[i].T[1] = (p2[1]-p1[1])/f[i].ds;
      f[i].T[2] = (p2[2]-p1[2])/f[i].ds;
      }
    for (i=0; i<npts-1; i++)
      {
      vtkMath::Cross(f[i].T, f[i+1].T, f[i].B);
      vtkMath::Normalize(f[i].B);
      }
    for (i=0; i<npts-1; i++)
      {
      
      vtkMath::Cross(f[i].B, f[i].T, f[i].N);
     
      kappa = sqrt(vtkMath::Distance2BetweenPoints(f[i+1].T,f[i].T))/f[i].ds;
      vtkDebugMacro(<< "Kappa: " << kappa);
      BB[0] = f[i+1].B[0] - f[i].B[0];
      BB[1] = f[i+1].B[1] - f[i].B[1];
      BB[2] = f[i+1].B[2] - f[i].B[2];
      tau = -vtkMath::Dot(BB, f[i].N)/f[i].ds;
      vtkDebugMacro(<< "Tau: " << tau);
      curvature->SetValue(linePts[i],kappa);
      torsion->SetValue(linePts[i],tau);
      }
      
    curvature->SetValue(linePts[npts-1],kappa);
    torsion->SetValue(linePts[npts-1],tau);     
    
    delete [] f;
    }
    
    output->GetPointData()->AddArray(curvature);
    output->GetPointData()->AddArray(torsion);
    if (this->ScalarOutput == VTK_CURVATURE)
      {
      output->GetPointData()->SetActiveScalars("Curvature");
      }
    else
      {
      output->GetPointData()->SetActiveScalars("Torsion");
      }
}

void vtkPolyLineCurvature::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os,indent);
  
//  os << indent << "Active Scalar:" << this->NeighborhoodSize << "\n";

}

