/******************* vtkUtils.c *********************/

#include "vtk_Utils.h"

/*===<VectorsToVTKPolyData.c>==================================================

     Title: convert Vectors To VTKPolyData for plotting

   Purpose: Input x and y axes vectors and convert to form
            appropriate for vtkXYPlotActor2D

      Call: vtkPolyData *VectorsToVTKData(float *pos, float *val, int numPts,
                                          float *xrange, float *yrange ) 
      
     Input: val = value (y-value)
            pos = position (x-value)
	    numPts = length of val, pos

    Output: xrange = range of x-values (xrange[0] = min, xrange[1] = max)
            yrange = range of y-values

    Return: vtkPolyData *

   History: 03_06_30 - lrf
*/
vtkPolyData *VectorsToVTKPolyData(vtkFloatingPointType *pos, 
				  vtkFloatingPointType *val, int numPts,
				  vtkFloatingPointType *xrange,
				  vtkFloatingPointType *yrange ) 
{
  float point[3] = {0,0,0};
  vtkPolyData *data = vtkPolyData::New();

  vtkPoints *positions = vtkPoints::New();
    positions->SetNumberOfPoints(numPts);

  vtkFloatArray *values = vtkFloatArray::New();
    values->SetNumberOfTuples(numPts);

  vtkFloatArray *pos1D = vtkFloatArray::New(); /* for finding x range */
    pos1D->SetNumberOfTuples(numPts);

  for (int i=0; i<numPts; i++)
    {
      point[0] = pos[i];
      positions->SetPoint(i,point);
      pos1D->SetValue(i,pos[i]); /* for finding x range */
      values->SetValue(i,val[i]);
    }
  data->SetPoints(positions);
  data->GetPointData()->SetScalars(values);

  // find ranges
  pos1D->GetRange(xrange,0) ;
  values->GetRange(yrange,0) ;

  // clean up
  positions->Delete(); 
  values->Delete();
  pos1D->Delete();

  return (data) ;
}

/*===<FloatArrayRange.c>====================================================

   input: vtkFloatArray[ntuple][ncomp]
  output: vtkFloatArray[ncomp][2]

*/

vtkFloatArray *FloatArrayRange( vtkFloatArray *fa )
{
  int ncomp = fa->GetNumberOfComponents() ;
  int ntupl = fa->GetNumberOfTuples() ;

  vtkFloatArray *range = vtkFloatArray::New(); 
  range->SetNumberOfComponents(2);
  range->SetNumberOfTuples(ncomp);

  vtkFloatingPointType *x = (vtkFloatingPointType *) 
     calloc((unsigned)ncomp,sizeof(vtkFloatingPointType));
  vtkFloatingPointType y[2] ;

  y[0] =  BIGVAL ;
  y[1] = -BIGVAL ;
  for (int j=0; j<ncomp; j++) range->SetTuple(j,y) ; 

  for (int i=0; i<ntupl; i++) {

    fa->GetTuple(i,x) ;

    for (int j=0; j<ncomp; j++) {

      range->GetTuple(j,y) ;
      if( x[j] <= y[0] ) y[0] = x[j] ;
      if( x[j] >  y[1] ) y[1] = x[j] ;
      range->SetTuple(j,y) ; 

    }
  }

  free(x) ;

  return (range) ;
}

/*===<FloatArray_Sum.c>========================================================

     Title: FLOATARRAY SUM

   Purpose: Sum over all ntupl ntuples for
            each of the ncomp components

      Call: vtkFloatArray *FloatArray_Sum( vtkFloatArray *meanfa , vtkFloatArray *fa )
      
     Input: vtkFloatArray fa[ntupl][ncomp]

    Output: vtkFloatArray meanfa[ncomp]

    Return: vtkFloatArray *meanfa

   History: 03_01_28 - lrf
*/
vtkFloatArray *FloatArray_Sum( vtkFloatArray *meanfa , vtkFloatArray *fa )
{
  int ncomp = fa->GetNumberOfComponents() ;
  int ntupl = fa->GetNumberOfTuples() ;

  complex *zv = zv_alloc(ntupl);

  vtkFloatingPointType *x  = (vtkFloatingPointType *) 
     calloc((unsigned)ntupl,sizeof(vtkFloatingPointType));
  vtkFloatingPointType *y  = (vtkFloatingPointType *) 
     calloc((unsigned)ncomp,sizeof(vtkFloatingPointType));

  for (int j=0; j<ncomp; j++) {
    for (int i=0; i<ntupl; i++) {
      fa->GetTuple(i,x) ;
    }
    y[j] = 0;
    for (int i=0; i<ntupl; ++i)
    {
       y[j] += x[i];
    }
  }

  free(x) ;

  return(meanfa) ;
}

/*===<Fill_PComplexArray.c>===================================================

  ncomp must be even below or watch out!
  note that the components are complex, 
  but the tuples are real (so to speak, in that their indices count correctly.)

*/
void Fill_PComplexArray( vtkFloatArray *fa , float fvalr, float fvali )
{
  int ncomp = fa->GetNumberOfComponents() ; // 2 * number of actual components
  int ntupl = fa->GetNumberOfTuples() ;

  vtkFloatingPointType *x  = (vtkFloatingPointType *) 
     calloc((unsigned)ncomp,sizeof(vtkFloatingPointType));

  int ncplx = ncomp/2 ;
  for (int i=0; i<ntupl; i++) {
    for (int j=0; j<ncplx; j++) {
      x[2*j]   = fvalr ;
      x[2*j+1] = fvali ;
    }
    fa->SetTuple(i,x) ;
  }

  free(x) ;
}

/*===<Fill_FloatArray.c>========================================================

     Title: FILL FLOATARRAY 

   Purpose: Fill up to maxcomp components of float array

      Call: void Fill_FloatArray( vtkFloatArray *fa, float fval, int maxcomp )
      
     Input: vtkFloatArray fa[ntupl][ncomp]

    Output: vtkFloatArray fa[ntupl][ncomp]

    Return: 

   History: 03_02_03 - lrf

*/
void Fill_FloatArray( vtkFloatArray *fa, float fval, int maxcomp )
{
  int ncomp = fa->GetNumberOfComponents() ;
  int ntupl = fa->GetNumberOfTuples() ;

  if(maxcomp<0) maxcomp = ncomp ;

  vtkFloatingPointType *x = (vtkFloatingPointType *) 
     calloc((unsigned)ncomp,sizeof(vtkFloatingPointType));
  vtkFloatingPointType *y = (vtkFloatingPointType *) 
     calloc((unsigned)ncomp,sizeof(vtkFloatingPointType));

  for (int i=0; i<ntupl; i++) {
    fa->GetTuple(i,y) ;
    for (int j=0;       j<maxcomp; j++) x[j] = fval ;
    for (int j=maxcomp; j<ncomp;   j++) x[j] = y[j] ;
    fa->SetTuple(i,x) ;
  }

  free(x) ;
  free(y) ;
}

/*===<Clear_FloatArray.c>====================================================*/

void Clear_FloatArray( vtkFloatArray *fa )
{
  Fill_FloatArray( fa, 0.0, 0 );
}

/*===<Print_FloatArray.c>====================================================*/

void Print_FloatArray( vtkFloatArray *fa , char *name, int pmode )
{
  int ncomp = fa->GetNumberOfComponents() ;
  int ntupl = fa->GetNumberOfTuples() ;

  vtkFloatingPointType *x  = (vtkFloatingPointType *) 
     calloc((unsigned)ncomp,sizeof(vtkFloatingPointType));

  for (int i=0; i<ntupl; i++) {
    fa->GetTuple(i,x) ;
    if(pmode==0) printf("\n");
    for (int j=0; j<ncomp; j++) {
      printf("%s[%2i][%2i] = %10.7g",name,i,j,x[j]) ;
      if(pmode==0) if(j!=(ncomp-1)&&ncomp>1) printf(",\n");
      if(pmode==1) if(j!=(ncomp-1)&&ncomp>1) printf(", ");
    }
    printf("\n");
  }

  free(x) ;
}
