/*
 *  	plug_diff.c
 *
 *	Larry Frank's Diffusion Toolbox
 *      A plugin application for Bob Cox's AFNI 
 *      (http://afni.nimh.nih.gov/afni)
 *  
 *      Lawrence R. Frank, PhD
 *      Department of Radiology
 *      UCSD Center for Functional MRI and
 *      VA San Diego Healthcare System 
 *      email: lfrank@ucsd.edu  
 *      phone: 858-822-0538
 *
 *      Copyright 2000-2004 by the
 *      Regents of the University of California. 
 *      All rights reserved.
 *
 *      Compiling this toolbox into your version of AFNI requires:
 *
 *      1) "tar xf plug_diff.tar" into your application directory ${APPDIR}
 *         (designed to be separate from any AFNI directories)
 *      2) Run the script ${APPDIR}/afni/diff/makelibs* 
 *         (this compiles all the libraries)
 *      3) cp ${APPDIR}/afni/diff/Makefile.lrf.(system) ${AFNI_PATH}/Makefile
 *      4) In the AFNI directory ${AFNI_PATH}, type
 *         "make lrf_plugins install_lrf_plugins"
 *         (Assuming AFNI is already compiled.  If not, type "make itall",
 *          then the line above)
 *      5) Fire up AFNI and look for the plugin called "Diffusion"
 *
 *      This toolkit uses the following stuff:
 *
 *           Source code:
 *
 *                 plug_diff.c, plug_diff.h (the main interface and header)
 *                 ~./diffuse = main diffusion routines
 *                 ~./utils   = general utilities
 *                 ~./vector  = vector math
 *                 ~./splash  = display tidbits
 *                 ~./docs    = documents
 *                 ~./vtk     = VTK routines
 *
 *            Libraries:
 *
 *                   ~./lib   = library lrftools.a (diffuse,vector)
 *                makelibs*   = script to generate lrftools library
 *               makeclean*   = clean up lrftools library
 *
 *            Utilities:
 *
 *                 ~./d2a     = diffusion data (from ses.psd) to afni conversion
 *                              (currently for UCSD formatted data only)
 *
 *                Setup: scripts of the form setup(system) are included
 *                       for one-stop shopping.
 *                       They require definition of ${AFNI_PATH} where
 *                       the AFNI source code sits:
 *
 *             setupLinux_g++ = sets up make for Linux system using g++
 *
 *             Usage:  1) type setupLinux_g++
 *                     2) cd ${AFNI_PATH}
 *                     3) make lrf_plugins install_lrf_plugins
 *
 *                   Currently "system" = "sgi","linux", or "mac"
 *                                  where   sgi = SGI  (IRIX 6.5, R10000 chip)
 *                                   and  linux = Linux ???
 *                              These are the only three operating systems that
 *                              I use, so that's all you get...
 * 
 *    History:  00_09_15  copy RWC's plug_render.c -> plug_diff.c
 *
 *
 * Make polytope consistent with hyper.e! (same as ses13 right now)
 *
 */

/* include files */

#include "plug_diff.h"

#include "mcw_graf.h"
#include "parser.h"
#include <ctype.h>
#include "dbtrace.h"
#include <assert.h>

#include "./splash/quotes.h"

#ifndef ALLOW_PLUGINS
#  error "Plugins not properly set up -- see machdep.h"
#endif

#undef DIFF_DEBUG
#define ONLY_AXIAL

#define TMP_EXIT {fprintf(stderr,"Temporary exit from program %s:\n",program_name);exit(0);}

/***********************************************************************
  Plugin to diffuse a volume dataset.  Makes a custom interface.
  Modified from plug_drawdset.c by RWCox - February 1999.
************************************************************************/

/*---------- prototypes for internal routines ----------*/

char * DIFF_main( PLUGIN_interface * ) ;                 /* called from AFNI */

void DIFF_make_widgets(void) ;                           /* initialization */

void DIFF_done_CB   ( Widget , XtPointer , XtPointer ) ; /* Done button */
void DIFF_draw_CB   ( Widget , XtPointer , XtPointer ) ; /* Draw button */
void DIFF_help_CB   ( Widget , XtPointer , XtPointer ) ; /* Help button */
void DIFF_reload_CB ( Widget , XtPointer , XtPointer ) ; /* Reload button */
void DIFF_choose_CB ( Widget , XtPointer , XtPointer ) ; /* Choose button */
void DIFF_xhair_CB  ( Widget , XtPointer , XtPointer ) ; /* See Xhairs toggle */
void DIFF_dynamic_CB( Widget , XtPointer , XtPointer ) ; /* DynaDraw toggle */
void DIFF_accum_CB  ( Widget , XtPointer , XtPointer ) ; /* Accumulate toggle */
void DIFF_angle_CB  ( MCW_arrowval * , XtPointer ) ;     /* Angle arrowvals */
void DIFF_param_CB  ( MCW_arrowval * , XtPointer ) ;     /* Cutout arrowvals */
void DIFF_precalc_CB( MCW_arrowval * , XtPointer ) ;     /* Precalc menu */
void DIFF_clip_CB   ( MCW_arrowval * , XtPointer ) ;     /* Clip arrowvals */

void DIFF_copyright_CB   ( Widget , XtPointer , XtPointer ) ; /* Copyright notice*/

void DIFF_xhair_recv( int,int , int *, void * ) ;        /* 29 Mar 1999 */

void   DIFF_choose_av_CB      ( MCW_arrowval * , XtPointer ) ; /* Sub-brick menus */
char * DIFF_choose_av_label_CB( MCW_arrowval * , XtPointer ) ;
void   DIFF_opacity_scale_CB  ( MCW_arrowval * , XtPointer ) ;

void DIFF_finalize_dset_CB( Widget , XtPointer , MCW_choose_cbs * ) ; /* dataset chosen */

int DIFF_reload_dataset(void) ; /* actual reloading work */
void DIFF_xhair_overlay(void)  ; /* make the crosshairs */

/* 31-Aug-02 */
DSET_stats *DIFF_dset_stats( THD_3dim_dataset * , THD_3dim_dataset *  ,
			     int , int , int , float , float  ) ;

DSET_stats *b0_dstats ;	/* stats of b=0 image.  31-Aug-02 */
DSET_stats *ftm_anat_dstats ;	/* stats of b=0 image.  31-Aug-02 */

/*----------------------------------------------------------------*/

static PLUGIN_interface * plint = NULL ;                 /* what AFNI sees */

void DIFF_textact_CB( Widget , XtPointer , XtPointer ) ; /* press Enter in a textfield */

/***********************************************************************
   Set up the interface to the user.  Note that we bypass the
   normal interface creation, and simply have the menu selection
   directly call the main function, which will create a custom
   set of interface widgets.
************************************************************************/
extern "C" PLUGIN_interface * PLUGIN_init( int ncall )
{
   /*void *handle;*/

   if( ncall > 0 ) return NULL ;  /* only one interface */

#if 0
   handle = dlopen("libvtkImaging.so", RTLD_LAZY);
   if( !handle ) { fprintf(stderr, "%s\n", dlerror()); return NULL; }
   handle = dlopen("libvtkGraphics.so", RTLD_LAZY);
   if( !handle ) { fprintf(stderr, "%s\n", dlerror()); return NULL; }
   handle = dlopen("libvtkCommon.so", RTLD_LAZY);
   if( !handle ) { fprintf(stderr, "%s\n", dlerror()); return NULL; }
   handle = dlopen("libvtkRendering.so", RTLD_LAZY);
   if( !handle ) { fprintf(stderr, "%s\n", dlerror()); return NULL; }
   handle = dlopen("libvtkFiltering.so", RTLD_LAZY);
   if( !handle ) { fprintf(stderr, "%s\n", dlerror()); return NULL; }
#endif

   plint = PLUTO_new_interface( "Diffusion" , NULL , NULL ,
                                PLUGIN_CALL_IMMEDIATELY ,
				reinterpret_cast<char * (*)()>(DIFF_main) ) ;

   PLUTO_add_hint( plint , "Diffusion Analysis" ) ;

   PLUTO_set_sequence( plint , "A:graphics" ) ;


   return plint ;
}

/***************************************************************************
                          Internal data structures
****************************************************************************/

#define NO_DATASET_STRING "[No Dataset is Loaded]"

/* Interface widgets */

static Widget shell=NULL , anat_rowcol , info_lab , choose_pb ;
static Widget done_pb , help_pb , draw_pb , reload_pb, copyright_pb ;
static MCW_arrowval * roll_av , * pitch_av , * yaw_av , * precalc_av ;
static MCW_bbox * xhair_bbox , * dynamic_bbox , * accum_bbox ;
static MCW_arrowval * choose_av , * opacity_scale_av ;

static char *DIFF_dummy_av_label[2] = { "[Nothing At All]" , "[Nothing At All]" } ;

static Widget top_rowcol , anat_frame ;

#define CLIP_RANGE 32767

static Widget range_lab ;
static MCW_arrowval * clipbot_av , * cliptop_av ;

static float  brickfac = 0.0 ;
static Widget range_faclab , clipbot_faclab , cliptop_faclab ;

void DIFF_graf_CB( MCW_graf * , void * ) ;
static MCW_graf    * opa_graf ;
static MCW_graf    * gry_graf ;
static MCW_pasgraf * his_graf ;

static char * xhair_bbox_label[1]   = { "See Xhairs" } ;
static char * dynamic_bbox_label[1] = { "DynaDraw"   } ;
static char * accum_bbox_label[1]   = { "Accumulate" } ;

/*----------------------------------------------------------------*/

/* Other data */

#define MODE_LOW       0
#define MODE_MEDIUM    1
#define MODE_HIGH      2

#define NUM_precalc    3

static char * precalc_strings[] = { "Low" , "Medium" , "High" } ;
/* static int    precalc_mode[]    = { PMODE_LOW,PMODE_MEDIUM,PMODE_HIGH } ; */

static MCW_DC * dc ;                   /* display context */
static Three_D_View * im3d ;           /* AFNI controller */
static THD_3dim_dataset * dset ;       /* The dataset!    */
static MCW_idcode         dset_idc ;   /* 31 Mar 1999     */
static THD_3dim_dataset *ftm_anat_dset ; /* 31-Aug-04    */
static MCW_idcode        ftm_anat_dset_idc ;   /* 31-Aug-04 */
static int new_dset = 0 ;              /* Is it new?      */
static int dset_ival = 0 ;             /* Sub-brick index */
static char dset_title[THD_MAX_NAME] ; /* Title string */

static MRI_IMAGE * grim=NULL ,
                 * opim=NULL ;      /* volumes to diffuse */
static void * diffuse_handle ;       /* diffusing struct */

static MRI_IMAGE * grim_showthru=NULL ,
                 * opim_showthru=NULL ;  /* 07 Jan 2000 */

#define FREEIM(x) if( (x) != NULL ){ mri_free(x); (x)=NULL; }

#define FREE_VOLUMES                                  \
  do{ FREEIM(grim) ; FREEIM(opim) ;                   \
      FREEIM(grim_showthru); FREEIM(opim_showthru) ; } while(0) ;

#define NEED_VOLUMES (grim == NULL || opim == NULL)

#ifdef SGI
static int   precalc_ival   = MODE_LOW    ;
#else
static int   precalc_ival   = MODE_MEDIUM ;
#endif

// temporary to check max value for montage display
static double max_magnitude = 600.0;

static int   dynamic_flag   = 0      ;
static int   accum_flag     = 0      ;
static float angle_roll     =   70.0 ;
static float angle_pitch    =  120.0 ;
static float angle_yaw      =    0.0 ;

static int xhair_flag  = 0    ;
static int xhair_ixold = -666 ;  /* remember the past */
static int xhair_jyold = -666 ;
static int xhair_kzold = -666 ;
static int xhair_omold = -666 ;  /* 02 Jun 1999 */

static int xhair_recv  = -1 ;    /* 29 Mar 1999 */

#define CHECK_XHAIR_MOTION ( im3d->vinfo->i1             != xhair_ixold || \
                             im3d->vinfo->j2             != xhair_jyold || \
                             im3d->vinfo->k3             != xhair_kzold || \
                             im3d->vinfo->xhairs_orimask != xhair_omold   )

static int new_data_loaded = 0 ;
static int diffuser_open   = 0 ;
static int npixels = 0 ;

/*------------------  Stuff for the image display window -----------------*/

static MCW_imseq * imseq      = NULL ;
static MRI_IMARR * diffusings = NULL ;

void DIFF_open_imseq( void ) ;
void DIFF_update_imseq( void ) ;
void DIFF_destroy_imseq( void ) ;
XtPointer DIFF_imseq_getim( int , int , XtPointer ) ;
void DIFF_seq_send_CB( MCW_imseq * , XtPointer , ISQ_cbs * ) ;

/*---------------- Stuff for Automate mode --------------*/

static MCW_bbox * automate_bbox ;

static int automate_flag = 0 ;
static char * automate_bbox_label[1]   = { "Automate" } ;

static MCW_arrowval * autoframe_av ;
static Widget autocompute_pb , autocancel_pb ;

void DIFF_autoflag_CB   (Widget , XtPointer , XtPointer) ; /* Automate toggle */
void DIFF_autocompute_CB(Widget , XtPointer , XtPointer) ; /* Compute pushbutton */
void DIFF_autocancel_CB (Widget , XtPointer , XtPointer) ; /* Cancel pushbutton */

/*-------------------------- Stuff for cutout logic ----------------------*/

#define MAX_CUTOUTS 9		/* need to keep for now */
#define MIN_OPACITY_SCALE 0.000

static char * mustdo_bbox_label[1] = { "Must Do" } ;

/*------------------  lrf's stuff for diffusion theory -----------------*/

char diff_choose_str[THD_MAX_NAME] ;

void DIFF_theory_CB( MCW_arrowval * , XtPointer ) ; /* Dif Method menu */

static MCW_arrowval *theory_av ;

#define THEORY_DTI   0
#define THEORY_SDV   1
#define THEORY_SHD   2
#define THEORY_FTM   3
#define THEORY_DST   4
#define THEORY_DTE   5

#define NUM_theory   6

static char * theory_strings[] = { "Diffusion Tensor Model", 
				   "Anisotropy in Multifiber Voxels" , 
				   "Spherical Harmonic Decomposition",
				   "Fiber Tract Mapping",
				   "Diffusion Simulation Fiber Tractography",
				   "Diffusion Tensor Estimation"};

static int theory_val = THEORY_DTI ;
static int show_theory_flag = 0 ;

static Widget show_theory_pb ;
static Widget show_vtkvis_pb ;

void DIFF_show_theory_CB(Widget , XtPointer , XtPointer) ; 
void DIFF_show_vtkvis_CB(Widget , XtPointer , XtPointer) ; 

/*-----------------   lrf's anisotropy stuff   ------------*/

char routine_name2[256] ;

static int dwi2dt_flag = 0;
static int dti_flag = 0 ;
static int shd_flag = 0 ;
static int swd_flag = 0 ;
static int ftm_flag = 0 ;

static int shd_scalars_flag = 0 ;
static int swd_scalars_flag = 0 ;
static int dti_scalars_flag = 0 ;
static int ftm_scalars_flag = 0 ;

static int dti_display_flag = 0 ;
static int shd_display_flag = 0 ;
static int swd_display_flag = 0 ;
static int ftm_display_flag = 0 ;

static int pos_mcalc_flag = 1 ;

DIF_state *current_dif_state = NULL;

void DIFF_load_shd_state(void) ;                /* load from widgets */
int  DIFF_shd_state_changed(void) ;             /* has it changed? */

static int shd_grid_fstep = 50 ;

/*-----------------   lrf's ftm stuff   ------------*/

static Widget ftm_choose_pb ;
static Widget ftm_anat_choose_pb ;

#ifdef EXPERIMENTAL
static int ftm1_flag = 0 ;
#else
static int ftm1_flag = 1;
#endif


static int ftm_dst_flag = 1;


static int ftm2_flag = 0 ;

FTM_state *current_ftm_state, *old_ftm_state ;

static int ftm_grid_fstep = 50 ;

/*-----------------   stuff for evaluation of expressions   ------------*/

static double atoz[26] ;  /* values of 'a', 'b', ..., 'z' in expressions */

#define N_IND  13  /* 'n' */
#define T_IND  19  /* 't' */
#define X_IND  23  /* 'x' */
#define Y_IND  24  /* 'y' */
#define Z_IND  25  /* 'z' */

float DIFF_evaluate( MCW_arrowval * ) ;

/*-------------------- Icon Pixmap for Image Window ----------------------*/

static Pixmap afni48ren_pixmap = XmUNSPECIFIED_PIXMAP ;
#define afni48ren_width 48
#define afni48ren_height 48                /* from file afni48ren.xbm */
static char afni48ren_bits[] = {
   0xff, 0xff, 0xc1, 0xc1, 0xff, 0xff, 0xff, 0x7f, 0x60, 0x00, 0xfe, 0xff,
   0xff, 0x0f, 0x30, 0x10, 0xf0, 0xff, 0xff, 0x01, 0x37, 0xf0, 0x80, 0xff,
   0x7f, 0xe0, 0x77, 0xe0, 0x07, 0xff, 0x7f, 0xfe, 0xe0, 0x00, 0x3f, 0xff,
   0x7f, 0x1e, 0xc0, 0x03, 0x38, 0xff, 0x3f, 0x00, 0x87, 0xe7, 0x01, 0xff,
   0x3f, 0xf0, 0x07, 0xe7, 0x0f, 0xfe, 0x3f, 0x7f, 0xc0, 0x04, 0x7e, 0xfe,
   0x3f, 0x0f, 0xe6, 0x67, 0x70, 0xfe, 0x3f, 0xe0, 0xa7, 0xe7, 0x03, 0xfe,
   0x1f, 0xfc, 0x21, 0x83, 0x3f, 0xfc, 0x9f, 0x1f, 0xe0, 0x00, 0xfc, 0xfc,
   0x9f, 0x83, 0xc7, 0xe1, 0xc1, 0xfc, 0x1f, 0xf0, 0x87, 0xe7, 0x0f, 0xfc,
   0x0f, 0x7f, 0x00, 0x07, 0x7e, 0xf8, 0xcf, 0x0f, 0xc0, 0x04, 0xf0, 0xf9,
   0xcf, 0x80, 0xe7, 0xe7, 0x81, 0xf9, 0x0f, 0xf0, 0xa7, 0xe3, 0x1f, 0xf8,
   0x0f, 0x7f, 0xe0, 0x00, 0xfe, 0xf9, 0xcf, 0x0f, 0xc0, 0x01, 0xe0, 0xf1,
   0xc7, 0x00, 0x87, 0xe3, 0x00, 0xf0, 0x07, 0xf8, 0x07, 0xe7, 0x1f, 0xf0,
   0x87, 0xff, 0xc0, 0x86, 0xff, 0xf1, 0xe7, 0x07, 0xe0, 0x03, 0xe0, 0xf3,
   0x77, 0x00, 0xe0, 0x00, 0x00, 0xe6, 0x03, 0x00, 0xc0, 0x01, 0x00, 0xe0,
   0x03, 0x00, 0x80, 0x03, 0x00, 0xe0, 0x03, 0xf8, 0x0f, 0xf8, 0x1f, 0xe0,
   0x81, 0xff, 0x3f, 0xfc, 0xff, 0xc1, 0xff, 0xff, 0x7f, 0xfe, 0xff, 0xff,
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x01, 0xf8, 0xff, 0xff, 0x1f, 0xc0,
   0xff, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x31, 0x74, 0x31, 0xc4, 0xe8, 0xc6,
   0xad, 0x67, 0xad, 0xb7, 0xcd, 0xda, 0xad, 0x57, 0xad, 0xb7, 0xad, 0xfa,
   0x31, 0x56, 0x2d, 0xc6, 0xad, 0xfa, 0xb5, 0x57, 0xad, 0xd7, 0xad, 0x8a,
   0xad, 0x37, 0xad, 0xb7, 0x6d, 0xda, 0xad, 0x77, 0xad, 0xb7, 0xed, 0xda,
   0x2d, 0x74, 0x31, 0xb4, 0xe8, 0xc6, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xab, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa};

/*-------------------- SHD overlay stuff ----------------------*/

int REPLICATE ;

#define INVALIDATE_OVERLAY do{ FREEIM(ovim) ; } while(0)

#define DO_OVERLAY   (func_dset != NULL && func_see_overlay)

#define TURNOFF_SHD_OVERLAY_WIDGETS                                   \
  do{ XmString xstr ;                                              \
      xstr = XmStringCreateLtoR( NO_DATASET_STRING ,                \
                                 XmFONTLIST_DEFAULT_TAG ) ;          \
      XtVaSetValues( wshdfunc_info_lab , XmNlabelString,xstr , NULL ) ;  \
      XmStringFree(xstr) ;                                             \
                                                                        \
      xstr = DIFF_range_label() ;                                        \
      XtVaSetValues( wshdfunc_range_label , XmNlabelString , xstr , NULL ) ; \
      XmStringFree(xstr) ;                                                 \
                                                                            \
      xstr = DIFF_autorange_label() ;                                        \
      XtVaSetValues( wshdfunc_range_bbox->wbut[0], XmNlabelString,xstr, NULL ) ; \
      XmStringFree(xstr) ;                                                    \
                                                                             \
      AV_SENSITIZE( wshdfunc_color_av  , False ) ;                             \
      AV_SENSITIZE( wshdfunc_thresh_av , False ) ;                            \
  } while(0)

#define TURNOFF_DTI_OVERLAY_WIDGETS                                   \
  do{ XmString xstr ;                                              \
      xstr = XmStringCreateLtoR( NO_DATASET_STRING ,                \
                                 XmFONTLIST_DEFAULT_TAG ) ;          \
      XtVaSetValues( wdtifunc_info_lab , XmNlabelString,xstr , NULL ) ;  \
      XmStringFree(xstr) ;                                             \
                                                                        \
      xstr = DIFF_range_label() ;                                        \
      XtVaSetValues( wdtifunc_range_label , XmNlabelString , xstr , NULL ) ; \
      XmStringFree(xstr) ;                                                 \
                                                                            \
      xstr = DIFF_autorange_label() ;                                        \
      XtVaSetValues( wdtifunc_range_bbox->wbut[0], XmNlabelString,xstr, NULL ) ; \
      XmStringFree(xstr) ;                                                    \
                                                                             \
      AV_SENSITIZE( wdtifunc_color_av  , False ) ;                             \
      AV_SENSITIZE( wdtifunc_thresh_av , False ) ;                            \
  } while(0)

#define TURNOFF_FTM_OVERLAY_WIDGETS                                   \
  do{ XmString xstr ;                                              \
      xstr = XmStringCreateLtoR( NO_DATASET_STRING ,                \
                                 XmFONTLIST_DEFAULT_TAG ) ;          \
      XtVaSetValues( wftmfunc_info_lab , XmNlabelString,xstr , NULL ) ;  \
      XmStringFree(xstr) ;                                             \
                                                                        \
      xstr = DIFF_range_label() ;                                        \
      XtVaSetValues( wftmfunc_range_label , XmNlabelString , xstr , NULL ) ; \
      XmStringFree(xstr) ;                                                 \
                                                                            \
      xstr = DIFF_autorange_label() ;                                        \
      XtVaSetValues( wftmfunc_range_bbox->wbut[0], XmNlabelString,xstr, NULL ) ; \
      XmStringFree(xstr) ;                                                    \
                                                                             \
      AV_SENSITIZE( wftmfunc_color_av  , False ) ;                             \
      AV_SENSITIZE( wftmfunc_thresh_av , False ) ;                            \
  } while(0)

void DIFF_func_dti_widgets(void) ;
void DIFF_func_shd_widgets(void) ;
void DIFF_func_ftm_widgets(void) ;
void DIFF_init_cmap(void) ;
void DIFF_reload_func_dset( Three_D_View *, int ) ;
void DIFF_reload_ftm_anat_dset( Three_D_View *, int ) ;
void DIFF_reload_diffuser(void) ;

/* DIF Overlay stuff */

static MCW_bbox *aniso_compute_bbox ;
static MCW_bbox *dif_overlay_bbox ;
static Widget    dif_overlay_label ;
static Widget    dif_overlay_pb ;

void DIFF_overlay_CB( Widget , XtPointer , XtPointer ) ;
void DIFF_overlay_chan_CB( Widget , XtPointer , XtPointer ) ;
void DIFF_aniso_compute_chan_CB( Widget , XtPointer , XtPointer ) ;
static int dif_overlay_flag = 0 ;
static int aniso_compute_flag = 0 ;

/*-------------------- DTI overlay stuff ----------------------*/

static char *DIFF_dti_scalar_av_label[N_DTI_SCALAR_METH] = { "Ave Diff Coeff" ,
							     "Frac Aniso Index" ,
							     "Rel Aniso Index" ,
							     "Lattice Aniso Index" ,
							     "Spher Diff Var" } ;

static MCW_arrowval *wdtifunc_scalar_av ;
static Widget        wdtifunc_open_pb ;

void DIFF_open_dtifunc_CB( Widget , XtPointer , XtPointer ) ;

static Widget wdtifunc_frame=NULL , wdtifunc_rowcol , wdtifunc_choose_pb ,
              wdtifunc_uber_rowcol , wdtifunc_info_lab , wdtifunc_vsep ;

static Widget wdtifunc_thr_rowcol , wdtifunc_thr_label , wdtifunc_thr_scale=NULL ,
              wdtifunc_thr_pval_label ;
static MCW_arrowval * wdtifunc_thr_top_av ;

static Widget wdtifunc_color_rowcol , wdtifunc_color_label ;
static MCW_pbar * wdtifunc_color_pbar=NULL ;
static MCW_arrowval * wdtifunc_color_av , * wdtifunc_thresh_av , * wdtifunc_colornum_av ;
static MCW_bbox * wdtifunc_color_bbox ;

static Widget wdtifunc_choices_rowcol , wdtifunc_choices_label ,
              wdtifunc_buck_frame , wdtifunc_buck_rowcol ,
              wdtifunc_opacity_frame , wdtifunc_opacity_rowcol ,
              wdtifunc_range_rowcol , wdtifunc_range_frame ;

static Widget wdtifunc_range_label ;
static MCW_arrowval * wdtifunc_opacity_av , * wdtifunc_range_av ;
static MCW_bbox * wdtifunc_see_overlay_bbox , * wdtifunc_range_bbox ;

static Widget wdtifunc_pbar_menu , wdtifunc_pbar_equalize_pb , wdtifunc_pbar_settop_pb ;
static MCW_arrowval * wdtifunc_pbar_palette_av ;
static MCW_arrowval * wdtifunc_pbar_mixshade_av ;  /* 21 Dec 1999 */

/*-------------------- SHD overlay stuff ----------------------*/

static char *DIFF_shd_scalar_av_label[N_SHD_SCALAR_METH] = { "Fiber channels" ,
							     "Even/Odd" ,
							     "FMI" ,
							     "Gen Rel Aniso Index",
							     "Gen Frac Aniso Index"} ;

static MCW_arrowval *wshdfunc_scalar_av ;
static Widget wshdfunc_open_pb ;
void DIFF_open_shdfunc_CB( Widget , XtPointer , XtPointer ) ;

static Widget wshdfunc_frame=NULL , wshdfunc_rowcol , wshdfunc_choose_pb ,
              wshdfunc_uber_rowcol , wshdfunc_info_lab , wshdfunc_vsep ;

static Widget wshdfunc_thr_rowcol , wshdfunc_thr_label , wshdfunc_thr_scale=NULL ,
              wshdfunc_thr_pval_label ;
static MCW_arrowval * wshdfunc_thr_top_av ;

static Widget wshdfunc_color_rowcol , wshdfunc_color_label ;
static MCW_pbar *wshdfunc_color_pbar=NULL ;
static MCW_arrowval *wshdfunc_color_av, *wshdfunc_thresh_av, *wshdfunc_colornum_av ;
static MCW_bbox * wshdfunc_color_bbox ;

static Widget wshdfunc_choices_rowcol , wshdfunc_choices_label ,
              wshdfunc_buck_frame , wshdfunc_buck_rowcol ,
              wshdfunc_opacity_frame , wshdfunc_opacity_rowcol ,
              wshdfunc_range_rowcol , wshdfunc_range_frame ;

static Widget wshdfunc_range_label ;
static MCW_arrowval *wshdfunc_opacity_av , *wshdfunc_range_av ;
static MCW_bbox *wshdfunc_lmview_see_overlay_bbox,*wshdfunc_scalars_see_overlay_bbox, *wshdfunc_range_bbox ;

static Widget wshdfunc_pbar_menu , wshdfunc_pbar_equalize_pb , wshdfunc_pbar_settop_pb ;
static MCW_arrowval * wshdfunc_pbar_palette_av ;
static MCW_arrowval * wshdfunc_pbar_mixshade_av ;  /* 21 Dec 1999 */

/*-------------------- FTM overlay stuff ----------------------*/

static Widget wftmfunc_open_pb ;
void DIFF_open_ftmfunc_CB( Widget , XtPointer , XtPointer ) ;

static Widget wftmfunc_frame=NULL , wftmfunc_rowcol , wftmfunc_choose_pb ,
              wftmfunc_uber_rowcol , wftmfunc_info_lab , wftmfunc_vsep ;

static Widget wftmfunc_thr_rowcol , wftmfunc_thr_label , wftmfunc_thr_scale=NULL ,
              wftmfunc_thr_pval_label ;
static MCW_arrowval * wftmfunc_thr_top_av ;

static Widget wftmfunc_color_rowcol , wftmfunc_color_label ;
static MCW_pbar * wftmfunc_color_pbar=NULL ;
static MCW_arrowval * wftmfunc_color_av , * wftmfunc_thresh_av , * wftmfunc_colornum_av ;
static MCW_bbox * wftmfunc_color_bbox ;

static Widget wftmfunc_choices_rowcol , wftmfunc_choices_label ,
              wftmfunc_buck_frame , wftmfunc_buck_rowcol ,
              wftmfunc_opacity_frame , wftmfunc_opacity_rowcol ,
              wftmfunc_range_rowcol , wftmfunc_range_frame ;

static Widget wftmfunc_range_label ;
static MCW_arrowval * wftmfunc_opacity_av , * wftmfunc_range_av ;
static MCW_bbox * wftmfunc_see_overlay_bbox , * wftmfunc_range_bbox ;

static Widget wftmfunc_pbar_menu , wftmfunc_pbar_equalize_pb , wftmfunc_pbar_settop_pb ;
static MCW_arrowval * wftmfunc_pbar_palette_av ;
static MCW_arrowval * wftmfunc_pbar_mixshade_av ;  /* 21 Dec 1999 */

/* 20-July-03 */
static MCW_bbox *fiber_compute_bbox ;
void DIFF_fiber_compute_chan_CB( Widget , XtPointer , XtPointer ) ;

/*------------------------ UCSD logo stuff ----------------------*/

Pixmap ucsd_logo_pixmap = XmUNSPECIFIED_PIXMAP ;
Widget ucsd_logo_picture ;
Pixel  ucsd_bg_pix , ucsd_fg_pix  ;

Pixmap corny_pixmap = XmUNSPECIFIED_PIXMAP ;
Widget corny_picture ;
Pixel  corny_bg_pix , corny_fg_pix  ;

/*------------------------    ----------------------*/

static Widget dataset_rowcol ;
static Widget overlay_rowcol ;

/*------------------------ Stuff for DTI viewer ----------------------*/

static Widget wdtifunc_scalars_frame ;
static Widget wdtifunc_scalars_rowcol ;
static Widget wdtifunc_scalars_label ;

MCW_imseq   *sdti_xyz , *sdti_yzx , *sdti_zxy , **sdti_new = NULL ;
MCW_grapher *gdti_xyz , *gdti_yzx , *gdti_zxy , **gdti_new = NULL ;
AFNI_imaging_widgets *dti_iwid = NULL;

/*------------------------ Stuff for SHD viewer ----------------------*/

static MCW_bbox *wshdfunc_lm_view_bbox , *wshdfunc_pos_mview_bbox ;
static MCW_arrowval *wshdfunc_shd_lval_av , *wshdfunc_shd_mval_av ;

static Widget wshdfunc_scalars_frame ;
static Widget wshdfunc_scalars_rowcol ;
static Widget wshdfunc_scalars_label ;

static Widget wshdfunc_lmview_frame ;
static Widget wshdfunc_lmview_rowcol ;
static Widget wshdfunc_lmview_label ;

static int pos_mview_flag = 0 ;

/*------------------------ Stuff for FTM viewer ----------------------*/

static MCW_bbox * wftmfunc_lm_view_bbox , * wftmfunc_pos_mview_bbox ;
static MCW_arrowval * wftmfunc_ftm_lval_av , * wftmfunc_ftm_mval_av ;

static Widget wftmfunc_lmview_frame ;
static Widget wftmfunc_lmview_rowcol ;
static Widget wftmfunc_lmview_label ;

/*------------------------ Stuff for anisotropy logic ----------------------*/

static Widget aniso_compute_pb , aniso_cancel_pb ;

static Widget achoices_rowcol , achoices_label ;
static Widget aniso_frame ;
static Widget aniso_rowcol ;
static Widget aniso_compute_rowcol ;
static Widget shd_choices_rowcol , shd_choices_label ;
static Widget minimal_shd_frame, minimal_shd_rowcol, minimal_shd_label;
static Widget minimal_shd_display_choose_pb, minimal_shd_display_pb;
static MCW_arrowval *dti_lval_av ;
static MCW_arrowval *dti_mval_av ;
static MCW_arrowval *dti_grid_av ;
static MCW_arrowval *shd_lval_av ;
static MCW_arrowval *shd_mval_av ;
static MCW_arrowval *shd_grid_av ;
static MCW_bbox *pos_mcalc_bbox ;
/*static MCW_bbox *dti_bbox , *sdv_bbox , *shd_bbox ;*/

static Widget spharm_frame ;
static Widget spharm_rowcol ;

static Widget aniso_label ;
static Widget spharm_label ;

/*------------------------ Stuff for fiber logic ----------------------*/

static Widget fiber_compute_pb , fiber_cancel_pb ;

static Widget fchoices_rowcol , fchoices_label ;
static Widget fiber_frame ;
static Widget fiber_rowcol ;
static Widget fiber_compute_rowcol ;
static MCW_arrowval *ftm_lval_av ;
static MCW_arrowval *ftm_mval_av ;
static MCW_arrowval *ftm_grid_av ;

static Widget fibtmap_frame ;
static Widget fibtmap_rowcol ;

static Widget fiber_anat_frame ;
static Widget fiber_anat_rowcol ;
static Widget fiber_anat_compute_rowcol ;

static Widget fiber_label ;
static Widget fibtmap_label ;
static Widget fiber_anat_label ;

/*-------------------------- end fiber logic ----------------------*/

extern void DIFF_pbarmenu_CB( Widget , XtPointer , XtPointer ) ;
extern void DIFF_pbarmenu_EV( Widget , XtPointer , XEvent * , Boolean * ) ;
extern void DIFF_palette_av_CB( MCW_arrowval * , XtPointer ) ;
extern void DIFF_mixshade_av_CB( MCW_arrowval * , XtPointer ) ;  /* 21 Dec 1999 */
extern void DIFF_set_pbar_top_CB( Widget , XtPointer , MCW_choose_cbs * ) ;

#define DEFAULT_FUNC_RANGE 10000.0

static int   func_use_autorange = 1   ;
static float func_threshold     = 0.5 ;
static float func_thresh_top    = 1.0 ;
static int   func_use_thresh    = 1   ;   /* not currently alterable */
static float func_color_opacity = 0.5 ;
static int   func_see_overlay   = 0   ;
static int   func_pos_mview   = 0   ;
static int   func_pos_mcalc   = 0   ;
static int   func_lm_view = 0   ;
static int   func_dti_overlay   = 0   ;	/* lrf */
static int   func_sdv_overlay   = 0   ;	/* lrf */
static int   func_new_shd       = 0   ;	/* lrf */
static int   func_new_ftm       = 0   ;	/* lrf */
static float func_clusters_rmm  = 1.0 ;
static float func_clusters_vmul = 200.0 ;
static int   func_posfunc       = 0   ;
static float func_range         = DEFAULT_FUNC_RANGE ;
static float func_autorange     = DEFAULT_FUNC_RANGE ;
static int   func_computed      = 0 ;

static int   func_showthru      = 0 ;  /* 07 Jan 2000 */
static int   func_showthru_pass = 0 ;

/* lrf's aniso */
static int shd_lval =  6 ;
static int shd_mval =  6 ;
static int shd_grid = 50 ;

/* lrf's fiber */
static int ftm_lval =  6 ;
static int ftm_mval =  6 ;
static int ftm_grid = 50 ;

static int new_ftm_anat_dset = 0 ;              /* 01-Sep-02 , Is it new?      */
static char ftm_anat_dset_title[THD_MAX_NAME] ; /* Title string */

/* lrf's dti viewer */
static int func_dti_lval = 6 ;
static int func_dti_mval = 6 ;

/* lrf's shd viewer */
static int func_shd_lval = 6 ;
static int func_shd_mval = 6 ;

#define NOSHADE 1
#define NOMIX   2
static int   func_mixshade      = 0 ;    /* 20 Dec 1999 */

static THD_3dim_dataset * func_dset = NULL ;
static THD_3dim_dataset *shd_dset = NULL;
static THD_3dim_dataset *dt2_dset = NULL;
static MCW_idcode         func_dset_idc ;   /* 31 Mar 1999 */
static int new_func_dset = 0 ;              /* 01-Sep-02 , Is it new?      */

static int func_color_ival  = 0 ;
static int func_thresh_ival = 0 ;

static byte func_rmap[256] , func_gmap[256] , func_bmap[256] ;
static int func_ncmap ;
static int func_cmap_set = 0 ;

static MRI_IMAGE * ovim ;

static char func_dset_title[THD_MAX_NAME] ; /* Title string */

char * DIFF_thresh_tlabel_CB( MCW_arrowval * , XtPointer ) ;
void DIFF_setup_color_pbar(void) ;
XmString DIFF_range_label(void) ;
XmString DIFF_autorange_label(void) ;

void DIFF_range_bbox_CB    ( Widget , XtPointer , XtPointer ) ;
void DIFF_color_bbox_CB    ( Widget , XtPointer , XtPointer ) ;
void DIFF_thr_scale_CB     ( Widget , XtPointer , XtPointer ) ;
void DIFF_thr_scale_drag_CB( Widget , XtPointer , XtPointer ) ;
void DIFF_see_dti_overlay_CB   ( Widget , XtPointer , XtPointer ) ;
void DIFF_see_shd_overlay_CB   ( Widget , XtPointer , XtPointer ) ;
void DIFF_pos_mview_CB   ( Widget , XtPointer , XtPointer ) ;
void DIFF_pos_mcalc_CB   ( Widget , XtPointer , XtPointer ) ;
void DIFF_finalize_func_CB ( Widget , XtPointer , MCW_choose_cbs * ) ;
void DIFF_finalize_ftm_anat_CB ( Widget , XtPointer , MCW_choose_cbs * ) ;

/*------------------- Stuff for DTI viewer logic ----------------------*/

void DIFF_scalar_av_CB( MCW_arrowval * , XtPointer ) ; /* Scalar choices */

/*------------------- Stuff for SHD viewer logic ----------------------*/

void DIFF_lm_view_CB ( Widget , XtPointer , XtPointer ) ;
void DIFF_wshdfunc_shd_lval_av_CB ( MCW_arrowval * , XtPointer ) ;
void DIFF_wshdfunc_shd_mval_av_CB ( MCW_arrowval * , XtPointer ) ;

/*------------------- Stuff for FTM viewer logic ----------------------*/

void DIFF_wftmfunc_ftm_lval_av_CB ( MCW_arrowval * , XtPointer ) ;
void DIFF_wftmfunc_ftm_mval_av_CB ( MCW_arrowval * , XtPointer ) ;

/*------------------- Stuff for anisotropy logic ----------------------*/

void DIFF_aniso_compute_CB(Widget , XtPointer , XtPointer) ; 
void DIFF_aniso_cancel_CB (Widget , XtPointer , XtPointer) ; 

/*void DIFF_dti_flag_CB ( Widget , XtPointer , XtPointer ) ;
  void DIFF_shd_flag_CB ( Widget , XtPointer , XtPointer ) ;*/

void DIFF_dti_compute_CB ( Widget , XtPointer , XtPointer ) ;
void DIFF_shd_compute_CB ( Widget , XtPointer , XtPointer ) ;
void DIFF_dwi2dt_compute_CB ( Widget, XtPointer, XtPointer ) ;

void DIFF_dti_lval_av_CB ( MCW_arrowval * , XtPointer ) ;
void DIFF_dti_mval_av_CB ( MCW_arrowval * , XtPointer ) ;
void DIFF_dti_grid_av_CB ( MCW_arrowval * , XtPointer ) ;

void DIFF_shd_lval_av_CB ( MCW_arrowval * , XtPointer ) ;
void DIFF_shd_mval_av_CB ( MCW_arrowval * , XtPointer ) ;
void DIFF_shd_grid_av_CB ( MCW_arrowval * , XtPointer ) ;

void DIFF_ftm_lval_av_CB ( MCW_arrowval * , XtPointer ) ;
void DIFF_ftm_mval_av_CB ( MCW_arrowval * , XtPointer ) ;
void DIFF_ftm_grid_av_CB ( MCW_arrowval * , XtPointer ) ;

void DIFF_aniso_CB  ( MCW_arrowval * , XtPointer ) ;

/*------------------- Stuff for dti display ----------------------*/

static Widget dti_controls_row1,dti_controls_row2, dti_controls_rowcol ;
static Widget dti_compute_scalars_pb ;
static Widget dti_cancel_scalars_pb ;
static Widget dti_display_scalars_pb ;
static Widget dti_save_scalars_pb ;
void DIFF_dti_cancel_scalars_CB(Widget , XtPointer , XtPointer) ; 
void DIFF_dti_compute_scalars_CB(Widget , XtPointer , XtPointer) ; 
void DIFF_dti_display_scalars_CB(Widget , XtPointer , XtPointer) ; 
void DIFF_dti_save_scalars_CB(Widget , XtPointer , XtPointer) ; 

/*------------------- Stuff for shd display ----------------------*/

static Widget shd_controls_row1,shd_controls_row2, shd_controls_rowcol ;
static Widget shd_compute_scalars_pb ;
static Widget shd_cancel_scalars_pb ;
static Widget shd_display_scalars_pb ;
static Widget shd_save_scalars_pb ;
void DIFF_shd_cancel_scalars_CB(Widget , XtPointer , XtPointer) ; 
void DIFF_shd_compute_scalars_CB(Widget , XtPointer , XtPointer) ; 
void DIFF_shd_display_scalars_CB(Widget , XtPointer , XtPointer) ; 
void DIFF_shd_save_scalars_CB(Widget , XtPointer , XtPointer) ; 

static Widget shd_display_rowcol ;
static Widget shd_display_pb ;
void DIFF_shd_display_CB(Widget , XtPointer , XtPointer) ; 
XtPointer compute_shd_imseq(int n, int type, FD_brick * br ) ; 

/*------------------- Stuff for lrf's header input ----------------------*/

int  header_exists ;
char nheader[RDB_HEADER_SIZE_BYTES];
/* char header[SIGNA_H]; */

/*------------------- Stuff for lrf's fiber logic ----------------------*/

void DIFF_fiber_compute_CB(Widget , XtPointer , XtPointer) ; 
void DIFF_fiber_cancel_CB (Widget , XtPointer , XtPointer) ; 

void DIFF_ftm1_flag_CB ( Widget , XtPointer , XtPointer ) ;
void DIFF_ftm1_compute_CB ( Widget , XtPointer , XtPointer ) ;

void DIFF_ftm_dst_compute_CB ( Widget , XtPointer , XtPointer ) ;

void DIFF_ftm2_flag_CB ( Widget , XtPointer , XtPointer ) ;
void DIFF_ftm2_compute_CB ( Widget , XtPointer , XtPointer ) ;

void DIFF_ftm_lval_av_CB ( MCW_arrowval * , XtPointer ) ;
void DIFF_ftm_mval_av_CB ( MCW_arrowval * , XtPointer ) ;
void DIFF_ftm_grid_av_CB ( MCW_arrowval * , XtPointer ) ;
void DIFF_fiber_CB  ( MCW_arrowval * , XtPointer ) ;

static int fiber_compute_flag = 0 ;

/*------------------ cyrus/ stuff for VTK window  -----------------*/

#define VTK_WINDOW_OPEN   1
#define VTK_WINDOW_CLOSED 0
static int VTK_window_status = VTK_WINDOW_CLOSED;
static int VTK_dirtybit = 1;   /* VTK window is initialized as needing update */

/* lrf - new FTM stuff 10-Mar-04 */
#define VTK_VECTOR_WINDOW_OPEN   1
#define VTK_VECTOR_WINDOW_CLOSED 0
static int vtk_vector_window_status = VTK_VECTOR_WINDOW_CLOSED;

/* compute dti, shd coefficients (dti by default)*/
static int DTI_COMPUTE_COEFFICIENTS = 0;
static int SHD_COMPUTE_COEFFICIENTS = 0;
static int VIEW_DATA = 1;

//float *current_vox_anat_data = NULL;
vtkPolyData *vGrid ;		/* to make visible to Launch_vtk_window call */

/*------------------  SHD VIEWER window -------------------------- */
int shd_img_count = 0;
int shd_img_count_ulay = 0;
MRI_IMAGE * SHD_overlay( int l, int m, int slice, FD_brick *br );

/*---------------------------------------------------------------------------*/

void DIFF_range_av_CB     ( MCW_arrowval * , XtPointer ) ;
void DIFF_thresh_top_CB   ( MCW_arrowval * , XtPointer ) ;
void DIFF_colornum_av_CB  ( MCW_arrowval * , XtPointer ) ;
void DIFF_color_opacity_CB( MCW_arrowval * , XtPointer ) ;
void DIFF_clusters_av_CB  ( MCW_arrowval * , XtPointer ) ;

void DIFF_color_pbar_CB( MCW_pbar * , XtPointer , int ) ;
void DIFF_set_thr_pval(void) ;

#define COLSIZE 20  /* to modify optmenus */

#undef FIX_SCALE_SIZE
#undef HIDE_SCALE
#ifdef FIX_SCALE_SIZE_PROBLEM
#  define FIX_SCALE_SIZE                                        \
     do{ unsigned long sel_height ;  XtPointer sel_ptr ;                  \
         if( wshdfunc_thr_scale != NULL ){                         \
           XtVaGetValues( wshdfunc_thr_scale ,                     \
                             XmNuserData , &sel_ptr , NULL ) ;  \
           sel_height = (unsigned long) sel_ptr ;                         \
           XtVaSetValues( wshdfunc_thr_scale ,                     \
                             XmNheight , sel_height , NULL ) ;  \
           XtManageChild(wshdfunc_thr_scale) ;                     \
       } } while(0)
#  define HIDE_SCALE \
     do{ if(wshdfunc_thr_scale != NULL) XtUnmanageChild(wshdfunc_thr_scale); } while(0)
#else
#  define FIX_SCALE_SIZE /* nada */
#  define HIDE_SCALE     /* nada */
#endif

/*-------------------------------------------------------------------*/

void QUOTE_popper(Widget w)
{
  char buf[2048] ;
  int nq ;

  nq = lrand48() % NUM_QUOTES ;

  strcpy( buf , quotes[nq] ) ;
  (void) MCW_popup_message( w , buf , MCW_USER_KILL | MCW_TIMER_KILL ) ;

  return ;
}

/*-----------------------------------------------------------------
  Event handler to find #3 button press for hidden popup
-------------------------------------------------------------------*/

void LRF_hidden_EV( Widget w , XtPointer cd ,
                    XEvent * ev , Boolean * continue_to_dispatch )
{
   Three_D_View * im3d = (Three_D_View *) cd ;

   if( ! IM3D_OPEN(im3d) ) EXRETURN ;

   /*** handle events ***/

   switch( ev->type ){

      /*----- take button press -----*/

      case ButtonPress:{
         XButtonEvent * event = (XButtonEvent *) ev ;

         if( event->button == Button3 ){ }
         else if( !NO_frivolities && event->button == Button1 ) QUOTE_popper(w) ;
         else if( !NO_frivolities && event->button == Button2 ) ;

      }
      break ;

   }

   EXRETURN ;
}

/*-------------------------------------------------------------------*/

/***************************************************************************
  Will be called from AFNI when user selects from Plugins menu.
****************************************************************************/

char * DIFF_main( PLUGIN_interface * plint )
{
   XmString xstr ;

   /* Disable Hawaii 5-0 splash by default */
   static int nosplash = 1 ;
   static double eltime , max_splash=1.0 ;
   char *e_asp ;

   /*-- Always enforce the aspect ratio --*/

   e_asp = my_getenv("AFNI_ENFORCE_ASPECT") ;
   if( !YESSISH(e_asp) ){
      putenv("AFNI_ENFORCE_ASPECT=YES") ;
   }

   /* Splash up the boys from 5-0 */

#ifdef ALL_FRIVOLITIES
   nosplash = 0 ;
#endif
#ifdef FRIVOLITIES
   nosplash = (getenv("AFNI_NOSPLASH") != NULL) ;
#endif
   if( !nosplash ){
     char * hh ;
     LRF_splashup() ;
     hh = getenv("AFNI_SPLASHTIME") ;
     if( hh != NULL ) max_splash = strtod(hh,NULL) ;
     iochan_sleep(1);
     XmUpdateDisplay(plint->im3d->vwid->top_shell);
   }

   eltime = COX_clock_time() ;
   while(1) {
     if( COX_clock_time()-eltime >= max_splash ){ 
       LRF_splashdown();      
       break;}
   }

   /*-- sanity checks --*/

   if( ! IM3D_OPEN(plint->im3d) ) return "AFNI Controller\nnot opened?!" ;

   if( diffuser_open ){
      XtMapWidget(shell) ;
      XRaiseWindow( XtDisplay(shell) , XtWindow(shell) ) ;
      return NULL ;
   }

   im3d = plint->im3d ;  /* save for local use */

   /*-- create widgets, first time through --*/

   if( shell == NULL ){
      dc = im3d->dc ;        /* save this too */
      DIFF_make_widgets() ;
   }

   /*-- set titlebar --*/

   { char ttl[PLUGIN_STRING_SIZE] ;
     sprintf( ttl , "%s AFNI Diffusion v. %s.%s", AFNI_controller_label(im3d),
	      ADP_MAJOR_VER, ADP_SVN_REV) ;
     XtVaSetValues( shell , XmNtitle , ttl , NULL ) ;
   }

   /*-- set some widget values --*/

   xstr = XmStringCreateLtoR( NO_DATASET_STRING ,
                              XmFONTLIST_DEFAULT_TAG ) ;
   XtVaSetValues( info_lab , XmNlabelString , xstr , NULL ) ;
   XmStringFree(xstr) ;

#ifdef EXPERIMENTAL
   xstr = XmStringCreateLtoR( "Min=?????? Max=??????" ,
                              XmFONTLIST_DEFAULT_TAG            ) ;
   XtVaSetValues( range_lab , XmNlabelString , xstr , NULL ) ;
   XmStringFree(xstr) ;

   AV_assign_ival( clipbot_av , -CLIP_RANGE ) ;
   AV_assign_ival( cliptop_av ,  CLIP_RANGE ) ;

   brickfac = 0.0 ;
   XtUnmanageChild( range_faclab   ) ;
   XtUnmanageChild( clipbot_faclab ) ;
   XtUnmanageChild( cliptop_faclab ) ;
#endif

   MCW_set_bbox( xhair_bbox   , 0 ) ; xhair_flag   = 0 ;
   MCW_set_bbox( dynamic_bbox , 0 ) ; dynamic_flag = 0 ;
   MCW_set_bbox( accum_bbox   , 0 ) ; accum_flag   = 0 ;

   AV_SENSITIZE( choose_av , False ) ;

   /* lrf's overlay */
   MCW_set_bbox( dif_overlay_bbox , 1 ) ; dif_overlay_flag = 0 ;

   /* lrf's aniso */

   /* 02-Sep-02
      MCW_set_bbox( dti_bbox , 0 ) ; dti_flag = 0 ;
      MCW_set_bbox( shd_bbox , 0 ) ; shd_flag = 0 ;
   */

   /* lrf's fiber */

   /* MCW_set_bbox( ftm1_bbox , 0 ) ; ftm1_flag = 0 ;
      MCW_set_bbox( ftm2_bbox , 0 ) ; ftm2_flag = 0 ; 
   */

   /*--- some of the function widgets, too ---*/

   if( wshdfunc_frame != NULL ){
      TURNOFF_SHD_OVERLAY_WIDGETS ;
   }

   /*-- pop the widget up --*/

   XtMapWidget(shell) ;
   PLUTO_cursorize(shell) ;

   /* lrf's diffusion state */

   current_dif_state = init_dif_state( EMPTY , NULL, EMPTY );
   /* set to default (DTI) on open 
    This works but files not sort correctly - must run through name check first */
   /* MCW_set_bbox( dif_overlay_bbox , current_dif_state->sta->overlay_chan) ; */

   /*-- misc initialization --*/

   dset          = NULL ;   /* not diffusing anything     */
   dset_ival     = 0 ;      /* if we were, it would be #0 */
   diffuser_open = 1 ;      /* diffuser is now open for business */
   imseq         = NULL ;   /* no image window is open yet */
   grim = opim   = NULL ;   /* don't have volumes to diffuse yet */
   diffuse_handle = NULL ;   /* don't have a diffuser yet */

   ovim          = NULL ;   /* no overlay volume yet */
   func_dset     = NULL ;   /* no functional dataset yet */

   new_data_loaded = 0 ;    /* not yet */

   grim_showthru = opim_showthru = NULL ; /* 07 Jan 2000 */

#ifdef EXPERIMENTAL
   set_MCW_pasgraf( his_graf , NULL ) ;  /* set histogram graph to 0's */
   redraw_MCW_pasgraf( his_graf ) ;
#endif

   xhair_ixold = -666 ; xhair_jyold = -666 ; xhair_kzold = -666 ;

   /* 29 Mar 1999: register to receive updates from AFNI */

#if 1
   xhair_recv = AFNI_receive_init( im3d ,
                                   RECEIVE_VIEWPOINT_MASK |
				   RECEIVE_DRAWNOTICE_MASK |
				   RECEIVE_DSETCHANGE_MASK ,
                                   reinterpret_cast<void (*)()>
				   (DIFF_xhair_recv), NULL  ,
				   "DIFF_xhair_recv") ;
#else
   xhair_recv = AFNI_receive_init( im3d ,
                                   RECEIVE_VIEWPOINT_MASK , DIFF_xhair_recv , NULL,
				   "DIFF_xhair_recv") ;
#endif

   MPROBE ;
   return NULL ;
}

/*------------------------------------------------------------------------
  Make the control popup for this thing
--------------------------------------------------------------------------*/

/*-- structures defining action buttons (at bottom of popup) --*/

#define NACT 4  /* number of action buttons */

static MCW_action_item DIFF_actor[NACT] = {

 {"Help",DIFF_help_CB,NULL,
  "Displays more help" , "Displays more help",0} ,

 {"Copyright",DIFF_copyright_CB,NULL,
  "Copyright Notice" , "Copyright Notice",0} ,

 /* {"Draw",DIFF_draw_CB,NULL,
    "(Re)Draw the image" , "(Re)Draw the image",0} , */

 {"Reload",DIFF_reload_CB,NULL,
  "Reload dataset values" , "Reload dataset values",0} ,

 {"done",DIFF_done_CB,NULL,
  "Close diffuser\nand image." , "Close windows",1}
} ;

#define SEP_HOR(ww)  XtVaCreateManagedWidget(                     \
                       "AFNI" , xmSeparatorWidgetClass , (ww) ,   \
                          XmNseparatorType , XmSINGLE_LINE ,      \
                          XmNinitialResourcesPersistent , False , \
                       NULL )

#define SEP_VER(ww) XtVaCreateManagedWidget(                      \
                       "AFNI" , xmSeparatorWidgetClass , (ww) ,   \
                          XmNseparatorType , XmDOUBLE_LINE ,      \
                          XmNorientation   , XmVERTICAL ,         \
                          XmNinitialResourcesPersistent , False , \
                       NULL )

void DIFF_make_widgets(void)
{
   XmString xstr ;
   char     str[64] ;
   Widget   hrc , vrc, vrc0, vtk ;
   int      ii ;
   char    *env ;
   float    val ;

   /***----- retrieve relevant environment variables, if any -----***/

   /* lrf - diffusion */

   env = getenv("AFNI_SHD_GRID_DELTA") ;
   if( env != NULL ){
      val = strtod(env,NULL) ;
      if( val > 0.0 && val < 100.0 ) shd_grid_fstep = (int)val ;
   }

   /***=============================================================*/

   /*** top level shell for window manager ***/

   shell =
      XtVaAppCreateShell(
			 "AFNI" ,"AFNI" , topLevelShellWidgetClass , dc->display ,
			 XmNtitle             , "AFNI Diffusion" , /* top of window */
			 XmNiconName          , "Diffusion"      , /* label on icon */
			 XmNdeleteResponse    , XmDO_NOTHING  ,   /* deletion handled below */
			 XmNallowShellResize  , True ,            /* let code resize shell? */
			 XmNmappedWhenManaged , False ,           /* must map it manually */
			 XmNinitialResourcesPersistent , False ,
			 NULL ) ;

   DC_yokify( shell , dc ) ; /* 14 Sep 1998 */

#ifndef DONT_INSTALL_ICONS
   if( afni48_good )             /* set icon pixmap */
      XtVaSetValues( shell ,
                        XmNiconPixmap , afni48_pixmap ,
                     NULL ) ;
#endif

   if( MCW_isitmwm(shell) )      /* remove some MWM functions */
      XtVaSetValues( shell ,
                       XmNmwmFunctions ,
                       MWM_FUNC_MOVE | MWM_FUNC_CLOSE | MWM_FUNC_MINIMIZE ,
                     NULL ) ;

   XmAddWMProtocolCallback(      /* make "Close" window menu work */
           shell ,
           XmInternAtom( dc->display , "WM_DELETE_WINDOW" , False ) ,
           DIFF_done_CB , (XtPointer) plint ) ;

   /*** horizontal rowcol to hold ALL interface stuff ***/

   top_rowcol =  XtVaCreateWidget(
                  "AFNI" , xmRowColumnWidgetClass , shell ,
                     XmNorientation  , XmHORIZONTAL ,
                     XmNpacking      , XmPACK_TIGHT ,
                     XmNadjustLast   , False ,
                     XmNadjustMargin , False ,
                     XmNtraversalOn  , False ,
                     XmNmarginWidth  , 0 ,
                     XmNmarginHeight , 0 ,
                     XmNinitialResourcesPersistent , False ,
                  NULL ) ;

   /*** vertical rowcolumn widget to hold anat interface stuff ***/

   anat_frame = XtVaCreateWidget(
                   "AFNI" , xmFrameWidgetClass , top_rowcol ,
                      XmNshadowType , XmSHADOW_ETCHED_IN ,
                      XmNshadowThickness , 5 ,
                      XmNtraversalOn , False ,
                      XmNinitialResourcesPersistent , False ,
                   NULL ) ;

   anat_rowcol = XtVaCreateWidget(
                  "AFNI" , xmRowColumnWidgetClass , anat_frame ,
                     XmNpacking     , XmPACK_TIGHT ,
                     XmNorientation , XmVERTICAL ,
                     XmNadjustLast  , False ,
                     XmNadjustMargin, False ,
                     XmNtraversalOn , False ,
                     XmNinitialResourcesPersistent , False ,
                  NULL ) ;

   /***=============================================================*/

   /*** label at top to let user know who we are ***/

   xstr = XmStringCreateLtoR( NO_DATASET_STRING ,
                              XmFONTLIST_DEFAULT_TAG ) ;
   info_lab = XtVaCreateManagedWidget(
                 "AFNI" , xmLabelWidgetClass , anat_rowcol ,
                    XmNlabelString , xstr ,
                    XmNrecomputeSize , False ,
                    XmNinitialResourcesPersistent , False ,
                 NULL ) ;
   XmStringFree(xstr) ;
   MCW_register_help( info_lab , "Shows dataset being diffused" ) ;

   /***** top row of widgets to choose dataset and sub-brick *****/

   SEP_HOR(anat_rowcol) ;  /* separator widget */

   hrc =  XtVaCreateWidget(
           "AFNI" , xmRowColumnWidgetClass , anat_rowcol ,
              XmNorientation  , XmHORIZONTAL ,
              XmNpacking      , XmPACK_TIGHT ,
              XmNadjustLast   , False ,
              XmNadjustMargin , False ,
              XmNtraversalOn  , False ,
              XmNmarginWidth  , 0 ,
              XmNmarginHeight , 0 ,
              XmNinitialResourcesPersistent , False ,
           NULL ) ;

   dataset_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , hrc ,
            XmNorientation , XmVERTICAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   /*** button to let user choose dataset to diffuse ***/

   xstr = XmStringCreateLtoR( "Choose Diffusion Dataset" , XmFONTLIST_DEFAULT_TAG ) ;
   choose_pb = XtVaCreateManagedWidget(
                  "AFNI" , xmPushButtonWidgetClass , dataset_rowcol ,
                     XmNalignment   , XmALIGNMENT_CENTER ,
                     XmNlabelString , xstr ,
                     XmNtraversalOn , False ,
                     XmNinitialResourcesPersistent , False ,
                  NULL ) ;
   XmStringFree(xstr) ;
   /*strcpy(diff_choose_str,"MAIN");
   XtAddCallback( choose_pb, XmNactivateCallback, DIFF_choose_CB, diff_choose_str ) ;*/
   XtAddCallback( choose_pb, XmNactivateCallback, DIFF_choose_CB, "MAIN" ) ;
   MCW_register_help( choose_pb ,
                      "Use this to popup a\n"
                      "'chooser' that lets\n"
                      "you select which\n"
                      "dataset to diffuse."
		      ) ;

   /*** menu to let user choose sub-brick to deal with ***/

   SEP_VER(hrc) ;

   XtManageChild( dataset_rowcol ) ;

   /*** ucsd logo ***/

   XtVaGetValues( info_lab ,
		  XmNforeground , &ucsd_fg_pix ,
		  XmNbackground , &ucsd_bg_pix ,
		  NULL ) ;

   /* UCSD Logo */

#ifdef UCSD_LOGO   

   if( ucsd_logo_pixmap == XmUNSPECIFIED_PIXMAP )
     ucsd_logo_pixmap = XCreatePixmapFromBitmapData(
						    XtDisplay(shell) ,
						    RootWindowOfScreen(XtScreen(shell)) ,
						    ucsd_logo_bits ,
						    ucsd_logo_width , ucsd_logo_height ,
						    ucsd_fg_pix , ucsd_bg_pix ,
						    DefaultDepthOfScreen(XtScreen(shell)) ) ;

   ucsd_logo_picture =
     XtVaCreateManagedWidget(
			     "dialog" , xmLabelWidgetClass , hrc ,
			     XmNlabelType        , XmPIXMAP ,
			     XmNalignment        , XmALIGNMENT_END ,
			     XmNlabelInsensitivePixmap , ucsd_logo_pixmap ,
			     XmNwidth        , ucsd_logo_width ,
			     XmNheight       , ucsd_logo_height ,
			     XmNmarginWidth  , 0 ,
			     XmNmarginHeight , 0 ,
			     XmNrecomputeSize, False ,
			     XmNtraversalOn  , False ,
			     XmNinitialResourcesPersistent , False ,
			     NULL ) ;

   XtVaSetValues(ucsd_logo_picture,XmNlabelPixmap,ucsd_logo_pixmap,NULL);

#endif

   /*** Hidden quotes ***/
#ifdef UCSD_LOGO
   XtInsertEventHandler( ucsd_logo_picture ,  /* handle events in picture */
			 0 | ButtonPressMask, /* button presses */
			 FALSE ,              /* nonmaskable events? */
			 LRF_hidden_EV ,      /* handler */
			 (XtPointer) im3d ,   /* client data */
			 XtListTail           /* last in queue */
			 ) ;
#endif
   /*** Button to open overlay panels ***/
#ifdef EXPERIMENTAL
   SEP_VER(hrc) ;

   overlay_rowcol =
      XtVaCreateWidget( "AFNI" , xmRowColumnWidgetClass , hrc ,
			XmNorientation , XmVERTICAL ,
			XmNalignment   , XmALIGNMENT_CENTER ,
			XmNpacking , XmPACK_TIGHT ,
			XmNtraversalOn , False ,
			XmNinitialResourcesPersistent , False ,
			NULL ) ;
   
   /* Overlay radio box and button */

   {
     dif_overlay_label =
       XtVaCreateManagedWidget("dialog" , xmLabelWidgetClass ,  
			       overlay_rowcol ,
			       LABEL_ARG("Diffusion Overlay") ,
			       XmNinitialResourcesPersistent , False ,
			       NULL ) ;

     xstr = XmStringCreateLtoR( "Open" , XmFONTLIST_DEFAULT_TAG ) ;

     dif_overlay_pb = 
       XtVaCreateManagedWidget(	"AFNI" , xmPushButtonWidgetClass , 
				overlay_rowcol ,
				XmNlabelString , xstr ,
				XmNtraversalOn , False ,
				XmNinitialResourcesPersistent , False ,
				NULL ) ;
     XmStringFree(xstr) ;
     XtAddCallback( dif_overlay_pb, 
		    XmNactivateCallback, DIFF_overlay_CB, NULL ) ;
     MCW_register_help( dif_overlay_pb ,
			"Use this to open/close diffusion overlay" ) ;

     dif_overlay_bbox = 
       new_LRF_bbox( overlay_rowcol ,
		     NBUT_DIFO , DIFO_labels ,
		     MCW_BB_radio_zero ,
		     MCW_BB_frame ,
		     LRF_BB_horiz ,
		     DIFF_overlay_chan_CB , NULL ) ;
   }

   /* end new overlay radio box */

   XtManageChild( overlay_rowcol ) ;
#endif
   XtManageChild(hrc) ;

   /***=============================================================*/

   /*** horizontal rowcol for graphs ***/

   SEP_HOR(anat_rowcol) ;  /* separator */

   hrc =  XtVaCreateWidget(
           "AFNI" , xmRowColumnWidgetClass , anat_rowcol ,
              XmNorientation , XmHORIZONTAL ,
              XmNpacking , XmPACK_TIGHT ,
              XmNadjustLast  , False ,
              XmNadjustMargin, False ,
              XmNtraversalOn , False ,
              XmNmarginWidth , 0 ,
              XmNmarginHeight, 0 ,
              XmNinitialResourcesPersistent , False ,
           NULL ) ;

#ifdef EXPERIMENTAL

   /*** passive graph to show data distribution ***/

   his_graf = new_MCW_pasgraf( hrc , im3d->dc , "Sqrt Histogram" ) ;
   his_graf->mode = PASGRAF_BAR ;

   MCW_reghelp_children( his_graf->topform ,
                         "The graph height is proportional to\n"
                         "the square-root of the histogram of\n"
                         "the input signal.\n"
                         "\n"
                         "N.B.: The histogram at 0 is not included\n"
                         "      in the scaling, since it tends to\n"
                         "      be huge.  The square-root is graphed\n"
                         "      to enhance the range of the plot."      ) ;

   SEP_VER(hrc) ;

   /***=============================================================*/

   /*** vertical rowcol for data value clipping ***/

   SEP_HOR(anat_rowcol) ;  /* separator */

   vrc0 =  XtVaCreateWidget(
           "AFNI" , xmRowColumnWidgetClass , hrc ,
	      XmNorientation , XmVERTICAL ,
              XmNpacking      , XmPACK_TIGHT ,
              XmNadjustLast   , False ,
              XmNadjustMargin , False ,
              XmNtraversalOn  , False ,
              XmNmarginWidth  , 0 ,
              XmNmarginHeight , 0 ,
              XmNinitialResourcesPersistent , False ,
           NULL ) ;

   /*** vertical rowcol for dataset range information labels ***/

   vrc = XtVaCreateWidget(
             "AFNI" , xmRowColumnWidgetClass , vrc0 ,
                XmNpacking     , XmPACK_TIGHT ,
                XmNorientation , XmVERTICAL ,
                XmNadjustLast  , False ,
                XmNadjustMargin, False ,
                XmNtraversalOn , False ,
                XmNmarginWidth , 0 ,
                XmNmarginHeight, 0 ,
                XmNinitialResourcesPersistent , False ,
             NULL ) ;

   /*** 1st label for dataset range information ***/

   xstr = XmStringCreateLtoR( "Min=?????? Max=??????" , XmFONTLIST_DEFAULT_TAG ) ;
   range_lab = XtVaCreateManagedWidget(
                 "AFNI" , xmLabelWidgetClass , vrc ,
                    XmNlabelString , xstr ,
                    XmNrecomputeSize , False ,
                    XmNinitialResourcesPersistent , False ,
                 NULL ) ;
   XmStringFree(xstr) ;

   MCW_register_help( range_lab ,
                      "Shows the range of the data stored\n"
                      "in the brick voxels.\n"
                      "\n"
                      "N.B.: These values are NOT scaled\n"
                      "      by any floating point\n"
                      "      brick scaling factor."
                    ) ;

   /*** 2nd label for scaled dataset range information ***/

   xstr = XmStringCreateLtoR( "[123456789 123456789]" , XmFONTLIST_DEFAULT_TAG ) ;
   range_faclab = XtVaCreateWidget(
                    "AFNI" , xmLabelWidgetClass , vrc ,
                       XmNlabelString , xstr ,
                       XmNrecomputeSize , False ,
                       XmNinitialResourcesPersistent , False ,
                    NULL ) ;
   XmStringFree(xstr) ;

   MCW_register_help( range_faclab ,
                      "Shows the range of data stored\n"
                      "in the brick, this time multiplied\n"
                      "by the brick's scaling factor."
                    ) ;

   XtManageChild(vrc) ;

   /*** arrowvals to get dataset clip levels ***/

   /*** vertical rowcol for Bot arrowval ***/

   vrc = XtVaCreateWidget(
             "AFNI" , xmRowColumnWidgetClass , vrc0 ,
                XmNpacking     , XmPACK_TIGHT ,
                XmNorientation , XmVERTICAL ,
                XmNadjustLast  , False ,
                XmNadjustMargin, False ,
                XmNtraversalOn , False ,
                XmNmarginWidth , 0 ,
                XmNmarginHeight, 0 ,
                XmNinitialResourcesPersistent , False ,
             NULL ) ;

   clipbot_av = new_MCW_arrowval( vrc , "Bot " ,
                                MCW_AV_downup , -CLIP_RANGE,CLIP_RANGE,-CLIP_RANGE ,
                                MCW_AV_editext , 0 ,
                                DIFF_clip_CB , NULL , NULL,NULL ) ;

   MCW_reghelp_children( clipbot_av->wrowcol ,
                         "All (unscaled) voxel values below\n"
                         "'Bot' will be increased to this\n"
                         "value.  The larger of 'Bot' and\n"
                         "'Min' is the left edge of the\n"
                         "brick graphs shown below."
                       ) ;

   xstr = XmStringCreateLtoR( "[-> 123456789]" , XmFONTLIST_DEFAULT_TAG ) ;
   clipbot_faclab = XtVaCreateWidget(
                    "AFNI" , xmLabelWidgetClass , vrc ,
                       XmNlabelString , xstr ,
                       XmNrecomputeSize , False ,
                       XmNinitialResourcesPersistent , False ,
                    NULL ) ;
   XmStringFree(xstr) ;

   MCW_register_help( clipbot_faclab ,
                      "Shows the scaled\nvalue of 'Bot'." ) ;

   XtManageChild(vrc) ;

   SEP_VER(hrc) ;

   /*** vertical rowcol for Top arrowval ***/

   vrc = XtVaCreateWidget(
             "AFNI" , xmRowColumnWidgetClass , vrc0 ,
                XmNpacking     , XmPACK_TIGHT ,
                XmNorientation , XmVERTICAL ,
                XmNadjustLast  , False ,
                XmNadjustMargin, False ,
                XmNtraversalOn , False ,
                XmNmarginWidth , 0 ,
                XmNmarginHeight, 0 ,
                XmNinitialResourcesPersistent , False ,
             NULL ) ;

   cliptop_av = new_MCW_arrowval( vrc , "Top " ,
                                MCW_AV_downup , -CLIP_RANGE,CLIP_RANGE, CLIP_RANGE ,
                                MCW_AV_editext , 0 ,
                                DIFF_clip_CB , NULL , NULL,NULL ) ;

   MCW_reghelp_children( cliptop_av->wrowcol ,
                         "All (unscaled) voxel values above\n"
                         "'Top' will be decreased to this\n"
                         "value.  The smaller of 'Top' and\n"
                         "'Max' is the right edge of the\n"
                         "brick graphs shown below."
                       ) ;

   xstr = XmStringCreateLtoR( "[-> 123456789]" , XmFONTLIST_DEFAULT_TAG ) ;
   cliptop_faclab = XtVaCreateWidget(
                    "AFNI" , xmLabelWidgetClass , vrc ,
                       XmNlabelString , xstr ,
                       XmNrecomputeSize , False ,
                       XmNinitialResourcesPersistent , False ,
                    NULL ) ;
   XmStringFree(xstr) ;

   MCW_register_help( clipbot_faclab ,
                      "Shows the scaled\nvalue of 'Top'." ) ;


   /***=============================================================*/ 
   
   /*** hold special button controls: work in progress : cyrus    ***/

   XtManageChild(vrc) ;
   XtManageChild(vrc0) ;

#endif /* EXPERIMENTAL */
   vtk =  XtVaCreateWidget(
           "AFNI" , xmRowColumnWidgetClass , hrc ,
	      XmNorientation , XmVERTICAL ,
              XmNpacking      , XmPACK_TIGHT ,
              XmNadjustLast  , False ,
              XmNadjustMargin, False ,
              XmNtraversalOn , False ,
              XmNmarginWidth , 0 ,
              XmNmarginHeight, 0 ,
              XmNinitialResourcesPersistent , False ,
           NULL ) ;



   /*** pushbutton to activate the vtk window ***/

   xstr = XmStringCreateLtoR( "Examine voxel data" , XmFONTLIST_DEFAULT_TAG ) ;
   show_vtkvis_pb = XtVaCreateManagedWidget(
                     "AFNI" , xmPushButtonWidgetClass , vtk ,
                        XmNlabelString , xstr ,
                        XmNtraversalOn , False ,
                        XmNinitialResourcesPersistent , False ,
                     NULL ) ;
   XmStringFree(xstr) ;
   XtAddCallback( show_vtkvis_pb, XmNactivateCallback, DIFF_show_vtkvis_CB, NULL ) ;
   MCW_register_help( show_vtkvis_pb ,
                      "Place cursor in image window and press left button."
                      "Window will display measured data."
		      );

   XtManageChild(vtk) ;
   XtManageChild(hrc) ;


   /***=============================================================*/

   /*** horizontal rowcol to hold anisotropy controls ***/

   SEP_HOR(anat_rowcol) ;  /* separator */

   achoices_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , anat_rowcol ,
            XmNorientation , XmHORIZONTAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNmarginHeight, 0 ,
            XmNmarginWidth , 0 ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   xstr = XmStringCreateLtoR( "" , XmFONTLIST_DEFAULT_TAG ) ;
   achoices_label =
      XtVaCreateManagedWidget(
         "AFNI" , xmLabelWidgetClass , achoices_rowcol ,
            XmNlabelString , xstr ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;
   XmStringFree(xstr) ;

   /*--- menus to control anisotropy choices ---*/

   aniso_frame =
      XtVaCreateWidget(
         "AFNI" , xmFrameWidgetClass , achoices_rowcol ,
            XmNshadowType , XmSHADOW_ETCHED_IN ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   aniso_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , aniso_frame ,
            XmNorientation , XmVERTICAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   aniso_label =
     XtVaCreateManagedWidget(
           "dialog" , xmLabelWidgetClass ,  aniso_rowcol,
            LABEL_ARG("Anisotropy") ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   MCW_register_help( aniso_label ,
                      "Anisotropy measures.\n\n" 
		      "1) Compute diffusion tensor at each voxel.\n"
		      "(See 'Diffusion Tensor Model' in Theory').\n\n" 
		      "2) Compute spherical harmonic transform at each voxel.\n"
		      "(See 'Spherical Harmonic Decomposition' in Theory')." ) ; 

   /*--- Anisotropy methods choices ---*/

   aniso_compute_bbox = 
       new_LRF_bbox( aniso_rowcol ,
		     NBUT_DIFC , DIFC_labels ,
		     MCW_BB_radio_zero ,
		     MCW_BB_noframe ,
		     LRF_BB_vert ,
		     DIFF_aniso_compute_chan_CB , NULL ) ;

   /*** pushbutton to activate the diffusion computation ***/

   aniso_compute_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , aniso_rowcol,
            XmNorientation , XmHORIZONTAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   xstr = XmStringCreateLtoR( "Compute" , XmFONTLIST_DEFAULT_TAG ) ;
   aniso_compute_pb = XtVaCreateManagedWidget(
                        "AFNI" , xmPushButtonWidgetClass , 
			aniso_compute_rowcol ,
                        XmNlabelString , xstr ,
                        XmNtraversalOn , False ,
                        XmNinitialResourcesPersistent , False ,
                     NULL ) ;
   XmStringFree(xstr) ;
   XtAddCallback( aniso_compute_pb, 
		  XmNactivateCallback, DIFF_aniso_compute_CB, NULL ) ;
   MCW_register_help( aniso_compute_pb ,
                      "Use this to start the\n"
                      "diffusion computation" ) ;

   /*** pushbutton to cancel the diffusion [not managed now] ***/

   xstr = XmStringCreateLtoR( " * CANCEL * " , XmFONTLIST_DEFAULT_TAG ) ;
   aniso_cancel_pb = XtVaCreateWidget(
                     "AFNI" , xmPushButtonWidgetClass , aniso_compute_rowcol ,
                        XmNlabelString , xstr ,
                        XmNtraversalOn , False ,
                        XmNinitialResourcesPersistent , False ,
                     NULL ) ;
   XmStringFree(xstr) ;
   XtAddCallback( aniso_cancel_pb, XmNactivateCallback, DIFF_aniso_cancel_CB, NULL ) ;

   XtManageChild( aniso_compute_rowcol ) ;
   XtManageChild( aniso_rowcol ) ;
   XtManageChild( aniso_frame ) ;

   /*--- menus to control SHD ---*/

   spharm_frame =
      XtVaCreateWidget(
         "AFNI" , xmFrameWidgetClass , achoices_rowcol ,
            XmNshadowType , XmSHADOW_ETCHED_IN ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   spharm_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , spharm_frame ,
            XmNorientation , XmVERTICAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   spharm_label =
     XtVaCreateManagedWidget(
           "dialog" , xmLabelWidgetClass ,  spharm_rowcol,
            LABEL_ARG("SHD Controls") ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   shd_lval_av =
     new_MCW_arrowval( spharm_rowcol , "Max L" , MCW_AV_downup ,
		       0 , 20 , shd_lval ,
		       MCW_AV_edittext , 0 ,
		       reinterpret_cast<void (*)()>(DIFF_shd_lval_av_CB),
		       NULL,NULL,NULL
		       ) ;
   MCW_reghelp_children(shd_lval_av->wrowcol ,
			"Use this to set the maximum L value\n"
			"for the Spherical Harmonic Transform"
			) ;

   shd_mval_av =
     new_MCW_arrowval( spharm_rowcol , "Max M" , MCW_AV_downup ,
		       0 , 20 , shd_mval,
		       MCW_AV_edittext , 0 ,
		       reinterpret_cast<void (*)()>(DIFF_shd_mval_av_CB),
		       NULL,NULL,NULL
		       ) ;
   MCW_reghelp_children(shd_mval_av->wrowcol ,
			"Use this to set the maximum M value\n"
			"for the Spherical Harmonic Transform"
			) ;

   { char * pos_mcalc_label[1]   = { "Calc M>=0 only?" } ;
   
     pos_mcalc_bbox = new_MCW_bbox( spharm_rowcol ,
				    1 , pos_mcalc_label ,
				    MCW_BB_check ,
				    MCW_BB_noframe ,
				    DIFF_pos_mcalc_CB , NULL ) ;
   }

   shd_grid_av =
     new_MCW_arrowval( spharm_rowcol , "Grid size" , MCW_AV_downup ,
		       50 , 200 , shd_grid,
		       MCW_AV_edittext , 0 ,
		       reinterpret_cast<void (*)()>(DIFF_shd_grid_av_CB),
		       NULL,NULL,NULL
		       ) ;
   shd_grid_av->fstep = shd_grid_fstep;
   MCW_reghelp_children(shd_grid_av->wrowcol ,
			"Use this to set the grid size\n"
			"for the Spherical Harmonic Transform"
			) ;
   XtAddCallback( shd_grid_av->wtext, XmNactivateCallback, DIFF_textact_CB, roll_av ) ;

   XtManageChild( spharm_rowcol ) ;
   XtManageChild( spharm_frame ) ;

   RC_SENSITIZE( spharm_rowcol , False ) ; 

   XtManageChild( achoices_rowcol ) ;

   /***=============================================================*/

   /*** horizontal rowcol to hold SHD viewer controls ***/

   SEP_HOR(anat_rowcol) ;  /* separator */

   shd_choices_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , anat_rowcol ,
            XmNorientation , XmHORIZONTAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNmarginHeight, 0 ,
            XmNmarginWidth , 0 ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   minimal_shd_frame =
      XtVaCreateWidget(
         "AFNI" , xmFrameWidgetClass , shd_choices_rowcol ,
            XmNshadowType , XmSHADOW_ETCHED_IN ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   minimal_shd_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , minimal_shd_frame ,
            XmNorientation , XmVERTICAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   minimal_shd_label =
     XtVaCreateManagedWidget(
           "dialog" , xmLabelWidgetClass ,  minimal_shd_rowcol,
            LABEL_ARG("SHD Display") ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   xstr = XmStringCreateLtoR( "Choose SHD Dataset" , XmFONTLIST_DEFAULT_TAG ) ;
   minimal_shd_display_choose_pb = XtVaCreateManagedWidget(
                   "AFNI" , xmPushButtonWidgetClass , minimal_shd_rowcol ,
		   XmNalignment   , XmALIGNMENT_CENTER ,
		   XmNpacking , XmPACK_TIGHT ,
		   XmNlabelString , xstr ,
		   XmNtraversalOn , False ,
		   XmNinitialResourcesPersistent , False ,
		   NULL ) ;
   XmStringFree(xstr) ;
   XtAddCallback( minimal_shd_display_choose_pb, XmNactivateCallback, DIFF_choose_CB, "OVER" ) ;
   xstr = XmStringCreateLtoR( "Display" , XmFONTLIST_DEFAULT_TAG ) ;
   minimal_shd_display_pb =
     XtVaCreateManagedWidget( "AFNI" , xmPushButtonWidgetClass , 
			      minimal_shd_rowcol ,
			      XmNlabelString , xstr ,
			      XmNtraversalOn , False ,
			      XmNinitialResourcesPersistent,False ,
			      NULL ) ;
   XmStringFree(xstr) ;
   XtAddCallback( minimal_shd_display_pb, XmNactivateCallback, 
		  DIFF_shd_display_CB, im3d ) ; 

   XtManageChild(shd_choices_rowcol);
   XtManageChild(minimal_shd_frame);
   XtManageChild(minimal_shd_rowcol);



   /***=============================================================*/

   /*** horizontal rowcol to hold fiber controls ***/

   SEP_HOR(anat_rowcol) ;  /* separator */

   fchoices_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , anat_rowcol ,
            XmNorientation , XmHORIZONTAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNmarginHeight, 0 ,
            XmNmarginWidth , 0 ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   xstr = XmStringCreateLtoR( "" , XmFONTLIST_DEFAULT_TAG ) ;
   fchoices_label =
      XtVaCreateManagedWidget(
         "AFNI" , xmLabelWidgetClass , fchoices_rowcol ,
            XmNlabelString , xstr ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;
   XmStringFree(xstr) ;

   /*--- Fiber track mapping computation controls ---*/

   fiber_frame =
      XtVaCreateWidget(
         "AFNI" , xmFrameWidgetClass , fchoices_rowcol ,
            XmNshadowType , XmSHADOW_ETCHED_IN ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   fiber_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , fiber_frame ,
            XmNorientation , XmVERTICAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   fiber_label =
     XtVaCreateManagedWidget(
           "dialog" , xmLabelWidgetClass ,  fiber_rowcol,
            LABEL_ARG("Fiber tract mapping") ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   MCW_register_help( fiber_label ,
                      "Fiber tract mapping.\n\n" 
		      "1) method 1.\n"
		      "(See 'ftm1' in Theory').\n\n" 
		      "2) method 2.\n"
		      "(See 'ftm2' in Theory')." ) ; 


   /*** button to let user choose dataset for underlay in vtk window ***/
   xstr = XmStringCreateLtoR( "Choose Anatomical Dataset" , XmFONTLIST_DEFAULT_TAG ) ;
   ftm_anat_choose_pb = XtVaCreateManagedWidget(
                   "AFNI" , xmPushButtonWidgetClass , fiber_rowcol ,
		   XmNalignment   , XmALIGNMENT_CENTER ,
		   XmNpacking , XmPACK_TIGHT ,
		   XmNlabelString , xstr ,
		   XmNtraversalOn , False ,
		   XmNinitialResourcesPersistent , False ,
		   NULL ) ;
   XmStringFree(xstr) ;
   XtAddCallback( ftm_anat_choose_pb, XmNactivateCallback, DIFF_choose_CB, "OVER" ) ;
   MCW_register_help( ftm_anat_choose_pb ,
                      "Use this to popup a\n"
                      "'chooser' that lets\n"
                      "you select which\n"
                      "dataset to track." ) ;

   /*** button to let user choose dataset to track ***/

   xstr = XmStringCreateLtoR( "Choose FTM (DTI) Dataset" , XmFONTLIST_DEFAULT_TAG ) ;
   ftm_choose_pb = XtVaCreateManagedWidget(
                   "AFNI" , xmPushButtonWidgetClass , fiber_rowcol ,
		   XmNalignment   , XmALIGNMENT_CENTER ,
		   XmNpacking , XmPACK_TIGHT ,
		   XmNlabelString , xstr ,
		   XmNtraversalOn , False ,
		   XmNinitialResourcesPersistent , False ,
		   NULL ) ;
   XmStringFree(xstr) ;
   XtAddCallback( ftm_choose_pb, XmNactivateCallback, DIFF_choose_CB, "OVER" ) ;
   MCW_register_help( ftm_choose_pb ,
                      "Use this to popup a\n"
                      "'chooser' that lets\n"
                      "you select which\n"
                      "dataset to track." ) ;

   /*--- Fiber methods choices ---*/
#ifdef EXPERIMENTAL
   fiber_compute_bbox = 
       new_LRF_bbox( fiber_rowcol ,
		     NBUT_FTMC , FTMC_labels ,
		     MCW_BB_radio_zero ,
		     MCW_BB_noframe ,
		     LRF_BB_vert ,
		     DIFF_fiber_compute_chan_CB , NULL ) ;
#endif
   /*** pushbutton to activate the fibtmap computation ***/

   fiber_compute_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , fiber_rowcol,
            XmNorientation , XmHORIZONTAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   xstr = XmStringCreateLtoR( "Initiate" , XmFONTLIST_DEFAULT_TAG ) ;
   fiber_compute_pb = XtVaCreateManagedWidget(
                        "AFNI" , xmPushButtonWidgetClass , 
			fiber_compute_rowcol ,
                        XmNlabelString , xstr ,
                        XmNtraversalOn , False ,
                        XmNinitialResourcesPersistent , False ,
                     NULL ) ;
   XmStringFree(xstr) ;
   XtAddCallback( fiber_compute_pb, XmNactivateCallback, DIFF_fiber_compute_CB, NULL ) ;
   MCW_register_help( fiber_compute_pb ,
                      "Use this to start the\n"
                      "fiber tract mapping computation" ) ;

   /*** pushbutton to cancel the diffusion [not managed now] ***/

   xstr = XmStringCreateLtoR( " * CANCEL * " , XmFONTLIST_DEFAULT_TAG ) ;
   fiber_cancel_pb = XtVaCreateWidget(
                        "AFNI" , xmPushButtonWidgetClass , 
			fiber_compute_rowcol ,
                        XmNlabelString , xstr ,
                        XmNtraversalOn , False ,
                        XmNinitialResourcesPersistent , False ,
                     NULL ) ;
   XmStringFree(xstr) ;
   XtAddCallback( fiber_cancel_pb, XmNactivateCallback, DIFF_fiber_cancel_CB, NULL ) ;

   XtManageChild( fiber_compute_rowcol ) ;
   XtManageChild( fiber_rowcol ) ;
   XtManageChild( fiber_frame ) ;

   /*--- Controls for FTM using SHD ---*/

   fibtmap_frame =
      XtVaCreateWidget(
         "AFNI" , xmFrameWidgetClass , fchoices_rowcol ,
            XmNshadowType , XmSHADOW_ETCHED_IN ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   fibtmap_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , fibtmap_frame ,
            XmNorientation , XmVERTICAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   fibtmap_label =
     XtVaCreateManagedWidget(
           "dialog" , xmLabelWidgetClass ,  fibtmap_rowcol,
            LABEL_ARG("Fiber Map Controls") ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   ftm_lval_av =
     new_MCW_arrowval( fibtmap_rowcol , "Max L" , MCW_AV_downup ,
		       0 , 20 , (int)(1.0*ftm_lval) ,
		       MCW_AV_edittext , 0 ,
		       reinterpret_cast<void (*)()>(DIFF_ftm_lval_av_CB),
		       NULL,NULL,NULL
		       ) ;
   MCW_reghelp_children(ftm_lval_av->wrowcol ,
			"Use this to set the maximum L value\n"
			"for the Spherical Harmonic Transform"
			) ;

   ftm_mval_av =
     new_MCW_arrowval( fibtmap_rowcol , "Max M" , MCW_AV_downup ,
		       0 , 20 , (int)(1.0*ftm_mval),
		       MCW_AV_edittext , 0 ,
		       reinterpret_cast<void (*)()>(DIFF_ftm_mval_av_CB),
		       NULL,NULL,NULL
		       ) ;
   MCW_reghelp_children(ftm_mval_av->wrowcol ,
			"Use this to set the maximum M value\n"
			"for the Spherical Harmonic Transform"
			) ;

   ftm_grid_av =
     new_MCW_arrowval( fibtmap_rowcol , "Grid size" , MCW_AV_downup ,
		       50 , 200 , (int)(1.0*ftm_grid),
		       MCW_AV_edittext , 0 ,
		       reinterpret_cast<void (*)()>(DIFF_ftm_grid_av_CB),
		       NULL,NULL,NULL
		       ) ;
   ftm_grid_av->fstep = ftm_grid_fstep;
   MCW_reghelp_children(ftm_grid_av->wrowcol ,
			"Use this to set the grid size\n"
			"for the Spherical Harmonic Transform"
			) ;
   XtAddCallback( ftm_grid_av->wtext, XmNactivateCallback, DIFF_textact_CB, roll_av ) ;

#ifdef EXPERIMENTAL
   XtManageChild( fibtmap_rowcol ) ;
   XtManageChild( fibtmap_frame ) ;

   RC_SENSITIZE( fibtmap_rowcol , False ) ; 
#endif

   /*--- Anatomical underlay for FTM ---*/
#ifdef UNNECESSARY
   fiber_anat_frame =
      XtVaCreateWidget(
         "AFNI" , xmFrameWidgetClass , fchoices_rowcol ,
            XmNshadowType , XmSHADOW_ETCHED_IN ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   fiber_anat_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , fiber_anat_frame ,
            XmNorientation , XmVERTICAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   fiber_anat_label =
     XtVaCreateManagedWidget(
           "dialog" , xmLabelWidgetClass ,  fiber_anat_rowcol,
            LABEL_ARG("FTM anatomical underlay") ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   MCW_register_help( fiber_anat_label ,
                      "Anatomicla dataset for fiber tract mapping.\n\n" );


   /*** button to let user choose associated anatomical dataset ***/

   xstr = XmStringCreateLtoR( "Choose Anatomical Dataset" , XmFONTLIST_DEFAULT_TAG ) ;
   ftm_anat_choose_pb = XtVaCreateManagedWidget(
                   "AFNI" , xmPushButtonWidgetClass , fiber_anat_rowcol ,
		   XmNalignment   , XmALIGNMENT_CENTER ,
		   XmNpacking , XmPACK_TIGHT ,
		   XmNlabelString , xstr ,
		   XmNtraversalOn , False ,
		   XmNinitialResourcesPersistent , False ,
		   NULL ) ;
   XmStringFree(xstr) ;
   XtAddCallback( ftm_anat_choose_pb, XmNactivateCallback, DIFF_choose_CB, "OVER" ) ;
   MCW_register_help( ftm_anat_choose_pb ,
                      "Use this to popup a\n"
                      "'chooser' that lets\n"
                      "you select which\n"
                      "dataset to track." ) ;

   XtManageChild( fiber_anat_rowcol ) ;
   XtManageChild( fiber_anat_frame ) ;
#endif /* UNNECESSARY */

   /* manage all ftm now */

   XtManageChild( fchoices_rowcol ) ;

   /***======================== lrf's diffusion theory ==========================*/

   /*** horizontal rowcol to hold miscellaneous display controls ***/

   SEP_HOR(anat_rowcol) ;  /* separator */

   hrc =  XtVaCreateWidget(
           "AFNI" , xmRowColumnWidgetClass , anat_rowcol ,
              XmNorientation , XmHORIZONTAL ,
              XmNpacking , XmPACK_TIGHT ,
              XmNadjustLast  , False ,
              XmNadjustMargin, False ,
              XmNtraversalOn , False ,
              XmNmarginWidth , 0 ,
              XmNmarginHeight, 0 ,
              XmNinitialResourcesPersistent , False ,
           NULL ) ;

   /*** option menu to choose diffusion processing method ***/

   theory_av = new_MCW_optmenu( hrc , "Theory " ,
				0 , NUM_theory-1 , theory_val,0 ,
				reinterpret_cast<void (*)()>
				(DIFF_theory_CB),
				NULL ,
				reinterpret_cast<char* (*)()>
				(MCW_av_substring_CB),
				theory_strings ) ;

   MCW_reghelp_children( theory_av->wrowcol ,
                         "Use this to set the choose the theory type"
                       ) ;

   SEP_VER(hrc) ;  /* separator */

   /*** pushbutton to activate the diffusion computation ***/

   xstr = XmStringCreateLtoR( "Show" , XmFONTLIST_DEFAULT_TAG ) ;
   show_theory_pb = XtVaCreateManagedWidget(
                     "AFNI" , xmPushButtonWidgetClass , hrc ,
                        XmNlabelString , xstr ,
                        XmNtraversalOn , False ,
                        XmNinitialResourcesPersistent , False ,
                     NULL ) ;
   XmStringFree(xstr) ;
   XtAddCallback( show_theory_pb, XmNactivateCallback, DIFF_show_theory_CB, NULL ) ;
   MCW_register_help( show_theory_pb ,
                      "Use this to show theory info" );

   XtManageChild(hrc) ;

   /***=============================================================*/

   /*** a set of action buttons below the line ***/

   SEP_HOR(anat_rowcol) ;

   (void) MCW_action_area( anat_rowcol , DIFF_actor , NACT ) ;

   help_pb      = (Widget) DIFF_actor[0].data ;
   copyright_pb = (Widget) DIFF_actor[1].data ;
   reload_pb    = (Widget) DIFF_actor[2].data ;
   done_pb      = (Widget) DIFF_actor[3].data ;

   /***=============================================================*/

   /*** that's all ***/

   XtManageChild(anat_rowcol) ;
   XtManageChild(anat_frame) ;

   XtManageChild(top_rowcol) ;
   XtRealizeWidget(shell) ;      /* will not be mapped */

   /*** 12 July 1999: make the overlay widgets now, instead of later ***/

   DIFF_func_dti_widgets() ;
   DIFF_func_shd_widgets() ;
   DIFF_func_ftm_widgets() ;

#if 0
   XtVaSetValues( anat_rowcol , XmNresizeWidth , False , NULL ) ;
#endif
   return ;
}

/*-------------------------------------------------------------------
  Callback for done button
---------------------------------------------------------------------*/

static int quit_first = 1 ;

void DIFF_done_timeout_CB( XtPointer client_data , XtIntervalId * id )
{
   MCW_set_widget_label( done_pb , "done" ) ;
   quit_first = 1 ;
   return ;
}

void DIFF_done_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   /** like AFNI itself, require two quick presses to exit **/

   if( w == done_pb && quit_first && diffusings != NULL ){
      MCW_set_widget_label( done_pb , "DONE " ) ;
      quit_first = 0 ;
      (void) XtAppAddTimeOut(
               XtWidgetToApplicationContext(done_pb) ,
               5000 , DIFF_done_timeout_CB , NULL ) ;
      return ;
   }

   if( xhair_recv >= 0 )  /* 29 Mar 1999 */
      AFNI_receive_control( im3d, xhair_recv,EVERYTHING_SHUTDOWN, NULL ) ;

   DIFF_destroy_imseq() ;      /* destroy the image window */

   if( wshdfunc_frame != NULL && XtIsManaged(wshdfunc_frame) )  /* close overlay */
      DIFF_open_shdfunc_CB(NULL,NULL,NULL) ;

   XtUnmapWidget( shell ) ; diffuser_open = 0 ; imseq = NULL ;

   if( dset      != NULL ) dset      = NULL ;
   if( func_dset != NULL ) func_dset = NULL ;

   FREE_VOLUMES ; INVALIDATE_OVERLAY ;
   MPROBE ;
   return ;
}

/**<load_gradient_file>***************************************************

   Purpose: Read and register gradient files 
      Call: int load_gradient_file( float *g , char *gdname, int skip1 )
    Return: 1 if   successfull
            0 if unsuccessfull

      Note: skip1 allows skipping over the 1st data point.
            This is necessary for data collected with ses13, 
	    which unwisely stored G=(0,0,0) as 1st data point.
	    hyper.e currently also stores (0,0,0) ; 02_02_21 lrf
*/
int load_gradient_file( float *g , char *gdname , int skip1 )
{
  char buf[256] ;
  char *text = NULL ;
  MRI_IMAGE *gdim = NULL, *flim ;
  float *far ;
  int ii,jj ;
  int debug = 0;
  int nskip = 0;

  gdim = mri_read_ascii( gdname ) ;
  if( gdim == NULL || gdim->ny < 2 ){
    sprintf(buf," \n"
	    "** Cannot read gradient file %s\n", gdname ) ;
    (void) MCW_popup_message( choose_pb , buf ,
			      MCW_USER_KILL | MCW_TIMER_KILL ) ;
    PLUTO_beep() ; return(0);
  }

  flim = mri_transpose(gdim) ; mri_free(gdim) ;
  far  = MRI_FLOAT_PTR(flim) ;

  /* Check limits and load grad values into dif */
  if(debug) 
    printf("load_gradient_file: Loading GX (nx=%i,ny=%i)...\n",
	   gdim->nx,gdim->ny);

  if(skip1) nskip = 1; 
  for( ii=0 ; ii < (flim->nvox - nskip) ; ii++ ) {
    jj = ii + nskip ;
    if( fabs(far[jj]) >= 33333.0 ) far[jj] = WAY_BIG ;
    g[ii] = far[jj] ;
    if(debug) printf("DIFF_reload_dataset: G[%i]=%f\n",ii,g[ii]);
  }
  if(debug) printf("done\n");
   
  PLUTO_register_timeseries( gdname , flim ) ; 
  mri_free(flim) ;
  printf("Successfully loaded gradient file %s\n",gdname);
  return(1);
}

void PrintVfileHeaderSummary( char *vheader )
{
  printf("Varian header contents:\n");
  
}

/*-------------------------------------------------------------------
   Load the data from the dataset into local arrays

   History: 06-Jan-02 lrf, now eliminates 1st grad pt G=(0,0,0) 
                           from gradient file if generated by ses13.
---------------------------------------------------------------------*/

int DIFF_reload_dataset(void)
{
   char *program_name = "DIFF_reload_dataset" ;

   int ii , nvox , vmin,vmax , cbot,ctop , ival,val , cutdone ;
   float fac ;
   void * var ;
   byte * gar ;
   MRI_IMAGE * vim ;
   XmString xstr ;
   char str[64] , buf[256] ;
   char gdxname[80],gdyname[80],gdzname[80];
   char headername[80], testname[80];
   int debug = 0;
   int skip1 = 0;
   char *scanner_type;

#define DO_3DSFD 0

   MCW_invert_widget(reload_pb) ;        /* flash a signal */

   /* start by tossing any old data */

#if !DO_3DSFD

   FREE_VOLUMES ;


   /* Check for existence of gradient files */
#if 0
   sprintf(gdxname,"%s.GDX",DSET_PREFIX(dset));
   sprintf(gdyname,"%s.GDY",DSET_PREFIX(dset));
   sprintf(gdzname,"%s.GDZ",DSET_PREFIX(dset));

   if(!THD_is_file(gdxname)) {	/* x grad */
     sprintf(buf," \n"
	     "** Corresponding gradient file %s\n"
	     "** doesn't exist!\n" , gdxname ) ;
     (void) MCW_popup_message( choose_pb , buf ,
			       MCW_USER_KILL | MCW_TIMER_KILL ) ;
     PLUTO_beep() ; 
     MCW_invert_widget(reload_pb) ;        /* flash a signal */
     return(0);
   };
   if(!THD_is_file(gdyname)) {	/* y grad */
     sprintf(buf," \n"
	     "** Corresponding gradient file %s\n"
	     "** doesn't exist!\n" , gdyname ) ;
     (void) MCW_popup_message( choose_pb , buf ,
			       MCW_USER_KILL | MCW_TIMER_KILL ) ;
     PLUTO_beep() ; 
     MCW_invert_widget(reload_pb) ;        /* flash a signal */
     return(0);
   };
   if(!THD_is_file(gdzname)) {	/* z grad */
     sprintf(buf," \n"
	     "** Corresponding gradient file %s\n"
	     "** doesn't exist!\n" , gdzname ) ;
     (void) MCW_popup_message( choose_pb , buf ,
			       MCW_USER_KILL | MCW_TIMER_KILL ) ;
     PLUTO_beep() ; 
     MCW_invert_widget(reload_pb) ;        /* flash a signal */
     return(0);
   };
#endif
#endif

   /* 19-Feb-03 Scanner type
      Currently only Varian type provides input.
      Need to alter dif2afni.c to put 
      SCANNER_TYPE_GE_LX in first note. */

   scanner_type = tross_Get_Note(dset,1);
   if(scanner_type==NULL) {
     scanner_type = (char *) calloc(80,sizeof(char));
     strcpy(scanner_type,"SCANNER_TYPE_GE_LX");
   }


   printf(" ************** HARDWIRING SCANNER TYPE TO GE_EXCITE for 3dsfd!!!! ************\n");
   strcpy(scanner_type,"SCANNER_TYPE_GE_EXCITE");

   if(!strcmp(scanner_type,"SCANNER_TYPE_VARIAN"))
     printf("Varian data detected.\n");
   if(!strcmp(scanner_type,"SCANNER_TYPE_GE_LX"))
     printf("GE LX data detected.\n");
   if(!strcmp(scanner_type,"SCANNER_TYPE_GE_EXCITE"))
     printf("GE EXCITE data detected.\n");

   /* end of scanner type */

   /* Check for existence of header file */
   sprintf(headername,"%s.header",DSET_PREFIX(dset));

   if (!THD_is_file(headername)) {	/* x grad */
     if (strcmp(scanner_type,"SCANNER_TYPE_GE_EXCITE")) {
       printf("** Corresponding header file %s not found\n" , headername ) ;
     }
     header_exists = 0;
   } else {
     printf("** Found header file %s\n" , headername ) ;
     header_exists = 1;
   };

   /* Read pfile header, if it exists */
   if(header_exists || !strcmp(scanner_type,"SCANNER_TYPE_GE_EXCITE")) {

     FILE *fin;
     int headlen ;

     if(!strcmp(scanner_type,"SCANNER_TYPE_GE_LX")) {

       char img_extent[256];

       if (!(fin=fopen(headername,"r"))) {
	 printf("Can't open header file %s!\n",headername);
	 dset = NULL;
	 return(0);
       }
       if ( (headlen=fread(nheader, sizeof(char), RDB_HEADER_SIZE_BYTES, fin))
	    != RDB_HEADER_SIZE_BYTES)
	 printf("header length %i too short (< %i)!\n",
		headlen,(int) RDB_HEADER_SIZE_BYTES);

       printf("Successfully read header file %s\n",headername);

#ifdef LINUX2
       printf("Linux system detected - swapping bytes in header ...\n");
       SwapPfileHeader(nheader);
#endif

       /* Print header info */
       PrintPfileHeaderSummary( nheader, img_extent );
       printf("Image extent: %s\n",img_extent);
       fclose(fin);

     } else if (!strcmp(scanner_type,"SCANNER_TYPE_GE_EXCITE")) {

       /* This one is done differently.
	  Load header name into header and
	  actually read in the header in init_img_state_GE_EXCITE
	  in init_dif_state.c */
       /* Actually, all .header data can now be in the AFNI header. */
       if (header_exists) {
	 sprintf(nheader,"file:%s",headername);
       }
       else {
	 if (tross_Get_Notecount(dset) >= 2) {
	   sprintf(nheader,"%s", tross_Get_Note(dset,2));
	 }
	 else {
	   sprintf(buf," \n"
		   "** Necessary header information in missing.\nYou probably need a .header file.**\n");
	   (void) MCW_popup_message( choose_pb , buf ,
				     MCW_USER_KILL | MCW_TIMER_KILL ) ;
	   dset = NULL;
	   return(0);
	 }
       }
     } else if (!strcmp(scanner_type,"SCANNER_TYPE_VARIAN")) {

       int VAR_HEADER_SIZE_BYTES = 1000 ;
       if (!(fin=fopen(headername,"r"))) {
	 printf("Can't open rawfile %s!\n",headername);
	 dset = NULL;
	 return(0);
       }
       if ((headlen=fread(nheader, sizeof(char), VAR_HEADER_SIZE_BYTES, fin))
	   != VAR_HEADER_SIZE_BYTES) 
	 { printf("header length %i too short (< %i)!\n",
		  headlen,VAR_HEADER_SIZE_BYTES);
	 }
       printf("Successfully read header file %s\n",headername);
       PrintVfileHeaderSummary( nheader );
       fclose(fin);

     } else {
       printf("Unknown scanner type %s!\n",scanner_type);
       exit(0);
     }


   } else {
     char *str = (char *) malloc( 1024 ) ;
       sprintf( str, "\nDIFF_reload_dataset: No header found!\n"
		"Proceed at your own risk!\n") ;
       PLUTO_popup_message(plint,str) ; free(str) ; 
       free(str) ;
       return(0);
   }

   /* Initialize the diffusion state */

   /* cyrus SHD/DTI crash */
   printf("Initializing dif_state (scanner_type = %s)\n",scanner_type);
   current_dif_state = init_dif_state( nheader, b0_dstats, scanner_type );
   if (!current_dif_state) {
     current_dif_state = init_dif_state( EMPTY , NULL, EMPTY );
     return(0);
   }
   printf(" ... done\n");

   /* Read and register gradient files */

   /* no provision made here yet for bad read (just message)
      maybe give option to regenerate directions in gdipat if
      no grads supplied? */


   /* Right now, only on excite is the pulse do the grad directions
      not include (0,0,0), (as it should be!) */
   if(MATCH_STR(scanner_type,"SCANNER_TYPE_GE_EXCITE")) {
     skip1 = 0;
   } else {
     printf("scanner_type is %s, so skipping 1st gradient pt (={0,0,0})\n",
	    scanner_type);
     skip1 = 1;
   }

   IMG_state *img = current_dif_state->img ;

   /* If file type is epi2 then hard code sampling type in case this
      was recorded incorrectly in the header. */
   if(MATCH_STR(img->psd_name,"epi2")) {
     img->difmode = DIFF_SAMP_GE;
   }

   char *default_tensor_path = getenv("AFNI_PLUGINPATH");
   char gradient_file[256] ;

   /* Set tensor file based on sampling type. */
   switch (img->difmode) {
   case DIFF_SAMP_NONE:
   case DIFF_SAMP_XYZ0:
     printf("Unsupported diffusion mode %i: unable to load this dataset.\n",
	    img->difmode);
     return(0);
     break;
   case DIFF_SAMP_GE:
     sprintf(gradient_file,"%s/tensor.dat",default_tensor_path);
     break;
   case DIFF_SAMP_TESS:
     if (img->dgrad->polytope == POLY_ICO) {
       sprintf(gradient_file,"%s/tensor_tess_ico.dat",default_tensor_path);
     }
     else {
       printf("Invalid input polytope type %i: unable to load this dataset.\n",
	      img->dgrad->polytope);
       return(0);
     }
     break;
   default:
     printf("Unknown sampling mode %d: unable to load this dataset.\n",
	    img->difmode);
     return(0);
   }

   /* Read tensor file to set gradient directions. */
   FILE *gfp;
   if ((gfp = fopen(gradient_file,"r")) == NULL) {
     printf("Problem opening gradient file %s\n",gradient_file);
     exit(0);
   }

#define MAXLEN 150

   int nvert = img->dgrad->nvert; /* set from .header file */
   char gfile_line[MAXLEN];
   int gfile_nvert = -1;
   int start_read = 0;
   float gxval,gyval,gzval;
   int gcount = 0;
   float *gdx = (float *)calloc(nvert, sizeof(float));
   float *gdy = (float *)calloc(nvert, sizeof(float));
   float *gdz = (float *)calloc(nvert, sizeof(float));

   char tempstring[MAXLEN];           /* buffer to access file */

   int read_skip = 1;
   char compstring[MAXLEN];           /* buffer used to open data file */ 
   long floc ;
   sprintf( compstring, "%d", nvert );
   int num_tensor_len = strlen( compstring );

   while( read_skip ) {
     floc = ftell(gfp) ;
     fgets( tempstring, MAXLEN, gfp );
     read_skip = strncmp( compstring, tempstring, num_tensor_len );
   }
   fseek(gfp,floc,SEEK_SET);

   while (nvert>gcount && fgets(gfile_line, MAXLEN, gfp) != NULL) {

     if(debug) printf("gfile_line = %s",gfile_line);
     if(!strncmp(gfile_line,"#",1)) {
       if(debug) printf("Comment line\n");
     } else if (!start_read) {
       sscanf(gfile_line,"%d",&gfile_nvert);
       if(debug) printf("gfile_nvert = %i\n",gfile_nvert);
     }

     if(start_read) {
       sscanf(gfile_line,"%f %f %f",&gdx[gcount],&gdy[gcount],&gdz[gcount]);
       if(debug) printf("gd = (%f,%f,%f)\n",gdx[gcount],gdy[gcount],gdz[gcount]);
       gcount++;
     }
     if(gfile_nvert==nvert) start_read = 1;

   }
   if(gcount!=nvert){
     printf("Number of gradient directions read (%i) not equal to nvert from header (%i)!!\a\n",
	    gcount,nvert);
     dset = NULL;
     return(0);
   }

   /* Set gradient directions just read. */
   for (ii=0; ii<nvert; ii++) {
     img->GX[ii] = gdx[ii];
     img->GY[ii] = gdy[ii];
     img->GZ[ii] = gdz[ii];
   }

   /* Gradients read, let setPattern in setB know ... */
   img->bpat_available = 1 ;

   /* The following was originally only for epi2 and 3dsfd, but now it
      should work for all types of scans. */
   /*   if(MATCH_STR(img->psd_name,"epi2")) { */
   if (1) {

     for (int i=0;i<img->dgrad->nvert;i++) {
       img->dgrad->gx[i] = img->GX[i];
       img->dgrad->gy[i] = img->GY[i];
       img->dgrad->gz[i] = img->GZ[i];
     }
     for (int i=0;i<img->nb0;i++) {
       img->gdxpat[i] = 0.0 ;
       img->gdypat[i] = 0.0 ;
       img->gdzpat[i] = 0.0 ;
     }
     for (int i=0;i<img->dgrad->nvert;i++) {
       img->gdxpat[i+img->nb0] = img->dgrad->gx[i] ;
       img->gdypat[i+img->nb0] = img->dgrad->gy[i] ;
       img->gdzpat[i+img->nb0] = img->dgrad->gz[i] ;
     }

     float *xyz = (float *) malloc(sizeof(float) * (3*img->dgrad->nvert));
     for (int i = 0; i < img->dgrad->nvert; i++) {
       xyz[3*i] = img->dgrad->gx[i];
       xyz[3*i+1] = img->dgrad->gy[i];
       xyz[3*i+2] = img->dgrad->gz[i];
     }
     img->dgrad->ntris = qhull_wrap(img->dgrad->nvert, xyz, &img->dgrad->tris);
     free(xyz);
     vGrid = vtk_LoadGradients(img->dgrad->gx,
			       img->dgrad->gy,
			       img->dgrad->gz,
			       img->dgrad->tris,
			       img->dgrad->ntris,
			       img->dgrad->nvert) ;
   } else {
     /* Dead code ... */

     /* Load up tris via sphtes */
     setPattern(img);

     vGrid = vtk_LoadGradients(img->dgrad->gx,
			       img->dgrad->gy,
			       img->dgrad->gz,
			       img->dgrad->tris,
			       2*(img->dgrad->nvert-2),
			       img->dgrad->nvert) ;
   }
     

   /* Dead code ... */
#if 0   
   if(MATCH_STR(img->psd_name,"3dsfd")) {
     /* Big hardwired hack for 3dsfd! */


     for (int i=0;i<img->nb0;i++) {
       img->gdxpat[i] = 0;
       img->gdypat[i] = 0;
       img->gdzpat[i] = 0;
     }
     for (int i=0;i<img->dgrad->nvert;i++) {
       printf("{gx,gy,gz} = {%f,%f,%f}\n",
	      img->dgrad->gx[i],
	      img->dgrad->gy[i],
	      img->dgrad->gz[i]);
       img->GX[i] = img->dgrad->gx[i];
       img->GY[i] = img->dgrad->gy[i];
       img->GZ[i] = img->dgrad->gz[i];
       img->gdxpat[i+img->nb0] = img->dgrad->gx[i] ;
       img->gdypat[i+img->nb0] = img->dgrad->gy[i] ;
       img->gdzpat[i+img->nb0] = img->dgrad->gz[i] ;
     }
   }
#endif

   /* This should be unnecessary now: these match by definition. */
#if !DO_3DSFD
   /* So let's make sure grads match up! 
      Check to 4 decimal places.  */
   
   int gdx_test, gdy_test, gdz_test;
   int gpx_test, gpy_test, gpz_test;

   for (int i=0; i<img->dgrad->nvert; i++) {  

     gdx_test = (int) (10000.*img->GX[i]);
     gdy_test = (int) (10000.*img->GY[i]);
     gdz_test = (int) (10000.*img->GZ[i]);
     gpx_test = (int) (10000.*img->gdxpat[i+img->nb0]) ;
     gpy_test = (int) (10000.*img->gdypat[i+img->nb0]) ;
     gpz_test = (int) (10000.*img->gdzpat[i+img->nb0]) ;

     if( (gdx_test != gpx_test) || (gdy_test != gpy_test) || (gdz_test != gpz_test) ) 
       {
	 printf("%s: Mismatch in grads found!\n\a",program_name);
	 printf("img->G{X,Y,Z} = (%f,%f,%f), img->gdpat{X,Y,Z} = (%f,%f,%f)\n",
		img->GX[i],
		img->GY[i],
		img->GZ[i],
		img->gdxpat[i+1],
		img->gdypat[i+1],
		img->gdzpat[i+1]) ;
	 printf("Exiting ...\n");
	 assert(false);
       }

     if(debug) printf("img->G{X,Y,Z} = (%f,%f,%f), img->gdpat{X,Y,Z} = (%f,%f,%f)\n",
		      img->GX[i],
		      img->GY[i],
		      img->GZ[i],
		      img->gdxpat[i+1],
		      img->gdypat[i+1],
		      img->gdzpat[i+1]) ;
    }
#endif

   /* Dead code ... */
   if(0) {	       /* TEST SECTION TO SEE IF I CAN EXTRACT FILE BY NAME */
     int ii,its;
     MRI_IMAGE *tstim = NULL;
     float *ts ;
     int debug = 0;

     sprintf(testname,"%s",gdzname);
     tstim = get_imarr_by_name(gdxname); 
     if( tstim == NULL ){
       sprintf(buf," \n"
	       "** Cannot find ts %s\n", testname ) ;
       (void) MCW_popup_message( choose_pb , buf ,
				 MCW_USER_KILL | MCW_TIMER_KILL ) ;
       PLUTO_beep() ; return(0);
     } 

     ts = MRI_FLOAT_PTR(tstim) ;
     if(debug) {
       for( ii=0 ; ii < tstim->nx ; ii++ )
	 printf("[%i] = %f\n",ii,ts[ii]);  
     }
   }

   /* Make sure the dataset is in memory */

   DSET_load(dset) ;
   vim = DSET_BRICK(dset,dset_ival) ; nvox = vim->nvox ;
   var = DSET_ARRAY(dset,dset_ival) ; brickfac = DSET_BRICK_FACTOR(dset,dset_ival) ;

   /* find data range, clip it, convert to bytes */

   grim = mri_new_conforming( vim , MRI_byte ) ;  /* new image data */
   gar  = MRI_BYTE_PTR(grim) ;

   switch( DSET_BRICK_TYPE(dset,dset_ival) ){

      case MRI_short:{
         short * sar = (short *) var ;

         vmin = vmax = sar[0] ;
         for( ii=1 ; ii < nvox ; ii++ ){
            val = sar[ii] ;
                 if( vmin > val ) vmin = val ;
            else if( vmax < val ) vmax = val ;
         }

#ifdef EXPERIMENTAL
         if( new_dset ){
            AV_assign_ival( clipbot_av , vmin ) ; cbot = vmin ;
            AV_assign_ival( cliptop_av , vmax ) ; ctop = vmax ;
         } else {
            cbot = MAX( clipbot_av->ival , vmin ) ;
            ctop = MIN( cliptop_av->ival , vmax ) ;
         }
#endif /* EXPERIMENTAL */

         fac  = (ctop > cbot) ? 255.9/(ctop-cbot) : 1.0 ;
         for( ii=0 ; ii < nvox ; ii++ ){
            val  = sar[ii] ;
            ival = (int) (fac * (val-cbot)) ; RANGE(ival,0,255) ; gar[ii] = ival ;
         }
      }
      break ;

      case MRI_byte:{
         byte * bar = (byte *) var ;

         vmin = vmax = bar[0] ;
         for( ii=1 ; ii < nvox ; ii++ ){
            val = bar[ii] ;
                 if( vmin > val ) vmin = val ;
            else if( vmax < val ) vmax = val ;
         }

#ifdef EXPERIMENTAL
         if( new_dset ){
            AV_assign_ival( clipbot_av , vmin ) ; cbot = vmin ;
            AV_assign_ival( cliptop_av , vmax ) ; ctop = vmax ;
         } else {
            cbot = MAX( clipbot_av->ival , vmin ) ;
            ctop = MIN( cliptop_av->ival , vmax ) ;
         }
#endif /* EXPERIMENTAL */

         fac  = (ctop > cbot) ? 255.9/(ctop-cbot) : 1.0 ;
         for( ii=0 ; ii < nvox ; ii++ ){
            val  = bar[ii] ;
            ival = (int) (fac * (val-cbot)) ; RANGE(ival,0,255) ; gar[ii] = ival ;
         }
      }
      break ;
   }

   /* set label showing data range */

   sprintf(str,"Min=%d Max=%d",vmin,vmax) ;
#ifdef EXPERIMENTAL
   xstr = XmStringCreateLtoR( str , XmFONTLIST_DEFAULT_TAG ) ;
   XtVaSetValues( range_lab , XmNlabelString , xstr , NULL ) ;
   XmStringFree(xstr) ;
#endif

   /* if brick is scaled, show the scaled labels */

   HIDE_SCALE ;

#ifdef EXPERIMENTAL
   if( brickfac != 0.0 && brickfac != 1.0 ){
      char minch[16] , maxch[16] ;

      AV_fval_to_char( vmin*brickfac , minch ) ;
      AV_fval_to_char( vmax*brickfac , maxch ) ;
      sprintf(str,"[%s %s]",minch,maxch) ;
      xstr = XmStringCreateLtoR( str , XmFONTLIST_DEFAULT_TAG ) ;
      XtVaSetValues( range_faclab , XmNlabelString , xstr , NULL ) ;
      XmStringFree(xstr) ;

      AV_fval_to_char( brickfac * clipbot_av->ival , minch ) ;
      sprintf(str,"[-> %s]",minch) ;
      xstr = XmStringCreateLtoR( str , XmFONTLIST_DEFAULT_TAG ) ;
      XtVaSetValues( clipbot_faclab , XmNlabelString , xstr , NULL ) ;
      XmStringFree(xstr) ;

      AV_fval_to_char( brickfac * cliptop_av->ival , maxch ) ;
      sprintf(str,"[-> %s]",maxch) ;
      xstr = XmStringCreateLtoR( str , XmFONTLIST_DEFAULT_TAG ) ;
      XtVaSetValues( cliptop_faclab , XmNlabelString , xstr , NULL ) ;
      XmStringFree(xstr) ;

      XtManageChild( range_faclab   ) ;
      XtManageChild( clipbot_faclab ) ;
      XtManageChild( cliptop_faclab ) ;
   } else {
      XtUnmanageChild( range_faclab   ) ;
      XtUnmanageChild( clipbot_faclab ) ;
      XtUnmanageChild( cliptop_faclab ) ;
   }
#endif

   FIX_SCALE_SIZE ;

   /* copy image data into opacity */

   opim = mri_to_byte( grim ) ;

   {  int hist[256] , nvox = grim->nvox , htop , ii ;
   /* float ofac = current_cutout_state.opacity_scale ; */

      /* do the histogram graph */
#ifdef EXPERIMENTAL
      MCW_histo_bytes( nvox , MRI_BYTE_PTR(grim) , hist ) ;

      htop = 0 ;
      for( ii=0 ; ii < GRAF_SIZE ; ii++ ){
         hist[ii] = hist[2*ii] + hist[2*ii+1]  ;            /* fold 2 into 1  */
         if( ii > 0 && hist[ii] > htop ) htop = hist[ii] ;  /* find max value */
      }

      if( htop == 0 ){
         set_MCW_pasgraf( his_graf , NULL ) ;
      } else {
         float scl = (GRAF_SIZE-0.44) / sqrt((double)htop) ;
         byte bhis[GRAF_SIZE] ;
         for( ii=0 ; ii < GRAF_SIZE ; ii++ )
            bhis[ii] = (hist[ii] > htop) ? GRAF_SIZE-1
                                         : (byte)(scl*sqrt((double)hist[ii])+0.49) ;

         set_MCW_pasgraf( his_graf , bhis ) ;
      }

      redraw_MCW_pasgraf( his_graf ) ;
#endif
   }
   /*--- Now deal with overlay, if need be ---*/

   func_computed = 0 ;  /* at this point, data is for grayscale diffusing */
   cutdone       = 0 ;  /* cutouts not done yet */

   if( DO_OVERLAY ){
      byte * gar , * opar , * ovar ;
      int nvox = grim->nvox , ii ;

      if( ovim == NULL ) DIFF_reload_func_dset(im3d,im3d->vinfo->fim_index) ;

      ovar = MRI_BYTE_PTR(ovim) ;

      if( !func_showthru ){  /* the old code: embed color into volume */

        gar  = MRI_BYTE_PTR(grim) ;
        opar = MRI_BYTE_PTR(opim) ;

        /* convert gar into index into the functional colormap */

        for( ii=0 ; ii < nvox ; ii++ ){
           if( ovar[ii] == 0 ) gar[ii] = gar[ii] >> 1   ;  /* gray */
           else                gar[ii] = 127 + ovar[ii] ;  /* color */
        }

        /* if needed, modify the opacity where there is color */

        if( func_color_opacity > 0.0 ){
           byte opac = (byte)(255.0 * MIN(func_color_opacity,1.0)) ;
           for( ii=0 ; ii < nvox ; ii++ )
              if( ovar[ii] != 0 ) opar[ii] = opac ;
        }

      } else {  /* 07 Jan 2000: make the showthru bricks instead */

        byte * garst , * oparst ;

        grim_showthru = mri_new_conforming( vim , MRI_byte ) ;
        opim_showthru = mri_new_conforming( vim , MRI_byte ) ;
        garst  = MRI_BYTE_PTR(grim_showthru) ;
        oparst = MRI_BYTE_PTR(opim_showthru) ;

        memset( garst  , 0 , sizeof(byte)*nvox ) ;
        memset( oparst , 0 , sizeof(byte)*nvox ) ;

        for( ii=0 ; ii < nvox ; ii++ ){       /* load values only if */
           if( ovar[ii] != 0 ){               /* there is overlay   */

              garst[ii]  = 127 + ovar[ii] ;   /* color index */
              oparst[ii] = 240 ;              /* mostly opaque */
           }
        }
      }

      func_computed = 1 ;  /* data is now set for color diffusing */
   }

   /*--- Other piddling details ---*/

   if( xhair_flag ) DIFF_xhair_overlay() ;

   MCW_invert_widget(reload_pb) ;  /* turn the signal off */

   new_dset = 0 ; new_data_loaded = 1 ;
   return(1);
}

/*-------------------------------------------------------------------
   Callback for the reload button
---------------------------------------------------------------------*/

void DIFF_reload_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   if( dset == NULL ){ XBell(dc->display,100) ; return ; }

   DIFF_reload_dataset() ;             /* load data again */

   return ;
}

/*-------------------------------------------------------------------
  Callback for help button
---------------------------------------------------------------------*/

void DIFF_help_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   (void ) new_MCW_textwin( info_lab ,
       "Version "
       ADP_MAJOR_VER
       "."
       ADP_SVN_REV
       "\n"
       "+++++++++++++++++++  D I F F U S I O N  A N A L Y S I S  ++++++++++++++++++\n"
       "\n"
       "This plugin is used to analyze diffusion weighted MRI data\n"
       "in a 3D dataset and overlay the (functional) results in color \n"
       "onto a grayscale (anatomical) dataset.\n"
       "\n"
       "General Notes:\n"
       "\n"
       "******************************** Input **********************************\n"
       "\n"
       "Input data structure:\n"
       "\n"
       "This plugin requires an AFNI 3D+diffusion dataset which can be\n"
       "generated from raw image files using the program 'dif2afni' that\n"
       "comes packaged with this plugin.  The raw data input into dif2afni\n"
       "is assumed to be in the standard UCSD format name.slice.dif where\n"
       "'.dif' means diffusion direction.  A set of gradient direction\n"
       "files name.01.gd[x,y,z] are also required.\n"
       "\n"
       "Note: For UCSD users collecting data with the pulse sequence 'ses.psd',\n"
       "the data format and the corresponding gradient direction files are\n"
       "automatically generated by the reconstruction program 'grses'.\n"
       "\n"
       "****************************** Anisotropy ********************************\n"
       "\n"
       "Anisotropy describes the existence of diffusion that depends upon the\n"
       "direction of measurement.  Several measures of anisotropy exist, but all\n"
       "are based on a comparison of the coefficients of expansion of the\n"
       "measured diffusion field in terms of some model.\n"
       "\n"
       "------------------------------------------------------------------------\n"
       "\n"
       "Diffusion Tensor Imaging (DTI) \n"
       "\n"
       "This is the standard model for analyzing anisotropic diffusion data.\n"
       "(e.g., Basser ..).\n"
       "This model is incorrect for voxels containing multiple fiber directions.\n"
       "(see Theory: Anisotropy in Multifiber Voxels)\n"
       "\n"
       "Algorithm:\n\n"
       "The diffusion tensor is calculated from an arbitrary number of\n"
       "measurements using the least-squares approach of Basser.\n"
       "(Basser ....).\n"
       "\n"
       "------------------------------------------------------------------------\n"
       "\n"
       "Spherical Harmonic Decomposition (SHD) \n"
       "\n"
       "This is a more general model for analyzing anisotropic diffusion data\n"
       "in voxels with multiple fiber directions (Frank, ....).\n"
       "(see Theory: Spherical Harmonic Decomposition)\n"
       "\n"
       "Algorithm:\n\n"
       "The spherical harmonic decomposition is acheived by calculating the\n"
       "spherical harmonic transform of the measured apparent diffusion\n"
       "coefficient. (Frank...)\n"
       "\n"
       "****************************** Orientation **********************************\n"
       "\n"
       "Orientation refers to the direction along which a fiber is pointing.\n"
       "For a single fiber voxel, this is usually chosen to be the direction\n"
       "of the largest (i.e., principal) eigenvector of the diffusion tensor.\n"
       "For multifiber voxels, the global and relative orientation of the fibers\n"
       "can be estimated if the number of fibers is known.\n"
       "\n"
       "------------------------------------------------------------------------\n"
       "\n"
       "Orientation from DTI \n"
       "\n"
       "------------------------------------------------------------------------\n"
       "\n"
       "Orientation from SHD \n"
       "\n"
       "************************** Fiber Tract Mapping ******************************\n"
       "\n"
       "Fiber tract mapping is the process of reconstructing fiber paths from\n"
       "the orientations of fibers in neighboring voxels.\n"
       "\n"
       "(This feature is currently unavailable)\n"
       "\n"
       "\n"
       "== == == == == == == == == == == == == == == == == == == == == == == == == ==\n"
       "\n"
       "Lawrence R. Frank, Ph.D.\n"
       "Department of Radiology\n"
       "VA San Diego Healthcare System and\n"
       "University of California, San Diego\n"
       "email: lfrank@ucsd.edu\n"  
       "\n"
       "Supported in part by VA Merit Award (MPP)\n"
       "\n"
       "Many thanks to Eman Ghobrial for assistance and support\n"
       "in solving a multitude of computer-related problems.\n"

    , TEXT_READONLY ) ;
   return ;
}

/*-------------------------------------------------------------------
  Callback for copyright button
---------------------------------------------------------------------*/

void DIFF_copyright_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   (void ) new_MCW_textwin( info_lab ,

       "++++++++++++++++++++++  Copyright Notice ++++++++++++++++++++++++\n"
       "\n"
       "Sorry, folks!  They made me put this in.\n"
       "The summary is:  It's free to use for research like\n"
       "any other AFNI Plugin, but you can't use it for profit.\n"
       "(If you find out how to do that, let me know!)\n"
       "\n"
       "------------------------------------------------------------------\n"
       "\n"
       "plug_diff.c\n"
       "Copyright  2000,  The Regents of the University of California.\n"
       "All Rights Reserved.\n"
       "\n"
       "Permission to use, copy, modify, and distribute any part of this\n"
       "diffusion analysis plugin for educational, research, and non-profit\n"
       "purposes, without fee, and without a written agreement is hereby\n"
       "granted, provided that the above copyright notice, this paragraph\n"
       "and the following three paragraphs appear in all copies.\n"
       "\n"
       "Those desiring to incorporate this diffusion analysis plugin into\n"
       "commercial products or use for commercial purposes should contact\n"
       "Technology Transfer & Intellectual Property Services, University of\n"
       "California, San Diego, 9500 Gilman Drive, MC - 0910, La Jolla, CA\n"
       "92093-0910, Phone: (858) 534-5815, FAX: (858) 534-7345.\n"
       "\n"
       "In no event shall the University of California be liable to any party\n"
       "for direct, indirect, special, incidental, or consequential damages,\n"
       "including, without limitation, lost profits, claims or demands, or\n"
       "business interruption, arising out of the use of this diffusion\n"
       "analysis plugin, even if the university of california has been advised\n"
       "of the possibility of such damage.\n"
       "\n"
       "The diffusion analysis plugin provided herein is on an 'as is' basis,\n"
       "and the University of California has no obligation to provide\n"
       "maintenance, support, updates, enhancements, or modifications.  The\n"
       "University of California makes no representations and extends no\n"
       "warranties of any kind, either implied or express, including, but not\n"
       "limited to, the implied warranties of merchantability or fitness for a\n"
       "particular purpose, or that the use of the diffusion analysis plugin\n"
       "will not infringe any patent, trademark or other rights.\n"
       "\n"

    , TEXT_READONLY ) ;
   return ;
}

/*-------------------------------------------------------------------
  Callback for (underlay) choose button.
  Criteria for datasets that can be diffused:
    - must be in current session
    - must have actual bricks of bytes or shorts
    - bricks must have cubical grid
    - bricks must be in RAI orientation
---------------------------------------------------------------------*/

/*-- 26 Apr 1999: relax requirement that dataset be axial --*/

#ifdef ONLY_AXIAL
# define IS_AXIAL(ds) ( ( (ds)->daxes->xxorient == ORI_R2L_TYPE ) && \
                        ( (ds)->daxes->yyorient == ORI_A2P_TYPE ) && \
                        ( (ds)->daxes->zzorient == ORI_I2S_TYPE )     )
#else
# define IS_AXIAL(ds) (1)
#endif

#define USEFUL_DIFF_DSET(ds)                                  \
    (( ISVALID_DSET(ds)                      )             && \
     ( DSET_INMEMORY(ds)                     )             && \
     ( DSET_BRICK_TYPE(ds,0) == MRI_short ||                  \
       DSET_BRICK_TYPE(ds,0) == MRI_byte  ||                  \
      (DSET_BRICK_TYPE(ds,0) == MRI_complex && complex_ok) || \
      (DSET_BRICK_TYPE(ds,0) == MRI_float && float_ok)))

static int                  ndsl = 0 ;
static PLUGIN_dataset_link * dsl = NULL ;

void DIFF_load_dsl( THD_3dim_dataset * mset , int float_ok, int complex_ok )
{
   THD_session * ss  = im3d->ss_now ;           /* current session */
   int           vv  = im3d->vinfo->view_type ; /* view type */
   THD_3dim_dataset * qset ;
   int id , nx,ny,nz ;
   int debug = 0;

   ndsl = 0 ; /* initialize */

   if( ISVALID_DSET(mset) ){
      nx = DSET_NX(mset) ; ny = DSET_NY(mset) ; nz = DSET_NZ(mset) ;
   } else {
      nx = ny = nz = 0 ;
   }

   /* scan anats and scan.  
      As of Version 2.55f.  There is no longer and 
      distinction between funcs and anat datasets*/

   for( id=0 ; id < ss->num_dsset ; id++ ){

      qset = ss->dsset[id][vv] ;

      if( !USEFUL_DIFF_DSET(qset) ) {
	continue ;   /* skip this one */
      }

      if( nx > 0 && DSET_NX(qset) != nx ) continue ;  /* must match */
      if( ny > 0 && DSET_NY(qset) != ny ) continue ;  /* brick size */
      if( nz > 0 && DSET_NZ(qset) != nz ) continue ;

      ndsl++ ;
      dsl = (PLUGIN_dataset_link *)
              XtRealloc( (char *) dsl , sizeof(PLUGIN_dataset_link)*ndsl ) ;

      make_PLUGIN_dataset_link( qset , dsl + (ndsl-1) ) ;  /* cf. afni_plugin.c */
   }

   /* scan funcs */
   if(debug) printf("\n\n DIFF_load_dsl:ss->num_dsset = %i\n",ss->num_dsset);

#ifdef OUTDATEDCODE
   for( id=0 ; id < ss->num_func ; id++ ){
      qset = ss->func[id][vv] ;

      if( !USEFUL_DIFF_DSET(qset) ) continue ;   /* skip this one */

      if( nx > 0 && DSET_NX(qset) != nx ) continue ;
      if( ny > 0 && DSET_NY(qset) != ny ) continue ;
      if( nz > 0 && DSET_NZ(qset) != nz ) continue ;

      ndsl++ ;
      dsl = (PLUGIN_dataset_link *)
              XtRealloc( (char *) dsl , sizeof(PLUGIN_dataset_link)*ndsl ) ;

      make_PLUGIN_dataset_link( qset , dsl + (ndsl-1) ) ;
   }
#endif

   return ;
}

int SET_SCALARS_FLAG( char *chan ) 
{
  int debug = 0;

  if(MATCH_STR(chan,"DTI")) {
    dti_scalars_flag = 1 ;	/* used to check for loaded dti dataset 
				   in DIFF_dti_compute_scalars_CB */
    if(debug) printf("SET_SCALARS_FLAG: setting dti_scalars_flag = %i\n",
		     dti_scalars_flag);
    return 1 ;
  }
  if(MATCH_STR(chan,"SHD")) {
    shd_scalars_flag = 1 ;	/* used to check for loaded shd dataset 
				   in DIFF_shd_compute_scalars_CB */
    if(debug) printf("SET_SCALARS_FLAG: setting shd_scalars_flag = %i\n",
		     shd_scalars_flag);
    return 1 ;
  }
  if(MATCH_STR(chan,"FTM")) {
    ftm_scalars_flag = 1 ;	/* used to check for loaded ftm dataset 
				   in DIFF_shd_compute_scalars_CB */
    if(debug) printf("SET_SCALARS_FLAG: setting ftm_scalars_flag = %i\n",
		     ftm_scalars_flag);
    return 1 ;
  }
    return 0 ;
}

char *MAKE_GOOD_PREFIX( THD_3dim_dataset *qset, char *suff )
{
  int  nn = 0;
  char *output_prefix;
  int  debug = 0;

  output_prefix = (char *) malloc( sizeof(char) * THD_MAX_PREFIX ) ;
  if(output_prefix == NULL) return NULL ;

  sprintf(output_prefix,"%s.%s",DSET_PREFIX(qset),suff);

  while(!PLUTO_prefix_ok(output_prefix)) {
    nn++ ;
    sprintf(output_prefix,"%s.%s%d",DSET_PREFIX(qset),suff,nn);
    if(debug) printf("output_prefix = %s\n",output_prefix);
  }

  return output_prefix ;

}

int CHECK_DSET_NAME( THD_3dim_dataset *qset , char *froot ) 
{ 
  char *program_name = "CHECK_DSET_NAME";
  char *stst ; 
  char qnam[THD_MAX_NAME] ;
  int  debug = 0;

  stst = strchr(DSET_PREFIX((qset)),'.');	
  sprintf(qnam,".%s",froot);

  if(debug) {
    printf("%s:\n",program_name);
    printf("DSET_PREFIX = %s\n",DSET_PREFIX((qset))); 
    printf("qnam = %s\n",qnam); 
    printf("stst = %s\n",stst); 
  }

  if(stst==NULL) {
    if(debug) printf("stst is NULL!\n");
    return(0);
  } else { 
    if(!strncmp(stst,qnam,4)) {
      if(debug) printf("Name matches desired type\n"); 
      return(1);
    } else {
      if(debug) printf("Name does not match desired type!\n"); 
      return(0);
    } 
  } 
}

#define MAX_NDSL 1024

void DIFF_choose_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   char *program_name = "DIFF_choose_CB";
   DIF_state *dif = current_dif_state ;
   char *curwin = (char *) client_data ; /* tells from which window chooser requested */

   int  vv = im3d->vinfo->view_type ; /* view type */
   THD_3dim_dataset *qset ;
   MCW_idcode midc ;
   int  id , ltop , llen , dofunc = 0;
   int  doanat_ftm=0;		/* 31-Aug-31 */
   static char ** strlist = NULL ;
   char buf[256] ;
   char qnam[THD_MAX_NAME] , label[THD_MAX_NAME] , qroot[THD_MAX_NAME] ;
   int  isl = -2 ;     /* 03 Apr 1999 */
   int  num_good_sets ;
   int  *dsindx = NULL;
   int  first_index = -1 ;	/* index of first occurance of correct file suffix */

   char *thdinfo;
   int fibmeth;
   int SHOW_DSET_INFO = 0;
   int debug = 0;
   
   /* Filename suffix for file check */

   if( (dif->sta->overlay_chan < 0) || !strcmp(curwin,"MAIN") ) { 
     strcpy(qroot,EMPTY_LABEL);
   } else {
     strcpy(qroot,DIFO_labels[dif->sta->overlay_chan]);
   }

   /* Copy into DIFF_choose_CB string 01-Sep-02 */

   dofunc = (w == wshdfunc_choose_pb || w == ftm_choose_pb 
	     || w == minimal_shd_display_choose_pb) ;     /* 22-July-03 */

   if (w == ftm_anat_choose_pb ) {     /* 31-Aug-04 */
     doanat_ftm = 1 ;
   }

   /* A hack(?) to make sure 
      DTI dataset loaded for FTM1
      SHD dataset loaded for FTM2
      22-July-03 */

   if(w == ftm_choose_pb) {

     if( !ftm1_flag && !ftm2_flag && !ftm_dst_flag ) {
       (void) MCW_popup_message( w ,
				 "\n** No fiber tracking method chosen! **\n",
				 MCW_USER_KILL | MCW_TIMER_KILL ) ;
       PLUTO_beep() ; return ;
     }

#ifdef EXPERIMENTAL
     fibmeth = MCW_val_bbox( fiber_compute_bbox ); 
#else
     fibmeth = 1;
#endif
     if(fibmeth==1) {
       strcpy(qroot,DIFO_labels[0]);
       strcpy(diff_choose_str,DIFO_labels[0]);
     }
     if(fibmeth==2) strcpy(qroot,DIFO_labels[1]);
     SET_SCALARS_FLAG("FTM") ;
   }

   /* if(w == wshdfunc_choose_pb) {
      dofunc = 1;
      strcpy(qroot,DIFO_labels[dif->sta->overlay_chan]);
      strcpy(diff_choose_str,qroot) ;
      } else if (w == choose_pb) {
      strcpy(qroot,EMPTY_LABEL);
      strcpy(diff_choose_str,"MAIN") ;
      } */

   if(debug) {
     printf("\n\nDIFF_choose_CB: overlay channel = %i\n"
	    "DIFF_choose_CB: qroot = %s\n"
	    "DIFF_choose_CB: diff_choose_str = %s\n"
	    "DIFF_choose_CB: curwin = %s\n",
	    dif->sta->overlay_chan,qroot,diff_choose_str,curwin);
   }

   /*-- decide if we want overlay (func) or underlay --*/

   if(debug) printf("%s: dofunc = %i, doanat_ftm = %i\n",program_name,dofunc,doanat_ftm);
   
   if( dofunc && !ISVALID_DSET(dset) ){
     (void) MCW_popup_message( w ,
			       "Can't choose overlay\nbefore underlay!" ,
			       MCW_USER_KILL | MCW_TIMER_KILL ) ;
     XBell(dc->display,100) ; return ;
   }

   if(debug) {printf("%s: Calling DIFF_load_dsl\n",program_name); PRINT_DSL(dsl);}
   if( dofunc )
     DIFF_load_dsl( dset , 1 , 1) ;
   else
     DIFF_load_dsl( NULL , 1 , 0) ;

   if(debug) {printf("%s: After loading DIFF_load_dsl\n",program_name); PRINT_DSL(dsl);}

   /* found nothing?  exit */

   if( ndsl < 1 ){
     (void) MCW_popup_message( w ,
			       "Didn't find\nany datasets\nto diffuse!" ,
			       MCW_USER_KILL | MCW_TIMER_KILL ) ;
     XBell(dc->display,100) ; return ;
   }

   /*--- loop over dataset links and patch their titles
         to include an indicator of the dataset type    ---*/

   ltop = 4 ;
   for( id=0 ; id < ndsl ; id++ ){
      llen = strlen(dsl[id].title) ;
      ltop = MAX(ltop,llen) ;
   }

   /* Check filenames to insure correct type - lrf */
   
   FREEUP(dsindx) ;
   if(debug) printf("%s: allocating dsindx of size = %i\n",program_name,ndsl);
   dsindx = (int *) calloc(ndsl,sizeof(int));

   if(debug) printf("%s: qroot check\n",program_name);

   if(strcmp(qroot,EMPTY_LABEL)) {

     if(debug) printf("%s: inside qroot!=EMPTY\n",program_name);

     /* Copy into DIFF_choose_CB string 01-Sep-02 */
     /*if(dofunc) strcpy(diff_choose_str,qroot) ;*/

     num_good_sets = 0;
     if(debug) printf("%s: Setting num_good_sets = %i\n",program_name,num_good_sets);

     for( id=0 ; id < ndsl ; id++ ){
       
       qset = PLUTO_find_dset( &(dsl[id].idcode) ) ;
       if( !CHECK_DSET_NAME( qset , qroot ) ) continue ;
       dsindx[ num_good_sets++ ] = id ;
       if(first_index==-1) first_index = id ; /* 1st occurance of correct suffix */
	
     }

     if(!num_good_sets) {

       sprintf(buf," \n"
	       "** No %s datasets found!\n", qroot ) ;
       (void) MCW_popup_message( choose_pb , buf ,
				 MCW_USER_KILL | MCW_TIMER_KILL ) ;
       PLUTO_beep() ; return ;

     } else {

       SET_SCALARS_FLAG( qroot ) ;

     }

   } else {			/* default settings - do we want this? */

     if(debug) printf("%s: inside qroot=EMPTY\n",program_name);

     num_good_sets = ndsl ;
     for( id=0 ; id < num_good_sets ; id++ ) dsindx[id] = id ;
     first_index = 0;

   }

   if(debug) printf("\n\t\t ***** DIFF_choose_CB: Starting id loop: ***** \n"
		    "\t\t num_good_sets = %i, first_index = %i\n",
		    num_good_sets,first_index);

   for( id=0 ; id < num_good_sets ; id++ ){
     qset = PLUTO_find_dset( &(dsl[dsindx[id]].idcode) ) ;

     if(SHOW_DSET_INFO) {
	thdinfo = THD_dataset_info(qset,-1);
     	printf("%s\n",thdinfo);
     }

     if(debug) printf("\t\t dsindx[%i] = %i, prefix = %s\n",
		      id,dsindx[id],DSET_PREFIX(qset));
      if( !ISVALID_DSET(qset) ) continue ;
      if( ISANAT(qset) ){
         if( ISANATBUCKET(qset) )         /* 30 Nov 1997 */
            sprintf(qnam,"%-*s [%s:%d]" ,
                    ltop,dsl[dsindx[id]].title ,
                    ANAT_prefixstr[qset->func_type] , DSET_NVALS(qset) ) ;

         else if( DSET_NUM_TIMES(qset) == 1 )
            sprintf(qnam,"%-*s [%s]" ,
                    ltop,dsl[dsindx[id]].title ,
                    ANAT_prefixstr[qset->func_type] ) ;

         else
            sprintf(qnam,"%-*s [%s:3D+t:%d]" ,
                    ltop,dsl[dsindx[id]].title ,
                    ANAT_prefixstr[qset->func_type] , DSET_NUM_TIMES(qset) ) ;

      } else {
         if( ISFUNCBUCKET(qset) )         /* 30 Nov 1997 */
            sprintf(qnam,"%-*s [%s:%d]" ,
                    ltop,dsl[dsindx[id]].title ,
                    FUNC_prefixstr[qset->func_type] , DSET_NVALS(qset) ) ;

         else if( DSET_NUM_TIMES(qset) == 1 )
            sprintf(qnam,"%-*s [%s]" ,
                    ltop,dsl[dsindx[id]].title ,
                    FUNC_prefixstr[qset->func_type] ) ;

         else
            sprintf(qnam,"%-*s [%s:3D+t:%d]" ,
                    ltop,dsl[dsindx[id]].title ,
                    FUNC_prefixstr[qset->func_type] , DSET_NVALS(qset) ) ;
      }

      if( DSET_COMPRESSED(qset) ) strcat(qnam,"z") ;

      strcpy( dsl[dsindx[id]].title , qnam ) ;
   }

   if(debug) printf("\n\t\t ***** DIFF_choose_CB: Finished with id loop ***** \n");
   if(debug) PRINT_DSL(dsl) ;

   /*--- make a popup chooser for the user to browse ---*/

   POPDOWN_strlist_chooser ;

   strlist = (char **) XtRealloc( (char *)strlist , sizeof(char *)*num_good_sets ) ;
   for( id=0 ; id < num_good_sets ; id++ ) strlist[id] = dsl[dsindx[id]].title ;

   /*-- 03 Apr 1999: set the initial selection in the chooser --*/

        if(  dofunc && func_dset != NULL ){ midc = func_dset_idc; isl = -1; }
   else if(  doanat_ftm && ftm_anat_dset != NULL ){ midc = ftm_anat_dset_idc ; isl = -1; }
   else if( !dofunc && dset      != NULL ){ midc = dset_idc     ; isl = -1; }

   if( isl == -1 ){
      for( id=0 ; id < num_good_sets ; id++ ){
         if( EQUIV_IDCODES(midc,dsl[dsindx[id]].idcode) ){ 
	   isl = dsindx[id] ; break ; 
	 }
      }
   }

   if(debug) {
     printf("midc.str = %s, isl = %i\n",midc.str,isl);
     for( id=0 ; id < num_good_sets ; id++ ) printf("strlist[%i] = %s\n",id,strlist[id]);
   } 

   /*
   if( dofunc ) {
     if(debug) printf("Loading dsl = %i\n",first_index);
     DIFF_load_dsl( dset , first_index ) ;
   } else {
     DIFF_load_dsl( NULL , 0 ) ;
     } */

   /*-- popup the chooser -- */

   sprintf( label , "AFNI Dataset from\nthe %s" , VIEW_typestr[vv] ) ;

   if(debug) 
     printf("DIFF_choose_CB: Calling MCW_choose_strlist: "
	    "num_good_sets = %i, first_index = %i, isl = %i\n",
	    num_good_sets,first_index,isl);

   if(debug) 
     for( id=0 ; id < num_good_sets ; id++ )
       printf("DIFF_choose_CB: dsindx[%i] = %i\n",id,dsindx[id]); 

   MCW_choose_strlist( w , label , num_good_sets , isl , strlist , 
                       (doanat_ftm) ? 
		       reinterpret_cast<void (*)()>
		       (DIFF_finalize_ftm_anat_CB)
		       : ((dofunc) ? 
			  reinterpret_cast<void (*)()>
			  (DIFF_finalize_func_CB)
			  : reinterpret_cast<void (*)()>
			  (DIFF_finalize_dset_CB) ),
		       (XtPointer) dsindx ) ;

   if(debug) 
     for( id=0 ; id < num_good_sets ; id++ )
       printf("DIFF_choose_CB: dsindx[%i] = %i\n",id,dsindx[id]); 

   return ;
}

/**<DIFF_finalize_func_CB>****************************************************

   Title: DIFF_finalize_func_CB 

   Usage: Called by  MCW_choose_strlist in DIFF_choose_CB 

   Call: DIFF_finalize_func_CB( Widget w, XtPointer fd, MCW_choose_cbs *cbs )

   History: 18-10-01 Now pass dsindx and use dsindx[id] instead of id
                     as the index of data sets.  This is to accomodate
		     partial lists used in the wfunc window chooser

******************************************************************************/
void DIFF_finalize_func_CB( Widget w, XtPointer fd, MCW_choose_cbs *cbs )
{
   int   *dsindx = (int *) fd ;
   int   id = dsindx[cbs->ival] ;

   THD_3dim_dataset *qset, *oset ;
   XmString         xstr ;
   char             str[2*THD_MAX_NAME] ;
   float            fac ;

   int debug = 0;

   if(debug) {
     printf("\nDIFF_finalize_func_CB: cbs->ival = %i\n",cbs->ival);
     printf("DIFF_finalize_func_CB: dsindx[0] = %i\n",dsindx[0]);
     printf("DIFF_finalize_func_CB: id = %i\n",id);
   }

   /* check for errors */

   if( ! diffuser_open ){ 
     POPDOWN_strlist_chooser ; XBell(dc->display,100) ; return ; 
   }

   if( id < 0 || id >= ndsl ){ XBell(dc->display,100) ; return ; }

   qset = PLUTO_find_dset( &(dsl[id].idcode) ) ;  /* the new dataset? */

   if( qset == NULL ){ XBell(dc->display,100) ; return ; }

   /* accept this dataset */

   FREE_VOLUMES ; INVALIDATE_OVERLAY ;

   oset          = func_dset ;
   func_dset     = qset ;
   func_dset_idc = qset->idcode ;  /* 31 Mar 1999 */

   if( DSET_NVALS(func_dset) == 20 )
     dt2_dset = qset;
   else
     shd_dset = qset;

   /* refit the sub-brick selector menus */

   if( oset == NULL ){ func_color_ival = 0 ; func_thresh_ival = 1 ; }

   if( func_color_ival >= DSET_NVALS(func_dset) )
     func_color_ival = DSET_NVALS(func_dset)-1;

   refit_MCW_optmenu( wshdfunc_color_av ,
                      0 ,                       /* new minval */
                      DSET_NVALS(func_dset)-1 , /* new maxval */
                      func_color_ival ,         /* new inival */
                      0 ,                       /* new decim? */
		      reinterpret_cast<char* (*)()>
                      (DIFF_choose_av_label_CB),/* text routine */
                      func_dset                 /* text data */
                    ) ;

   AV_SENSITIZE( wshdfunc_color_av , (DSET_NVALS(func_dset) > 1) ) ;

   if( func_thresh_ival >= DSET_NVALS(func_dset) )
     func_thresh_ival = DSET_NVALS(func_dset)-1;

   refit_MCW_optmenu( wshdfunc_thresh_av ,
                      0 ,                       /* new minval */
                      DSET_NVALS(func_dset)-1 , /* new maxval */
                      func_thresh_ival ,        /* new inival */
                      0 ,                       /* new decim? */
		      reinterpret_cast<char* (*)()>
                      (DIFF_choose_av_label_CB),/* text routine */
                      func_dset                 /* text data */
                    ) ;

   AV_SENSITIZE( wshdfunc_thresh_av , (DSET_NVALS(func_dset) > 1) ) ;

   /* write the informational label */

   strcpy( func_dset_title , dsl[id].title ) ;
   printf("\nDIFF_finalize_func_CB: ++++++++++++++++ func_dset_title = %s\n",func_dset_title);
   fac = DSET_BRICK_FACTOR(func_dset,func_color_ival) ;

   if( fac == 0.0 || fac == 1.0 ){
      strcpy(str,func_dset_title) ;
   } else {
      char abuf[16] ;
      AV_fval_to_char( fac , abuf ) ;
      sprintf(str,"%s [* %s]", func_dset_title , abuf ) ;
   }
   xstr = XmStringCreateLtoR( str , XmFONTLIST_DEFAULT_TAG ) ;
   XtVaSetValues( wshdfunc_info_lab , XmNlabelString , xstr , NULL ) ;
   XmStringFree(xstr) ;

   /* fix the range labels */

   xstr = DIFF_range_label() ;
   XtVaSetValues( wshdfunc_range_label , XmNlabelString , xstr , NULL ) ;
   XmStringFree(xstr) ;

   xstr = DIFF_autorange_label() ;
   XtVaSetValues( wshdfunc_range_bbox->wbut[0], XmNlabelString,xstr , NULL ) ;
   XmStringFree(xstr) ;

   /* fix the p-value label */

   DIFF_set_thr_pval() ;

   /* read the new data */

   DIFF_reload_func_dset(im3d,id) ; /* 01-Sep-02 */

   new_func_dset = 1 ;           /* flag it as new */

   if(debug) printf("DIFF_finalize_func_CB: diff_choose_str = %s\n",
		    diff_choose_str);

   return ;
}

void DIFF_finalize_dset_CB( Widget w, XtPointer fd, MCW_choose_cbs * cbs )
{
   int              ii,id = cbs->ival ;
   THD_3dim_dataset *qset ;
   XmString         xstr ;
   char             str[2*THD_MAX_NAME] ;
   float            fac ;

   /* check for errors */

   if( ! diffuser_open ){ POPDOWN_strlist_chooser ; XBell(dc->display,100) ; return ; }

   if( id < 0 || id >= ndsl ){ XBell(dc->display,100) ; return ; }

   qset = PLUTO_find_dset( &(dsl[id].idcode) ) ;  /* the new dataset? */

   if( qset == NULL ){ XBell(dc->display,100) ; return ; }

   /* if there was an existing diffuser, kill it off */

   if( diffuse_handle != NULL ){
     /* destroy_MDIF_diffuser(diffuse_handle) ; */
      diffuse_handle = NULL ; func_cmap_set = 0 ;
   }

   FREE_VOLUMES ; INVALIDATE_OVERLAY ;

   /* accept this dataset */

   dset = qset ;
   dset_idc = qset->idcode ;  /* 31 Mar 1999 */

   npixels = 256 ;                             /* size of image to diffuse */
   npixels = MAX( npixels , DSET_NX(dset) ) ;
   npixels = MAX( npixels , DSET_NY(dset) ) ;
   npixels = MAX( npixels , DSET_NZ(dset) ) ;

   /* refit the sub-brick selector menu */

   if( dset_ival >= DSET_NVALS(dset) ) dset_ival = DSET_NVALS(dset)-1 ;

   refit_MCW_optmenu( choose_av ,
                      0 ,                       /* new minval */
                      DSET_NVALS(dset)-1 ,      /* new maxval */
                      dset_ival ,               /* new inival */
                      0 ,                       /* new decim? */
		      reinterpret_cast<char* (*)()>
                      (DIFF_choose_av_label_CB),/* text routine */
                      dset                      /* text data */
                    ) ;

   AV_SENSITIZE( choose_av , (DSET_NVALS(dset) > 1) ) ;

   /* write the informational label */

   strcpy( dset_title , dsl[id].title ) ;
   fac = DSET_BRICK_FACTOR(dset,dset_ival) ;

   if( fac == 0.0 || fac == 1.0 ){
      strcpy(str,dset_title) ;
   } else {
      char abuf[16] ;
      AV_fval_to_char( fac , abuf ) ;
      sprintf(str,"%s [* %s]", dset_title , abuf ) ;
   }
   xstr = XmStringCreateLtoR( str , XmFONTLIST_DEFAULT_TAG ) ;
   XtVaSetValues( info_lab , XmNlabelString , xstr , NULL ) ;
   XmStringFree(xstr) ;

   /* if the existing overlay dataset doesn't match this one, kill the overlay */

   if( func_dset != NULL && ( DSET_NX(dset) != DSET_NX(func_dset) ||
                              DSET_NY(dset) != DSET_NY(func_dset) ||
                              DSET_NZ(dset) != DSET_NZ(func_dset)   ) ){

      INVALIDATE_OVERLAY ;
      func_dset = NULL ;

      TURNOFF_SHD_OVERLAY_WIDGETS ;
      TURNOFF_DTI_OVERLAY_WIDGETS ;
      TURNOFF_FTM_OVERLAY_WIDGETS ;

      (void) MCW_popup_message( choose_pb ,
                                   " \n"
                                   "** New underlay dataset did  **\n"
                                   "** not match dimensions of   **\n"
                                   "** existing overlay dataset, **\n"
                                   "** so the latter was removed **\n" ,
                                MCW_USER_KILL | MCW_TIMER_KILL ) ;
   }

   /* read the new data */

   new_dset = 1 ;           /* flag it as new */
   /* load the data */
   if (DIFF_reload_dataset()) {
     /* Create mask off b=0 image on input.  1 Sept 2002 
	b0_dstats is a global (ech!) that hold stats of
	the b0 images. */

     /* Average brain intensity */

     b0_dstats = DIFF_dset_stats( dset , NULL , 0, 0, 0, 0.0 , 0.0 ) ;
     PRINT_DSTATS(b0_dstats) ;

     /* Create mask of brain intensities > thresh*mean(brain) 
	for each slice*/

     b0_dstats->mask = THD_makemask(dset, 0, b0_dstats->thresh, b0_dstats->max ) ; 
   }
   return ;
}

/* Load ftm anatomical dataset. Sept 4, 2004 */

void Load_FTM_Anat_DataSet()
{
    char *program_name = "Load_FTM_Anat_DataSet";
    float loc[3];
    int   sdims[3] ;
    int   debug = 0;

      if( ftm_anat_dset != NULL ) {

	DSET_load( ftm_anat_dset ); 
	if(!DSET_LOADED(ftm_anat_dset) ){
	  fprintf(stderr,"*** Error: can't load data from %s!\n"
		  "Returning from %s\n",
		  DSET_FILECODE(ftm_anat_dset),program_name) ;
	  return ;
	}

	/* Retrieve dimensions */

	int nvox = DSET_NVOX(ftm_anat_dset);
	int nbrk = DSET_NVALS(ftm_anat_dset);
	int nx   = DSET_NX(ftm_anat_dset);
	int ny   = DSET_NY(ftm_anat_dset);
	int nz   = DSET_NZ(ftm_anat_dset);
	int nxy  = nx * ny;		

	float dx = DSET_DX(ftm_anat_dset),
	      dy = DSET_DY(ftm_anat_dset),
	      dz = DSET_DZ(ftm_anat_dset);

	LOAD_VEC3(sdims,nx,ny,nz);

	fprintf(stderr,"%s: Loading anat dataset '%s': \n"
		"\t orientation =  (%s,%s,%s)\n"
		"\t origin (x,y,z) =  (%f,%f,%f)\n"
		"\t dimensions (nx,ny,nz) =  (%i,%i,%i)\n"
		"\t increments (dx,dy,dz) =  (%f,%f,%f)\n",
		program_name,DSET_PREFIX(ftm_anat_dset),
		ORIENT_typestr[ftm_anat_dset->daxes->xxorient],
		ORIENT_typestr[ftm_anat_dset->daxes->yyorient],
		ORIENT_typestr[ftm_anat_dset->daxes->zzorient],
		DSET_XORG(ftm_anat_dset), DSET_YORG(ftm_anat_dset),
		DSET_ZORG(ftm_anat_dset),nx,ny,nz,dx,dy,dz);

	/********************* vtk *********************/

	/* The image polydata */
	vtkStructuredPoints *Image = vtkStructuredPoints::New();
        Image->SetScalarType( VTK_FLOAT );

	/*
	 * We use the Dicom axes as our canonical coordinate system,
	 * which are x = R->L, y = A->P, z = I->S
	 */

	if (ftm_anat_dset->daxes->xxorient == ORI_R2L_TYPE ||
	    ftm_anat_dset->daxes->xxorient == ORI_L2R_TYPE)
	  {
	    if (ftm_anat_dset->daxes->yyorient == ORI_P2A_TYPE ||
		ftm_anat_dset->daxes->yyorient == ORI_A2P_TYPE)
	      {
		Image->SetDimensions(nx, ny, nz);
		Image->SetSpacing(fabs(dx), fabs(dy), fabs(dz));
	      }
	    else  //yyorient is I2S or S2I
	      {
		Image->SetDimensions(nx, nz, ny);
		Image->SetSpacing(fabs(dx), fabs(dz), fabs(dy));
	      }
	  }
	else if (ftm_anat_dset->daxes->xxorient == ORI_P2A_TYPE ||
		 ftm_anat_dset->daxes->xxorient == ORI_A2P_TYPE)
	  {
	    if (ftm_anat_dset->daxes->yyorient == ORI_R2L_TYPE ||
		ftm_anat_dset->daxes->yyorient == ORI_L2R_TYPE)
	      {
		Image->SetDimensions(ny, nx, nz);
		Image->SetSpacing(fabs(dy), fabs(dx), fabs(dz));
	      }
	    else  //yyorient is I2S or S2I
	      {
		Image->SetDimensions(nz, nx, ny);
		Image->SetSpacing(fabs(dz), fabs(dx), fabs(dy));
	      }
	  }
	else  //xxorient is I2S or S2I
	  {
	    if (ftm_anat_dset->daxes->yyorient == ORI_P2A_TYPE ||
		ftm_anat_dset->daxes->yyorient == ORI_A2P_TYPE)
	      {
		Image->SetDimensions(nz, ny, nx);
		Image->SetSpacing(fabs(dz), fabs(dy), fabs(dx));
	      }
	    else  //yyorient is R2L or L2R
	      {
		Image->SetDimensions(ny, nz, nx);
		Image->SetSpacing(fabs(dy), fabs(dz), fabs(dx));
	      }
	  }


	/*
	 * To compute the origin in terms of Dicom coordinates, we use
	 * the axes' to_dicomm transformation
	 */
	THD_fvec3 origin_dset;
	LOAD_FVEC3(origin_dset, 
		   DSET_XORG(ftm_anat_dset), 
		   DSET_YORG(ftm_anat_dset), 
		   DSET_ZORG(ftm_anat_dset));
	if (dx < 0.0)
	  {
	    origin_dset.xyz[0] = origin_dset.xyz[0] + dx * (nx - 1);
	  }
	if (dy < 0.0)
	  {
	    origin_dset.xyz[1] = origin_dset.xyz[1] + dy * (ny - 1);
	  }
	if (dz < 0.0)
	  {
	    origin_dset.xyz[2] = origin_dset.xyz[2] + dz * (nz - 1);
	  }

	THD_fvec3 origin_dicom = THD_3dmm_to_dicomm(ftm_anat_dset, 
						    origin_dset);
	Image->SetOrigin((vtkFloatingPointType) origin_dicom.xyz[0],
			 (vtkFloatingPointType) origin_dicom.xyz[1],
			 (vtkFloatingPointType) origin_dicom.xyz[2]);

	/* The image intensities */

	vtkFloatArray *NewIntensity = vtkFloatArray::New();
	NewIntensity->SetNumberOfTuples(nvox);
	
	//if(!ISANAT(ftm_anat_dset)) {
	//  fprintf(stderr,"%s: ftm_anat_dset not an ANAT dset!\n",program_name); 
	//  exit(0);
	//}

	int typeDset = DSET_BRICK_TYPE( ftm_anat_dset,0 );

	switch(typeDset) {

	  case MRI_short:{

	    /* voxel intensity */
	    short *sval = (short *) DSET_ARRAY(ftm_anat_dset,0) ;

	    for (int i=0; i < nx; ++i)
	      {
		for (int j=0; j < ny; ++j)
		  {
		    for (int k=0; k < nz; ++k)
		      {

			int afniIndex = DSET_ixyz_to_index(ftm_anat_dset,
							   i,j,k);

			/* Load up the polydata */
			float fval = (float) sval[afniIndex];

			/* 
			 * As with origin, use the to_dicom transform
			 * to transform positon into dicom coordinates,
			 * to eliminate effect of axis orientation
			 */
			THD_fvec3 position_dset;
			LOAD_FVEC3(position_dset,
				   i * dx + DSET_XORG(ftm_anat_dset),
				   j * dy + DSET_YORG(ftm_anat_dset),
				   k * dz + DSET_ZORG(ftm_anat_dset));

			THD_fvec3 position_dicom =
			  THD_3dmm_to_dicomm(ftm_anat_dset, position_dset);

			int vtkIndex = Image->FindPoint(
				(vtkFloatingPointType) position_dicom.xyz[0],
				(vtkFloatingPointType) position_dicom.xyz[1],
				(vtkFloatingPointType) position_dicom.xyz[2]);
			NewIntensity->SetTuple(vtkIndex,&fval);
		      }
		  }
	      }


	    Image->GetPointData()->SetScalars(NewIntensity);
	  }
	    NewIntensity->Delete() ;

	  break;

	  case MRI_float:{

	    /* voxel intensity */
	    float *sval = (float *) DSET_ARRAY(ftm_anat_dset,0) ;

	    for (int i=0; i < nx; ++i)
	      {
		for (int j=0; j < ny; ++j)
		  {
		    for (int k=0; k < nz; ++k)
		      {

			int afniIndex = DSET_ixyz_to_index(ftm_anat_dset,
							   i,j,k);

			/* Load up the polydata */
			float fval = (float) sval[afniIndex];

			/* 
			 * As with origin, use the to_dicom transform
			 * to transform positon into dicom coordinates,
			 * to eliminate effect of axis orientation
			 */
			THD_fvec3 position_dset;
			LOAD_FVEC3(position_dset,
				   i * dx + DSET_XORG(ftm_anat_dset),
				   j * dy + DSET_YORG(ftm_anat_dset),
				   k * dz + DSET_ZORG(ftm_anat_dset));

			THD_fvec3 position_dicom =
			  THD_3dmm_to_dicomm(ftm_anat_dset, position_dset);

			int vtkIndex = Image->FindPoint(
				(vtkFloatingPointType) position_dicom.xyz[0],
				(vtkFloatingPointType) position_dicom.xyz[1],
				(vtkFloatingPointType) position_dicom.xyz[2]);
			NewIntensity->SetTuple(vtkIndex,&fval);
		      }
		  }
	      }


	    Image->GetPointData()->SetScalars(NewIntensity);
	  }
	    NewIntensity->Delete() ;

	  break;

	  default: 
	    fprintf(stderr,
		    "%s: Intensities should be floats!\n"
		    "(something else found)\n",program_name);
	  break;
	}
	fprintf(stderr , "done\n");
      
	vtkFloatingPointType range[2];
	Image->GetScalarRange(range);
	fprintf(stderr,"%s: Anatomical image range = (%e,%e)\n",
		program_name,range[0],range[1]) ;

	/* Visualize */

	fprintf(stderr,"Visualizing %dx%dx%d dataset ...\n", nx, ny, nz);
	Send_IData_to_vtk_vector_window(Image,ftm_anat_dset);

      } else {
	fprintf(stderr , "Anatomical dataset to view data not loaded\n");
      }
}

/**<DIFF_finalize_ftm_anat_CB>****************************************************

   Title: DIFF_finalize_ftm_anat_CB

   Usage: Called by  MCW_choose_strlist in DIFF_choose_CB 

   Call: DIFF_finalize_ftm_anat_CB( Widget w, XtPointer fd, MCW_choose_cbs *cbs )

   History: 31-08-04 Now pass dsindx and use dsindx[id] instead of id
                     as the index of data sets.  This is to accomodate
		     partial lists used in the wfunc window chooser

******************************************************************************/
void DIFF_finalize_ftm_anat_CB( Widget w, XtPointer fd, MCW_choose_cbs *cbs )
{
   char *program_name = "DIFF_finalize_ftm_anat_CB";
   int   *dsindx = (int *) fd ;
   int   id = dsindx[cbs->ival] ;

   THD_3dim_dataset *qset, *oset ;
   XmString         xstr ;
   char             str[2*THD_MAX_NAME] ;
   float            fac ;

   int debug = 0;

   if(debug) {
     printf("\nDIFF_finalize_func_CB: cbs->ival = %i\n",cbs->ival);
     printf("DIFF_finalize_func_CB: dsindx[0] = %i\n",dsindx[0]);
     printf("DIFF_finalize_func_CB: id = %i\n",id);
   }

   /* check for errors */

   if( ! diffuser_open ){ 
     POPDOWN_strlist_chooser ; XBell(dc->display,100) ; return ; 
   }

   if( id < 0 || id >= ndsl ){ XBell(dc->display,100) ; return ; }

   qset = PLUTO_find_dset( &(dsl[id].idcode) ) ;  /* the new dataset? */

   if( qset == NULL ){ XBell(dc->display,100) ; return ; }

   /* accept this dataset */

   FREE_VOLUMES ; INVALIDATE_OVERLAY ;

   oset          = ftm_anat_dset ;
   ftm_anat_dset     = qset ;
   ftm_anat_dset_idc = qset->idcode ;  /* 31 Mar 1999 */

   /* refit the sub-brick selector menus */

   if( oset == NULL ){ func_color_ival = 0 ; func_thresh_ival = 1 ; }

   if( func_color_ival >= DSET_NVALS(ftm_anat_dset) )
     func_color_ival = DSET_NVALS(ftm_anat_dset)-1;

   refit_MCW_optmenu( wshdfunc_color_av ,
                      0 ,                       /* new minval */
                      DSET_NVALS(ftm_anat_dset)-1 , /* new maxval */
                      func_color_ival ,         /* new inival */
                      0 ,                       /* new decim? */
		      reinterpret_cast<char* (*)()>
                      (DIFF_choose_av_label_CB),/* text routine */
                      ftm_anat_dset             /* text data */
                    ) ;

   AV_SENSITIZE( wshdfunc_color_av , (DSET_NVALS(ftm_anat_dset) > 1) ) ;

   if( func_thresh_ival >= DSET_NVALS(ftm_anat_dset) )
     func_thresh_ival = DSET_NVALS(ftm_anat_dset)-1;

   refit_MCW_optmenu( wshdfunc_thresh_av ,
                      0 ,                       /* new minval */
                      DSET_NVALS(ftm_anat_dset)-1 , /* new maxval */
                      func_thresh_ival ,        /* new inival */
                      0 ,                       /* new decim? */
		      reinterpret_cast<char* (*)()>
                      (DIFF_choose_av_label_CB),/* text routine */
                      ftm_anat_dset             /* text data */
                    ) ;

   AV_SENSITIZE( wshdfunc_thresh_av , (DSET_NVALS(ftm_anat_dset) > 1) ) ;

   /* write the informational label */

   strcpy( ftm_anat_dset_title , dsl[id].title ) ;
   if(debug) printf("\nDIFF_finalize_func_CB: ++++++++++++++++ ftm_anat_dset_title = %s\n",
		    ftm_anat_dset_title);
   fac = DSET_BRICK_FACTOR(ftm_anat_dset,func_color_ival) ;

   if( fac == 0.0 || fac == 1.0 ){
      strcpy(str,ftm_anat_dset_title) ;
   } else {
      char abuf[16] ;
      AV_fval_to_char( fac , abuf ) ;
      sprintf(str,"%s [* %s]", ftm_anat_dset_title , abuf ) ;
   }
   xstr = XmStringCreateLtoR( str , XmFONTLIST_DEFAULT_TAG ) ;
   XtVaSetValues( wshdfunc_info_lab , XmNlabelString , xstr , NULL ) ;
   XmStringFree(xstr) ;

   /* fix the range labels */

   xstr = DIFF_range_label() ;
   XtVaSetValues( wshdfunc_range_label , XmNlabelString , xstr , NULL ) ;
   XmStringFree(xstr) ;

   xstr = DIFF_autorange_label() ;
   XtVaSetValues( wshdfunc_range_bbox->wbut[0], XmNlabelString,xstr , NULL ) ;
   XmStringFree(xstr) ;

   /* fix the p-value label */

   DIFF_set_thr_pval() ;

   /* read the new data */

   DIFF_reload_ftm_anat_dset(im3d,id) ; /* 01-Sep-02 */

   new_ftm_anat_dset = 1 ;           /* flag it as new */

   /* New loader for ftm anat dset - Sept 4, 2004 */
//   Load_FTM_Anat_DataSet();

   if(debug) printf("%s: diff_choose_str = %s\n",program_name,diff_choose_str);

   return ;
}

/*-------------------------------------------------------------------
  lrf's diffusion:

  Callback for theory menu	
---------------------------------------------------------------------*/

void DIFF_theory_CB( MCW_arrowval * av , XtPointer cd )
{
   int theory_ival = av->ival ;

   /* does nothing - really just need title */

   return ;
}

/*-------------------------------------------------------------------
  Callback for precalc menu
---------------------------------------------------------------------*/

void DIFF_precalc_CB( MCW_arrowval * av , XtPointer cd )
{
   precalc_ival = av->ival ;
   return ;
}

/*-----------------------------------------------------------------------
   Callback for xhair toggle button
-------------------------------------------------------------------------*/

#define CHECK_XHAIR_ERROR                                               \
  do{ if( xhair_flag && dset!=NULL &&                                   \
          ! EQUIV_DATAXES(dset->daxes,im3d->wod_daxes) ){               \
        MCW_set_bbox( xhair_bbox , 0 ) ; xhair_flag = 0 ;               \
        (void) MCW_popup_message( xhair_bbox->wrowcol ,                 \
                                     "Can't overlay AFNI crosshairs\n"  \
                                     "because dataset grid and AFNI\n"  \
                                     "viewing grid don't coincide."   , \
                                  MCW_USER_KILL | MCW_TIMER_KILL ) ;    \
        XBell(dc->display,100) ; return ;                               \
     } } while(0)

void DIFF_xhair_CB( Widget w , XtPointer cd , XtPointer call_data )
{
   int old_xh = xhair_flag ;

   xhair_flag = MCW_val_bbox( xhair_bbox ) ;
   if( old_xh == xhair_flag ) return ;

   CHECK_XHAIR_ERROR ;
   FREE_VOLUMES ;

   xhair_ixold = -666 ; xhair_jyold = -666 ; xhair_kzold = -666 ; /* forget */

   /*   if( cd == NULL && dynamic_flag && diffuse_handle != NULL )
      DIFF_draw_CB(NULL,NULL,NULL) ; */

   return ;
}
/* cyrus: callback to update vtk window */
/*-------------------------------------------------------------------------
  29 Mar 1999: called by AFNI when the user changes the crosshair location
  30 Mar 1999: or when the user draws into the controller image window
---------------------------------------------------------------------------*/

void DIFF_xhair_recv( int why , int np , int * ijk , void * junk )
{
   char *program_name = "DIFF_xhair_recv" ;
   int debug = 1;
   switch( why ){

      /*-- change of crosshair location --*/

      case RECEIVE_VIEWPOINT:{

	 /* update VTK window of click on dataset */

	 if( VTK_window_status == VTK_WINDOW_OPEN ) {
	   int ix, jy, kz;
	   int nx, ny, nz;

 	   /* integer coordinates of current point */
    	   ix = im3d->vinfo->i1 ; nx = DSET_NX(dset);
    	   jy = im3d->vinfo->j2 ; ny = DSET_NY(dset); 
    	   kz = im3d->vinfo->k3 ; nz = DSET_NZ(dset);

//    	   if( ix < 0 || ix >= nx ) return ;  /* error, outside of data bounds */
//      	   if( jy < 0 || jy >= ny ) return ;  /* error */
//    	   if( kz < 0 || kz >= nz ) return ;  /* error */

	   float dx = fabs(dset->daxes->xxdel) ,
	         dy = fabs(dset->daxes->yydel) ,
	         dz = fabs(dset->daxes->zzdel)  ;

	   if(debug) {
	     fprintf(stderr,"%s: (dx,dy,dz) = (%f,%f,%f)\n",program_name,dx,dy,dz); 
	   }

	   /* Update vtk window
	      pass needed data, ie. picked location.  any other params */

	   if(debug) {
	      fprintf(stderr ,"updated open vtk window with new location\n" );
	      fprintf(stderr ,"\tinteger coords of new point (%d, %d, %d)\n",
		      ix, jy,kz );
	      fprintf(stderr ,"\tdimensions of MRI_IMAGE (%d, %d, %d)\n",
		      nx, ny, nz );
	      fprintf(stderr ,"\tdata spacing: (%f, %f, %f)\n",
		      dx, dy, dz );
	   }

	   /* This code will later drive the VTK window.  
	      On click, initiate a call to subroutine that visualizes the 
	      DTI coefficients or the SHD coefficients. */
	   int ijkvox; 
	   ijkvox = THREE_TO_IJK(ix, jy, kz, nx, nx*ny); 
           Send_xhair_to_vtk_window( ijkvox, dset, shd_dset, dt2_dset, 
	     current_dif_state, vGrid );

	 if( vtk_vector_window_status == VTK_VECTOR_WINDOW_OPEN ) {
	   send_xhair_to_vtk_vector_window(im3d->vinfo->xi,
					   im3d->vinfo->yj,
					   im3d->vinfo->zk);
	 }

         if( !xhair_flag || !dynamic_flag || diffuse_handle == NULL ) return ;

         CHECK_XHAIR_ERROR ;

         if( CHECK_XHAIR_MOTION ){
            FREE_VOLUMES ;
            /* DIFF_draw_CB(NULL,NULL,NULL) ; */
         }
        }
      }
      return ;

      /*-- user drew something --*/

      case RECEIVE_DRAWNOTICE:{   /* 30 Mar 1999 */

         int doit=0 ;

         if( EQUIV_DSETS(im3d->anat_now,dset) ||    /* can't tell if user */
             EQUIV_DSETS(im3d->fim_now,dset)    ){  /* is drawing on anat */

            doit = 1 ; FREE_VOLUMES ;

         }

         if( EQUIV_DSETS(im3d->anat_now,func_dset) ||    /* or is drawing */
             EQUIV_DSETS(im3d->fim_now,func_dset)    ){  /* on the fim    */

            doit = 1 ; INVALIDATE_OVERLAY ;

            /* 15 Jun 1999: fix the range labels */

            { XmString xstr ;
              xstr = DIFF_range_label() ;
              XtVaSetValues( wshdfunc_range_label , XmNlabelString , xstr , NULL ) ;
              XmStringFree(xstr) ;

              xstr = DIFF_autorange_label() ;
              XtVaSetValues( wshdfunc_range_bbox->wbut[0], XmNlabelString,xstr , NULL ) ;
              XmStringFree(xstr) ;
            }
         }

         /* if( doit && dynamic_flag && diffuse_handle != NULL )
            DIFF_draw_CB(NULL,NULL,NULL) ; */
      }
      return ;

      /*-- dataset pointers have changed --*/

      case RECEIVE_DSETCHANGE:{   /* 31 Mar 1999 */

         if( dset != NULL )
            dset = PLUTO_find_dset( &dset_idc ) ;

         if( func_dset != NULL )
            func_dset = PLUTO_find_dset( &func_dset_idc ) ;

         if( ftm_anat_dset != NULL )
	   ftm_anat_dset = PLUTO_find_dset( &ftm_anat_dset_idc ) ;

         FREE_VOLUMES ; INVALIDATE_OVERLAY ;

         (void) MCW_popup_message( reload_pb ,
                                     "********** NOTICE ***********\n"
                                     "* Session rescan has forced *\n"
                                     "* purge of dataset brick(s) *\n"
                                     "* from memory.              *\n"
                                     "*****************************" ,
                                   MCW_USER_KILL | MCW_TIMER_KILL     ) ;
      }
      return ;

   }  /* end of switch on "why" */

   return ;
}
#if 0
void DIFF_xhair_recv( int why , int np , int * ijk , void * junk )
{
   char *program_name = "DIFF_xhair_recv" ;
   int debug = 1;
   switch( why ){

      /*-- change of crosshair location --*/

      case RECEIVE_VIEWPOINT:{

	 /* update VTK window of click on dataset */

	 if( VTK_window_status == VTK_WINDOW_OPEN ) {
	   int ix, jy, kz;
	   int nx, ny, nz;

 	   /* integer coordinates of current point */
    	   ix = im3d->vinfo->i1 ; nx = DSET_NX(dset);
    	   jy = im3d->vinfo->j2 ; ny = DSET_NY(dset); 
    	   kz = im3d->vinfo->k3 ; nz = DSET_NZ(dset);

//    	   if( ix < 0 || ix >= nx ) return ;  /* error, outside of data bounds */
//      	   if( jy < 0 || jy >= ny ) return ;  /* error */
//    	   if( kz < 0 || kz >= nz ) return ;  /* error */

	   float dx = fabs(dset->daxes->xxdel) ,
	         dy = fabs(dset->daxes->yydel) ,
	         dz = fabs(dset->daxes->zzdel)  ;

	   if(debug) {
	     fprintf(stderr,"%s: (dx,dy,dz) = (%f,%f,%f)\n",program_name,dx,dy,dz); 
	   }

	   /* Update vtk window
	      pass needed data, ie. picked location.  any other params */

	   if(debug) {
	      fprintf(stderr ,"updated open vtk window with new location\n" );
	      fprintf(stderr ,"\tinteger coords of new point (%d, %d, %d)\n",
		      ix, jy,kz );
	      fprintf(stderr ,"\tdimensions of MRI_IMAGE (%d, %d, %d)\n",
		      nx, ny, nz );
	      fprintf(stderr ,"\tdata spacing: (%f, %f, %f)\n",
		      dx, dy, dz );
	   }

	   /* This code will later drive the VTK window.  
	      On click, initiate a call to subroutine that just prints out the 
	      DTI coefficients or the SHD coefficients. */

   	   THD_3dim_dataset *new_dset = NULL;
	   int cancelcode = 0;
	   char * output_prefix;

               // receive flags showing which data to be visualized: 
	       // signal, odf, shd, or tensor
          int voxel_signal_on, voxel_odf_on, voxel_shd_on, voxel_dt2_on;
	  receive_visualFlag_from_vtk_window( &voxel_signal_on,
	    &voxel_odf_on, &voxel_shd_on, &voxel_dt2_on );

//	   if( VIEW_DATA && !SHD_COMPUTE_COEFFICIENTS) {
	   if( voxel_signal_on) {
	     if (func_dset) {
	       // new code
	       /* integer coordinates of current point */
	       int fnx = DSET_NX(func_dset);
	       int fny = DSET_NY(func_dset);
	       int fnz = DSET_NZ(func_dset);
	       float fxi = im3d->vinfo->xi;
	       float fyj = im3d->vinfo->yj;
	       float fzk = im3d->vinfo->zk;

	       THD_ivec3 iv = THD_3dmm_to_3dind(func_dset, 
						TEMP_FVEC3(fxi, fyj, fzk));

	       int ii = THREE_TO_IJK(iv.ijk[0], iv.ijk[1], iv.ijk[2],
				     fnx, fnx*fny);
	       if( debug ) {
		 fprintf(stderr, "Data (vox %d located @ %d %d %d):\n",
			 ii, iv.ijk[0], iv.ijk[1], iv.ijk[2] );
	       }
	       
	       int ndif = DSET_NVALS(func_dset);
	       if (ndif == 9 || ndif == 25 || ndif == 49) {
		 /* SHD data */
		 float ** fptr = NULL;
		 fptr = (float **) malloc (sizeof (float *) * ndif );

		 for (int kk = 0; kk < ndif; kk++) {
		   fptr[kk] = (float*) DSET_ARRAY(func_dset, kk);
		 }
	       }
	       else if (ndif == 20) {
		 /* DTI data */

		 /* only load eigenvalues */
		 float **fptr = NULL;
		 float *fxar = NULL;
		 int eval_offset = 6;
		 fptr = (float **) malloc (sizeof (float *) * 3);
		 fxar = (float *) malloc (sizeof (float) * 3);

		 for (int kk = 0; kk < 3; kk++) {
		   fptr[kk] = (float*) DSET_ARRAY(func_dset, eval_offset+kk);
		   fxar[kk] = fptr[kk][ii];
		 }
	       }
	     }
	     // #ifdef OLD_CODE
	      int ijkvox; 
	      ijkvox = THREE_TO_IJK(ix, jy, kz, nx, nx*ny); 

	      if( debug ) {
		fprintf(stderr, "Data (vox %d located @ %d %d %d):\n",
			ijkvox, ix, jy, kz );
	      }
	      if( dset != NULL ) {

		DSET_load( dset ); 

		if(debug) {
			if( ISANAT( dset ) )
			  printf("Anat dataset \n" );
			else if( ISFUNC( dset ) )
			  printf("Func dataset \n" );
			else
			  printf("Error: dataset type not known \n" );
		}
		fprintf(stderr, "Current dataset: %s\n", DSET_PREFIX(dset));
		int nvox     = DSET_NVOX(dset);
		int ndif     = DSET_NVALS( dset );
		int typeDset = DSET_BRICK_TYPE( dset, 0 );
		if( typeDset != MRI_short )
		  fprintf(stderr,"Debug: check to make sure we are looking at shorts\n");

		int nvert = current_dif_state->img->dgrad->nvert ;
		printf("%s: nvox = %i, ndif = %i, nvert = %i\n",program_name,nvox,ndif,nvert);

		/* allocate memory */
		short ** sptr = NULL;
		float *  fxar = NULL;

		int kk,ii;

		sptr = (short **) malloc (sizeof (short *) * nvert );
		if( sptr == NULL ) 
		  fprintf(stderr, "Debug: (sptr == NULL) We have a problem, outcome uncertain\n");

		fxar = (float *) malloc (sizeof (float) * nvert );
		if( fxar == NULL ) 
		  fprintf(stderr, "Debug: (fxar == NULL) We have a problem, outcome uncertain\n");

		/* point to the dataset
		   voxel #i at time #k is sptr[k][i];
		   for i = 0 to nvox-1
		   k = 0 to nvert -1 */

		/* Note we skip the first array, which is the b=0 image.
		   There are ndif nvals */
		for( kk = 0; kk < nvert; kk++ ) sptr[kk] = (short*) DSET_ARRAY(dset, kk+1);

		/* now load data for correct voxel (int ii =vox) we are interested in */
		ii = ijkvox;
		for(kk=0; kk<nvert; kk++) fxar[kk] = (float) sptr[kk][ii];
		
		/* current_dif_state should contain the values that we computed  */
		if(debug)
		  for( kk =0; kk < nvert; kk++ ) 
		    fprintf(stderr ,"data[%i] = %f, voxel location x(%d) y(%d) slice %d \n",
			    kk,fxar[kk], ix, jy, kz );

		vtkFloatArray *vSignal = vtk_LoadSignal(fxar, nvert);
		vtkPolyData *vGridSig = MapSignal2VertexGrid(vGrid, vSignal) ;

		Send_Data_to_vtk_window(vGridSig); 

		vGridSig->Delete() ;
		vSignal->Delete() ;
		FREEUP(fxar) ;
		FREEUP(sptr) ;
	      }
	      else {
	         fprintf(stderr , "functional dataset to view data not loaded\n");
	      }
	      // #endif
	   }
//	   else if( DTI_COMPUTE_COEFFICIENTS ) {
	   if( voxel_dt2_on ) {

	      int ijkvox; 
	      ijkvox = THREE_TO_IJK(ix, jy, kz, nx, nx*ny); 

	      if( debug ) {
		fprintf(stderr, "DTI coefficients (vox %d located @ %d %d %d):\n",
			ijkvox, ix, jy, kz );
	      }
	      if( dt2_dset != NULL ) {

		DSET_load( dt2_dset ); 

		if(debug) {
			if( ISANAT( dt2_dset ) )
			  printf("Anat dataset \n" );
			else if( ISFUNC( dt2_dset ) )
			  printf("Func dataset \n" );
			else
			  printf("Error: dataset type not known \n" );
		}

		int ndif     = DSET_NVALS( dt2_dset );
		int typeDset = DSET_BRICK_TYPE( dt2_dset, 0 );
		if( typeDset != MRI_float )
		  fprintf(stderr,"Debug: check to make sure we are looking at floats\n");

		/* allocate memory */
		float ** fptr = NULL;
		float *  fxar = NULL;
		float *  val  = NULL;
		int kk,ii, nbrik;
		
		nbrik = current_dif_state->dti->prc->edim;
		if(debug) fprintf(stderr ,"eigenvalue vector dimension: %d\n", nbrik );

		fptr = (float **) malloc (sizeof (float *) * ndif );
		if( fptr == NULL ) 
		  fprintf(stderr, "Debug: (fptr == NULL) We have a problem, outcome uncertain\n");   

		fxar = (float *) malloc (sizeof (float) * ndif );
		if( fxar == NULL ) 
		  fprintf(stderr, "Debug: (fxar == NULL) We have a problem, outcome uncertain\n");
		
		/* point to the dataset
		   voxel #i at time #k is fptr[k][i];
		   for i =0 to nvox-1
		   k =0 to ndif -1 */

		for( kk = 0; kk < ndif; kk++ ) fptr[kk] = (float*) DSET_ARRAY(dt2_dset, kk);

		/* now load data for correct voxel (int ii =vox) we are interested in */
		ii = ijkvox;
		if(debug) printf("Loading fxar for voxel %i\n",ii); fflush(stdout);
		for( kk =0; kk < ndif; kk++ ) fxar[kk] = fptr[kk][ii];
		
		/* current_dif_state should contain the values that we computed  */
		if(debug)
		  fprintf(stderr ,"eigenvalues %f %f %f, voxel location x(%d) y(%d) slice %d \n",
			  fxar[6], fxar[7], fxar[8], ix, jy, kz );
	        
		 Send_DT2_to_vtk_window( fxar );
                 FREEUP( fxar );
		 FREEUP( fptr );
	      }
	      else {
	         fprintf(stderr , "functional dataset to compute DTI coefficients not loaded\n");
	      }
	   }
//	   if ( SHD_COMPUTE_COEFFICIENTS ) {
	   if ( voxel_odf_on || voxel_shd_on ) {
	      int ijkvox; 
	      ijkvox = THREE_TO_IJK(ix, jy, kz, nx, nx*ny); 
	      
	      if( debug ) {
		   fprintf(stderr, "SHD coefficients:\n" );
		   fprintf(stderr, "\tretrieving coords from image plane: ");  
		   fprintf(stderr, "(i=%d j=%d k=%d) ijkvox=%d\n", ix, jy, kz, ijkvox );  
	      }
	      if( shd_dset != NULL ) {

		 DSET_load( shd_dset ); 

		if(debug) {
			if( ISANAT( shd_dset ) )
			  printf("Anat dataset \n" );
			else if( ISFUNC( shd_dset ) )
			  printf("Func dataset \n" );
			else
			  printf("Error: dataset type not known \n" );
		}

		int ndif = DSET_NVALS( shd_dset );
		int typeDset = DSET_BRICK_TYPE( shd_dset, 0 );
		if( typeDset != MRI_complex )
		  fprintf(stderr, 
			  "Debug: check to make sure we are looking at complex data\n");

		complex** cptr = NULL;
		complex*  cxar = NULL;
		complex*  cval  = NULL;
		int kk,ii, nbrik;
		
		nbrik = current_dif_state->shd->calc->nlm;
		if(debug) fprintf(stderr ,"#shd coeffiecnts: %d\n", nbrik );
	
		cptr = (complex**) malloc (sizeof (complex*) * ndif );
		if( cptr == NULL ) 
		  fprintf(stderr,
			  "Debug: (cptr == NULL) We have a problem, outcome uncertain\n");
		cxar = (complex*) malloc (sizeof (complex) * ndif );
		if( cxar == NULL ) 
		  fprintf(stderr,
			  "Debug: (cxar == NULL) We have a problem, outcome uncertain\n");

		/* point to the dataset
		   voxel #i at time #k is fptr[k][i];
		   for i =0 to nvox-1
		   k =0 to ndif -1 */

		for( kk = 0; kk < ndif; kk++ ) cptr[kk] = (complex*) DSET_ARRAY(shd_dset, kk);

		/* now load data for correct voxel (int ii =vox) we are interested in */
		ii = ijkvox;
		for( kk =0; kk < ndif; kk++ ) cxar[kk] = cptr[kk][ii];
		
		/* for now, just print out data
		   HACK FOR NOW,  seems correct but need to look into what is actually 
		   happening here */
  		current_dif_state->shd->calc->Alm = alm_alloc( current_dif_state->shd->calc->Lmax ) ;
  		current_dif_state->shd->calc->Olm = alm_alloc( current_dif_state->shd->calc->Lmax ) ;

		if( debug && current_dif_state->shd->calc->Alm == NULL )
		  fprintf(stderr,  "error: shd->Alm is not initialized yet\n" );

		zvtoAlm( current_dif_state->shd->calc->Alm, cxar, 
			 current_dif_state->shd->calc->Lmax );
		alm2olm(current_dif_state->shd->calc->Alm,
			current_dif_state->shd->calc->Olm,
			current_dif_state->shd->calc->Lmax);
		if(debug) {
		  fprintf(stderr , "Alm values for voxel location x(%d) y(%d) slice %d: \n",	
			  ix, jy, kz );
		  printAlm( current_dif_state->shd->calc->Alm,
			    current_dif_state->shd->calc->Lmax );
		}
		if( voxel_shd_on )
		Send_Alm_to_vtk_window(current_dif_state->shd->calc->Alm,
				       current_dif_state->shd->calc->Lmax);
                if( voxel_odf_on )
		Send_Olm_to_vtk_window(current_dif_state->shd->calc->Olm,
				       current_dif_state->shd->calc->Lmax);
	      }
	      else {
	         fprintf(stderr , "functional dataset to compute SHD coefficients not loaded\n");
	      }
	   }
	 }

	 if( vtk_vector_window_status == VTK_VECTOR_WINDOW_OPEN ) {
	   send_xhair_to_vtk_vector_window(im3d->vinfo->xi,
					   im3d->vinfo->yj,
					   im3d->vinfo->zk);
	 }

         if( !xhair_flag || !dynamic_flag || diffuse_handle == NULL ) return ;

         CHECK_XHAIR_ERROR ;

         if( CHECK_XHAIR_MOTION ){
            FREE_VOLUMES ;
            /* DIFF_draw_CB(NULL,NULL,NULL) ; */
         }
      }
      return ;

      /*-- user drew something --*/

      case RECEIVE_DRAWNOTICE:{   /* 30 Mar 1999 */

         int doit=0 ;

         if( EQUIV_DSETS(im3d->anat_now,dset) ||    /* can't tell if user */
             EQUIV_DSETS(im3d->fim_now,dset)    ){  /* is drawing on anat */

            doit = 1 ; FREE_VOLUMES ;

         }

         if( EQUIV_DSETS(im3d->anat_now,func_dset) ||    /* or is drawing */
             EQUIV_DSETS(im3d->fim_now,func_dset)    ){  /* on the fim    */

            doit = 1 ; INVALIDATE_OVERLAY ;

            /* 15 Jun 1999: fix the range labels */

            { XmString xstr ;
              xstr = DIFF_range_label() ;
              XtVaSetValues( wshdfunc_range_label , XmNlabelString , xstr , NULL ) ;
              XmStringFree(xstr) ;

              xstr = DIFF_autorange_label() ;
              XtVaSetValues( wshdfunc_range_bbox->wbut[0], XmNlabelString,xstr , NULL ) ;
              XmStringFree(xstr) ;
            }
         }

         /* if( doit && dynamic_flag && diffuse_handle != NULL )
            DIFF_draw_CB(NULL,NULL,NULL) ; */
      }
      return ;

      /*-- dataset pointers have changed --*/

      case RECEIVE_DSETCHANGE:{   /* 31 Mar 1999 */

         if( dset != NULL )
            dset = PLUTO_find_dset( &dset_idc ) ;

         if( func_dset != NULL )
            func_dset = PLUTO_find_dset( &func_dset_idc ) ;

         if( ftm_anat_dset != NULL )
	   ftm_anat_dset = PLUTO_find_dset( &ftm_anat_dset_idc ) ;

         FREE_VOLUMES ; INVALIDATE_OVERLAY ;

         (void) MCW_popup_message( reload_pb ,
                                     "********** NOTICE ***********\n"
                                     "* Session rescan has forced *\n"
                                     "* purge of dataset brick(s) *\n"
                                     "* from memory.              *\n"
                                     "*****************************" ,
                                   MCW_USER_KILL | MCW_TIMER_KILL     ) ;
      }
      return ;

   }  /* end of switch on "why" */

   return ;
}
#endif

/*-------------------------------------------------------------------------
   callbacks for other toggle buttons
---------------------------------------------------------------------------*/

void DIFF_dynamic_CB( Widget w , XtPointer client_data , XtPointer call_data )
{
   dynamic_flag = MCW_val_bbox( dynamic_bbox ) ;
   return ;
}

void DIFF_accum_CB( Widget w , XtPointer client_data , XtPointer call_data )
{
   accum_flag = MCW_val_bbox( accum_bbox ) ;
   return ;
}

/*-----------------------------------------------------------------------
   Overlay some lines showing the crosshair location.
   Note that this function assumes that the current anat dataset
   in AFNI is defined on exactly the same grid as the diffusing dataset.
-------------------------------------------------------------------------*/
#define GR(i,j,k) gar[(i)+(j)*nx+(k)*nxy]
#define OP(i,j,k) oar[(i)+(j)*nx+(k)*nxy]

#define GXH_GRAY  255
#define GXH_COLOR 127
#define OXH       255

void DIFF_xhair_overlay(void)
{
   int ix,jy,kz , nx,ny,nz,nxy , ii , gap , om ;
   byte * gar , * oar ;
   byte   gxh ,   oxh=OXH ;

   if( grim == NULL || opim == NULL ) return ;  /* error */

   gxh = (func_computed) ? GXH_COLOR : GXH_GRAY ;

   CHECK_XHAIR_ERROR ;

   ix = im3d->vinfo->i1 ; nx = grim->nx ;
   jy = im3d->vinfo->j2 ; ny = grim->ny ; nxy = nx * ny ;
   kz = im3d->vinfo->k3 ; nz = grim->nz ;

   om = im3d->vinfo->xhairs_orimask ;  /* 02 Jun 1999 */

   if( ix < 0 || ix >= nx ) return ;  /* error */
   if( jy < 0 || jy >= ny ) return ;  /* error */
   if( kz < 0 || kz >= nz ) return ;  /* error */

   gap = im3d->vinfo->crosshair_gap ;
   gar = MRI_BYTE_PTR(grim) ;
   oar = MRI_BYTE_PTR(opim) ;

   /* 02 Jun 1999: allow for partial crosshair drawing */

   if( (om & ORIMASK_LR) != 0 ){
      for( ii=0 ; ii < nx ; ii++ ){
         if( abs(ii-ix) > gap ){ GR(ii,jy,kz) = gxh ; OP(ii,jy,kz) = oxh ; }
      }
   }

   if( (om & ORIMASK_AP) != 0 ){
      for( ii=0 ; ii < ny ; ii++ ){
         if( abs(ii-jy) > gap ){ GR(ix,ii,kz) = gxh ; OP(ix,ii,kz) = oxh ; }
      }
   }

   if( (om & ORIMASK_IS) != 0 ){
      for( ii=0 ; ii < nz ; ii++ ){
         if( abs(ii-kz) > gap ){ GR(ix,jy,ii) = gxh ; OP(ix,jy,ii) = oxh ; }
      }
   }

   xhair_ixold = ix ; xhair_jyold = jy ; xhair_kzold = kz ;  /* memory */
   xhair_omold = om ;                                        /* 02 Jun 1999 */
   return ;
}

/*------------------------------------------------------------------
   Called when the user changes a graph - just signals that we
   need to reload the data.
--------------------------------------------------------------------*/

void DIFF_graf_CB( MCW_graf * gp , void * cd )
{
   FREE_VOLUMES ;  /* free the volumes, will force reloading at redraw */
   return ;
}

/*-------------------------------------------------------------------
  Callback for clipping arrowvals
---------------------------------------------------------------------*/

void DIFF_clip_CB( MCW_arrowval * av , XtPointer cd )
{
   FREE_VOLUMES ;  /* free the volumes, will force reloading */

   if( clipbot_av->ival >= cliptop_av->ival ){
      if( av == clipbot_av )
         AV_assign_ival( clipbot_av , cliptop_av->ival - 1 ) ;
      else
         AV_assign_ival( cliptop_av , clipbot_av->ival + 1 ) ;
   }

   /* if brick is scaled, re-show the scaled labels */

   if( brickfac != 0.0 && brickfac != 1.0 ){
      char minch[16] , maxch[16] , str[64] ;
      XmString xstr ;

      if( av == clipbot_av ){
         AV_fval_to_char( brickfac * clipbot_av->ival , minch ) ;
         sprintf(str,"[-> %s]",minch) ;
         xstr = XmStringCreateLtoR( str , XmFONTLIST_DEFAULT_TAG ) ;
         XtVaSetValues( clipbot_faclab , XmNlabelString , xstr , NULL ) ;
         XmStringFree(xstr) ;
      } else {
         AV_fval_to_char( brickfac * cliptop_av->ival , maxch ) ;
         sprintf(str,"[-> %s]",maxch) ;
         xstr = XmStringCreateLtoR( str , XmFONTLIST_DEFAULT_TAG ) ;
         XtVaSetValues( cliptop_faclab , XmNlabelString , xstr , NULL ) ;
         XmStringFree(xstr) ;
      }
   }

   return ;
}

/*-----------------------------------------------------------------------
   Evaluate an arrowval string that may be a number or may be
   a more complicated arithmetic expression.  Input variables are
   stored in the global array atoz[] ;
-------------------------------------------------------------------------*/

float DIFF_evaluate( MCW_arrowval * av )
{
   PARSER_code * pcode ;
   char * str , * cpt ;
   float val ;

   /* get the string to convert */

   if( av        == NULL ) return 0.0 ;        /* these cases should */
   if( av->wtext == NULL ) return av->fval ;   /* never happen       */

   str = XmTextFieldGetString( av->wtext ) ;
   if( str == NULL || str[0] == '\0' ){ XtFree(str) ; return 0.0 ; }

   /* try a regular numerical conversion */

   val = strtod( str , &cpt ) ;

   for( ; *cpt != '\0' && isspace(*cpt) ; cpt++ ) ; /* skip blanks */

   if( *cpt == '\0' ){ XtFree(str); AV_assign_fval(av,val); return val; }

   /* try to parse an expression */

   pcode = PARSER_generate_code( str ) ;
   if( pcode == NULL ){ XtFree(str) ; return 0.0 ; }

   val = PARSER_evaluate_one( pcode , atoz ) ; free(pcode) ;

   XtFree(str) ; return val ;
}

/*-------------------------------------------------------------------------
   When the user presses Enter in a textfield that might have an
   expression entered, evaluate the expression numerically and
   load that number back into the textfield.
---------------------------------------------------------------------------*/

void DIFF_textact_CB( Widget wtex, XtPointer client_data, XtPointer call_data )
{
   MCW_arrowval * av         = (MCW_arrowval *) client_data ;
   XmAnyCallbackStruct * cbs = (XmAnyCallbackStruct *) call_data ;
   float sval ;
   int iv ;

   MCW_invert_widget(wtex) ;

   sval = DIFF_evaluate( av ) ;
   AV_assign_fval( av , sval ) ;

   MCW_invert_widget(wtex) ;
   return ;
}

/**********************************************************************
   Stuff to deal with the image display window
***********************************************************************/

/*-----------------------------------------------------------------------
   Open an image display window.
-------------------------------------------------------------------------*/

static int any_rgb_images ;

void DIFF_open_imseq( void )
{
   int ntot , ii ;
   int debug = 0;

   if( imseq != NULL      ||
       diffusings == NULL || IMARR_COUNT(diffusings) == 0 ) return ;

   ntot = IMARR_COUNT(diffusings) ;

   any_rgb_images = 0 ;
   for( ii=0 ; ii < ntot ; ii++ ){
      if( IMARR_SUBIMAGE(diffusings,ii) != NULL &&
          IMARR_SUBIMAGE(diffusings,ii)->kind == MRI_rgb ){

         any_rgb_images = 1 ; break ;
      }
   }

   imseq = open_MCW_imseq(dc,
			  reinterpret_cast<void* (*)()>(DIFF_imseq_getim),
			  NULL ) ;

   drive_MCW_imseq( imseq , isqDR_clearstat , NULL ) ;

   { ISQ_options opt ;       /* change some options from the defaults */

     ISQ_DEFAULT_OPT(opt) ;
     opt.save_one = False ;  /* change to Save:bkg */
     opt.save_pnm = False ;
     drive_MCW_imseq( imseq , isqDR_options      , (XtPointer) &opt ) ;
     drive_MCW_imseq( imseq , isqDR_periodicmont , (XtPointer) 0    ) ;
   }

   /* make it popup */

   drive_MCW_imseq( imseq , isqDR_realize, NULL ) ;

   drive_MCW_imseq( imseq , isqDR_title, "AFNI Diffusions" ) ;

   /* lrf - turn on wid always */
   if( ntot == 1 ) {
     if(debug) printf("DIFF_open_imseq: ntot=%d, num_total=%d num_series=%d\n",
		      ntot,imseq->status->num_total , imseq->status->num_series ) ;
     drive_MCW_imseq( imseq , isqDR_onoffwid , (XtPointer) isqDR_onwid ) ;
     /* drive_MCW_imseq( imseq , isqDR_onoffwid , (XtPointer) isqDR_offwid ) ; */
   } else {
     drive_MCW_imseq( imseq , isqDR_onoffwid , (XtPointer) isqDR_onwid ) ;
   }

   drive_MCW_imseq( imseq , isqDR_reimage , (XtPointer) (ntot-1) ) ; 

#ifndef DONT_INSTALL_ICONS
   if( afni48_good && afni48ren_pixmap == XmUNSPECIFIED_PIXMAP ){
      Pixel bg_pix , fg_pix  ;

      XtVaGetValues( info_lab ,
		     XmNforeground , &fg_pix ,
		     XmNbackground , &bg_pix ,
                     NULL ) ;

      afni48ren_pixmap = XCreatePixmapFromBitmapData(
                            XtDisplay(shell) ,
                            RootWindowOfScreen(XtScreen(shell)) ,
                            afni48ren_bits , afni48ren_width , afni48ren_height ,
                            bg_pix , fg_pix ,
                            DefaultDepthOfScreen(XtScreen(shell)) ) ;

   }
   if( afni48_good )
         drive_MCW_imseq( imseq,isqDR_icon , (XtPointer) afni48ren_pixmap ) ;
#endif

   return ;
}

void DIFF_update_imseq( void )
{
   int ntot , ii ;
   int debug = 0;

   if( imseq == NULL ){ DIFF_open_imseq() ; return ; }
   if( diffusings == NULL || IMARR_COUNT(diffusings) == 0 ) return ;

   ntot = IMARR_COUNT(diffusings) ;

   any_rgb_images = 0 ;
   for( ii=0 ; ii < ntot ; ii++ ){
      if( IMARR_SUBIMAGE(diffusings,ii) != NULL &&
          IMARR_SUBIMAGE(diffusings,ii)->kind == MRI_rgb ){

         any_rgb_images = 1 ; break ;
      }
   }

   drive_MCW_imseq( imseq , isqDR_newseq , NULL ) ;

   if(debug) printf("DIFF_update_imseq: ntot = %i\n",ntot);

   /* lrf - turn on wid always */

   if( ntot == 1 ) {
     if(debug) printf("DIFF_update_imseq: ntot=%d, num_total=%d num_series=%d\n",
		      ntot,imseq->status->num_total , imseq->status->num_series ) ;
     drive_MCW_imseq( imseq , isqDR_onoffwid , (XtPointer) isqDR_onwid ) ;
     /* drive_MCW_imseq( imseq , isqDR_onoffwid , (XtPointer) isqDR_offwid ) ; */
   } else {
     drive_MCW_imseq( imseq , isqDR_onoffwid , (XtPointer) isqDR_onwid ) ;
   }

   drive_MCW_imseq( imseq , isqDR_reimage , (XtPointer)(ntot-1) ) ; 

   return ;
}

void DIFF_destroy_imseq( void )
{
   if( imseq == NULL ) return ;
   drive_MCW_imseq( imseq , isqDR_destroy , NULL ) ;
   return ;
}

/*------------------------------------------------------------------
   Routine to provide data to the imseq.
   Just returns the control information, or the selected image.
--------------------------------------------------------------------*/

XtPointer DIFF_imseq_getim( int n , int type , XtPointer handle )
{
   int ntot = 0 ;

   if( diffusings != NULL ) ntot = IMARR_COUNT(diffusings) ;
   if( ntot < 1 ) ntot = 1 ;

   /*--- send control info ---*/

   if( type == isqCR_getstatus ){
      MCW_imseq_status * stat = myXtNew( MCW_imseq_status ) ;  /* will be free-d */
                                                               /* when imseq is */
                                                               /* destroyed    */
      stat->num_total  = ntot ;
      stat->num_series = stat->num_total ;
      stat->send_CB    = reinterpret_cast<void (*)()>(DIFF_seq_send_CB);
      stat->parent     = NULL ;
      stat->aux        = NULL ;

      stat->transforms0D = &(GLOBAL_library.registered_0D) ;
      stat->transforms2D = &(GLOBAL_library.registered_2D) ;

      return (XtPointer) stat ;
   }

   /*--- no overlay, never ---*/

   if( type == isqCR_getoverlay ) return NULL ;

   /*--- return a copy of a diffused image
         (since the imseq will delete it when it is done) ---*/

   if( type == isqCR_getimage || type == isqCR_getqimage ){
      MRI_IMAGE * im = NULL , * rim ;

      if( diffusings != NULL ){
         if( n < 0 ) n = 0 ; else if( n >= ntot ) n = ntot-1 ;
         rim = IMARR_SUBIMAGE(diffusings,n) ;
         if( any_rgb_images )
            im = mri_to_rgb( rim ) ;
         else
            im = mri_to_mri( rim->kind , rim ) ;
      }
      return (XtPointer) im ;
   }

   return NULL ; /* should not occur, but who knows? */
}

/*---------------------------------------------------------------------------
   Routine called when the imseq wants to send a message.
   In this case, all we need to handle is the destroy message,
   so that we can free some memory.
-----------------------------------------------------------------------------*/

void DIFF_seq_send_CB( MCW_imseq * seq , XtPointer handle , ISQ_cbs * cbs )
{
  int debug = 0;

  if(debug) printf("DIFF_seq_send_CB: *************************GOT HERE!\n\a");

   switch( cbs->reason ){
      case isqCR_destroy:{
         myXtFree(imseq) ; imseq = NULL ;
      }
      break ;
   }
   return ;
}

/*----------------------------------------------------------------------------
   Automation callbacks
------------------------------------------------------------------------------*/

void DIFF_autoflag_CB( Widget w , XtPointer client_data , XtPointer call_data )
{
   int flag = MCW_val_bbox( automate_bbox ) ;
   XtSetSensitive( autocompute_pb , (Boolean) flag ) ;
   return ;
}

/*--------------------------------------------------------------------------
   Drive the automatic computation
----------------------------------------------------------------------------*/

static int autokill ;

void DIFF_autocompute_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   int it , ntime = autoframe_av->ival ;
   float scl = 100.0/ntime ;
   Widget autometer ;

   automate_flag = 1 ;
   if( ! accum_flag ){
      DESTROY_IMARR(diffusings) ;
   }

   atoz[N_IND] = ntime ;

   autometer = MCW_popup_meter( shell , METER_TOP_WIDE ) ;

   XtManageChild( autocancel_pb ) ; AFNI_add_interruptable( autocancel_pb ) ;
   autokill = 0 ;

   for( it=0 ; it < ntime ; it++ ){
      atoz[T_IND] = it ;
      AV_assign_ival( autoframe_av , it+1 ) ;

      /* DIFF_draw_CB(NULL,NULL,NULL) ; */

      if( it < ntime-1 ){
         AFNI_process_interrupts(autocancel_pb) ;
         if( autokill ) break ;
      }

      MCW_set_meter( autometer , (int)(scl*(it+1)) ) ;
   }

   MCW_popdown_meter( autometer ) ;

   /*-- done: turn off automation --*/

   MCW_set_bbox( automate_bbox , 0 ) ;
   XtSetSensitive( autocompute_pb , False ) ;

   XtUnmanageChild( autocancel_pb ) ; AFNI_add_interruptable(NULL) ;

   automate_flag = 0 ;
   return ;
}

/*--------------------------------------------------------------------------
   Set a flag to cancel the automatic computation
----------------------------------------------------------------------------*/

void DIFF_autocancel_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   if( autokill ){ XBell(dc->display,100) ; return ; }
   autokill = 1 ;
}

/*--------------------------------------------------------------------------
   Drive the anisotropy computation
----------------------------------------------------------------------------*/

int anisokill ; 

void DIFF_aniso_compute_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
  /* int it , ntime = shd_lval_av->ival ;
   float scl = 100.0/ntime ;
   Widget anisometer ;*/

   int debug = 0;
   int nvox;

   /*aniso_compute_flag = 1 ;*/	/* do we use this? */

   /* Check input integrity */

   if(dset==NULL) {
      (void) MCW_popup_message( w ,
				"\n** No dataset loaded! **\n",
                                MCW_USER_KILL | MCW_TIMER_KILL ) ;
      PLUTO_beep() ; return ;
   }

   if(!dti_flag && !shd_flag && !dwi2dt_flag) {
      (void) MCW_popup_message( w ,
				"\n** No anisotropy method(s) chosen! **\n",
                                MCW_USER_KILL | MCW_TIMER_KILL ) ;
      PLUTO_beep() ; return ;
   }
   nvox = dset->daxes->nxx * dset->daxes->nyy * dset->daxes->nzz ;

   /* Choose method */

   if(dti_flag) {
     if(debug) printf("DTI computation enabled: (dti_flag = %i)\n",dti_flag);
     DIFF_dti_compute_CB(aniso_cancel_pb,NULL,NULL) ;
   }

   if(shd_flag) {
     if(debug) printf("SHD computation enabled: (shd_flag = %i)\n",shd_flag);
     if(swd_flag) {if(debug) printf("SWD computation enabled: (swd_flag = %i)\n",swd_flag);}
     DIFF_shd_compute_CB(aniso_cancel_pb,NULL,NULL) ;
   }

   if (dwi2dt_flag) {
     DIFF_dwi2dt_compute_CB(aniso_cancel_pb,NULL,NULL);
   }

   /*-- done: turn off diffusion --*/

   /*aniso_compute_flag = 0 ;*/
   return ;
}

/*--------------------------------------------------------------------------
   Set a flag to cancel the dti scalar computation
----------------------------------------------------------------------------*/

void DIFF_dti_cancel_scalars_CB( Widget w, XtPointer client_data, XtPointer call_data )
{

   if( anisokill ){ XBell(dc->display,100) ; return ; }
   anisokill = 1 ;

}

/*--------------------------------------------------------------------------
   Set a flag to cancel the shd scalar computation
----------------------------------------------------------------------------*/

void DIFF_shd_cancel_scalars_CB( Widget w, XtPointer client_data, XtPointer call_data )
{

   if( anisokill ){ XBell(dc->display,100) ; return ; }
   anisokill = 1 ;

}

/*--------------------------------------------------------------------------
   Set a flag to cancel the anisotropy computation
----------------------------------------------------------------------------*/

void DIFF_aniso_cancel_CB( Widget w, XtPointer client_data, XtPointer call_data )
{

   if( anisokill ){ XBell(dc->display,100) ; return ; }
   anisokill = 1 ;

}

/*--------------------------------------------------------------------------
   Drive the fiber computation
----------------------------------------------------------------------------*/

int fiberkill ; 

void  DIFF_fiber_compute_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   /* int it , ntime = shd_lval_av->ival ;
      float scl = 100.0/ntime ;
      Widget anisometer ;*/

   int debug = 1;
   int nvox;

   fiber_compute_flag = 1 ;	/* do we use this? */

   /* Check input integrity */

   if(dset==NULL) {
      (void) MCW_popup_message( w ,
				"\n** No dataset loaded! **\n",
                                MCW_USER_KILL | MCW_TIMER_KILL ) ;
      PLUTO_beep() ; return ;
   }

  if(!ftm1_flag && !ftm2_flag && !ftm_dst_flag ) {
      (void) MCW_popup_message( w ,
				"\n** No fiber tracking method(s) chosen! **\n",
                                MCW_USER_KILL | MCW_TIMER_KILL ) ;
      PLUTO_beep() ; return ;
   }

   nvox = dset->daxes->nxx * dset->daxes->nyy * dset->daxes->nzz ;

   /* Choose method */

   if(ftm1_flag) {
     if(debug) printf("FTM1 computation enabled: (ftm1_flag = %i)\n",ftm1_flag);
     DIFF_ftm1_compute_CB(fiber_cancel_pb,NULL,NULL) ;
   }

   if(ftm2_flag) {
     if(debug) printf("FTM2 computation enabled: (ftm2_flag = %i)\n",ftm2_flag);
     DIFF_ftm2_compute_CB(fiber_cancel_pb,NULL,NULL) ;
   }

   if(ftm_dst_flag) {
     if(debug) printf("FTM_DST computation enabled: (ftm_dst_flag = %i)\n",
         ftm_dst_flag);
 //    if( xhari_recv	 
     DIFF_ftm_dst_compute_CB(fiber_cancel_pb,NULL,NULL) ;
//printf("===========DIFF-ftm-dst=======xhair_recv=%d\n",xhair_recv);   
//     xhair_recv = AFNI_receive_init( im3d ,
//                                   RECEIVE_VIEWPOINT_MASK , DIFF_xhair_recv , NULL,
//				   "DIFF_xhair_recv") ;

   }
 

   /*-- done: turn off fiber tracking --*/

   fiber_compute_flag = 0 ;
   return ;
}

/*--------------------------------------------------------------------------
   Set a flag to cancel the fiber computation
----------------------------------------------------------------------------*/

void DIFF_fiber_cancel_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   if( fiberkill ){ XBell(dc->display,100) ; return ; }
   fiberkill = 1 ;
}

/*--------------------------------------------------------------------------
   Drive the VTK visualization viewer 
----------------------------------------------------------------------------*/
// current cyrus
void DIFF_show_vtkvis_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   int debug = 0;
   int nvox = 0;
   char run_program[128];

      
   //system(run_program);

   /* tmp code to get things started */
    /* Check input integrity */

    if(dset==NULL) {
       (void) MCW_popup_message( w ,
				 "\n** No dataset loaded! **\n",
				 MCW_USER_KILL | MCW_TIMER_KILL ) ;
       PLUTO_beep() ; return ;
    }

    /* we're probably not going to do this, this creates a new DATAset. 
       want we want is to give us control of a new window powered by VTK  
       make a new directory called vtk.  have all the vtk code there. cyrus
       DIFF_dti_compute_CB(aniso_cancel_pb,NULL,NULL) ;  */

    /* launch vtk window */

    Launch_vtk_window(shell); 

    switch( VTK_window_status ) {
      case VTK_WINDOW_CLOSED:
        /*(void) MCW_popup_message( w ,
				 "\n** vtk window is closed: need to open! **\n",
				 MCW_USER_KILL | MCW_TIMER_KILL ) ;
				 fprintf(stderr, "opening vtk window\n" );*/
	 VTK_window_status = VTK_WINDOW_OPEN;

         break;
      case VTK_WINDOW_OPEN:
        /*(void) MCW_popup_message( w ,
				 "\n** vtk window is opened: just need to refresh! **\n",
				 MCW_USER_KILL | MCW_TIMER_KILL ) ;*/
	 break;
      default:
        (void) MCW_popup_message( w ,
				  "\n** vtk window is neither closed or opened: something is wrong! **\n",
				  MCW_USER_KILL | MCW_TIMER_KILL ) ;

	 break;
  }

   /* update status of VTK window */
   nvox = DSET_NVOX(dset);
   fprintf(stderr , "nvox:= %d\n", nvox );

   return ;
}

/*--------------------------------------------------------------------------
   Drive the diffusion theory viewer
----------------------------------------------------------------------------*/

void DIFF_show_theory_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   int goodviewer ;
   int debug = 0;

#define   ACRO_VIEWER 0		/* acroread  */
#define     GV_VIEWER 1		/* ghostview */
#define SHOWPS_VIEWER 2		/* showps    */
#define    PDF_VIEWER 3		/* xpdf 17-Feb-04  */

   int TEXT_VIEWER = PDF_VIEWER;

   char theory_str[10000];	/* command line */
   char viewer[256];
   char paper_name[256];
   char paper_type[15];
   char docs_dir[32];
   char *lrfpath = getenv("AFNI_LRF_PATH") ;

   show_theory_flag = 1 ;

   switch (TEXT_VIEWER) {
      default:{
	sprintf(viewer,"xpdf");
	sprintf(paper_type,"pdf");
      }
      case PDF_VIEWER:{
	sprintf(viewer,"xpdf");
	sprintf(paper_type,"pdf");
      }
      break;
      case ACRO_VIEWER:{
	sprintf(viewer,"acroread");
	sprintf(paper_type,"pdf");
      }
      break;
      case GV_VIEWER:{
	sprintf(viewer,"ghostview");
	sprintf(paper_type,"ps");
      }
      break;
      case SHOWPS_VIEWER:{
	sprintf(viewer,"showps ");
	sprintf(paper_type,"ps");
      }
      break;
   }

   sprintf(docs_dir,"docs");

   switch ( theory_av->ival ) {

      default:

      case THEORY_DTI:{
	if(debug) printf("inside DIFF_show_theory_CB: ival = %i\n",THEORY_DTI);
	sprintf(paper_name,"bastin");
      }
      break;

      case THEORY_SDV:{
	if(debug) printf("inside DIFF_show_theory_CB: ival = %i\n",THEORY_SDV);
	sprintf(paper_name,"Frank_dvar_MRM");
      }
      break;

      case THEORY_SHD:{
	if(debug) printf("inside DIFF_show_theory_CB: ival = %i\n",THEORY_SHD);
	sprintf(paper_name,"Frank_dsph_MRM");
      }
      break;

      case THEORY_FTM:{
	if(debug) printf("inside DIFF_show_theory_CB: ival = %i\n",THEORY_FTM);
	sprintf(paper_name,"Basser_IEICE_02");
      }
      break;

      case THEORY_DST:{
	if(debug) printf("inside DIFF_show_theory_CB: ival = %i\n",THEORY_DST);
	sprintf(paper_name,"dst_ieee_tmi");
      }
      break;

      case THEORY_DTE:{
	if(debug) printf("inside DIFF_show_theory_CB: ival = %i\n",THEORY_DTE);
	sprintf(paper_name,"ISMRM_2006_DTI");
      }
      break;

   }

   sprintf(theory_str,"%s %s/%s/%s.%s&",viewer,lrfpath,docs_dir,paper_name,paper_type);
      
   if(debug) printf("theory_str = %s\n",theory_str);
   goodviewer = system(theory_str);
   if(debug) printf("goodviewer = %i\n",goodviewer );
   show_theory_flag = 0 ;

   return ;
}

/*--------------------------------------------------------------------------
   What happens when the user selects a sub-brick from a menu
----------------------------------------------------------------------------*/

void DIFF_choose_av_CB( MCW_arrowval * av , XtPointer cd )
{
   XmString xstr ;
   char str[2*THD_MAX_NAME] ;

   /*--- selection of a underlay sub-brick ---*/

   if( av == choose_av && dset != NULL && av->ival < DSET_NVALS(dset) ){

      float fac = DSET_BRICK_FACTOR(dset,av->ival) ;

      if( fac == 0.0 || fac == 1.0 ){  /* rewrite the informational label */
         strcpy(str,dset_title) ;
      } else {
         char abuf[16] ;
         AV_fval_to_char( fac , abuf ) ;
         sprintf(str,"%s [* %s]", dset_title , abuf ) ;
      }
      xstr = XmStringCreateLtoR( str , XmFONTLIST_DEFAULT_TAG ) ;
      XtVaSetValues( info_lab , XmNlabelString , xstr , NULL ) ;
      XmStringFree(xstr) ;

      dset_ival = av->ival ;   /* read this sub-brick    */
      new_dset = 1 ;           /* flag it as new         */
      FREE_VOLUMES ;           /* free the internal data */
      DIFF_reload_dataset() ;  /* load the data          */
      MCW_invert_widget(reload_pb) ;  /* lrf - hack added 8/5/01 to turn the signal off */

      /* draw */

      /* if( diffuse_handle != NULL ) DIFF_draw_CB(NULL,NULL,NULL) ;  */

   /*--- selection of overlay color sub-brick ---*/

   } else if( av == wshdfunc_color_av && func_dset != NULL && av->ival < DSET_NVALS(func_dset) ){

      float fac = DSET_BRICK_FACTOR(func_dset,av->ival) ;

      if( fac == 0.0 || fac == 1.0 ){  /* rewrite the informational label */
         strcpy(str,func_dset_title) ;
      } else {
         char abuf[16] ;
         AV_fval_to_char( fac , abuf ) ;
         sprintf(str,"%s [* %s]", func_dset_title , abuf ) ;
      }
      xstr = XmStringCreateLtoR( str , XmFONTLIST_DEFAULT_TAG ) ;
      XtVaSetValues( wshdfunc_info_lab , XmNlabelString , xstr , NULL ) ;
      XmStringFree(xstr) ;

      func_color_ival = av->ival ;

      /* fix the range labels */

      xstr = DIFF_range_label() ;
      XtVaSetValues( wshdfunc_range_label , XmNlabelString , xstr , NULL ) ;
      XmStringFree(xstr) ;

      xstr = DIFF_autorange_label() ;
      XtVaSetValues( wshdfunc_range_bbox->wbut[0], XmNlabelString,xstr , NULL ) ;
      XmStringFree(xstr) ;

      INVALIDATE_OVERLAY ;

   /*--- selection of overlay threshold sub-brick ---*/

   } else if( av == wshdfunc_thresh_av && func_dset != NULL && av->ival < DSET_NVALS(func_dset) ){

      func_thresh_ival = av->ival ;

      /* fix the range label */

      xstr = DIFF_range_label() ;
      XtVaSetValues( wshdfunc_range_label , XmNlabelString , xstr , NULL ) ;
      XmStringFree(xstr) ;

      /* fix the p-value label */

      DIFF_set_thr_pval() ;

      INVALIDATE_OVERLAY ;
   }

   return ;
}

/*---------------------------------------------------------------------------
   Make a label for a sub-brick selector menu
-----------------------------------------------------------------------------*/

char * DIFF_choose_av_label_CB( MCW_arrowval * av , XtPointer cd )
{
   static char blab[32] ;
   THD_3dim_dataset * dset = (THD_3dim_dataset *) cd ;
   static char * lfmt[3] = { "#%1d %-14.14s" , "#%2d %-14.14s" , "#%3d %-14.14s"  } ;
   static char * rfmt[3] = { "%-14.14s #%1d" , "%-14.14s #%2d" , "%-14.14s #%3d"  } ;

   if( ISVALID_3DIM_DATASET(dset) ){

#ifdef USE_RIGHT_BUCK_LABELS
      if( DSET_NVALS(dset) < 10 )
        sprintf(blab, rfmt[0] , DSET_BRICK_LABEL(dset,av->ival) , av->ival ) ;
      else if( DSET_NVALS(dset) < 100 )
        sprintf(blab, rfmt[1] , DSET_BRICK_LABEL(dset,av->ival) , av->ival ) ;
      else
        sprintf(blab, rfmt[2] , DSET_BRICK_LABEL(dset,av->ival) , av->ival ) ;
#else
      if( DSET_NVALS(dset) < 10 )
        sprintf(blab, lfmt[0] , av->ival , DSET_BRICK_LABEL(dset,av->ival) ) ;
      else if( DSET_NVALS(dset) < 100 )
        sprintf(blab, lfmt[1] , av->ival , DSET_BRICK_LABEL(dset,av->ival) ) ;
      else
        sprintf(blab, lfmt[2] , av->ival , DSET_BRICK_LABEL(dset,av->ival) ) ;
#endif
   }
   else
      sprintf(blab," #%d ",av->ival) ;  /* should not happen! */

   return blab ;
}

/*----------------------------------------------------------------------------------
   Make the widgets for the DTI functional overlay
------------------------------------------------------------------------------------*/

void DIFF_func_dti_widgets(void)
{

  printf("\aDIFF_func_dti_widgets not available!\n") ;

}

/*----------------------------------------------------------------------------------
   Make the widgets for the FTM functional overlay
------------------------------------------------------------------------------------*/

void DIFF_func_ftm_widgets(void)
{
  printf("\aDIFF_func_ftm_widgets not available!\n") ;
}

/*----------------------------------------------------------------------------------
   Make the widgets for the SHD functional overlay
------------------------------------------------------------------------------------*/

void DIFF_func_shd_widgets(void)
{
   XmString xstr ;
   Widget   wqqq ;
   int      sel_height ;

   /* top level managers */

   wshdfunc_vsep = SEP_VER(top_rowcol) ;
   XtUnmanageChild(wshdfunc_vsep ) ;

   wshdfunc_frame = XtVaCreateWidget(
                   "AFNI" , xmFrameWidgetClass , top_rowcol ,
                      XmNshadowType , XmSHADOW_ETCHED_IN ,
                      XmNshadowThickness , 5 ,
                      XmNtraversalOn , False ,
                      XmNinitialResourcesPersistent , False ,
                   NULL ) ;

   wshdfunc_uber_rowcol = XtVaCreateWidget(
                    "AFNI" , xmRowColumnWidgetClass , wshdfunc_frame ,
                       XmNorientation , XmVERTICAL ,
                       XmNpacking , XmPACK_TIGHT ,
                       XmNtraversalOn , False ,
                       XmNinitialResourcesPersistent , False ,
                    NULL ) ;

   xstr = XmStringCreateLtoR( NO_DATASET_STRING ,
                              XmFONTLIST_DEFAULT_TAG ) ;
   wshdfunc_info_lab = XtVaCreateManagedWidget(
                      "AFNI" , xmLabelWidgetClass , wshdfunc_uber_rowcol ,
                         XmNlabelString , xstr ,
                         XmNrecomputeSize , False ,
                         XmNinitialResourcesPersistent , False ,
                      NULL ) ;
   XmStringFree(xstr) ;

   SEP_HOR(wshdfunc_uber_rowcol) ;

   xstr = XmStringCreateLtoR( "Choose DTI Dataset" , XmFONTLIST_DEFAULT_TAG ) ;
   wshdfunc_choose_pb = XtVaCreateManagedWidget(
                  "AFNI" , xmPushButtonWidgetClass , wshdfunc_uber_rowcol ,
                     XmNalignment   , XmALIGNMENT_CENTER ,
                     XmNlabelString , xstr ,
                     XmNtraversalOn , False ,
                     XmNinitialResourcesPersistent , False ,
                  NULL ) ;
   XmStringFree(xstr) ;

   /*strcpy(diff_choose_str,DIFO_labels[0]);
   XtAddCallback( wshdfunc_choose_pb, XmNactivateCallback, DIFF_choose_CB,diff_choose_str) ;*/

   XtAddCallback( wshdfunc_choose_pb, XmNactivateCallback, DIFF_choose_CB,"OVER") ;

   SEP_HOR(wshdfunc_uber_rowcol) ;

   wshdfunc_rowcol = XtVaCreateWidget(
                    "AFNI" , xmRowColumnWidgetClass , wshdfunc_uber_rowcol ,
                       XmNorientation , XmHORIZONTAL ,
                       XmNpacking , XmPACK_TIGHT ,
                       XmNtraversalOn , False ,
                       XmNinitialResourcesPersistent , False ,
                    NULL ) ;

   /*------------------------ 1st column: threshold stuff ------------------------*/

   wshdfunc_thr_rowcol = XtVaCreateWidget(
                        "AFNI" , xmRowColumnWidgetClass , wshdfunc_rowcol ,
                           XmNorientation , XmVERTICAL ,
                           XmNpacking , XmPACK_TIGHT ,
                           XmNmarginHeight, 0 ,
                           XmNmarginWidth , 0 ,
                           XmNtraversalOn , False ,
                           XmNinitialResourcesPersistent , False ,
                        NULL ) ;

   xstr = XmStringCreateLtoR( "Thresh" , XmFONTLIST_DEFAULT_TAG ) ;
   wshdfunc_thr_label = XtVaCreateManagedWidget(
                       "AFNI" , xmLabelWidgetClass , wshdfunc_thr_rowcol ,
                          XmNlabelString , xstr ,
                          XmNrecomputeSize , False ,
                          XmNinitialResourcesPersistent , False ,
                       NULL ) ;
   XmStringFree(xstr) ;

 { int smax , stop , decim , sstep ;                  /* 30 Nov 1997:       */
   decim = THR_TOP_EXPON ;                            /* compute parameters */
   smax  = (int)( pow(10.0,decim) + 0.001 ) ;         /* for scale display. */
   stop  = smax - 1 ;
   sstep = smax / 1000 ;  if( sstep < 1 ) sstep = 1 ;

#ifdef BOXUP_SCALE
   wqqq = XtVaCreateManagedWidget(
           "AFNI" , xmFrameWidgetClass , wshdfunc_thr_rowcol ,
            XmNshadowType , XmSHADOW_ETCHED_IN ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
          NULL ) ;
#else
   wqqq = wshdfunc_thr_rowcol ;
#endif

#if 1
   MCW_widget_geom( anat_frame , &sel_height , NULL,NULL,NULL ) ;
   sel_height -= (74 + 24*MAX_CUTOUTS) ;  /* shorter allows for widgets below */
#else
   sel_height = 290 ;                     /* a hardwired approach */
#endif

   wshdfunc_thr_scale =
      XtVaCreateManagedWidget(
         "scale" , xmScaleWidgetClass , wqqq ,
            XmNminimum , 0 ,
            XmNmaximum , stop ,
            XmNscaleMultiple , sstep ,
            XmNdecimalPoints , decim ,
            XmNshowValue , True ,
            XmNvalue , (int)(smax*func_threshold) ,
            XmNorientation , XmVERTICAL ,
            XmNheight , sel_height ,
            XmNborderWidth , 0 ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;
  }

#ifdef FIX_SCALE_SIZE_PROBLEM
   XtVaSetValues( wshdfunc_thr_scale , XmNuserData , (XtPointer) sel_height , NULL ) ;
#endif

   XtAddCallback( wshdfunc_thr_scale , XmNvalueChangedCallback ,
                  DIFF_thr_scale_CB , NULL ) ;

   XtAddCallback( wshdfunc_thr_scale , XmNdragCallback ,
                  DIFF_thr_scale_drag_CB , NULL ) ;

   /** label for computed p-value, under scale **/

   xstr = XmStringCreateLtoR( THR_PVAL_LABEL_NONE , XmFONTLIST_DEFAULT_TAG ) ;
   wshdfunc_thr_pval_label =
      XtVaCreateManagedWidget(
         "AFNI" , xmLabelWidgetClass , wshdfunc_thr_rowcol ,
            XmNlabelString , xstr ,
            XmNrecomputeSize , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;
   XmStringFree(xstr) ;

   /** optmenu to choose top value for scale **/

   wshdfunc_thr_top_av = new_MCW_arrowval(wshdfunc_thr_rowcol ,
					  "**" ,
					  MCW_AV_optmenu ,
					  0,THR_TOP_EXPON,0 ,
					  MCW_AV_notext , 0 ,
					  reinterpret_cast<void (*)()>
					  (DIFF_thresh_top_CB),
					  NULL ,
					  reinterpret_cast<char* (*)()>
					  (DIFF_thresh_tlabel_CB),
					  NULL ) ;
   XtManageChild(wshdfunc_thr_rowcol) ;

   /*--------------- column 2: color chooser stuff ------------------------------*/

   wshdfunc_color_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , wshdfunc_rowcol ,
            XmNorientation , XmVERTICAL ,
            XmNmarginHeight, 0 ,
            XmNmarginWidth , 0 ,
            XmNpacking , XmPACK_TIGHT ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   xstr = XmStringCreateLtoR( "Color" , XmFONTLIST_DEFAULT_TAG ) ;
   wshdfunc_color_label =
      XtVaCreateManagedWidget(
         "AFNI" , xmLabelWidgetClass , wshdfunc_color_rowcol ,
            XmNlabelString , xstr ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;
   XmStringFree(xstr) ;

   /**-- Popup menu to control some facets of the pbar --**/

   wshdfunc_pbar_menu = XmCreatePopupMenu( wshdfunc_color_label , "menu" , NULL , 0 ) ;

   XtInsertEventHandler( wshdfunc_color_label ,     /* handle events in label */

                               0
                             | ButtonPressMask   /* button presses */
                            ,
                            FALSE ,              /* nonmaskable events? */
                            DIFF_pbarmenu_EV ,   /* handler */
                            NULL ,               /* client data */
                            XtListTail           /* last in queue */
                          ) ;

   (void) XtVaCreateManagedWidget(
            "dialog" , xmLabelWidgetClass , wshdfunc_pbar_menu ,
               LABEL_ARG("-- Cancel --") ,
               XmNrecomputeSize , False ,
               XmNinitialResourcesPersistent , False ,
            NULL ) ;

   (void) XtVaCreateManagedWidget(
            "dialog" , xmSeparatorWidgetClass , wshdfunc_pbar_menu ,
             XmNseparatorType , XmSINGLE_LINE , NULL ) ;

   wshdfunc_pbar_equalize_pb =
      XtVaCreateManagedWidget(
         "dialog" , xmPushButtonWidgetClass , wshdfunc_pbar_menu ,
            LABEL_ARG("Equalize Spacing") ,
            XmNmarginHeight , 0 ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   XtAddCallback( wshdfunc_pbar_equalize_pb , XmNactivateCallback ,
                  DIFF_pbarmenu_CB , im3d ) ;

   wshdfunc_pbar_settop_pb =
      XtVaCreateManagedWidget(
         "dialog" , xmPushButtonWidgetClass , wshdfunc_pbar_menu ,
            LABEL_ARG("Set Top Value") ,
            XmNmarginHeight , 0 ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   XtAddCallback( wshdfunc_pbar_settop_pb , XmNactivateCallback ,
                  DIFF_pbarmenu_CB , im3d ) ;

   (void) XtVaCreateManagedWidget(
            "dialog" , xmSeparatorWidgetClass , wshdfunc_pbar_menu ,
             XmNseparatorType , XmSINGLE_LINE , NULL ) ;

 { static char * pb_dum_label[2] = { "Dummy" , "Dummy" } ;
   wshdfunc_pbar_palette_av = new_MCW_arrowval(
                             wshdfunc_pbar_menu ,     /* parent Widget */
                             "Set Pal " ,          /* label */
                             MCW_AV_optmenu ,      /* option menu style */
                             0 ,                   /* first option */
                             1 ,                   /* last option */
                             0 ,                   /* initial selection */
                             MCW_AV_readtext ,     /* ignored but needed */
                             0 ,                   /* ditto */
			     reinterpret_cast<void (*)()>
                             (DIFF_palette_av_CB), /* callback when changed */
                             NULL ,                /* data for above */
			     reinterpret_cast<char* (*)()>
                             (MCW_av_substring_CB),/* text creation routine */
                             pb_dum_label          /* data for above */
                           ) ;
   }

   (void) XtVaCreateManagedWidget(
            "dialog" , xmSeparatorWidgetClass , wshdfunc_pbar_menu ,
             XmNseparatorType , XmSINGLE_LINE , NULL ) ;

 { static char * pb_dum_label[3] = { "Normal" , "NoShade" , "NoMix" } ;  /* 21 Dec 1999 */
   wshdfunc_pbar_mixshade_av = new_MCW_arrowval(
                             wshdfunc_pbar_menu ,     /* parent Widget */
                             "Mixing  " ,          /* label */
                             MCW_AV_optmenu ,      /* option menu style */
                             0 ,                   /* first option */
                             2 ,                   /* last option */
                             0 ,                   /* initial selection */
                             MCW_AV_readtext ,     /* ignored but needed */
                             0 ,                   /* ditto */
			     reinterpret_cast<void (*)()>
                             (DIFF_mixshade_av_CB),/* callback when changed */
                             NULL ,                /* data for above */
			     reinterpret_cast<char* (*)()>
                             (MCW_av_substring_CB),/* text creation routine */
                             pb_dum_label          /* data for above */
                           ) ;
   }

   if( GPT != NULL && PALTAB_NUM(GPT) > 0 ){
      refit_MCW_optmenu(wshdfunc_pbar_palette_av ,
			0 ,                     /* new minval */
			PALTAB_NUM(GPT)-1 ,     /* new maxval */
			0 ,                     /* new inival */
			0 ,                     /* new decim? */
			reinterpret_cast<char* (*)()>
			(AFNI_palette_label_CB),/* text routine */
			NULL                    /* text data */
                        ) ;
   } else {
      XtUnmanageChild( wshdfunc_pbar_palette_av->wrowcol ) ;
   }

   /**-- Color pbar to control intensity-to-color mapping --**/

 { float pmin=-1.0 , pmax=1.0 ;
   int npane = INIT_panes_sgn ;  /* from afni.h */

   sel_height -= 22 ;            /* a little shorter than the scale */

   wshdfunc_color_pbar = new_MCW_pbar(
                        wshdfunc_color_rowcol ,        /* parent */
                        dc ,                        /* display */
                        npane ,                     /* number panes */
                        sel_height / npane ,        /* init pane height */
                        pmin , pmax ,               /* value range */
			reinterpret_cast<void (*)()>
                        (DIFF_color_pbar_CB),       /* callback */
                        NULL                ) ;     /* callback data */

   wshdfunc_color_pbar->parent       = NULL ;
   wshdfunc_color_pbar->mode         = 0 ;
   wshdfunc_color_pbar->npan_save[0] = INIT_panes_sgn ;  /* from afni.h */
   wshdfunc_color_pbar->npan_save[1] = INIT_panes_pos ;
   wshdfunc_color_pbar->hide_changes = INIT_panes_hide ;

   DIFF_setup_color_pbar() ;  /* other setup stuff */

   (void) XtVaCreateManagedWidget(
            "AFNI" , xmSeparatorWidgetClass , wshdfunc_color_rowcol ,
                XmNseparatorType , XmSINGLE_LINE ,
            NULL ) ;

   wshdfunc_colornum_av = new_MCW_arrowval(
                       wshdfunc_color_rowcol ,
		       "#" ,
		       MCW_AV_optmenu ,
		       NPANE_MIN , NPANE_MAX , npane ,
		       MCW_AV_notext , 0 ,
		       reinterpret_cast<void (*)()>
		       (DIFF_colornum_av_CB),
		       NULL ,
		       NULL,NULL ) ;

   if( NPANE_MAX >= COLSIZE )
      AVOPT_columnize( wshdfunc_colornum_av , 1+(NPANE_MAX+1)/COLSIZE ) ;
  }

   /*--- toggle button to control posfunc option for pbar ---*/

 { char * color_bbox_label[1] = { "Pos?" } ;
   wshdfunc_color_bbox = new_MCW_bbox( wshdfunc_color_rowcol ,
                                      1 , color_bbox_label ,
                                      MCW_BB_check ,
                                      MCW_BB_noframe ,
                                      DIFF_color_bbox_CB , NULL ) ;
 }

   XtManageChild(wshdfunc_color_rowcol) ;

   /*---------- Column 3:  Choices controls -------------------------------*/

   wshdfunc_choices_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , wshdfunc_rowcol ,
            XmNorientation , XmVERTICAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNmarginHeight, 0 ,
            XmNmarginWidth , 0 ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   xstr = XmStringCreateLtoR( "Choices" , XmFONTLIST_DEFAULT_TAG ) ;
   wshdfunc_choices_label =
      XtVaCreateManagedWidget(
         "AFNI" , xmLabelWidgetClass , wshdfunc_choices_rowcol ,
            XmNlabelString , xstr ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;
   XmStringFree(xstr) ;

   /*--- 30 Nov 1997: sub-brick menus ---*/

   wshdfunc_buck_frame =
      XtVaCreateWidget(
         "AFNI" , xmFrameWidgetClass , wshdfunc_choices_rowcol ,
            XmNshadowType , XmSHADOW_ETCHED_IN ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   wshdfunc_buck_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , wshdfunc_buck_frame ,
            XmNorientation , XmVERTICAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   /*--- Sub-brick selectors for Color & Threshold ---*/
   /*    (Actual labels are set when used)            */

   wshdfunc_color_av = new_MCW_arrowval(
                          wshdfunc_buck_rowcol    ,  /* parent Widget */
                          "Color" ,               /* label */
                          MCW_AV_optmenu ,        /* option menu style */
                          0 ,                     /* first option */
                          1 ,                     /* last option */
                          0 ,                     /* initial selection */
                          MCW_AV_readtext ,       /* ignored but needed */
                          0 ,                     /* decimal shift */
			  reinterpret_cast<void (*)()>
                          (DIFF_choose_av_CB),    /* callback when changed */
                          NULL ,                  /* data for above */
			  reinterpret_cast<char* (*)()>
                          (MCW_av_substring_CB),  /* text creation routine */
                          DIFF_dummy_av_label     /* data for above */
                        ) ;

   wshdfunc_thresh_av = new_MCW_arrowval(
                          wshdfunc_buck_rowcol    ,  /* parent Widget */
                          "Thr  " ,               /* label */
                          MCW_AV_optmenu ,        /* option menu style */
                          0 ,                     /* first option */
                          1 ,                     /* last option */
                          0 ,                     /* initial selection */
                          MCW_AV_readtext ,       /* ignored but needed */
                          0 ,                     /* decimal shift */
			  reinterpret_cast<void (*)()>
                          (DIFF_choose_av_CB),    /* callback when changed */
                          NULL ,                  /* data for above */
			  reinterpret_cast<char* (*)()>
                          (MCW_av_substring_CB),  /* text creation routine */
                          DIFF_dummy_av_label     /* data for above */
                        ) ;

   XtManageChild( wshdfunc_buck_rowcol ) ;
   XtManageChild( wshdfunc_buck_frame ) ;

   AV_SENSITIZE( wshdfunc_color_av  , False ) ;  /* turn these off for now */
   AV_SENSITIZE( wshdfunc_thresh_av , False ) ;

   /*--- DTI controls - dti scalars ---*/

   wdtifunc_scalars_frame =
      XtVaCreateWidget(
         "AFNI" , xmFrameWidgetClass , wshdfunc_choices_rowcol ,
            XmNshadowType , XmSHADOW_ETCHED_IN ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   wdtifunc_scalars_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , wdtifunc_scalars_frame ,
            XmNorientation , XmVERTICAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   wdtifunc_scalars_label =
     XtVaCreateManagedWidget(
           "dialog" , xmLabelWidgetClass ,  wdtifunc_scalars_rowcol,
            LABEL_ARG("DTI Scalar Controls") ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   /*--- toggle switches to control if we see function ---*/

   { char * see_overlay_label[1]   = { "See Overlay" } ;

   wdtifunc_see_overlay_bbox = new_MCW_bbox( wdtifunc_scalars_rowcol ,
					     1 , see_overlay_label ,
					     MCW_BB_check ,
					     MCW_BB_noframe ,
					     DIFF_see_dti_overlay_CB , im3d ) ;

   wdtifunc_scalar_av = new_MCW_arrowval(
			  wdtifunc_scalars_rowcol , /* parent Widget */
                          "Scalar" ,               /* label */
                          MCW_AV_optmenu ,         /* option menu style */
                          0 ,                      /* first option */
                          N_DTI_SCALAR_METH-1 ,         /* last option */
                          0 ,                      /* initial selection */
                          MCW_AV_readtext ,        /* ignored but needed */
                          0 ,                      /* decimal shift */
			  reinterpret_cast<void (*)()>
                          (DIFF_scalar_av_CB),     /* callback when changed */
                          NULL ,                   /* data for above */
			  reinterpret_cast<char* (*)()>
                          (MCW_av_substring_CB),   /* text creation routine */
                          DIFF_dti_scalar_av_label     /* data for above */
			  ) ;

     MCW_register_help( wdtifunc_scalar_av->wlabel ,
			"Compute scalar quantity from DTI data" ) ;

     MCW_set_bbox( wdtifunc_range_bbox , pos_mcalc_flag ) ;
   
     /* Buttons for dti control */

     dti_controls_rowcol =
       XtVaCreateWidget(
			"AFNI" , xmRowColumnWidgetClass , wdtifunc_scalars_rowcol ,
			XmNorientation , XmVERTICAL ,
			XmNpacking , XmPACK_TIGHT ,
			XmNtraversalOn , False ,
			XmNinitialResourcesPersistent , False ,
			NULL ) ;

     dti_controls_row1 =
       XtVaCreateWidget(
			"AFNI" , xmRowColumnWidgetClass , dti_controls_rowcol,
			XmNorientation , XmHORIZONTAL ,
			XmNpacking , XmPACK_TIGHT ,
			XmNtraversalOn , False ,
			XmNinitialResourcesPersistent , False ,
			NULL ) ;

     dti_controls_row2 =
       XtVaCreateWidget(
			"AFNI" , xmRowColumnWidgetClass , dti_controls_rowcol,
			XmNorientation , XmHORIZONTAL ,
			XmNpacking , XmPACK_TIGHT ,
			XmNtraversalOn , False ,
			XmNinitialResourcesPersistent , False ,
			NULL ) ;


     /* Compute */

     xstr = XmStringCreateLtoR( "Compute" , XmFONTLIST_DEFAULT_TAG ) ;
     dti_compute_scalars_pb = XtVaCreateManagedWidget(
					      "AFNI" , xmPushButtonWidgetClass , 
					      dti_controls_row1 ,
					      XmNlabelString , xstr ,
					      XmNtraversalOn , False ,
					      XmNinitialResourcesPersistent,False ,
					      NULL ) ;
     XmStringFree(xstr) ;
     XtAddCallback( dti_compute_scalars_pb, XmNactivateCallback, 
		    DIFF_dti_compute_scalars_CB, im3d ) ; 

     MCW_register_help( dti_compute_scalars_pb,
			"Use this to compute the\n"
			"scalar DTI components" ) ;

     /*** pushbutton to cancel scalar computation [not managed now] ***/

     xstr = XmStringCreateLtoR( " * CANCEL * " , XmFONTLIST_DEFAULT_TAG ) ;
     dti_cancel_scalars_pb = XtVaCreateWidget(
                     "AFNI" , xmPushButtonWidgetClass , dti_controls_row1,
                        XmNlabelString , xstr ,
                        XmNtraversalOn , False ,
                        XmNinitialResourcesPersistent , False ,
                     NULL ) ;
     XmStringFree(xstr) ;
     XtAddCallback( dti_cancel_scalars_pb, 
		    XmNactivateCallback, 
		    DIFF_dti_cancel_scalars_CB, NULL ) ;

     /* Display */

     xstr = XmStringCreateLtoR( "Display" , XmFONTLIST_DEFAULT_TAG ) ;
     dti_display_scalars_pb = XtVaCreateManagedWidget(
					      "AFNI" , xmPushButtonWidgetClass , 
					      dti_controls_row2 ,
					      XmNlabelString , xstr ,
					      XmNtraversalOn , False ,
					      XmNinitialResourcesPersistent,False ,
					      NULL ) ;
     XmStringFree(xstr) ;
     XtAddCallback( dti_display_scalars_pb, XmNactivateCallback, 
		    DIFF_dti_display_scalars_CB, im3d ) ; 

     MCW_register_help( dti_display_scalars_pb,
			"Use this to display the\n"
			"scalar DTI components" ) ;
     /* Save */

     xstr = XmStringCreateLtoR( "Save" , XmFONTLIST_DEFAULT_TAG ) ;
     dti_save_scalars_pb = XtVaCreateManagedWidget(
					      "AFNI" , xmPushButtonWidgetClass , 
					      dti_controls_row2 ,
					      XmNlabelString , xstr ,
					      XmNtraversalOn , False ,
					      XmNinitialResourcesPersistent,False ,
					      NULL ) ;
     XmStringFree(xstr) ;
     XtAddCallback( dti_save_scalars_pb, XmNactivateCallback, 
		    DIFF_dti_save_scalars_CB, im3d ) ; 

     MCW_register_help( dti_save_scalars_pb,
			"Use this to save the\n"
			"scalar DTI components" ) ;

     /* Manage */

     XtManageChild( dti_controls_row1 ) ;
     XtManageChild( dti_controls_row2 ) ;
     XtManageChild( dti_controls_rowcol ) ;

   }

   /*--- end DTI controls ---*/

   /*--- SHD controls - shd scalars ---*/

   wshdfunc_scalars_frame =
      XtVaCreateWidget(
         "AFNI" , xmFrameWidgetClass , wshdfunc_choices_rowcol ,
            XmNshadowType , XmSHADOW_ETCHED_IN ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   wshdfunc_scalars_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , wshdfunc_scalars_frame ,
            XmNorientation , XmVERTICAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   wshdfunc_scalars_label =
     XtVaCreateManagedWidget(
           "dialog" , xmLabelWidgetClass ,  wshdfunc_scalars_rowcol,
            LABEL_ARG("SHD Scalar Controls") ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   /*--- toggle switches to control if we see function ---*/

   { char * see_overlay_label[1]   = { "See Overlay" } ;

   wshdfunc_scalars_see_overlay_bbox = new_MCW_bbox( wshdfunc_scalars_rowcol ,
					     1 , see_overlay_label ,
					     MCW_BB_check ,
					     MCW_BB_noframe ,
					     DIFF_see_shd_overlay_CB , im3d ) ;

   wshdfunc_scalar_av = new_MCW_arrowval(
			  wshdfunc_scalars_rowcol ,/* parent Widget */
                          "Scalar" ,               /* label */
                          MCW_AV_optmenu ,         /* option menu style */
                          0 ,                      /* first option */
                          N_SHD_SCALAR_METH-1 ,	   /* last option */
                          0 ,                      /* initial selection */
                          MCW_AV_readtext ,        /* ignored but needed */
                          0 ,                      /* decimal shift */
			  reinterpret_cast<void (*)()>
                          (DIFF_scalar_av_CB),     /* callback when changed */
                          NULL ,                   /* data for above */
			  reinterpret_cast<char* (*)()>
                          (MCW_av_substring_CB),   /* text creation routine */
                          DIFF_shd_scalar_av_label /* data for above */
			  ) ;

     MCW_register_help( wshdfunc_scalar_av->wlabel ,
			"Compute scalar quantity from SHD data" ) ;

     MCW_set_bbox( wshdfunc_range_bbox , pos_mcalc_flag ) ;
   
     /* Buttons for shd control */

     shd_controls_rowcol =
       XtVaCreateWidget(
			"AFNI" , xmRowColumnWidgetClass , wshdfunc_scalars_rowcol ,
			XmNorientation , XmVERTICAL ,
			XmNpacking , XmPACK_TIGHT ,
			XmNtraversalOn , False ,
			XmNinitialResourcesPersistent , False ,
			NULL ) ;

     shd_controls_row1 =
       XtVaCreateWidget(
			"AFNI" , xmRowColumnWidgetClass , shd_controls_rowcol,
			XmNorientation , XmHORIZONTAL ,
			XmNpacking , XmPACK_TIGHT ,
			XmNtraversalOn , False ,
			XmNinitialResourcesPersistent , False ,
			NULL ) ;

     shd_controls_row2 =
       XtVaCreateWidget(
			"AFNI" , xmRowColumnWidgetClass , shd_controls_rowcol,
			XmNorientation , XmHORIZONTAL ,
			XmNpacking , XmPACK_TIGHT ,
			XmNtraversalOn , False ,
			XmNinitialResourcesPersistent , False ,
			NULL ) ;


     /* Compute */

     xstr = XmStringCreateLtoR( "Compute" , XmFONTLIST_DEFAULT_TAG ) ;
     shd_compute_scalars_pb = XtVaCreateManagedWidget(
					      "AFNI" , xmPushButtonWidgetClass , 
					      shd_controls_row1 ,
					      XmNlabelString , xstr ,
					      XmNtraversalOn , False ,
					      XmNinitialResourcesPersistent,False ,
					      NULL ) ;
     XmStringFree(xstr) ;
     XtAddCallback( shd_compute_scalars_pb, XmNactivateCallback, 
		    DIFF_shd_compute_scalars_CB, im3d ) ; 

     MCW_register_help( shd_compute_scalars_pb,
			"Use this to compute the\n"
			"scalar SHD components" ) ;

     /*** pushbutton to cancel scalar computation [not managed now] ***/

     xstr = XmStringCreateLtoR( " * CANCEL * " , XmFONTLIST_DEFAULT_TAG ) ;
     shd_cancel_scalars_pb = XtVaCreateWidget(
                     "AFNI" , xmPushButtonWidgetClass , shd_controls_row1,
                        XmNlabelString , xstr ,
                        XmNtraversalOn , False ,
                        XmNinitialResourcesPersistent , False ,
                     NULL ) ;
     XmStringFree(xstr) ;
     XtAddCallback( shd_cancel_scalars_pb, 
		    XmNactivateCallback, 
		    DIFF_shd_cancel_scalars_CB, NULL ) ;

     /* Display */

     xstr = XmStringCreateLtoR( "Display" , XmFONTLIST_DEFAULT_TAG ) ;
     shd_display_scalars_pb = XtVaCreateManagedWidget(
					      "AFNI" , xmPushButtonWidgetClass , 
					      shd_controls_row2 ,
					      XmNlabelString , xstr ,
					      XmNtraversalOn , False ,
					      XmNinitialResourcesPersistent,False ,
					      NULL ) ;
     XmStringFree(xstr) ;
     XtAddCallback( shd_display_scalars_pb, XmNactivateCallback, 
		    DIFF_shd_display_scalars_CB, im3d ) ; 

     MCW_register_help( shd_display_scalars_pb,
			"Use this to display the\n"
			"scalar SHD components" ) ;
     /* Save */

     xstr = XmStringCreateLtoR( "Save" , XmFONTLIST_DEFAULT_TAG ) ;
     shd_save_scalars_pb = XtVaCreateManagedWidget(
					      "AFNI" , xmPushButtonWidgetClass , 
					      shd_controls_row2 ,
					      XmNlabelString , xstr ,
					      XmNtraversalOn , False ,
					      XmNinitialResourcesPersistent,False ,
					      NULL ) ;
     XmStringFree(xstr) ;
     XtAddCallback( shd_save_scalars_pb, XmNactivateCallback, 
		    DIFF_shd_save_scalars_CB, im3d ) ; 

     MCW_register_help( shd_save_scalars_pb,
			"Use this to save the\n"
			"scalar SHD components" ) ;

     /* Manage */

     XtManageChild( shd_controls_row1 ) ;
     XtManageChild( shd_controls_row2 ) ;
     XtManageChild( shd_controls_rowcol ) ;

   }

   /*--- end SHD scalars controls ---*/


   /*--- SHD controls - lmview ---*/

   wshdfunc_lmview_frame =
      XtVaCreateWidget(
         "AFNI" , xmFrameWidgetClass , wshdfunc_choices_rowcol ,
            XmNshadowType , XmSHADOW_ETCHED_IN ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   wshdfunc_lmview_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , wshdfunc_lmview_frame ,
            XmNorientation , XmVERTICAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   wshdfunc_lmview_label =
     XtVaCreateManagedWidget(
           "dialog" , xmLabelWidgetClass ,  wshdfunc_lmview_rowcol,
            LABEL_ARG("LM View Controls") ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   /*--- toggle switches to control if we see function ---*/

   { char * see_overlay_label[1]   = { "See Overlay" } ;
     char * lm_view_label[1] = { "Restrict LM View" } ;
     char * pos_mview_label[1]   = { "View M>=0 only?" } ;

     wshdfunc_lmview_see_overlay_bbox = new_MCW_bbox( wshdfunc_lmview_rowcol ,
                                            1 , see_overlay_label ,
                                            MCW_BB_check ,
                                            MCW_BB_noframe ,
                                            DIFF_see_dti_overlay_CB , NULL ) ;

     wshdfunc_lm_view_bbox = new_MCW_bbox( wshdfunc_lmview_rowcol ,
                                              1 , lm_view_label ,
                                              MCW_BB_check ,
                                              MCW_BB_noframe ,
                                              DIFF_lm_view_CB , NULL ) ;

     wshdfunc_shd_lval_av =
       new_MCW_arrowval(wshdfunc_lmview_rowcol , "Max L" , MCW_AV_downup ,
			0 , 20 , func_shd_lval ,
			MCW_AV_edittext , 0 ,
			reinterpret_cast<void (*)()>
			(DIFF_wshdfunc_shd_lval_av_CB),
			im3d ,
			NULL , NULL
			) ;
     MCW_reghelp_children( wshdfunc_shd_lval_av->wrowcol ,
			   "Use this to set the maximum L value\n"
			   "of the Spherical Harmonic Transform to view"
			   ) ;

     wshdfunc_shd_mval_av =
       new_MCW_arrowval(wshdfunc_lmview_rowcol , "Max M" , MCW_AV_downup ,
			0 , 20 , func_shd_mval,
			MCW_AV_edittext , 0 ,
			reinterpret_cast<void (*)()>
			(DIFF_wshdfunc_shd_mval_av_CB),
			im3d ,
			NULL,NULL
			) ;
     MCW_reghelp_children(wshdfunc_shd_mval_av->wrowcol ,
			  "Use this to set the maximum M value\n"
			  "of the Spherical Harmonic Transform to view"
			  ) ;
   
     AV_SENSITIZE( wshdfunc_shd_lval_av , False ) ;
     AV_SENSITIZE( wshdfunc_shd_mval_av , False ) ;

     wshdfunc_pos_mview_bbox = new_MCW_bbox( wshdfunc_lmview_rowcol ,
                                            1 , pos_mview_label ,
                                            MCW_BB_check ,
                                            MCW_BB_noframe ,
                                            DIFF_pos_mview_CB , NULL ) ;

     MCW_set_bbox( wshdfunc_range_bbox , pos_mcalc_flag ) ;
     /* BB_SENSITIZE( wshdfunc_pos_mview_bbox, pos_mcalc_flag ) ; */

     shd_display_rowcol =
       XtVaCreateWidget(
			"AFNI" , xmRowColumnWidgetClass , wshdfunc_lmview_rowcol ,
			XmNorientation , XmHORIZONTAL ,
			XmNpacking , XmPACK_TIGHT ,
			XmNtraversalOn , False ,
			XmNinitialResourcesPersistent , False ,
			NULL ) ;

     /* Display */

     xstr = XmStringCreateLtoR( "Display" , XmFONTLIST_DEFAULT_TAG ) ;
     shd_display_pb =
       XtVaCreateManagedWidget( "AFNI" , xmPushButtonWidgetClass , 
				shd_display_rowcol ,
				XmNlabelString , xstr ,
				XmNtraversalOn , False ,
				XmNinitialResourcesPersistent,False ,
				NULL ) ;
     XmStringFree(xstr) ;
     XtAddCallback( shd_display_pb, XmNactivateCallback, 
		    DIFF_shd_display_CB, im3d ) ; 

     MCW_register_help( shd_display_pb,
			"Use this to display\n"
			"the shd components" ) ;

     XtManageChild( shd_display_rowcol ) ;

   }

   /* end SHD lmview controls */

   /*--- FTM controls - lmview ---*/

   wftmfunc_lmview_frame =
      XtVaCreateWidget(
         "AFNI" , xmFrameWidgetClass , wshdfunc_choices_rowcol ,
            XmNshadowType , XmSHADOW_ETCHED_IN ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   wftmfunc_lmview_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , wftmfunc_lmview_frame ,
            XmNorientation , XmVERTICAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   wftmfunc_lmview_label =
     XtVaCreateManagedWidget(
           "dialog" , xmLabelWidgetClass ,  wftmfunc_lmview_rowcol,
            LABEL_ARG("FTM View Controls") ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;


#ifdef CORNY_LOGO   

   /* Corny Logo */
   if( corny_pixmap == XmUNSPECIFIED_PIXMAP )
     corny_pixmap = 
       XCreatePixmapFromBitmapData(XtDisplay(shell) ,
				   RootWindowOfScreen(XtScreen(shell)) ,
				   corny_bits ,
				   corny_width , corny_height ,
				   ucsd_fg_pix , ucsd_bg_pix ,
				   DefaultDepthOfScreen(XtScreen(shell)) ) ;

   corny_picture =
     XtVaCreateManagedWidget("dialog" , xmLabelWidgetClass , wftmfunc_lmview_rowcol,
			     XmNlabelType        , XmPIXMAP ,
			     XmNalignment        , XmALIGNMENT_CENTER ,
			     XmNlabelInsensitivePixmap , corny_pixmap ,
			     XmNwidth        , corny_width ,
			     XmNheight       , corny_height ,
			     XmNmarginWidth  , 0 ,
			     XmNmarginHeight , 0 ,
			     XmNrecomputeSize, False ,
			     XmNtraversalOn  , False ,
			     XmNinitialResourcesPersistent , False ,
			     NULL ) ;

   XtVaSetValues(corny_picture,XmNlabelPixmap,corny_pixmap,NULL);

#endif

   /* end FTM controls */

   /* Manage overlay */

   XtManageChild( wshdfunc_lmview_rowcol ) ;
   XtManageChild( wshdfunc_scalars_rowcol ) ;
   XtManageChild( wdtifunc_scalars_rowcol ) ;
   XtManageChild( wftmfunc_lmview_rowcol ) ;

   /*--- range controls ---*/

   wshdfunc_range_frame =
      XtVaCreateManagedWidget(
         "AFNI" , xmFrameWidgetClass , wshdfunc_choices_rowcol ,
            XmNshadowType , XmSHADOW_ETCHED_IN ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   wshdfunc_range_rowcol =
      XtVaCreateWidget(
         "AFNI" , xmRowColumnWidgetClass , wshdfunc_range_frame ,
            XmNorientation , XmVERTICAL ,
            XmNpacking , XmPACK_TIGHT ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;

   /*--- label to show the data ranges from the overlay dataset ---*/

   xstr = DIFF_range_label() ;  /* make a dummy label */
   wshdfunc_range_label =
      XtVaCreateManagedWidget(
         "AFNI" , xmLabelWidgetClass , wshdfunc_range_rowcol ,
            XmNrecomputeSize , False ,
            XmNlabelString , xstr ,
            XmNtraversalOn , False ,
            XmNinitialResourcesPersistent , False ,
         NULL ) ;
   XmStringFree(xstr) ;

   /*--- toggle button to control automatic range scaling for pbar ---*/

 { char * range_bbox_label[1] = { "autoRange:xxxxxxxxx" } ;

   wshdfunc_range_bbox =
      new_MCW_bbox( wshdfunc_range_rowcol ,
                    1 , range_bbox_label ,
                    MCW_BB_check ,
                    MCW_BB_noframe ,
                    DIFF_range_bbox_CB , NULL ) ;

   MCW_set_bbox( wshdfunc_range_bbox , 1 ) ;

   xstr = DIFF_autorange_label() ;
   XtVaSetValues( wshdfunc_range_bbox->wbut[0], XmNlabelString,xstr , NULL ) ;
   XmStringFree(xstr) ;
 }

   /*--- arrowval to provide user control for pbar scaling ---*/

   wshdfunc_range_av =
     new_MCW_arrowval(wshdfunc_range_rowcol ,  /* parent */
		      NULL ,                /* label */
		      MCW_AV_downup ,       /* arrow directions */
		      0  ,                  /* min value */
		      9999999 ,             /* max value */
		      (int)(func_range) ,   /* init value */
		      MCW_AV_editext ,      /* input/output text display */
		      0 ,                   /* decimal shift */
		      reinterpret_cast<void (*)()>
		      (DIFF_range_av_CB),   /* routine to call when button */
		      NULL ,                /* is pressed, and its data */
		      NULL,NULL             /* no special display */
		      ) ;
   AV_SENSITIZE( wshdfunc_range_av , False ) ;

   XtManageChild( wshdfunc_range_rowcol ) ;
   XtManageChild( wshdfunc_range_frame ) ;

   XtManageChild( wshdfunc_choices_rowcol ) ;

   XtManageChild( wshdfunc_rowcol ) ;
   XtManageChild( wshdfunc_uber_rowcol ) ;

#if 0
   XtVaSetValues( wshdfunc_uber_rowcol , XmNresizeWidth , False , NULL ) ;
#endif

   return ;
}

/*---------------------------------------------------------------------------------
   Initialize the functional colormap
-----------------------------------------------------------------------------------*/

void DIFF_init_cmap(void)
{
   int ii , nc ;

   for( ii=0 ; ii < 127 ; ii++ )                             /* gray part */
      func_rmap[ii] = func_gmap[ii] = func_bmap[ii] = 2*ii ;

   func_rmap[127] = func_gmap[127] = func_bmap[127] = 255 ;  /* white pixel */

   nc = MIN( dc->ovc->ncol_ov , 129 ) ;   /* don't allow colormap overflow */

   for( ii=1 ; ii < nc ; ii++ ){                             /* color part */
      func_rmap[127+ii] = DCOV_REDBYTE(dc,ii) ;              /* [skips #0] */
      func_gmap[127+ii] = DCOV_GREENBYTE(dc,ii) ;
      func_bmap[127+ii] = DCOV_BLUEBYTE(dc,ii) ;
   }

   func_ncmap = 127 + nc ;                                   /* size of map */

   if( diffuse_handle != NULL ){
     /* MDIF_set_rgbmap( diffuse_handle, func_ncmap, func_rmap,func_gmap,func_bmap ) ; */
       func_cmap_set = 1 ;
   } else {
       func_cmap_set = 0 ;
   }

   return ;
}

/*---------------------------------------------------------------------------
   Makes the labels for the thresh_top menu (narrower than the default)
-----------------------------------------------------------------------------*/

char * DIFF_thresh_tlabel_CB( MCW_arrowval * av , XtPointer junk )
{
   static char tlab[8] ;
   sprintf(tlab,"%d",av->ival) ;
   return tlab ;
}

/*---------------------------------------------------------------------------
   Initialize the memory of the color selector
-----------------------------------------------------------------------------*/

void DIFF_setup_color_pbar(void)
{
  MCW_pbar * pbar = wshdfunc_color_pbar ;
  int np , i , jm , lcol ;

  jm   = pbar->mode ;
  lcol = dc->ovc->ncol_ov - 1 ;

  /** load the 'save' values for all possible pane counts **/

  for( np=NPANE_MIN ; np <= NPANE_MAX ; np++ ){

      for( i=0 ; i <= np ; i++ ){
         pbar->pval_save[np][i][0] = INIT_pval_sgn[np][i] ;  /* from afni.h */
         pbar->pval_save[np][i][1] = INIT_pval_pos[np][i] ;
      }

      for( i=0 ; i <  np ; i++ ){
         pbar->ovin_save[np][i][0] = MIN( lcol , INIT_ovin_sgn[np][i] ) ;
         pbar->ovin_save[np][i][1] = MIN( lcol , INIT_ovin_pos[np][i] ) ;
      }
  }

  /** load the values for the current pane count **/

  np = pbar->num_panes ;
  jm = pbar->mode ;

  for( i=0 ; i <= np ; i++ ) pbar->pval[i]     = pbar->pval_save[np][i][jm] ;
  for( i=0 ; i <  np ; i++ ) pbar->ov_index[i] = pbar->ovin_save[np][i][jm] ;

  pbar->update_me = 1 ;
  return ;
}

/*-------------------------------------------------------------------------
  Create label for range display in function control panel
---------------------------------------------------------------------------*/

XmString DIFF_range_label(void)
{
   char fim_minch[10]  = " --------" , fim_maxch[10]  = " --------" ,
        thr_minch[10]  = " --------" , thr_maxch[10]  = " --------"   ;
   char buf[256] , qbuf[16] ;
   XmString xstr ;
   int iv ;

   if( ISVALID_DSET(func_dset) && ISVALID_STATISTIC(func_dset->stats) ){

      iv = func_color_ival ;

      if( DSET_VALID_BSTAT(func_dset,iv) ){
         AV_fval_to_char( func_dset->stats->bstat[iv].min , qbuf ) ;
         sprintf( fim_minch , "%9.9s" , qbuf ) ;
         AV_fval_to_char( func_dset->stats->bstat[iv].max , qbuf ) ;
         sprintf( fim_maxch , "%9.9s" , qbuf ) ;
      }

      iv = func_thresh_ival ;

      if( DSET_VALID_BSTAT(func_dset,iv) ){
         AV_fval_to_char( func_dset->stats->bstat[iv].min , qbuf ) ;
         sprintf( thr_minch , "%9.9s" , qbuf ) ;
         AV_fval_to_char( func_dset->stats->bstat[iv].max , qbuf ) ;
         sprintf( thr_maxch , "%9.9s" , qbuf ) ;
      }
   }

   sprintf( buf , "Color %s:%s\nThr   %s:%s" ,
            fim_minch,fim_maxch, thr_minch,thr_maxch ) ;

   xstr = XmStringCreateLtoR( buf , XmFONTLIST_DEFAULT_TAG ) ;

   return xstr ;
}

/*------------------------------------------------------------------------
   Find the autorange and make a label for the autorange control widget
--------------------------------------------------------------------------*/

XmString DIFF_autorange_label(void)
{
   XmString xstr ;
   float rrr = DEFAULT_FUNC_RANGE ;
   char buf[32] , qbuf[16] ;

   if( ISVALID_DSET(func_dset) ){

      RELOAD_STATS(func_dset) ;
      if( ISVALID_STATISTIC(func_dset->stats) ){
         float s1 , s2 ; int iv ;

         iv = func_color_ival ;

         if( DSET_VALID_BSTAT(func_dset,iv) ){
            s1  = fabs(func_dset->stats->bstat[iv].min) ,
            s2  = fabs(func_dset->stats->bstat[iv].max) ;
            rrr = (s1<s2) ? s2 : s1 ;
            if( rrr == 0.0 ) rrr = 1.0 ;
         }
      }
   }

   func_autorange = rrr ;
   AV_fval_to_char( rrr , qbuf ) ;
   sprintf( buf , "autoRange:%s" , qbuf ) ;
   xstr = XmStringCreateLtoR( buf , XmFONTLIST_DEFAULT_TAG ) ;

   return xstr ;
}

/*-----------------------------------------------------------------------------
   Set the p-value label at the bottom of the threshold slider
-------------------------------------------------------------------------------*/

void DIFF_set_thr_pval(void)
{
   float thresh , pval ;
   int   dec ;
   char  buf[16] ;

   if( !ISVALID_DSET(func_dset) ) return ;

   /* get the "true" threshold (scaled up from being in [0,1]) */

   thresh = func_threshold * func_thresh_top ;

   /* get the p-value that goes with this threshold, for this functional dataset */

#if 0
   if( ISFUNCBUCKET(func_dset) )
      pval = THD_stat_to_pval( thresh ,
                               DSET_BRICK_STATCODE(func_dset,func_thresh_ival) ,
                               DSET_BRICK_STATAUX (func_dset,func_thresh_ival)  ) ;

   else if( func_thresh_ival == DSET_THRESH_VALUE(func_dset) )
      pval = THD_stat_to_pval( thresh , func_dset->func_type ,
                                        func_dset->stat_aux   ) ;

   else
      pval = -1.0 ;
#else
   pval = THD_stat_to_pval( thresh ,
                            DSET_BRICK_STATCODE(func_dset,func_thresh_ival) ,
                            DSET_BRICK_STATAUX (func_dset,func_thresh_ival)  ) ;
#endif

   if( pval < 0.0 ){
      strcpy( buf , THR_PVAL_LABEL_NONE ) ;
   } else {
      if( pval == 0.0 ){
         strcpy( buf , "p = 0" ) ;
      } else if( pval >= 0.9999 ){
         strcpy( buf , "p = 1" ) ;
      } else if( pval >= 0.0010 ){
         char qbuf[16] ;
         sprintf( qbuf , "%5.4f" , pval ) ;
         strcpy( buf , qbuf+1 ) ;
      } else {
         int dec = (int)(0.999 - log10(pval)) ;
         pval = pval * pow( 10.0 , (double) dec ) ;  /* between 1 and 10 */
         if( dec < 10 ) sprintf( buf , "%3.1f-%1d" ,      pval, dec ) ;
         else           sprintf( buf , "%1d.-%2d"  , (int)pval, dec ) ;
      }
   }
   MCW_set_widget_label( wshdfunc_thr_pval_label , buf ) ;
   return ;
}

/*------------------------------------------------------------------------------
   When the user alters the threshold [at the end]
--------------------------------------------------------------------------------*/

void DIFF_thr_scale_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   XmScaleCallbackStruct * cbs = (XmScaleCallbackStruct *) call_data ;
   float fff ;

   fff = THR_FACTOR * cbs->value ;  /* between 0 and 1 now */
   if( fff >= 0.0 && fff <= 1.0 ) func_threshold = fff ; else return ;
   DIFF_set_thr_pval() ;

   INVALIDATE_OVERLAY ;
   return ;
}

/*---------------------------------------------------------------------------
  When the user drags the slider (just update the p-value)
-----------------------------------------------------------------------------*/

void DIFF_thr_scale_drag_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   XmScaleCallbackStruct * cbs = (XmScaleCallbackStruct *) call_data ;
   float fff ;

   fff = THR_FACTOR * cbs->value ;  /* between 0 and 1 now */
   if( fff >= 0.0 && fff <= 1.0 ) func_threshold = fff ; else return ;
   DIFF_set_thr_pval() ;

   return ;
}


/*---------------------------------------------------------------------------
  Called when the user toggles the autorange button
-----------------------------------------------------------------------------*/

void DIFF_range_bbox_CB( Widget w, XtPointer cd, XtPointer cb)
{
   int newauto = MCW_val_bbox(wshdfunc_range_bbox) ;

   if( newauto == func_use_autorange ) return ;  /* no change? */

   func_use_autorange = newauto ;

   func_range = (newauto) ? (func_autorange)
                          : (wshdfunc_range_av->fval) ;

   AV_SENSITIZE( wshdfunc_range_av , ! newauto ) ;

   INVALIDATE_OVERLAY ;
   return ;
}


/*---------------------------------------------------------------------------
  Called when the user changes the function range
-----------------------------------------------------------------------------*/

void DIFF_range_av_CB( MCW_arrowval * av , XtPointer cd )
{
   func_range = av->fval ;

   INVALIDATE_OVERLAY ;
   return ;
}

/*---------------------------------------------------------------------------
   Called to change the scaling on the threshold slider
-----------------------------------------------------------------------------*/

void DIFF_thresh_top_CB( MCW_arrowval * av , XtPointer cd )
{
   static float dval[9] = { 1.0 , 10.0 , 100.0 , 1000.0 , 10000.0 ,
                            100000.0 , 1000000.0 , 10000000.0 , 100000000.0 } ;
   int decim ;
   float tval ;

   tval = dval[av->ival] ; if( tval <= 0.0 ) tval = 1.0 ;

   decim = (2*THR_TOP_EXPON) - (int)(THR_TOP_EXPON + 0.01 + log10(tval)) ;
   if( decim < 0 ) decim = 0 ;

   XtVaSetValues( wshdfunc_thr_scale, XmNdecimalPoints, decim, NULL ) ;

   func_thresh_top = tval ;
   DIFF_set_thr_pval() ;

   INVALIDATE_OVERLAY ;
   return ;
}

/*---------------------------------------------------------------------------
  Called by the pbar routines when the pbar is altered.
-----------------------------------------------------------------------------*/

void DIFF_color_pbar_CB( MCW_pbar * pbar , XtPointer cd , int reason )
{
   FIX_SCALE_SIZE ;
   INVALIDATE_OVERLAY ;
   return ;
}

/*------------------------------------------------------------------------------
  Called to change the number of panels in the pbar
--------------------------------------------------------------------------------*/

void DIFF_colornum_av_CB( MCW_arrowval * av , XtPointer cd )
{
   HIDE_SCALE ;
   alter_MCW_pbar( wshdfunc_color_pbar , av->ival , NULL ) ;
   FIX_SCALE_SIZE ;
   INVALIDATE_OVERLAY ;
   return ;
}

/*---------------------------------------------------------------------------
   Called when the user toggles the posfunc button
-----------------------------------------------------------------------------*/

void DIFF_color_bbox_CB( Widget w, XtPointer cd, XtPointer cb)
{
   int jm , newpos=MCW_val_bbox(wshdfunc_color_bbox) ;

   if( newpos == func_posfunc ) return ;  /* no change? */

   func_posfunc = newpos ;
   jm = wshdfunc_color_pbar->mode = (newpos) ? 1 : 0 ;  /* pbar mode */

   HIDE_SCALE ;
   alter_MCW_pbar( wshdfunc_color_pbar , wshdfunc_color_pbar->npan_save[jm] , NULL ) ;
   FIX_SCALE_SIZE ;

   /* set the count on the pbar control arrowval to match */

   AV_assign_ival( wshdfunc_colornum_av , wshdfunc_color_pbar->npan_save[jm] ) ;

   INVALIDATE_OVERLAY ;
   return ;
}

/*------------------------------------------------------------------------------
   Called to set the overlay opacity factor
--------------------------------------------------------------------------------*/

void DIFF_color_opacity_CB( MCW_arrowval * av , XtPointer cd )
{
   int ofs = func_showthru ;
   func_color_opacity = 0.1 * av->ival ;
   func_showthru = (av->ival == 11) ;    /* 07 Jan 2000 */
   INVALIDATE_OVERLAY ;

   if( func_showthru != ofs ) FREE_VOLUMES ;
   return ;
}

/*---------------------------------------------------------------------------
   Called to toggle visibility of the overlay
-----------------------------------------------------------------------------*/

/* This is from AFNI_see_func_CB in afni_func.c */

void DIFF_see_dti_overlay_CB( Widget w, XtPointer cd, XtPointer cb)
{
   Three_D_View * im3d = (Three_D_View *) cd ;
   int old_val , new_val ;
   int debug = 0 ;

ENTRY("DIFF_see_dti_overlay_CB") ;

   if( ! IM3D_VALID(im3d) ) {
     if(debug) printf("Inside DIFF_see_dti_overlay_CB: invalid data set\n") ;
     EXRETURN ;
   }

   old_val = (im3d->vinfo->func_visible) ? 1 : 0 ;
   /* new_val = MCW_val_bbox( im3d->vwid->view->see_func_bbox ) ; */
   new_val = MCW_val_bbox( wdtifunc_see_overlay_bbox );

   if(debug) printf("Inside DIFF_see_dti_overlay_CB: old_val = %i, new_val = %i \n",
		    old_val,new_val) ;

   if( old_val != new_val ){
     im3d->vinfo->func_visible = (new_val == 1) ? True : False ;
     if( ! ISVALID_3DIM_DATASET(im3d->fim_now) ){            
       im3d->vinfo->func_visible = False ;
       MCW_set_bbox( im3d->vwid->view->see_func_bbox , 0 ) ; 
     }
     AFNI_set_viewpoint( im3d , -1,-1,-1 , REDISPLAY_OVERLAY ) ;  /* redraw */
   }

   RESET_AFNI_QUIT(im3d) ;
   EXRETURN ;
}

/*
void DIFF_see_dti_overlay_CB( Widget w, XtPointer cd, XtPointer cb)
{
   int newsee = MCW_val_bbox(wshdfunc_lmview_see_overlay_bbox) ;

   if( newsee == func_see_overlay ) return ;

   func_see_overlay = newsee ;
   INVALIDATE_OVERLAY ; FREE_VOLUMES ;
   return ;
} */

/*---------------------------------------------------------------------------
   Called to toggle visibility of the overlay

   I just cut and pasted from the DIFF_see_dti..routing above.
   No idea if this (or even that one) works, does anything, etc.
-----------------------------------------------------------------------------*/

/* This is from AFNI_see_func_CB in afni_func.c */

void DIFF_see_shd_overlay_CB( Widget w, XtPointer cd, XtPointer cb)
{
   Three_D_View * im3d = (Three_D_View *) cd ;
   int old_val , new_val ;
   int debug = 0 ;

ENTRY("DIFF_see_shd_overlay_CB") ;

   if( ! IM3D_VALID(im3d) ) {
     if(debug) printf("Inside DIFF_see_shd_overlay_CB: invalid data set\n") ;
     EXRETURN ;
   }

   old_val = (im3d->vinfo->func_visible) ? 1 : 0 ;
   /* new_val = MCW_val_bbox( im3d->vwid->view->see_func_bbox ) ; */
   new_val = MCW_val_bbox( wshdfunc_scalars_see_overlay_bbox );

   if(debug) printf("Inside DIFF_see_shd_overlay_CB: old_val = %i, new_val = %i \n",
		    old_val,new_val) ;

   if( old_val != new_val ){
     im3d->vinfo->func_visible = (new_val == 1) ? True : False ;
     if( ! ISVALID_3DIM_DATASET(im3d->fim_now) ){            
       im3d->vinfo->func_visible = False ;
       MCW_set_bbox( im3d->vwid->view->see_func_bbox , 0 ) ; 
     }
     AFNI_set_viewpoint( im3d , -1,-1,-1 , REDISPLAY_OVERLAY ) ;  /* redraw */
   }

   RESET_AFNI_QUIT(im3d) ;
   EXRETURN ;
}

/*---------------------------------------------------------------------------
   Called to toggle whether or not cutouts affect the overlay
-----------------------------------------------------------------------------*/

void DIFF_pos_mview_CB( Widget w, XtPointer cd, XtPointer cb)
{
   int newmview = MCW_val_bbox(wshdfunc_pos_mview_bbox) ;

   pos_mview_flag = newmview ;

   /* Can only view negative m if calculated. */

   if(pos_mcalc_flag==1)
     MCW_set_bbox( wshdfunc_pos_mview_bbox , 1 ) ;
   else
     MCW_set_bbox( wshdfunc_pos_mview_bbox , pos_mview_flag ) ;

   func_pos_mview = newmview ;

   return ;
}

/*----------------------------------------------------------------------------
   Called to choose L and M values to view
-----------------------------------------------------------------------------*/

void DIFF_lm_view_CB( Widget w, XtPointer cd, XtPointer cb)
{
   int cc , newlm = MCW_val_bbox(wshdfunc_lm_view_bbox) ;

   if( newlm == func_lm_view ) return ;

   func_lm_view = newlm ;

   AV_SENSITIZE( wshdfunc_shd_lval_av , newlm ) ;
   AV_SENSITIZE( wshdfunc_shd_mval_av , newlm ) ;

   INVALIDATE_OVERLAY ;

   return ;
}

void DIFF_clusters_av_CB( MCW_arrowval * av , XtPointer cd )
{
   int cc ;

   INVALIDATE_OVERLAY ;

   return ;
}

/*------------------------------------------------------------------------------
   Called to set the [anisotropy] ? factor
--------------------------------------------------------------------------------*/

void DIFF_aniso_CB( MCW_arrowval * av , XtPointer cd )
{
   int ofs = func_showthru ;
   func_color_opacity = 0.1 * av->ival ;
   func_showthru = (av->ival == 11) ;
   INVALIDATE_OVERLAY ;

   if( func_showthru != ofs ) FREE_VOLUMES ;
   return ;
}

/*------------------------------------------------------------------------------
   Called to set the [fiber] ? factor
--------------------------------------------------------------------------------*/

void DIFF_fiber_CB( MCW_arrowval * av , XtPointer cd )
{
   int ofs = func_showthru ;
   func_color_opacity = 0.1 * av->ival ;
   func_showthru = (av->ival == 11) ;
   INVALIDATE_OVERLAY ;

   if( func_showthru != ofs ) FREE_VOLUMES ;
   return ;
}

void DIFF_shd_lval_av_CB( MCW_arrowval * av , XtPointer cd )
{
   Three_D_View *im3d = (Three_D_View *) cd ;

   INVALIDATE_OVERLAY ;

   /* If L is changed, assume we want M = L, as it 
      should be in almost all cases.
      (if not, then you have to adjust M yourself! */

   AV_assign_ival(shd_mval_av,av->ival);

   /* If L falls below M, adjust M down */

   if(av->ival < shd_mval_av->ival) 
     AV_assign_ival(shd_mval_av,av->ival);

   return ;
}

void DIFF_shd_mval_av_CB( MCW_arrowval * av , XtPointer cd )
{
   int cc ;

   INVALIDATE_OVERLAY ;

   /* If M goes above L, adjust M down to L */

   if(av->ival > shd_lval_av->ival) 
     AV_assign_ival(av,shd_lval_av->ival);

   return ;
}

/* Transfer Spharm Control parameters to SHD_calc_state */

void get_spharm_control_vals( DIF_state *dif )
{
  char *program_name = "get_spharm_control_vals";
  int debug = 0;

  SHD_calc_state *calc = dif->shd->calc ;
  SHD_view_state *view = dif->shd->view ;

  int doPosM = MCW_val_bbox(pos_mcalc_bbox) ;

  /* Calculation values */

  /* L and M max from bboxes */

  calc->Lmax = MIN(shd_lval_av->ival,CALC_Lmax_MAX) ;
  calc->Mmax = MIN(shd_mval_av->ival,CALC_Mmax_MAX) ;

  if(debug) printf("%s: calc->Lmax = %i, calc->Mmax = %i\n",
		   program_name,calc->Lmax,calc->Mmax);

  /* L and M min. Fixed Lmin, and Mmin from bbox */

  calc->Lmin = 0 ;
  switch( doPosM ){

     case 0:{
       calc->full_mcalc_range = 1;
       calc->Mmin = -calc->Mmax ; 
     }
     break ;

     case 1:{
       calc->full_mcalc_range = 0;
       calc->Mmin = 0 ; 
     }
     break ;
  }

  /* Grid size */

  calc->shd_grid_size = shd_grid_av->ival ;
  calc->npolr = calc->shd_grid_size ;
  calc->nazim = calc->shd_grid_size ;

  if(debug) printf("%s: calc->shd_grid_size = %i, calc->nazim = %i, calc->npolr = %i\n",
		   program_name,calc->shd_grid_size,calc->nazim,calc->npolr);

  /* Viewing values */

  /* L and M max from bboxes */

  view->Lmax = MIN(shd_lval_av->ival,CALC_Lmax_MAX) ;
  view->Mmax = MIN(shd_mval_av->ival,CALC_Mmax_MAX) ;

  if(debug) printf("%s: view->Lmax = %i, view->Mmax = %i\n",
		   program_name,view->Lmax,view->Mmax);

  /* L and M min. Fixed Lmin, and Mmin from bbox */
  view->Lmin = 0 ;
  switch( doPosM ){

     case 0:{
       view->full_mview_range = 1;
       view->Mmin = -view->Mmax ; 
     }
     break ;

     case 1:{
       view->full_mview_range = 0;
       view->Mmin = 0 ; 
     }
     break ;
  }

}

/*---------------------------------------------------------------------------
   Called to toggle whether or not cutouts affect the overlay
-----------------------------------------------------------------------------*/

void DIFF_pos_mcalc_CB( Widget w, XtPointer cd, XtPointer cb)
{
   int newmcalc = MCW_val_bbox(pos_mcalc_bbox) ;

   pos_mcalc_flag = newmcalc ;

   /* Can only view negative m if calculated. */

   if(pos_mcalc_flag==1)
     MCW_set_bbox( wshdfunc_pos_mview_bbox , 1 ) ;
   else
     MCW_set_bbox( wshdfunc_pos_mview_bbox , 0 ) ;

   /*   if(pos_mcalc_flag)
	BB_SENSITIZE( wshdfunc_pos_mview_bbox, False ) ;
	else
	BB_SENSITIZE( wshdfunc_pos_mview_bbox, True ) ; */

   func_pos_mcalc = newmcalc ;

   return ;
}

void DIFF_shd_grid_av_CB( MCW_arrowval * av , XtPointer cd )
{
   int cc ;

   INVALIDATE_OVERLAY ;

   return ;
}


/*------------------------------------------------------------------------------
  DTI Viewer stuff
--------------------------------------------------------------------------------*/


/*------------------------------------------------------------------------------
  SHD Viewer stuff
--------------------------------------------------------------------------------*/

void DIFF_wshdfunc_shd_lval_av_CB( MCW_arrowval * av , XtPointer cd )
{
   DIF_state *dif = current_dif_state ;

   INVALIDATE_OVERLAY ;

   /* L to view can only be as high as L computed */

   if(av->ival > shd_lval_av->ival) 
     AV_assign_ival(wshdfunc_shd_lval_av,shd_lval_av->ival);

   /* If L is changed, assume we want M = L, as it 
      should be in almost all cases.
      (if not, then you have to adjust M yourself! */

   AV_assign_ival(wshdfunc_shd_mval_av,av->ival);

   /* If L falls below M, adjust M down */

   if(av->ival < wshdfunc_shd_mval_av->ival) 
     AV_assign_ival(wshdfunc_shd_mval_av,av->ival);

   dif->shd->view->nl = av->ival;
   DIFF_shd_display_CB( NULL , cd , NULL );

   return ;
}

void DIFF_wshdfunc_shd_mval_av_CB( MCW_arrowval * av , XtPointer cd )
{
   DIF_state *dif = current_dif_state ;

   INVALIDATE_OVERLAY ;

   /* M to view can only be as high as M computed */

   if(av->ival > shd_mval_av->ival) 
     AV_assign_ival(wshdfunc_shd_mval_av,shd_mval_av->ival);

   /* If M goes above L, adjust M down to L */

   if(av->ival > wshdfunc_shd_lval_av->ival) {
     AV_assign_ival(av,wshdfunc_shd_lval_av->ival);
   }

   dif->shd->view->nm = av->ival;
   DIFF_shd_display_CB( NULL , cd , NULL );

   return ;
}


/*------------------------------------------------------------------------------
  FTM Viewer stuff
--------------------------------------------------------------------------------*/

void DIFF_wftmfunc_ftm_lval_av_CB( MCW_arrowval * av , XtPointer cd )
{
   int cc ;

   INVALIDATE_OVERLAY ;

   /* L to view can only be as high as L computed */

   if(av->ival > ftm_lval_av->ival) 
     AV_assign_ival(wftmfunc_ftm_lval_av,ftm_lval_av->ival);

   /* If L is changed, assume we want M = L, as it 
      should be in almost all cases.
      (if not, then you have to adjust M yourself! */

   AV_assign_ival(wftmfunc_ftm_mval_av,av->ival);

   /* If L falls below M, adjust M down */

   if(av->ival < wftmfunc_ftm_mval_av->ival) 
     AV_assign_ival(wftmfunc_ftm_mval_av,av->ival);

   return ;
}

void DIFF_wftmfunc_ftm_mval_av_CB( MCW_arrowval * av , XtPointer cd )
{
   int cc ;

   INVALIDATE_OVERLAY ;

   /* M to view can only be as high as M computed */

   if(av->ival > ftm_mval_av->ival) 
     AV_assign_ival(wftmfunc_ftm_mval_av,ftm_mval_av->ival);

   /* If M goes above L, adjust M down to L */

   if(av->ival > wftmfunc_ftm_lval_av->ival) {
     AV_assign_ival(av,wftmfunc_ftm_lval_av->ival);
   }

   return ;
}

/*------------------------------------------------------------------------------
   DTI computation using 3dDWItoDT
--------------------------------------------------------------------------------*/

void DIFF_dwi2dt_compute_CB( Widget wcancel, XtPointer client_data,
			     XtPointer call_data ) {
   char *output_prefix;
   char *modified_output_prefix;
   char *modified_dset_prefix;
   char *system_call_string;
   char *full_path_filename;
   int nvert = current_dif_state->img->dgrad->nvert;
   int i;

   Widget popup;

   THD_3dim_dataset *new_dset = NULL;

   /* make a temporary file to hold the gradient information */
   char *tensorfile;
   tensorfile = (char *) malloc( sizeof(char) * THD_MAX_PREFIX );
   sprintf(tensorfile,"tensorXXXXXX");
   
   
   FILE *fp;
   if (!mktemp(tensorfile)) {
     printf("couldn't create necessary temporary file\n");
     return;
   }
   fp = fopen(tensorfile, "w");
   for (i = 0; i < nvert; i++) {
     fprintf(fp, "%f %f %f\n",
	     current_dif_state->img->GX[i],
	     current_dif_state->img->GY[i],
	     current_dif_state->img->GZ[i]);
   }
   fclose(fp);

   modified_output_prefix = (char *) malloc( sizeof(char) * THD_MAX_PREFIX ) ;
   modified_dset_prefix = (char *) malloc( sizeof(char) * THD_MAX_PREFIX ) ;
   system_call_string = (char *) malloc(sizeof(char)*3*THD_MAX_PREFIX);
   full_path_filename = (char *) malloc( sizeof(char) * THD_MAX_PREFIX ) ;

   if( dset == NULL ){ XBell(dc->display,100) ; return ; }
   output_prefix = MAKE_GOOD_PREFIX( dset, "DTI" ) ;

   sprintf(modified_output_prefix,"%s+orig",output_prefix);
   sprintf(modified_dset_prefix,"%s+orig",DSET_PREFIX(dset));
   sprintf(system_call_string,
	   "3dDWItoDT -automask -eigs -prefix %s %s %s",
	   modified_output_prefix, tensorfile, modified_dset_prefix);

   popup = MCW_popup_message( aniso_compute_pb ,
		      "\n** Computing ... please be patient **\n",
		      MCW_CALLER_KILL) ;

   XtDestroyWidget( popup );

   if (system(system_call_string)) {
     printf("something's broken about 3dDWItoDT\n");
     return;
   }

   remove(tensorfile);

   if (DSET_DIRNAME(dset)[strlen(DSET_DIRNAME(dset)) - 1] != '/') {
     printf("something's broken about 3dDWItoDT\n");
     return;
   }
   sprintf(full_path_filename, "%s%s", DSET_DIRNAME(dset), modified_output_prefix);
   new_dset = THD_open_dataset(full_path_filename);
   PLUTO_add_dset( plint , new_dset , DSET_ACTION_MAKE_CURRENT ) ;
   printf("3dDWItoDT complete\n");
   return;
}

/*------------------------------------------------------------------------------
   DTI computation
--------------------------------------------------------------------------------*/

void DIFF_dti_compute_CB( Widget wcancel, XtPointer client_data, XtPointer call_data )
{
   int i,j ;
   int nx,ny,nxy,nsl,ndif,nvox ;
   Mask_state *ms ;
   float maskval ;

   IMG_state     *img = current_dif_state->img ;
   DTI_lsq_state *lsq = current_dif_state->dti->lsq ;
   DTI_prc_state *prc = current_dif_state->dti->prc ; 

   char *output_prefix;
   THD_3dim_dataset *new_dset = NULL;
   int cancelcode = 0;
   int debug = 1;
   int debug2 = 0;

   sprintf(routine_name2,"%s","DIFF_dti_compute_CB");

   /* Check input integrity */

   if( dset == NULL ){ XBell(dc->display,100) ; return ; }

   /* Retrieve dimensions */

   /* 
   ndif = DSET_NVALS(dset);
   nx   = DSET_NX(dset);
   ny   = DSET_NY(dset);
   nsl  = DSET_NZ(dset);
   nxy = nx * ny;		
   */
   nvox = DSET_NVOX(dset);

   if(debug) printf("DIFF_dti_compute_CB:\n"
		    "nbvals = %i\n"
		    "ndif = %i\n"
		    "(nx,ny) = (%i,%i)\n"
		    "nsl = %i\n"
		    "nvox = %i\n",
		    img->nbvals,img->ndif,img->nx,img->ny,img->nsl,nvox);

   output_prefix = MAKE_GOOD_PREFIX( dset, "DTI" ) ;

   if(debug) printf("# sub-bricks = %i\n",DSET_NVALS(dset));

   /* Compute mask value as mean voxel intensity over all b=0 slices */

   ms = init_mask_state( dset, 0, 0, NULL, NULL ) ;
   mask( dset, ms, 0 ); 
   if(debug) printf("Maskval = %f\n",ms->mean); 

   /* set mask value in diffusion state */
   prc->maskval = ms->mean ;

   /* Setup of DTI */

   printf("Found %i b=0 image(s) \n"
	  "Found %i diffusion directions \n"
	  "Found %i b-value(s)\n"
	  "Found %i slice(s)\n"
	  "Total #images = %i\n",img->nb0,img->dgrad->nvert,img->nbvals,img->nsl,img->nim);
   if(img->ndif<7) {
     printf("%s: Not enough directions found (%i) - Need 6!\n",
	    routine_name2,img->ndif);
     exit(0);
   }
   /* Initialize DTI */

   if(!current_dif_state->sta->dti_initialized) {
     if(debug) printf("DIFF_dti_compute_CB:  Initializing DTI ...!\n");
     init_dti( current_dif_state ) ;
   }

   /* Calculate B matrix from diffusion gradient pattern */
   
   printf("Generating b matrix...\n");
   setB( current_dif_state );
   printf("...done\n");

   /* Put B into lsq */

   for (j=0; j<current_dif_state->dti->lsq->lsqdim; j++) { /* size of alpha */
     for (i=0; i<prc->nimsl; i++) { /* ndif*nbvals */  
       lsq->tsim[j][i] = prc->Bmat[i][j] ;
       if(debug2) printf("lsq->tsim[%i][%i] = %f\n",j,i,lsq->tsim[j][i]);
     }
   }

   /* Drive the DTI diffusion computations */

   /* nbrik = current_dif_state->dti->prc->edim
      + current_dif_state->dti->prc->nten,*/

   printf("dset directory name: %s\n", dset->dblk->diskptr->directory_name);
   new_dset = MAKER_4D_to_typed_difbuc(wcancel, shell, 
				       dset , 
				       b0_dstats->mask,
				       output_prefix ,
				       MRI_float , 
				       reinterpret_cast<void (*)()>
				       (DIFF_aniso_meth),
				       METH_DTI , 
				       &cancelcode , 
				       (void *)current_dif_state );

   /* Check output integrity */

   if( new_dset == NULL || cancelcode ){	/* something amiss ... */

      char buf[256] ;
      sprintf(buf," \n"
                  "** Problem writing file %s!\n" 
                  "** (Routine: %s)" 
                  "\n", 
	      output_prefix,routine_name2) ;
      (void) MCW_popup_message( choose_pb , buf ,
                                MCW_USER_KILL | MCW_TIMER_KILL ) ;

      current_dif_state->sta->dti_performed = 0;

      PLUTO_beep() ; return ;

   } else {			/* everything ok ... */

     current_dif_state->sta->dti_performed = 1;

     /* Write out data and add to current session */

     PLUTO_add_dset( plint , new_dset , DSET_ACTION_MAKE_CURRENT ) ;
   }

   return;
}

/*------------------------------------------------------------------------------
   SHD computation
--------------------------------------------------------------------------------*/

void DIFF_shd_compute_CB( Widget wcancel, 
			  XtPointer client_data, 
			  XtPointer call_data )
{
   int i,j ;
   int nvox ;
   Mask_state *ms ;
   float maskval ;

   DIF_status    *sta = current_dif_state->sta ;
   IMG_state     *img = current_dif_state->img ;
   DTI_lsq_state *lsq = current_dif_state->dti->lsq ;
   DTI_prc_state *prc = current_dif_state->dti->prc ;

   char *output_prefix;
   THD_3dim_dataset *new_dset = NULL;
   int cancelcode = 0;
   int debug = 1;

   THD_3dim_dataset *dt4_dset = NULL; // 4th-order diffusion tensor
   char *dt4_prefix;

   sprintf(routine_name2,"%s","DIFF_shd_compute_CB");

   /* Check input integrity */

   if( dset == NULL ){ XBell(dc->display,100) ; return ; }

   /* Retrieve dimensions */

   nvox = DSET_NVOX(dset);

   if(debug) printf("DIFF_shd_compute_CB:\n"
		    "nbvals = %i\n"
		    "ndif = %i\n"
		    "(nx,ny) = (%i,%i)\n"
		    "nsl = %i\n"
		    "nvox = %i\n",
		    img->nbvals,img->ndif,img->nx,img->ny,img->nsl,nvox);

   if(swd_flag) {
     output_prefix = MAKE_GOOD_PREFIX( dset, "SWD" ) ; /* 16-Jul-04 */
   } else {
     output_prefix = MAKE_GOOD_PREFIX( dset, "SHD" ) ;
   }
   if(debug) printf("# sub-bricks = %i\n",DSET_NVALS(dset));

   /* Compute mask value as mean voxel intensity over all b=0 slices */

   ms = init_mask_state( dset, 0, 0, NULL, NULL ) ;
   mask( dset, ms, 0 ); 
   if(debug) printf("Maskval = %f\n",ms->mean); 

   /* set mask value in diffusion state */
   prc->maskval = ms->mean ;

   /* Setup of SHD */

   printf("Found %i b=0 image(s) \n"
	  "Found %i diffusion directions \n"
	  "Found %i b-value(s)\n"
	  "Found %i slice(s)\n"
	  "Total #images = %i\n",img->nb0,img->dgrad->nvert,img->nbvals,img->nsl,img->nim);
   if(img->ndif<7) {
     printf("%s: Not enough directions found (%i) - Need 6!\n",routine_name2,img->ndif);
     exit(0);
   }

   /* Initialize SHD */

   if( !current_dif_state->sta->shd_initialized ) {

     if(debug) printf("DIFF_shd_compute_CB:  Initializing SHD ...!\n");
     
     get_spharm_control_vals( current_dif_state ) ;
     init_shd( current_dif_state ) ;
   }

   /* Calculate B matrix from diffusion gradient pattern */
   
   printf("Generating b matrix...\n");
   setB( current_dif_state );
   printf("...done\n");

   /* Put B into lsq */

#if 0
     for (j=0; j<lsq->lsqdim; j++) { /* size of alpha */
       for (i=0; i<prc->nimsl; i++) { /* ndif*nbvals */  
	 lsq->tsim[j][i] = prc->Bmat[i][j] ;
       }
     } 
#endif

   /* Drive the SHD diffusion computations */
   /* Now allow SWD computation.  16-Jul-04 (lrf) */

   if(swd_flag) {

     printf("plug_diff: Calling SWD Maker with nlm = %i, nrad = %i\n",
	    current_dif_state->shd->calc->nrad,
	    current_dif_state->shd->calc->nlm); 

     new_dset = MAKER_4D_to_typed_swdbuc(wcancel, shell, 
					 dset , 
					 b0_dstats->mask,
					 output_prefix ,
					 MRI_complex , 
					 reinterpret_cast<void (*)()>
					 (DIFF_aniso_meth),
					 METH_SWD_SWD,
					 &cancelcode , 
					 (void *)current_dif_state );
   } else {

     printf("plug_diff: Calling SHD Maker with nlm = %i\n",
	  current_dif_state->shd->calc->nlm); 

     /* nbrik = current_dif_state->shd->calc->nlm */
     new_dset = MAKER_4D_to_typed_difbuc(wcancel, shell, 
					 dset , 
					 b0_dstats->mask,
					 output_prefix ,
					 MRI_complex , 
					 reinterpret_cast<void (*)()>
					 (DIFF_aniso_meth), 
					 METH_SHD_SHD,
					 &cancelcode , 
					 (void *)current_dif_state );
   }

   printf("DIFF_shd_compute_CB: cancelcode=%i\n",cancelcode);

   /* Check output integrity */

   if( new_dset == NULL || cancelcode ){	/* something amiss ... */

      char buf[256] ;
      sprintf(buf," \n"
                  "** Problem writing file %s!\n" 
                  "** (Routine: %s)" 
                  "\n", 
	      output_prefix,routine_name2) ;
      (void) MCW_popup_message( choose_pb , buf ,
                                MCW_USER_KILL | MCW_TIMER_KILL ) ;

      current_dif_state->sta->shd_performed = 0;

      PLUTO_beep() ; return ;

   } else {

       /* If successful, set flag */
       current_dif_state->sta->shd_performed = 1;

       /* Write out data and add to current session */

       PLUTO_add_dset( plint , new_dset , DSET_ACTION_MAKE_CURRENT ) ;

       // calculate 4th-order diffusion tensor from SHD coefficients
       dt4_prefix = MAKE_GOOD_PREFIX( dset, "DT4" );
       dt4_dset = shd2tensor4th( new_dset, dt4_prefix );
       if( dt4_dset == NULL ) printf( "** not enough memory for dt4!!\n");
               // write out dataset and add to current session
       else
         PLUTO_add_dset( plint, dt4_dset, DSET_ACTION_MAKE_CURRENT );
   }

   return;
}

/*------------------------------------------------------------------------------
  FTM2 computation
--------------------------------------------------------------------------------*/

void DIFF_ftm2_compute_CB( Widget wcancel, 
			   XtPointer client_data, 
			   XtPointer call_data )
{
  Mask_state *ms ;
  char *output_prefix;
  THD_3dim_dataset *new_dset = NULL;
  int cancelcode = 0;
  int debug = 0;

  sprintf(routine_name2,"%s","DIFF_ftm2_compute_CB");

  /* Check input integrity */

  if( dset == NULL ){ XBell(dc->display,100) ; return ; }

  /* Retrieve dimensions */

  output_prefix = MAKE_GOOD_PREFIX( dset, "FTM" ) ;

  /* Compute mask value as mean voxel intensity over all b=0 slices */

  ms = init_mask_state( dset, 0, 0, NULL, NULL ) ;
  mask( dset, ms, 0 ); 

  /* Initialize the fiber diffusion computations */

 // cyrus:  init_ftm_state( current_dif_state );
  init_ftm_state( current_dif_state->dti, current_dif_state->shd );

  /* Drive the fiber diffusion computations */

  new_dset = MAKER_4D_to_typed_ftmbuc( wcancel, shell, 
				       dset, im3d,
				       b0_dstats->mask,
				       output_prefix ,
				       MRI_float , 
				       1 ,
				       /*DIFF_fiber_meth , */
				       METH_FTM_SHD , 
				       &cancelcode , 
				       (void *)current_dif_state );

  /* Check output integrity */

  if( new_dset == NULL || cancelcode ){	/* something amiss ... */

    char buf[256] ;
    sprintf(buf," \n"
	    "** Problem writing file %s!\n" 
	    "** (Routine: %s)" 
	    "\n", 
	    output_prefix,routine_name2) ;
    (void) MCW_popup_message( choose_pb , buf ,
			      MCW_USER_KILL | MCW_TIMER_KILL ) ;

    current_dif_state->sta->ftm_performed = 0;

    PLUTO_beep() ; return ;

  } else {			/* everything ok ... */

    current_dif_state->sta->ftm_performed = 1;

    /* Write out data and add to current session */

    PLUTO_add_dset( plint , new_dset , DSET_ACTION_MAKE_CURRENT ) ;

  }

  return;

}

/*------------------------------------------------------------------------------
  FTM1 computation
--------------------------------------------------------------------------------*/

void DIFF_ftm1_compute_CB( Widget wcancel, 
			   XtPointer client_data, 
			   XtPointer call_data )
{
  char *program_name = "DIFF_ftm1_compute_CB";
  char buf[256] ;
  int debug = 0;
  int debug2 = 0;

  if(!ftm_scalars_flag) {
     (void) MCW_popup_message( wcancel ,
			       "\n** No DTI dataset chosen! **\n",
			       MCW_USER_KILL | MCW_TIMER_KILL ) ;
     PLUTO_beep() ; return ;
  }

  Launch_vtk_vector_window(shell, func_dset); 

  switch( vtk_vector_window_status ) {
  case VTK_VECTOR_WINDOW_CLOSED:
    /*(void) MCW_popup_message( wcancel ,
			      "\n** vtk window is closed: need to open! **\n",
			      MCW_USER_KILL | MCW_TIMER_KILL ) ; 
			      fprintf(stderr, "opening vtk window\n" );*/
    vtk_vector_window_status = VTK_VECTOR_WINDOW_OPEN;
    break;

  case VTK_VECTOR_WINDOW_OPEN:
    /*(void) MCW_popup_message( wcancel ,
			      "\n** vtk window is opened: just need to refresh! **\n",
			      MCW_USER_KILL | MCW_TIMER_KILL ) ;*/
    break;
  default:
    (void) MCW_popup_message( wcancel ,
			      "\n** vtk window is neither closed or opened:"
			      "something is wrong! **\n",
			      MCW_USER_KILL | MCW_TIMER_KILL ) ;
    break;
  }

  if( vtk_vector_window_status == VTK_VECTOR_WINDOW_OPEN ) {

    int   ijk,ix,jy,kz;
    float loc[3];
    int   sdims[3] ;

    if( VIEW_DATA ) {

      if( func_dset != NULL ) {	/* The vector data */

	DSET_load( func_dset ); 
	if( ! DSET_LOADED(func_dset) ){
	  fprintf(stderr,"*** Error: can't load data from %s!\n"
		  "Returning from %s\n",
		  DSET_FILECODE(func_dset),program_name) ;
	  return ;
	}

	/* Retrieve dimensions */

	int nvox = DSET_NVOX(func_dset);
	int nbrk = DSET_NVALS(func_dset);
	int nx   = DSET_NX(func_dset);
	int ny   = DSET_NY(func_dset);
	int nz   = DSET_NZ(func_dset);
	int nxy  = nx * ny;		

	float dx = DSET_DX(func_dset),
	      dy = DSET_DY(func_dset),
	      dz = DSET_DZ(func_dset);

	LOAD_VEC3(sdims,nx,ny,nz);

	fprintf(stderr,"%s: Loading func dataset '%s': \n"
		"\t orientation =  (%s,%s,%s)\n"
		"\t origin =  (%f,%f,%f)\n"
		"\t dimensions (nx,ny,nz) =  (%i,%i,%i)\n"
		"\t increments (dx,dy,dz) =  (%f,%f,%f)\n",
		program_name,DSET_PREFIX(func_dset),
		ORIENT_typestr[func_dset->daxes->xxorient],
		ORIENT_typestr[func_dset->daxes->yyorient],
		ORIENT_typestr[func_dset->daxes->zzorient],
		DSET_XORG(func_dset),DSET_YORG(func_dset),DSET_ZORG(func_dset),
		nx,ny,nz,dx,dy,dz);

	fprintf(stderr,"%s: Loading eigvectors ... \n",program_name); 

	/********************* vtk *********************/

	float vec[3],val[3];
	int ncmp = 3;          /* Number of eigenvector components */
	int eval_offset = 0 ;  /* Location of first eigenvector component in .DTI buc */
	int reverse_evecs = 0; /* AHB to account for wrong eval sign!!! FIX!!! */

	int evec_offset = 3 + reverse_evecs*6;  /* Location of first eigevector component in .DTI buc */

	switch (nbrk) {
	case 20:
	  eval_offset = 6;
	  evec_offset = 9;
	  break;
	default:
	case 12:
	  eval_offset = 0;
	  evec_offset = 3;
	  break;
	}
	/* The eigenvector polydata */
	vtkStructuredPoints *Evecs = vtkStructuredPoints::New();

	/* The locations */
	//vtkPoints *NewPoints = vtkPoints::New();
	//NewPoints->SetNumberOfPoints(nvox) ;

	/* The actual eigenvalues */

	vtkFloatArray *NewScalars = vtkFloatArray::New();
	NewScalars->SetNumberOfTuples(nvox);

	/* The actual eigenvectors */

	// Note: SetNumberOfComponents() must come before SetNumberOfTuples()!
	vtkFloatArray *NewVectors = vtkFloatArray::New();
	NewVectors->SetNumberOfComponents(3);
	NewVectors->SetNumberOfTuples(nvox);
	
	if(!ISANAT(dset)) {
	  fprintf(stderr,"%s: dset not an ANAT dset!\n",program_name); 
	  exit(0);
	}

	int typeDset = DSET_BRICK_TYPE( func_dset, evec_offset );

	switch(typeDset) {

	  case MRI_float:{

	    float *ev11,*ev12,*ev13 ;
	    float *eval1,*eval2,*eval3 ;
	    float fai ;
	    float aveD ;

	    /*
	     * We use the Dicom axes as our canonical coordinate system,
	     * which are x = R->L, y = A->P, z = I->S
	     */

	    if (func_dset->daxes->xxorient == ORI_R2L_TYPE ||
		func_dset->daxes->xxorient == ORI_L2R_TYPE)
	      {
		if (func_dset->daxes->yyorient == ORI_P2A_TYPE ||
		    func_dset->daxes->yyorient == ORI_A2P_TYPE)
		  {
		    Evecs->SetDimensions(nx, ny, nz);
		    Evecs->SetSpacing(fabs(dx), fabs(dy), fabs(dz));
		  }
		else  //yyorient is I2S or S2I
		  {
		    Evecs->SetDimensions(nx, nz, ny);
		    Evecs->SetSpacing(fabs(dx), fabs(dz), fabs(dy));
		  }
	      }
	    else if (func_dset->daxes->xxorient == ORI_P2A_TYPE ||
		     func_dset->daxes->xxorient == ORI_A2P_TYPE)
	      {
		if (func_dset->daxes->yyorient == ORI_R2L_TYPE ||
		    func_dset->daxes->yyorient == ORI_L2R_TYPE)
		  {
		    Evecs->SetDimensions(ny, nx, nz);
		    Evecs->SetSpacing(fabs(dy), fabs(dx), fabs(dz));
		  }
		else  //yyorient is I2S or S2I
		  {
		    Evecs->SetDimensions(nz, nx, ny);
		    Evecs->SetSpacing(fabs(dz), fabs(dx), fabs(dy));
		  }
	      }
	    else  //xxorient is I2S or S2I
	      {
		if (func_dset->daxes->yyorient == ORI_P2A_TYPE ||
		    func_dset->daxes->yyorient == ORI_A2P_TYPE)
		  {
		    Evecs->SetDimensions(nz, ny, nx);
		    Evecs->SetSpacing(fabs(dz), fabs(dy), fabs(dx));
		  }
		else  //yyorient is R2L or L2R
		  {
		    Evecs->SetDimensions(ny, nz, nx);
		    Evecs->SetSpacing(fabs(dy), fabs(dz), fabs(dx));
		  }
	      }

	    /*
	     * To compute the origin in terms of Dicom coordinates, we use
	     * the axes' to_dicomm transformation
	     */
	    THD_fvec3 origin_dset;
	    LOAD_FVEC3(origin_dset, 
		       DSET_XORG(func_dset), 
		       DSET_YORG(func_dset), 
		       DSET_ZORG(func_dset));
	    if (dx < 0.0)
	      {
		origin_dset.xyz[0] = origin_dset.xyz[0] + dx * (nx - 1);
	      }
	    if (dy < 0.0)
	      {
		origin_dset.xyz[1] = origin_dset.xyz[1] + dy * (ny - 1);
	      }
	    if (dz < 0.0)
	      {
		origin_dset.xyz[2] = origin_dset.xyz[2] + dz * (nz - 1);
	      }

	    THD_fvec3 origin_dicom = THD_3dmm_to_dicomm(func_dset, 
							origin_dset);

	    Evecs->SetOrigin((vtkFloatingPointType) origin_dicom.xyz[0],
			     (vtkFloatingPointType) origin_dicom.xyz[1],
			     (vtkFloatingPointType) origin_dicom.xyz[2]);

	    for (int i=0; i < nx; ++i)
	      {
		for (int j=0; j < ny; ++j)
		  {
		    for (int k=0; k < nz; ++k)
		      {			
			int afniIndex = DSET_ixyz_to_index(func_dset,i,j,k);

			//GET_LOC(loc,afniIndex,dset);

			//if(debug) {
			//  if(!(afniIndex%(dset->daxes->nxx*dset->daxes->nxx))) PRINT_LOC(loc,afniIndex,dset);
			//}

			/* Eigenvalues */

			eval1 = (float *) DSET_ARRAY(func_dset,eval_offset) ;
			eval2 = (float *) DSET_ARRAY(func_dset,eval_offset+1) ;
			eval3 = (float *) DSET_ARRAY(func_dset,eval_offset+2) ;
			LOAD_VEC3(val,eval1[afniIndex],eval2[afniIndex],eval3[afniIndex]);
			fai = calc_fai(val,3) ;	/* calculate fai */
			//if(fai>0.) printf("fai = %f\n",fai);

			/* Principle eigenvector */
			if(debug2) printf("loading buckets (%i,%i,%i) of voxel %i\n",
					  evec_offset,evec_offset+1,evec_offset+2,afniIndex);

			ev11 = (float *) DSET_ARRAY(func_dset,evec_offset  ) ;
			ev12 = (float *) DSET_ARRAY(func_dset,evec_offset+1) ;
			ev13 = (float *) DSET_ARRAY(func_dset,evec_offset+2) ;
			LOAD_VEC3(vec,ev11[afniIndex],ev12[afniIndex],ev13[afniIndex]);

			if(ev12[afniIndex]<0) NEGATE_VEC3(vec); /* Keep evecs in +y direction (05-01-27, lrf) */

			/* Load up the polydata */

			/* 
			 * As with origin, use the to_dicom transform
			 * to transform positon into dicom coordinates,
			 * to eliminate effect of axis orientation
			 */
			THD_fvec3 position_dset;
			LOAD_FVEC3(position_dset,
				   i * dx + DSET_XORG(func_dset),
				   j * dy + DSET_YORG(func_dset),
				   k * dz + DSET_ZORG(func_dset));

			THD_fvec3 position_dicom =
			  THD_3dmm_to_dicomm(func_dset, position_dset);

			int vtkIndex = Evecs->FindPoint(
				(vtkFloatingPointType) position_dicom.xyz[0],
				(vtkFloatingPointType) position_dicom.xyz[1],
				(vtkFloatingPointType) position_dicom.xyz[2]);

			NewScalars->SetTuple(vtkIndex,&fai);
			NewVectors->SetTuple(vtkIndex,vec);
		      }
		  }
	      }

	    //Evecs->SetPoints(NewPoints);
	    Evecs->GetPointData()->SetScalars(NewScalars);
	    Evecs->GetPointData()->SetVectors(NewVectors);

	  }
	    //NewPoints->Delete() ;
	  NewScalars->Delete() ;
	  NewVectors->Delete() ;

	  if(debug) Evecs->Print(cout) ;

	  break;
	  default: 
	    fprintf(stderr,
		    "%s:  Eigenvectors should be floats!\n"
		    "(something else found)\n",program_name);
	  break;
	}
	fprintf(stderr , "done\n");
      
	/* Visualize */

	fprintf(stderr,"Visualizing ...\n");
	Send_VData_to_vtk_vector_window(Evecs);
	Evecs->Delete();

      } else {
	fprintf(stderr , "functional dataset to view data not loaded\n");
      }

      /* Load the anatomical data */

      Load_FTM_Anat_DataSet();

    }
  }

}

/*-------------------------------------------------------------------------
   Start a view (12-3, 23-1, or 31-2)
---------------------------------------------------------------------------*/

static char * AFNI_image_help =
 "Button 1: Set crosshair location\n"
 "Button 3: Pop up image menu\n\n"
 "Shift/Ctrl/Alt + Button 3\n"
 "will open up the Disp/Mont/Save\n"
 "control panels, respectively." ;

static char * AFNI_arrowpad_help =
   "Click arrows to scroll crosshair position\n"
   "Click button to open/close crosshair gap " ;

static char * AFNI_arrowpad_hint[] = {
  "Scroll crosshairs down" ,
  "Scroll crosshairs up" ,
  "Scroll crosshairs left" ,
  "Scroll crosshairs right" ,
  "Close/open crosshairs gap"
} ;

/*----------------------------------------------------------------------

   Title: DIF_brick_to_mri

   Purpose: Routine to extract a plane of data from a 3D brick
            (used as a "get_image" routine for an MCW_imseq)

   History: lrf May 24, 2001: rwc's AFNI_brick_to_mri -> DIF_brick_to_mri

------------------------------------------------------------------------*/
/* cyrus : This seems like the place the overlay is being computed */
XtPointer DIF_brick_to_mri( int n , int type , FD_brick * br )
{
   MRI_IMAGE * im ;
   MCW_imseq_status * stat ;
   int i1,j2,k3;
   int jb,bb , dd1,dd2,tt1,tt2 ;
   AFNI_view_info *vinfo = im3d->vinfo;
   int current_slice = vinfo->k3;
   SHD_view_state *vshd = current_dif_state->shd->view ;
   int debug = 0;

   /* printf("DIF_brick_to_mri: current coords = (%i,%i,%i) \n",
      vinfo->i1,vinfo->j2,vinfo->k3); */

ENTRY("DIF_brick_to_mri") ;

if(PRINT_TRACING){ char str[256] ; sprintf(str,"n=%d type=%d",n,type) ; STATUS(str) ; }

   /*-------------------------------------------------*/
   /*-------- May 1996: graph callbacks first --------*/

   if( type == graCR_getstatus ){
      MCW_grapher_status * grstat = myXtNew( MCW_grapher_status ) ;

      grstat->num_total  = grstat->num_series = br->dset->dblk->nvals ;
      grstat->nx         = br->n1 ;
      grstat->ny         = br->n2 ;
      grstat->nz         = br->n3 ;

      grstat->send_CB    = reinterpret_cast<void (*)()>(AFNI_gra_send_CB);
      grstat->parent     = (XtPointer) br ;
      grstat->aux        = NULL ;

      grstat->transforms0D = & (GLOBAL_library.registered_0D) ;
      grstat->transforms1D = & (GLOBAL_library.registered_1D) ;

      strcpy( grstat->namecode , br->namecode ) ;

      RETURN( (XtPointer) grstat ) ;
   }

   if( type == graCR_getseries ){
      im = FD_brick_to_series( n , br ) ;
      RETURN( (XtPointer) im ) ;
   }

   /*----------------------------------------*/
   /*-------- Now do imseq callbacks --------*/

   if( n < 0 || n >= br->n3 ) RETURN(NULL) ;

   /*--- overlay # n ---*/

   if( type == isqCR_getoverlay  ){

STATUS("get overlay") ;

 if(debug) 
   printf("DIF_brick_to_mri: overlay n=%d, view (nl,nm) = (%i,%i)\n",
	   n,vshd->nl,vshd->nm) ;
      
      im = AFNI_overlay( n , br ) ;
      RETURN( (XtPointer) im ) ;
   }

   /*--- status ---*/

   if( type == isqCR_getstatus ){

STATUS("get status") ;

      stat = myXtNew( MCW_imseq_status ) ;

      stat->num_total  = br->n3 ;
      stat->num_series = br->n3 ;
      stat->send_CB    = reinterpret_cast<void (*)()>(AFNI_seq_send_CB);
      stat->parent     = (XtPointer) br ;
      stat->aux        = NULL ;

      stat->transforms0D = & (GLOBAL_library.registered_0D) ;
      stat->transforms2D = & (GLOBAL_library.registered_2D) ;

      RETURN( (XtPointer) stat ) ;
   }

   /*--- 26 Feb 2001: return a memplot drawing struct ---*/

   if( type == isqCR_getmemplot ){
      RETURN( (XtPointer) NULL ) ;
   }

   /*--- underlay image # n ---*/

   if( type == isqCR_getimage || type == isqCR_getqimage ){
      Three_D_View * im3d = (Three_D_View *) br->parent ;
      int ival = 0 ;

      /*** decide which 3D brick to extract data from (ival) ***/

      if( DSET_NUM_TIMES(br->dset) > 1 ){ /* a time-dependent brick */
         ival = im3d->vinfo->time_index ;
         if( ival >= br->dset->dblk->nvals ) ival = br->dset->dblk->nvals -1 ;

      } else if( ISANAT(br->dset) ){      /* an anatomy brick */
         if( ISANATBUCKET(br->dset) ) {     /* 30 Nov 1997 */
            ival = im3d->vinfo->anat_index ;
	 }
         else {
            ival = ANAT_ival_zero[br->dset->func_type] ;  /* get default data */
	 }
      } else if( ISFUNC(br->dset) ){      /* a functional brick */

        if( ISFUNCBUCKET(br->dset) ){     /* 30 Nov 1997 */
           ival = im3d->vinfo->fim_index ;
        }
        else {
           if( /*im3d->vinfo->showfunc_type == SHOWFUNC_THR  &&*/  /* showing thr */
               ISFUNC_UNDERLAY(im3d->vinfo->underlay_type) &&  /* as underlay? */
               FUNC_HAVE_THR(br->dset->func_type) ){

              ival = FUNC_ival_thr[br->dset->func_type] ; /* get thresh data */
           } else {
              ival = FUNC_ival_fim[br->dset->func_type] ; /* get fim data */
           }
        }
     }

      if( type == isqCR_getqimage ) ival = -1 ; /* get empty image */

if(PRINT_TRACING)
{ char str[256] ;
  sprintf(str,"getting image n1=%d n2=%d ival=%d",br->n1,br->n2,ival) ;
  STATUS(str) ; }

      LOAD_DSET_VIEWS(im3d) ;  /* 02 Nov 1996 */

      if( ival > 0 && GLOBAL_library.have_dummy_dataset && lrand48()%2 == 1 ){
         RETURN(NULL) ;  /* 23 Apr 2001: test of imseq.c NULL image handler */
      }

      /* Replicate current slice in underlay - lrf*/

      if(REPLICATE) {
	n = current_slice ;	
	if(debug) printf("DIF_brick_to_mri: replicate = %i (n=%i)\n",REPLICATE,n);
      }

      im = FD_warp_to_mri( n , ival , br ) ; /* actually get image from dataset */

      if( ival < 0 ) RETURN( (XtPointer) im ) ;  /* return fake image */

      /* Allow for thresholding of underlay image */

      if( im != NULL && im3d->vinfo->underlay_type == UNDERLAY_THRFUNC &&
          ISFUNC(br->dset) && FUNC_HAVE_THR(br->dset->func_type) &&
          im3d->vinfo->func_threshold > 0.0 ){

         MRI_IMAGE * im_thr , * qim ;
         double thresh ;
         int jval ;

         if( ISFUNCBUCKET(br->dset) )          /* 30 Nov 1997 */
            jval = im3d->vinfo->thr_index ;
         else
            jval = FUNC_ival_thr[br->dset->func_type] ;

         if( jval != ival ){
            im_thr = FD_warp_to_mri( n , jval ,  br ) ;
            if( im_thr == NULL ) im_thr = im ;
         } else {
            im_thr = im ;  /* if already have threshold data */
         }

         if( ! AFNI_GOOD_FUNC_DTYPE(im_thr->kind) ){ /* should not occur */
            qim = mri_to_float( im_thr ) ;           /* but if it does,  */
            if( im_thr != im ) mri_free( im_thr ) ;  /* make an OK type  */
            im_thr = qim ;
         }

         thresh = im3d->vinfo->func_threshold * im3d->vinfo->func_thresh_top ;

         if( im_thr->kind == MRI_short )
            thresh *= FUNC_scale_short[br->dset->func_type];
         else if( im_thr->kind == MRI_byte )
            thresh *= FUNC_scale_byte[br->dset->func_type];

         mri_threshold( -thresh , thresh , im_thr , im ) ;
         if( im_thr != im ) mri_free( im_thr ) ;
      }

      /* Load value of current pixel into display label */
      /* April 1996: only if image is at current slice  */

#ifdef ALLOW_BKGD_LAB
      { char buf[32] = "\0" ;
        AFNI_set_valabel( br , n , im , buf ) ;
        if( buf[0] != '\0' ){
           if( im3d->vinfo->underlay_type == UNDERLAY_ANAT )
              strcpy( im3d->vinfo->anat_val , buf ) ;
           else
              im3d->vinfo->anat_val[0] = '\0' ;

           if( im->kind != MRI_complex ){
              char qbuf[32] = "bg =" ;
              strcat(qbuf,buf) ; strcpy(buf,qbuf) ;
           }
           MCW_set_widget_label( im3d->vwid->imag->pop_bkgd_lab , buf ) ;
           XtManageChild( im3d->vwid->imag->pop_bkgd_lab ) ;
        }
      }
#endif

      RETURN( (XtPointer) im ) ;
   }

STATUS("get something else, but I don't care!") ;

   RETURN( NULL ) ;
}

/*------------------------------------------------------------------------------

   Title: SHD COMPUTE SCALARS

    Type: Callback

    Call: void DIFF_shd_compute_scalars_CB( Widget w , 
				            XtPointer client_data , 
				            XtPointer call_data )

   Purpose: Compute SHD scalars

-----------------------------------------------------------------------------*/

void DIFF_shd_compute_scalars_CB( Widget w , 
				  XtPointer client_data , 
				  XtPointer call_data )
{
   Three_D_View *im3d = (Three_D_View *) client_data ;
   DIF_state    *dif = current_dif_state ;
   int METH_SHD_SCALAR ;
   int meth_scalar_dim;

   int nx,ny,nxy,nsl,ndif,nvox,nbrk ;
   char buf[256] ;
   Mask_state *ms ;
   float maskval ;
   char prebuf[80];

   char *output_prefix;
   THD_3dim_dataset *new_dset = NULL;
   int cancelcode = 0;
   int debug = 0;

   sprintf(routine_name2,"%s","DIFF_shd_compute_scalars_CB");

   /* Check input integrity */

   if( dset == NULL ){ XBell(dc->display,100) ; return ; }

   if(!shd_scalars_flag) {
      (void) MCW_popup_message( w ,
				"\n** No SHD dataset chosen! **\n",
                                MCW_USER_KILL | MCW_TIMER_KILL ) ;
      PLUTO_beep() ; return ;
   }

   /* Retrieve dimensions */

   ndif = DSET_NVALS(dset);	/* diff directions are stored in time axis */
   nx   = DSET_NX(dset);
   ny   = DSET_NY(dset);
   nsl  = DSET_NZ(dset);
   nvox = DSET_NVOX(dset);
   nbrk = DSET_NVALS(dset) ;

   nxy = nx * ny;		

   if(debug) printf("DIFF_shd_compute_scalars_CB:\n"
		    "ndif = %i\n"
		    "(nx,ny) = (%i,%i)\n"
		    "nsl = %i\n"
		    "nvox = %i\n"
		    ,ndif,nx,ny,nsl,nvox);


   if(debug) printf("# sub-bricks = %i\n",nbrk);

   /* Compute mask value as mean voxel intensity over all b=0 slices */

   ms = init_mask_state( dset, 0, 0, NULL, NULL ) ;
   mask( dset, ms, 0 ); 
   if(debug) printf("Maskval = %f\n",ms->mean); 

   /* Choose scalar */

   /* Maybe not best way?  The "+1" is to move past SHD to scalar values
      so the first "case" below is never reached.  We should consolidate these!! */

   METH_SHD_SCALAR = wshdfunc_scalar_av->ival + METH_SHD_OFFSET + 1; 

   switch ( METH_SHD_SCALAR ) {

      default: 
	printf("\aplug_diff: Unrecognized value %i of METH_SHD_SCALAR! \n",
	       METH_SHD_SCALAR);
	printf("...returning...\n");
	return;

      case METH_SHD_SHD:{
	if(debug) printf("inside DIFF_shd_compute_scalars_CB: ival = %i\n",
			 METH_SHD_SHD);

	/* Initialize SHD */

	if(!current_dif_state->sta->shd_shd_initialized) {
	  if(debug) printf("Initializing SHD ...!\n");
	  init_shd_scalars( METH_SHD_SCALAR , current_dif_state ) ;
	}

	meth_scalar_dim = current_dif_state->shd->calc->nlm;	/* NL*NM */
      }
      break;

      case METH_SHD_FIB:{
	if(debug) printf("inside DIFF_shd_compute_scalars_CB: ival = %i\n",
			 METH_SHD_FIB);

	/* Initialize FIB */

	if(!current_dif_state->sta->shd_ftm_initialized) {
	  if(debug) printf("Initializing FIB ...!\n");
	  init_shd_scalars( METH_SHD_SCALAR , current_dif_state ) ;
	}

	/* iso, fib1, fibn */
	meth_scalar_dim = current_dif_state->shd->calc->nl; 
      }
      break;


      case METH_SHD_EOS:{
	if(debug) printf("inside DIFF_shd_compute_scalars_CB: ival = %i\n",
			 METH_SHD_EOS);

	/* Initialize EOS */

	if(!current_dif_state->sta->shd_eos_initialized) {
	  if(debug) printf("Initializing EOS ...!\n");
	  init_shd_scalars( METH_SHD_SCALAR , current_dif_state ) ;
	}

	/* meth_scalar_dim = current_dif_state->shd->calc->neos;*/
	meth_scalar_dim = 2;	/* even/odd */
      }
      break;

      case METH_SHD_FMI:{
	if(debug) printf("inside DIFF_shd_compute_scalars_CB: ival = %i\n",
			 METH_SHD_FMI);

	/* Initialize FMI */

	if(!current_dif_state->sta->shd_fmi_initialized) {
	  if(debug) printf("Initializing FMI ...!\n");
	  init_shd_scalars( METH_SHD_SCALAR , current_dif_state ) ;
	}

	meth_scalar_dim = 1;	/* fmi */

      }
      break;

      case METH_SHD_GRA:{
	if(debug) printf("inside DIFF_shd_compute_scalars_CB: ival = %i\n",
			 METH_SHD_GRA);

	/* Initialize GRA */

	if(!current_dif_state->sta->shd_gra_initialized) {
	  if(debug) printf("Initializing GRA ...!\n");
	  init_shd_scalars( METH_SHD_SCALAR , current_dif_state ) ;
	}

	meth_scalar_dim = 1;	/* rai */

      }
      break;

      case METH_SHD_GFA:{
	if(debug) printf("inside DIFF_shd_compute_scalars_CB: ival = %i\n",
			 METH_SHD_GFA);

	/* Initialize GFA */

	if(!current_dif_state->sta->shd_gfa_initialized) {
	  if(debug) printf("Initializing GFA ...!\n");
	  init_shd_scalars( METH_SHD_SCALAR , current_dif_state ) ;
	}

	meth_scalar_dim = 1;	/* rai */

      }
      break;

   }

   sprintf(prebuf,"%s",
	   SHD_OUTPUT_PREFIX_CHOICES[ METH_SHD_SCALAR - METH_SHD_OFFSET]);
   output_prefix = MAKE_GOOD_PREFIX( dset, prebuf ) ;

   /* set mask value in diffusion state */
   current_dif_state->dti->prc->maskval = ms->mean ;

   /* Drive the scalar diffusion computations */

   if(METH_SHD_SCALAR<0) {

     printf("DIFF_shd_compute_scalars_CB: "
	    "Incorrect value of METH_SHD_SCALAR!... returning ...\a\n"); 

   } else {

     if(debug) 
       printf("Calling MAKER_4D_to_typed_difbuc with METH_SHD_SCALAR = %i,\n"
	      "METH_SCALAR_DIM = %i, DSET_PREFIX = %s\n",
	      METH_SHD_SCALAR, meth_scalar_dim,DSET_PREFIX(func_dset)); 

     /* nbrik = meth_scalar_dim */
     new_dset = MAKER_4D_to_typed_difbuc(shd_cancel_scalars_pb, shell, 
					 func_dset, 
					 b0_dstats->mask, 
					 output_prefix ,
					 MRI_float , 
					 reinterpret_cast<void (*)()>
					 (DIFF_aniso_meth),
					 METH_SHD_SCALAR , 
					 &cancelcode , 
					 (void *)current_dif_state ) ;

     if(debug) printf("%s: back from MAKER_4D_to_typed_difbuc\n",__FILE__); 

   }


   /* Check output integrity */

   if( new_dset == NULL || cancelcode ){	/* something amiss ... */

      sprintf(buf," \n"
                  "** Problem writing file %s!\n" 
                  "** (Routine: %s)" 
                  "\n", 
	      output_prefix,routine_name2) ;
      (void) MCW_popup_message( choose_pb , buf ,
                                MCW_USER_KILL | MCW_TIMER_KILL ) ;
      PLUTO_beep() ; return ;

       switch ( wshdfunc_scalar_av->ival ) {

       default:
       break;

       case METH_SHD_SHD:{ current_dif_state->sta->shd_shd_performed = 0; } /* redundant? */
       break;

       case METH_SHD_FIB:{ current_dif_state->sta->shd_ftm_performed = 0; }
       break;

       case METH_SHD_EOS:{ current_dif_state->sta->shd_eos_performed = 0; }
       break;

       case METH_SHD_FMI:{ current_dif_state->sta->shd_fmi_performed = 0; }
       break;

       case METH_SHD_GRA:{ current_dif_state->sta->shd_gra_performed = 0; }
       break;

       case METH_SHD_GFA:{ current_dif_state->sta->shd_gfa_performed = 0; }
       break;

       }

      return ;			/* go out silently */

     } else {			/* everything ok ... */

       switch ( wshdfunc_scalar_av->ival ) {

       default:
       break;

       case METH_SHD_SHD:{ current_dif_state->sta->shd_shd_performed = 1; }
       break;

       case METH_SHD_FIB:{ current_dif_state->sta->shd_ftm_performed = 1; }
       break;

       case METH_SHD_EOS:{ current_dif_state->sta->shd_eos_performed = 1; }
       break;

       case METH_SHD_FMI:{ current_dif_state->sta->shd_fmi_performed = 1; }
       break;

       case METH_SHD_GRA:{ current_dif_state->sta->shd_gra_performed = 1; }
       break;

       case METH_SHD_GFA:{ current_dif_state->sta->shd_gfa_performed = 1; }
       break;

       }

       /* Write out data and add to current session */
       PLUTO_add_dset( plint , new_dset , DSET_ACTION_MAKE_CURRENT ) ;
     }

   return;

}

/*----------------------------------------------------------------------------

   Title: DTI COMPUTE SCALARS

    Type: Callback

    Call: void DIFF_dti_compute_scalars_CB( Widget w , 
				            XtPointer client_data , 
				            XtPointer call_data )

   Purpose: Compute DTI scalars

-----------------------------------------------------------------------------*/

void DIFF_dti_compute_scalars_CB( Widget w , 
				  XtPointer client_data , 
				  XtPointer call_data )
{
   Three_D_View *im3d = (Three_D_View *) client_data ;
   DIF_state  *dif = current_dif_state ;
   THD_3dim_dataset *qset = NULL;
   Mask_state *ms ;

   int        METH_DTI_SCALAR ;
   float      maskval ;
   char       buf[256], prebuf[80], *output_prefix;
   int        cancelcode = 0;

   int debug = 0;

   sprintf(routine_name2,"%s","DIFF_dti_compute_scalars_CB");

   /* Check input integrity */

   if( func_dset == NULL ){ XBell(dc->display,100) ; return ; }

   if(!dti_scalars_flag) {
      (void) MCW_popup_message( w ,
				"\n** No DTI dataset chosen! **\n",
                                MCW_USER_KILL | MCW_TIMER_KILL ) ;
      PLUTO_beep() ; return ;
   }

   if(debug) printf("# sub-bricks = %i\n",DSET_NVALS(func_dset));

   /* Compute mask value as mean voxel intensity over all b=0 slices.
      This uses dset, the anatomical dataset. */

   ms = init_mask_state( dset, 0, 0, NULL, NULL ) ;
   mask( dset, ms, 0 ); 
   if(debug) printf("Maskval = %f\n",ms->mean); 

   /* Choose scalar */

   METH_DTI_SCALAR = wdtifunc_scalar_av->ival + METH_DTI_OFFSET + 1;
   switch ( METH_DTI_SCALAR ) {

      default: printf("\aplug_diff: Unrecognized value %i of METH_DTI_SCALAR! \n",
		      METH_DTI_SCALAR);
	       printf("...returning...\n");
	       return;

      case METH_DTI_AVE:{
	if(debug) printf("inside DIFF_dti_compute_scalars_CB: METH_DTI_AVE\n");

	/* Initialize AVE */

	if(!current_dif_state->sta->dti_ave_initialized) {
	  if(debug) printf("Initializing AVE ...!\n");
	  init_dti_scalars( METH_DTI_SCALAR , current_dif_state ) ;
	}

      }
      break;

      case METH_DTI_FAI:{
	if(debug) printf("inside DIFF_dti_compute_scalars_CB: METH_DTI_FAI\n");

	/* Initialize FAI */

	if(!current_dif_state->sta->dti_fai_initialized) {
	  if(debug) printf("Initializing FAI ...!\n");
	  init_dti_scalars( METH_DTI_SCALAR , current_dif_state ) ;
	}

      }
      break;

      case METH_DTI_RAI:{
	if(debug) printf("inside DIFF_dti_compute_scalars_CB: METH_DTI_RAI\n");

	/* Initialize RAI */

	if(!current_dif_state->sta->dti_rai_initialized) {
	  if(debug) printf("Initializing RAI ...!\n");
	  init_dti_scalars( METH_DTI_SCALAR , current_dif_state ) ;
	}

      }
      break;

      case METH_DTI_LAT:{
	if(debug) printf("inside DIFF_dti_compute_scalars_CB: METH_DTI_LAT\n");

	/* Initialize LAT */

	if(!current_dif_state->sta->dti_lat_initialized) {
	  if(debug) printf("Initializing LAT ...!\n");
	  init_dti_scalars( METH_DTI_SCALAR , current_dif_state ) ;
	}

      }
      break;

      case METH_DTI_SDV:{
	if(debug) printf("inside DIFF_dti_compute_scalars_CB: METH_DTI_SDV\n");

	/* Initialize SDV */

	if(!current_dif_state->sta->dti_sdv_initialized) {
	  if(debug) printf("Initializing SDV ...!\n");
	  init_dti_scalars( METH_DTI_SCALAR , current_dif_state ) ;
	}

      }
      break;

   }

   sprintf(prebuf,"%s",
	   DTI_OUTPUT_PREFIX_CHOICES[ METH_DTI_SCALAR - METH_DTI_OFFSET]);
   output_prefix = MAKE_GOOD_PREFIX( dset, prebuf ) ;

   /* set mask value in diffusion state */
   current_dif_state->dti->prc->maskval = ms->mean ;

   /* Drive the scalar diffusion computations */

   if(METH_DTI_SCALAR<0) {

     printf("DIFF_dti_compute_scalars_CB: "
	    "Incorrect value of METH_DTI_SCALAR!... returning ...\a\n"); 

   } else {

     if(debug) printf("Calling MAKER_4D_to_typed_difbuc with "
		      "METH_DTI_SCALAR = %i, DSET_PREFIX = %s\n",
		      METH_DTI_SCALAR,DSET_PREFIX(func_dset));
     /* nbrik = 1  */
     qset = MAKER_4D_to_typed_difbuc(dti_cancel_scalars_pb, shell, 
				     func_dset, 
				     b0_dstats->mask, 
				     output_prefix , 
				     MRI_float , 
				     reinterpret_cast<void (*)()>
				     (DIFF_aniso_meth), 
				     METH_DTI_SCALAR , 
				     &cancelcode, 
				     (void *)current_dif_state );

     if(debug) printf("back from MAKER_4D_to_typed_difbuc\n"); 

   }

   /* Check output integrity */

   if( qset == NULL || cancelcode ){	/* something amiss ... */

      sprintf(buf," \n"
                  "** Problem writing file %s!\n" 
                  "** (Routine: %s)" 
                  "\n", 
	      output_prefix,routine_name2) ;
      (void) MCW_popup_message( choose_pb , buf ,
                                MCW_USER_KILL | MCW_TIMER_KILL ) ;
      PLUTO_beep() ; return ;

      switch ( wdtifunc_scalar_av->ival ) {

      default:
	 break ;
      case METH_DTI_AVE:{ current_dif_state->sta->dti_ave_performed = 0; }
	break;
      case METH_DTI_FAI:{ current_dif_state->sta->dti_fai_performed = 0; }
	break;
      case METH_DTI_RAI:{ current_dif_state->sta->dti_rai_performed = 0; }
	break;
      case METH_DTI_LAT:{ current_dif_state->sta->dti_lat_performed = 0; }
	break;
      case METH_DTI_SDV:{ current_dif_state->sta->dti_sdv_performed = 0; }
	break;

      }

      return ;			/* go out silently */

     } else {			/* everything ok ... */

       switch ( wdtifunc_scalar_av->ival ) {

       default:
	 break ;
       case METH_DTI_AVE:{ current_dif_state->sta->dti_ave_performed = 1; }
	 break;
       case METH_DTI_FAI:{ current_dif_state->sta->dti_fai_performed = 1; }
	 break;
       case METH_DTI_RAI:{ current_dif_state->sta->dti_rai_performed = 1; }
	 break;
       case METH_DTI_LAT:{ current_dif_state->sta->dti_lat_performed = 1; }
	 break;
       case METH_DTI_SDV:{ current_dif_state->sta->dti_sdv_performed = 1; }
	 break;

       }

       /* Write out data and add to current session */
       PLUTO_add_dset( plint , qset , DSET_ACTION_MAKE_CURRENT ) ;
     }

   return;
}

/*------------------------------------------------------------------------------

   Title: SHD SAVE SCALARS

    Type: Callback

    Call: void DIFF_shd_save_scalars_CB( Widget w , 
				            XtPointer client_data , 
				            XtPointer call_data )

   Purpose: Save SHD scalars

--------------------------------------------------------------------------------*/
void DIFF_shd_save_scalars_CB( Widget w , 
			       XtPointer client_data , 
			       XtPointer call_data )
{
   Three_D_View *im3d = (Three_D_View *) client_data ;
   DIF_state *dif = current_dif_state ;
   SHD_calc_state *calc = dif->shd->calc ;

   int debug = 0;
   int ll,mm;
   float coeff ;

   printf("DIFF_shd_save_scalars_CB: computing...\n");
   DSET_load(func_dset) ;            /* make sure is in memory */

   for( ll=calc->Lmin ; ll<=calc->Lmax ; ll++ ) {

     for( mm=-ll ; mm<=ll ; mm++ ) {

       if(debug) printf("abs(Alm[%i][%i]) = %f\n",ll,mm,coeff);

     }
   }

   printf("DIFF_shd_save_scalars_CB: ...done!\n");
   

}

/*------------------------------------------------------------------------------

   Title: DTI SAVE SCALARS

    Type: Callback

    Call: void DIFF_dti_save_scalars_CB( Widget w , 
				            XtPointer client_data , 
				            XtPointer call_data )

   Purpose: Save DTI scalars

--------------------------------------------------------------------------------*/

void DIFF_dti_save_scalars_CB( Widget w , 
				  XtPointer client_data , 
				  XtPointer call_data )
{
   Three_D_View *im3d = (Three_D_View *) client_data ;
   DIF_state *dif = current_dif_state ;

   printf("DIFF_dti_save_scalars_CB: Not available yet!\a\n");
}

/*------------------------------------------------------------------------------

   Title: SHD DISPLAY SCALARS

    Type: Callback

    Call: void DIFF_shd_display_scalars_CB( Widget w , 
				            XtPointer client_data , 
				            XtPointer call_data )

   Purpose: Display montage of multiple images for SHD scalar overlay.
            Sort of hardwired currently .

   History: mainly from AFNI_view_xyz_CB() in afni.c

--------------------------------------------------------------------------------*/

void DIFF_shd_display_scalars_CB( Widget w , 
				  XtPointer client_data , 
				  XtPointer call_data )
{
  printf(" DIFF_shd_display_scalars_CB not available yet\n");
}

/*------------------------------------------------------------------------------

   Title: DTI DISPLAY SCALARS

    Type: Callback

    Call: void DIFF_dti_display_scalars_CB( Widget w , 
				            XtPointer client_data , 
				            XtPointer call_data )

   Purpose: Display montage of multiple images for DTI scalar overlay.
            Sort of hardwired currently .

   History: mainly from AFNI_view_xyz_CB() in afni.c

--------------------------------------------------------------------------------*/

void DIFF_dti_display_scalars_CB( Widget w , 
				  XtPointer client_data , 
				  XtPointer call_data )
{
   Three_D_View *im3d = (Three_D_View *) client_data ;
   DIF_state *dif = current_dif_state ;

   int debug = 0;

   MCW_imseq   * sxyz , * syzx , * szxy , ** snew = NULL ;
   MCW_grapher * gxyz , * gyzx , * gzxy , ** gnew = NULL ;
   Widget        pboff , pb_xyz , pb_yzx , pb_zxy ;
   Widget        groff , gr_xyz , gr_yzx , gr_zxy ;
   FD_brick    * brnew ;
   int mirror=0 ;

   char *dti_viewer_title = "DTI Viewer" ;
   int  mp[5] ;			/* montage params */
   char mcol[128] = "\0" ;
   int  mww, mhh, msp, mgap ;
   int  mont_wsize, mont_hsize ; /* x and y size of montage array */

ENTRY("DIFF_dti_display_scalars_CB") ;

   if(dif!=NULL) { 
     printf("\aDIFF_dti_display_scalars_CB: dif->img->nsl = %i\n", 
	    dif->img->nsl) ;
   }

   if( ! IM3D_VALID(im3d) ) {
     printf("\aDIFF_dti_display_scalars_CB: Invalid im3d - returning \n");
     EXRETURN ;
   }

    dti_iwid = im3d->vwid->imag ;

    if(debug) {
      printf("DIFF_dti_display_scalars_CB: dti_display_flag = %i\n",
	     dti_display_flag);
      if( dti_iwid == NULL ){
	printf("DIFF_dti_display_scalars_CB: dti_iwid == NULL\n");
      } else {
	printf("DIFF_dti_display_scalars_CB: dti_iwid != NULL\n");
      }
    }

    if( dti_display_flag ) EXRETURN ; 
    dti_display_flag = 1;

    sdti_xyz = im3d->s123 ; gdti_xyz = im3d->g123 ;
    sdti_yzx = im3d->s231 ; gdti_yzx = im3d->g231 ;
    sdti_zxy = im3d->s312 ; gdti_zxy = im3d->g312 ;

    pb_xyz = dti_iwid->image_xyz_pb ;
    pb_yzx = dti_iwid->image_yzx_pb ;
    pb_zxy = dti_iwid->image_zxy_pb ;

    gr_xyz = dti_iwid->graph_xyz_pb ;
    gr_yzx = dti_iwid->graph_yzx_pb ;
    gr_zxy = dti_iwid->graph_zxy_pb ;

    /* handle case of button press of already
       open window by bringing that window to the top */

    if( w == pb_xyz && sxyz != NULL ){
       if( ISQ_REALZ(sxyz) )
          XMapRaised( XtDisplay(sxyz->wtop) , XtWindow(sxyz->wtop) ) ;
       EXRETURN ;
    } else if( w == pb_yzx && syzx != NULL ){
       if( ISQ_REALZ(syzx) )
          XMapRaised( XtDisplay(syzx->wtop) , XtWindow(syzx->wtop) ) ;
       EXRETURN ;
    } else if( w == pb_zxy && szxy != NULL ){
       if( ISQ_REALZ(szxy) )
          XMapRaised( XtDisplay(szxy->wtop) , XtWindow(szxy->wtop) ) ;
       EXRETURN ;
    } else if( w == gr_xyz && gxyz != NULL ){
       if( GRA_REALZ(gxyz) )
          XMapRaised( XtDisplay(gxyz->fdw_graph) , XtWindow(gxyz->fdw_graph) ) ;
       EXRETURN ;
    } else if( w == gr_yzx && gyzx != NULL ){
       if( GRA_REALZ(gyzx) )
          XMapRaised( XtDisplay(gyzx->fdw_graph) , XtWindow(gyzx->fdw_graph) ) ;
       EXRETURN ;
    } else if( w == gr_zxy && gzxy != NULL ){
       if( GRA_REALZ(gzxy) )
          XMapRaised( XtDisplay(gzxy->fdw_graph) , XtWindow(gzxy->fdw_graph) ) ;
       EXRETURN ;
    }

if(debug) printf("DIFF_dti_display_scalars_CB: GOT HERE!\n\a");

    /* button pressed and window not open, so prepare to open it */

    if( w == pb_xyz && sxyz == NULL ){
       snew  = &(im3d->s123) ;
       brnew = im3d->b123_ulay ;
       pboff = pb_xyz ;
       mirror= GLOBAL_argopt.left_is_left ;

    } else if( w == pb_yzx && syzx == NULL ){
       snew  = &(im3d->s231) ;
       brnew = im3d->b231_ulay ;
       pboff = pb_yzx ;

    } else if( w == pb_zxy && szxy == NULL ){
       snew  = &(im3d->s312) ;
       brnew = im3d->b312_ulay ;
       pboff = pb_zxy ;
       mirror= GLOBAL_argopt.left_is_left ;

    } else if( w == gr_xyz && gxyz == NULL ){
       gnew  = &(im3d->g123) ;
       brnew = im3d->b123_ulay ;
       pboff = gr_xyz ;
       mirror= GLOBAL_argopt.left_is_left ;

    } else if( w == gr_yzx && gyzx == NULL ){
       gnew  = &(im3d->g231) ;
       brnew = im3d->b231_ulay ;
       pboff = gr_yzx ;

    } else if( w == gr_zxy && gzxy == NULL ){
       gnew  = &(im3d->g312) ;
       brnew = im3d->b312_ulay ;
       pboff = gr_zxy ;
       mirror= GLOBAL_argopt.left_is_left ;

    } else {			/* lrf -  */
       snew  = &(im3d->s123) ;
       brnew = im3d->b123_ulay ;
       pboff = pb_xyz ;
       mirror= GLOBAL_argopt.left_is_left ;
    }
    /*  EXRETURN ; */ /* something funny */

    /** Mar 1997: don't open if x or y dimension is 1 **/

    if( brnew->n1 < 2 || brnew->n2 < 2 ) EXRETURN  ;

    SHOW_AFNI_PAUSE ;

    if( snew != NULL ){
      MCW_invert_widget(pboff) ;
      REPLICATE = 0;
      *snew = open_MCW_imseq(im3d->dc,
			     reinterpret_cast<void *(*)()>(DIF_brick_to_mri),
			     (XtPointer) brnew);

      (*snew)->parent = (XtPointer) im3d ;

      /* INIT_BKGD_LAB(im3d) ; */ /* lrf doesn't work! */

      drive_MCW_imseq( *snew, isqDR_imhelptext, (XtPointer) AFNI_image_help ) ;
      drive_MCW_imseq( *snew, isqDR_arrowpadon, (XtPointer) AFNI_arrowpad_help ) ;
      drive_MCW_imseq( *snew, isqDR_arrowpadhint , (XtPointer) AFNI_arrowpad_hint ) ;
      drive_MCW_imseq( *snew, isqDR_realize, NULL ) ;
      drive_MCW_imseq( *snew, isqDR_title, (XtPointer) dti_viewer_title );
      drive_MCW_imseq( *snew, isqDR_periodicmont, 0) ; /* enforce non-periodic
							  more - lrf */

      im3d->vinfo->crosshair_visible = 0;

      if(debug) printf("DIF_brick_to_mri: im3d->vinfo->xhairs_periodic =%d\n",
		       im3d->vinfo->xhairs_periodic);

      /* Montage size */

      { 
	double  sq_nsl = sqrt((double)dif->img->nsl) ;
	int    isq_nsl = (int) sq_nsl ;
	double   d_nsl = sq_nsl - isq_nsl ;

	mont_wsize = isq_nsl ;
	if(d_nsl<=0.5) mont_hsize = mont_wsize + 1;
	if(d_nsl> 0.5) mont_hsize = mont_wsize;
     
      }

      printf("\amontage size: mont_wsize = %i, mont_hsize = %i\n",
	     mont_wsize,mont_hsize) ;

      if(debug) printf("DIFF_dti_display_scalars_CB:im3d->vinfo->xhairs_all = %i\n",im3d->vinfo->xhairs_all);

      /* controller in mp[4] is 0 for no reason - lrf */

      mww = mont_wsize ; mhh = mont_hsize ; msp  = 1 ; mgap = 1 ;
      mp[0] = mww ; mp[1] = mhh ; mp[2] = msp ; mp[3] = mgap ;
      mp[4] = DC_find_overlay_color(GLOBAL_library.controllers[0]->dc,mcol);
      drive_MCW_imseq( *snew, isqDR_setmontage, (XtPointer) mp ) ;

STATUS("realizing new image viewer") ;

      drive_MCW_imseq( *snew, isqDR_realize, NULL ) ;

      /* 09 Oct 1998: force L-R mirroring on axial and coronal images? */

      if( mirror ){
         ISQ_options opt ;

STATUS("setting image view to be L-R mirrored") ;

         ISQ_DEFAULT_OPT(opt) ;
         opt.mirror = TRUE ;
         drive_MCW_imseq( *snew,isqDR_options  ,(XtPointer) &opt ) ;
      }

#ifdef USE_SIDES
#define LL 0
#define RR 1
#define AA 2
#define PP 3
#define SS 4
#define II 5
      if( !AFNI_yesenv("AFNI_NO_SIDES_LABELS") ){
         static char * ssix[6] = { "Left"     , "Right"     ,
                                   "Anterior" , "Posterior" ,
                                   "Superior" , "Inferior"   } ;
         char * ws[4] ;

         if( *snew == im3d->s123 ){
           ws[0] = ssix[RR]; ws[1] = ssix[AA]; ws[2] = ssix[LL]; ws[3] = ssix[PP];
         } else if( *snew == im3d->s231 ){
           ws[0] = ssix[AA]; ws[1] = ssix[SS]; ws[2] = ssix[PP]; ws[3] = ssix[II];
         } else if( *snew == im3d->s312 ){
           ws[0] = ssix[RR]; ws[1] = ssix[SS]; ws[2] = ssix[LL]; ws[3] = ssix[II];
         } else {
           ws[0] = ws[1] = ws[2] = ws[3] = NULL ;
         }

STATUS("setting image viewer 'sides'") ;

         drive_MCW_imseq( *snew,isqDR_winfosides,(XtPointer)ws ) ;

      }
#undef LL
#undef RR
#undef AA
#undef PP
#undef SS
#undef II
#endif

      AFNI_toggle_drawing( im3d ) ;

#ifndef DONT_INSTALL_ICONS
      if( afni48_good ){
         Pixmap pm = XmUNSPECIFIED_PIXMAP ;

              if( w == pb_xyz ) pm = afni48axi_pixmap ;
         else if( w == pb_yzx ) pm = afni48sag_pixmap ;
         else if( w == pb_zxy ) pm = afni48cor_pixmap ;

         drive_MCW_imseq( *snew,isqDR_icon , (XtPointer) pm ) ;
      }
#endif
    }

    /** Don't forget to send information like the reference timeseries ... **/

    if( gnew != NULL && DSET_GRAPHABLE(brnew->dset) ){
       MCW_grapher * gr ;

       MCW_invert_widget(pboff) ;
       REPLICATE = 0;
       gr = new_MCW_grapher(im3d->dc,
			    reinterpret_cast<void *(*)()>(DIF_brick_to_mri),
			    (XtPointer) brnew);
       drive_MCW_grapher( gr, graDR_title, (XtPointer) im3d->window_title );
       drive_MCW_grapher( gr, graDR_addref_ts, (XtPointer) im3d->fimdata->fimref );
       drive_MCW_grapher( gr, graDR_setignore, (XtPointer) im3d->fimdata->init_ignore );
       drive_MCW_grapher( gr, graDR_polort, (XtPointer) im3d->fimdata->polort );
       drive_MCW_grapher( gr, graDR_setindex , (XtPointer) im3d->vinfo->time_index );

       if( im3d->type == AFNI_IMAGES_VIEW )
          drive_MCW_grapher( gr , graDR_fim_disable , NULL ) ; /* 19 Oct 1999 */

       if( mirror )                                            /* 12 Jul 2000 */
          drive_MCW_grapher( gr , graDR_mirror , (XtPointer) 1 ) ;

STATUS("realizing new grapher") ;

       drive_MCW_grapher( gr , graDR_realize , NULL ) ;

       *gnew = gr ;
       (*gnew)->parent = (XtPointer) im3d ;

#ifndef DONT_INSTALL_ICONS
      if( afni48_good ){
         Pixmap pm = XmUNSPECIFIED_PIXMAP ;

              if( w == gr_xyz ) pm = afni48graaxi_pixmap ;
         else if( w == gr_yzx ) pm = afni48grasag_pixmap ;
         else if( w == gr_zxy ) pm = afni48gracor_pixmap ;

         drive_MCW_grapher( gr , graDR_icon , (XtPointer) pm ) ;
      }
#endif
    }

   /*-- force a jump to the viewpoint of the current location --*/

   AFNI_set_viewpoint( im3d , -1,-1,-1 , REDISPLAY_OVERLAY ) ;

   SHOW_AFNI_READY ;
   RESET_AFNI_QUIT(im3d) ;

   MPROBE ;
   EXRETURN ;
}

/*---------------------------------------------------------------------------
   DTI undisplay
-----------------------------------------------------------------------------*/

void DIFF_dti_undisplay_CB( Widget w ,
			    XtPointer client_data, 
			    XtPointer call_data )
{
  Three_D_View * im3d = (Three_D_View *) client_data ;
  MCW_imseq   ** sold = NULL ;

  printf("\aundisplay unavailable!\n");

}

/*-----------------------------------------------------------------------------

   compute SHD Image Sequence for SHD display

   montage of single image for SHD overlay.

-----------------------------------------------------------------------------*/
XtPointer compute_shd_imseq(int n, int type, FD_brick * br )  
{
  MRI_IMAGE * im ;
  MCW_imseq_status * stat ;
  int i1,j2,k3;
  int jb,bb , dd1,dd2,tt1,tt2 ;
  AFNI_view_info *vinfo = im3d->vinfo;
  int current_slice = vinfo->k3;
  SHD_view_state *vshd = current_dif_state->shd->view ;
  int debug = 0;

  /* printf("DIF_brick_to_mri: current coords = (%i,%i,%i) \n",
     vinfo->i1,vinfo->j2,vinfo->k3); */

  ENTRY("DIF_brick_to_mri") ;

  if(PRINT_TRACING){ char str[256] ; sprintf(str,"n=%d type=%d",n,type) ; STATUS(str) ; }

  /*-------------------------------------------------*/
  /*-------- May 1996: graph callbacks first --------*/
  if( type == graCR_getstatus ){
    MCW_grapher_status * grstat = myXtNew( MCW_grapher_status ) ;

    grstat->num_total  = grstat->num_series = br->dset->dblk->nvals ;
    grstat->nx         = br->n1 ;
    grstat->ny         = br->n2 ;
    grstat->nz         = br->n3 ;

    grstat->send_CB    = reinterpret_cast<void (*)()>(AFNI_gra_send_CB);
    grstat->parent     = (XtPointer) br ;
    grstat->aux        = NULL ;

    grstat->transforms0D = & (GLOBAL_library.registered_0D) ;
    grstat->transforms1D = & (GLOBAL_library.registered_1D) ;

    strcpy( grstat->namecode , br->namecode ) ;

    RETURN( (XtPointer) grstat ) ;
  }

  if( type == graCR_getseries ){
    RETURN( (XtPointer) im ) ;
  }

  /*----------------------------------------*/
  /*-------- Now do imseq callbacks --------*/

  if( n < 0 || n >= br->n3 ) RETURN(NULL) ;

  /*--- overlay # n ---*/

  if( type == isqCR_getoverlay  ){

    STATUS("get overlay") ;

    if(debug) 
      printf("DIF_brick_to_mri: overlay n=%d, view (nl,nm) = (%i,%i)\n",
	     n,vshd->nl,vshd->nm) ;

#if 0
    if( shd_img_count_ulay == 11 || (shd_img_count_ulay > 20 && shd_img_count_ulay < 24) || 
	( shd_img_count_ulay > 30 &&   shd_img_count_ulay < 36 ) || 
	( shd_img_count_ulay > 40 &&   shd_img_count_ulay < 48 ) ||
	( shd_img_count_ulay > 50 &&   shd_img_count_ulay < 60 ) || 
	( shd_img_count_ulay > 60 &&   shd_img_count_ulay < 66 )
        ) {
      n = -1;
    }
    else {
      n = current_slice;
    }
#else
    if( shd_img_count_ulay == 9 || (shd_img_count_ulay > 16 && shd_img_count_ulay < 20) || 
	( shd_img_count_ulay > 24 &&   shd_img_count_ulay < 30 ) || 
	( shd_img_count_ulay > 32 &&   shd_img_count_ulay < 40 ) ||
	( shd_img_count_ulay > 40 ) ) { 

    /*    if( shd_img_count_ulay != 40 ) { */
      n = -1;
    }
    else {
      n = current_slice;
    }
#endif
    shd_img_count_ulay++;
#if 0
    if( shd_img_count_ulay == 66 )
#else
    if( shd_img_count_ulay == 45 )
#endif
      shd_img_count_ulay = 0;

    int l = 0, m = 0;
    /* compute which overlay we are drawing */
    int in = shd_img_count_ulay-1;
#if 0
    int shd_row = in/11;
    l = 5 - shd_row;
    int shd_col = in % 11;
    m = -5 + shd_col;
#else
    int shd_row = in/9;
    l = 4 - shd_row;
    int shd_col = in % 9;
    m = -4 + shd_col;
#endif

    if( n == -1 )
      im = NULL;
    else
      im = SHD_overlay( l, m, n, br );

    /*im = AFNI_overlay( n , br ) ;*/

    /* write new Get SHD overlay code here */

    RETURN( (XtPointer) im ) ;
  }

  /*--- status ---*/

  if( type == isqCR_getstatus ){

    STATUS("get status") ;

    stat = myXtNew( MCW_imseq_status ) ;

    stat->num_total  = br->n3 ;
    stat->num_series = br->n3 ;
    stat->send_CB    = reinterpret_cast<void (*)()>(AFNI_seq_send_CB);
    stat->parent     = (XtPointer) br ;
    stat->aux        = NULL ;

    stat->transforms0D = & (GLOBAL_library.registered_0D) ;
    stat->transforms2D = & (GLOBAL_library.registered_2D) ;

    RETURN( (XtPointer) stat ) ;
  }

  /*--- 26 Feb 2001: return a memplot drawing struct ---*/

  if( type == isqCR_getmemplot ){
    RETURN( (XtPointer) NULL ) ;
  }

  /*--- underlay image # n ---*/

  if( type == isqCR_getimage || type == isqCR_getqimage ){
    Three_D_View * im3d = (Three_D_View *) br->parent ;
    int ival = 0 ;

    /*** decide which 3D brick to extract data from (ival) ***/

    if( DSET_NUM_TIMES(br->dset) > 1 ){ /* a time-dependent brick */
      ival = im3d->vinfo->time_index ;
      if( ival >= br->dset->dblk->nvals ) ival = br->dset->dblk->nvals -1 ;

    } else if( ISANAT(br->dset) ){      /* an anatomy brick */
      if( ISANATBUCKET(br->dset) ) {     /* 30 Nov 1997 */
	ival = im3d->vinfo->anat_index ;
      }
      else {
	ival = ANAT_ival_zero[br->dset->func_type] ;  /* get default data */
      }
    } else if( ISFUNC(br->dset) ){      /* a functional brick */

      if( ISFUNCBUCKET(br->dset) ){     /* 30 Nov 1997 */
	ival = im3d->vinfo->fim_index ;
      }
      else {
	if( /*im3d->vinfo->showfunc_type == SHOWFUNC_THR  &&*/  /* showing thr */
	   ISFUNC_UNDERLAY(im3d->vinfo->underlay_type) &&  /* as underlay? */
	   FUNC_HAVE_THR(br->dset->func_type) ){

	  ival = FUNC_ival_thr[br->dset->func_type] ; /* get thresh data */
	} else {
	  ival = FUNC_ival_fim[br->dset->func_type] ; /* get fim data */
	}
      }
    }

    if( type == isqCR_getqimage ) ival = -1 ; /* get empty image */

    if(PRINT_TRACING)
      { char str[256] ;
      sprintf(str,"getting image n1=%d n2=%d ival=%d",br->n1,br->n2,ival) ;
      STATUS(str) ; }

    LOAD_DSET_VIEWS(im3d) ;  /* 02 Nov 1996 */

    if( ival > 0 && GLOBAL_library.have_dummy_dataset && lrand48()%2 == 1 ){
      RETURN(NULL) ;  /* 23 Apr 2001: test of imseq.c NULL image handler */
    }

    /* Replicate current slice in underlay - lrf*/

    if(REPLICATE) {
      n = current_slice ;	
      if(debug) printf("DIF_brick_to_mri: replicate = %i (n=%i)\n",REPLICATE,n);
    }

#if 0
    if( shd_img_count == 12 || (shd_img_count > 21 && shd_img_count < 25) || 
	( shd_img_count > 31 && shd_img_count < 37 ) || (shd_img_count > 41  &&  shd_img_count < 49 ) ||
	( shd_img_count > 51 && shd_img_count < 61 ) || (shd_img_count > 61 && shd_img_count < 67 )
        )
#else
    if( shd_img_count == 10 || (shd_img_count > 17 && shd_img_count < 21) || 
	( shd_img_count > 25 &&   shd_img_count < 31 ) || 
	( shd_img_count > 33 &&   shd_img_count < 41 ) ||
	( shd_img_count > 41 ) )
#endif
      ival = -1;
    shd_img_count++;
#if 0
    if( shd_img_count == 67 )
#else
    if( shd_img_count == 46 )
#endif
      shd_img_count = 1;
      
    im = FD_warp_to_mri( n , ival , br ) ; /* actually get image from dataset */

    if( ival < 0 ) RETURN( (XtPointer) im ) ;  /* return fake image */

    /* Allow for thresholding of underlay image */

    if( im != NULL && im3d->vinfo->underlay_type == UNDERLAY_THRFUNC &&
	ISFUNC(br->dset) && FUNC_HAVE_THR(br->dset->func_type) &&
	im3d->vinfo->func_threshold > 0.0 ){

      MRI_IMAGE * im_thr , * qim ;
      double thresh ;
      int jval ;

      if( ISFUNCBUCKET(br->dset) )          /* 30 Nov 1997 */
	jval = im3d->vinfo->thr_index ;
      else
	jval = FUNC_ival_thr[br->dset->func_type] ;

      if( jval != ival ){
	im_thr = FD_warp_to_mri( n , jval ,  br ) ;
	if( im_thr == NULL ) im_thr = im ;
      } else {
	im_thr = im ;  /* if already have threshold data */
      }

      if( ! AFNI_GOOD_FUNC_DTYPE(im_thr->kind) ){ /* should not occur */
	qim = mri_to_float( im_thr ) ;           /* but if it does,  */
	if( im_thr != im ) mri_free( im_thr ) ;  /* make an OK type  */
	im_thr = qim ;
      }

      thresh = im3d->vinfo->func_threshold * im3d->vinfo->func_thresh_top ;

      if( im_thr->kind == MRI_short )
	thresh *= FUNC_scale_short[br->dset->func_type];
      else if( im_thr->kind == MRI_byte )
	thresh *= FUNC_scale_byte[br->dset->func_type];

      mri_threshold( -thresh , thresh , im_thr , im ) ;
      if( im_thr != im ) mri_free( im_thr ) ;
    }

    /* Load value of current pixel into display label */
    /* April 1996: only if image is at current slice  */

#ifdef ALLOW_BKGD_LAB
    { char buf[32] = "\0" ;
    AFNI_set_valabel( br , n , im , buf ) ;
    if( buf[0] != '\0' ){
      if( im3d->vinfo->underlay_type == UNDERLAY_ANAT )
	strcpy( im3d->vinfo->anat_val , buf ) ;
      else
	im3d->vinfo->anat_val[0] = '\0' ;

      if( im->kind != MRI_complex ){
	char qbuf[32] = "bg =" ;
	strcat(qbuf,buf) ; strcpy(buf,qbuf) ;
      }
      MCW_set_widget_label( im3d->vwid->imag->pop_bkgd_lab , buf ) ;
      XtManageChild( im3d->vwid->imag->pop_bkgd_lab ) ;
    }
    }
#endif

    RETURN( (XtPointer) im ) ;
  }

  STATUS("get something else, but I don't care!") ;
}


MRI_IMAGE * SHD_overlay( int l,int m,int  n, FD_brick * br )
{
  MRI_IMAGE * im = NULL;
  register short * oar;
  int debug = 1;

  if(debug)
    printf("Calling SHD_overlay with param l=%d, m=%d, n=%d\n", l,m,n );

  STATUS("new overlay is created");
  im = mri_new( br->n1, br->n2, MRI_short );
  oar = MRI_SHORT_PTR(im);

  /* coordinate bounds */
  int nx,ny,nz;
   
  nx = DSET_NX(dset);
  ny = DSET_NY(dset);
  nz = DSET_NZ(dset);

  if(debug)
    printf("coord bounds: nx=%d, ny=%d, nz=%d\n", nx,ny,nz );

  if( func_dset != NULL ) {
    DSET_load( func_dset ); 
  }

  int ndif = DSET_NVALS( func_dset );
  int typeDset = DSET_BRICK_TYPE( func_dset, 0 );
  if( typeDset != MRI_complex )
    fprintf(stderr , "Debug: check to make sure we are looking at complex data\n");
                                                                                                               
  /* allocate memory */
  complex** cptr = NULL;
  complex*  cxar = NULL;
  complex*  cval  = NULL;
  int kk,ii, nbrik;

  cptr = (complex**) malloc (sizeof (complex*) * ndif );
  if( cptr == NULL )
    fprintf(stderr, "Debug: (cptr == NULL) We have a problem, outcome uncertain\n");

  cxar = (complex*) malloc (sizeof (complex) * ndif );
  if( cxar == NULL )
    fprintf(stderr, "Debug: (cxar == NULL) We have a problem, outcome uncertain\n");

  /* point to the dataset
     voxel #i at time #k is fptr[k][i];
     for i =0 to nvox-1
     k =0 to ndif -1 */
  for( kk = 0; kk < ndif; kk++ ) {
    cptr[kk] = (complex*) DSET_ARRAY(func_dset, kk);
  }

  /* for now, just print out data
     if( !current_dif_state->sta->shd_initialized ) init_shd( current_dif_state );
     HACK FOR NOW,  seems correct but need to look into what is actually
     happening here */
  current_dif_state->shd->calc->Alm = alm_alloc( current_dif_state->shd->calc->Lmax ) ;

  int jj;
  /*  FILE *slice_file;
      slice_file = fopen("slice_overlay.dat", "w"); */
  for( jj = 0; jj < nx*ny; jj++ )
    {
      /* now load data for correct voxel (int ii =vox) we are interested in */
      ii = jj + (nx*ny*(nz-n-1));
      for( kk =0; kk < ndif; kk++ ) {
	cxar[kk] = cptr[kk][ii];
      }

      zvtoAlm( current_dif_state->shd->calc->Alm, cxar,
               current_dif_state->shd->calc->Lmax );

      /* grab data for n=0, l=0 for currentslice.
	 printf("Alm for n =0, n =1: (%13.6g, %13.6g)\n", current_dif_state->shd->calc->Alm[n][l].r, 
	 current_dif_state->shd->calc->Alm[n][l].i );

	 compute complex ABSolute value */
      double comp_abs = zabs(zcplx(current_dif_state->shd->calc->Alm[l][m].r, 
				   current_dif_state->shd->calc->Alm[l][m].i )  );
      if (comp_abs > max_magnitude) {
	max_magnitude = comp_abs;
      }
      if(debug > 2)
	printf("Complex ABS value: %lf\n", comp_abs);

      // GOOD GRIEF!!! this is hardcoded!
      //      oar[jj] = (short int) (19*comp_abs);
      // changing to another hardcoded value!!
      if (comp_abs > 0.20 * max_magnitude) {
	oar[jj] = (short int) (17.0*comp_abs/max_magnitude + 24);
      }
      else {
	oar[jj] = 0;
      }

      /* oar[jj] = 16*(jj%128)/128 + 25; */
      /*      fprintf(slice_file, "%d\n", oar[jj]); */

      if(debug > 2 && oar[jj] != 0)
	printf("Complex ABS value: %lf; Short value: %d\n", comp_abs, oar[jj]);
    }

  /*  fclose(slice_file); */
  if (debug) {
    printf("maximum complex abs value: %lf\n", max_magnitude);
  }

  RETURN( im );
}
/*-----------------------------------------------------------------------------

   SHD display

   Display montage of single image for SHD overlay.

-----------------------------------------------------------------------------*/
/* cyrus montage */
void DIFF_shd_display_CB( Widget w , XtPointer client_data , XtPointer call_data )
{
   Three_D_View * im3d = (Three_D_View *) client_data ;
   DIF_state *dif = current_dif_state ;

   MCW_imseq   * sxyz , * syzx , * szxy , ** snew = NULL ;
   MCW_grapher * gxyz , * gyzx , * gzxy , ** gnew = NULL ;
   Widget        pboff , pb_xyz , pb_yzx , pb_zxy ;
   Widget        groff , gr_xyz , gr_yzx , gr_zxy ;
   FD_brick    * brnew ;
   int mirror=0 ;

   char *shd_viewer_title = "SHD Viewer: In Developement" ;
   int  mp[5] ;			/* montage params */
   char mcol[128] = "\0" ;
   int  mww, mhh, msp, mgap ;

   int mont_wsize, mont_hsize ;	/* x and y size of montage array */

ENTRY("DIFF_shd_display_CB") ;

   if(dif!=NULL) { 
     printf("\aDIFF_shd_display_CB: dif->img->nx = %i\n",dif->img->nx);
   }

   if( ! IM3D_VALID(im3d) ) {
     printf("\aDIFF_shd_display_CB: Invalid im3d - returning \n");
     EXRETURN ;
   }

    sxyz = im3d->s123 ; gxyz = im3d->g123 ;
    syzx = im3d->s231 ; gyzx = im3d->g231 ;
    szxy = im3d->s312 ; gzxy = im3d->g312 ;

    pb_xyz = im3d->vwid->imag->image_xyz_pb ;
    pb_yzx = im3d->vwid->imag->image_yzx_pb ;
    pb_zxy = im3d->vwid->imag->image_zxy_pb ;

    gr_xyz = im3d->vwid->imag->graph_xyz_pb ;
    gr_yzx = im3d->vwid->imag->graph_yzx_pb ;
    gr_zxy = im3d->vwid->imag->graph_zxy_pb ;

    /* handle case of button press of already
       open window by bringing that window to the top */

    if( w == pb_xyz && sxyz != NULL ){
       if( ISQ_REALZ(sxyz) )
          XMapRaised( XtDisplay(sxyz->wtop) , XtWindow(sxyz->wtop) ) ;
       EXRETURN ;
    } else if( w == pb_yzx && syzx != NULL ){
       if( ISQ_REALZ(syzx) )
          XMapRaised( XtDisplay(syzx->wtop) , XtWindow(syzx->wtop) ) ;
       EXRETURN ;
    } else if( w == pb_zxy && szxy != NULL ){
       if( ISQ_REALZ(szxy) )
          XMapRaised( XtDisplay(szxy->wtop) , XtWindow(szxy->wtop) ) ;
       EXRETURN ;
    } else if( w == gr_xyz && gxyz != NULL ){
       if( GRA_REALZ(gxyz) )
          XMapRaised( XtDisplay(gxyz->fdw_graph) , XtWindow(gxyz->fdw_graph) ) ;
       EXRETURN ;
    } else if( w == gr_yzx && gyzx != NULL ){
       if( GRA_REALZ(gyzx) )
          XMapRaised( XtDisplay(gyzx->fdw_graph) , XtWindow(gyzx->fdw_graph) ) ;
       EXRETURN ;
    } else if( w == gr_zxy && gzxy != NULL ){
       if( GRA_REALZ(gzxy) )
          XMapRaised( XtDisplay(gzxy->fdw_graph) , XtWindow(gzxy->fdw_graph) ) ;
       EXRETURN ;
    }

    /* button pressed and window not open, so prepare to open it */

    if( w == pb_xyz && sxyz == NULL ){
       snew  = &(im3d->s123) ;
       brnew = im3d->b123_ulay ;
       pboff = pb_xyz ;
       mirror= GLOBAL_argopt.left_is_left ;

    } else if( w == pb_yzx && syzx == NULL ){
       snew  = &(im3d->s231) ;
       brnew = im3d->b231_ulay ;
       pboff = pb_yzx ;

    } else if( w == pb_zxy && szxy == NULL ){
       snew  = &(im3d->s312) ;
       brnew = im3d->b312_ulay ;
       pboff = pb_zxy ;
       mirror= GLOBAL_argopt.left_is_left ;

    } else if( w == gr_xyz && gxyz == NULL ){
       gnew  = &(im3d->g123) ;
       brnew = im3d->b123_ulay ;
       pboff = gr_xyz ;
       mirror= GLOBAL_argopt.left_is_left ;

    } else if( w == gr_yzx && gyzx == NULL ){
       gnew  = &(im3d->g231) ;
       brnew = im3d->b231_ulay ;
       pboff = gr_yzx ;

    } else if( w == gr_zxy && gzxy == NULL ){
       gnew  = &(im3d->g312) ;
       brnew = im3d->b312_ulay ;
       pboff = gr_zxy ;
       mirror= GLOBAL_argopt.left_is_left ;

    } else {			/* lrf -  */
       snew  = &(im3d->s123) ;
       brnew = im3d->b123_ulay ;
       pboff = pb_xyz ;
       mirror= GLOBAL_argopt.left_is_left ;
    }
    /*  EXRETURN ; */ /* something funny */

    /** Mar 1997: don't open if x or y dimension is 1 **/

    if( brnew->n1 < 2 || brnew->n2 < 2 ) EXRETURN  ;

    SHOW_AFNI_PAUSE ;

    if(  snew != NULL ){
      MCW_invert_widget(pboff) ;
      REPLICATE = 1;
      shd_img_count = 0;
      shd_img_count_ulay = 0;
      *snew = open_MCW_imseq(im3d->dc,
			     reinterpret_cast<void *(*)()>(compute_shd_imseq),
			     (XtPointer) brnew );

      (*snew)->parent = (XtPointer) im3d ;

      /* INIT_BKGD_LAB(im3d) ; */ /* lrf doesn't work! */

      drive_MCW_imseq( *snew, isqDR_imhelptext,    (XtPointer) AFNI_image_help ) ;
      drive_MCW_imseq( *snew, isqDR_arrowpadoff,    (XtPointer) AFNI_arrowpad_help ) ;
      drive_MCW_imseq( *snew, isqDR_arrowpadhint , (XtPointer) AFNI_arrowpad_hint ) ;
      drive_MCW_imseq( *snew, isqDR_realize, NULL ) ;
      drive_MCW_imseq( *snew, isqDR_title, (XtPointer) shd_viewer_title );
      drive_MCW_imseq( *snew, isqDR_periodicmont, (XtPointer)1 );
      drive_MCW_imseq( *snew, isqDR_button2_disable, NULL); 
      drive_MCW_imseq( *snew, isqDR_zoombut, 0); 
      drive_MCW_imseq( *snew, isqDR_penbbox, 0); 
      drive_MCW_imseq( *snew, isqDR_plot_plot, (XtPointer)1); 
      drive_MCW_imseq( *snew, isqDR_ignore_redraws, (XtPointer)1); 

      /* Montage size */

      /* wshdfunc_shd_mval_av */
      /* wshdfunc_shd_lval_av */

      mont_wsize = dif->shd->view->nm ;
      mont_hsize = dif->shd->view->nl ;

      printf("\amontage size: mont_wsize = %i, mont_hsize = %i\n",
	     mont_wsize,mont_hsize - 1) ;

      /* controller in mp[4] is 0 for no reason - lrf */

      mww = mont_wsize ; mhh = mont_hsize-1 ; msp  = 1 ; mgap = 1 ;
      mp[0] = mww ; mp[1] = mhh ; mp[2] = msp ; mp[3] = mgap ;
      mp[4] = DC_find_overlay_color(GLOBAL_library.controllers[0]->dc,mcol);
      drive_MCW_imseq( *snew, isqDR_setmontage, (XtPointer) mp ) ;

STATUS("realizing new image viewer") ;

      drive_MCW_imseq( *snew, isqDR_realize, NULL ) ;

      /* 09 Oct 1998: force L-R mirroring on axial and coronal images? */

      if( mirror ){
         ISQ_options opt ;

STATUS("setting image view to be L-R mirrored") ;

         ISQ_DEFAULT_OPT(opt) ;
         opt.mirror = TRUE ;
         drive_MCW_imseq( *snew,isqDR_options  ,(XtPointer) &opt ) ;
      }

#ifdef USE_SIDES
#define LL 0
#define RR 1
#define AA 2
#define PP 3
#define SS 4
#define II 5
      if( !AFNI_yesenv("AFNI_NO_SIDES_LABELS") ){
         static char * ssix[6] = { "Left"     , "Right"     ,
                                   "Anterior" , "Posterior" ,
                                   "Superior" , "Inferior"   } ;
         char * ws[4] ;

         if( *snew == im3d->s123 ){
           ws[0] = ssix[RR]; ws[1] = ssix[AA]; ws[2] = ssix[LL]; ws[3] = ssix[PP];
         } else if( *snew == im3d->s231 ){
           ws[0] = ssix[AA]; ws[1] = ssix[SS]; ws[2] = ssix[PP]; ws[3] = ssix[II];
         } else if( *snew == im3d->s312 ){
           ws[0] = ssix[RR]; ws[1] = ssix[SS]; ws[2] = ssix[LL]; ws[3] = ssix[II];
         } else {
           ws[0] = ws[1] = ws[2] = ws[3] = NULL ;
         }

STATUS("setting image viewer 'sides'") ;

         drive_MCW_imseq( *snew,isqDR_winfosides,(XtPointer)ws ) ;

      }
#undef LL
#undef RR
#undef AA
#undef PP
#undef SS
#undef II
#endif

      AFNI_toggle_drawing( im3d ) ;

#ifndef DONT_INSTALL_ICONS
      if( afni48_good ){
         Pixmap pm = XmUNSPECIFIED_PIXMAP ;

              if( w == pb_xyz ) pm = afni48axi_pixmap ;
         else if( w == pb_yzx ) pm = afni48sag_pixmap ;
         else if( w == pb_zxy ) pm = afni48cor_pixmap ;

         drive_MCW_imseq( *snew,isqDR_icon , (XtPointer) pm ) ;
      }
#endif
    }

    /** Don't forget to send information like the reference timeseries ... **/

    if( gnew != NULL && DSET_GRAPHABLE(brnew->dset) ){
       MCW_grapher * gr ;

       MCW_invert_widget(pboff) ;
       REPLICATE = 1;
       gr = new_MCW_grapher( im3d->dc ,
			     reinterpret_cast<void *(*)()>(DIF_brick_to_mri),
			     (XtPointer) brnew );
       drive_MCW_grapher( gr, graDR_title, (XtPointer) im3d->window_title );
       drive_MCW_grapher( gr, graDR_addref_ts, (XtPointer) im3d->fimdata->fimref );
       drive_MCW_grapher( gr, graDR_setignore, (XtPointer) im3d->fimdata->init_ignore );
       drive_MCW_grapher( gr, graDR_polort, (XtPointer) im3d->fimdata->polort );
       drive_MCW_grapher( gr, graDR_setindex , (XtPointer) im3d->vinfo->time_index );

       if( im3d->type == AFNI_IMAGES_VIEW )
          drive_MCW_grapher( gr , graDR_fim_disable , NULL ) ; /* 19 Oct 1999 */

       if( mirror )                                            /* 12 Jul 2000 */
          drive_MCW_grapher( gr , graDR_mirror , (XtPointer) 1 ) ;

STATUS("realizing new grapher") ;

       drive_MCW_grapher( gr , graDR_realize , NULL ) ;

       *gnew = gr ;
       (*gnew)->parent = (XtPointer) im3d ;

#ifndef DONT_INSTALL_ICONS
      if( afni48_good ){
         Pixmap pm = XmUNSPECIFIED_PIXMAP ;

              if( w == gr_xyz ) pm = afni48graaxi_pixmap ;
         else if( w == gr_yzx ) pm = afni48grasag_pixmap ;
         else if( w == gr_zxy ) pm = afni48gracor_pixmap ;

         drive_MCW_grapher( gr , graDR_icon , (XtPointer) pm ) ;
      }
#endif
    }

   /*-- force a jump to the viewpoint of the current location --*/

   AFNI_set_viewpoint( im3d , -1,-1,-1 , REDISPLAY_OVERLAY ) ;

   SHOW_AFNI_READY ;
   RESET_AFNI_QUIT(im3d) ;

   MPROBE ;
   EXRETURN ;
}

/*------------------------------------------------------------------------------
   SHD undisplay
--------------------------------------------------------------------------------*/

void DIFF_shd_undisplay_CB( Widget w ,
			    XtPointer client_data, 
			    XtPointer call_data )
{
  Three_D_View * im3d = (Three_D_View *) client_data ;
  MCW_imseq   ** sold = NULL ;

  printf("\aundisplay unavailable!\n");
  /*  sold  = &(im3d->s123) ;
      drive_MCW_imseq( sold , isqDR_unrealize, NULL ) ; */
}

void DIFF_ftm_lval_av_CB( MCW_arrowval * av , XtPointer cd )
{
   int cc ;

   INVALIDATE_OVERLAY ;

   /* If L is changed, assume we want M = L, as it 
      should be in almost all cases.
      (if not, then you have to adjust M yourself! */

   AV_assign_ival(ftm_mval_av,av->ival);

   /* If L falls below M, adjust M down */

   if(av->ival < ftm_mval_av->ival) {
     AV_assign_ival(ftm_mval_av,av->ival);
   }

   return ;
}

void DIFF_ftm_mval_av_CB( MCW_arrowval * av , XtPointer cd )
{
   int cc ;

   INVALIDATE_OVERLAY ;

   /* If M goes above L, adjust M down to L */

   if(av->ival > ftm_lval_av->ival) {
     AV_assign_ival(av,ftm_lval_av->ival);
   }

   return ;
}

void DIFF_ftm_grid_av_CB( MCW_arrowval * av , XtPointer cd )
{
   int cc ;

   INVALIDATE_OVERLAY ;

   return ;
}


/*-------------------------------------------------------------------------------
  Event handler to find #3 button press for pbar popup, and popup the menu
---------------------------------------------------------------------------------*/

void DIFF_pbarmenu_EV( Widget w , XtPointer cd ,
                       XEvent * ev , Boolean * continue_to_dispatch )
{
   static int old_paltab_num = 0 ;

   switch( ev->type ){
      case ButtonPress:{
         XButtonEvent * event = (XButtonEvent *) ev ;
         if( event->button == Button3 ){

            /* in case the user read in any new palette, add them to menu */

            if( GPT != NULL && PALTAB_NUM(GPT) > old_paltab_num ){
               refit_MCW_optmenu( wshdfunc_pbar_palette_av ,
				  0 ,                     /* new minval */
				  PALTAB_NUM(GPT)-1 ,     /* new maxval */
				  0 ,                     /* new inival */
				  0 ,                     /* new decim? */
				  reinterpret_cast<char *(*)()>
				  (AFNI_palette_label_CB) , /* text routine */
				  NULL                    /* text data */
                                 ) ;
               XtManageChild( wshdfunc_pbar_palette_av->wrowcol ) ;
               old_paltab_num = PALTAB_NUM(GPT) ;
            }

            XmMenuPosition( wshdfunc_pbar_menu , event ) ; /* where */
            XtManageChild ( wshdfunc_pbar_menu ) ;         /* popup */
         }
      }
      break ;
   }
   return ;
}

/*--------------------------------------------------------------------------------
  Callbacks for all actions in the pbar popup
----------------------------------------------------------------------------------*/

void DIFF_pbarmenu_CB( Widget w , XtPointer cd , XtPointer cbs )
{
   MCW_pbar * pbar ;
   int npane , jm , ii ;
   double pmax , pmin ;
   float pval[NPANE_MAX+1] ;

   pbar  = wshdfunc_color_pbar ;
   npane = pbar->num_panes ;
   jm    = pbar->mode ;
   pmax  = pbar->pval_save[npane][0][jm] ;
   pmin  = pbar->pval_save[npane][npane][jm] ;

   /*--- Equalize spacings ---*/

   if( w == wshdfunc_pbar_equalize_pb ){
      for( ii=0 ; ii <= npane ; ii++ )
         pval[ii] = pmax - ii * (pmax-pmin)/npane ;

      HIDE_SCALE ;
      alter_MCW_pbar( pbar , 0 , pval ) ;
      FIX_SCALE_SIZE ;
      INVALIDATE_OVERLAY ;
   }

   /*--- Set top value ---*/

   else if( w == wshdfunc_pbar_settop_pb ){
      MCW_choose_integer( wshdfunc_choices_label ,
                          "Pbar Top" , 0 , 99999 , 1 ,
                          reinterpret_cast<void (*)()>(DIFF_set_pbar_top_CB), NULL  ) ;
   }

   return ;
}

void DIFF_palette_av_CB( MCW_arrowval * av , XtPointer cd )
{
   if( GPT == NULL || av->ival < 0 || av->ival >= PALTAB_NUM(GPT) ) return ;

   HIDE_SCALE ;
   load_PBAR_palette_array( wshdfunc_color_pbar ,             /* cf. afni_setup.c */
                            PALTAB_ARR(GPT,av->ival) , 0 ) ;
   FIX_SCALE_SIZE ;

   INVALIDATE_OVERLAY ;
   return ;
}

void DIFF_mixshade_av_CB( MCW_arrowval * av , XtPointer cd )  /* 21 Dec 1999 */
{
   func_mixshade = av->ival ;
   return ;
}

void DIFF_set_pbar_top_CB( Widget w , XtPointer cd , MCW_choose_cbs * cbs )
{
   MCW_pbar * pbar ;
   float pval[NPANE_MAX+1] ;
   double pmax , fac ;
   int ii ;

   if( ! diffuser_open ){ POPDOWN_integer_chooser; XBell(dc->display,100); return; }

   pmax = cbs->fval ; if( pmax <= 0.0 ) return ;           /* illegal */
   pbar = wshdfunc_color_pbar ;
   fac  = pmax / pbar->pval[0] ; if( fac == 1.0 ) return ; /* no change */

   for( ii=0 ; ii <= pbar->num_panes ; ii++ )
      pval[ii] = fac * pbar->pval[ii] ;

   HIDE_SCALE ;
   alter_MCW_pbar( pbar , 0 , pval ) ;
   FIX_SCALE_SIZE ;

   INVALIDATE_OVERLAY ;
   return ;
}

/*----------------------------------------------------------------------------
   Load color index data from the functional dataset into the local overlay array
   (heavily adapted from AFNI_func_overlay in afni_func.c).
------------------------------------------------------------------------------*/

void DIFF_reload_func_dset(Three_D_View * im3d, int id)
{
   MRI_IMAGE * cim , * tim ;
   void      * car , * tar ;
   float      cfac ,  tfac ;
   int ii , nvox , num_lp , lp ;
   byte * ovar ;
   MCW_pbar * pbar = wshdfunc_color_pbar ;
   byte fim_ovc[NPANE_MAX+1] ;
   float fim_thr[NPANE_MAX] , scale_factor , thresh ;
   THD_session * sess ;
   int iss;
   int debug = 0;

   if( func_dset == NULL ) return ;  /* error */

   if(debug) printf("DIFF_reload_func_dset: DSET_PREFIX(func_dset) = %s\n",
		    DSET_PREFIX(func_dset));

   INVALIDATE_OVERLAY ;              /* toss old overlay, if any */

   DSET_load(func_dset) ;            /* make sure is in memory */
   

   cim  = DSET_BRICK(func_dset,func_color_ival) ; nvox = cim->nvox ; /* color brick */
   car  = DSET_ARRAY(func_dset,func_color_ival) ;
   cfac = DSET_BRICK_FACTOR(func_dset,func_color_ival) ;
   if( cfac == 0.0 ) cfac = 1.0 ;

   tim  = DSET_BRICK(func_dset,func_thresh_ival) ;                   /* thresh brick */
   tar  = DSET_ARRAY(func_dset,func_thresh_ival) ;
   tfac = DSET_BRICK_FACTOR(func_dset,func_thresh_ival) ;
   if( tfac == 0.0 ) tfac = 1.0 ;

   ovim = mri_new_conforming( cim , MRI_byte ) ;                     /* new overlay */
   ovar = MRI_BYTE_PTR(ovim) ;

   scale_factor = func_range ;                                       /* for map from */
   if( scale_factor == 0.0 || func_use_autorange )                   /* pbar to data */
      scale_factor = func_autorange ;                                /* value range  */

   num_lp = pbar->num_panes ;
   for( lp=0 ; lp < num_lp ; lp++ ) fim_ovc[lp] = pbar->ov_index[lp] ; /* top to bottom */
   fim_ovc[num_lp] = (func_posfunc) ? (0) : (fim_ovc[num_lp-1]) ;      /* off the bottom */

   thresh = func_threshold * func_thresh_top / tfac ;  /* threshold in tar[] scale */

   /*--- Load the overlay image with the color overlay index ---*/

   if( thresh < 1.0 || !func_use_thresh ){  /*--- no thresholding needed ---*/

      switch( cim->kind ){

         case MRI_short:{
            short * sar = (short *) car ;

            for( lp=0 ; lp < num_lp ; lp++ )                          /* pbar in     */
               fim_thr[lp] = scale_factor * pbar->pval[lp+1] / cfac ; /* car[] scale */

            for( ii=0 ; ii < nvox ; ii++ ){
               if( sar[ii] == 0 ){
                  ovar[ii] = 0 ;
               } else {
                  for( lp=0 ; lp < num_lp && sar[ii] < fim_thr[lp] ; lp++ ) ; /*nada*/
                  ovar[ii] = fim_ovc[lp] ;
               }
            }
         }
         break ;

         case MRI_float:{
            float * sar = (float *) car ;

            for( lp=0 ; lp < num_lp ; lp++ )
               fim_thr[lp] = scale_factor * pbar->pval[lp+1] / cfac ;

            for( ii=0 ; ii < nvox ; ii++ ){
               if( sar[ii] == 0.0 ){
                  ovar[ii] = 0 ;
               } else {
                  for( lp=0 ; lp < num_lp && sar[ii] < fim_thr[lp] ; lp++ ) ; /*nada*/
                  ovar[ii] = fim_ovc[lp] ;
               }
            }
         }
         break ;

         case MRI_byte:{
            byte * sar = (byte *) car ;

            for( lp=0 ; lp < num_lp ; lp++ )
               if( pbar->pval[lp+1] <= 0.0 )
                  fim_thr[lp] = 0 ;
               else
                  fim_thr[lp] = scale_factor * pbar->pval[lp+1] / cfac ;

            for( ii=0 ; ii < nvox ; ii++ ){
               if( sar[ii] == 0 ){
                  ovar[ii] = 0 ;
               } else {
                  for( lp=0 ; lp < num_lp && sar[ii] < fim_thr[lp] ; lp++ ) ; /*nada*/
                  ovar[ii] = fim_ovc[lp] ;
               }
            }
         }
         break ;
      }             /*--- end of no thresholding case ---*/

   } else {         /*--- start of thresholding ---*/

      switch( cim->kind ){

         case MRI_short:{
            short * sar = (short *) car ;
            short * qar = (short *) tar ;
            short   thr = (short) thresh ;

            for( lp=0 ; lp < num_lp ; lp++ )
               fim_thr[lp] = scale_factor * pbar->pval[lp+1] / cfac ;

            for( ii=0 ; ii < nvox ; ii++ ){
               if( (qar[ii] > -thr && qar[ii] < thr) || sar[ii] == 0.0 ){
                  ovar[ii] = 0 ;
               } else {
                  for( lp=0 ; lp < num_lp && sar[ii] < fim_thr[lp] ; lp++ ) ; /*nada*/
                  ovar[ii] = fim_ovc[lp] ;
               }
            }
         }
         break ;

         case MRI_float:{
            float * sar = (float *) car ;
            float * qar = (float *) tar ;
            float   thr = (float) thresh ;

            for( lp=0 ; lp < num_lp ; lp++ )
               fim_thr[lp] = scale_factor * pbar->pval[lp+1] / cfac ;

            for( ii=0 ; ii < nvox ; ii++ ){
               if( (qar[ii] > -thr && qar[ii] < thr) || sar[ii] == 0 ){
                  ovar[ii] = 0 ;
               } else {
                  for( lp=0 ; lp < num_lp && sar[ii] < fim_thr[lp] ; lp++ ) ; /*nada*/
                  ovar[ii] = fim_ovc[lp] ;
               }
            }
         }
         break ;

         case MRI_byte:{
            byte * sar = (byte *) car ;
            byte * qar = (byte *) tar ;
            byte   thr = (byte) thresh ;

            for( lp=0 ; lp < num_lp ; lp++ )
               if( pbar->pval[lp+1] <= 0.0 )
                  fim_thr[lp] = 0 ;
               else
                  fim_thr[lp] = scale_factor * pbar->pval[lp+1] / cfac ;

            for( ii=0 ; ii < nvox ; ii++ ){
               if( qar[ii] < thr || sar[ii] == 0 ){
                  ovar[ii] = 0 ;
               } else {
                  for( lp=0 ; lp < num_lp && sar[ii] < fim_thr[lp] ; lp++ ) ; /*nada*/
                  ovar[ii] = fim_ovc[lp] ;
               }
            }
         }
         break ;
      }
   }

   /*----- if ordered, remove clusters -----*/

   if( func_lm_view ){
      int nx=ovim->nx , ny=ovim->ny , nz=ovim->nz , ptmin,iclu ;
      float dx = fabs(func_dset->daxes->xxdel) ,
            dy = fabs(func_dset->daxes->yydel) ,
            dz = fabs(func_dset->daxes->zzdel)  ;
      float rmm  = wshdfunc_shd_lval_av->fval ,
            vmul = wshdfunc_shd_mval_av->fval ;
      MCW_cluster_array * clar ;
      MCW_cluster * cl ;

      if( (rmm >= dx || rmm >= dy || rmm >= dz) && vmul > (dx*dy*dz) ){
         ptmin = (int) (vmul / (dx*dy*dz) + 0.99);
         clar  = MCW_find_clusters( nx,ny,nz , dx,dy,dz , MRI_byte,ovar , rmm ) ;
         if( clar != NULL ){
            for( iclu=0 ; iclu < clar->num_clu ; iclu++ ){
               cl = clar->clar[iclu] ;
               if( cl->num_pt >= ptmin )  /* put back into array */
                  MCW_cluster_to_vol( nx,ny,nz , MRI_byte,ovar , cl ) ;
            }
            DESTROY_CLARR(clar) ;
         }
      }
   }  /* end of cluster removal */


   /* Make this func dset current. 1-Sep-02 */
   
   im3d->vinfo->func_num = id - 1;
   AFNI_initialize_view( im3d->fim_now , im3d ) ;

   return ;
}

/*----------------------------------------------------------------------------
   Load color index data from the functional dataset into the local overlay array
   (heavily adapted from AFNI_func_overlay in afni_func.c).
------------------------------------------------------------------------------*/

void DIFF_reload_ftm_anat_dset(Three_D_View * im3d, int id)
{
   char *program_name = "DIFF_reload_ftm_anat_dset";
   MRI_IMAGE * cim , * tim ;
   void      * car , * tar ;
   float      cfac ,  tfac ;
   int ii , nvox , num_lp , lp ;
   byte * ovar ;
   MCW_pbar * pbar = wshdfunc_color_pbar ;
   byte fim_ovc[NPANE_MAX+1] ;
   float fim_thr[NPANE_MAX] , scale_factor , thresh ;
   THD_session * sess ;
   int iss;
   int debug = 0;

   if( ftm_anat_dset == NULL ) return ;  /* error */

   if(debug) printf("%s: DSET_PREFIX(ftm_anat_dset) = %s\n",
		    program_name,DSET_PREFIX(ftm_anat_dset));

   INVALIDATE_OVERLAY ;              /* toss old overlay, if any */

   DSET_load(ftm_anat_dset) ;            /* make sure is in memory */
   

   cim  = DSET_BRICK(ftm_anat_dset,func_color_ival) ; nvox = cim->nvox ; /* color brick */
   car  = DSET_ARRAY(ftm_anat_dset,func_color_ival) ;
   cfac = DSET_BRICK_FACTOR(ftm_anat_dset,func_color_ival) ;
   if( cfac == 0.0 ) cfac = 1.0 ;

   tim  = DSET_BRICK(ftm_anat_dset,func_thresh_ival) ;                   /* thresh brick */
   tar  = DSET_ARRAY(ftm_anat_dset,func_thresh_ival) ;
   tfac = DSET_BRICK_FACTOR(ftm_anat_dset,func_thresh_ival) ;
   if( tfac == 0.0 ) tfac = 1.0 ;

   ovim = mri_new_conforming( cim , MRI_byte ) ;                     /* new overlay */
   ovar = MRI_BYTE_PTR(ovim) ;

   scale_factor = func_range ;                                       /* for map from */
   if( scale_factor == 0.0 || func_use_autorange )                   /* pbar to data */
      scale_factor = func_autorange ;                                /* value range  */

   num_lp = pbar->num_panes ;
   for( lp=0 ; lp < num_lp ; lp++ ) fim_ovc[lp] = pbar->ov_index[lp] ; /* top to bottom */
   fim_ovc[num_lp] = (func_posfunc) ? (0) : (fim_ovc[num_lp-1]) ;      /* off the bottom */

   thresh = func_threshold * func_thresh_top / tfac ;  /* threshold in tar[] scale */

   /*--- Load the overlay image with the color overlay index ---*/

   if( thresh < 1.0 || !func_use_thresh ){  /*--- no thresholding needed ---*/

      switch( cim->kind ){

         case MRI_short:{
            short * sar = (short *) car ;

            for( lp=0 ; lp < num_lp ; lp++ )                          /* pbar in     */
               fim_thr[lp] = scale_factor * pbar->pval[lp+1] / cfac ; /* car[] scale */

            for( ii=0 ; ii < nvox ; ii++ ){
               if( sar[ii] == 0 ){
                  ovar[ii] = 0 ;
               } else {
                  for( lp=0 ; lp < num_lp && sar[ii] < fim_thr[lp] ; lp++ ) ; /*nada*/
                  ovar[ii] = fim_ovc[lp] ;
               }
            }
         }
         break ;

         case MRI_float:{
            float * sar = (float *) car ;

            for( lp=0 ; lp < num_lp ; lp++ )
               fim_thr[lp] = scale_factor * pbar->pval[lp+1] / cfac ;

            for( ii=0 ; ii < nvox ; ii++ ){
               if( sar[ii] == 0.0 ){
                  ovar[ii] = 0 ;
               } else {
                  for( lp=0 ; lp < num_lp && sar[ii] < fim_thr[lp] ; lp++ ) ; /*nada*/
                  ovar[ii] = fim_ovc[lp] ;
               }
            }
         }
         break ;

         case MRI_byte:{
            byte * sar = (byte *) car ;

            for( lp=0 ; lp < num_lp ; lp++ )
               if( pbar->pval[lp+1] <= 0.0 )
                  fim_thr[lp] = 0 ;
               else
                  fim_thr[lp] = scale_factor * pbar->pval[lp+1] / cfac ;

            for( ii=0 ; ii < nvox ; ii++ ){
               if( sar[ii] == 0 ){
                  ovar[ii] = 0 ;
               } else {
                  for( lp=0 ; lp < num_lp && sar[ii] < fim_thr[lp] ; lp++ ) ; /*nada*/
                  ovar[ii] = fim_ovc[lp] ;
               }
            }
         }
         break ;
      }             /*--- end of no thresholding case ---*/

   } else {         /*--- start of thresholding ---*/

      switch( cim->kind ){

         case MRI_short:{
            short * sar = (short *) car ;
            short * qar = (short *) tar ;
            short   thr = (short) thresh ;

            for( lp=0 ; lp < num_lp ; lp++ )
               fim_thr[lp] = scale_factor * pbar->pval[lp+1] / cfac ;

            for( ii=0 ; ii < nvox ; ii++ ){
               if( (qar[ii] > -thr && qar[ii] < thr) || sar[ii] == 0.0 ){
                  ovar[ii] = 0 ;
               } else {
                  for( lp=0 ; lp < num_lp && sar[ii] < fim_thr[lp] ; lp++ ) ; /*nada*/
                  ovar[ii] = fim_ovc[lp] ;
               }
            }
         }
         break ;

         case MRI_float:{
            float * sar = (float *) car ;
            float * qar = (float *) tar ;
            float   thr = (float) thresh ;

            for( lp=0 ; lp < num_lp ; lp++ )
               fim_thr[lp] = scale_factor * pbar->pval[lp+1] / cfac ;

            for( ii=0 ; ii < nvox ; ii++ ){
               if( (qar[ii] > -thr && qar[ii] < thr) || sar[ii] == 0 ){
                  ovar[ii] = 0 ;
               } else {
                  for( lp=0 ; lp < num_lp && sar[ii] < fim_thr[lp] ; lp++ ) ; /*nada*/
                  ovar[ii] = fim_ovc[lp] ;
               }
            }
         }
         break ;

         case MRI_byte:{
            byte * sar = (byte *) car ;
            byte * qar = (byte *) tar ;
            byte   thr = (byte) thresh ;

            for( lp=0 ; lp < num_lp ; lp++ )
               if( pbar->pval[lp+1] <= 0.0 )
                  fim_thr[lp] = 0 ;
               else
                  fim_thr[lp] = scale_factor * pbar->pval[lp+1] / cfac ;

            for( ii=0 ; ii < nvox ; ii++ ){
               if( qar[ii] < thr || sar[ii] == 0 ){
                  ovar[ii] = 0 ;
               } else {
                  for( lp=0 ; lp < num_lp && sar[ii] < fim_thr[lp] ; lp++ ) ; /*nada*/
                  ovar[ii] = fim_ovc[lp] ;
               }
            }
         }
         break ;
      }
   }

   /*----- if ordered, remove clusters -----*/

   if( func_lm_view ){
      int nx=ovim->nx , ny=ovim->ny , nz=ovim->nz , ptmin,iclu ;
      float dx = fabs(ftm_anat_dset->daxes->xxdel) ,
            dy = fabs(ftm_anat_dset->daxes->yydel) ,
            dz = fabs(ftm_anat_dset->daxes->zzdel)  ;
      float rmm  = wshdfunc_shd_lval_av->fval ,
            vmul = wshdfunc_shd_mval_av->fval ;
      MCW_cluster_array * clar ;
      MCW_cluster * cl ;

      if( (rmm >= dx || rmm >= dy || rmm >= dz) && vmul > (dx*dy*dz) ){
         ptmin = (int) (vmul / (dx*dy*dz) + 0.99);
         clar  = MCW_find_clusters( nx,ny,nz , dx,dy,dz , MRI_byte,ovar , rmm ) ;
         if( clar != NULL ){
            for( iclu=0 ; iclu < clar->num_clu ; iclu++ ){
               cl = clar->clar[iclu] ;
               if( cl->num_pt >= ptmin )  /* put back into array */
                  MCW_cluster_to_vol( nx,ny,nz , MRI_byte,ovar , cl ) ;
            }
            DESTROY_CLARR(clar) ;
         }
      }
   }  /* end of cluster removal */


   /* Make this func dset current. 1-Sep-02 */
   
   im3d->vinfo->func_num = id - 1;
   AFNI_initialize_view( im3d->fim_now , im3d ) ;

   return ;
}

/*---------------------------------------------------------------------------
  Callback for opacity scale factor selection
-----------------------------------------------------------------------------*/

void DIFF_opacity_scale_CB( MCW_arrowval * av , XtPointer cd )
{
   if( av->fval < MIN_OPACITY_SCALE ) AV_assign_fval(av,MIN_OPACITY_SCALE) ;
   /* if( cd == NULL && dynamic_flag && diffuse_handle != NULL )
      DIFF_draw_CB(NULL,NULL,NULL) ; */
   return ;
}

/*---------------------------------------------------------------------------
  Callback for diffusion anisotropy computations
-----------------------------------------------------------------------------*/

void DIFF_aniso_compute_chan_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
  char *program_name = "DIFF_aniso_compute_chan_CB";
  int debug = 0;

  aniso_compute_flag = MCW_val_bbox( aniso_compute_bbox ) ;

  current_dif_state->sta->compute_chan = (int)(LOG2(aniso_compute_flag)) ; /* (1,2,4)->(0,1,2) */

  if(debug) {
    printf("%s: cds->compute_chan = %i, aniso_compute_flag = %i\n",
	   program_name,
	   current_dif_state->sta->compute_chan,
	   aniso_compute_flag ) ;
    printf("%s: SWD_COMPUTE = %i\n",program_name,SWD_COMPUTE);
  }

  switch (aniso_compute_flag) {

    case DTI_COMPUTE:{
      dwi2dt_flag = 0;
      dti_flag = 1 ;
      shd_flag = 0 ;
      swd_flag = 0 ;
    }
    break;

    case SHD_COMPUTE:{
      dwi2dt_flag = 0;
      dti_flag = 0 ;
      shd_flag = 1 ;
      swd_flag = 0 ;
    }
    break;

    case SWD_COMPUTE:{
      dwi2dt_flag = 0;
      dti_flag = 0 ;
      shd_flag = 1 ;		/* also needed */
      swd_flag = 1 ;
    }
    break;

  case DWI2DT_COMPUTE:{
    dwi2dt_flag = 1;
    dti_flag = 0;
    shd_flag = 0;
    swd_flag = 0;
  }
    break;
  }

  RC_SENSITIZE( spharm_rowcol , shd_flag ) ; 

}

/*---------------------------------------------------------------------------
  Callback for fiber computations
-----------------------------------------------------------------------------*/

void DIFF_fiber_compute_chan_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
  int debug = 0;

  char *program_name="DIFF_fiber_compute_chan_CB";

#ifdef EXPERIMENTAL
  fiber_compute_flag = MCW_val_bbox( fiber_compute_bbox ) ;
#else
  fiber_compute_flag = 1;
#endif

  if(debug) printf("Inside %s: fiber_compute_flag = %i\n",
		   program_name,fiber_compute_flag);

  current_dif_state->sta->compute_chan = (int)(LOG2(fiber_compute_flag)) ; /* (1,2,4)->(0,1,2) */

  if(debug) 
    printf("%s: cds->compute_chan = %i, fiber_compute_flag = %i\n",
	   program_name,current_dif_state->sta->compute_chan,fiber_compute_flag ) ;

  switch (fiber_compute_flag) {

    case FTM1_COMPUTE:{
      ftm1_flag = 1 ;
      ftm2_flag = 0 ;
    }
    break;

    case FTM2_COMPUTE:{
      ftm1_flag = 0 ;
      ftm2_flag = 1 ;
    }
    break;

  }

#ifdef EXPERIMENTAL
  RC_SENSITIZE( fibtmap_rowcol , ftm2_flag ) ; 
#endif

}

/*---------------------------------------------------------------------------
  Callback for diffusion overlay
-----------------------------------------------------------------------------*/

void DIFF_overlay_chan_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
  int debug = 0;

  if( XtIsManaged(wshdfunc_frame) ) {
    dif_overlay_flag = MCW_val_bbox( dif_overlay_bbox ) ;
    /* (1,2,4)->(0,1,2) */
    current_dif_state->sta->overlay_chan = (int)(LOG2(dif_overlay_flag)) ;
  } else {
    current_dif_state->sta->overlay_chan = DTI_OVERLAY ;
  }
  if(debug) 
    printf("DIFF_overlay_chan_CB: cds->ochan = %i, dif_overlay_flag = %i\n",
	   current_dif_state->sta->overlay_chan,dif_overlay_flag ) ;

  switch (dif_overlay_flag) {

    default:
      break;

    case DTI_OVERLAY:{
      strcpy(diff_choose_str,DIFO_labels[0]);
      if( XtIsManaged(wshdfunc_frame) )
	DIFF_open_dtifunc_CB(w, client_data, call_data ) ;
      printf(" done\n");
    }
    break;

    case SHD_OVERLAY:{
      strcpy(diff_choose_str,DIFO_labels[1]);
      if( XtIsManaged(wshdfunc_frame) )
	DIFF_open_shdfunc_CB(w, client_data, call_data ) ;
    }
    break;

    case FTM_OVERLAY:{
      strcpy(diff_choose_str,DIFO_labels[2]);
      if( XtIsManaged(wshdfunc_frame) )
	DIFF_open_ftmfunc_CB(w, client_data, call_data ) ;
    }
    break;

  }

}

/*---------------------------------------------------------------------------
   Open or close the SHD overlay control panel
-----------------------------------------------------------------------------*/

void DIFF_overlay_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   XmString xstr ;

   if( wshdfunc_frame == NULL ) DIFF_func_shd_widgets() ;  /* need to make them */

   if( XtIsManaged(wshdfunc_frame) ){          /* if open, close */

      XtUnmanageChild(wshdfunc_vsep ) ;
      XtUnmanageChild(wshdfunc_frame) ;
      xstr = XmStringCreateLtoR( "Open" , XmFONTLIST_DEFAULT_TAG ) ;
      XtVaSetValues( dif_overlay_pb , XmNlabelString , xstr , NULL ) ;
      XmStringFree(xstr) ;

   } else {                                 /* if closed, open */

      HIDE_SCALE ;
      XtManageChild(wshdfunc_vsep ) ;
      XtManageChild(wshdfunc_frame) ;
      update_MCW_pbar( wshdfunc_color_pbar ) ; /* may need to be redrawn */
      FIX_SCALE_SIZE ;
      DIFF_init_cmap() ;                    /* setup the colormap */
      xstr = XmStringCreateLtoR( "Close" , XmFONTLIST_DEFAULT_TAG ) ;
      XtVaSetValues( dif_overlay_pb , XmNlabelString , xstr , NULL ) ;
      XmStringFree(xstr) ;

      DIFF_overlay_chan_CB( w, client_data, call_data ) ;
   }

   /* Default L and M to view are those computed */
   AV_assign_ival( wshdfunc_shd_lval_av , shd_lval_av->ival );
   AV_assign_ival( wshdfunc_shd_mval_av , shd_mval_av->ival );

   if(0) {			/* old style */
     MCW_invert_widget(wshdfunc_open_pb) ;       /* a flag */
   }
   return ;
}

/*---------------------------------------------------------------------------
   Open or close the DTI overlay control panel
-----------------------------------------------------------------------------*/

void DIFF_open_dtifunc_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   XmString xstr ;
   int debug = 0;

   if(debug) printf("DIFF_open_dtifunc_CB:\n") ;

   if( wshdfunc_frame == NULL ) DIFF_func_shd_widgets() ;  /* need to make them */

   if( XtIsManaged(wshdfunc_lmview_frame) ){ /* if others open close others */
      XtUnmanageChild(wshdfunc_lmview_frame) ;
   }
   if( XtIsManaged(wshdfunc_scalars_frame) ){ /* if others open close others */
      XtUnmanageChild(wshdfunc_scalars_frame) ;
   }
   if( XtIsManaged(wftmfunc_lmview_frame) ){ /* if others open close others */
      XtUnmanageChild(wftmfunc_lmview_frame) ;
   }

   if( !XtIsManaged(wdtifunc_scalars_frame) ){ /* open, if not already */
      HIDE_SCALE ;
      XtManageChild(wdtifunc_scalars_frame) ;
      update_MCW_pbar( wshdfunc_color_pbar ) ; /* may need to be redrawn */
      FIX_SCALE_SIZE ;
      DIFF_init_cmap() ;                    /* setup the colormap */
   }

   xstr = XmStringCreateLtoR( "Choose DTI Dataset" , XmFONTLIST_DEFAULT_TAG ) ;
   XtVaSetValues( wshdfunc_choose_pb , XmNlabelString , xstr , NULL ) ;
   XmStringFree(xstr) ;

   /* Default L and M to view are those computed */
   /*   AV_assign_ival( wshdfunc_shd_lval_av , shd_lval_av->ival );
	AV_assign_ival( wshdfunc_shd_mval_av , shd_mval_av->ival );  */

   if(0) {			/* old style */
     MCW_invert_widget(wshdfunc_open_pb) ;       /* a flag */
   }
   return ;
}

/*---------------------------------------------------------------------------
   Open or close the SHD overlay control panel
-----------------------------------------------------------------------------*/

void DIFF_open_shdfunc_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   XmString xstr ;

   if( wshdfunc_frame == NULL ) DIFF_func_shd_widgets() ;  /* need to make them */

   if( XtIsManaged(wdtifunc_scalars_frame) ){ /* if others open close others */
      XtUnmanageChild(wdtifunc_scalars_frame) ;
   }
   if( XtIsManaged(wftmfunc_lmview_frame) ){ /* if others open close others */
      XtUnmanageChild(wftmfunc_lmview_frame) ;
   }

   if( !XtIsManaged(wshdfunc_scalars_frame) ){  /* open, if not already */
      HIDE_SCALE ;
      XtManageChild(wshdfunc_scalars_frame) ;
      update_MCW_pbar( wshdfunc_color_pbar ) ; /* may need to be redrawn */
      FIX_SCALE_SIZE ;
      DIFF_init_cmap() ;                    /* setup the colormap */
   }
   if( !XtIsManaged(wshdfunc_lmview_frame) ){  /* open, if not already */
      XtManageChild(wshdfunc_lmview_frame) ;
   }

   xstr = XmStringCreateLtoR( "Choose SHD Dataset" , XmFONTLIST_DEFAULT_TAG ) ;
   XtVaSetValues( wshdfunc_choose_pb , XmNlabelString , xstr , NULL ) ;
   XmStringFree(xstr) ;

   /* Default L and M to view are those computed */
   AV_assign_ival( wshdfunc_shd_lval_av , shd_lval_av->ival );
   AV_assign_ival( wshdfunc_shd_mval_av , shd_mval_av->ival );

   if(0) {			/* old style */
     MCW_invert_widget(wshdfunc_open_pb) ;       /* a flag */
   }

   return ;
}

/*---------------------------------------------------------------------------
   Open or close the FTM overlay control panel
-----------------------------------------------------------------------------*/

void DIFF_open_ftmfunc_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   XmString xstr ;

   if( wshdfunc_frame == NULL ) DIFF_func_shd_widgets() ;  /* need to make them */

   if( XtIsManaged(wdtifunc_scalars_frame) ){ /* if others open close others */
      XtUnmanageChild(wdtifunc_scalars_frame) ;
   }
   if( XtIsManaged(wshdfunc_lmview_frame) ){ /* if others open close others */
      XtUnmanageChild(wshdfunc_lmview_frame) ;
   }
   if( XtIsManaged(wshdfunc_scalars_frame) ){ /* if others open close others */
      XtUnmanageChild(wshdfunc_scalars_frame) ;
   }

   if( !XtIsManaged(wftmfunc_lmview_frame) ){  /* open, if not already */
      HIDE_SCALE ;
      XtManageChild(wftmfunc_lmview_frame) ;
      update_MCW_pbar( wshdfunc_color_pbar ) ; /* may need to be redrawn */
      FIX_SCALE_SIZE ;
      DIFF_init_cmap() ;                    /* setup the colormap */
   }

   xstr = XmStringCreateLtoR( "Choose FTM Dataset" , XmFONTLIST_DEFAULT_TAG ) ;
   XtVaSetValues( wshdfunc_choose_pb , XmNlabelString , xstr , NULL ) ;
   XmStringFree(xstr) ;

   /* Default L and M to view are those computed */
   AV_assign_ival( wftmfunc_ftm_lval_av , ftm_lval_av->ival );
   AV_assign_ival( wftmfunc_ftm_mval_av , ftm_mval_av->ival );

   if(0) {			/* old style */
     MCW_invert_widget(wftmfunc_open_pb) ;       /* a flag */
   }

   return ;
}

/*--------------------------------------------------------------------------
  Computation of scalar parameter from DTI data
----------------------------------------------------------------------------*/

void DIFF_scalar_av_CB( MCW_arrowval * av , XtPointer cd )
{
   int debug = 0;

   switch ( av->ival ) {

      default:

      case METH_DTI_AVE:{
	if(debug) printf("inside DIFF_scalar_av_CB: ival = %i\n",METH_DTI_AVE);
      }
      break;

      case METH_DTI_FAI:{
	if(debug) printf("inside DIFF_scalar_av_CB: ival = %i\n",METH_DTI_FAI);
      }
      break;

      case METH_DTI_RAI:{
	if(debug) printf("inside DIFF_scalar_av_CB: ival = %i\n",METH_DTI_RAI);
      }
      break;

      case METH_DTI_LAT:{
	if(debug) printf("inside DIFF_scalar_av_CB: ival = %i\n",METH_DTI_LAT);
      }
      break;

      case METH_DTI_SDV:{
	if(debug) printf("inside DIFF_scalar_av_CB: ival = %i\n",METH_DTI_SDV);
      }
      break;

   }
}

DSET_stats *DIFF_dset_stats( THD_3dim_dataset *input_dset , 
			     THD_3dim_dataset *mask_dset  ,
			     int miv, int ivtop, int ivbot ,
			     float mask_bot, float mask_top )
{
   double sum , sigma , vmin, vmax;
   float *sumar=NULL , *sigmar=NULL , *vminar=NULL , *vmaxar=NULL ;
   int   iv , niv, mcount , nvox , ii , sigmait , nvals ;
   int   doall = 1;
   char * tag , * str , buf[64] , abuf[32],sbuf[32] ;
   byte * mmm ;
   DSET_stats *dstat = NULL ;

   char * cname=NULL ;  /* 06 Aug 1998 */
   int    cdisk=0 ;     /* 22 Aug 2000 */

   /* Check input */

   if( !ISVALID_DSET(input_dset) ) return NULL ;

   DSET_load(input_dset) ;	/* Make sure data is loaded */

   nvox = DSET_NVOX(input_dset) ;

   if( mask_dset == NULL ) {	/* no mask found - create unit mask */

     int ityp , nbytes , nvals , ival, ii ;
     void *bp ;
     byte *new_brick ;

     printf("DIFF_dset_stats: Mask input NULL - "
	    "creating unit mask ... ");

     mask_dset = EDIT_empty_copy( input_dset ) ; 
     
     EDIT_dset_items( mask_dset ,
		      ADN_datum_all, MRI_byte,
		      ADN_none ) ;
     nvals = DSET_NVALS(mask_dset) ;

      for( ival=0 ; ival < nvals ; ival++) /* get memory for bricks */
      {				
         ityp      = DSET_BRICK_TYPE(mask_dset,ival) ;
         nbytes    = DSET_BRICK_BYTES(mask_dset,ival) ; /* how much data */
         new_brick = (byte *)malloc( nbytes ) ;
	 for( ii=0 ; ii < nvox ; ii++ ) new_brick[ii] = 1 ;
         EDIT_substitute_brick( mask_dset , ival , MRI_byte , new_brick ) ;
         EDIT_BRICK_FACTOR(mask_dset,ival,0.0) ;        /* brick factor  */
      }

      /* Initialize mask range.
         This is default for unit mask.
         You'll need to do more for input masks */

      mask_bot = 0.0;
      mask_top = 1.0;

      printf("done\n");

   } else {

     mask_dset = PLUTO_copy_dset( input_dset , "tmp" ) ;  

   }

   /*-- build a byte mask array --*/

   mmm = THD_makemask(mask_dset, miv, mask_bot, mask_top ) ;
   mcount = THD_countmask(nvox,mmm);

   if( mcount == 0 ){
      free(mmm) ;
      printf("*** No voxels survive the masking operations! ***") ;
      return NULL ;
   }
   sigmait = (mcount > 1) ;

   /*-- compute statistics --*/

   nvals = DSET_NVALS(input_dset) ;

   if( doall ){
      sumar  = (float *) malloc( sizeof(float) * nvals ) ;
      sigmar = (float *) malloc( sizeof(float) * nvals ) ;
      vminar = (float *) malloc( sizeof(float) * nvals ) ;
      vmaxar = (float *) malloc( sizeof(float) * nvals ) ;
   }

   for( iv=ivbot ; iv <=ivtop ; iv++ ){
      sum = sigma = 0.0 ;                         /* 13 Dec 1999 */
      switch( DSET_BRICK_TYPE(input_dset,iv) ){

         default:
            free(mmm) ; if( doall ){ free(sumar) ; free(sigmar) ; 
	                             free(vminar) ; free(vmaxar) ;}
            printf("*** Can't use source dataset -- illegal data type! ***") ;
            return NULL ;

         case MRI_short:{
            short * bar = (short *) DSET_ARRAY(input_dset,iv) ;
            float  mfac = DSET_BRICK_FACTOR(input_dset,iv) ;
	    short  val ;
            if( mfac == 0.0 ) mfac = 1.0 ;

	    vmin = vmax = bar[0] ;
            for( ii=0 ; ii < nvox ; ii++ ) 
	      if( mmm[ii] ) {
		val = bar[ii];
		sum += val ;
		     if( vmin > val ) vmin = val ;
		else if( vmax < val ) vmax = val ;
	      }
            sum = sum / mcount ;

            if( sigmait ){
               for( ii=0 ; ii < nvox ; ii++ )
                  if( mmm[ii] ) sigma += SQR(bar[ii]-sum) ;
               sigma = mfac * sqrt( sigma/(mcount-1) ) ;
            }
            sum = mfac * sum ;
         }
         break ;

         case MRI_byte:{
            byte * bar = (byte *) DSET_ARRAY(input_dset,iv) ;
            float mfac = DSET_BRICK_FACTOR(input_dset,iv) ;
	    short  val ;
            if( mfac == 0.0 ) mfac = 1.0 ;

            for( ii=0 ; ii < nvox ; ii++ ) 
	      if( mmm[ii] ) {
		val = bar[ii];
		sum += val ;
		     if( vmin > val ) vmin = val ;
		else if( vmax < val ) vmax = val ;
	      }
            sum = sum / mcount ;

            if( sigmait ){
               for( ii=0 ; ii < nvox ; ii++ )
                  if( mmm[ii] ) sigma += SQR(bar[ii]-sum) ;
               sigma = mfac * sqrt( sigma/(mcount-1) ) ;
            }
            sum = mfac * sum ;
         }
         break ;

         case MRI_float:{
            float * bar = (float *) DSET_ARRAY(input_dset,iv) ;
            float mfac = DSET_BRICK_FACTOR(input_dset,iv) ;
            if( mfac == 0.0 ) mfac = 1.0 ;

            for( ii=0 ; ii < nvox ; ii++ ) if( mmm[ii] ) sum += bar[ii] ;
            sum = sum / mcount ;

            if( sigmait ){
               for( ii=0 ; ii < nvox ; ii++ )
                  if( mmm[ii] ) sigma += SQR(bar[ii]-sum) ;
               sigma = mfac * sqrt( sigma/(mcount-1) ) ;
            }
            sum = mfac * sum ;
         }
         break ;
      }

      if( doall ){ sumar[iv] = sum ; sigmar[iv] = sigma ; 
                   vminar[iv] = vmin ; vmaxar[iv] = vmax ;}
   }

   free(mmm) ;

   /* For now, sum over sub-bricks here.
      We're currently not using sumar,sigmar */

   sum = sigma = 0.;
   niv = ivtop - ivbot + 1;
   for( iv=ivbot ; iv <=ivtop ; iv++ ){
       sum += sumar[iv]/niv ;
     sigma += sigmar[iv]/niv ;
   }

   /*-- Create and load struct to hold info --*/

   dstat = (DSET_stats *) malloc (sizeof(DSET_stats));
   if(dstat==NULL) { str = (char *) malloc( 1024 ) ;
     sprintf( str , "DIFF_dset_stats: Problem allocating DSET_stats!\n") ;
     PLUTO_popup_message(plint,str) ; free(str) ; }

   dstat->min    = vmin ;
   dstat->minar  = vminar ;
   dstat->max    = vmax ;
   dstat->maxar  = vmaxar ;
   dstat->mean   = sum ;
   dstat->meanar = sumar ;
   dstat->sigma  = sigma ;
   dstat->sigmar = sigmar ;
   dstat->count  = mcount ;
   /*dstat->thresh = 1.5*sum ;*/	/* This works for human brain  */
   dstat->thresh = .5*sum ;	/* mouse brain */

   /*-- send report --*/

   if( doall ){
      str = (char *) malloc( 1024 + 64*nvals ) ;
      sprintf(str," ****** ROI statistics ****** \n"
                  " Source  = %s [all sub-bricks] \n"
                  " Mask    = %s [%s]" ,
              DSET_FILECODE(input_dset) ,
              DSET_FILECODE(mask_dset)  , DSET_BRICK_LABEL(mask_dset,miv) ) ;
      if( mask_bot <= mask_top ){
         sprintf(buf," [range %g .. %g]" , mask_bot , mask_top ) ;
         strcat(str,buf) ;
      }
      strcat(str," \n") ;
      sprintf(buf," Count   = %d voxels\n",mcount) ; strcat(str,buf) ;
      for( iv=ivbot ; iv <=ivtop ; iv++ ){
         AV_fval_to_char( sumar[iv]  , abuf ) ;
         AV_fval_to_char( sigmar[iv] , sbuf ) ;
         sprintf(buf," Average = %9.9s  Sigma = %9.9s [%s]  \n",
                 abuf,sbuf , DSET_BRICK_LABEL(input_dset,iv) ) ;
         strcat(str,buf) ;
      }
      (void) MCW_popup_message( shell , str ,
			       MCW_USER_KILL | MCW_TIMER_KILL ) ;

      free(str) ; free(sumar) ; free(sigmar) ;

   } else if( mask_bot <= mask_top ){
      str = (char *) malloc( 1024 ) ;
      sprintf( str , " *** ROI Statistics *** \n"
                     " Source  = %s [%s] \n"
                     " Mask    = %s [%s] [range %g .. %g] \n"
                     " Count   = %d voxels \n"
                     " Average = %g \n"
                     " Sigma   = %g " ,
               DSET_FILECODE(input_dset) , DSET_BRICK_LABEL(input_dset,ivbot) ,
               DSET_FILECODE(mask_dset)  , DSET_BRICK_LABEL(mask_dset,miv)    ,
               mask_bot , mask_top , mcount , sum , sigma ) ;
      PLUTO_popup_message(plint,str) ;
      free(str) ;

   } else {
      str = (char *) malloc( 1024 ) ;
      sprintf( str , " *** ROI Statistics *** \n"
                     " Source  = %s [%s] \n"
                     " Mask    = %s [%s] \n"
                     " Count   = %d voxels \n"
                     " Average = %g \n"
                     " Sigma   = %g " ,
               DSET_FILECODE(input_dset) , DSET_BRICK_LABEL(input_dset,ivbot) ,
               DSET_FILECODE(mask_dset)  , DSET_BRICK_LABEL(mask_dset,miv)    ,
               mcount , sum , sigma ) ;
      PLUTO_popup_message(plint,str) ;
      free(str) ;
   }

   return dstat ;

}


/*--------------------------------------------------------------------------
  fiber tracking method: diffusion simulation tractography (DST)
----------------------------------------------------------------------------*/
void DIFF_ftm_dst_compute_CB( Widget wcancel, 
                                  XtPointer client_data, 
				  XtPointer call_data )
{
  send_3dview_to_vtk_vector_window( im3d, dset, current_dif_state );
}

